//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzwangyingqiao@corp.netease.com
//	Author: xqiao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<AugmentedReality.fx>
///	@path	~/System/Shader/
///	@date	2018/03/13
///	@brief	AugmentedReality shader from HEX studio

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

@_UseRGBATexture
{
	Name = UseRGBATexture;
	Description = "Use RGBA Texture";
	Default = false;
};

cbuffer Batch
{
	float4 cDisplayToCamera  = { 1.0f, 1.0f, 0.0f, 0.0f };
	float4 UVTransform		 = { 1.0f, 0.0f, 0.0f, 1.0f };
};

///////////////// samplers //////////////////
texture2D TextureYBuffer : layout(t0);
sampler TextureYBuffer_Sampler : layout(s0)
{
	Texture = TextureYBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D TextureCbCrBuffer : layout(t1);
sampler TextureCbCrBuffer_Sampler : layout(s1)
{
	Texture = TextureCbCrBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D TextureRGBABuffer : layout(t2);
sampler TextureRGBABuffer_Sampler : layout(s2)
{
	Texture = TextureRGBABuffer;
	Filter = Linear;
	Address = Clamp;
};
//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;

	if (!@_UseRGBATexture)
	{
		vPos.y = 1 - vPos.y;
	}

	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.HPosition.z = (EPSILON*5)*OUT.HPosition.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.

	OUT.TC.x = cDisplayToCamera.x * IN.BaseTC.x + cDisplayToCamera.z;
	OUT.TC.y = cDisplayToCamera.y * IN.BaseTC.y + cDisplayToCamera.w;

	if (@_VULKAN)
	{
		float2 t = (OUT.TC - 0.5) * 2.0;
		OUT.TC.x = dot(t, UVTransform.xy);
		OUT.TC.y = dot(t, UVTransform.zw);
		OUT.TC = OUT.TC * 0.5 + 0.5;
	}

	return OUT;
}

float4 ARResolvePS(VertexScreenOutput IN) : SV_Target
{
	if (@_UseRGBATexture)
	{
		float4 outColor = TextureRGBABuffer.Sample(TextureRGBABuffer_Sampler, IN.TC).rgba;
		outColor.w = 0;
		return outColor;
	}
	else
	{
		float bright = TextureYBuffer.Sample(TextureYBuffer_Sampler, IN.TC).r;
		float2 color = TextureCbCrBuffer.Sample(TextureCbCrBuffer_Sampler, IN.TC).rg;
		float4 outColor = float4(bright+1.402*color.y-0.701, bright-0.3441*color.x-0.7141*color.y+0.5291,
			bright+1.772*color.x-0.886, 0);
		return outColor;
	}	
}

technique ARResolve
{
	VertexShader = ScreenVS();
	PixelShader = ARResolvePS();
}

