//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<LocalHdr.fx>
///	@path	~/System/Shader/
///	@date	2014/11/05
///	@brief	Local HDR.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////

///////////////// const buffer //////////////////
#define BrightThreshold		BloomParam.x
#define LocalRate			BloomParam.y
#define BloomHack			BloomParam.z

cbuffer Batch
{
	float4 BloomParam						;
	float4 BloomTint						;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D BaseMapPoint : layout(t0);
sampler BaseMapPoint_Sampler : layout(s0)
{
	Texture = BaseMapPoint;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D bloomMap0 : layout(t1);
sampler bloomMap0_Sampler : layout(s1)
{
	Texture = bloomMap0;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D bloomMap1 : layout(t2);
sampler bloomMap1_Sampler : layout(s2)
{
	Texture = bloomMap1;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D bloomMap2 : layout(t3);
sampler bloomMap2_Sampler : layout(s3)
{
	Texture = bloomMap2;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
///////////////// io signatures //////////////////
struct VSInput 
{
	float4 Position		: POSITION;
	float2 BaseTC		: TEXCOORD0;
};

struct VSOutputBrightPass
{
	float4 HPosition	: SV_Position;
	float4 TC0			: TEXCOORD0;
	float4 TC1			: TEXCOORD1;
};

struct VSOutput
{
	float4 HPosition	: SV_Position;
	float2 TC			: TEXCOORD0;
};


/////////////////// BrightPassFilter ////////////////////
VSOutput ScreenVS( VSInput IN )
{
	VSOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, EPSILON, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

VSOutputBrightPass BrightPassFilterVS(VSInput IN)
{
	VSOutputBrightPass OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, EPSILON, 1.0);

	OUT.TC0.xy = IN.BaseTC + float2(-1.5, -0.5) * ScreenInfo.zw;
	OUT.TC0.zw = IN.BaseTC + float2(0.5, -1.5) * ScreenInfo.zw;
	OUT.TC1.xy = IN.BaseTC + float2( 1.5, 0.5) * ScreenInfo.zw;
	OUT.TC1.zw = IN.BaseTC + float2( -0.5, 1.5) * ScreenInfo.zw;

	return OUT;
}

float4 BrightPassFilterPS(VSOutputBrightPass IN) : SV_Target
{
	float3 sum = float3(0, 0, 0);

#define BrightPassSample(i, tc) \
	float4 S##i = BaseMapPoint.Sample(BaseMapPoint_Sampler, tc); \
	float3 hack##i = S##i.rgb * BloomHack;\
	S##i.rgb = (4 * DecodeBloomFactor(S##i.a)) * S##i.rgb - BrightThreshold; \
	sum += max(S##i.rgb, hack##i); // BrightPassSample

	BrightPassSample(0,  IN.TC0.xy);
	BrightPassSample(1,  IN.TC0.zw);
	BrightPassSample(2,  IN.TC1.xy);
	BrightPassSample(3,  IN.TC1.zw);
#undef BrightPassSample

	return float4(sum / 4.0, 1.0);
}

technique BrightPassFilter
{
	VertexShader = BrightPassFilterVS();
	PixelShader = BrightPassFilterPS();
}

/////////////////// FinalPass ////////////////////
float4 FinalPS(VSOutput IN) : SV_Target
{
	half3 cBloom0 = SampleColorMap(bloomMap0, bloomMap0_Sampler, IN.TC.xy).rgb;
	half3 cBloom1 = SampleColorMap(bloomMap1, bloomMap1_Sampler, IN.TC.xy).rgb;
	half3 cBloom2 = SampleColorMap(bloomMap2, bloomMap2_Sampler, IN.TC.xy).rgb;
	float3 bloom = (cBloom0 + cBloom1 + cBloom2) * BloomTint.rgb;

	return float4(bloom, 1.0);
}

technique FinalPass
{
	VertexShader = ScreenVS();
	PixelShader = FinalPS();
}

