
#pragma shader draw
#include "Constants.fxh"
#include "Common.fxh"
texture2D tTextureIn : layout(t0);
sampler sTextureIn : layout(s0)
{
	Texture = tTextureIn;
	Filter = Point;
	Address = Clamp;
};

RWTexture2D<float4> TextureOut0	: layout(u0);
RWTexture2D<float4> TextureOut1	: layout(u1);
RWTexture2D<float4> TextureOut2	: layout(u2);
RWTexture2D<float4> TextureOut3	: layout(u3);

#define GROUP_SIZE    8

groupshared float4 Samples[GROUP_SIZE * GROUP_SIZE];
cbuffer Shader
{
	float4	cInvSize				;
	int   	cMipNum					;
};
int SignedRightShift(int x, const int bitshift)
{
	if (bitshift > 0)
	{
		return x << (bitshift);
	}
	else if (bitshift < 0)
	{
		return x >> (-bitshift);
	}
	return x;
}
int2 InitialTilePixelPositionForReduction2x2(const int TileSizeLog2, int SharedArrayId)
{
	int x = 0;
	int y = 0;

	[unroll]
	for (int i = 0; i < TileSizeLog2; i++)
	{
		const int DestBitId = TileSizeLog2 - 1 - i;
		const int DestBitMask = 1 << DestBitId;
		x |= DestBitMask & SignedRightShift(SharedArrayId, int(DestBitId) - int(i * 2 + 0));
		y |= DestBitMask & SignedRightShift(SharedArrayId, int(DestBitId) - int(i * 2 + 1));
	}

	return int2(x, y);
}

[numthreads(GROUP_SIZE, GROUP_SIZE, 1)]
void HierarchicalZReduction(int3 GroupID : SV_GroupThreadID, int3 GlobalID : SV_DispatchThreadID)
{
	int2 Idx = GlobalID.xy;
	int groupThreadIndex = GroupID.x*GROUP_SIZE+GroupID.y;
	int2 GroupThreadId = InitialTilePixelPositionForReduction2x2(3, groupThreadIndex);
	int2 DispatchThreadId = GROUP_SIZE * (GlobalID.xy/GROUP_SIZE) + GroupThreadId;
	float2 uv = (DispatchThreadId + float2(0.5, 0.5)) * cInvSize.zw;
	float4 depthSamples;
	if(@_METAL)
	{
		TextureOut0.Store(DispatchThreadId, float4(1,0,0,1));
	}else
	{
		depthSamples.x = DecodeDepth(tTextureIn.SampleLevel(sTextureIn, uv * cInvSize.xy + float2(-0.25f, -0.25f) * cInvSize.zw, 0));
		depthSamples.y = DecodeDepth(tTextureIn.SampleLevel(sTextureIn, uv * cInvSize.xy + float2( 0.25f, -0.25f) * cInvSize.zw, 0));
		depthSamples.z = DecodeDepth(tTextureIn.SampleLevel(sTextureIn, uv * cInvSize.xy + float2(-0.25f,  0.25f) * cInvSize.zw, 0));
		depthSamples.w = DecodeDepth(tTextureIn.SampleLevel(sTextureIn, uv * cInvSize.xy + float2( 0.25f,  0.25f) * cInvSize.zw, 0));	
		
		float furthestZ = min(min(depthSamples.x,depthSamples.y), min(depthSamples.z,depthSamples.w));
		Samples[groupThreadIndex] = furthestZ;
		TextureOut0.Store(DispatchThreadId, furthestZ.xxxx);
		[unroll]
		for (int MipLevel = 1; MipLevel < cMipNum; ++MipLevel)
		{
			const int TileSize = GROUP_SIZE / (1 << MipLevel);
			const int ReduceBankSize = TileSize * TileSize;
			
			// LDS has been written before.
			if (MipLevel == 1)
			{
				GroupMemoryBarrierWithGroupSync();
			}
			float4 parentFurthestDeviceZ;
			parentFurthestDeviceZ[0] = furthestZ;
			if(groupThreadIndex < ReduceBankSize)
			{
				[unroll]
				for (int i = 1; i < 4; i++)
				{
					int index = groupThreadIndex + i * ReduceBankSize;
					parentFurthestDeviceZ[i] = Samples[index];
				}
				furthestZ = min(min(parentFurthestDeviceZ.x,  parentFurthestDeviceZ.y),  min(parentFurthestDeviceZ.z,  parentFurthestDeviceZ.w));
				DispatchThreadId = DispatchThreadId >> 1;
				if(@_METAL)
				{
					if(MipLevel == 1)
					{
						TextureOut1.Store(DispatchThreadId, furthestZ.xxxx);
					}
					else if(MipLevel == 2)
					{
						TextureOut2.Store(DispatchThreadId, furthestZ.xxxx);
					}
					else if(MipLevel == 3)
					{
						TextureOut3.Store(DispatchThreadId, furthestZ.xxxx);
					}
				}else
				{
					if(MipLevel == 1)
					{
						TextureOut1[DispatchThreadId]=furthestZ;
					}
					else if(MipLevel == 2)
					{
						TextureOut2[DispatchThreadId]=furthestZ;
					}
					else if(MipLevel == 3)
					{
						TextureOut3[DispatchThreadId]=furthestZ;
					}
				}
				
				Samples[groupThreadIndex] = furthestZ;
			}
		}
	}	
}


technique HierarchicalZTech
{
	ComputeShader = HierarchicalZReduction();
}

