//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Decal.fx>
///	@path	~/System/Shader/
///	@date	2015/04/08
///	@brief	Decal rendering.

#pragma shader draw
#pragma shader category("Screen")
#define ENABLE_MIXMAP						(@_EnableReflection)
#define ENABLE_SCREEN_SHADOWMAP				(@_ReceiveShadow && @_CSM && @_CastSunShadow && @_LOD_LEVEL <= LOD_Level_2)
#define USE_SHADOWMAP_SAMPLE				(@_LOD_LEVEL <= LOD_Level_2 && @_CastSunShadow)
#include "Constants.fxh"
#include "Common.fxh"


//////////////////////////////////Properties////////////////////////////////////////
@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask texture enabled?";
	AffectedTex = MaskTexture;
	Visible = false;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Visible = false;
	Default = false;
};

@_SlopeTestEnable
{
	Name = _SlopeTestEnable;
	Description = "Is slope test enabled?";
	Visible = false;
}
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = false;
};
@_ReceiveShadow
{
	Name = ReceiveShadow;
	Description = "for WeaponPreview";
	Default = true;
	Visible = false;
};
// constants
cbuffer Batch
{
	float4x3 World 				: World ;
	float4   InstanceColor		: Color ;

	float4 cPointCloud[6]			: PointCloudData;

	float4 cShadowBias				: ShadowBias				; 

};
#define SHADOW_PIXELSIZE (cShadowBias.w)
cbuffer Shader
{
	// Internal parameters
	float4 _UVTransform			 = { 1, 1, 0, 0 };
	float4 _Color				 = { 1, 1, 1, 1 };
	float4 _vWBasisX			;
	float4 _vWBasisY			;
	float4 _vWBasisZ			;
	float4 _XAxis				;
	float4 _YAxis				;
	float4 _ZAxis				;
	float3 _Origin				;

	// Editable parameters
	float  BloomScale				= 0.5;
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
};

// textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D tMixMap : layout(t5) = "Reflection";
sampler sMixSampler : layout(s5)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D DepthBuffer : SceneDepth : layout(t6);
sampler DepthBuffer_Sampler : layout(s6)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D tEnvMap : EnvMap : layout(t7) = "White";
sampler sEnvSampler : layout(s7)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};
texture2D tShadowMap : ShadowMap : layout(t8) = "White";
SamplerComparisonState sShadowMapSampler : layout(s8)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tShadowMask : ShadowMask : layout(t9) = "White";
sampler sShadowMaskSampler : layout(s9)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Decal vertex input.
#include "BRDF.fxh"
#include "Lighting.fxh"
struct VertexAssembleDecal
{
		//EVF_P3F_C4B_N4B_T2F
	float4		Position 		: POSITION;
    
	half4		VertexColor 	: COLOR;
	
	half4		Normal 			: NORMAL;

	float2		TexCoord		: TEXCOORDN;
	
	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;
};

/// Decal output.
struct VertexDecalOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	float3		AxisX			: TEXCOORDN;
	float3		AxisY			: TEXCOORDN;
	float3		AxisZ			: TEXCOORDN;
	half4		VertexColor		: COLOR;				//	Per-vertex color.
	half4		ViewDirection	: TEXCOORDN;
	float3		WorldPosition	: TEXCOORDN;
		[@_CastSunShadow]
	float4 		LitUVZ			: TEXCOORDN; 
};

VertexDecalOutput DecalVS(in VertexAssembleDecal IN)
{
	VertexDecalOutput OUT;

	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	OUT.ViewDirection.xyz =  WorldPosition.xyz-CameraPos.xyz ;
	OUT.ViewDirection.w = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);
	


	OUT.AxisX = -(_XAxis.xyz);

	OUT.AxisY = - (_YAxis.xyz);

	OUT.AxisZ =-(_ZAxis.xyz);

	

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Decal rendering
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = _vWBasisZ.xyz + (_vWBasisX.xyz * screenTC.x) + (_vWBasisY.xyz * screenTC.y);
	return pos * depth;
}
float3 CalcPosDerivative(float2 xPair, float2 yPair, float2 zPair)
{
	float3 deriv =	_vWBasisX.xyz * (xPair.y * zPair.y - xPair.x * zPair.x) +
					_vWBasisY.xyz * (yPair.y * zPair.y - yPair.x * zPair.x) +
					_vWBasisZ.xyz * (zPair.y - zPair.x);
	return deriv;
}

void InitFogLightingSpec(/*in float2 stc,*/in float3 ViewDirection,in float FogT, in out float4 OUT)
{
	//issue : IN.DiffLighting.w is different to VolumetricFog-off
// 	if (@_FOG && @_VolumetricFog)
// 	{
// 		// float2 spos = IN.HPosition.xy;
// 		// [@_D3D9]
// 		// spos += 0.5;
// 		// float2 stc = spos * ScreenInfo.zw;
// 		float depth = IN.DiffLighting.w;
// 		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
// 		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
// 	}
	if (@_FOG && !@_VolumetricFog)
	{
		float4 fogFinalColor = float4(0,0,0,0);
		FogLightingAndScatter(FogT, fogFinalColor.xyz, ViewDirection, SunDirection);
		OUT.rgb = lerp(fogFinalColor.xyz * OUT.a, OUT.rgb, FogT);
	}
}
void GetHexShadow(float3 LitUVZ, int layerIndex, out half InDynamicShadow)
{
	float inShadow;
	float2 pInLitTC = LitUVZ.xy;
	float litZ = LitUVZ.z;
	half2 inRange = abs(pInLitTC - 0.5);
	half ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
	float2 vInvShadowMapWH = float2(SHADOW_PIXELSIZE,SHADOW_PIXELSIZE);
	if(!@_CSM)
	{
		if(@_LOD_LEVEL <= LOD_Level_1)
		{
			hardware_pcf_4x4(pInLitTC, litZ, vInvShadowMapWH, inShadow);
		}else
		{
			hardware_pcf_2x2(pInLitTC, litZ, vInvShadowMapWH, inShadow);
		}
	}
	else if(layerIndex == 0)
	{
		if(@_LOD_LEVEL <= LOD_Level_1)
		{
	#if MATERIAL_SHADINGMODEL_HAIR
			inShadow = hardware_pcf_array_4x4(pInLitTC, litZ, vInvShadowMapWH, 0);
	#else
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 0);
	#endif
		}
		else
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 0);
		}
	}else if(layerIndex == 1)
	{
		float tempShadow;
		if(@_LOD_LEVEL <= SD_SHADOW_HIGHQUALITY_LEVEL)
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 1);
			tempShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
		}
		else
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 1);
			tempShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
		}
		inShadow = max(inShadow, tempShadow);
	}else{
		inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
	}
	inShadow *= saturate((ShadowRate*10));
	InDynamicShadow = inShadow;
}

half GetDynamicShadowForSun(in VertexDecalOutput IN)
{
	half InShadow = 0;
	//get dynamic shadow
	if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		InShadow = tShadowMask.Sample(sShadowMaskSampler, tc).r;
	}
	else if (@_CastSunShadow && (@_REC_SHADOW || @_DEFERRED_SHADING || USE_SHADOWMAP_SAMPLE) && @_LOD_LEVEL <= LOD_Level_2)
	{
		float3 LitUVZ = IN.LitUVZ.xyz / IN.LitUVZ.w;
		LitUVZ.z = min(1.0-10*HALF_EPSILON, LitUVZ.z);
		GetHexShadow(LitUVZ, 1, InShadow);
	}
	InShadow = min(1.0, InShadow);

		InShadow *= SHADOW_RATE;
	return InShadow;
}
#include "BRDF.fxh"
#include "ShadingModel.fxh"
float4 DecalPS(VertexDecalOutput IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;

	float2 stc = spos * ScreenInfo.zw;
	float depth = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc));
	float3 worldPos = CalcHomogeneousPos(depth, stc);
	worldPos += CameraPos.xyz;

	// volume intersection test
	float3 dir = worldPos - _Origin.xyz;
	float3 volumePos = float3(dot(dir, _XAxis.xyz), dot(dir, _YAxis.xyz), dot(dir, _ZAxis.xyz));
	clip(0.5f - abs(volumePos.xyz));

	// normal slope test
	if (@_SlopeTestEnable)
	{
		// try reconstructing normal accuratly from depth buffer
		// 5 taps for each direction: | z | x | * | y | w |
		float4 offset = float4(ScreenInfo.z * 2, ScreenInfo.w * 2, ScreenInfo.z, ScreenInfo.w);
		float4 H;
		H.x = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc - float2(offset.z, 0)));
		H.y = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc + float2(offset.z, 0)));
		H.z = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc - float2(offset.x, 0)));
		H.w = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc + float2(offset.x, 0)));
		float2 he = abs(H.xy * H.zw * rcp(2 * H.zw - H.xy) - depth);
		float2 xPair = IN.HPosition.xx - float2(2, 1);
		if (he.x > he.y)
		{
			xPair = IN.HPosition.xx + float2(1, 2);
			H.zx = H.yw;
		}
		float3 hDeriv = CalcPosDerivative(xPair, IN.HPosition.yy, H.zx);

		float4 V;
		V.x = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc - float2(0, offset.w)));
		V.y = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc + float2(0, offset.w)));
		V.z = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc - float2(0, offset.y)));
		V.w = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc + float2(0, offset.y)));

		float2 ve = abs(V.xy * V.zw * rcp(2 * V.zw - V.xy) - depth);
		float2 yPair = IN.HPosition.yy - float2(2, 1);
		if (ve.x > ve.y)
		{
			yPair = IN.HPosition.yy + float2(1, 2);
			V.zx = V.yw;
		}
		float3 vDeriv = CalcPosDerivative(IN.HPosition.xx, yPair, V.zx);

		// return by normal: cut if slope > 70 degree
		float3 normal = normalize(cross(hDeriv, vDeriv));
		// clip(dot(normal, normalize(_ZAxis.xyz)) - 0.342f);
	}

	float2 uv = volumePos.xy * float2(1, -1) + float2(0.5, 0.5);
	float alphaMask = 1;
	alphaMask = distance((frac(uv) -0.5),0);
	alphaMask = 1-alphaMask;
	

	uv = uv * _UVTransform.xy + _UVTransform.zw;

	float4 OUT = float4(0,0,0,0);

	float2 diffuseTC = uv ;



	float4 colormap =SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);


	OUT.a = colormap.a;
	OUT.rgb =colormap.rgb*colormap.rgb*_Color.rgb;
	OUT.rgb/=PI;
	// OUT *= ModulateColor * InstanceColor*_Color;
	// alphaMask = saturate(ceil(alphaMask*alphaMask*2-(1-_Color.a)*2-0.7))    ;

	// OUT.rgb *= ModulateColor * InstanceColor*_Color;
	

half InShadow = 0;
	if(USE_SHADOWMAP_SAMPLE)
	{
		InShadow = GetDynamicShadowForSun(IN);
#if USE_PLAYER_SHADOWMAP
		if(@_TRANSPARENT && @_CSM)
		{
			float3 LitUVZ2 = IN.LitUVZ2.xyz / IN.LitUVZ2.w;
			LitUVZ2.z = min(1.0-10*HALF_EPSILON, LitUVZ2.z);
			half calcShadow = 0;
			GetHexShadow(LitUVZ2, 0, calcShadow);
			InShadow = max(calcShadow, InShadow);
		}
#endif
	}else if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		float2 shadowAndAO = tShadowMask.Sample(sShadowMaskSampler, tc).rg;
		InShadow = max(shadowAndAO.x, InShadow);
		// GTAO = shadowAndAO.y;
	}

	half3 N = half3(0,0,0);	
	///////sampleMix//////////////////////////
	half Roughness = 1.0;
	half Metallic = 0;
	half AO = 1;
	half3 SpecularColor =lerp(0.05*0.08, OUT.rgb, Metallic);
	if (@_EnableReflection)
	{
	half3 Mix = SampleColorMap(tMixMap, sMixSampler, diffuseTC).xyz;
	Roughness =1-max(0.05,(Mix.r*_Color.a)) ;
	Metallic = Mix.g;
	AO = Mix.b;
	}
	//////sampleNormal/////////////////////
	float3 Normal =normalize(IN.AxisZ);
	if(@_NormalMapEnable)
	{
		float3 Binormal = normalize(IN.AxisX);	
		float3 Tangent = normalize(IN.AxisY) ;	
	


		float3 TangentNormal = tNormalMap.SampleBias(sNormalSampler,diffuseTC, 0.0).xyz;

	TangentNormal.xy = 1-TangentNormal.xy;
	
	TangentNormal =TangentNormal*2-1;
	TangentNormal.xy*=3.0;
	TangentNormal = lerp(half3(0,0,1),TangentNormal, saturate((_Color.a-0.8)*5) );
		//-0.8*5 为了让normal变平发生在渐隐之前
	
	N = TangentNormal.x*Binormal + TangentNormal.y*Tangent+TangentNormal.z*Normal;
	
	}
	else
	{
	N = Normal;	
	

	}
	
	alphaMask = saturate((1+alphaMask-(1-_Color.a)*5))    ;
	N = normalize(N);
	half3 V = normalize(-IN.ViewDirection.xyz);
	
	half NoV = saturate(dot(N,V));
	half NoL = saturate(dot(N,SunDirection.xyz)) ;
	
	half3 GILighting = half3(1,1,1);
	
	DynamicGILight(N, GILighting, InShadow);
	//////Speclighting////
		/////IBL//////
	half3 sampleEnvSpecular = half3(0,0,0);
	
	if(@_EnableReflection)
	{
		half3 IBLSpec = half3(0,0,0);
	half3 R = reflect(-V, N);
		
		half MIP_ROUGHNESS = 0.17;
		half level =Roughness/MIP_ROUGHNESS;
		half fSign = R.z > 0;
		half fSign2 = fSign*2 - 1;
		R.xy /= R.z*fSign2 + 1;
		R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
		
		half4 srcColor;
		srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy,level);
		sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) ;
	
	sampleEnvSpecular = EnvBRDFApprox(SpecularColor ,Roughness,NoV)*sampleEnvSpecular;

	}
	
/////////////GGX//////////////////	
	half3 BRDF =half3(0,0,0);
	if(@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		BRDF  = min(SimpleBRDF_SPEC(SunDirection.xyz,N,V,NoV,NoL,SpecularColor,Roughness)*SunColor,16) ;
	}
	
	OUT.rgb= OUT.rgb*(NoL*SunColor.rgb*(1-InShadow)+GILighting);
	OUT.rgb+=BRDF+sampleEnvSpecular  ;
	OUT.a *= alphaMask ;
	OUT.rgb *= OUT.a;
	InitFogLightingSpec(IN.ViewDirection.xyz,IN.ViewDirection.w,OUT);
	
	// OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);
	// OUT.rgb = alphaMask;
	return OUT;
}

technique Shading
{
	VertexShader = DecalVS();
	PixelShader = DecalPS();
}
