//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Deferred.fx>
///	@path	~/System/Shader/
///	@date	2019/08/29
///	@brief	Deferred rendering.

#pragma shader draw
#define IS_DEFERRED_LIGHTTING 1

#include "Constants.fxh"
#include "Common.fxh"
#include "BRDF.fxh"
#include "SDF.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_NormalQualityPBRLight
{ 
	Name = NormalQualityPBRLight;
	Description = "use simple pbr mode";  
	Default = true;
};
@_HighQualityPBRLight
{
	Name = HighQualityPBRLight;
	Description = "use hight quality pbr mode";
	Default = true;
};
@_EnableSunLighting
{
	Name = EnableSunLighting;
	Description = "use hight quality pbr mode";
	Default = true;	
};
@_ShadowMaskEnable
{ 
	Name = ShadowMaskEnable;
	Description = "Do we have shadow mask?";  
	Visible = true;
};
@_SkinProfileMapEnable
{
	Name = SkinProfileMapEnable;
	Description = "Do we have skin profile map";  
	Visible = true;
};
@_EnableSSS
{
	Name = EnableSSS;
	Description = "SSS Rendering";
	Default = false;
	Visible = true;
};
@_SSSTranslucency
{
	Name = SSSTranslucency;
	Description = "Do translucency lighting";  
	Default = false;
	Visible = true;
};
@_EnableIESLight
{
	Name = EnableIESLight;
	Description = "Do IES lighting";
	Default = false;
	Visible = true;
};
@_IsSphericalLight
{
	Name = IsSphericalLight;
	Description = "Is spherical light";
	Default = false;
	Visible = true;
};
@_TwoSided
{
	Name = TwoSided;
	Description = "Is two sided";
	Default = false;
	Visible = true;
};
@_IsDiskLight
{
	Name = IsDiskLight;
	Description = "Is two sided";
	Default = false;
	Visible = true;
}
@_IsSpotLightVolume
{
	Name = IsSpotLightVolume;
	Description = "Is spot light volume";
	Default = false;
	Visible = true;
};

@_EnableSeparateLighting
{
	Name = EnableSeparateLighting;
	Description = "Enable Separate Lighting";
}；

@_EnableSSGI
{
	Name = EnableSSGI;
	Description = "Enable SSGI";
};

@_EnableVXGI
{
	Name = EnableVXGI;
	Description = "Enable VXGI";
	Default = false;
};
@_EnableGTAO
{
	Name = EnableGTAO;
	Description = "Enable GTAO";
	Default = false;
}
///////////////// const buffer //////////////////
cbuffer Batch
{
	float4x3 World 					;

	float4 cWBasisX		: WBasisX 	;
	float4 cWBasisY		: WBasisY	;
	float4 cWBasisZ		: WBasisZ	;
	float4 cLightData[4]			;
	float3 cRectAxisZ				;
	float  cLightYScaleBelow1		;
	float3 cLightTangent			;
	float  cSelfShadowOffset		;
	float3 cLightBitangent			;
	float  cDebugFlag 				;
	float4x4 cLastViewProjTex		;
};

#define LightPosRange			cLightData[0]
#define LightColorAtten  		cLightData[1]
#define LightDir				cLightData[2].xyz
#define FallOffCosHalfThetaPHi  cLightData[3].xyz

[inputColor(1)]
texture2D tGBufferB : GBufferMapB : layout(t0);
sampler sGBufferB : layout(s0)
{
	Texture = tGBufferB;
	Filter = Point;
	Address = Clamp;
};

[inputColor(2)]
texture2D tGBufferC : GBufferMapC : layout(t1);
sampler sGBufferC : layout(s1)
{
	Texture = tGBufferC;
	Filter = Point;
	Address = Clamp;
};

texture2D tSampleGBufferC : GBufferMapC : layout(t1);
sampler sSampleGBufferC : layout(s1)
{
	Texture = tSampleGBufferC;
	Filter = Point;
	Address = Clamp;
};
[inputColor(3)]
texture2D tGBufferD : GBufferMapD : layout(t2);
sampler sGBufferD : layout(s2)
{
	Texture = tGBufferD;
	Filter = Point;
	Address = Clamp;
};

[inputColor(4)]
texture2D tGBufferE : GBufferMapE : layout(t3);
sampler sGBufferE : layout(s3)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};

[inputColor(5)]
texture2D tGBufferF : GBufferMapF : layout(t4);
sampler sGBufferF : layout(s4)
{
	Texture = tGBufferF;
	Filter = Point;
	Address = Clamp;
};
texture2D tSampleGBufferF : GBufferMapF : layout(t4);
sampler sSampleGBufferF : layout(s4)
{
	Texture = tSampleGBufferF;
	Filter = Point;
	Address = Clamp;
};

[inputColor(6)]
texture2D tDepthMap : SceneDepth : layout(t5);
sampler sDepthSampler : layout(s5)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tEnvMap : EnvMap : layout(t6) = "White";
sampler sEnvSampler : layout(s6)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

// Skin profile map
texture2D tSkinProfileMap : layout(t7) = "White";
sampler sSkinProfileMapSampler : layout(s7)
{
	Texture = tSkinProfileMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

// IES light profile
texture2D tIESTexture : layout(t8) = "White";
sampler sIESSampler : layout(s8)
{
	Texture = tIESTexture;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

// LTC matrix texture
texture2D tLTCMatTexture : layout(t9) = "White";
sampler sLTCMatSampler : layout(s9)
{
	Texture = tLTCMatTexture;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

// LTC GF texture
texture2D tLTCGFTexture : layout(t10) = "White";
sampler sLTCGFSampler : layout(s10)
{
	Texture = tLTCGFTexture;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tGIAO : layout(t8);
sampler sGIAO : layout(s8)
{
	Texture = tGIAO;
	Filter = Linear;
	Address = Clamp;
};

texture2D tPreDepth : layout(t11);
sampler sPreDepth : layout(s11)
{
	Texture = tPreDepth;
	Filter = Point;
	Address = Clamp;
};
texture2D tGTAO : layout(t9);
sampler sGTAO : layout(s9)
{
	Texture = tGTAO;
	Filter = Linear;
	Address = Clamp;
};
////////////////////////////////////Lighting Function////////////////////////////////////
#include "Lighting.fxh"
#include "RectLighting.fxh"
#include "DiskLighting.fxh"
#include "ShadingModel.fxh"
///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/// light volume vertex to vs
struct VertexAssembleLightVolume
{
	float3		Position		: POSITION;
};

// vertex output to pixel shader.
struct VertexOutputLightVolume
{
	float4		HPosition		: SV_Position;
};

///////////////// helper functions //////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}

#define DecodeGBuffer(_GBufferBaseColor, _GBufferNormal, BaseColor, Roughness, N, Metallic, SSS, InShadow)\
{\
	BaseColor = _GBufferBaseColor.rgb;\
	DecodeFloatAndBit(_GBufferBaseColor.a, Roughness, SSS);\
	N = normalize(DecodeNormal(_GBufferNormal.rgb));\
	DecodeFloatAB(_GBufferNormal.a, Metallic, InShadow);\
}

#define DistanceToBox(p, Layer) length(max(max(p-(cMaxXYZ##Layer-cVolumeSize##Layer), (cMinXYZ##Layer+cVolumeSize##Layer)-p), 0))

#define Map(p, Layer, sdf, level)\
{\
	half3 xyzId = WorldPosToXYZId(p, cVolumeSize##Layer, cMinXYZ##Layer);\
	half fc = frac(xyzId.z);\
	xyzId.z -= fc;\
	half2 uv; XYZIdToUV(xyzId, uv);\
	half4 t1 = tSDFLinearMap##Layer.SampleLevel(sSDFLinearSampler##Layer, uv, level);\
	xyzId.z += 1; XYZIdToUV(xyzId, uv);\
	half4 t2 = tSDFLinearMap##Layer.SampleLevel(sSDFLinearSampler##Layer, uv, level);\
	sdf = lerp(t1, t2, fc);\
}

void ComputeTangent(in half3 vNormal, out half3 tan0, out half3 tan1)
{
	tan0 = float3( 1, 0, 0 );
	tan1 = cross( vNormal, tan0 );
	half lg1 = dot( tan1, tan1 );
	if ( lg1 < HALF_EPSILON )
	{
		tan0 = half3( 0, 0, 1 );
		tan1 = cross( vNormal, tan0 );
		tan1 = normalize( tan1 );
	}
	else
		tan1 *= rsqrt(lg1);
	tan0 = normalize( cross(tan1,vNormal) );
}

///////////////// indirect lighting /////////////////
struct LightingOutput
{
	float4 FinalResult : SV_Target0;
	[@_EnableSSS]
	float4 SpecularLighting : SV_Target5;
};
void InitLightingOutput(in out LightingOutput result)
{
	result.FinalResult = float4(0,0,0,0);
	[@_EnableSSS]
	result.SpecularLighting = float4(0,0,0,0);
}

/////////////////// sun lighting ////////////////////
struct DirectAndIndirect
{
	float4 FinalResult : SV_Target0;
	float4 InDirectResult : SV_Target5; // 非sss存储间接光，如果是sss存储specular，蛋疼的带宽
};
void InitDirectAndIndirect(in out DirectAndIndirect result)
{
	result.FinalResult = float4(0,0,0,0);
	result.InDirectResult = float4(0,0,0,0);
}
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

LightingOutput IndirectLightingPS(v2f IN)
{
	LightingOutput OUT;

	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		tGBufferF.Fetch(sGBufferF, stc));

	OUT.FinalResult.rgb = GBuffer.DiffuseColor.rgb;
	OUT.FinalResult.a = 0;
	
	return OUT;
}

technique IndirectLighting
{
	VertexShader = ScreenVS();
	PixelShader = IndirectLightingPS();
}
/////////////////// sun lighting ////////////////////
void OutputLightingHelper(in GBufferData GBuffer, in LightingResult LitResult, out LightingOutput OUT)
{
	//OUTPUT
	OUT.FinalResult = float4(0, 0, 0, 0);
	bool enableSSS = false;
	if (@_EnableSSS)
	{
		OUT.SpecularLighting = float4(0, 0, 0, 0);
		if(GBuffer.ShadingModelID == SHADINGMODELID_SSS)
			enableSSS = true;
	}
	if (enableSSS)
	{
		OUT.FinalResult.rgb = LitResult.DiffuseLighting;
		[@_EnableSSS]//always true
		OUT.SpecularLighting.rgb = LitResult.SpecularLighting;
		EncodeHDR(OUT.FinalResult);
	}
	else
	{
		OUT.FinalResult.rgb = LitResult.DiffuseLighting * GBuffer.DiffuseColor + LitResult.SpecularLighting;
		EncodeHDR(OUT.FinalResult);
	}
}
DirectAndIndirect SunLightingPS(v2f IN)
{
	DirectAndIndirect OUT;
	InitDirectAndIndirect(OUT);

	//1, GBufferData
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	half3 V = normalize( -D );
	float3 WorldPos = CameraPos.xyz + D;
	float4 gbufferF = float4(0, 0, 0, 0);
	if(!@_EnableSeparateLighting)
	{
		gbufferF = tGBufferF.Fetch(sGBufferF, stc);
	}
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		gbufferF);
	half NoV = saturate( dot(GBuffer.WorldNormal, V) );

	//2, lighting 
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	//sun
	{	
		int PBRQualityLevel = PBRQualityLevel_None;
		if (@_HighQualityPBRLight)
			PBRQualityLevel = PBRQualityLevel_High;
		else if(@_NormalQualityPBRLight)
			PBRQualityLevel = PBRQualityLevel_Normal;
		LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);
		IntegrateBxDF(LitResult, GBuffer, LitData);
	}

	LitResult.DiffuseLighting *= ThunderRate;
	LitResult.SpecularLighting *= ThunderRate;

	//GILighting & IBL reflection
	// 开了全场景ssr就不需要这边求ibl了，ssr那边会处理
	if(!@_SSR)
	{
		half3 R = reflect( -V, GBuffer.WorldNormal );
		if(GBuffer.ShadingModelID == SHADINGMODELID_PREINTEGRATED_SKIN)
		{
			GBuffer.CustomData.xyz = normalize(DecodeNormal(GBuffer.CustomData.xyz));
			R = reflect( -V, GBuffer.CustomData.xyz );
		}
		// R = normalize(WorldPos + R * ReflectionPos.w - ReflectionPos.xyz);
		if (@_EnvMapEnable)
		{
			half3 IBLSpec = IBL_Specular(GBuffer, R, NoV);
			if(@_EnableSeparateLighting)
			{
				OUT.InDirectResult.a = 0;
				OUT.InDirectResult.rgb = IBLSpec;
			}else
			{
				LitResult.SpecularLighting += IBLSpec; // 因为没走ssgi和gtao，所以也就无所谓乘ao了
			}
			
		}
	}
	if(!@_EnableSeparateLighting)
	{
		LitResult.SpecularLighting += GBuffer.GILighting; // 在gbufferGen的时候已经乘上DiffuseColor了，然后也一样无所谓乘ao了
	}
	bool enableSSS = false;
	if (@_EnableSSS)
	{
		if(GBuffer.ShadingModelID == SHADINGMODELID_SSS)
			enableSSS = true;
	}
	if (enableSSS)
	{
		OUT.FinalResult.rgb = LitResult.DiffuseLighting;
		OUT.InDirectResult.rgb += LitResult.SpecularLighting; // 这样针对sss的就完美拆分成diffuse和specular了
	}
	else
	{
		OUT.FinalResult.rgb = LitResult.DiffuseLighting * GBuffer.DiffuseColor + LitResult.SpecularLighting;	
	}
	
	EncodeHDR(OUT.FinalResult); // 间接光也需要EncodeHDR
	return OUT;
}

technique SunLighting
{
	VertexShader = ScreenVS();
	PixelShader = SunLightingPS();
}

float4 DebugDrawTechPS(v2f IN) : SV_Target
{
	//1, GBufferData
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		tGBufferF.Fetch(sGBufferF, stc));
	
	float4 OUT = float4(1, 0, 0, 1);
	if ((int)cDebugFlag == 1)								//1-base color
	{
		OUT.rgb = GBuffer.BaseColor.rgb;
	}
	else if((int)cDebugFlag == 2)							//2-Metallic
	{
		OUT.rgb = GBuffer.Metallic.rrr;
	}
	else if ((int)cDebugFlag == 3)							//3-roughness
	{
		OUT.rgb = GBuffer.Roughness.rrr;
	}
	else if ((int)cDebugFlag == 4)							//4-diffuse color
	{
		OUT.rgb = GBuffer.DiffuseColor.rgb;
	}
	else if ((int)cDebugFlag == 5)							//5-specular color
	{
		OUT.rgb = GBuffer.SpecularColor.rgb;
	}
	else if ((int)cDebugFlag == 6)							//6-normal
	{
		OUT.rgb = 0.5+0.5*GBuffer.WorldNormal;
	}
	else
	{
		float2 AB = abs(floor((ScreenInfo.xy*0.5 - spos.xy) / 20));
		OUT.rgb = (AB.x == AB.y)? half3(0,0,0) : half3(1,0,0);
	}
	EncodeHDR(OUT);
	return OUT;
}

technique DebugDrawTech
{
	VertexShader = ScreenVS();
	PixelShader = DebugDrawTechPS();
}

// ////////////////////////////////////LightVolumeStencil////////////////////////////////////
VertexOutputLightVolume LightVolumeVS( VertexAssembleLightVolume IN )
{
	VertexOutputLightVolume OUT;

	// Transform position.
	float4 worldPosition = float4(MulMat3x4(World, float4(IN.Position, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, worldPosition);

	//disable farplane culling
	OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

	return OUT;
}

float4 LightVolumeStencilPS( VertexOutputLightVolume IN ) : SV_Target
{
	return float4(0, 0, 1, 1);
}

technique LightVolumeStencil
{
	VertexShader = LightVolumeVS();
	PixelShader = LightVolumeStencilPS();
}

////////////////////////////////PointLighting////////////////////////////////////////
LightingOutput PointLightingPS( VertexOutputLightVolume IN ) : SV_Target
{
	LightingOutput OUT;
	InitLightingOutput(OUT);
	
	//1, GBufferData
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		tGBufferF.Fetch(sGBufferF, stc));

	//2, Lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	float3 V = normalize( -D );
	int PBRQualityLevel = PBRQualityLevel_None;
	if (@_HighQualityPBRLight)
		PBRQualityLevel = PBRQualityLevel_High;
	else if(@_NormalQualityPBRLight)
		PBRQualityLevel = PBRQualityLevel_Normal;
	half InShadow = 0;
	if (@_ShadowMaskEnable)
		InShadow = GBuffer.InShadow;
	LightingData LitData = GetPointLightData(WorldPos, GBuffer, V, InShadow, PBRQualityLevel, cLightData);
	if (@_IsSphericalLight)
		LitData = GetSphericalLightData(WorldPos, GBuffer, V, InShadow, PBRQualityLevel, cLightData);
	if (@_SSSTranslucency)
		LitData.TranslucencyLighting = true;
	if (@_EnableIESLight)
	{
		float2 profileUV = ComputeLightProfileMultiplier(LitData.L, LightDir, cLightTangent.xyz, cLightBitangent.xyz);
		LitData.Atten *= tIESTexture.Sample(sIESSampler, profileUV).r;
	}
	IntegrateBxDF(LitResult, GBuffer, LitData);
	
	//3, output
	OutputLightingHelper(GBuffer, LitResult, OUT);
	return OUT;
}

technique PointLightingTech
{
	VertexShader = LightVolumeVS();
	PixelShader = PointLightingPS();
}

////////////////////////////////SpotLighting////////////////////////////////////////
LightingOutput SpotLightingPS( VertexOutputLightVolume IN ) : SV_Target
{
	LightingOutput OUT;
	InitLightingOutput(OUT);

	//1, GBufferData
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		tGBufferF.Fetch(sGBufferF, stc));

	//2, Lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	float3 V = normalize( -D );
	int PBRQualityLevel = PBRQualityLevel_None;
	if (@_HighQualityPBRLight)
		PBRQualityLevel = PBRQualityLevel_High;
	else if(@_NormalQualityPBRLight)
		PBRQualityLevel = PBRQualityLevel_Normal;
	half InShadow = 0;
	if (@_ShadowMaskEnable)
		InShadow = GBuffer.InShadow;
	LightingData LitData = GetSpotLightData(WorldPos, GBuffer, V, InShadow, PBRQualityLevel, cLightData);
	if (@_SSSTranslucency)
		LitData.TranslucencyLighting = true;
	if (@_IsSphericalLight)
		LitData.IsSphericalLight = true;
	if (@_EnableIESLight)
	{
		float2 profileUV = ComputeLightProfileMultiplier(LitData.L, LightDir, cLightTangent.xyz, cLightBitangent.xyz);
		LitData.Atten *= tIESTexture.Sample(sIESSampler, profileUV).r;
	}
	IntegrateBxDF(LitResult, GBuffer, LitData);

	//3, output
	OutputLightingHelper(GBuffer, LitResult, OUT);
	return OUT;
}

technique SpotLightingTech
{
	VertexShader = LightVolumeVS();
	PixelShader = SpotLightingPS();
}

////////////////////////////////RectLighting////////////////////////////////////////
LightingOutput RectLightingPS( VertexOutputLightVolume IN )
{
	LightingOutput OUT;
	[@_HLSL]
	OUT = (LightingOutput)0;
	
	//1, GBufferData
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		tGBufferF.Fetch(sGBufferF, stc));

	//2, Lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	float3 V = normalize( -D );
	int PBRQualityLevel = PBRQualityLevel_None;
	if (@_HighQualityPBRLight)
		PBRQualityLevel = PBRQualityLevel_High;
	else if(@_NormalQualityPBRLight)
		PBRQualityLevel = PBRQualityLevel_Normal;
	half InShadow = 0;
	if (@_ShadowMaskEnable)
		InShadow = GBuffer.InShadow;
	bool TwoSided = false;
	if (@_TwoSided)
		TwoSided = true;
	LightingData LitData = GetRectLightingData(WorldPos, GBuffer, V, InShadow, PBRQualityLevel, TwoSided, cLightData, cRectAxisZ);
	if (@_SSSTranslucency)
		LitData.TranslucencyLighting = true;
	if (@_IsDiskLight)		
	{
		LitData.IsRectLight = false;
		LitData.IsDiskLight = true;
	}
	if (@_EnableIESLight)
	{
		float2 profileUV = ComputeLightProfileMultiplier(LitData.L, cRectAxisZ, cLightTangent.xyz, cLightBitangent.xyz);
		LitData.Atten *= tIESTexture.Sample(sIESSampler, profileUV).r;
	}
	IntegrateBxDF(LitResult, GBuffer, LitData);
	
	//3, output
	OutputLightingHelper(GBuffer, LitResult, OUT);

	return OUT;
}

technique RectLightingTech
{
	VertexShader = LightVolumeVS();
	PixelShader = RectLightingPS();
}
LightingOutput MergeLightingPS(v2f IN)
{
	half4 SSGIAndAO = half4(0,0,0,1);
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	if (@_EnableSSGI || @_EnableVXGI)
	{
		SSGIAndAO = tGIAO.Sample(sGIAO, stc);
		SSGIAndAO.rgb *= GIInfo.x;
	}
	LightingOutput OUT;
	OUT.FinalResult = float4(0, 0, 0, 0);
	float3 diffuseColor = float3(0, 0, 0);
	float4 giLighting = tSampleGBufferF.Sample(sSampleGBufferF, stc);
	if (@_EnableSSGI || @_EnableVXGI)
	{
		float4 baseColor = tSampleGBufferC.Sample(sSampleGBufferC, stc);
		diffuseColor = Diffuse_Lambert(baseColor.rgb - baseColor.rgb * giLighting.a);
	}
	if(@_EnableGTAO)
	{
		float ao = tGTAO.Sample(sGTAO, stc).a;
		SSGIAndAO.a = min(SSGIAndAO.a, ao);
	}
	OUT.FinalResult.rgb = giLighting.rgb * SSGIAndAO.a + SSGIAndAO.rgb * diffuseColor;
	EncodeHDR(OUT.FinalResult);
	return OUT;
}

technique MergeLighting
{
	VertexShader = ScreenVS();
	PixelShader = MergeLightingPS();
}
