//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<DeferredDecal.fx>
///	@path	~/System/Shader/
///	@date	2019/08/29
///	@brief	Deferred rendering.

#pragma shader draw
#define IS_DEFERRED_LIGHTTING 1
#define DEFERRED_PRE_AO						1 // Ƿao

#include "Constants.fxh"
#include "DeferredShadingCommon.fxh"
#include "BRDF.fxh"
#include "SDF.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_IsOverlayBaseColor
{
	Name = IsOverlayBaseColor;
	Description = "is this decal overlay base color?";  
	Default = true;
	Predict = true;
};
@_IsOverlayNormal
{
	Name = IsOverlayNormal;
	Description = "is this decal overlay normal?";
	Default = true;
};
@_IsOverlayMetallicSpecularRoughness
{ 
	Name = IsOverlayMetallicSpecularRoughness;
	Description = "is this decal overlay metallic/specular/roughness?";  
	Default = true;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

//r-smoothness;g-metallic;b-specular;
texture2D tMixMap : layout(t1) = "Reflection";
sampler sMixSampler : layout(s1)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tNormalMap : layout(t2) = "NormalMap";
sampler sNormalSampler : layout(s2)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

[inputColor(6)]
texture2D tDepthMap : SceneDepth : layout(t5);
sampler sDepthSampler : layout(s5)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 cWBasisX		: WBasisX 	;
	float4 cWBasisY		: WBasisY	;
	float4 cWBasisZ		: WBasisZ	;
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
#endif
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		_AxisX			: TEXCOORD0;	//u begin
	float4		_AxisY			: TEXCOORD1;	//v begin
	float4		_AxisZ			: TEXCOORD2;	//u len
	float4		_CenterPos		: TEXCOORD3;	//v len
	float4		_Param			: TEXCOORD4;	//fade alpha
};

#include "Lighting.fxh"
struct VertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float3		AxisX			: TEXCOORDN;

	float3		AxisY			: TEXCOORDN;

	float3		AxisZ			: TEXCOORDN;

	float3		CenterPos		: TEXCOORDN;

	float4 		UVBegLen		: TEXCOORDN;

	float 		AxisYLength		: TEXCOORDN;

	float4 		TintColorAndFadeAlpha	: TEXCOORDN;
};

VertexShadingOutput DecalVS(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	float3 WorldPosition = IN._CenterPos.xyz + IN.Position.x * IN._AxisX.xyz + IN.Position.y * IN._AxisY.xyz 
		+ IN.Position.z * IN._AxisZ.xyz;
	OUT.HPosition = MulMat4x4(ViewProj, float4(WorldPosition,1));
	OUT.AxisX = IN._AxisX.xyz; float lenX = length(OUT.AxisX); OUT.AxisX /= lenX * lenX;
	OUT.AxisY = IN._AxisY.xyz; float lenY = length(OUT.AxisY); OUT.AxisY /= lenY * lenY;
	OUT.AxisZ = IN._AxisZ.xyz; float lenZ = length(OUT.AxisZ); OUT.AxisZ /= lenZ * lenZ;
	OUT.CenterPos = IN._CenterPos.xyz;
	OUT.UVBegLen = float4(IN._AxisX.w, IN._AxisY.w, IN._AxisZ.w, IN._CenterPos.w);
	OUT.AxisYLength = lenY;
	OUT.TintColorAndFadeAlpha = IN._Param;
	return OUT;
}
//////////////////////////////////Pixel Output////////////////////////////////////////
struct DecalOutput
{
	float4 GBufferA 	: SV_Target0;   // GBufferB[R8G8B8A8]		: Emission, Alpha
	[@_IsOverlayMetallicSpecularRoughness]
	float4 GBufferB 	: SV_Target1;	// GBufferB[R8G8B8A8]		: Roughness, Metallic, AO, Alpha
	[@_IsOverlayBaseColor]
	float4 GBufferC 	: SV_Target2;	// GBufferC[R8G8B8A8]		: BaseColor.rgb, Alpha
	[@_IsOverlayNormal]
	float4 GBufferE 	: SV_Target4;	// GBufferE[R8G8B8A8]		: N.xyz, Alpha
	[@_IsOverlayBaseColor]
	float4 GBufferF 	: SV_Target5;	// GBufferF[R8G8B8A8]		: InDirection.xyz, Alpha
};
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}
DecalOutput DecalPS(VertexShadingOutput IN)
{
	DecalOutput OUT;

	//1, WorldPos & GeometryNormal
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;

	//2, decal pos
	float3 posDir = WorldPos - IN.CenterPos;
	float3 decalPos = float3(dot(posDir, IN.AxisX), dot(posDir, IN.AxisY), dot(posDir, IN.AxisZ));
	clip(float3(0.5,0.5,0.5)-abs(decalPos));

	//3, slop test, clip if > 70 degree
	//clip(dot(IN.AxisY * IN.AxisYLength, GeometryNormal) - 0.342);

	//4, output 
	float DecalFading = saturate(4 - 8 * abs(decalPos.y));
	float2 UV = (decalPos.xz  + float2(0.5, 0.5)) * IN.UVBegLen.zw + IN.UVBegLen.xy;
	float Alpha = 0;
	float metallic = 0;
	float AO = 1;
	if (@_IsOverlayMetallicSpecularRoughness)
	{
		float4 rawData = tMixMap.Sample(sMixSampler, UV);
		OUT.GBufferB.r = 1 - rawData.r;//Roughness
		OUT.GBufferB.g = rawData.g;//Metallic
		OUT.GBufferB.b = rawData.b;//AO
		OUT.GBufferB.a = 1;
		Alpha = rawData.a;
		metallic = OUT.GBufferB.g;
		AO = OUT.GBufferB.b;
	}
	//normal
	float3 X = normalize(IN.AxisX);
	float3 Y = normalize(IN.AxisY);
	float3 Z = normalize(IN.AxisZ);
	X = cross(Y, Z);
	Z = cross(X, -Y);
	float3 decalNormal = Y;
	if (@_IsOverlayNormal)
	{
		float4 rawData = tNormalMap.Sample(sNormalSampler, UV);
		float3 tangentNormal = rawData.rgb * 2 - 1;
		tangentNormal.z = saturate(tangentNormal.z);
		decalNormal = normalize(tangentNormal.x * X + tangentNormal.y * Z + tangentNormal.z * Y);
		OUT.GBufferE.rgb = EncodeNormal(decalNormal);
		OUT.GBufferE.a = 1;
		Alpha = rawData.a;
	}
	if (@_IsOverlayBaseColor)
	{
		float4 rawData = tBaseMap.Sample(sBaseSampler, UV);
		OUT.GBufferC.rgb = DecodeGamma(rawData);
		rawData.rgb *= IN.TintColorAndFadeAlpha.rgb;
		OUT.GBufferC.a = 1;
		Alpha = rawData.a;
		//GILighting
		half3 GILighting = half3(0,0,0);
		half pointCloudShadow = 0;
		DynamicGILight(decalNormal, GILighting, pointCloudShadow);
		half4 SSGIandAO = half4(0,0,0,1);
		GILighting = GILighting * AO;
		OUT.GBufferF.rgb = GILighting * (OUT.GBufferC.rgb - OUT.GBufferC.rgb * metallic) / PI;
		OUT.GBufferF.a = 1;
	}
	Alpha *= IN.TintColorAndFadeAlpha.a * DecalFading;
	//apply alpha
	OUT.GBufferA = float4(0, 0, 0, Alpha);
	if (@_IsOverlayMetallicSpecularRoughness)
	{
		OUT.GBufferB.rgb *= Alpha;
		OUT.GBufferB.a = Alpha;
	}
	if (@_IsOverlayNormal)
	{
		OUT.GBufferE.rgb *= Alpha;
		OUT.GBufferE.a = Alpha;
	}
	if (@_IsOverlayBaseColor)
	{
		OUT.GBufferC.rgb *= Alpha;
		OUT.GBufferC.a = Alpha;
		OUT.GBufferA.rgb *= Alpha;
		OUT.GBufferA.a = Alpha;
		OUT.GBufferF.rgb *= Alpha;
		OUT.GBufferF.a = Alpha;
	}
	return OUT;
}

technique Decal
{
	VertexShader = DecalVS();
	PixelShader = DecalPS();
}

