//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_star.fx>
///	@path	~/System/Shader/
///	@date	2014/10/27
///	@brief	star.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 cTest		 = {1, 1, 1, 1};
};

///////////////// io signatures //////////////////
struct a2v
{
	float4 Position 	: POSITION;		//in world space
	float2 uv			: TEXCOORD0;	//uv
};

struct v2f
{
	float4 Position : SV_Position;
	float2 uv		: TEXCOORD0;
};

///////////////// general pass //////////////////
v2f DiskShadowVS(a2v IN)
{
	v2f OUT;

	[@_HLSL]
	OUT = (v2f)0;

	//apply size
	OUT.Position = MulMat4x4(ViewProj, float4(IN.Position.xyz,1.0));
	OUT.uv = IN.uv;

	return OUT;
}

float4 DiskShadowPS(v2f IN) : SV_Target
{
	half2 pos = 2.0*IN.uv - cTest.xy;
	half dis = sqrt(dot(pos, pos));
	half alpha = 1 - dis;
	alpha = clamp(alpha, 0, 0.6);
	return float4(0, 0, 0, alpha);
}

///////////////// techniques //////////////////
technique DiskShadow
{
	VertexShader = DiskShadowVS();
	PixelShader = DiskShadowPS();
}
///////////////// eof //////////////////
