///	@file	<Basic_VAT_Soft.fx>
///	@path	~/System/Shader/Effect
///	@date	201912
/// @LastChange 202011
///	@brief	Effect shading model With VAT.

/*
	Basic VAT Soft
	Normal是用于Distortion的
	可以不给Normal贴图，但是不应该删除Normal相关实现
*/

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

#define ENABLE_ANIMATION_LERP				1

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
};

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = true;
};

@_VatNormalMapEnable
{ 
	Name = VatNormalMapEnable;
	Description = "Using VAT Normal Map";  
	AffectedTex = tVatNormalMap;
	Fallback = true;
	Visible = true;
};

@_ScriptControl
{ 
	Name = ScriptControl;
	Description = "if Particle System Control VAT Play time";  
	Fallback = true;
	Visible = true;
};

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Basic
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;
	float4 EffectParam			: Param  ;
	float4 SubUV				: Param2 ;
	float4 cTintColor1			: TintColor1 ;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	float  BloomScale			 = 0.5;
	float2 _Padding				 = { 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 fadeColor			 = { 0, 0, 0, 0 };

	float  cVAT_BoundingBoxMax	 = 0;
	float  cVAT_BoundingBoxMin	 = 0;
	uint   cVAT_NumberOfFrames	 = 24;
	float  cVAT_Speed			 = 24;
	uint   cVAT_SrcTextureWidth  = 128;
	uint   cVAT_SrcTextureHeight = 128;
	uint   cVAT_TextureSize 	 = 128;
 	int    cVAT_LoopTimes		 = 1;
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D ScreenTexture : SceneColor : layout(t2) = "Transparent";
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D MaskMap : layout(t4) = "White";
sampler MaskMap_Sampler : layout(s4)
{
	Texture = MaskMap;
	Filter = Linear;
	Address = FromTexture;
};
// VAT Part
texture2D tVatPosMap : layout(t5) = "Black";
sampler sVatPosSampler : layout(s5)
{
	Texture = tVatPosMap;
	Filter = Point;
	Address = Wrap;
	SRGB = false;
};
texture2D tVatNormalMap : layout(t6) = "Black";
sampler sVatNormalSampler : layout(s6)
{
	Texture = tVatNormalMap;
	Filter = Point;
	Address = Wrap;
	SRGB = false;
};
// VAT Part End

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Lighting.fxh"

//////////////////////////////////Assemble Input////////////////////////////////////////
struct EffectVertexAssemble
{
	float4		Position 		: POSITION;

	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
	[@_INSTANCED]
	float4		EffectColor		: TEXCOORD6;
	[@_INSTANCED]
	float4		EffectParam		: TEXCOORD7;
};

//////////////////////////////////////////////////////////////////////////
// Intermediate results.

/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate and depth for surface.
	half3		WorldNormal		: TEXCOORDN;			//	World Normal
	half3		ViewDirection	: TEXCOORDN;			//	View direction
	float4		VertexColor		: TEXCOORDN;			//	Per-vertex color.	Used in HDR Scale...
	float3		WorldPosition   : TEXCOORDN;
	[@_INSTANCED]
	float4		EffectColor		: TEXCOORDN;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	[@_INSTANCED]
	float4		EffectParam		: TEXCOORDN;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;
};
void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{	
	OUT.HPosition = float4(0,0,0,0);
	OUT.TexCoord = float4(0,0,0,0);
	OUT.WorldNormal = half3(0,0,0);
	OUT.ViewDirection = half3(0,0,0);
	OUT.VertexColor = half4(1,1,1,1);
	OUT.WorldPosition = float3(0,0,0);
 
	[@_INSTANCED]
	OUT.EffectColor = float4(0,0,0,0);
	
	[@_INSTANCED]
	OUT.EffectParam = float4(0,0,0,0);

	[@_TANGENT] 
	OUT.WorldTangent = half4(0,0,0,0);
	[@_TANGENT]
	OUT.WorldBinormal= half4(0,0,0,0);
}

struct EffectShadingParameters
{
    half4 BaseColor;
    half Metallic;
    half Roughness;
    half Specular;
    half3 WorldNormal;
	half Depth;
};

void InitEffectShadingParameters(in out EffectShadingParameters MaterialParamters)
{
    MaterialParamters.BaseColor = half4(1,1,1,1);
    MaterialParamters.Metallic = 0;
    MaterialParamters.Roughness = 1;
    MaterialParamters.Specular = 0.5;
    MaterialParamters.WorldNormal = half3(1,0,0);
	MaterialParamters.Depth = 0;
}

float4 GetEffectColor(in VertexShadingOutput IN)
{
	[@_INSTANCED]
	return IN.EffectColor * EffectColor;

	return EffectColor;
}

float4 GetEffectColorVS(in EffectVertexAssemble IN)
{
	[@_INSTANCED]
	return IN.EffectColor * EffectColor;

	return EffectColor;
}

float4 GetEffectParam(in VertexShadingOutput IN)
{
	if (@_INSTANCED)
		return IN.EffectParam;
	else
		return EffectParam;
}

float4 GetEffectParamVS(in EffectVertexAssemble IN)
{
	if (@_INSTANCED)
		return IN.EffectParam;
	else
		return EffectParam;
}

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.

half3 GetLocalNormal(in EffectVertexAssemble IN)
{
	if(@_LIGHT_MAP && @_VatNormalMapEnable)
	{
		half4 effectParam = GetEffectParamVS(IN);

		half ScaledSizeX =(half )cVAT_SrcTextureWidth / (half)cVAT_TextureSize;
		half ScaledSizeY =(half)cVAT_SrcTextureHeight / (half)cVAT_TextureSize;
			
		half DiscreteTime = 0.0;
		if(@_ScriptControl)
			DiscreteTime = effectParam.x  * cVAT_LoopTimes;
		else
		 	DiscreteTime = frac(GameTime  * (half)cVAT_Speed / cVAT_NumberOfFrames);

		half timeInFrames = ((floor(DiscreteTime * cVAT_NumberOfFrames)) / cVAT_NumberOfFrames);
		half timeInFrames2 = timeInFrames + 1.0 / (half)cVAT_NumberOfFrames;

		half3 textureN1 =  tVatNormalMap.SampleLevel(sVatNormalSampler ,half2(IN.TexCoord2.x * ScaledSizeX,(timeInFrames + IN.TexCoord2.y)* ScaledSizeY),0).xyz; 
		half3 textureN = textureN1;
		
		// Double sample for better effect
		if(ENABLE_ANIMATION_LERP)
		{
			half3 textureN2 =  tVatNormalMap.SampleLevel(sVatNormalSampler ,half2(IN.TexCoord2.x * ScaledSizeX,(timeInFrames2 + IN.TexCoord2.y) * ScaledSizeY),0).xyz; 
			half LerpValue = -1 * (timeInFrames - (DiscreteTime)) / (timeInFrames2 - timeInFrames);
			textureN = lerp(textureN1, textureN2, LerpValue);
		}
		
		//Transform from textureInfo To LocalSpace
		textureN *= 2;
		textureN -= 1;
		textureN.xy *= -1;
		textureN.xyz = textureN.xzy;

		return textureN.xyz;
	}
	else
	{
		//Fallback Output Src State Normal
		float3 LocalNormal = (float3)Expand((half3)IN.Normal.xyz);
		return normalize(MulMat3x4((float3x3)World,LocalNormal).xyz);
	}
}


half3 GetLocalPositionOffset(in EffectVertexAssemble IN)
{	
	if(@_LIGHT_MAP)
	{
		half4 effectParam = GetEffectParamVS(IN);
		half ScaledSizeX =	(half)cVAT_SrcTextureWidth / (half)cVAT_TextureSize;
		half ScaledSizeY =	(half)cVAT_SrcTextureHeight / (half)cVAT_TextureSize;

		half DiscreteTime = 0.0;
		if(@_ScriptControl)
			DiscreteTime = frac(effectParam.x * cVAT_LoopTimes);
		else
		 	DiscreteTime = frac(GameTime  * (half)cVAT_Speed / cVAT_NumberOfFrames);

		half timeInFrames = ((floor(DiscreteTime * (cVAT_NumberOfFrames - 1.0))) / cVAT_NumberOfFrames);
		half timeInFrames2 = timeInFrames + 1.0 / (half)cVAT_NumberOfFrames;
		
		half4 texturePos1 = tVatPosMap.SampleLevel(sVatPosSampler ,half2(IN.TexCoord2.x * ScaledSizeX,(timeInFrames + IN.TexCoord2.y)*ScaledSizeY),0); // Work But Scale
		half4 texturePos = texturePos1;

		if(ENABLE_ANIMATION_LERP)
		{
			half4 texturePos2 = tVatPosMap.SampleLevel(sVatPosSampler ,half2(IN.TexCoord2.x * ScaledSizeX,(timeInFrames2 + IN.TexCoord2.y)*ScaledSizeY),0); 
			half LerpValue = -1 * (timeInFrames - (DiscreteTime)) / (timeInFrames2 - timeInFrames);
			texturePos = lerp(texturePos1, texturePos2, LerpValue);
		}
		
		//Transform To LocalSpace
		half expand = cVAT_BoundingBoxMax / 1000 - cVAT_BoundingBoxMin / 1000;
		texturePos.xyz  *= (expand * 1);
		texturePos.xyz += cVAT_BoundingBoxMin / 1000;
		texturePos.xy *= -1;
		texturePos.xyz = texturePos.xzy;

		return texturePos.xyz;
	}
	else
	{
		//VAT Asset Must Have UV2(@_LIGHT_MAP)
		return float3(0,0,0);
	}
}
VertexShadingOutput EffectVertexShading(in EffectVertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);

	float4 LocalPosition = IN.Position;
	float3 LocalNormal = (float3)Expand((half3)IN.Normal.xyz);

	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, LocalPosition.xyz);
			if(@_VatNormalMapEnable)
				LocalNormal = Matrix4x3TransformN(Mx, My, Mz, LocalNormal);
		}
	}


	float3 localPositionOffset = GetLocalPositionOffset(IN);
	// LocalNormal = GetLocalNormal(IN);
	if (@_INSTANCED)
	{
		OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, LocalPosition.xyz + localPositionOffset.xyz ),1);
		OUT.WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, LocalNormal));
		OUT.TexCoord.xy = IN.TexCoord;
		OUT.EffectParam = IN.EffectParam;
		OUT.EffectColor = IN.EffectColor;
	}
	else
	{
		OUT.WorldPosition = float4(MulMat3x4(World, float4(LocalPosition.xyz + localPositionOffset.xyz, 1)).xyz, 1);
		OUT.WorldNormal = normalize(MulMat3x4((float3x3)World,LocalNormal).xyz); //May get Wrong Normal with (float3x3)World
		OUT.TexCoord.xy = IN.TexCoord * SubUV.xy + SubUV.zw;
	}

	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz,1));
	OUT.HPosition.z -= DepthBias / (DepthBias + OUT.HPosition.w);

	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = FogTransmittance(OUT.WorldPosition.xyz - CameraPos.xyz);

	OUT.ViewDirection = normalize(CameraPos.xyz - OUT.WorldPosition.xyz);

	if (@_VERTEXCOLOR)
		OUT.VertexColor = GetVertexColor(IN.Color);

	[@_HDR]
	OUT.VertexColor.rgb *= HDRScale;

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Utilities
void InitFogLightingEffect(/*in float2 stc,*/in VertexShadingOutput IN, in out float4 OUT)
{
	//issue : IN.DiffLighting.w is different to VolumetricFog-off
// 	if (@_FOG && @_VolumetricFog)
// 	{
// 		// float2 spos = IN.HPosition.xy;
// 		// [@_D3D9]
// 		// spos += 0.5;
// 		// float2 stc = spos * ScreenInfo.zw;
// 		float depth = IN.DiffLighting.w;
// 		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
// 		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
// 	}
	if (@_FOG && !@_VolumetricFog)
	{
		float4 fogFinalColor = float4(0,0,0,0);
		FogLightingAndScatter(IN.TexCoord.w, fogFinalColor.xyz, IN.ViewDirection, SunDirection);
		OUT.rgb = lerp(fogFinalColor.xyz * OUT.a, OUT.rgb, IN.TexCoord.w);
	}
}

void SoftParticle(in VertexShadingOutput IN,in out EffectShadingParameters materialParamters, float4 screenCoord )
{
	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

	float depth = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc));
	depth *= CameraInfo.y;

	float fade = saturate((depth - IN.TexCoord.z) * InvFadeDistance);

	materialParamters.BaseColor.rgb = lerp(materialParamters.BaseColor.rgb,fadeColor.rgb,1-fade.xxx);
	materialParamters.BaseColor *= fade;
}

//////////////////////////////////////////////////////////////////////////
// technique Clip.
void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT = float4(1,0,1,1);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;

	float Mask = SampleColorMap(MaskMap, MaskMap_Sampler, IN.TexCoord.xy).r;
	float4 effectColor = GetEffectColor(IN);	
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	
	OUT *= ModulateColor * effectColor * IN.VertexColor * Mask;
	materialParamters.BaseColor = OUT;
}

float4 EffectClipShading(in VertexShadingOutput IN) : SV_Target
{
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectClipShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 EffectDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.05, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectDebugShading();
}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
