//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Breaker.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

#pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Breaker
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 TwistUVScolling0		 = { 1, 1, 0, 0 };
	float4 TwistUVScolling1		 = { 1, 1, 0, 0 };
	float4 MaskUVTransform		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float  BloomScale			 = 0.5;
	float  _Padding				 = 0;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cFov					 = 1.0;
};

texture2D TwistTexture : layout(t1) = "White";
sampler TwistTexture_Sampler : layout(s1)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D MaskTexture : layout(t2) = "White";
sampler MaskTexture_Sampler : layout(s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Breaker.

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT;
	float4 effectParam = GetEffectParam(IN);
	float2 twistTC0 = IN.TexCoord.xy * TwistUVScolling0.xy + effectParam.y * TwistUVScolling0.zw;
	float2 twist0 = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC0).xy;
	float2 twistTC1 = IN.TexCoord.xy * TwistUVScolling1.xy + effectParam.y * TwistUVScolling1.zw;
	float2 twist1 = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC1).xy;
	float2 twist = (twist0 + twist1) * TwistStrength;

	float2 maskTC = IN.TexCoord.xy * MaskUVTransform.xy + MaskUVTransform.zw;
	maskTC = saturate(maskTC + twist * effectParam.w);
	float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;

	OUT = ModulateColor * GetEffectColor(IN) * IN.VertexColor;
	OUT.a *= mask;

	materialParamters.BaseColor = OUT;

}
#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
