//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Basic.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

// #define FORCE_DISABLE_SOFTPARTICLE 1
// #define FORCE_DISABLE_DISTORTION 1

#pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_MODULATED
{
	Name = Modulated;
	Description = "Distortion effect";
	Visible = true;
}

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask texture enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Basic
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 Layer0Color			 = { 1, 1, 1, 1 };
	float4 Layer0UVScolling		 = { 1, 1, 0, 0 };
	float4 Layer1Color			 = { 1, 1, 1, 1 };
	float4 Layer1UVScolling		 = { 1, 1, 0, 0 };
	float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float  BloomScale			 = 0.5;
	float3 _Padding				 = { 0, 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D Layer0Texture : layout(t0) = "White";
sampler Layer0Texture_Sampler : layout(s0, s1)
{
	Texture = Layer0Texture;
	Filter = Linear;
	Address = Wrap;
};

texture2D Layer1Texture : layout(t2) = "White";
sampler Layer1Texture_Sampler : layout(s2, s3)
{
	Texture = Layer1Texture;
	Filter = Linear;
	Address = Wrap;
};

texture2D MaskTexture : layout(t4) = "White";
sampler MaskTexture_Sampler : layout(s4)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D DepthBuffer : SceneDepth : layout(t5);
sampler DepthBuffer_Sampler : layout(s5)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"
//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Basic

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT;
	float4 effectParam = GetEffectParam(IN);
	
	float2 TC0 = IN.TexCoord.xy * Layer0UVScolling.xy + effectParam.y * Layer0UVScolling.zw;
	float2 TC1 = IN.TexCoord.xy * Layer1UVScolling.xy + effectParam.y * Layer1UVScolling.zw;

	float4 Color0 = SampleColorAlphaMap(Layer0Texture, Layer0Texture_Sampler, TC0);
	float4 Color1 = SampleColorAlphaMap(Layer1Texture, Layer1Texture_Sampler, TC1);

	Color0 *= Layer0Color;
	Color1 *= Layer1Color;

	if (@_MODULATED)
	{
		OUT = Color0 * Color1;
	}
	else
	{
		OUT = lerp(Color0, Color1, saturate(Color1.a));
	}

	if (@_MaskEnable)
	{
		float2 maskUV = IN.TexCoord.xy * MaskUVScolling.xy + effectParam.y * MaskUVScolling.zw;
		float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskUV).r;
		OUT.a *= mask;
	}

	OUT *= GetEffectColor(IN) * IN.VertexColor;

	materialParamters.BaseColor = OUT;

}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
