//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Panner.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

#pragma shader override("@_SOFT", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_TwistEnable
{
	Name = TwistEnable;
	Description = "Is twisting enabled?";
	AffectedTex = TwistTexture;
	Visible = true;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
};

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Panner
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float2 DiffuseUVOffset		 = { 0, 0 };
	float4 MaskUVTransform		 = { 1, 1, 0, 0 };
	float  BloomScale			 = 0.5;
	float3 _Padding				 = { 0, 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D MaskTexture : layout(t2) = "White";
sampler MaskTexture_Sampler : layout(s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D TwistTexture : layout(t3) = "White";
sampler TwistTexture_Sampler : layout(s3)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D ScreenTexture : SceneColor : layout(t4) = "Transparent";
sampler ScreenTexture_Sampler : layout(s4)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t5);
sampler DepthBuffer_Sampler : layout(s5)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"
//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Panner.

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT;
	float4 effectParam = GetEffectParam(IN);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + effectParam.w * DiffuseUVScolling.zw;
	diffuseTC += DiffuseUVOffset.xy;
	if (@_TwistEnable)
	{
		float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + effectParam.y * TwistUVScolling.zw;
		float2 twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;
		diffuseTC += twist * TwistStrength;
	}

	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	OUT *= ModulateColor * GetEffectColor(IN) * IN.VertexColor;

	float2 maskTC = IN.TexCoord.xy * MaskUVTransform.xy + MaskUVTransform.zw;
	float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;
	OUT.a *= mask;

	materialParamters.BaseColor = OUT;
}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
