//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Stretch.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

//#pragma shader override("@_SOFT", "false")
//#pragma shader override("@_DISTORTION", "false")

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
}

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = true;
}


//////////////////////////////////////////////////////////////////////////
// Effect shading model - Stretch
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float2 StretchCenter		 = { 0.5, 0.5 };
	float  Contrast				 = 1.0;
	float  BloomScale			  = 0.5;
	float  FogStrength			  = 0;
	float  DepthBias			  = 0;
	float  HDRScale				  = 2.0;
	float  InvFadeDistance		 = 1;
	float4 fadeColor			 = { 0, 0, 0, 0 };
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D ScreenTexture : SceneColor : layout(t2) = "Transparent";
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D MaskTexture : layout(t4) = "White";
sampler MaskTexture_Sampler : layout(s4)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Wrap;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"
//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Stretch.

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT;
	float4 effectParam = GetEffectParam(IN);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + effectParam.y * DiffuseUVScolling.zw;
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	OUT *= ModulateColor * GetEffectColor(IN) * IN.VertexColor;
	OUT.a = pow(OUT.a, Contrast);

	float stretch = effectParam.w;
	float2 maskTC = (IN.TexCoord.xy - StretchCenter) * stretch + StretchCenter;

	float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, saturate(maskTC)).r;
	OUT.a *= mask;

	materialParamters.BaseColor = OUT;

}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
