//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<CubeMap.fx>
///	@path	~/System/Shader/
///	@date	2015/08/05
///	@brief	Cube map rendering.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader draw

@_ReversedZ
{
	Name = ReversedZ;
	Default = false;
};

@_HAS_CUBEMAP
{
	Name = HasCubemap;
	AffectedTex = cubeMap;
	Description = "Do we has cubemap";
	Visible = true;
};

cbuffer Batch
{
	float4	CubeMapParam	 = { 1, 1, 1, 1 };
	//world
	float4x3 World			: World ;
};

texture2D cubeMap : layout(t0) = "White";
sampler cubeMapSampler : layout(s0)
{
	Texture = cubeMap;
	Filter = Linear;
	Address = Wrap;
};


///////////////// io signatures //////////////////
struct a2v
{
	float4 Position : POSITION; //in object space
	float2 baseTC   : TEXCOORD0;
};

struct v2f_hdr
{
	float4 Position	: SV_Position;  //in projection space
	float2 mapTC	: TEXCOORD0;
};


///////////////// general pass //////////////////

v2f_hdr CubeMapVS(a2v IN)
{
	v2f_hdr OUT;

	[@_HLSL]
	OUT = (v2f_hdr)0;

	// Position in screen space.
	float4 vPos = IN.Position;
	float4 worldPos = float4(MulMat3x4(World, vPos).xyz, 1);
	OUT.Position = MulMat4x4(ViewProj, worldPos);
	[@_ReversedZ]
	OUT.Position.z = (EPSILON*5)*OUT.Position.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.
	[!@_ReversedZ]
	OUT.Position.z = (1.0-EPSILON*5)*OUT.Position.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.

	OUT.mapTC = IN.baseTC;

	return OUT;
}

float4 CubeMapPS(v2f_hdr IN) : SV_Target
{
	float4 result = float4(0.2, 0.2, 0.2, 0);

	if (@_HAS_CUBEMAP)
	{
		result.rgb = SampleColorMap(cubeMap, cubeMapSampler, IN.mapTC).rgb * CubeMapParam.x;
		if (@_HDR)
		{
			result.rgb = DecodeGamma(result.rgb);
		}
	}

	return result;
}


///////////////// techniques //////////////////

technique CubeMapTech
{
	VertexShader = CubeMapVS();
	PixelShader = CubeMapPS();
}

///////////////// eof //////////////////
