#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

// ///////////////// const buffer //////////////////
cbuffer Shader
{
	float2		cPosMove				:	layout(c0.xy);
	float		cDissipation			:   layout(c0.z) = 0.999;
	float 		cRenderTime				:	layout(c0.w) = 0.016;
	float4		cDensityUV				:	layout(c1);
	float4		cAffectNum				:	layout(c2) = 1;
	//x,y:center; z:radius; w:intensity
	float4		cAffectParam[10]		:	layout(c3);
	
	float2		cAffectNumImpulse		:	layout(c13) = 1;
	//x,y:center; z:radius; w:intensity
	float4		cAffectParamImpulse[10]		:	layout(c14);
};

texture2D tDensityMap : layout(t0);
sampler sDensitySampler : layout(s0)
{
	Texture = tDensityMap;
	Filter = Point;
	Address = Clamp;
 	SRGB = false;
};

texture2D tVelocityMap : layout(t1);
sampler sVelocitySampler : layout(s1)
{
	Texture = tVelocityMap;
	Filter = Point;
	Address = Clamp;
 	SRGB = false;
};

texture2D tPressureMap : layout(t2);
sampler sPressureSampler : layout(s2)
{
	Texture = tPressureMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tDivergenceMap : layout(t3);
sampler sDivergenceSampler : layout(s3)
{
	Texture = tDivergenceMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tBurnTimeMap : layout(t4);
sampler sBurnTimeSampler : layout(s4)
{
	Texture = tBurnTimeMap;
	Filter = Linear;
	Address = Clamp;
};

@_IsUpdateBurn
{
	Name = IsUpdateBurn;
	Description = "is burn updating enabled on advect pass"; 
	Default = false;
	Visible = true;
};
@_IsSyncPos
{
	Name = IsSyncPos;
	Description = "is syncPlayer Position"; 
	Default = false;
	Visible = true;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 Position : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////// general pass //////////////////
v2f FoliageVS(a2v IN)
{
	v2f OUT;
	OUT.Position = float4(IN.Position.xy*2-1, IN.Position.z, 1.0);
	OUT.TC = IN.BaseTC;
	OUT.TC.y = 1.0 - OUT.TC.y;
	return OUT;
}

float4 ImpulsePS(v2f IN) : SV_Target
{
	float4 velocity = tVelocityMap.Sample(sVelocitySampler, IN.TC);
	float2 res = float2(0, 0);
	float atten = 0;
	float burnRes = velocity.z;
	//calc line force
	for (int i = 0; i < cAffectNum.x; i++)
	{
		float2 center = cAffectParam[i].xy;
		float2 dir = float2(sin(cAffectParam[i].z), cos(cAffectParam[i].z));
		float radius = 0.01;
		float attenNew = saturate((radius - length(IN.TC - center))/radius * cAffectParam[i].w);
		if (atten < attenNew)
		{
			atten = attenNew;
			res = dir;//float2(cAffectParam[i].z, cAffectParam[i].w);
		}
	}
	//calc circle force
	for (int j = int(cAffectNum.x); j < cAffectNum.y; j++)
	{
		float2 center = cAffectParam[j].xy;
		float radius = cAffectParam[j].z;
		float intensity = cAffectParam[j].w;
		float2 dir = IN.TC - center;
		float dist = length(dir);
 		float attenNew = saturate((radius - dist) / radius * intensity);
		if (radius - 0.01 > dist)
			attenNew = 0;
		if (atten < attenNew)
		{
			atten = attenNew;
			res = dir;
		}
	}
	//burn
	//for (int l = int(cAffectNum.z); l < cAffectNum.w; l++)
	//{
		//float2 center = cAffectParam[l].xy;
		//float radius = cAffectParam[l].z;
		//float intensity = cAffectParam[l].w;
		//float2 dir = IN.TC - center;
		//float dist = length(dir);
		//float attenNew = saturate((radius - dist) / radius  * intensity);
		//burnRes = max(attenNew, burnRes);
		//if (dist < radius)
		//{
		//	burnRes = 1.0;
		//	velocity.w = 10.0;
		//}
	//}
	return float4(lerp(velocity.xy, res, atten), burnRes, velocity.w);
}

float4 AdvectPS(v2f IN)	: SV_Target
{			
	const float BurnSpeed = 0.5;
	float4 oldV = float4(0,0,0,0);
	
	float2 readTC = IN.TC;
	float readMask = 1.0;
	if (@_IsSyncPos)
	{
		readTC = readTC + cPosMove;
		float2 mask2 = step(abs(readTC - 0.5), float2(0.5,0.5));
		readMask = mask2.x * mask2.y;
	}
	float2 currTC = IN.TC;
		
	oldV = tVelocityMap.Sample(sVelocitySampler, readTC) * readMask;
	float2 coord = readTC - oldV.xy / 256.0;//(* the speed of advect)
	float4 newV = tVelocityMap.Sample(sVelocitySampler, coord);
	
	//impulse pass
	{
		float4 velocity = newV;
		float2 res = float2(0, 0);
		float atten = 0;
		float burnRes = velocity.z;
		//calc line force
		for (int i = 0; i < cAffectNumImpulse.x; i++)
		{
			float2 center = cAffectParamImpulse[i].xy;
			float2 dir = float2(sin(cAffectParamImpulse[i].z), cos(cAffectParamImpulse[i].z));
			float radius = 0.01;
			float attenNew = saturate((radius - length(currTC - center))/radius * cAffectParamImpulse[i].w);
			if (atten < attenNew)
			{
				atten = attenNew;
				res = dir;//float2(cAffectParamImpulse[i].z, cAffectParamImpulse[i].w);
			}
		}
		//calc circle force
		for (int j = int(cAffectNumImpulse.x); j < cAffectNumImpulse.y; j++)
		{
			float2 center = cAffectParamImpulse[j].xy;
			float radius = cAffectParamImpulse[j].z;
			float intensity = cAffectParamImpulse[j].w;
			float2 dir = currTC - center;
			float dist = length(dir);
			float attenNew = saturate((radius - dist) / radius * intensity);
			if (radius - 0.01 > dist)
				attenNew = 0;
			if (atten < attenNew)
			{
				atten = attenNew;
				res = dir;
			}
		}
		newV = float4(lerp(velocity.xy, res, atten), burnRes, velocity.w);
	}
	
	float2 resVel = cDissipation * newV.xy;
	//for burn diffusion
	float inverseSize = 1.0/256.0;
	if(!@_IsUpdateBurn)
	{
		return float4(resVel.xy, oldV.z, oldV.w);
	} 
	else
	{
		float4 vN = tVelocityMap.Sample(sVelocitySampler, readTC + float2(0, inverseSize));
		float4 vS = tVelocityMap.Sample(sVelocitySampler, readTC + float2(0, -inverseSize));
		float4 vE = tVelocityMap.Sample(sVelocitySampler, readTC + float2(inverseSize, 0));
		float4 vW = tVelocityMap.Sample(sVelocitySampler, readTC + float2(-inverseSize, 0));
		// burntime control
		float tN = vN.w;
		float tS = vS.w;
		float tE = vE.w;
		float tW = vW.w;
		for (int l = int(cAffectNum.z); l < cAffectNum.w; l++)
		{
			float2 center = cAffectParam[l].xy;
			float radius = cAffectParam[l].z;
			float intensity = cAffectParam[l].w;
			float2 dir = currTC - center;
			float dist = length(dir);
			if (dist < radius)
			{
				oldV.z = 1.0;
				oldV.w = intensity;
			}
		}
	
		float density = tDensityMap.Sample(sDensitySampler, currTC * cDensityUV.xy + cDensityUV.zw).x;
		float rate = cRenderTime * 0.8;
		float burnRate = oldV.z;
		if (burnRate >= 0.0) {
			vN.z *= step(vN.z,1.01) * step(0.001,vN.w);
			vS.z *= step(vS.z,1.01) * step(0.001,vS.w);
			vE.z *= step(vE.z,1.01) * step(0.001,vE.w);
			vW.z *= step(vW.z,1.01) * step(0.001,vN.w);
			burnRate += (vN.z * rate +  vS.z * rate + vE.z * rate + vW.z * rate) * density;
			//burnRate = saturate(burnRate);
		}
	
		float newTime = max(max(max(tN, tS), max(tE, tW)), oldV.w);
		newTime = min(newTime,oldV.w + cRenderTime * 1.05);
		newTime -= cRenderTime;
		newTime = max(newTime,0.0);
	
		//burnRate = density;
	
		return float4(resVel.xy, 0, 0);
	}
}

float4 DivergencePS(v2f IN)	: SV_Target
{			
	// Find neighboring velocities:
	float inverseSize = 1.0/256.0;
	float4 vN = tVelocityMap.Sample(sVelocitySampler, IN.TC + float2(0, inverseSize));
	float4 vS = tVelocityMap.Sample(sVelocitySampler, IN.TC + float2(0, -inverseSize));
	float4 vE = tVelocityMap.Sample(sVelocitySampler, IN.TC + float2(inverseSize, 0));
	float4 vW = tVelocityMap.Sample(sVelocitySampler, IN.TC + float2(-inverseSize, 0));
	// with velocity * 2 - 1 and result *= halfcell  (halfcell = 0.5)
	float result = (vE.x - vW.x + vN.y - vS.y) * 0.5;
	return float4(result, 0, 0, 1);
}

float4 PressurePS(v2f IN)	: SV_Target
{			
	// Find neighboring pressure:
	float inverseSize = 1.0/256.0;
	float pN = tPressureMap.Sample(sPressureSampler, IN.TC + float2(0, inverseSize)).x;
	float pS = tPressureMap.Sample(sPressureSampler, IN.TC +  float2(0, -inverseSize)).x;
	float pE = tPressureMap.Sample(sPressureSampler, IN.TC +  float2(inverseSize, 0)).x;
	float pW = tPressureMap.Sample(sPressureSampler, IN.TC +  float2(-inverseSize, 0)).x;
	float bC = tDivergenceMap.Sample(sDivergenceSampler, IN.TC).x;
	float result = (pW + pE + pS + pN - bC) * 0.25;
	return float4(result, 0, 0, 1);
}

float4 SubtractGradient(v2f IN)	: SV_Target
{			
	// Find neighboring pressure:
	// Move (* 2 - 1) to grad calculation
	float inverseSize = 1.0/256.0;
	float pN = tPressureMap.Sample(sPressureSampler, IN.TC + float2(0, inverseSize)).x;
	float pS = tPressureMap.Sample(sPressureSampler, IN.TC + float2(0, -inverseSize)).x;
	float pE = tPressureMap.Sample(sPressureSampler, IN.TC + float2(inverseSize, 0)).x;
	float pW = tPressureMap.Sample(sPressureSampler, IN.TC + float2(-inverseSize, 0)).x;

	float4 oldV = tVelocityMap.Sample(sVelocitySampler, IN.TC);
	float2 grad = float2(pE - pW, pN - pS);
	float2 newV = oldV.xy - grad;
	return float4(newV, oldV.z, oldV.w);
}

float4 MovePS(v2f IN)	: SV_Target
{
	float2 uv = step(abs(IN.TC + cPosMove - 0.5), float2(0.5,0.5));
	return tVelocityMap.Sample(sVelocitySampler, IN.TC + cPosMove) * uv.x * uv.y;
}

///////////////// techniques //////////////////
technique Move
{
	VertexShader = FoliageVS();
	PixelShader = MovePS();
}

technique FluidAdvect
{
	VertexShader = FoliageVS();
	PixelShader = AdvectPS();
}

technique FluidImpulse
{
	VertexShader = FoliageVS();
	PixelShader = ImpulsePS();
}

technique FluidDivergence
{
	VertexShader = FoliageVS();
	PixelShader = DivergencePS();
}

technique FluidPressure
{
	VertexShader = FoliageVS();
	PixelShader = PressurePS();
}

technique FluidSubtractGradient
{
	VertexShader = FoliageVS();
	PixelShader = SubtractGradient();
}

///////////////// eof //////////////////
