//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_star.fx>
///	@path	~/System/Shader/
///	@date	2014/10/27
///	@brief	star.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

@_EnableVolumetricCloud
{
	Name = EnableVolumetricCloud;
	Description = "Use EnableVolumetricCloud";
	Visible = true;	
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float2 cPixelSizeInClipSpace		;
	float  cFade						;
	float  cPlaceHolder					;
	float2 cSunPos						;
	float2 cPlaceHolder2				;
	float3 cTintColor 					 = {1, 1, 1};
};

///////////////// io signatures //////////////////
struct a2v
{
	float4 Position 	: POSITION;		//in clip space (-1,1)
	float2 uv			: TEXCOORD0;	//uv
	float3 Color		: TEXCOORD1;	//pixel size 0-255
};

struct v2f
{
	float4 Position : SV_Position;
	float3 Color    : TEXCOORD0;
	float2 uv		: TEXCOORD1;
};

texture2D tFlareMap : layout(t0) = "Transparent";
sampler sFlareSampler : layout(s0)
{
	Texture = tFlareMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tCloudMap : CloudMap : layout(t1);
sampler sCloudMap : layout(s1)
{
	Texture = tCloudMap;
	Filter = Linear;
	Address = Clamp;
};

///////////////// general pass //////////////////
v2f FlareVS(a2v IN)
{
	v2f OUT;

	[@_HLSL]
	OUT = (v2f)0;

	//apply size
	OUT.Position = IN.Position;
	OUT.Position.z = 0;
	OUT.Position.w = 1.0-EPSILON*5;

	OUT.Color = IN.Color * cFade;
	OUT.uv = IN.uv;

	return OUT;
}

float4 FlarePS(v2f IN) : SV_Target
{
	float3 texCol = tFlareMap.Sample(sFlareSampler, IN.uv).rgb * cTintColor.rgb;
	if (@_EnableVolumetricCloud)
	{
		float4 ScatteringAndAlpha = tCloudMap.Sample(sCloudMap, cSunPos);
		float cloudOccluder = saturate(1-ScatteringAndAlpha.a*3);
		texCol *= cloudOccluder;
	}
	
	return float4(texCol.rgb * IN.Color, 0);
}

///////////////// techniques //////////////////
technique FlareTech
{
	VertexShader = FlareVS();
	PixelShader = FlarePS();
}
///////////////// eof //////////////////
