//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_all.fx>
///	@path	~/System/Shader/
///	@date	2014/10/24
///	@brief	HDR SKY.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"
#include "MonteCarlo.fxh"

///////////////// Properties //////////////////
@_HAS_PROBEA
{
	Name = HasProbeA;
	AffectedTex = tProbeA;
	Description = "Do we has tProbeA";
	Visible = true;
};

@_HAS_PROBEB
{
	Name = HasProbeB;
	AffectedTex = tProbeB;
	Description = "Do we has tProbeB";
	Visible = true;
};

@_HAS_PROBEC
{
	Name = HasProbeC;
	AffectedTex = tProbeC;
	Description = "Do we has tProbeC";
	Visible = true;
};

@_HAS_TOD_PROBEA
{
	Name = HasTODProbeA;
	AffectedTex = tNormalProbeA;
	Description = "Do we has tTODProbeA";
	Visible = true;
};

@_HAS_TOD_PROBEB
{
	Name = HasTODProbeB;
	AffectedTex = tNormalProbeB;
	Description = "Do we has tTODProbeB";
	Visible = true;
};

@_HAS_TOD_PROBEC
{
	Name = HasTODProbeC;
	AffectedTex = tNormalProbeC;
	Description = "Do we has tTODProbeC";
	Visible = true;
};

@_HAS_SKY_UV_MAP
{
	Name = HasSkyUVMap;
	Description = "Do we need baking sky";
	AffectedTex = tSkyUVMap;
	Visible = true;
};
@_DAY_LIGHTING
{
	Name = DayLighting;
	Description = "Shall we apply day lighting?";
	Visible = true;
};
@_NIGHT_LIGHTING
{
	Name = NightLighting;
	Description = "Shall we apply night lighting?";
	Visible = true;
};
@_DOME_CLOUD
{
	Name = usingDomeCloud;
	AffectedTex = DomeCloudMap;
	Description = "Using Dome Cloud texture";
	Visible = true;
};
@_ENABLE_SIMPLE_MIP
{
	Name = EnableSimpleMip;
	Description = "simple mip generate";
	Visible = true;
	Default = false;
}
@_ENABLE_SAMPLE_SKY
{
	Name = EnableSimpleMip;
	Description = "enable sample sky";
	Visible = true;
	Default = false;
}
///////////////// const buffer //////////////////
cbuffer Batch
{
	float   cWeightA								;
	float   cWeightB								;
	float   cWeightC								;
	float   cMipLevelA								;
	float   cMipLevelB								;
	float   cMipLevelC								;
	float   cPlaceHold1								;
	float   cPlaceHold2								;
	float4 SkyDome_PartialRayleighInScatteringConst ;
	float4 SkyDome_PhaseFunctionConstants 			;
	float4 SkyDome_PartialMieInScatteringConst 		;
	float4 SkyDome_NightSkyColBase 					;
	float4 SkyDome_NightSkyColDelta					;
	float4 SkyDome_NightMoonDirSize 				;
	float4 SkyDome_NightMoonInnerCoronaColorScale 	;
	float4 SkyDome_NightMoonOuterCoronaColorScale 	;
	float4 SkyDome_MapAndFogParam					;
	float4 DomeCloudMovement						;
	float4 DomeCloudUVScaleDensity					;
	float4 DomeCloudAmbientColorSunScale			;
	float3 RealSunDir								;
	float2 cInvSize									;
};

///////////////// samplers //////////////////
texture2D tProbeA : layout(t0) = "Transparent";
sampler sProbeA : layout(s0)
{
	Texture = tProbeA;
	Filter = Linear;
	Address = Clamp;
};
texture2D tProbeB : layout(t1) = "Transparent";
sampler sProbeB : layout(s1)
{
	Texture = tProbeB;
	Filter = Linear;
	Address = Clamp;
};
texture2D tProbeC : layout(t2) = "Transparent";
sampler sProbeC : layout(s2)
{
	Texture = tProbeC;
	Filter = Linear;
	Address = Clamp;
};
texture2D tNormalProbeA : layout(t3) = "Transparent";
sampler sNormalProbeA : layout(s3)
{
	Texture = tNormalProbeA;
	Filter = Point;
	Address = Clamp;
};
texture2D tNormalProbeB : layout(t4) = "Transparent";
sampler sNormalProbeB : layout(s4)
{
	Texture = tNormalProbeB;
	Filter = Point;
	Address = Clamp;
};
texture2D tNormalProbeC : layout(t5) = "Transparent";
sampler sNormalProbeC : layout(s5)
{
	Texture = tNormalProbeC;
	Filter = Point;
	Address = Clamp;
};
texture2D tSkyUVMap : layout(t6) = "Transparent";
sampler sSkyUVMap : layout(s6)
{
	Texture = tSkyUVMap;
	Filter = Linear;
	Address = Clamp;
};
texture2D skyDomeTextureMie : layout(t7) = "White";
sampler skyDomeSamplerMie : layout(s7)
{
	Texture = skyDomeTextureMie;
	Filter = Linear;
	Address = Clamp;
};
texture2D skyDomeTextureRayleigh : layout(t8) = "White";
sampler skyDomeSamplerRayleigh : layout(s8)
{
	Texture = skyDomeTextureRayleigh;
	Filter = Linear;
	Address = Clamp;
};

texture2D DomeCloudMap : layout(t9) = "Transparent";
sampler DomeCloudSampler : layout(s9)
{
	Texture = DomeCloudMap;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = None;
	Address = Wrap;
};
texture2D tInputRT : layout(t10);
sampler sInputRT_Sampler : layout(s10)
{
	Texture = tInputRT;
	Filter = Linear;
	Address = Clamp;
};
texture2D tSkyColorMap : layout(t11) = "Transparent";
sampler sSkyColorMap : layout(s11)
{
	Texture = tSkyColorMap;
	Filter = Linear;
	Address = Clamp;
};
///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////// general pass //////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float3 CalcSkyColor(half2 baseTC)
{
	half3 skyColor = half3(0, 0, 0);
	/*if(normalLength > 0.3) // sb filter
	{
		return skyColor;
	}*/
	
	float2 skyUV = tSkyUVMap.SampleLevel(sSkyUVMap, baseTC, 0).rg;
	float angleA = skyUV.x * PI * 2;
	float angleB = skyUV.y * PI;
	skyUV.y *= 2;
	float sinAngleB = sin(angleB);
	float3 skyDir = float3(cos(angleA) * sinAngleB, cos(angleB), sin(angleA) * sinAngleB); // is equal worldPos - cameraPos
	//return skyDir;
	if (@_DAY_LIGHTING)
	{
		float4 ColorMie = SampleColorMap(skyDomeTextureMie, skyDomeSamplerMie, skyUV);
		float4 ColorRayleigh = SampleColorMap(skyDomeTextureRayleigh, skyDomeSamplerRayleigh, skyUV);
		
		ColorMie.xyz = RGBMDecode(ColorMie, SkyDome_PartialRayleighInScatteringConst.w);
		ColorRayleigh.xyz = RGBMDecode(ColorRayleigh, SkyDome_PartialRayleighInScatteringConst.w);
		float miePart_g_2 = SkyDome_PhaseFunctionConstants.x;
		float miePart_g2_1 = SkyDome_PhaseFunctionConstants.y;

		float cosine = dot(RealSunDir, skyDir);
		float cosine2 = cosine * cosine;

		//float miePhase = miePart * ( 1.0 + cosine2 ) / pow( 1.0 + g2 - 2.0 * g * cosine, 1.5 );
		float miePhase = (1.0 + cosine2) * pow(miePart_g2_1 + miePart_g_2 * cosine, -1.5);
		float rayleighPhase = 0.75 * (1.0 + cosine2);	
		skyColor += ColorRayleigh.xyz * SkyDome_PartialRayleighInScatteringConst.xyz * rayleighPhase;
		skyColor += ColorMie.xyz * SkyDome_PartialMieInScatteringConst.xyz * miePhase;
	}
	if (@_NIGHT_LIGHTING)
	{
		float gr = saturate(skyDir.y * SkyDome_NightSkyColBase.w + SkyDome_NightSkyColDelta.w);
		gr *= 2 - gr;
		gr *= 2 - gr;
		skyColor += SkyDome_NightSkyColBase.xyz + SkyDome_NightSkyColDelta.xyz * gr;
			
		// add inner and outer moon corona
		float m = 1 - dot(skyDir, SkyDome_NightMoonDirSize.xyz);
		float innerScale = 1.0 / (1.05 + m * SkyDome_NightMoonInnerCoronaColorScale.w);
		innerScale = lerp(0.0, 1.0, max(0, innerScale-0.04));
		skyColor += SkyDome_NightMoonInnerCoronaColorScale.xyz * innerScale;
		skyColor += SkyDome_NightMoonOuterCoronaColorScale.xyz * (1.0 / (1.05 + m * SkyDome_NightMoonOuterCoronaColorScale.w));
	}
	if(@_DOME_CLOUD)
	{
		const float CloudLayerHeight = 8000.0;
		const float CloudUVSize = 16000.0;
		const float cPhaseG = 0.85;
		const float CloudLayerFlatten = 0.17;
		
		float2 CloudLayerUVScrolling = -DomeCloudMovement.zx * 0.002 * DomeCloudMovement.w;
		float2 CloudMaskUVScrolling = -CloudLayerUVScrolling * 0.5;

		float CloudMaskUVScale = DomeCloudUVScaleDensity.x;
		float CloudLayerUVScale = DomeCloudUVScaleDensity.y;
		float CloudLayer1Density = DomeCloudUVScaleDensity.z;
		float CloudLayer2Density = DomeCloudUVScaleDensity.w;
		float3 AmbientColor = DomeCloudAmbientColorSunScale.xyz;
		float SunIntensityScale = DomeCloudAmbientColorSunScale.w / 100.0;

		//Calc UV
		float3 startPos = float3(CameraPos.x, 0, CameraPos.z) + skyDir * CloudLayerHeight / (lerp(skyDir.y, 1, CloudLayerFlatten));
		startPos = -startPos / CloudUVSize;
		float2 cloudUV = startPos.zx * CloudLayerUVScale + GameTime * CloudLayerUVScrolling;
		float2 maskUV = startPos.zx * CloudMaskUVScale + GameTime * CloudMaskUVScrolling ;

		//Calc Density
		float4 skyMapInfo = (DomeCloudMap.SampleLevel(DomeCloudSampler, cloudUV, 0));
		float4 skyMapMask = DecodeGamma(DomeCloudMap.SampleLevel(DomeCloudSampler, maskUV, 0));
		float Layer1 = skyMapInfo.x * skyMapMask.z * CloudLayer1Density;
		float Layer2 = skyMapInfo.y * (skyMapMask.z + 1) * CloudLayer2Density;
		float TotalDensity = (Layer1 + Layer2);

		//Calc Lighting
		half VoL = dot(SunDirection.xyz, skyDir);
		float lightingDensity = lightEnergy(VoL, TotalDensity, cPhaseG);
		float4 ResultColor = float4(AmbientColor + lightingDensity * SunColor * ThunderRate * SunIntensityScale , saturate(TotalDensity));
		
		if(skyDir.y > 0)
			skyColor = lerp(skyColor, ResultColor.xyz, ResultColor.w);
	}
	if (@_HDR)
		skyColor *= SkyDome_PhaseFunctionConstants.z;
	if (@_Weather == WEATHER_RAIN)
		skyColor *= (1 - 0.8*RainRate);
	float fogT = saturate(SkyDome_MapAndFogParam.w);
	float3 fogFinalColor = float3(0,0,0);
	SkyFogLightingAndScatter(fogFinalColor, -skyDir, SunDirection.xyz);
	skyColor = lerp(skyColor, fogFinalColor, fogT);
	return skyColor;
}
float3 ImportanceSampleGGX2(float2 E, float Roughness, float3 N)
{
	float m = Roughness * Roughness;
	float a2 = m * m;
	float Phi = 2 * PI * E.x;
	float CosTheta = sqrt((1 - E.y) / (1 + (a2 - 1) * E.y));
	float SinTheta = sqrt(1 - CosTheta * CosTheta);

	float3 H;
	H.x = SinTheta * cos(Phi);
	H.y = SinTheta * sin(Phi);
	H.z = CosTheta;

	return H;
}
float3 PreFilterMap(float2 spos, float roughness, float3 R)
{
	static const int sampleNum = 16;
	float totalWeight = 0;
	uint2 randomSeed = ComputeRandomSeed(uint2(spos), 0); // only one frame
	float3 V = R;
	float3 N = R;
	float3 filterColor = float3(0, 0, 0);
	float3 UpVector = abs(N.y) < 0.999 ? float3(0, 1, 0) : float3(1, 0, 0);
	float3 TangentX = normalize(cross(UpVector, N));
	float3 TangentY = cross(N, TangentX);
	[unroll]
	for(int i = 0; i < sampleNum; ++i)
	{
		float2 randomUV = Hammersley16(i, sampleNum, randomSeed);
		float3 H = ImportanceSampleGGX2(randomUV, roughness, N);
		H = TangentX * H.x + TangentY * H.y + N * H.z;
		float3 L = 2 * dot(V, H) * H - V;

		float NoL = saturate(dot(N, L));
		if(NoL > 0.0f)
		{
			/*float NoH = saturate(dot(N, H));
			float VoH = saturate(dot(V, H));
			float pdf = D_GGX(roughness, NoH) * NoH / (4 * VoH);
			float omegaS = 1 / (sampleNum * pdf);
			float omegaP = 4 * PI / (6.0 * envmapSize * envMapSize);
			float mipBias = 1.0f;
			float mipLevel = max(0.5 * log2(omegaS / omegaP) + mipBias, 0.0f);*/
			float mipLevel = 0; // use trick first		
			half fSign = L.z > 0;
			half fSign2 = fSign * 2 - 1;
			L.xy /= L.z * fSign2 + 1;
			L.xy = L.xy * half2(0.25, -0.25) + (0.25 + 0.5 * fSign);
			filterColor += RGBMDecode(tInputRT.SampleLevel(sInputRT_Sampler, L.xy, 0), MAX_PBR_RGBM_VALUE) * NoL;
			totalWeight += NoL;
		}
	}
	return filterColor / totalWeight;
}
#define SqureLength(A)  dot(A, A)

float4 MergeProbePS(v2f IN) : SV_Target
{
	half4 Color = half4(0, 0, 0, 0);
	//input
	half2 baseTC = IN.TC.xy;
	if(@_ENABLE_SIMPLE_MIP && @_HAS_SKY_UV_MAP)
	{
		float2 skyUV = tSkyUVMap.SampleLevel(sSkyUVMap, baseTC, 0).rg;
		float angleA = skyUV.x * PI * 2;
		float angleB = skyUV.y * PI;
		skyUV.y *= 2;
		float sinAngleB = sin(angleB);
		float3 skyDir = float3(cos(angleA) * sinAngleB, cos(angleB), sin(angleA) * sinAngleB);
		Color.rgb = PreFilterMap(IN.HPosition.xy, 0.17 * cMipLevelA, skyDir); // roughness = 0.17 * cMipLevelA
		RGBMEncode(Color.rgb, MAX_PBR_RGBM_VALUE, Color);
		return Color;
	}
	float normalLength = 1;
	//texture
	if(@_HAS_TOD_PROBEA)
	{
		half4 srcColor = tProbeA.SampleLevel(sProbeA, baseTC, cMipLevelA).rgba;
		float4 srcNormal = tNormalProbeA.SampleLevel(sNormalProbeA, baseTC, cMipLevelA).rgba;
		float3 worldNormal = DecodeNormal(srcNormal.rgb);
		normalLength = min(normalLength, SqureLength(srcNormal.rgb));
		half3 probeColor = SunColor.rgb * saturate(dot(SunDirection.xyz, worldNormal)) * srcColor.rgb;// + FogColor.rgb * (1 - srcNormal.a);	
		probeColor += (SunFogAndAmbientGI(worldNormal) * LITMAP_GISCALE * srcColor.rgb);
		Color.rgb += probeColor * cWeightA;
	}
	else if(@_HAS_PROBEA)
	{
		half4 srcColor = tProbeA.SampleLevel(sProbeA, baseTC, cMipLevelA).rgba;
		Color.rgb += RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) * cWeightA;
	}

	if(@_HAS_TOD_PROBEB)
	{
		half4 srcColor = tProbeB.SampleLevel(sProbeB, baseTC, cMipLevelB).rgba;
		float4 srcNormal = tNormalProbeB.SampleLevel(sNormalProbeB, baseTC, cMipLevelB).rgba;
		float3 worldNormal = DecodeNormal(srcNormal.rgb);
		normalLength = min(normalLength, SqureLength(srcNormal.rgb));
		half3 probeColor = SunColor.rgb * saturate(dot(SunDirection.xyz, worldNormal)) * srcColor.rgb;// + FogColor.rgb * (1 - srcNormal.a);
		probeColor += (SunFogAndAmbientGI(worldNormal) * LITMAP_GISCALE * srcColor.rgb);
		Color.rgb += probeColor * cWeightB;
	}
	else if(@_HAS_PROBEB)
	{
		half4 srcColor = tProbeB.SampleLevel(sProbeB, baseTC, cMipLevelB).rgba;
		Color.rgb += RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) * cWeightB;
	}

	if(@_HAS_TOD_PROBEC)
	{
		half4 srcColor = tProbeC.SampleLevel(sProbeC, baseTC, cMipLevelC).rgba;
		float4 srcNormal = tNormalProbeC.SampleLevel(sNormalProbeC, baseTC, cMipLevelC).rgba;
		float3 worldNormal = DecodeNormal(srcNormal.rgb);
		normalLength = min(normalLength, SqureLength(srcNormal.rgb));
		half3 probeColor = SunColor.rgb * saturate(dot(SunDirection.xyz, worldNormal)) * srcColor.rgb;// + FogColor.rgb * (1 - srcNormal.a);
		probeColor += (SunFogAndAmbientGI(worldNormal) * LITMAP_GISCALE * srcColor.rgb);
		Color.rgb += probeColor * cWeightC;
	}
	else if(@_HAS_PROBEC)
	{
		half4 srcColor = tProbeC.SampleLevel(sProbeC, baseTC, cMipLevelC).rgba;
		Color.rgb += RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) * cWeightC;
	}
	if(normalLength < 0.3)
		Color.rgb += tSkyColorMap.Sample(sSkyColorMap, baseTC).rgb;
	
	RGBMEncode(Color.rgb, MAX_PBR_RGBM_VALUE, Color);
	return Color;
}

///////////////// techniques //////////////////
technique MergeProbeTech
{
	VertexShader = ScreenVS();
	PixelShader = MergeProbePS();
}

float4 BakingSkyPS(v2f IN) : SV_Target
{
	half4 Color = half4(0, 0, 0, 1);
	//input
	half2 baseTC = IN.TC.xy;

	Color.rgb = CalcSkyColor(baseTC);
	return Color;
}

technique SkyProbeRenderTech
{
	VertexShader = ScreenVS();
	PixelShader = BakingSkyPS();
}
///////////////// eof //////////////////

