//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_all.fx>
///	@path	~/System/Shader/
///	@date	2014/10/24
///	@brief	HDR SKY.

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

#pragma shader draw

///////////////// Properties //////////////////
@_DAY_LIGHTING
{
	Name = DayLighting;
	Description = "Shall we apply day lighting?";
	Visible = true;
};
@_NIGHT_LIGHTING
{
	Name = NightLighting;
	Description = "Shall we apply day lighting?";
	Visible = true;
};
@_HAS_SKYMAP
{
	Name = HasSkymap;
	AffectedTex = skyMap;
	Description = "Do we has skymap";
	Visible = true;
};
@_DOME_CLOUD
{
	Name = usingDomeCloud;
	AffectedTex = DomeCloudMap;
	Description = "Using Dome Cloud texture";
	Visible = true;
}
//%_HDR, system property

///////////////// const buffer //////////////////
#define SkyDome_GAMMA 	SkyDome_PartialMieInScatteringConst.w
#define SkyDome_RGBS_K	SkyDome_PartialRayleighInScatteringConst.w
#define SkyDome_HDR 	SkyDome_PhaseFunctionConstants.z
// 1.0f / ( nightSkyZenithColShift - minNightSkyZenithGradient(-0.1) )
#define SkyDome_NightSkyZenithColShiftX	SkyDome_NightSkyColBase.w
//-minNightSkyZenithGradient(-0.1) / ( nightSkyZenithColShift - minNightSkyZenithGradient(-0.1) ) 
#define SkyDome_NightSkyZenithColShiftY	SkyDome_NightSkyColDelta.w
#define cameraPos		CameraPos.xyz
#define farPlane		CameraInfo.y
#define MapBeginU		SkyDome_MapAndFogParam.x
#define MapInvLenV		SkyDome_MapAndFogParam.y
#define SkyFogBegin 	SkyDome_MapAndFogParam.z
#define SkyFogScale 	SkyDome_MapAndFogParam.w
#define MoonBlend		SkyDome_NightMoonColor.w

cbuffer Batch
{
	//day sky
	float4 SkyDome_PartialMieInScatteringConst 		;
	float4 SkyDome_PartialRayleighInScatteringConst ;
	float4 SkyDome_PhaseFunctionConstants 			;

	//night sky
		// nightSkyHorizonCol * nightSkyHorizonScale
	float4 SkyDome_NightSkyColBase 					;
		// nightSkyZenithCol * nightSkyZenithScale - nightSkyHorizonCol * nightSkyHorizonScale 
	float4 SkyDome_NightSkyColDelta					;
		// {nightMoonDirection, 25.0f - 24.0f * clamp_tpl( nightMoonSize, 0.0f, 1.0f)}
	float4 SkyDome_NightMoonDirSize 				;
		// nightMoonColor
	float4 SkyDome_NightMoonColor 					;
		// nightMoonInnerCoronaColor, nightMoonInnerCoronaScale
	float4 SkyDome_NightMoonInnerCoronaColorScale 	;
		// nightMoonOuterCoronaColor, nightMoonOuterCoronaScale
	float4 SkyDome_NightMoonOuterCoronaColorScale 	;
		// skymap and skyfog param, [beginU,lenV,SkyFogBegin, SkyFogEndBengin]
	float4 SkyDome_MapAndFogParam					;

	//world
	float4x3 World 									: World	;

	float  SkyDome_BloomScale						 = 1.0;

	float3 RealSunDir								;

	//Dome Cloud
	float4 DomeCloudMovement;
	float4 DomeCloudUVScaleDensity;
	float4 DomeCloudAmbientColorSunScale;
};

///////////////// samplers //////////////////
texture2D skyDomeTextureMie : layout(t0) = "White";
sampler skyDomeSamplerMie : layout(s0)
{
	Texture = skyDomeTextureMie;
	Filter = Linear;
	Address = Clamp;
};
texture2D skyDomeTextureRayleigh : layout(t1) = "White";
sampler skyDomeSamplerRayleigh : layout(s1)
{
	Texture = skyDomeTextureRayleigh;
	Filter = Linear;
	Address = Clamp;
};
texture2D skyDomeTextureMoon : layout(t2) = "Transparent";
sampler skyDomeSamplerMoon : layout(s2)
{
	Texture = skyDomeTextureRayleigh;
	Filter = Linear;
	Address = Clamp;
};
texture2D skyMap : layout(t3) = "Transparent";
sampler skyMapSampler : layout(s3)
{
	Texture = skyMap;
	Filter = Linear;
	Address = Wrap;
};
texture2D moonMap : layout(t4) = "Transparent";
sampler moonMapSampler : layout(s4)
{
	Texture = moonMap;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = None;
	Address = Clamp;
};
texture2D DomeCloudMap : layout(t5) = "Transparent";
sampler DomeCloudSampler : layout(s5)
{
	Texture = DomeCloudMap;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = None;
	Address = Wrap;
};



///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 baseTC   : TEXCOORD0;
};

struct v2f_hdr
{
	float4 Position			: SV_Position;  //in projection space
	float4 packedTC			: TEXCOORD0;
	float3 skyDir			: TEXCOORD1;
	float2 mapTC			: TEXCOORD2;
	float fogT 				: TEXCOORD3;
};

///////////////// general pass //////////////////

v2f_hdr SkyHDRVS(a2v IN)
{
	v2f_hdr OUT;
	
	[@_HLSL]
	OUT = (v2f_hdr)0;

	// Position in screen space.
	float4 vPos = IN.Position;
	float4 worldPos = float4(MulMat3x4(World, vPos).xyz, 1);
	OUT.Position = MulMat4x4(ViewProj, worldPos);
	OUT.Position.z = (EPSILON * 10)*OUT.Position.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.
	//ISSUE:Depth will be Diff with Screen After SkyPass
	//Affect PostProcessing
	float2 baseTC = IN.baseTC;
	float2 moonTC = float2(0.0, 0.0);

	if (@_NIGHT_LIGHTING)
	{
		float3 NX = cross( float3( 0, 1, 0 ), SkyDome_NightMoonDirSize.xyz );
		float3 NY = cross( SkyDome_NightMoonDirSize.xyz, NX );
		NX = normalize( -NX );
		NY = normalize( -NY );
		moonTC = float2( dot( NX, vPos.xyz ), dot( NY, vPos.xyz ) ) * SkyDome_NightMoonDirSize.w;
		moonTC = moonTC*0.5 + 0.5;

		if ( dot( vPos.xyz, SkyDome_NightMoonDirSize.xyz ) < 0 )
			moonTC = -100000.0;
	}

	[@_HAS_SKYMAP]
	{
		float2 mapTC = baseTC;
		mapTC.x += MapBeginU;
		mapTC.y *= MapInvLenV;
		mapTC.y  = min(0.998, mapTC.y);//can't set clamp for address_v now...
		OUT.mapTC.xy = mapTC.xy;
	}

	OUT.packedTC = float4(baseTC, moonTC.yx);
	OUT.skyDir = vPos.xyz;
	OUT.fogT = FogTransmittance(vPos.xyz * CameraInfo.y);
	return OUT;
}

float4 SkyHDRPS(v2f_hdr IN) : SV_Target
{
	float2 baseTC = IN.packedTC.xy;
	float2 moonTC = IN.packedTC.wz;
	float3 skyDir = normalize(IN.skyDir);
	half4 raw = half4(1,1,1,1);
	float4 OUT = float4(0, 0, 0, 0);
	float Alpha = 1;


	// if (@_DAY_LIGHTING)
	// {
	// 	float4 ColorMie = SampleColorMap(skyDomeTextureMie, skyDomeSamplerMie, baseTC.xy);
	// 	float4 ColorRayleigh = SampleColorMap(skyDomeTextureRayleigh, skyDomeSamplerRayleigh, baseTC.xy);

	// 	//if (!@_SUPPORT_F16)
	// 	{
	// 		ColorMie.xyz = RGBMDecode( ColorMie, SkyDome_RGBS_K );
	// 		ColorRayleigh.xyz = RGBMDecode( ColorRayleigh, SkyDome_RGBS_K );
	// 	}
	// 	float miePart_g_2 = SkyDome_PhaseFunctionConstants.x;  // miePart*(-2.0*g)
	// 	float miePart_g2_1 = SkyDome_PhaseFunctionConstants.y; // miePart*(1+g2)

	// 	float cosine = dot( RealSunDir, skyDir );
	// 	float cosine2 = cosine * cosine;

	// 	//float miePhase = miePart * ( 1.0 + cosine2 ) / pow( 1.0 + g2 - 2.0 * g * cosine, 1.5 );
	// 	float miePhase = ( 1.0 + cosine2 ) * pow( miePart_g2_1 + miePart_g_2 * cosine, -1.5 );
	// 	float rayleighPhase = 0.75 * ( 1.0 + cosine2 );
		
	// 	// no rayleigh scattering if we have skymap
	// 	[!@_HAS_SKYMAP]
	// 	OUT.xyz += ColorRayleigh.xyz * SkyDome_PartialRayleighInScatteringConst.xyz * rayleighPhase;

	// 	// always have mie scattering
	// 	OUT.xyz += ColorMie.xyz * SkyDome_PartialMieInScatteringConst.xyz * miePhase*10;
	// }

	// if (@_NIGHT_LIGHTING)  // add horizontal night sky gradient
	// {
	// 	if(!@_HAS_SKYMAP)
	// 	{
	// 		// sky
	// 		float gr = saturate( skyDir.y * SkyDome_NightSkyZenithColShiftX + SkyDome_NightSkyZenithColShiftY );
	// 		gr *= 2 - gr;
	// 		gr *= 2 - gr;
	// 		OUT.xyz += SkyDome_NightSkyColBase.xyz + SkyDome_NightSkyColDelta.xyz * gr;
				
	// 		// add inner and outer moon corona
	// 		float m = 1 - dot( skyDir, SkyDome_NightMoonDirSize.xyz );
	// 		float innerScale = 1.0 / ( 1.05 + m * SkyDome_NightMoonInnerCoronaColorScale.w );
	// 		innerScale = lerp(0.0, 1.0, max(0,innerScale-0.04) );
	// 		OUT.xyz += SkyDome_NightMoonInnerCoronaColorScale.xyz * innerScale;
	// 		OUT.xyz += SkyDome_NightMoonOuterCoronaColorScale.xyz * (1.0 / ( 1.05 + m * SkyDome_NightMoonOuterCoronaColorScale.w ));
	// 	}
		
	// }

	//Moon:
	// float4 moonAlbedo = moonMap.SampleLevel(moonMapSampler, moonTC.xy, 0);
	// moonAlbedo.xyz *= moonAlbedo.xyz;
	// float3 moonLighting = SkyDome_NightMoonColor.xyz * moonAlbedo.xyz;
	// OUT.xyz = OUT.xyz*min(1.0,MoonBlend-moonAlbedo.a) + moonLighting;
	// Alpha = saturate(1.0 - moonAlbedo.a);

	if (@_HAS_SKYMAP)
	{
		 raw = SampleColorMap(skyMap, skyMapSampler, float2(IN.mapTC.x,IN.mapTC.y));
		//degamma
		raw.rgb = pow(raw,2) ;
		OUT.xyz +=(raw.rgb*10*raw.a);
		// raw = DecodeGamma(raw);
	}

	//Sky High Layer
	if(@_DOME_CLOUD)
	{
		const float CloudLayerHeight = 8000.0;
		const float CloudUVSize = 16000.0;
		const float cPhaseG = 0.85;
		const float CloudLayerFlatten = 0.17;
		
		float2 CloudLayerUVScrolling = -DomeCloudMovement.zx * 0.002 * DomeCloudMovement.w;
		float2 CloudMaskUVScrolling = -CloudLayerUVScrolling * 0.5;

		float CloudMaskUVScale = DomeCloudUVScaleDensity.x;
		float CloudLayerUVScale = DomeCloudUVScaleDensity.y;
		float CloudLayer1Density = DomeCloudUVScaleDensity.z;
		float CloudLayer2Density = DomeCloudUVScaleDensity.w;
		float3 AmbientColor = DomeCloudAmbientColorSunScale.xyz;
		float SunIntensityScale = DomeCloudAmbientColorSunScale.w / 100.0;

		//Calc UV
		float3 viewDir = IN.skyDir.xyz;
		float3 startPos = float3(CameraPos.x,0,CameraPos.z) + viewDir * CloudLayerHeight / (lerp(viewDir.y,1,CloudLayerFlatten));
		startPos = -startPos / CloudUVSize;
		float2 cloudUV = startPos.zx * CloudLayerUVScale + GameTime * CloudLayerUVScrolling;
		float2 maskUV = startPos.zx * CloudMaskUVScale + GameTime * CloudMaskUVScrolling ;

		//Calc Density
		float4 skyMapInfo = (DomeCloudMap.SampleLevel(DomeCloudSampler, cloudUV, 0));
		float4 skyMapMask = DecodeGamma(DomeCloudMap.SampleLevel(DomeCloudSampler, maskUV, 0));
		float Layer1 = skyMapInfo.x * skyMapMask.z * CloudLayer1Density;
		float Layer2 = skyMapInfo.y * (skyMapMask.z+1) * CloudLayer2Density;
		float TotalDensity = (Layer1 + Layer2);

		//Calc Lighting
		half VoL = dot(SunDirection.xyz, viewDir);
		float lightingDensity = lightEnergy(VoL,TotalDensity,cPhaseG);
		float4 ResultColor = float4(AmbientColor + lightingDensity * SunColor * ThunderRate * SunIntensityScale , saturate(TotalDensity));
		
		if(viewDir.y>0)
			OUT.xyz = lerp(OUT.xyz , ResultColor.xyz , ResultColor.w);
	}

	if (@_HDR)
		OUT.xyz *= SkyDome_HDR;
	// rain rate
	// if (@_Weather == WEATHER_RAIN)
	// 	OUT.rgb *= 1 - 0.8*RainRate;

	// calc fog
	if (@_FOG && !@_VolumetricFog && !@_DEFERRED_SHADING)
	{
		//float fogT = saturate((baseTC.y - SkyFogBegin) * SkyFogScale);
		float fogT = IN.fogT * IN.fogT;
		float3 fogFinalColor = float3(0,0,0);
		SkyFogLightingAndScatter(fogFinalColor, -skyDir, SunDirection.xyz);
		OUT.rgb = lerp(fogFinalColor , OUT.rgb,saturate(fogT+(1-baseTC.y)*SkyFogBegin) );
	}
	else if (@_FOG && @_DEFERRED_SHADING)
	{
		float fogT = saturate(SkyFogScale);
		float3 fogFinalColor = float3(0,0,0);
		SkyFogLightingAndScatter(fogFinalColor, -skyDir, SunDirection.xyz);
		OUT.rgb = lerp(OUT.rgb, fogFinalColor, fogT);
	}

	// screen color tint
	OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);

	//hdr, calc fog in linear space
	EncodeHDR(OUT.rgb);
	OUT.a = Alpha;
	/*float2 baseModifyUV = IN.packedTC.xy;
	baseModifyUV.y/=2;
	return float4(baseModifyUV, 0, 1);*/
	return OUT;
}

///////////////// techniques //////////////////
//ZEnable          = TRUE;
//ZWriteEnable     = FALSE;
//CullMode 		 = NONE;
//FOGENABLE 		 = FALSE;

//ALPHATESTENABLE  = FALSE;
//ALPHABLENDENABLE = FALSE;
technique SkyHDRTech
{
	VertexShader = SkyHDRVS();
	PixelShader = SkyHDRPS();
}
///////////////// eof //////////////////
