//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_star.fx>
///	@path	~/System/Shader/
///	@date	2014/10/27
///	@brief	star.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////

///////////////// const buffer //////////////////
cbuffer Batch
{
	//worldviewproj
	float4x3 World 						: World	;

	float Star_Size						;
	float Star_Random 					;
	float Star_Intensity				;
	float Star_GAMMA					;
		// {nightMoonDirection, 25.0f - 24.0f * clamp_tpl( nightMoonSize, 0.0f, 1.0f)}
	float4 SkyDome_NightMoonDirSize 	;
};

///////////////// io signatures //////////////////
struct a2v
{
	float4 Position 	: POSITION;
	float4 ColorAndSize	: COLOR;
	float3 Offset		: NORMAL;		//quad center to vertice, normalize
	float2 uv			: TEXCOORD0;
};

struct v2f
{
	float4 Position : SV_Position;
	float4 Color    : COLOR;
	float2 uv		: TEXCOORD0;
};

///////////////// general pass //////////////////
float GetFlickerAmount(in float2 pos)
{
	return abs(0.1*sin(pos.x + GameTime * frac(pos.y*357)));
}

v2f StarsVS(a2v IN)
{
	v2f OUT;

	[@_HLSL]
	OUT = (v2f)0;

	//apply size
	float4 pos = IN.Position;
	float sSize = IN.ColorAndSize.w * Star_Size * 2.;
	pos.xyz += Expand(IN.Offset.xyz) * sSize;

	// Position in screen space.
	float4 worldPos = float4(MulMat3x4(World, pos).xyz, 1);
	OUT.Position = MulMat4x4(ViewProj, worldPos);
	OUT.Position.z = (EPSILON*5)*OUT.Position.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.

	// brightness
	//float4 ColAndMag = IN.ColAndMag;
	//float appMag = 6.5 + ColAndMag.w * (-1.44 - 6.5);
	float brightness = GetFlickerAmount(IN.Position.xy);// * pow(5.0, (-appMag - 1.44) / 2.5);
	OUT.Color = float4(brightness * IN.ColorAndSize.xyz, brightness);
	OUT.uv = IN.uv;

	return OUT;
}

float4 StarsPS(v2f IN) : SV_Target
{
	float2 uv = IN.uv;
	float2 distCenter = 3*uv.xy;
	float scale = exp(-dot(distCenter, distCenter));

	float3 colCool = IN.Color.xyz;
	float3 colHot = IN.Color.w;
	float3 colFinal = Star_Intensity * (colCool * scale + 5 * colHot * pow(scale, 10));
	float4 Color = float4(colFinal.xyz, 0.f);
	
	// Re-scale range
	//colFinal.xyz *= PS_HDR_RANGE_ADAPT_MAX;
	
	if(!@_HDR)
	{
		Color.xyz = 1 - exp( -Star_GAMMA * Color.xyz );
	}

	// rain rate
	if (@_Weather == WEATHER_RAIN)
		Color.rgb *= 1 - 0.8*RainRate;

	// screen color tint
	Color.rgb = lerp(Color.rgb, ScreenColor.rgb, ScreenColor.a);

	return Color;
}

///////////////// techniques //////////////////
technique StarHDRTech
{
	VertexShader = StarsVS();
	PixelShader = StarsPS();
}
///////////////// eof //////////////////
