//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GPUParticle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	GPU particle shader.

#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1
#include "Constants.fxh"
#include "Common.fxh"


#define USE_DEBUG_TECHNIQUE 0
#define USE_DEBUG_CLIPPING 0
#define FORCE_ENABLE_FOG true
#pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

#pragma shader draw


@_UserNormalMapEnable
{
	Name = UserNormalMapEnablecity;
	Description = "Using NormalMap";
	Default = false;
}
@_UserMixMapEnable
{
	Name = UserMixMapEnablecity;
	Description = "Using MixMap";
	Default = false;
}
@_UserSinkByAlphaEnable
{
	Name = UserSinkByAlphaEnable;
	Description = "SinkToGroundByAlphaValue";
	Default = false;
}
@_VELOCITY
{
	Name = Velocity;
	Description = "Facing Vecloity";
	Visible = false;
}
@_DISABLEDIFFUSE
{
	Name = DisableDiffuse;
	Description = "Disable Diffuse";
	Visible = false;
	Fallback = false;
}

// constants
cbuffer Batch
{
	float4x3 World 				: World;
	float4   InstanceColor		: Color;
	float4 	cPointCloud[6]		: PointCloudData				;
	float4 	SubUV				;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  BloomScale			 = 0.5;
	float3 _Param0				;
	float4 _Param1				;
	float4 _ColorCurve[8]		;
	float4 _ParamCurve[8]		;	// xy : size, z : emission	
	float  NormalScale			 < Widget = "FloatSlider"; Min = "0"; Max = "5.0"; Step = "0.01"; Precision = "3"; > = 1;
	float  Roughness			 < Widget = "FloatSlider"; Min = "0.08"; Max = "0.92"; Step = "0.01"; Precision = "3"; > = 0.9;
	float  cNormalMapBias		 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;
	float2 UVScale				 = { 1, 1 };
	float  Metallic				  < Widget = "FloatSlider"; Min = "0.001"; Max = "0.999"; Step = "0.01"; Precision = "3"; > = 0.1;
	float  AlphaScale			  < Widget = "FloatSlider"; Min = "0.0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 1;
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  InvFadeDistance		 = 1;
	float4 fadeColor			 = { 0, 0, 0, 0 };
	float  SinkDepth			 = 0.1f;
	float  YOffset			 	 = 0.05f;
	float  Thickness			 < Widget = "FloatSlider"; Min = "0.2"; Max = "1.0"; Step = "0.01"; Precision = "3"; > = 1.0;
	float  Scattering			 < Widget = "FloatSlider"; Min = "0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 0.0;
	float  ScatteringPhaseG	 	 < Widget = "FloatSlider"; Min = "0.005"; Max = "0.995"; Step = "0.001"; Precision = "3"; > = 0.7;
};

#define LocalCameraPos   (_Param0.xyz)
#define SizeScale        (_Param1.xy)

// Particle textures.
texture2D _PositionTexture : layout(t0);
sampler _PositionTexture_Sampler : layout(s0)
{
	Texture = _PositionTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D _PropertyTexture : layout(t1);
sampler _PropertyTexture_Sampler : layout(s1)
{
	Texture = _PropertyTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D _SecPropertyTexture : layout(t2);
sampler _SecPropertyTexture_Sampler : layout(s2)
{
	Texture = _PropertyTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D DiffuseTexture : layout(t3) = "White";
sampler DiffuseTexture_Sampler : layout(s3)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D tNormalMap : layout(t5) = "NormalMap";
sampler sNormalMap_Sampler : layout(s5)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = FromTexture;
};


texture2D tShadowMask : ShadowMask : layout(t6) = "White";
sampler sShadowMaskSampler : layout(s6)
{
	Texture = tShadowMask;
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tMixMap : layout(t7) = "Reflection";
sampler sMixSampler : layout(s7)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};


struct VertexAssembleCustom
{
	float4	Position	: POSITION;
	[@_NORMAL]
	float4	Normal 		: NORMAL;
	[@_TANGENT] 
	float4		Tangent	: TANGENT;
	[@_TANGENT]
	float4		Binormal: BINORMAL;
	float2	TexCoord	: TEXCOORD0;
	[@_LIGHT_MAP]
	float2	TexCoord2	: TEXCOORD1;
	float4	InstancedData : TEXCOORD3;
};

struct DrawVSOutput
{
	float4	Position	: SV_Position;
	float4	Color		: COLOR;
	float3	TexCoord	: TEXCOORDN;					// xy: texcord, z: inv emission.
};


#include "Effect_Common.fxh"
#include "Effect.fxh"
#include "quaternion.fxh"

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{	
	InitBaseColorCommon(IN,materialParamters);
	materialParamters.BaseColor.rgb *= IN.VertexColor.rgb;
	//materialParamters.BaseColor.a = 1;
	
	if (@_DISABLEDIFFUSE)
	{
		materialParamters.BaseColor = float4(1,1,1,1);
	}

	if (@_UserMixMapEnable && @_LOD_LEVEL < LOD_Level_1)
	{
		float4 rawData = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord));
		materialParamters.Roughness = 1 - rawData.r;//Roughness
		materialParamters.Metallic = rawData.g;//Metallic
	} else {
		materialParamters.Metallic = Metallic;
		materialParamters.Roughness = Roughness;
	}

	if (@_NORMAL)
		materialParamters.WorldNormal = normalize(IN.WorldNormal.xyz);
	else
		materialParamters.WorldNormal = float3(0,1,0);
		
	if(@_TANGENT && @_UserNormalMapEnable && @_LOD_LEVEL < LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalMap_Sampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		[@_DIRECTX]//always use bc5 for pc
		tangentNormal.z = 1;
		tangentNormal = (2 * tangentNormal) - 1;
		//tangentNormal.xy *= cBumpIns;
		materialParamters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * materialParamters.WorldNormal.xyz;
		materialParamters.WorldNormal = normalize(materialParamters.WorldNormal);
	}

}

#include "Effect_Shading.fxh"

float4 quaternion_from_float3(float3 value)
{
	float w = sqrt(1.0 - saturate(dot(value,value)));
	return float4(value,w);
}


DrawVSOutput DrawVS(VertexAssembleCustom IN)
{
	DrawVSOutput OUT;

	float2 ptc = IN.InstancedData.xy;
	[@_GLSL]
	ptc.y = 1 - ptc.y;
	
	float4 position = _PositionTexture.SampleLevel(_PositionTexture_Sampler, ptc, 0);
	if (position.w >= 1.0f)
	{
		OUT.Position = float4(2, 2, 2, 1);
		OUT.Color = 0;
		OUT.TexCoord = 0;
		return OUT;
	}

	float progress = position.w * 7.0;
	float base = floor(progress);
	float fraction = progress - base;

	int i = int(base);
	float4 colorSample = lerp(_ColorCurve[i], _ColorCurve[i + 1], fraction);
	float4 paramSample = lerp(_ParamCurve[i], _ParamCurve[i + 1], fraction);

	float4 property = _PropertyTexture.SampleLevel(_PropertyTexture_Sampler, ptc, 0);
	float2 size = property.xy * paramSample.xy * SizeScale;
	
	float4 secProperty = _SecPropertyTexture.SampleLevel(_SecPropertyTexture_Sampler, ptc, 0);
	float4 orientation = quaternion_from_float3(float3(property.zw,secProperty.x));
	
	{//free facing
		float3 localSpacePosition = IN.Position * 5.0f;
		float3x3 rotationMat = quaternion_to_matrix3(orientation);
		localSpacePosition = mul(rotationMat,localSpacePosition);
		localSpacePosition += position.xyz;
		
		float4 WorldPosition = float4(MulMat3x4(World, float4(localSpacePosition.xyz, 1)).xyz, 1);
		OUT.Position = MulMat4x4(ViewProj, WorldPosition);
	}

	OUT.Color = colorSample;
	OUT.TexCoord.xy = float2(IN.TexCoord.x, 1 - IN.TexCoord.y);
	OUT.TexCoord.z = 1 - paramSample.z;								// inv emission

	return OUT;
}


struct VertexShadingOutputGPUParticle
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	half4		VertexColor		: COLOR;				//	Per-vertex color.
	half3		ViewDirection	: TEXCOORDN;
	half3		WorldPosition	: TEXCOORDN;
	[@_NORMAL]
	half3		WorldNormal		: TEXCOORDN;			//	World Normal
	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;
};

void InitVertexShadingOutputGPUParticle(in out VertexShadingOutputGPUParticle data)
{
	data.HPosition = float4(0,0,0,0);
	data.TexCoord = float4(0,0,0,0);
	data.VertexColor = float4(0,0,0,0);
	data.ViewDirection = float3(0,0,0);
	data.WorldPosition = float3(0,0,0);
	if (@_NORMAL)
		data.WorldNormal = float3(0,0,0);
	if (@_TANGENT)
		data.WorldTangent = float4(0,0,0,0);
	if (@_TANGENT)
		data.WorldBinormal = float4(0,0,0,0);
}

VertexShadingOutputGPUParticle ParticleLightingVS(VertexAssembleCustom IN)
{
	VertexShadingOutputGPUParticle OUT;
	
	InitVertexShadingOutputGPUParticle(OUT);

	float2 ptc = IN.InstancedData.xy;
	[@_GLSL]
	ptc.y = 1 - ptc.y;
	
	float4 position = _PositionTexture.SampleLevel(_PositionTexture_Sampler, ptc, 0);
	if (position.w >= 1.0f)
	{
		OUT.HPosition = float4(2, 2, 2, 1);
		OUT.VertexColor = 0;
		OUT.TexCoord = 0;
		return OUT;
	}

	float progress = position.w * 7.0;
	float base = floor(progress);
	float fraction = progress - base;

	int i = int(base);
	float4 colorSample = lerp(_ColorCurve[i], _ColorCurve[i + 1], fraction);
	float4 paramSample = lerp(_ParamCurve[i], _ParamCurve[i + 1], fraction);

	float4 property = _PropertyTexture.SampleLevel(_PropertyTexture_Sampler, ptc, 0);
	float2 size = property.xy * paramSample.xy * SizeScale;
	float4 secProperty = _SecPropertyTexture.SampleLevel(_SecPropertyTexture_Sampler, ptc, 0);
	float4 orientation = quaternion_from_float3(float3(property.zw,secProperty.x));

	//free facing
	float3 localSpacePosition = IN.Position * size.x;
	float3x3 rotationMat = quaternion_to_matrix3(orientation);
	localSpacePosition = mul(rotationMat,localSpacePosition);
	localSpacePosition += position.xyz;
		
	float4 WorldPosition = float4(MulMat3x4(World, float4(localSpacePosition.xyz, 1)).xyz, 1);
	WorldPosition.y += YOffset;
	if (@_UserSinkByAlphaEnable)
		WorldPosition.y -= SinkDepth * (1.0 - colorSample.a);
	OUT.HPosition = MulMat4x4(ViewProj, float4(WorldPosition.xyz,1));
	OUT.WorldPosition = WorldPosition.xyz ;
	OUT.ViewDirection =  CameraPos.xyz - WorldPosition.xyz;

	OUT.VertexColor = colorSample;
	
	OUT.TexCoord.xy = float2(IN.TexCoord.x, IN.TexCoord.y);
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = FogTransmittance(OUT.WorldPosition.xyz - CameraPos.xyz);
	
	if(@_NORMAL)
	{
		OUT.WorldNormal = mul(rotationMat,(IN.Normal.xyz * 2) - 1);
	}
	if (@_TANGENT)
	{
		OUT.WorldTangent.xyz = mul(rotationMat,IN.Tangent.xyz);
		OUT.WorldBinormal.xyz = mul(rotationMat,IN.Binormal.xyz);
	}
	
	return OUT;
}


float4 DrawPS(DrawVSOutput IN) : SV_Target
{
	float4 OUT;
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	OUT *= IN.Color * ModulateColor * InstanceColor;
	OUT.rgb *= OUT.a;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);
	
	return OUT;
}

float4 GPUMeshParticleShading(VertexShadingOutputGPUParticle IN):SV_Target
{

	VertexShadingOutput temp;
	InitVertexShadingOutput(temp);
	temp.HPosition = IN.HPosition;	//	Position in homogeneous space. For output to rasterizer.
	temp.TexCoord = IN.TexCoord;
	if (@_NORMAL)
		temp.WorldNormal = IN.WorldNormal;
	temp.ViewDirection = normalize(IN.ViewDirection);
	temp.VertexColor = IN.VertexColor;
	temp.WorldPosition = IN.WorldPosition;
	if (@_TANGENT)
		temp.WorldTangent = IN.WorldBinormal;
	if (@_TANGENT)
		temp.WorldBinormal = IN.WorldBinormal;
	float3 debugNormal = float3(1,0,0);
	
	if (@_NORMAL)
		debugNormal = IN.WorldNormal;
	
	debugNormal = normalize(debugNormal);
	float4 debugOut;
	debugOut.xyz = debugNormal * 0.5 + 0.5;
	debugOut.a = 1;
	//return debugOut;
	float4 OUT = EffectLightingPixelShading(temp,float4(0,0,0,0));
	return OUT;
}

technique Shading
{
	VertexShader = ParticleLightingVS();
	PixelShader = GPUMeshParticleShading();
}

//////////////////////////////////GBuffer////////////////////////////////////////
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	
	float4 GBufferB 	: SV_Target1;	
	float4 GBufferC 	: SV_Target2;	
	float4 GBufferD 	: SV_Target3;	
	float4 GBufferE 	: SV_Target4;	
	float4 GBufferF  	: SV_Target5;	
	float4 GBufferG 	: SV_Target6;	
};

GBufferOUTPUT GBufferGenPixelShading(VertexShadingOutputGPUParticle IN)
{

	GBufferOUTPUT OUT;
	GBufferData GBuffer;
	
	VertexShadingOutput temp;
	InitVertexShadingOutput(temp);
	temp.HPosition = IN.HPosition;	//	Position in homogeneous space. For output to rasterizer.
	temp.TexCoord = IN.TexCoord;
	if (@_NORMAL)
		temp.WorldNormal = IN.WorldNormal;
	temp.ViewDirection = normalize(IN.ViewDirection);
	temp.VertexColor = IN.VertexColor;
	temp.WorldPosition = IN.WorldPosition;
	if (@_TANGENT)
		temp.WorldTangent = IN.WorldBinormal;
	if (@_TANGENT)
		temp.WorldBinormal = IN.WorldBinormal;
		
	// 1. Init Shading Parameters Struct	
    EffectShadingParameters materialParamters;
    InitEffectShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(temp,materialParamters);

	// 2.1 Value Check
	materialParamters.Roughness = max(Roughness,MIN_ROUGHNESS);
	
	MaterialShadingParameters defferedMatParamters;
	InitShadingParameters(defferedMatParamters);
	defferedMatParamters.BaseColor = materialParamters.BaseColor.rgb;
	defferedMatParamters.Metallic = materialParamters.Metallic;
	defferedMatParamters.Roughness = materialParamters.Roughness;
	defferedMatParamters.Specular = materialParamters.Specular;
	defferedMatParamters.SpecularColor = materialParamters.BaseColor.rgb * (materialParamters.Metallic);
	defferedMatParamters.DiffuseColor = materialParamters.BaseColor.rgb * (1 - materialParamters.Metallic);
	defferedMatParamters.WorldNormal = materialParamters.WorldNormal;
	defferedMatParamters.Alpha = materialParamters.BaseColor.a;
	defferedMatParamters.AO = 1;
	
	if (@_DISABLEDIFFUSE)
	{
		defferedMatParamters.BaseColor = float3(1,1,1);
		defferedMatParamters.Alpha = 1;
	}

	// GI Lighting For Particle
	FetchPrecomputedLighingParticle(defferedMatParamters);

	SetGBufferForShadingModel(GBuffer, defferedMatParamters);
	EncodeGBufferData(GBuffer, OUT.GBufferA, OUT.GBufferB, OUT.GBufferC, OUT.GBufferD, OUT.GBufferE, OUT.GBufferF);

	/*
		HPosition on Metal
		will return {x,y,z,1/w} on PS Stage when input is {x,y,z,w} on VS 
		using TexCoord.z instead of HPosition.w
	*/
	float linearZ = IN.TexCoord.z * CameraInfo.z;

	OUT.GBufferG = EncodeDepth(linearZ);
	OUT.GBufferA.w = EncodeDepth(linearZ);
	
	float clipValue = 1.0;
	if (@_PARTICLECLIPPING)
	{
		float alpha = defferedMatParamters.Alpha;
		//clip(alpha - ALPHA_REF);
		clipValue = alpha - ALPHA_REF;
	}
	if (@_PARTICLECLIPPING)
		clip(clipValue);
	return OUT;
}

technique GBufferGen
{
	VertexShader = ParticleLightingVS();
	PixelShader = GBufferGenPixelShading();
}

struct HeightmapVertexShadingOutput
{
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		SceneDepthTC	: TEXCOORDN;			//	Scene Depth
};

//viewProjVS is light view now
HeightmapVertexShadingOutput HeightmapGenVS(VertexAssembleCustom IN)
{
	HeightmapVertexShadingOutput OUT;

	float2 ptc = IN.InstancedData.xy;
	[@_GLSL]
	ptc.y = 1 - ptc.y;
	
	float4 position = _PositionTexture.SampleLevel(_PositionTexture_Sampler, ptc, 0);
	if (position.w >= 1.0f)
	{
		OUT.HPosition = float4(2, 2, 2, 1);
		//SceneDepthTC
		OUT.SceneDepthTC.xy = OUT.HPosition.zw;
		OUT.SceneDepthTC.zw = IN.TexCoord.xy;
		return OUT;
	}

	float progress = position.w * 7.0;
	float base = floor(progress);
	float fraction = progress - base;

	int i = int(base);
	float4 colorSample = lerp(_ColorCurve[i], _ColorCurve[i + 1], fraction);
	float4 paramSample = lerp(_ParamCurve[i], _ParamCurve[i + 1], fraction);

	float4 property = _PropertyTexture.SampleLevel(_PropertyTexture_Sampler, ptc, 0);
	float2 size = property.xy * paramSample.xy * SizeScale;
	
	float4 secProperty = _SecPropertyTexture.SampleLevel(_SecPropertyTexture_Sampler, ptc, 0);
	float4 orientation = quaternion_from_float3(float3(property.zw,secProperty.x));
	
	{//free facing
		float3 localSpacePosition = IN.Position * 5.0f;
		float3x3 rotationMat = quaternion_to_matrix3(orientation);
		localSpacePosition = mul(rotationMat,localSpacePosition);
		localSpacePosition += position.xyz;
		
		float4 WorldPosition = float4(MulMat3x4(World, float4(localSpacePosition.xyz, 1)).xyz, 1);
		OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	}

	OUT.SceneDepthTC.xy = OUT.HPosition.zw;
	OUT.SceneDepthTC.zw = IN.TexCoord.xy;

	return OUT;
}

[target("R32F")]
float4 HeightmapGenPS(in HeightmapVertexShadingOutput IN) : SV_Target
{
#ifndef NO_SHADOW_CLIPPING
	float clipValue = 1.0;
	if (@_PARTICLECLIPPING)
	{
	
		half4 baseColor = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, IN.SceneDepthTC.zw);
		float alpha = baseColor.a;
		//clip(alpha - ALPHA_REF);
		clipValue = alpha - ALPHA_REF;
	}
	if (@_DISABLEDIFFUSE)
	{
		clipValue = 1;
	}
	if (@_PARTICLECLIPPING)
		clip(clipValue);
#endif
	float depth = IN.SceneDepthTC.x / IN.SceneDepthTC.y;
	float4 rt = EncodeDepth(depth);
	return rt;
}

technique HeightmapGen
{ 
	VertexShader = HeightmapGenVS();
	PixelShader = HeightmapGenPS();
}
// Clip
float4 ClipPixelShading(VertexShadingOutputGPUParticle IN) : SV_Target
{
	VertexShadingOutput temp;
	InitVertexShadingOutput(temp);
	temp.HPosition = IN.HPosition;	//	Position in homogeneous space. For output to rasterizer.
	temp.TexCoord = IN.TexCoord;
	if (@_NORMAL)
		temp.WorldNormal = IN.WorldNormal;
	temp.ViewDirection = normalize(IN.ViewDirection);
	temp.VertexColor = IN.VertexColor;
	temp.WorldPosition = IN.WorldPosition;
	if (@_TANGENT)
		temp.WorldTangent = IN.WorldBinormal;
	if (@_TANGENT)
		temp.WorldBinormal = IN.WorldBinormal;
		
		// 1. Init Shading Parameters Struct	
    EffectShadingParameters materialParamters;
    InitEffectShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(temp,materialParamters);
	
	if (@_PARTICLECLIPPING)
	{	
		half a = materialParamters.BaseColor.a;
		clip(a - ALPHA_REF);
	}

	return float4(0, 0, 0, 0);
}

technique Clip
{	VertexShader = ParticleLightingVS();
	PixelShader = ClipPixelShading();
}


// float4 EffectDebugShading(in VertexShadingOutput IN) : SV_Target
// {
// 	float4 OUT = float4(0.1, 0.05, 0.05, 0.1);
// 	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
// 	return OUT;
// }

#ifndef USE_DEBUG_TECHNIQUE
#define USE_DEBUG_TECHNIQUE 1
#endif

#if USE_DEBUG_TECHNIQUE
technique Debug
{
	VertexShader = ParticleLightingVS();
	PixelShader = GPUMeshParticleShading();
}
#endif


// void FetchPrecomputedLighingParticle(in out MaterialShadingParameters MaterialParameters)
// {
// 	//Modified From PBR_Common:FetchPrecomputedLighing()
// 	//No LightMap For Particle
// 	//Only Calc Dymnamic GI Lighting as All PBR Material On Deferred
// 	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
// 	MaterialParameters.GILighting *=  MaterialParameters.AO;
// }
