//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GPUParticle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	GPU particle shader.
#pragma shader draw

@_VELOCITY
{
	Name = Velocity;
	Description = "Facing Vecloity";
	Visible = false;
}

@_UserFrameSequence
{
	Name = UserFrameSequence;
	Description = "Facing Vecloity";
	Visible = true;
}

@_UserLightingShading
{
	Name = UserLightingShading;
	Description = "Lighting Shading";
	Visiable = true;
	Default = false;
}

#pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

// constants
cbuffer Batch
{
	float4x3 World 				: World;
	float4   InstanceColor		: Color;
	float4 	cPointCloud[6]		: PointCloudData;
};

cbuffer Shader
{
	float4 ModulateColor		< Widget = "Color4"; >  = { 1, 1, 1, 1 };
	float HDRScale				= 2.0;
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };

	float  FrameSequenceSpeed		= 1.0;
	float2 FrameSequenceTileSize	= { 1 , 1 };
	float  Thickness			 < Widget = "FloatSlider"; Min = "0.2"; Max = "1.0"; Step = "0.01"; Precision = "3"; > = 1.0;
	float  AlphaScale			  < Widget = "FloatSlider"; Min = "0.0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 1;
	float  Metallic				  < Widget = "FloatSlider"; Min = "0.001"; Max = "0.999"; Step = "0.01"; Precision = "3"; > = 0;
	float  Roughness			 < Widget = "FloatSlider"; Min = "0.08"; Max = "0.92"; Step = "0.01"; Precision = "3"; > = 0.9;
	float  NormalScale			 < Widget = "FloatSlider"; Min = "-5.0"; Max = "5.0"; Step = "0.01"; Precision = "3"; > = 1;
	float  Scattering			 < Widget = "FloatSlider"; Min = "0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 0.0;
	float  ScatteringPhaseG	 	 < Widget = "FloatSlider"; Min = "0.005"; Max = "0.995"; Step = "0.001"; Precision = "3"; > = 0.7;	

	float3 _Param0				;
	float4 _Param1				;
	float4 _ColorCurve[8]		;
	float4 _ParamCurve[8]		;	// xy : size, z : emission
};

#define LocalCameraPos   (_Param0.xyz)
#define SizeScale        (_Param1.xy)

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

// GPU Particle textures.
texture2D _PositionTexture : layout(t0);
sampler _PositionTexture_Sampler : layout(s0)
{
	Texture = _PositionTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D _PropertyTexture : layout(t1);
sampler _PropertyTexture_Sampler : layout(s1)
{
	Texture = _PropertyTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D _SecPropertyTexture : layout(t2);
sampler _SecPropertyTexture_Sampler : layout(s2)
{
	Texture = _PropertyTexture;
	Filter = Point;
	Address = Clamp;
};

// Particle Usual Texture
texture2D DiffuseTexture : layout(t3) = "White";
sampler DiffuseTexture_Sampler : layout(s3)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D tNormalMap : layout(t4) = "Normal";
sampler sNormalMap_Sampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
};


struct VertexAssemble
{
	float2	TexCoord	: TEXCOORDN;
	float4	Position	: POSITION;
};

struct DrawVSOutput
{
	float4	Position	: SV_Position;
	float4	Color		: COLOR;
	float4	TexCoord	: TEXCOORDN;					// xy: texcord, z: inv emission. w: Life
	float4 ViewDirection: TEXCOORDN;
	float4 WorldPosition: TEXCOORDN;
};

void InitFogLightingSpec(/*in float2 stc,*/in float3 ViewDirection,in float FogT, in out float4 OUT)
{
	//issue : IN.DiffLighting.w is different to VolumetricFog-off
// 	if (@_FOG && @_VolumetricFog)
// 	{
// 		// float2 spos = IN.HPosition.xy;
// 		// [@_D3D9]
// 		// spos += 0.5;
// 		// float2 stc = spos * ScreenInfo.zw;
// 		float depth = IN.DiffLighting.w;
// 		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
// 		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
// 	}
	if (@_FOG && !@_VolumetricFog)
	{
		float4 fogFinalColor = float4(0,0,0,0);
		FogLightingAndScatter(FogT, fogFinalColor.xyz, ViewDirection, SunDirection);
		OUT.rgb = lerp(fogFinalColor.xyz * OUT.a, OUT.rgb, FogT);
	}
}

DrawVSOutput DrawVS(VertexAssemble IN)
{
	DrawVSOutput OUT;

	float2 ptc = IN.Position.xy;
	[@_GLSL]
	ptc.y = 1 - ptc.y;
	
	float4 position = _PositionTexture.SampleLevel(_PositionTexture_Sampler, ptc, 0);
	if (position.w >= 1.0f)
	{
		OUT.Position = float4(2, 2, 2, 1);
		OUT.Color = 0;
		OUT.TexCoord = 0;
		return OUT;
	}

	float progress = position.w * 7.0;
	float base = floor(progress);
	float fraction = progress - base;

	int i = int(base);
	float4 colorSample = lerp(_ColorCurve[i], _ColorCurve[i + 1], fraction);
	float4 paramSample = lerp(_ParamCurve[i], _ParamCurve[i + 1], fraction);

	float4 property = _PropertyTexture.SampleLevel(_PropertyTexture_Sampler, ptc, 0);
	float2 size = property.xy * paramSample.xy * SizeScale;
	
	float4 secProperty = _SecPropertyTexture.SampleLevel(_SecPropertyTexture_Sampler, ptc, 0);
	float3 rotation = float3(property.zw,secProperty.x);
	float4 WorldPosition = float4(0,0,0,0);
	if (@_VELOCITY)
	{
		// Facing velocity
		float3 velocityDir = abs(property.zzw) - 2;
		float ysign = property.z < 0 ? -1 : 1;
		velocityDir.y = sqrt(1 - dot(velocityDir.xz, velocityDir.xz)) * ysign;
		float3 cameraDir = LocalCameraPos - position.xyz;
		float3 right = normalize(cross(cameraDir, velocityDir)) * size.x;
		float3 up = velocityDir * size.y;

		float2 dc = IN.TexCoord * 2 - 1;
		position.xyz += dc.x * right + dc.y * up;
		WorldPosition = float4(MulMat3x4(World, float4(position.xyz, 1)).xyz, 1);
		OUT.Position = MulMat4x4(ViewProj, WorldPosition);
	}
	else
	{
		// Facing camera
		WorldPosition = float4(MulMat3x4(World, float4(position.xyz, 1)).xyz, 1);
		OUT.Position = MulMat4x4(ViewProj, WorldPosition);

		// rotation
		float rot = property.z + position.w * property.w;
		float2 right;
		sincos(rot, right.y, right.x);
		float2 up = float2(-right.y, right.x);

		float2 dc = IN.TexCoord * 2 - 1;
		float2 offset = dc.x * right + dc.y * up;
		OUT.Position.xy += offset * size;
	}

	OUT.Color = colorSample;
	OUT.Color.xyz *= HDRScale;
	OUT.TexCoord.xy = float2(IN.TexCoord.x, 1 - IN.TexCoord.y);
	OUT.TexCoord.z = 1 - paramSample.z;								// inv emission
	OUT.TexCoord.w = position.w;									// Particle Life
	OUT.ViewDirection.xyz = normalize( CameraPos.xyz - WorldPosition.xyz );
	OUT.ViewDirection.w = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);
	OUT.WorldPosition = WorldPosition;
	return OUT;
}

float4 DrawPS(DrawVSOutput IN)
{
	float4 OUT = float4(0,0,0,0);
	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;

	if(@_UserFrameSequence)
	{
		float CurrentIndex = floor(FrameSequenceTileSize.x * FrameSequenceTileSize.y * IN.TexCoord.w * FrameSequenceSpeed);
		float2 AnimatedUV = float2(0,0);
		AnimatedUV.x = fmod(CurrentIndex , FrameSequenceTileSize.x);
		AnimatedUV.y = floor(CurrentIndex / FrameSequenceTileSize.x);
		diffuseTC = (diffuseTC + AnimatedUV) / FrameSequenceTileSize;
	}

	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	OUT *= IN.Color * ModulateColor * InstanceColor;
	OUT.rgb *= OUT.a;
	OUT.a *= IN.TexCoord.z;

    EncodeHDRParticle(OUT);

	InitFogLightingSpec(IN.ViewDirection.xyz,IN.ViewDirection.w,OUT);

	return OUT;
}

#define __FILE_PARTICLE_COMMON_FXH__

struct ParticleShadingParameters
{
    half4 BaseColor;
    half Metallic;
    half Roughness;
    half Specular;
    half3 WorldNormal;
	half Depth;
	half3 Emission;
};

struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	[@_NORMAL]
	half3		Normal 			: NORMAL;
	half4		VertexColor		: COLOR;				//	Per-vertex color.
	half3		ViewDirection	: TEXCOORDN;
	float3		WorldPosition	: TEXCOORDN;
};

void InitParticleShadingParameters(in out ParticleShadingParameters MaterialParamters)
{
    MaterialParamters.BaseColor = half4(1,1,1,1);
    MaterialParamters.Metallic = 0;
    MaterialParamters.Roughness = 1;
    MaterialParamters.Specular = 0.5;
    MaterialParamters.WorldNormal = half3(1,0,0);
	MaterialParamters.Depth = 0;
	MaterialParamters.Emission = half3(0,0,0);
    // MaterialParamters.Alpha = 1;
}

void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{
	OUT.HPosition = float4(0,0,0,0);
	OUT.TexCoord = float4(0,0,0,0);
	OUT.VertexColor = half4(0,0,0,0);
	OUT.ViewDirection = half3(0,0,0);
	OUT.WorldPosition = half3(0,0,0);
	if (@_NORMAL)
		OUT.Normal = half3(0,0,0);
}

float3 SafeNormalizeF3(float3 vin,float3 safeValue)
{
	float len = length(vin);
	return (len < 0.01) ? safeValue : vin / len;
}

float2 SafeNormalizeF2(float2 vin,float2 safeValue)
{
	float len = length(vin);
	return (len < 0.01) ? safeValue : vin / len;
}

void GetNormalD(in VertexShadingOutput IN,out float3 Normal,out float3 BiNormal,out float3 Tangent)
{
	/*
		Check for NaN, when using WorldPosition as source to calc Normal
		due to Precision ddx() ddy() may return float3(0,0,0) that will cause NaN
	*/
	float3 P = IN.WorldPosition;
	float3 dx = SafeNormalizeF3(ddx(P),float3(1,1,0));
	float3 dy = SafeNormalizeF3(ddy(P),float3(0,-1,-1));
	float2 Edge1uv = SafeNormalizeF2(ddx(IN.TexCoord.xy),float2(1,0));
	float2 Edge2uv = SafeNormalizeF2(ddy(IN.TexCoord.xy),float2(0,1));
	float3 N = float3(0,0,0);
	if (@_NORMAL)
	{
		N = IN.Normal;
	} else {
		N = SafeNormalizeF3(cross(dy, dx),float3(0,-1,0));
	}

	float cp = Edge1uv.y * Edge2uv.x - Edge1uv.x * Edge2uv.y;
    float mul =  cp == 0 ? 1 : 1.0f / cp;

    Tangent = SafeNormalizeF3((dx * -Edge2uv.y + dy * Edge1uv.y) * mul,float3(1,0,0));
	BiNormal = SafeNormalizeF3((dx * -Edge2uv.x + dy * Edge1uv.x) * mul,float3(0,0,1));

	Normal = N;	
}

// void GetNormalDD(in VertexShadingOutput IN,out float3 Normal,out float3 BiNormal,out float3 Tangent)
// {
// 	/*
// 		Check for NaN, when using WorldPosition as source to calc Normal
// 		due to Precision ddx() ddy() may return float3(0,0,0) that will cause NaN
// 	*/
// 	float3 P = IN.WorldPosition;
// 	float3 dx = normalize(ddx(P));
// 	float3 dy = normalize(ddy(P));
// 	float2 Edge1uv = normalize(ddx(IN.TexCoord.xy));
// 	float2 Edge2uv = normalize(ddy(IN.TexCoord.xy));
// 	float3 N = float3(0,0,0);
// 	if (@_NORMAL)
// 	{
// 		N = normalize(IN.Normal);
// 	} else {
// 		//when cross(dx,dy) == 0,normalize will return NaN
// 		N = normalize(cross(dy, dx));
// 		// N = any(isnan(N)) ? float3(0,-1,0) : N;
// 	}

// 	// From ParticleSystem's Distortion
// 	// if (@_GLSL || @_SPIRV)
// 	// {
// 	// 	// ddx/ddy is incorrect on some android devices.
// 	// 	half len = length(N);
// 	// 	N = (len == 0) ? half3(0, 1, 0) : N / len;
// 	// }
// 	// else
// 	// {
// 	// N = normalize(N);
// 	// }

// 	float cp = Edge1uv.y * Edge2uv.x - Edge1uv.x * Edge2uv.y;
//     float mul = (bool)(cp == 0) ? 1.0 : 1.0f / cp;

//     Tangent = normalize((dx * -Edge2uv.y + dy * Edge1uv.y) * mul);
// 	// Tangent = length(Tangent) < 0.01 ? float3(1,0,0) : normalize(Tangent);
// 	BiNormal = normalize((dx * -Edge2uv.x + dy * Edge1uv.x) * mul);
// 	// BiNormal = length(BiNormal) < 0.01 ? float3(0,0,1) : normalize(BiNormal);

	
// 	Normal = N;	
// }

void InitWorldNormal(in VertexShadingOutput IN,in out ParticleShadingParameters materialParamters)
{
	//Get Normal using ddx() / ddy()
	float3 Normal = half3(0,1,0);
	float3 BiNormal = half3(1,0,0);
	float3 Tangent = half3(0,0,1);
	GetNormalD(IN,Normal,BiNormal,Tangent);

	float4 rst = tNormalMap.SampleBias(sNormalMap_Sampler, IN.TexCoord.xy, 0);
	float3 tangentNormal = rst.xyz;
	[@_DIRECTX]//always use bc5 for pc
	tangentNormal.z = 1;
	tangentNormal = (2 * tangentNormal) - 1;
	tangentNormal.xy *= NormalScale;
	
	Normal = normalize(Tangent * tangentNormal.x + BiNormal * tangentNormal.y + Normal * tangentNormal.z);

	materialParamters.WorldNormal = Normal * (NormalScale > 0 ? 1 : -1);	
	
}

void InitBaseColorCommon(in VertexShadingOutput IN,in out ParticleShadingParameters materialParamters)
{
	// BaseColor
	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	// float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + GameTime * TwistUVScolling.zw;
	// float2 twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;
	// diffuseTC += twist * TwistStrength;
	float4 baseColor = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	baseColor.rgb = DecodeGamma(baseColor.rgb);
	
	materialParamters.BaseColor = baseColor;	
}

void InitFogLightingParticle(/*in float2 stc,*/in VertexShadingOutput IN, in out float4 OUT)
{
	//issue : IN.DiffLighting.w is different to VolumetricFog-off
// 	if (@_FOG && @_VolumetricFog)
// 	{
// 		// float2 spos = IN.HPosition.xy;
// 		// [@_D3D9]
// 		// spos += 0.5;
// 		// float2 stc = spos * ScreenInfo.zw;
// 		float depth = IN.DiffLighting.w;
// 		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
// 		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
// 	}
	if (@_FOG && !@_VolumetricFog)
	{
		float4 fogFinalColor = float4(0,0,0,0);
		FogLightingAndScatter(IN.TexCoord.w, fogFinalColor.xyz, IN.ViewDirection, SunDirection);
		OUT.rgb = lerp(fogFinalColor.xyz * OUT.a, OUT.rgb, IN.TexCoord.w);
	}
}

// half henyeyGreenstein(half lightDotEye, half g)
// {
// 	half g2 = g * g;

// 	//fast HG Function <- but not right
// 	return ((1.0f - g2) / (1.0f + g2 - 2.0f * g * lightDotEye)) * 0.099577f;
// }

#include "Particle_Shading.fxh"

float4 DrawPSLighting(DrawVSOutput IN)
{
	ParticleShadingParameters materialParamters;
	InitParticleShadingParameters(materialParamters);

	VertexShadingOutput VIN;
	InitVertexShadingOutput(VIN);
	VIN.TexCoord.w = IN.ViewDirection.w;
	VIN.ViewDirection = IN.ViewDirection.xyz;
	VIN.WorldPosition = IN.WorldPosition.xyz;
	VIN.TexCoord = IN.TexCoord;
	
	InitBaseColorCommon(VIN,materialParamters);
	materialParamters.BaseColor *= ModulateColor;

	InitWorldNormal(VIN,materialParamters);

	return LightingShading(VIN,materialParamters);
}

float4 GPUParticlePixelShading(DrawVSOutput IN) : SV_Target
{
	if(@_UserLightingShading == false)
		return DrawPS(IN);
	else
		return DrawPSLighting(IN);
	return float4(0,0,0,0);
}

technique Shading
{
	VertexShader = DrawVS();
	PixelShader = GPUParticlePixelShading();
}
