//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<PostProcess.fx>
///	@path	~/System/Shader/
///	@date	2013/04/24
///	@brief	Post process utils.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

// Source target.
texture2D tBaseMap : layout(t0) = "Transparent";
sampler sBaseMap_Sampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tMakeup0 : layout(t1) = "Gray";
sampler sMakeup0Sampler : layout(s1)
{
	Texture = tMakeup0;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tMakeup1 : layout(t2) = "Gray";
sampler sMakeup1Sampler : layout(s2)
{
	Texture = tMakeup1;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tMakeup2 : layout(t3) = "Gray";
sampler sMakeup2Sampler : layout(s3)
{
	Texture = tMakeup2;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tMakeup3 : layout(t4) = "Gray";
sampler sMakeup3Sampler : layout(s4)
{
	Texture = tMakeup3;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tMakeup4 : layout(t5) = "Gray";
sampler sMakeup4Sampler : layout(s5)
{
	Texture = tMakeup4;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

// texture2D tMakeup5 : layout(t6) = "Gray";
// sampler sMakeup5Sampler : layout(s6)
// {
// 	Texture = tMakeup5;
// 	Filter = Linear;
// 	Address = Clamp;
// 	SRGB = false;
// };

cbuffer Shader
{
	float4 cUVTransform0		 = 	{0,0,1,1};
	float4 cDyeing0				 = {1,0,0,0};
	float4 cUVTransform1		 = {0,0,1,1};
	float4 cDyeing1				 = {1,0,0,0};
	float4 cUVTransform2		 = {0,0,1,1};
	float4 cDyeing2				 = {1,0,0,0};
	float4 cUVTransform3		 = {0,0,1,1};
	float4 cDyeing3				 = {1,0,0,0};
	float4 cUVTransform4		 = {0,0,1,1};
	float4 cDyeing4				 = {1,0,0,0};
};

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 FaceMakeupPS(in VertexScreenOutput IN) : SV_Target
{
	float4 OUT = float4(0,0,0,0);

	// vanila gaussian blur
	OUT = tBaseMap.Sample(sBaseMap_Sampler, IN.TC.xy);
	// OUT.rgb = DecodeGamma(OUT.rgb);
	half3 diff;
	float2 uv = IN.TC * cUVTransform0.xy + cUVTransform0.zw;;
	half4 makeup = tMakeup0.Sample(sMakeup0Sampler, uv);
	uv = step(abs(uv - 0.5), float2(0.5,0.5));
	makeup = makeup * (127.5 / 128.0);
	makeup.rgb = DyeingYIQ(makeup.rgb, half3(cDyeing0.xy, 1));
	OUT.rgb = saturate(OUT.rgb + (makeup.rgb * 2 - 1) * (cDyeing0.z * makeup.a * uv.x * uv.y));

	uv = IN.TC * cUVTransform1.xy + cUVTransform1.zw;
	makeup = tMakeup1.Sample(sMakeup1Sampler, uv);
	uv = step(abs(uv - 0.5), float2(0.5,0.5));
	makeup = makeup * (127.5 / 128.0);
	makeup.rgb = DyeingYIQ(makeup.rgb, half3(cDyeing1.xy, 1));
	OUT.rgb = saturate(OUT.rgb + (makeup.rgb * 2 - 1) * (cDyeing1.z * makeup.a * uv.x * uv.y));
	
	uv = IN.TC * cUVTransform2.xy + cUVTransform2.zw;
	makeup = tMakeup2.Sample(sMakeup2Sampler, uv);
	uv = step(abs(uv - 0.5), float2(0.5,0.5));
	makeup = makeup * (127.5 / 128.0);
	makeup.rgb = DyeingYIQ(makeup.rgb, half3(cDyeing2.xy, 1));
	OUT.rgb = saturate(OUT.rgb + (makeup.rgb * 2 - 1) * (cDyeing2.z * makeup.a * uv.x * uv.y));
	
	uv = IN.TC * cUVTransform3.xy + cUVTransform3.zw;
	makeup = tMakeup3.Sample(sMakeup3Sampler, uv);
	uv = step(abs(uv - 0.5), float2(0.5,0.5));
	makeup = makeup * (127.5 / 128.0);
	makeup.rgb = DyeingYIQ(makeup.rgb, half3(cDyeing3.xy, 1));
	OUT.rgb = saturate(OUT.rgb + (makeup.rgb * 2 - 1) * (cDyeing3.z * makeup.a * uv.x * uv.y));

	uv = IN.TC * cUVTransform4.xy + cUVTransform4.zw;
	makeup = tMakeup4.Sample(sMakeup4Sampler, uv);
	uv = step(abs(uv - 0.5), float2(0.5,0.5));
	makeup = makeup * (127.5 / 128.0);
	makeup.rgb = DyeingYIQ(makeup.rgb, half3(cDyeing4.xy, 1));
	OUT.rgb = saturate(OUT.rgb + (makeup.rgb * 2 - 1) * (cDyeing4.z * makeup.a * uv.x * uv.y));

	// OUT.rgb = EncodeGamma(OUT.rgb);
	return OUT;
}

technique GenTexture
{
	VertexShader = ScreenVS();
	PixelShader = FaceMakeupPS();
}
