// {HEX} GI
#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "DeferredShadingCommon.fxh"

@_PreGIEnable
{ 
	Name = PreGIEnable;
	Description = "Do we have Pre GI?";
	Default = false;
};
@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};
@_BlurX
{
	Name = BlurX;
	Default = false;
	Visible = false;
}
@_EnableVPL
{
	Name = EnableVPL;
	Default = false;
	Visible = false;
}
@_EnableSSGI
{
	Name = EnableSSGI;
	Description = "Enable SSGI";
	Default = false;
};
@_EnableVXGI
{
	Name = EnableVXGI;
	Description = "Enable VXGI";
	Default = false;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX				: WBasisX ;
	float4 vWBasisY				: WBasisY ;
	float4 vWBasisZ				: WBasisZ ;
	float4 cRTSize				;			//xy-gi rt size, zw-pingpong rt size
	float  cFrameIndex			;
	float2 cAOIntensity			 = {1, 1}; // x:aoBiasStatic, y:aoBiasDynamic
	float4 cRTAndHZUVScale		 = {1, 1, 1, 1}; // RTWidth/2/HZWidth
	float4x4 cProjTex			;
	float4x3 cWorldToView		: View ;
	float4x4 cShadowViewProjTex	;
	float4x4 cShadowProjTex 	;
	float4x3 cShadowWorldToView ;
	float  cVPLIntensity		;
};

texture2D tDepthBufferLinear : HierarchicalZ : layout(t0);
sampler sDepthBufferSamplerLinear : layout(s0)
{
	Texture = tDepthBufferLinear;
	Filter = Point;
	Address = Clamp;
};
texture2D tGBufferE : layout(t1);
sampler sGBufferE : layout(s1)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};
texture2D tSceneColorMap : layout(t2);
sampler sSceneColorSampler : layout(s2)
{
	Texture = tSceneColorMap;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};
texture2D tSceneDepthBuffer : SceneDepth : layout(t3);
sampler sSceneDepthBufferSampler : layout(s3)
{
	Texture = tSceneDepthBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D tVPLColorMap : layout(t4);
sampler sVPLColorMap : layout(s4)
{
	Texture = tVPLColorMap;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};
texture2D tVPLNormalMap : layout(t5);
sampler sVPLNormalMap : layout(s5)
{
	Texture = tVPLNormalMap;
	Filter = Point;
	Address = Clamp;
};
texture2D tVPLDepthMap : layout(t6);
sampler sVPLDepthMap : layout(s6)
{
	Texture = tVPLDepthMap;
	Filter = Point;
	Address = Clamp;
};
texture2D tVXGI : layout(t7) = "Black";
sampler sVXGI : layout(s7)
{
	Texture = tVXGI;
	Filter = Linear;
	Address = Clamp;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position		: POSITION; // in object space
	float2 BaseTC		: TEXCOORD0;
};

struct v2f
{
	float4 HPosition	: SV_Position;
	float2 TC			: TEXCOORD0;
};

#define GetLuminanceGI(color) dot(color.rgb, half3(0.3, 0.59, 0.11))
// [ Duff et al. 2017, "Building an Orthonormal Basis, Revisited" ]
/*float3x3 GetTangentBasis( float3 TangentY )
{
	const float Sign = 2 * step(0, TangentY.y) - 1;
	const float a = -1 / ( Sign + TangentY.y );
	const float b = TangentY.x * TangentY.z * a;
	
	float3 TangentX = { 1 + Sign * a * TangentY.x * TangentY.x, Sign * b, -Sign * TangentY.x };
	float3 TangentZ = { b,  Sign + a * TangentY.z * TangentY.z, -TangentY.z };

	return float3x3( TangentX, TangentY, TangentZ);
}*/
float3x3 GetTangentBasis( float3 TangentZ )
{
	const float Sign = TangentZ.z >= 0 ? 1 : -1;
	const float a = -rcp( Sign + TangentZ.z );
	const float b = TangentZ.x * TangentZ.y * a;
	
	float3 TangentX = { 1 + Sign * a * TangentZ.x * TangentZ.x, Sign * b, -Sign * TangentZ.x };
	float3 TangentY = { b,  Sign + a * TangentZ.y * TangentZ.y, -TangentZ.y };

	return float3x3(TangentX.x, TangentX.y, TangentX.z, TangentY.x, TangentY.y, TangentY.z, TangentZ.x, TangentZ.y, TangentZ.z);
}
float2 UniformSampleDiskConcentric( float2 E )
{
	float2 p = 2 * E - 1;
	float Radius;
	float Phi;
	if( abs( p.x ) > abs( p.y ) )
	{
		Radius = p.x;
		Phi = (PI/4) * (p.y / p.x);
	}
	else
	{
		Radius = p.y;
		Phi = (PI/2) - (PI/4) * (p.x / p.y);
	}
	return float2( Radius * cos( Phi ), Radius * sin( Phi ) );
}
// function
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}

uint3 Rand3DPCG16(int3 p)
{
	// taking a signed int then reinterpreting as unsigned gives good behavior for negatives
	uint3 v = uint3(p);
	v = v * 1664525u + 1013904223u;
	v.x += v.y*v.z;
	v.y += v.z*v.x;
	v.z += v.x*v.y;
	v.x += v.y*v.z;
	v.y += v.z*v.x;
	v.z += v.x*v.y;

	// only top 16 bits are well shuffled
	return v >> 16u;
}
uint2 ComputeRandomSeed(uint2 PixelPosition)
{
	return Rand3DPCG16(int3(PixelPosition, cFrameIndex)).xy;
}
/** Reverses all the 32 bits. */
uint ReverseBits32( uint bits )
{
	bits = ( bits << 16) | ( bits >> 16);
	bits = ( (bits & uint(16711935)) << 8 ) | ( (bits & ‭uint(4278255360‬)) >> 8 );
	bits = ( (bits & ‭uint(252645135‬)) << 4 ) | ( (bits & ‭uint(4042322160‬)) >> 4 );
	bits = ( (bits & ‭uint(858993459‬)) << 2 ) | ( (bits & ‭uint(3435973836‬)) >> 2 );
	bits = ( (bits & ‭uint(1431655765‬)) << 1 ) | ( (bits & ‭uint(2863311530‬)) >> 1 );
	return bits;
}
float2 Hammersley16( uint Index, uint NumSamples, uint2 Random )
{
	float E1 = frac( (float)Index / NumSamples + float( Random.x ) * (1.0 / 65536.0) );
	float E2 = float( ( ReverseBits32(Index) >> 16 ) ^ Random.y ) * (1.0 / 65536.0);
	return float2( E1, E2 );
}
float3 ComputeL(float3 N, float2 E)
{
	float3x3 TangentBasis = GetTangentBasis(N);
	
	float3 L;
	L.xy = UniformSampleDiskConcentric( E );
	L.z = sqrt( saturate(1 - dot( L.xy, L.xy ) ));
	L = mul( L, TangentBasis );

	return L;
}

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.HPosition.z = OUT.HPosition.w;

	OUT.TC = IN.BaseTC;
	return OUT;
}

half4 HiZRayMarching(float3 rayStartScreen, float3 rayStepScreen, float compareTolerance, float stepOffset)
{
	float3 rayStartUVz = float3(rayStartScreen.xy * cRTAndHZUVScale.xy, rayStartScreen.z);
	float3 rayStepUVz = float3(rayStepScreen.xy * cRTAndHZUVScale.xy, rayStepScreen.z);
	static const half stepNum = 8;  // 默认八步吧
	static const half batchSize = 4;
	const float stepRatio = 1.0 / stepNum;
	float stepCompareTolerance = compareTolerance * stepRatio;
	float level = 1;
	rayStepUVz *= stepRatio;
	float3 rayUVz = rayStartUVz + rayStepUVz * stepOffset;
	float4 multipleSampleDepthDiff;
	bool4 bMutipleSampleHit;
	bool findHit = false;
	int i;
	[unroll]
	for(i = 0; i < stepNum; i = i + 4)
	{
		float2 samplesUV[4];
		float4 samplesZ;
		float4 samplesMip;
		[unroll]
		for(int j = 0; j < batchSize; ++j)
		{
			samplesUV[j] = rayUVz.xy + (i + j + 1) * rayStepUVz.xy;
			samplesZ[j] = rayUVz.z + (i + j + 1) * rayStepUVz.z;
		}
		samplesMip.xy = level;
		level += 1;
		samplesMip.zw = level;
		level += 1;

		float4 sampleDepth;
		[unroll]
		for(int k = 0; k < batchSize; ++k)
		{
			sampleDepth[k] = tDepthBufferLinear.SampleLevel(sDepthBufferSamplerLinear, samplesUV[k], samplesMip[k]).r;
		}
		multipleSampleDepthDiff = samplesZ - sampleDepth;
		bMutipleSampleHit = abs(multipleSampleDepthDiff + stepCompareTolerance) < stepCompareTolerance;
		findHit = any(bMutipleSampleHit);
		if(findHit)
		{
			break;
		}
	}
	if(findHit)
	{
		float4 hitTime;
		if(@_D3D11)
		{
			hitTime = bMutipleSampleHit.xyzw ? float4(1, 2, 3, 4) : 5; // metal may error...
		}else
		{
			hitTime.x = bMutipleSampleHit.x ? 1 : 5;
			hitTime.y = bMutipleSampleHit.y ? 2 : 5;
			hitTime.z = bMutipleSampleHit.z ? 3 : 5;
			hitTime.w = bMutipleSampleHit.w ? 4 : 5;
		}
		float time1 = i + min(min(min(hitTime.x, hitTime.y), hitTime.z), hitTime.w);
		float3 hitUVz = rayUVz + rayStepUVz * time1;
		hitUVz.xy *= cRTAndHZUVScale.zw;
		half3 sampleColor = tSceneColorMap.Sample(sSceneColorSampler, hitUVz.xy, 0).rgb * 4;
		sampleColor = sampleColor / (1 + GetLuminanceGI(sampleColor));
		return half4(sampleColor, 1);
	}
	return half4(0, 0, 0, 0);
}

float GetStepScreenFactorToClipAtScreenEdge(float2 startUV, float2 stepUV)
{
	float2 rayStartScreen = 2 * startUV - 1;
	float2 rayStepScreen = stepUV * 2;
	const float rayStepScreenInvFactor = 0.5 * length(rayStepScreen);
	const float2 S = 1 - max(abs(rayStepScreen + rayStartScreen * rayStepScreenInvFactor) - rayStepScreenInvFactor, 0.0f) / abs(rayStepScreen);
	const float RayStepFactor = min(S.x, S.y) / rayStepScreenInvFactor;

	return RayStepFactor;
}
half4 RSMMarching(float3 rayStartScreen, float3 rayStepScreen, float stepOffset, float3 worldNormal)
{
	static const half stepNum = 8;
	static const half batchSize = 4;
	const float stepRatio = 1.0 / stepNum;
	float level = 0;
	rayStepScreen *= stepRatio;
	float3 stepRay = normalize(rayStepScreen);
	float3 rayUVz = rayStartScreen + rayStepScreen * stepOffset;
	bool4 bMutipleSampleHit;
	bool findHit = false;
	int i;
	[unroll]
	for(i = 0; i < stepNum; i = i + 4)
	{
		float2 samplesUV[4];
		float4 samplesZ;
		[unroll]
		for(int j = 0; j < batchSize; ++j)
		{
			samplesUV[j] = rayUVz.xy + (i + j + 1) * rayStepScreen.xy;
			samplesZ[j] = rayUVz.z + (i + j + 1) * rayStepScreen.z;
		}
		float4 sampleDepth;
		[unroll]
		for(int k = 0; k < batchSize; ++k)
		{
			sampleDepth[k] = tVPLDepthMap.SampleLevel(sVPLDepthMap, samplesUV[k], 0).r;
		}
		float hit1 = dot(stepRay, normalize(float3(samplesUV[0], sampleDepth[0]) - rayStartScreen));
		float hit2 = dot(stepRay, normalize(float3(samplesUV[1], sampleDepth[1]) - rayStartScreen));
		float hit3 = dot(stepRay, normalize(float3(samplesUV[2], sampleDepth[2]) - rayStartScreen));
		float hit4 = dot(stepRay, normalize(float3(samplesUV[3], sampleDepth[3]) - rayStartScreen));
		bMutipleSampleHit = saturate(float4(hit1, hit2, hit3, hit4) - 0.99) > 0;
		findHit = any(bMutipleSampleHit);
		if(findHit)
		{
			break;
		}
	}
	if(findHit)
	{
		float4 hitTime;
		if(@_D3D11)
		{
			hitTime = bMutipleSampleHit.xyzw ? float4(1, 2, 3, 4) : 5; // metal may error...
		}else
		{
			hitTime.x = bMutipleSampleHit.x ? 1 : 5;
			hitTime.y = bMutipleSampleHit.y ? 2 : 5;
			hitTime.z = bMutipleSampleHit.z ? 3 : 5;
			hitTime.w = bMutipleSampleHit.w ? 4 : 5;
		}
		float time1 = i + min(min(min(hitTime.x, hitTime.y), hitTime.z), hitTime.w);
		float3 hitUVz = rayUVz + rayStepScreen * time1;
		half3 sampleColor = tVPLColorMap.Sample(sVPLColorMap, hitUVz.xy, 0).rgb;
		float3 normal = DecodeNormal(tVPLNormalMap.Sample(sVPLNormalMap, hitUVz.xy, 0).rgb);
		sampleColor = sampleColor / (1 + GetLuminanceGI(sampleColor));
		
		return half4(sampleColor * saturate(1 - dot(normal, worldNormal)), 1);
	}
	return half4(0, 0, 0, 0);
}
half4 VPLGICalc(float depth, float2 stc, float3 worldNormal, float stepOffset, uint2 randomSeed)
{
	static const int RAY_COUNT = 4;
	float3 worldPos = CalcHomogeneousPos(depth, stc);
	worldPos += CameraPos.xyz;
	float4 shadowViewProjUVZ = mul(float4(worldPos, 1), cShadowViewProjTex);
	shadowViewProjUVZ.xyz /= shadowViewProjUVZ.w;
	half2 inRange = abs(shadowViewProjUVZ.xy - 0.5);
	float inRate = min(0.5-inRange.x, 0.5-inRange.y);
	if(inRate <= 0)
	{
		return half4(0, 0, 0, 0);
	}
	float3 rayStartScreen = shadowViewProjUVZ.xyz;
	half4 result = half4(0, 0, 0, 0);
	[unroll]
	for(int i = 0; i < RAY_COUNT; ++i)
	{
		float2 hammersleyE = Hammersley16(i, RAY_COUNT, randomSeed);
		half3 ray = ComputeL(worldNormal, hammersleyE);
		float4 rayEndClip = mul(float4(worldPos + ray * (1 - depth), 1), cShadowViewProjTex);//MulMat4x4(cShadowProjTex, float4(ray, 0))+ float4(rayStartScreen, 1);
		float3 rayEndScreen = rayEndClip.xyz / rayEndClip.w;
		float3 rayStepScreen = rayEndScreen - rayStartScreen;
		rayStepScreen *= GetStepScreenFactorToClipAtScreenEdge(rayStartScreen.xy, rayStepScreen.xy);

		result += RSMMarching(rayStartScreen, rayStepScreen, stepOffset, worldNormal);
	}
	result  = result / RAY_COUNT;
	result.rgb = result.rgb / (1 - GetLuminanceGI(result.rgb)) * cVPLIntensity;
	return result;
}
float4 SSGICalc(v2f IN)
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	
	half2 tc = spos * cRTSize.zw;
	float depth = DecodeDepth(tSceneDepthBuffer.SampleLevel(sSceneDepthBufferSampler, tc, 0));
	if(depth > 1 - EPSILON)
	{
		return float4(0, 0, 0, 0.9960784); //254/255
	}
	static const int RAY_COUNT = 4;
	float4 normalAndShadow = tGBufferE.Sample(sGBufferE, tc);
	int filterFlag;
	DecodeFilter(normalAndShadow.a, filterFlag);

	normalAndShadow.xyz = DecodeNormal(normalAndShadow.xyz);
	half3 vNormal = mul(normalAndShadow.xyz, (float3x3)cWorldToView);

	uint2 randomSeed = ComputeRandomSeed(uint2(spos));
	float stepOffset = InterleavedGradientNoise(spos, cFrameIndex);
	stepOffset -= 0.9;

	// 初始化射线数据
	float startDeviceZ = GetDeviceZ(depth);
	float3 rayStartScreen = float3(tc, startDeviceZ); // 注意uv已经是变换到0-1

	half4 result = half4(0, 0, 0, 0);
	[unroll]
	for(int i = 0; i < RAY_COUNT; ++i)
	{
		float2 hammersleyE = Hammersley16(i, RAY_COUNT, randomSeed);
		half3 ray = ComputeL(vNormal, hammersleyE); // view的ray
		float4 rayEndClip = MulMat4x4(cProjTex, float4(ray, 0))+ float4(rayStartScreen, 1);

		float3 rayEndScreen = rayEndClip.xyz / rayEndClip.w;
		float3 rayDepthScreen = 0.5 * (rayStartScreen.xyz + mul(float4(0, 0, 1, 0), cProjTex).xyz); // notice
		float3 rayStepScreen = rayEndScreen - rayStartScreen;

		rayStepScreen *= GetStepScreenFactorToClipAtScreenEdge(rayStartScreen.xy, rayStepScreen.xy);
		
		float compareTolerance = max(abs(rayStepScreen.z), (rayStartScreen.z - rayDepthScreen.z)*2);

		result += HiZRayMarching(rayStartScreen, rayStepScreen, compareTolerance, stepOffset);
	}
	result  = result / RAY_COUNT;
	result.rgb = result.rgb / (1 - GetLuminanceGI(result.rgb)) * cAOIntensity.x;
	if(@_EnableVPL)
	{
		float4 vplColor = VPLGICalc(depth, tc, normalAndShadow.xyz, stepOffset, randomSeed);
		result.rgb = lerp(result.rgb, vplColor.rgb, saturate((GetLuminanceGI(vplColor.rgb) - GetLuminanceGI(result.rgb))*100)); // trick
		//result.a = max(result.a, vplColor.a);
		//result.a = vplColor.a;
		//result.rgb = vplColor.rgb;
	}
	result.a = saturate(1 - result.a * cAOIntensity.y);	
	result.a = EncodeShadowAndFilter(result.a, filterFlag);
	return result;
	
}

float4 GICombinePS(v2f IN) : SV_Target
{
	// TODO: Simply output one of the GI results. How to actually blend them?
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	half2 tc = spos * cRTSize.zw;

	if (@_EnableSSGI)
	{
		return SSGICalc(IN);
	}

	if (@_EnableVXGI)
	{
		return tVXGI.Sample(sVXGI, tc);
	}
	return float4(0, 0, 0, 0);
}

technique GICombine
{
	VertexShader = ScreenVS();
	PixelShader = GICombinePS();
}
// {HEX} end
