#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "DeferredShadingCommon.fxh"
@_PreAOEnable
{ 
	Name = PreAOEnable;
	Description = "Do we have Pre AO?";  
	Default = false;
};
@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Default = false;
	Visible = false;
};
@_EnableGI
{
	Name = EnableGI;
	Description = "Enable GI";
	Default = false;
};
///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX				: WBasisX ;
	float4 vWBasisY				: WBasisY ;
	float4 vWBasisZ				: WBasisZ ;
	float4 cRTSize				;			//xy-ao rt size, zw-pingpong rt size
	float2 cClipToView			 = {0, 0};//tan(0.5*fovy)*w/h, tan(0.5*fovy)
	float2 cViewToClip 			 = {0, 0};//cot(0.5*fovy)*h/w, cot(0.5*fovy)
	float2 cPixelSize			;
	float2 cRadius				;
	float4x4 cLastViewProjTex	;
	float4x3 cWorldToView		: View ;
	float cFrameIndex			;
};

texture2D tDepthBufferLinear : SceneDepth : layout(t0);
sampler sDepthBufferSamplerLinear : layout(s0)
{
	Texture = tDepthBufferLinear;
	Filter = Linear;
	Address = Clamp;
};
texture2D tCurrAOMap : layout(t1);
sampler sCurrAOSampler : layout(s1)
{
	Texture = tCurrAOMap;
	Filter = Linear;
	Address = Clamp;
};
texture2D tCurrShadowMap : layout(t2);
sampler sCurrShadowSampler : layout(s2)
{
	Texture = tCurrShadowMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tGBufferE : layout(t3);
sampler sGBufferE : layout(s3)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};
texture2D tPreAOMap : layout(t4);
sampler sPreAOSampler : layout(s4)
{
	Texture = tPreAOMap;
	Filter = Linear;
	Address = Clamp;
};
texture2D tVBuffer : VelocityMap : layout(t5);
sampler sVBufferSampler : layout(s5)
{
	Texture = tVBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D tGIAO : layout(t6);
sampler sGIAO : layout(s6)
{
	Texture = tGIAO;
	Filter = Point;
	Address = Clamp;
};
///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

// function
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}
float3 uvToViewPosWithDepth(float2 uv, float depth)
{
	float2 vPosInClip = uv * 2 - 1;
	[!@_GLSL]
	vPosInClip.y = -vPosInClip.y;
	float3 viewPos;
	viewPos.z = depth * CameraInfo.y;
	viewPos.xy = vPosInClip * cClipToView * viewPos.z;
	return viewPos;
}

float3 uvToViewPosLinear(float2 uv)
{
	float depth = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, uv));
	return uvToViewPosWithDepth(uv, depth);
}

float4 EncodeAODepth(float ao, float depth)
{
	if(@_GLSL)
		return float4(FloatToR8G8(ao), FloatToR8G8(depth));
	else
		return float4(ao, depth, 1, 1);
}

float2 DecodeAODepth(float4 raw)
{
	if(@_GLSL)
		return float2(R8G8ToFloat(raw.xy), R8G8ToFloat(raw.zw));
	else
		return raw.xy;
}

float2 HDecodeAODepth(float4 raw)
{
	if (@_ISRGBA8 || @_GLSL)
		return float2(R8G8ToFloat(raw.xy), R8G8ToFloat(raw.zw));
	else
		return raw.xy;
}

float3 rebuildNormalFromDepth(float3 origPos, float2 uv)
{
	float3 posXN = uvToViewPosLinear(uv + float2(-ScreenInfo.z, 0));		//-x 
	float3 posX = uvToViewPosLinear(uv + float2(ScreenInfo.z, 0));		//+x

	float3 posYN = uvToViewPosLinear(uv + float2(0, ScreenInfo.w));		//-y
	float3 posY = uvToViewPosLinear(uv + float2(0, -ScreenInfo.w));		//+y
	if (@_GLSL)
	{
		float3 temp = posYN;
		posYN = posY;
		posY = temp;
	}

	float3 _ddx1 = posX-origPos;
	float3 _ddx2 = origPos - posXN;
	float3 _ddx = lerp(_ddx1, _ddx2, step(abs(_ddx2.z),abs(_ddx1.z)));

	float3 _ddy1 = posY-origPos;
	float3 _ddy2 = origPos - posYN;
	float3 _ddy = lerp(_ddy1, _ddy2, step(abs(_ddy2.z),abs(_ddy1.z)));
	float3 N = -normalize(cross(_ddx, _ddy));
	return N;
}

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.HPosition.z = OUT.HPosition.w;

	OUT.TC = IN.BaseTC;
	return OUT;
}

float gtaoSqrt(float x)
{
	float t = asfloat(3757216765 + (asint(x) >> 1));
	return 0.5 * (t + x / t);
}

half2 RayMarch(int stepCount, half2 screenPos, half2 sliceDirTexelSize, half stepRadius, half noiseOffset, 
	half twoOverSquareRadius, half thinkness, half3 vPos, half3 vDir)
{
	half2 h = half2(-1, -1);
	[unroll]
	for(int i = 0; i < stepCount; ++i)
	{
		half2 uvOffset = sliceDirTexelSize * max(stepRadius * (i + noiseOffset), 1 + i);
		uvOffset.y = - uvOffset.y;
		half4 uv = screenPos.xyxy + half4(uvOffset.xy, -uvOffset.xy);
		half3 ds = uvToViewPosLinear(uv.xy) - vPos;
		half3 dt = uvToViewPosLinear(uv.zw) - vPos;
		half2 dsdt = half2(dot(ds, ds), dot(dt, dt));
		half2 dsdtLength = rsqrt(dsdt + half2(0.0001, 0.0001));
		//half2 tt = dsdt + half2(0.0001, 0.0001);
		//half2 dsdtLength = half2(gtaoSqrt(tt.x), gtaoSqrt(tt.y));
		half2 dsdtCos = half2(dot(ds, vDir), dot(dt, vDir)) * dsdtLength;
		half2 falloff = saturate(dsdt * twoOverSquareRadius);
		half2 weight = step(dsdtCos.xy, h.xy);
		h.xy = lerp(dsdtCos, h, falloff)*(1-weight) + lerp(dsdtCos, h, thinkness)*weight;
		//h.xy = dsdtCos.xy > h.xy ? lerp(dsdtCos, h, falloff) : lerp(dsdtCos, h, thinkness);
	}
	return h;
}

float4 GTAOCalc(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	
	half2 tc = spos * cRTSize.zw;
	float depth = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, tc));
	if(depth > 1.0 - EPSILON)
	{
		return float4(0, 0, 0, 1);
	}
	
	half3 vPos = uvToViewPosWithDepth(tc, depth); // view pos
	int filterFlag = NO_NEED_FILTER;
	half4 giao = half4(0, 0, 0, 1);
	if(@_EnableGI)
	{
		giao = tGIAO.Sample(sGIAO, tc);
		DecodeFilter(giao.a, filterFlag);
	}
	half3 vNormal;
	if(@_NormalMapEnable)//CACL NORMAL
	{
		float4 normalAndShadow = tGBufferE.Sample(sGBufferE, tc);	
		DecodeFilter(normalAndShadow.a, filterFlag);
		normalAndShadow.xyz = DecodeNormal(normalAndShadow.xyz);
		vNormal = mul(normalAndShadow.xyz, (float3x3)cWorldToView);
		vNormal.z = -vNormal.z;
		vNormal = normalize(vNormal);
	}else
	{
		vNormal = rebuildNormalFromDepth(vPos, tc); // view normal
	}

	half3 vDir = normalize(half3(0,0,0) - vPos.xyz);
	half vDirXYDot = dot(vDir.xy, vDir.xy);
	half radius = 10 * cRadius.x;
	int stepCount = 4;
	int directionCount = 4;
	half thinkness = 1;
	half screenRadius = max(min(10 * radius / vPos.z, 512), stepCount);
	half twoOverSquareRadius = 300 * cRadius.y / radius / radius;
	half stepRadius = screenRadius / (stepCount + 1);
	half noiseOffset;
	half noiseDir;
	//noiseOffset = ((int(spos.y) - int(spos.x)) & 3) / 4.0;
	noiseOffset = frac((spos.y - spos.x)/4.0);
	noiseDir = InterleavedGradientNoise(spos, cFrameIndex);
	//noiseDir = GetRandom(spos);
	half piOver = PI / directionCount;
	half noiseAngleDir = noiseDir * piOver;
	half ao = 0;
	half2 h;
	[unroll]
	for(int dirNum = 0; dirNum < directionCount; ++dirNum)
	{
		half angle = dirNum * piOver + noiseAngleDir;
		half2 cosSin = half2(cos(angle), sin(angle));
		half3 sliceDir = half3(cosSin.xy, 0);
		half wallDarkingCorrection = dot(vNormal, cross(vDir, sliceDir)) * vDirXYDot;
		wallDarkingCorrection = wallDarkingCorrection * wallDarkingCorrection;
		half2 sliceDirTexelSize = sliceDir.xy * cRTSize.zw;
		
		h = RayMarch(stepCount, tc, sliceDirTexelSize, stepRadius, noiseOffset, twoOverSquareRadius, thinkness, vPos, vDir);
		
		half3 normalSlicePlane = normalize(cross(sliceDir, vDir));
		half3 tangentSlicePlane = cross(vDir, normalSlicePlane);
		half3 normalProjected = vNormal - normalSlicePlane * dot(vNormal, normalSlicePlane);
		half projLength = length(normalProjected);
		half cosN = clamp(dot(normalProjected, vDir)/projLength, -1.0, 1.0);
		half n = -sign(dot(normalProjected, tangentSlicePlane)) * acos(cosN);

		h = acos(clamp(h, -1, 1));
		
		h.x = n + max(-h.x-n, -HALF_PI); // h1 negative
		h.y = n + min(h.y-n, HALF_PI);

		half sinN = sin(n);
		half2 h2 = 2 * h;
		half2 innerIntegral = -cos(h2 - n) + cosN + h2 * sinN;
		
		ao += (projLength + wallDarkingCorrection) * 0.25 * (innerIntegral.x + innerIntegral.y);
	}
	ao /= directionCount;
	ao = saturate(ao);

	return float4(giao.rgb, EncodeShadowAndFilter(min(pow(ao, 2), giao.a), filterFlag));
}
technique CalcGTAO
{
	VertexShader = ScreenVS();
	PixelShader = GTAOCalc();
}

float4 FilterAO(v2f IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * cPixelSize;
	half aoAndPreAO = saturate(tCurrAOMap.Sample(sCurrAOSampler, stc).a);
	if(!@_PreAOEnable)
	{
		return float4(0, 0, 0, aoAndPreAO);
	}
	const int2 sampleOffset[8] = {int2(-1.0, -1.0), int2(0.0, -1.0), int2(1.0, -1.0), int2(-1.0, 0.0), int2(1.0, 0.0), int2(-1.0, 1.0), int2(0.0, 1.0), int2(1.0, 1.0)};
	half sampleAOs[8];
	[unroll]
	for(int i = 0; i < 8; ++i)
	{
		sampleAOs[i] = saturate(tCurrAOMap.Sample(sCurrAOSampler, stc + sampleOffset[i]*cPixelSize).a);
	}
	half4 m1 = aoAndPreAO;
	half4 m2 = aoAndPreAO * aoAndPreAO;
	[unroll]
	for(int j = 0; j < 8; ++j) {
		m1 += sampleAOs[j];
		m2 += sampleAOs[j] * sampleAOs[j];
	}
	half4 mean = m1 / 9.0;
	half4 stddev = sqrt((m2 / 9.0) - mean * mean);
	half aabb = cRadius.x;
	half minAO = min(mean - aabb * stddev, aoAndPreAO);
	half maxAO = max(mean + aabb * stddev, aoAndPreAO);
	
	float depth = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, stc));
	float3 worldPos = CalcHomogeneousPos(depth, stc);
	worldPos += CameraPos.xyz;
	float4 lastPos = MulMat4x4(cLastViewProjTex, float4(worldPos,1.0));
	float2 lastUV = (lastPos.xy / lastPos.w);

	half2 inRange = abs(lastUV - 0.5);
	half preAORate = step(max(inRange.x, inRange.y), 0.5);
	//float preDepth = DecodeDepth(tPreDepthMap.Sample(sPreDepthSampler, lastUV));
	float diffDepth = 1;//step(depth - preDepth,0.001);
	if (@_VelocityMapEnable)
	{
		float2 motionVector = abs(tVBuffer.SampleLevel(sVBufferSampler, stc, 0).xy);
		diffDepth = step(motionVector.x+motionVector.y, EPSILON);
	}

	half preAO = saturate(tPreAOMap.Sample(sPreAOSampler, lastUV).a);
	

	preAO = clamp(preAO, minAO, maxAO);
	half finalAO = lerp(aoAndPreAO.x, preAO, saturate(cRadius.y)* preAORate * diffDepth);
	//return float4(aoAndPreAO.y,aoAndPreAO.y,aoAndPreAO.y,1);
	return float4(0, 0, 0, finalAO);
}
technique FilterGTAO
{
	VertexShader = ScreenVS();
	PixelShader = FilterAO();
}

float4 GTAOMerge(v2f IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float shadow = tCurrShadowMap.Sample(sCurrShadowSampler, stc).r;
	//get border Pixels
	stc += float2(0.5, 0.5) * cPixelSize;
	float ao = saturate(tCurrAOMap.Sample(sCurrAOSampler, stc).a);
	//float finalAO = lerp(ao*1.2, ao, shadow);
	return float4(shadow, ao, 0, 0);
}

technique MergeGTAO
{
	VertexShader = ScreenVS();
	PixelShader = GTAOMerge();
}

float4 GTAOBlend(v2f IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float3 color = float3(0,0,0);
	float4 normalAndAo = tGBufferE.Sample(sGBufferE, stc);
	normalAndAo.rgb = DecodeNormal(normalAndAo.rgb);
	//get border Pixels
	float2 aoStc = stc + float2(0.5, 0.5) * cPixelSize;
	float ao = saturate(tCurrAOMap.Sample(sCurrAOSampler, aoStc).a);
	float NoL = saturate(dot(normalAndAo.rgb, SunDirection.rgb)) * (1 - normalAndAo.a);
	ao = saturate(ao + (1 - 1/max(1, GetLuminance(SunColor.rgb)))*NoL);
	//return float4(ao, ao, ao, 1);
	return float4(0, 0, 0, 1 - ao);
}

technique BlendGTAO
{
	VertexShader = ScreenVS();
	PixelShader = GTAOBlend();
}
