//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liuzhongyuan@corp.netease.com
//	Author: Jrenliu
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GpuCloth.fx>
///	@path	~/System/Shader/
///	@date	2020/03/22
///	@brief	GPU cloth shader.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader draw

#define INCLUDE_GPU_PHYSICS_VS
#define INCLUDE_GPU_PHYSICS_DATA
#define INCLUDE_GPU_PHYSICS_SKINNING_TOOL
#include "GpuClothCommon.fxh"

struct ClothInitOutput
{
    float4 PrePositionResult   : SV_Target0;
    float4 PrePrePositionResult: SV_Target1;
};

[target("RGBA32F")]
ClothInitOutput ClothInitPs(in SimpleVSOutput IN)
{
    ClothInitOutput OUT;
    float3 skinning;
    uint skeletonStart = (uint)tFabricTexture.SampleLevel(sFabricTextureSampler, IN.TexCoord, 0).z;
    float4 static_position = tStaticPositionTexture.SampleLevel(sStaticPositionTextureSampler, IN.TexCoord, 0);
    {
        float4 fweight = tWeightTexture.SampleLevel(sWeightTextureSampler, IN.TexCoord, 0);
        float4 findex = tIndexTexture.SampleLevel(sIndexTextureSampler, IN.TexCoord, 0);
        uint4 index = (uint)(255.5 * findex);
        SkinningTex(skeletonStart, fweight, index , static_position.xyz, skinning);
    }
    // Clear the vel.
    OUT.PrePositionResult = float4(skinning, 0);
    OUT.PrePrePositionResult = float4(skinning, 0);
    return OUT;
}

technique ClothInit
{
    VertexShader = SimpleVS();
    PixelShader = ClothInitPs();
}
