//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Gray.fx>
///	@path	~/System/Shader/
///	@date	2015/04/08
///	@brief	Gray.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// const buffer //////////////////
cbuffer Batch
{
	float3 cMaskColor			;
	float cGrayPercent 			;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tBaseMap : layout(t0);
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/////////////////// HDRSampleLumInitial ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 GrayBlendPS(v2f IN) : SV_Target
{
	half3 Source = SampleColorMap(tBaseMap, sBaseSampler, IN.TC.xy).rgb;
	half d = GetLuminance(Source);
	half dd = d * d;
	half3 Result = (cMaskColor*(d-dd)+dd);
	
	Result = lerp(Source, Result, cGrayPercent);

	return float4(Result, 1.0);
}

technique GrayTech
{
	VertexShader = ScreenVS();
	PixelShader = GrayBlendPS();
}


