#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4 cInvSize	;
};

texture2D tInputRT : layout(t0);
sampler sInputRT_Sampler : layout(s0)
{
	Texture = tInputRT;
	Filter = Point;
	Address = Clamp;
};

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};
VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 HierarchicalZPS( VertexScreenOutput IN ) : SV_Target
{
	float4 depthSamples;
	
	depthSamples.x = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(-0.25f, -0.25f) * cInvSize.zw));
	depthSamples.y = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(0.25f, -0.25f) * cInvSize.zw));
	depthSamples.z = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(-0.25f, 0.25f) * cInvSize.zw));
	depthSamples.w = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(0.25f, 0.25f) * cInvSize.zw));
	
	
	float furthestZ = min(min(depthSamples.x,depthSamples.y), min(depthSamples.z,depthSamples.w));
	return furthestZ.xxxx;
}

float HierarchicalZToDepthPS(VertexScreenOutput IN) : SV_Depth
{
	float4 depthSamples;

	depthSamples.x = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(-0.25f, -0.25f) * cInvSize.zw));
	depthSamples.y = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(0.25f, -0.25f) * cInvSize.zw));
	depthSamples.z = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(-0.25f, 0.25f) * cInvSize.zw));
	depthSamples.w = DecodeDepth(tInputRT.Sample(sInputRT_Sampler, IN.TC * cInvSize.xy + float2(0.25f, 0.25f) * cInvSize.zw));

	float furthestZ = min(min(depthSamples.x,depthSamples.y), min(depthSamples.z,depthSamples.w));
	return furthestZ;
}

technique HierarchicalZTech
{
	VertexShader = ScreenVS();
	PixelShader = HierarchicalZPS();
}

technique HierarchicalZToDepthTech
{
	VertexShader = ScreenVS();
	PixelShader = HierarchicalZToDepthPS();
}
