//============================================
// Uncharted settings
// Slope = 0.63;
// Toe = 0.55;
// Shoulder = 0.47;
// BlackClip= 0;
// WhiteClip = 0.01;

// HP settings
// Slope = 0.65;
// Toe = 0.63;
// Shoulder = 0.45;
// BlackClip = 0;
// WhiteClip = 0;

// Legacy settings
// Slope = 0.98;
// Toe = 0.3;
// Shoulder = 0.22;
// BlackClip = 0;
// WhiteClip = 0.025;

// ACES settings
// Slope = 0.91;
// Toe = 0.53;
// Shoulder = 0.23;
// BlackClip = 0;
// WhiteClip = 0.035;
//===========================================
#define FilmSlope 0.91
#define FilmToe 0.53
#define FilmShoulder 0.23
#define FilmBlackClip 0
#define FilmWhiteClip 0.035

half3 ACES_U4( half3 LinearColor ) 
{
	static const float3x3 XYZ_2_AP0_MAT =
	{
		 1.0498110175, 0.0000000000,-0.0000974845,
		-0.4959030231, 1.3733130458, 0.0982400361,
		 0.0000000000, 0.0000000000, 0.9912520182
	};

	static const float3x3 AP1_2_XYZ_MAT = 
	{
		0.6624541811, 0.1340042065, 0.1561876870,
		0.2722287168, 0.6740817658, 0.0536895174,
		-0.0055746495, 0.0040607335, 1.0103391003
	};

	static const float3x3 AP0_2_AP1_MAT = //mul( AP0_2_XYZ_MAT, XYZ_2_AP1_MAT );
	{
		 1.4514393161, -0.2365107469, -0.2149285693,
		-0.0765537734,  1.1762296998, -0.0996759264,
		 0.0083161484, -0.0060324498,  0.9977163014
	};

	static const float3x3 AP1_2_AP0_MAT = //mul( AP1_2_XYZ_MAT, XYZ_2_AP0_MAT );
	{
		 0.6954522414,  0.1406786965,  0.1638690622,
		 0.0447945634,  0.8596711185,  0.0955343182,
		-0.0055258826,  0.0040252103,  1.0015006723
	};

	static const float3 AP1_RGB2Y =
	{
		0.2722287168, //AP1_2_XYZ_MAT[0][1],
		0.6740817658, //AP1_2_XYZ_MAT[1][1],
		0.0536895174 //AP1_2_XYZ_MAT[2][1]
	};

	const float3x3 AP1_2_AP0 = mul( XYZ_2_AP0_MAT, AP1_2_XYZ_MAT );
	
	float3 ColorAP1 = LinearColor;
	//float3 ColorAP1 = mul( sRGB_2_AP1, float3(LinearColor) );
	float3 ColorAP0 = mul( AP1_2_AP0, ColorAP1 );
	// Use ACEScg primaries as working space
	float3 WorkingColor = mul( AP0_2_AP1_MAT, ColorAP0 );
	WorkingColor = max( 0, WorkingColor );
	// Pre desaturate
	WorkingColor = lerp( dot( WorkingColor, AP1_RGB2Y ), WorkingColor, 0.96 );
	
	const half ToeScale			= 1 + FilmBlackClip - FilmToe;
	const half ShoulderScale	= 1 + FilmWhiteClip - FilmShoulder;
	
	const float InMatch = 0.18;
	const float OutMatch = 0.18;

	float ToeMatch;
	if( FilmToe > 0.8 )
	{
		// 0.18 will be on straight segment
		ToeMatch = ( 1 - FilmToe  - OutMatch ) / FilmSlope + log10( InMatch );
	}
	else
	{
		// 0.18 will be on toe segment

		// Solve for ToeMatch such that input of InMatch gives output of OutMatch.
		const float bt = ( OutMatch + FilmBlackClip ) / ToeScale - 1;
		ToeMatch = log10( InMatch ) - 0.5 * log( (1+bt)/(1-bt) ) * (ToeScale / FilmSlope);
	}

	float StraightMatch = ( 1 - FilmToe ) / FilmSlope - ToeMatch;
	float ShoulderMatch = FilmShoulder / FilmSlope - StraightMatch;
	
	half3 LogColor = log10( WorkingColor );
	half3 StraightColor = FilmSlope * ( LogColor + StraightMatch );
	
	half3 ToeColor		= (    -FilmBlackClip ) + (2 *      ToeScale) / ( 1 + exp( (-2 * FilmSlope /      ToeScale) * ( LogColor -      ToeMatch ) ) );
	half3 ShoulderColor	= ( 1 + FilmWhiteClip ) - (2 * ShoulderScale) / ( 1 + exp( ( 2 * FilmSlope / ShoulderScale) * ( LogColor - ShoulderMatch ) ) );

	ToeColor		= LogColor <      ToeMatch ?      ToeColor : StraightColor;
	ShoulderColor	= LogColor > ShoulderMatch ? ShoulderColor : StraightColor;

	half3 t = saturate( ( LogColor - ToeMatch ) / ( ShoulderMatch - ToeMatch ) );
	t = ShoulderMatch < ToeMatch ? 1 - t : t;
	t = (3-2*t)*t*t;
	half3 ToneColor = lerp( ToeColor, ShoulderColor, t );

	// Post desaturate
	ToneColor = lerp( dot( float3(ToneColor), AP1_RGB2Y ), ToneColor, 0.93 );

	// Returning positive AP1 values
	return ToneColor;
}

half3 SRGB2AP1(half3 LinearColor)
{
	static const float3x3 XYZ_2_AP1_MAT =
	{
		1.6410233797, -0.3248032942, -0.2364246952,
		-0.6636628587,  1.6153315917,  0.0167563477,
		0.0117218943, -0.0082844420,  0.9883948585
	};

	static const float3x3 D65_2_D60_CAT =
	{
		1.01303,    0.00610531, -0.014971,
		0.00769823, 0.998165,   -0.00503203,
		-0.00284131, 0.00468516,  0.924507
	};

	static const float3x3 sRGB_2_XYZ_MAT =
	{
		0.4124564, 0.3575761, 0.1804375,
		0.2126729, 0.7151522, 0.0721750,
		0.0193339, 0.1191920, 0.9503041
	};

	const float3x3 sRGB_2_AP1 = mul( XYZ_2_AP1_MAT, mul( D65_2_D60_CAT, sRGB_2_XYZ_MAT ) );
	return mul( sRGB_2_AP1, LinearColor );
}

half3 AP12SRGB(half3 ColorAP1)
{
	static const float3x3 D60_2_D65_CAT =
	{
		0.987224,   -0.00611327, 0.0159533,
		-0.00759836,  1.00186,    0.00533002,
		0.00307257, -0.00509595, 1.08168
	};

	static const float3x3 AP1_2_XYZ_MAT = 
	{
		0.6624541811, 0.1340042065, 0.1561876870,
		0.2722287168, 0.6740817658, 0.0536895174,
		-0.0055746495, 0.0040607335, 1.0103391003
	};

	static const float3x3 XYZ_2_sRGB_MAT =
	{
		3.2409699419, -1.5373831776, -0.4986107603,
		-0.9692436363,  1.8759675015,  0.0415550574,
		0.0556300797, -0.2039769589,  1.0569715142
	};

	const float3x3 AP1_2_sRGB = mul( XYZ_2_sRGB_MAT, mul( D60_2_D65_CAT, AP1_2_XYZ_MAT ) );

	return mul( AP1_2_sRGB, ColorAP1 );
}

float3 RRTAndODTFit(float3 v)
{
    float3 a = v * (v + 0.0245786) - 0.000090537;
    float3 b = v * (0.983729 * v + 0.4329510) + 0.238081;
    return a / b;
}

// stephen Hill
half3 ACES_StephenHill(half3 color)
{
	// sRGB => XYZ => D65_2_D60 => AP1 => RRT_SAT
	static const float3x3 ACESInputMat =
	{
		0.59719, 0.35458, 0.04823,
		0.07600, 0.90834, 0.01566,
		0.02840, 0.13383, 0.83777
	};

	// ODT_SAT => XYZ => D60_2_D65 => sRGB
	static const float3x3 ACESOutputMat =
	{
		1.60475, -0.53108, -0.07367,
		-0.10208,  1.10813, -0.00605,
		-0.00327, -0.07276,  1.07602
	};

	color = mul(ACESInputMat, color);	
	// Apply RRT and ODT
	color = RRTAndODTFit(color);
	color = mul(ACESOutputMat, color);
	return color;
}

// knarkowicz
half3 ACES_Knarkowicz( float3 x )
{
	float a = 2.51;
	float b = 0.03;
	float c = 2.43;
	float d = 0.59;
	float s = 0.14;
	return (x * ( a * x + b)) / (x * (c * x + d) + s);
}

// knarkowicz inverse (analytical reference)
half3 ACES_Knarkowicz_Inverse_Ref(half3 x)
{
	half a = 2.51;
	half b = 0.03;
	half c = 2.43;
	half d = 0.59;
	half s = 0.14;
	half3 aa = a - x * c;
	half3 bb = b - x * d;
	half3 cc = -x * s;
	return (-bb + sqrt(bb * bb - 4 * aa * cc)) / (2 * aa);
}

// knarkowicz inverse (fitted curve)
half3 ACES_Knarkowicz_Inverse_Approx(half3 x)
{
	return ((-0.074349 * x + 0.175004) * x + 0.076950) * x / ((-0.686161 * x + 0.665234) * x + 0.045533);
}

//#define ACESFilm ACES_U4
//#define ACESFilm ACES_StephenHill
#define ACESFilm ACES_Knarkowicz

#define ACESFilmInverse ACES_Knarkowicz_Inverse_Approx

// Aces tone mapping for 1000 nits display, combined with a Messiah specific IDT.
// 0.18(mid gray) -> 20 nits, 1.0(paper white) -> 172 nits, 8.0(max) -> 1238 nits
half3 ACES_Tone_1000nits(half3 x)
{
	return (200.276765 * x + 1.379809) * x / ((0.039350 * x + 0.958296) * x + 0.171154);
}

