//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Common.fxh>
///	@path	~/System/Shader/Include/
///	@date	2010/05/04
///	@brief	Common shader calculations.

#ifndef __FILE_COMMON_FXH__
#define __FILE_COMMON_FXH__

#pragma shader fallback("@_DISSOLVING", "true")
#pragma shader fallback("@_LOD_LEVEL", "0")

#pragma shader predict("@_FOG", "true")
#pragma shader predict("@_DEFERRED_SHADING", "false")
#pragma shader predict("@_HDR", "false")
#pragma shader predict("@_ENCODEHDR", "false")
#pragma shader predict("@_SKELETON", "false")

//@_CastSunShadow
#pragma shader predict("@_VolumetricFog", "false")
#pragma shader predict("@_Weather", "0")
#pragma shader predict("@_RIPPLE", "false")
#pragma shader predict("@_EnvMapEnable", "true")
#pragma shader predict("@_ReflectionMapEnable", "true")
#pragma shader predict("@_DEBUG", "0")
//@_LOD_LEVEL
//@_DF_SKIN
//@_CLIPPING
//@_TRANSPARENT
//@_NoFarPlaneClip		//vsonly
//@_NoNearPlaneClip		//vsonly
#pragma shader predict("@_Emissive", "true") //deferred only

//@_Billboard
//@_ReceiveDynamicShadow
#pragma shader fallback("@_LitCount", "3")

//@_TANGENT				// Has tangent stream.
//@_VERTEXCOLOR			// Vertex has per-vertex color item.
//@_LIGHT_MAP,			// Has second UV for light map.
//@_INSTANCED,			// Object is rendered as instanced.

//Platform
#define IS_WINDOWS			(@_D3D9 || @_D3D11)
#define IS_IOS				@_METAL
#define IS_ANDROID			(@_GLSL || @_SPRIV)

// Epsilon.
#define EPSILON							0.000001
#define HALF_EPSILON					0.0001
#ifndef PI
	#define PI 3.1415926535897932384626433832795
#endif
#ifndef HALF_PI
	#define HALF_PI 1.5707963267948966192313216916398
#endif
#define MAX_PBR_RGBM_VALUE				16.0
#define MAX_EMISSIVE_VALUE				16.0
// Matrix multiplications.
#define	MulMat4x4(Mat4x4, Vec)			mul(Vec, Mat4x4)
#define	MulMat3x4(Mat3x4, Vec)			mul(Vec, Mat3x4)

// Linear blend 4 primitives.
#define LinearBlend(W, A, B, C, D)		(W[0]*A + W[1]*B + W[2]*C + W[3]*D)

#define LOD_Level_0 0
#define LOD_Level_1 1
#define LOD_Level_2 2
#define LOD_Level_3 3

#define BOARD_DEPTH_BIAS 1.5

#if SD_ENABLE_WEATHER
	//#define WEATHER_NORM 0
	#define WEATHER_SNOW 1
	#define WEATHER_RAIN 2
#else
	//#define WEATHER_NORM 0
	#define WEATHER_SNOW 4//(@_Weather == WEATHER_SNOW) always return false
	#define WEATHER_RAIN 4//(@_Weather == WEATHER_RAIN) always return false
#endif
#define ENABLE_IBL	(@_EnableIBL || @_Weather == WEATHER_RAIN) && (@_EnvMapEnable)
#define ENABLE_IBL_REF (@_EnableIBL || @_Weather == WEATHER_RAIN)

//////////////////////////////////////////////////////////////////////////
#include "ACES.fxh"

#define Expand(a) (a*2 -1)
#define CompressNormal(a) (a*0.5+0.5)

#define EncodeGamma(x) (sqrt(x))
#define DecodeGamma(x) ((x) * (x))

// #define EncodeNormal(a) (a*0.5+0.5)
// #define DecodeNormal(a) (a*2-1)

#define EncodeNormal(a) (a*127.0/255.0 + 127.0/255.0)
#define DecodeNormal(a) (a*255.0/127.0 - 1.0)
 
#define OPTEncodeNormal(orig) \
{\
	float base = min(min(abs(orig.x), abs(orig.y)), abs(orig.z));\
	base = max((1+EPSILON)/127.0, base);\
	float decode = floor(base*127.0) / 127.0;\
	float scale = decode / base;\
	orig *= scale;\
}

#define EncodeFloatRGBA(v, enc) \
{ \
	enc = float4(1.0, 255.0, 65025.0, 16581375.0) * v;\
	enc = frac(enc);\
	enc -= enc.yzww * float4(1.0/255.0,1.0/255.0,1.0/255.0,0.0);\
}

#define DecodeFloatRGBA(rgba) dot( rgba, float4(1.0, 1.0/255.0, 1.0/65025.0, 1.0/16581375.0) )

#define EncodeFloatRG(v, enc) \
{ \
	enc = frac(float2(1.0, 255.0) * v); \
	enc -= enc.yy * float2(1.0/255.0, 0.0); \
}

#define DecodeFloatRG(rg) dot(rg, float2(1.0, 1.0/255.0))

#define EncodeSolidBloomFactor(bloom) step(1, bloom)

#define EncodeTransparentBloomFactor(alpha, bloom)  ((floor(saturate(alpha) * 127)*2 + step(1, bloom)) / 255)

#define DecodeBloomFactor(a) (frac(a * 127.5 + 0.001) * 2)

#define EncodeFloatAndBit(A, B)  ((floor(saturate(A) * 127)*2 + B) / 255)

#define DecodeFloatAndBit(x, A, B) { B = frac(x * 127.5) * 2; A = (x-B/255.0) * 255.0/254.0; }

#define EncodeFloatAB(A, B)  ((floor(saturate(A) * 15)*16 + B*15) / 255)

#define DecodeFloatAB(x, A, B) { B = frac(x*255/16)*16/15; A = (x-B*15/255) * 255.0 / 240.0;}

#define RGBMEncode(srcColor, fMAX_RGBM_VALUE, rgbm) \
{\
	float3 _Color = srcColor / fMAX_RGBM_VALUE;\
	rgbm.a = saturate( max( max( _Color.r, _Color.g ), max( _Color.b, 1.1/65025 ) ) );\
	rgbm.a = sqrt(rgbm.a);\
	rgbm.a = ceil( rgbm.a * 255.0 ) / 255.0;\
	float aa = rgbm.a * rgbm.a;\
	rgbm.rgb = _Color / aa;\
}

#define RGBMDecode(rgbm, fMAX_RGBM_VALUE) (rgbm.rgb * (rgbm.a * rgbm.a * fMAX_RGBM_VALUE))

#if SD_FAKE_HDR
	#define EncodeHDR(Color) \
	{\
		if(@_HDR && @_ENCODEHDR)\
		{\
			Color.rgb /= SD_SCENE_MAX_LUMINANCE;\
			Color.rgb = EncodeGamma(Color.rgb);\
		}\
		else if(!@_HDR)\
		{\
			Color.rgb = ACES_Knarkowicz(Color.rgb);\
			Color.rgb = EncodeGamma(Color.rgb);\
		}\
		else\
		{Color.rgb = min(Color.rgb, SD_SCENE_MAX_LUMINANCE);}\
	}
#else
	#define EncodeHDR(Color) \
	{\
		if(@_HDR && @_ENCODEHDR)\
		{\
			Color.rgb /= SD_SCENE_MAX_LUMINANCE;\
			Color.rgb = EncodeGamma(Color.rgb);\
		}\
		else if(!@_HDR)\
		{\
			Color.rgb = EncodeGamma(Color.rgb);\
		}\
		else\
		{Color.rgb = min(Color.rgb, SD_SCENE_MAX_LUMINANCE);}\
	}
#endif

#define EncodeHDRParticle(Color)\
{\
	Color.rgb = DecodeGamma(Color.rgb);\
	if(@_HDR && @_ENCODEHDR)\
	{\
		Color.rgb /= SD_SCENE_MAX_LUMINANCE;\
		Color.rgb = sqrt(Color.rgb);\
	}\
	else\
	{Color.rgb = min(Color.rgb, SD_SCENE_MAX_LUMINANCE);}\
}

half4 DecodeHDR(half4 color)
{
	if(@_ENCODEHDR)
	{
		color.rgb *= color.rgb;
		color.rgb *= SD_SCENE_MAX_LUMINANCE;
	}
	return color;
}
half3 DecodeHDR(half3 color)
{
	if(@_ENCODEHDR)
	{
		color *= color;
		color *= SD_SCENE_MAX_LUMINANCE;
	}
	return color;
}
half DecodeHDR(half color)
{
	if(@_ENCODEHDR)
	{
		color *= color;
		color *= SD_SCENE_MAX_LUMINANCE;
	}
	return color;	
}

#define	SampleColorMap(T, S, C)			T.Sample(S, C.xy)
#define	SampleColorAlphaMap(T, S, C)	T.SampleRGBA(S, C.xy)
#define	SampleNormalMap(T, S, C)		(half2)T.Sample(S, C.xy).yx

#define	SampleGDepthMap(T, S, C)		((float)T.Sample(S, C.xy).x)
#define	SampleGNormalMap(T, S, C)		((half4)T.Sample(S, C.xy))
#define	GetGNormalMap(T, S, C)			(Expand(SampleGNormalMap(T, S, C)))

#define Square(x) ((x)*(x))
#define Pow2(x) ((x)*(x))

float2 LightMapUV(in float4 Texcoord)
{
	[@_ES20]
	return Texcoord.xy;
	[!@_ES20]
	return Texcoord.zw;
}

float2 BaseMapUV(in float4 Texcoord)
{
	[@_ES20]
	return Texcoord.zw;
	[!@_ES20]
	return Texcoord.xy;
}

float2 BaseMapUV(in float2 Texcoord)
{
	return Texcoord.xy;
}

// Wrap to get vertex color.
float4 GetVertexColor(in float4 Color)
{
	if (@_GLSL || @_METAL || @_SPIRV || @_PS4)
	{
		// swizzeling
		return Color.bgra;
	}
	else
	{
		return Color;
	}
}

// Wrap to encode depth.
float4 EncodeDepth(in float depth)
{
	if (@_ENCODEHDR)
	{
		float4 rgba;
		EncodeFloatRGBA((depth - 0.000001), rgba);
		return rgba;
	}
	else
	{
		return float4(depth, 1, 1, 1);
	}
}

// Wrap to decode depth.
float DecodeDepth(in float4 rgba)
{
	if (@_ENCODEHDR)
	{
		return DecodeFloatRGBA(rgba);
	}
	else
	{
		return rgba.x;
	}
}

////////////////////////////////////////////////////////////////////////
// Random

#define Random2(tc) frac(dot(tc.xyx * tc.xyy, float3(20.390625f, 60.703125f, 2.4281209f)))

float GetRandom( in float2 xy)
{
	return frac(52.9829189 * frac(dot(xy, float2( 0.06711056, 0.00583715))));
}

float Noise2(in float2 tc, in float f /* frequency */)
{
	float2 p = tc * f;
	float2 d = frac(p);
	float2 p00 = p - d;
	float2 p01 = p00 + float2(0, 1);
	float2 p10 = p00 + float2(1, 0);
	float2 p11 = p00 + float2(1, 1);
	return lerp(lerp(Random2(p00), Random2(p01), d[1]), lerp(Random2(p10), Random2(p11), d[1]), d[0]);
}

////////////////////////////////////////////////////////////////////////
// Skinning

// Get 3x4 matrix from skeleton data.
void GetSkeletonMatrix4x3( uint idx, out float4 Mx, out float4 My, out float4 Mz )
{
	Mx = SkeletonData[3*idx];
	My = SkeletonData[3*idx+1];
	Mz = SkeletonData[3*idx+2];
}
// Using 4x3 matrix to transform a position.
float3 Matrix4x3TransformP( float4 Mx, float4 My, float4 Mz, float3 P )
{
	return float3(dot(Mx.xyz, P), dot(My.xyz, P), dot(Mz.xyz, P)) + float3(Mx.w, My.w, Mz.w);
}
// Using 4x3 matrix to transform a normal.
float3 Matrix4x3TransformN( float4 Mx, float4 My, float4 Mz, float3 N )
{
	return float3(dot(Mx.xyz, N), dot(My.xyz, N), dot(Mz.xyz, N));
}

// Get dual quaternion from skeleton data.
void GetSkeletonDualQuaternion( uint idx, out float4 Qn, out float4 Qd )
{
	Qn = SkeletonData[2*idx];
	Qd = SkeletonData[2*idx+1];
}
// Using dual-quaternion to transform a position.
float3 DualQuaternionTransformP( float4 Qn, float4 Qd, float3 P )
{
	return P + 2*(cross(Qn.xyz, cross(Qn.xyz, P) + Qn.w*P) + (Qn.w*Qd.xyz - Qd.w*Qn.xyz + cross(Qn.xyz, Qd.xyz)))/(dot(Qn, Qn)+EPSILON);
}
// Using dual-quaternion to transform a normal.
float3 DualQuaternionTransformN( float4 Qn, float4 Qd, float3 N )
{
	return N + 2*(cross(Qn.xyz, cross(Qn.xyz, N) + Qn.w*N))/(dot(Qn, Qn)+EPSILON);
}

////////////////////////////////////////////////////////////////////////
// Color Space

// Get the Lab luminance of an RGB color.
#define GetLuminance(rgb) dot(rgb, half3(0.2126, 0.7152, 0.0722))

// Convert from RGB to HSVL.
half4 RGB2HSVL( half3 rgb )
{
	half4 K = half4(0, -0.3333, 0.6666, -1);
	half4 p = lerp(half4(rgb.bg, K.wz), half4(rgb.gb, K.xy), step(rgb.b, rgb.g));
	half4 q = lerp(half4(p.xyw, rgb.r), half4(rgb.r, p.yzx), step(p.x, rgb.r));

	half c = min(q.w, q.y);
	half d = q.x-c;
	return half4(abs(q.z+(q.w-q.y)/(6*d+HALF_EPSILON)), d/(q.x+HALF_EPSILON), q.x, (q.x+c)/2);
}
// Convert from HSV to RGB.
half3 HSV2RGB( half3 hsv )
{
	half4 K = half4(1, 0.6666, 0.3333, 3);
	half3 p = abs(6*frac(hsv.xxx+K.xyz)-K.www);
	return hsv.z * lerp(K.xxx, saturate(p-K.xxx), hsv.y);
}

// Get the dyeing result based on H for color rgb.
float3 Dyeing(float3 rgb, in float H)
{
	float4 p = lerp(float4(rgb.bg, -1, 0.6666666), float4(rgb.gb, 0, -0.3333333), step(rgb.b, rgb.g));
	float4 q = lerp(float4(p.xyw, rgb.r), float4(rgb.r, p.yzx), step(p.x, rgb.r));
	float d = q.x - min(q.w, q.y);

	return q.x * lerp(float3(1,1,1), saturate(abs(6*frac(abs(q.z + (q.w-q.y) / (6*d+1.0e-7)) + H + float3(1,0.6666666,0.3333333)) - 3) - 1),
		d / (q.x+1.0e-7));
}
half3 RGB2YIQ(in half3 rgb)
{
	half3 yiq;
	yiq.r = dot(half3(0.299, 0.587, 0.114), rgb);
	yiq.g = dot(half3(0.596, -0.274, -0.322), rgb);
	yiq.b = dot(half3(0.211, -0.523, 0.312), rgb);
	return yiq;
}

half RemapValeRange(half input, half inputLow, half inputHigh,half inputTargetLow,half inputTargetHeight)
{
	return (input-inputLow)/(inputHigh-inputLow) * (inputTargetHeight-inputTargetLow) + inputTargetLow;
}

half3 BlendAngleCorrectedNormal(in half3 BaseNomral, in half3 AdditionalNormal)
{
	BaseNomral.b +=1 ;
	AdditionalNormal.rg *=-1;
	half3 blend  = dot(BaseNomral,AdditionalNormal);
	blend *= BaseNomral;
	AdditionalNormal*=BaseNomral.b;
	blend -=AdditionalNormal;
	return  blend;
}

half3 YIQ2RGB(in half3 yiq)
{
	half3 rgb;
	rgb.r = dot(half3(1, 0.956, 0.621), yiq);
	rgb.g = dot(half3(1, -0.272, -0.647), yiq);
	rgb.b = dot(half3(1, -1.106, 1.703), yiq);
	return saturate(rgb);
}
half3 DyeingYIQ(in half3 rgb, in half3 uwv)
{
	half3 yiq = RGB2YIQ(rgb);
	half3 dye = yiq;
	dye.r = yiq.r; // dot((1,0,0), yiq)
	dye.g = dot(half3(0, uwv.x, uwv.y), yiq);
	dye.b = dot(half3(0, -uwv.y, uwv.x), yiq);
	dye *= uwv.zzz;
	return YIQ2RGB(dye);
}

// Get the gray result for color rgb.
float3 Graying(float3 rgb, in float R)
{
	float gray = dot(rgb, float3(0.2126, 0.7152, 0.0722));
	return rgb * R + float3(gray, gray, gray) * (1.0 - R);
}

half2 HalfToR8G8(half x)
{
	half k = x * 255;
	return half2( floor(k) / 255.0, frac(k) );
}
half R8G8ToHalf(half2 x)
{
	return x.r + x.g/255.0;
}

float2 FloatToR8G8(float x)
{
	float k = x * 255;
	return float2( floor(k) / 255.0, frac(k) );
}

float R8G8ToFloat(float2 x)
{
	return x.r + x.g/255.0;
}

half CalcVignetting(half2 uv, half cVignettingDarkenss)
{
	half2 offset = uv - half2(0.5,0.5);
	half Vignetting = max(0, 1 - cVignettingDarkenss * dot(offset, offset));
	Vignetting *= Vignetting;
	if (@_HDR)
		Vignetting *= Vignetting;
	return Vignetting;
}
half cheapContrast(half x, half contrast)
{	
	return saturate(lerp(0 - contrast, contrast + 1, x));
}

void heightLerp(in half heightA, in half heightB, in half sharpness, in out half blendAlpha, out half blendHeight)
{
	// half finalMask = blendMask * 2 + (height - 1);
	// finalMask = saturate(finalMask);
	// finalMask = cheapContrast(finalMask, contrast);
	// return lerp(A, B, finalMask);
	half maskA = heightA * (1 - blendAlpha) + (1 - blendAlpha);
	half maskB = heightB * blendAlpha + blendAlpha;
	half heightDist = max(maskA, maskB) - max(sharpness, 0.001);
	maskA = max(maskA - heightDist, 0);
	maskB = max(maskB - heightDist, 0);
	blendAlpha = maskB / (maskA + maskB);
	blendHeight = lerp(heightA, heightB, blendAlpha);
}

half LinearSine(float speed)
{
	speed = frac(speed);
	half A = speed * 2;
	half B = (1 - speed) * 2;
	return lerp(A, B, floor(A));
}

float Rayleigh(float mu)
{
  return 3.0 / 4.0 * 1.0 / (4.0*PI)*(1.0 + mu * mu);
}

float Mie(float mu, float g)
{
  // Henyey-Greenstein phase function
  return (1.0 - g * g) / ((4.0 * PI) * pow(1.0 + g * g - 2.0 * g * mu, 1.5));
}

#define LightRangeMask(D2, InvSquareRange) ( Square(saturate(1 - Square(D2 * InvSquareRange))) )
#define CalcDistanceAttenuation(D2, LightData) ( 1 / (D2*LightData[1].w+1) )

float FogTransmittance(in float3 cameraToWorld)
{
	float D = max(0, length(cameraToWorld) - FOG_OFFSET);
	//half tLen = FOG_DENSITY * exp(-k*(CameraPos.y-BaseH)) * D * (1-exp(-k*DIR.y)) / (k*DIR.y)
	//FOG_EXP_BASE = FOG_DENSITY * exp(-k*(CameraPos.y-BaseH))/k
	// k = 5 / FOG_HEIGHT
	float tLen = FOG_EXP_BASE;
	if (abs(cameraToWorld.y)>EPSILON) 
		tLen *= (1-exp(-FOG_EXP_K * cameraToWorld.y)) / (cameraToWorld.y);
	else//todo: remove "/k" from FOG_EXP_BASE
		tLen *= FOG_EXP_K;
	tLen = max(0, tLen);
	float T = exp(-D * tLen);
	return max(0.0001, lerp(FOG_MINT, 0.999, T));
}

struct MaterialShadingParameters
{
    half3 BaseColor;
	half Metallic;
	half Roughness;
	half Specular;
	half3 SpecularColor;
	half3 DiffuseColor;
	half3 WorldNormal;
	half3 Emission;
	// Precomputed lihgting value, sampled from lightmap(static models) or SHVolume/PointCloud/global light probe(dynamic models)
    // indirect lighting + baked point lights's direct lighting
	half3 GILighting;
    // Precomputed shadow factor, sampled from lightmap(static models) or SHVolume/PointCloud/global light probe(dynamic models)
	half InShadow;
	half AO;
    half Alpha;
	half3 SubsurfaceColor;
	half Opacity;
	half Thickness;
	half4 CustomData;
	half3 WorldOffset;
	half RainOcclusion;
};

void InitShadingParameters(in out MaterialShadingParameters MaterialParameters)
{
	MaterialParameters.BaseColor = half3(1, 1, 1);
	MaterialParameters.Metallic = 0;
	MaterialParameters.Roughness = 1.0;
	MaterialParameters.Specular = 0.5;
	MaterialParameters.SpecularColor = half3(0.04,0.04,0.04);
	MaterialParameters.DiffuseColor = half3(1,1,1)/PI;
	MaterialParameters.WorldNormal = half3(0, 1, 0);
	MaterialParameters.Emission = half3(0, 0, 0);
	MaterialParameters.GILighting = half3(0, 0, 0);
	MaterialParameters.InShadow = 0;
	MaterialParameters.AO = 1;
	MaterialParameters.Alpha = 1;
	MaterialParameters.SubsurfaceColor = half3(0, 0, 0);// half3(1, 1, 1)
	MaterialParameters.Opacity = 0.5;
	MaterialParameters.Thickness = 0;
	MaterialParameters.CustomData = half4(0, 0, 0, 0);
	MaterialParameters.WorldOffset = half3(0, 0, 0);
	MaterialParameters.RainOcclusion = 1;
}

// half distanceSquared(half2 A, half2 B)
// {
//     A -= B;
//     return dot(A, A);
// }
float distanceSquared(float2 A, float2 B)
{
    A -= B;
    return dot(A, A);
}
// half distanceSquared(half3 A, half3 B)
// {
//     A -= B;
//     return dot(A, A);
// }
float distanceSquared(float3 A, float3 B)
{
    A -= B;
    return dot(A, A);
}

float GetDeviceZ(float depth)
{
	float fMinusN = CameraInfo.y - CameraInfo.x;
	return (CameraInfo.x / depth - CameraInfo.x) / fMinusN;
}

// high frequency dither pattern appearing almost random without banding steps
//note: from "NEXT GENERATION POST PROCESSING IN CALL OF DUTY: ADVANCED WARFARE"
//      http://advances.realtimerendering.com/s2014/index.html
// Epic extended by FrameId
// ~7 ALU operations (2 frac, 3 mad, 2 *)
// @return 0..1
float InterleavedGradientNoise( float2 uv, float FrameId )
{
	uv += FrameId * (float2(47, 17) * 0.695f);

	const float3 magic = float3( 0.06711056f, 0.00583715f, 52.9829189f );
	return frac(magic.z * frac(dot(uv, magic.xy)));
}

half henyeyGreenstein(half lightDotEye, half g)
{
	half g2 = g * g;

	//fast HG Function <- but not right
	// return ((1.0f - g2) / (1.0f + g2 - 2.0f * g * lightDotEye)) * 0.099577f;

	//src HG Function
	return ((1.0f - g2) / pow((1.0f + g2 - 2.0f * g * lightDotEye), 1.5f)) * 0.079577f;

	//src Messiah HG Function
	// return ( (1.0f - g2) / pow((1.0f + g2 - 2.0f * g * lightDotEye), 1.5f)) * 0.25f;
}

half PhaseFunction(half g, half CosTheta, half Ext)
{
	half w1 = 1.0 * 50.0;
	half rst = w1 * Ext * henyeyGreenstein(CosTheta, g);

	// half w2 = cCloudLightingPhaseControl.y * 50.0;
	// rst += w2 * Ext * henyeyGreenstein(CosTheta, 0.1 * g);

	//	float w0 = LightingControl.y * 0.2;
	//rst += Ext * 0.1 * w0 * henyeyGreenstein(CosTheta , g);

	return rst;
}

half lightEnergy(float lightDotEye, float densitySample,float cPhaseG)
{
	/*0611M_Remove unusual Const*/
	float rst = PhaseFunction(cPhaseG, lightDotEye, 1.0);			// Get Phase Function
	// rst *= saturate(exp(-densitySample)); 						// Beer Lambert's Law //densitySample is Pixel Density not Density Alone point toward sun.
	return rst;
}
#endif
