//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Constants.fxh>
///	@path	~/System/Shader/Include/
///	@date	2010/05/04
///	@brief	Shader constants definition.

#ifndef __FILE_CONSTANTS_FXH__
#define __FILE_CONSTANTS_FXH__
#include "Definitions.fxh"
#include "Config.fxh"

// Global constants.
cbuffer Global
{
	float4x4	ViewProj			: ViewProj 				;
	float4x4	ShadowViewProjTexs0	: ShadowViewProjTexs0	;
	float4x4	ShadowViewProjTexs1	: ShadowViewProjTexs1	;
	float4x4	ShadowViewProjTexs2	: ShadowViewProjTexs2	;
	float4		CSMShadowBiases		: CSMShadowBiases		;
	float4		CSMShadowSlopBiases : CSMShadowSlopBiases 	;
	float4		CameraPos			: CameraPos 			;
	float4		CameraInfo			: CameraInfo			; // near, far, invfar, fov
	float4		ScreenInfo			: ScreenInfo			;
	float4		ScreenColor			: ScreenColor			;
	float4		FogInfo				: FogInfo				;
	float4		FogColor			: FogColor				;
	float4		EnvInfo				: EnvInfo				;
	float4		SunDirection		: SunDirection			;
	float4		SunColor			: SunColor				;
	float4		AmbientColor		: AmbientColor			;
	float4		ShadowColor			: ShadowColor			;
	float4		ReflectionPos		: ReflectionPos			;
	float4		Misc				: Misc					;
	float4		VolumetricFogParam	: VolumetricFogParam	;
	float4x4	ShadowViewProjTexs3	: ShadowViewProjTexs3	;

	float4		VTParam0			: VTParam0				;
	float4		VTParam1			: VTParam1				;
	float4		VTParam2			: VTParam2				;
	float4		VTViewpoint			: VTViewpoint			;
	float4		VTIndMapUVTransform	: VTIndMapUVTransform	;

	float4		SunFogColor			: SunFogColor			;

	float4		WindParam			: WindParam				;
	float4		ScreenMotionGray	: ScreenMotionGray		;//mode {[0, 0.3) spread, [0.3, 0.6) none, [0.6, 1] shrink}, speed, MaxDistande, StartTime
	float4 		GIInfo				: GIInfo				; // x=renderOption aoBiasStatic

	float4 		VXGIParam[8]		: VXGIParam				; // VXGI
	float4 		GlobalBurnParam		: GlobalBurnParam		= {0, 0, 0,0};
	float4 		CSMCacheIndexs		: CSMCacheIndexs		;
};

#define ScreenMGrayModeSpread	(ScreenMotionGray.x < -0.1)
#define ScreenMGrayModeShrink	(ScreenMotionGray.x > 0.1)
#define ScreenMGraySpeed 		(ScreenMotionGray.y)
#define ScreenMGrayMaxDistance 	(ScreenMotionGray.z)
#define ScreenMGrayStartTime 	(ScreenMotionGray.w)

#define GameTime					(CameraPos.w)
#define LITMAP_GISCALE				(AmbientColor.w)
#define RainRate					(EnvInfo.x) //[0-1]
#define SnowRate					(EnvInfo.y) //[0-1]
#define ThunderRate					(EnvInfo.z) //[1-10]
#ifndef ReflectionProbeLum
	#define ReflectionProbeLum		(EnvInfo.w)
#endif
#ifndef REFLECTIONPOS 
	#define  REFLECTIONPOS 			(ReflectionPos)
#endif

#define RainNormalXZScale			0.3
#define RainNormalYScale			0.1
#define RainNormalTimeScale			0.1

#define FOG_MINT					(FogColor.w)
#define FOG_OFFSET					(FogInfo.x)
#define FOG_DENSITY 				(FogInfo.y)
#define INVFOG_HEIGHT2				(FogInfo.z) //-1 / (2.0*FOG_HEIGHT)
#define FOG_BASE2					(FogInfo.w) //1.0 - (CameraPos.y-FOG_BASE) / FOG_HEIGHT

#define FOG_EXP_K					(FogInfo.z) //5.0/FOG_HEIGHT
#define FOG_EXP_BASE				(FogInfo.w) //d*exp(-k*(CameraPos.y-BaseH))/k

#define SHADOW_RATE 				(ShadowColor.w)
#define RenderToMip					(SunDirection.w)

#define ReflectionScale				(Misc.x)
#if SD_HAS_FOCUS_DISTANCE
	#define FOCUS_DISTANCE			(Misc.y)
#else
	#define FOCUS_DISTANCE			(300)
#endif
#define RainNormalRate				(Misc.z)
#define TemporalAASample			(Misc.w)

// for deferred
#define AO_OFFSET					ShadowColor.x
#define AO_SCALE					ShadowColor.y


#define VTBlockSizeInUV VTParam0.x
#define VTBlockContentSizeInUV VTParam0.y
#define VTBlockMarginInUV VTParam0.z
#define VTTerrainSizeInBlock VTParam0.w

#define VTBlendRange VTParam1.x
#define VTLastLayerSizeInBlock VTParam1.y
#define VTLodLayerCount VTParam1.z
#define VTIndirectionLayerSizeInBlock VTParam1.w

#define LocalHeightmapSize VTParam2.x
#define LocalHeightmapMinHeight VTParam2.y
#define LocalHeightmapMaxHeight VTParam2.z
#define VTBlockContentSizeInPixel VTParam2.w

#define VTViewpointInTerrain (VTViewpoint.xy)
#define VTViewpointInWorld (VTViewpoint.zw)

#define GIScale GIInfo.x

// LightPosRange, LightColorAtten, LightDir, FallOffCosHalfThetaPHi
cbuffer Lights
{
	float4		Lights0[4]			: Lights0				;
	float4		Lights1[4]			: Lights1				;
	float4		Lights2[4]			: Lights2				;
};

cbuffer DebugBuffer
{
	float4 		DebugFlag			: DebugData				;
};

#define LightMapSize 				(DebugFlag.y)
#define DiffuseMapSize 				(DebugFlag.z)
#define PsInstructions				(DebugFlag.w)
#define PsTextureLods				(DebugFlag.w)
#define TiangleDensity				(DebugFlag.w)
#define TiangleCount				(DebugFlag.w)

cbuffer Skeleton
{
	// Skinned constants.
	// 384/3 = 128 bones for hard skinning & linear blend skinning.
	// 384/2 = 192 bones for dual quaternion skinning.
	float4 SkeletonData[384]		: SkeletonData 			;
};

#endif//__FILE_CONSTANTS_FXH__

