//	--------------------------------------------------------------------
//	Copyright(C) 2006~2019 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<ShadingMode.fxh>
///	@path	Shader/
///	@date	2019/06/27
///	@brief	shading models definition

#ifndef __FILE_DEFERRED_SHADING_COMMON_FXH__
#define __FILE_DEFERRED_SHADING_COMMON_FXH__

#include "Common.fxh"
#include "BRDF.fxh"


#define SHADINGMODELID_NEED_CUSTOMDATA		0x10
#define SHADINGMODELID_NOGI					0x20
#define SHADINGMODELID_UNLIT				0
#define SHADINGMODELID_DEFAULT_LIT			1
#define SHADINGMODELID_DEFAULT_LIT_NO_GI	(1|SHADINGMODELID_NOGI)
#define SHADINGMODELID_PREINTEGRATED_SKIN	(2|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_SSS					(3|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_HAIR					(4|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_FUR					(5|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_EYE					(6|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_CLEAR_COAT			(7|SHADINGMODELID_NEED_CUSTOMDATA)
#define SHADINGMODELID_TWOSIDED_FOLIAGE		(8|SHADINGMODELID_NEED_CUSTOMDATA)
#define SHADINGMODELID_CLOTH				(9|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_LOD_LIT				10
#define SHADINGMODELID_IMP_LIT				(11|SHADINGMODELID_NEED_CUSTOMDATA)
#define SHADINGMODEL_EYETEAR				(12|SHADINGMODELID_NEED_CUSTOMDATA|SHADINGMODELID_NOGI)
#define SHADINGMODELID_NUM					0xff //<=0xF

#define NO_NEED_FILTER						0
#define NEED_FILTER							1


struct GBufferData
{
    int ShadingModelID;
    half3 BaseColor;
	half Metallic;
    half Specular;
	half Roughness;
	half3 WorldNormal;
	half3 Emission;
    // Precomputed shadow factor, sampled from lightmap(static models) or SHVolume/PointCloud/global light probe(dynamic models)
	half InShadow;
	half AO;
    // other datas
    half4 CustomData;
	half3 DiffuseColor;	//already apply Diffuse_Lambert(1/pi), yes it's brdf ready!
	half3 SpecularColor;
	half3 GILighting;
};

float EncodeShadingModelID(int ShadingModelID)
{
	return (float)ShadingModelID / (float)0xFF;
}
int DecodeShadingModelID(float channelVal)
{
	return (int)round(channelVal * (float)0xFF);
}
half3 EncodeBaseColor(half3 BaseColor)
{
	//TODO: use sRGB on the render target to give more precision to the darks
	return BaseColor;
}
half3 DecodeBaseColor(half3 BaseColor)
{
	//TODO: use sRGB on the render target to give more precision to the darks
	return BaseColor;
}
half3 EncodeSubsurfaceColor(half3 SubsurfaceColor)
{
	return SubsurfaceColor;
}
half3 ExtractSubsurfaceColor(GBufferData GBuffer)
{
	return GBuffer.CustomData.rgb;
}

half3 EncodeWorldOffset(half3 offset)
{
	return (offset + 8) / 16.0;
}
half3 DecodeWorldOffset(half3 offset)
{
	return offset * 16 - 8;
}

float EncodeShadowAndFilter(float shadow, int flag)
{
	int n1 = (int)round(shadow * 255.0);
	return (float)((n1 & 0xFE) | (flag & 0x01)) / (float)0xFF;
}
void DecodeShadowAndFilter(float shadowAndFilter, out float shadow, out int flag)
{
	int n = (int)round(shadowAndFilter * (float)0xFF);
	shadow = (float)(n & 0xFE) / 255.0f;
	flag = n & 0x01;
}
void DecodeShadow(float shadowAndFilter, out float shadow)
{
	int n = (int)round(shadowAndFilter * (float)0xFF);
	shadow = (float)(n & 0xFE) / 255.0f;
}
void DecodeFilter(float shadowAndFilter, out int flag)
{
	int n = (int)round(shadowAndFilter * (float)0xFF);
	flag = n & 0x01;
}

/*
    GBufferA[R16G16B16A16F]	: Emission.rgb
	GBufferB[R8G8B8A8]		: ShadingModelID, Specular, Roughness, InShadow
	GBufferC[R8G8B8A8]		: BaseColor.rgb, Metallic
	GBufferD[R8G8B8A8]		: CustomData
	GBufferE[R8G8B8A8]		: N.xyz, IBLIntensity
	GBufferF[R10G10B10A2]	: TempRT for OnChipMem optimization
	GBufferG[R32F]			: Depth

	[HEX]
    GBufferA[R10G10B10A2]	: Emission.rgb with lighting result
	GBufferB[R8G8B8A8]		: Roughness, Metallic, AO, ShadingModelID
	GBufferC[R8G8B8A8]		: BaseColor.rgb, Specular
	GBufferD[R8G8B8A8]		: CustomData
	GBufferE[R8G8B8A8]		: N.xyz, InShadow
	GBufferF[R10G10B10A2]	: GILighting//TempRT for OnChipMem optimization
	GBufferG[R32F]			: Depth
*/
void EncodeGBufferData(GBufferData GBuffer, out float4 OutGBufferA, out float4 OutGBufferB, out float4 OutGBufferC, out float4 OutGBufferD, out float4 OutGBufferE, out float4 OutGBufferF)
{
#if MATERIAL_SHADINGMODEL_UNLIT
	OutGBufferA.rgb = GBuffer.Emission.rgb;
	OutGBufferA.a = 1;
    OutGBufferB = float4(1, 0, 1, EncodeShadingModelID(GBuffer.ShadingModelID));
	OutGBufferC = 0;
	OutGBufferD = 0;
	OutGBufferE = 0;
	OutGBufferF.a = 0;
	OutGBufferF.rgb = GBuffer.GILighting * GBuffer.DiffuseColor;
#else
	OutGBufferA.rgb = GBuffer.Emission.rgb;
	OutGBufferA.a = 1;

	OutGBufferB.r = GBuffer.Roughness;
	OutGBufferB.g = GBuffer.Metallic;
	OutGBufferB.b = GBuffer.AO;
	OutGBufferB.a = EncodeShadingModelID(GBuffer.ShadingModelID);

	OutGBufferC.rgb = EncodeBaseColor(GBuffer.BaseColor);
	OutGBufferC.a = GBuffer.Specular;

	OutGBufferD = GBuffer.CustomData;

	OutGBufferE.rgb = EncodeNormal(GBuffer.WorldNormal);
#ifdef HAS_WORLD_POSITION_OFFSET
	OutGBufferE.a = EncodeShadowAndFilter(GBuffer.InShadow, NEED_FILTER);
#else
	OutGBufferE.a = EncodeShadowAndFilter(GBuffer.InShadow, NO_NEED_FILTER);
#endif
	OutGBufferF = half4(0,0,0,GBuffer.Metallic);
	OutGBufferF.rgb = GBuffer.GILighting * GBuffer.DiffuseColor;
#endif
}

GBufferData DecodeGBufferData(float4 InGBufferB, float4 InGBufferC, float4 InGBufferD, float4 InGBufferE, float4 InGBufferF)
{
	GBufferData GBuffer;
	//GBufferA emission, lighting result

	//GBufferB
	GBuffer.Roughness 		= InGBufferB.r;
	GBuffer.Metallic		= InGBufferB.g;
	GBuffer.AO				= InGBufferB.b;
	GBuffer.ShadingModelID 	= DecodeShadingModelID(InGBufferB.a);

	//GBufferC
	GBuffer.BaseColor 		= DecodeBaseColor(InGBufferC.rgb);
	GBuffer.Specular		= InGBufferC.a;

	//GBufferD
	GBuffer.CustomData 		= ((GBuffer.ShadingModelID & SHADINGMODELID_NEED_CUSTOMDATA) != 0) ? InGBufferD : half4(0,0,0,0);

	//GBufferE
	GBuffer.WorldNormal 	= DecodeNormal(InGBufferE.xyz);
	GBuffer.WorldNormal 	= normalize(GBuffer.WorldNormal);

	float shadow;
	DecodeShadow(InGBufferE.a, shadow);
	GBuffer.InShadow 		= shadow;

	//GBufferF
	GBuffer.GILighting 		= InGBufferF.rgb;

	// derived from BaseColor, Metalness, Specular
	{
		GBuffer.SpecularColor = lerp(0.08 * GBuffer.Specular.xxx, GBuffer.BaseColor, GBuffer.Metallic);
		GBuffer.DiffuseColor = Diffuse_Lambert(GBuffer.BaseColor - GBuffer.BaseColor * GBuffer.Metallic);
	}
	return GBuffer;
}

void SetGBufferForShadingModel(in out GBufferData GBuffer, in const MaterialShadingParameters MaterialParameters)
{
	GBuffer.BaseColor = MaterialParameters.BaseColor;
	GBuffer.Metallic = MaterialParameters.Metallic;
	GBuffer.Specular = MaterialParameters.Specular;
	GBuffer.Roughness = max(MIN_ROUGHNESS, MaterialParameters.Roughness);
	GBuffer.WorldNormal = MaterialParameters.WorldNormal;
	GBuffer.Emission = MaterialParameters.Emission;
	GBuffer.InShadow = MaterialParameters.InShadow;
	GBuffer.AO = MaterialParameters.AO;//GetLuminance(MaterialParameters.GILighting);
	GBuffer.CustomData = MaterialParameters.CustomData;

#if MATERIAL_SHADINGMODEL_UNLIT
	GBuffer.ShadingModelID = SHADINGMODELID_UNLIT;
#elif MATERIAL_SHADINGMODEL_DEFAULT_LIT
	GBuffer.ShadingModelID = SHADINGMODELID_DEFAULT_LIT;
#elif MATERIAL_SHADINGMODEL_PREINTEGRATED_SKIN
	GBuffer.ShadingModelID = SHADINGMODELID_PREINTEGRATED_SKIN;
	GBuffer.CustomData.rgb = EncodeNormal(GBuffer.CustomData.rgb);
	GBuffer.CustomData.a = MaterialParameters.Thickness;
#elif MATERIAL_SHADINGMODEL_SSS
	GBuffer.ShadingModelID = SHADINGMODELID_SSS;
	// sss的CustomData看起来没用，也就是GBufferD空着
	GBuffer.CustomData.rgb = EncodeNormal(GBuffer.CustomData.rgb);
	GBuffer.CustomData.a = MaterialParameters.Thickness;
#elif MATERIAL_SHADINGMODEL_HAIR
	GBuffer.ShadingModelID = SHADINGMODELID_HAIR;
#elif MATERIAL_SHADINGMODEL_FUR
	GBuffer.ShadingModelID = SHADINGMODELID_FUR;
#elif MATERIAL_SHADINGMODEL_EYE
	GBuffer.ShadingModelID = SHADINGMODELID_EYE;
	GBuffer.CustomData.a = MaterialParameters.Opacity;
#elif MATERIAL_SHADINGMODEL_EYETEAR
	GBuffer.ShadingModelID = SHADINGMODELID_DEFAULT_LIT;
	//GBuffer.CustomData.a = MaterialParameters.Alpha;	
#elif MATERIAL_SHADINGMODEL_CLEAR_COAT
	GBuffer.ShadingModelID = SHADINGMODELID_CLEAR_COAT;
#elif MATERIAL_SHADINGMODEL_TWOSIDED_FOLIAGE
	GBuffer.ShadingModelID = SHADINGMODELID_TWOSIDED_FOLIAGE;
	GBuffer.CustomData.rgb = EncodeWorldOffset(MaterialParameters.WorldOffset); // foliage customedata用于植被飘动的位置offset记录，从而计算阴影的时候可以反变换回去求自阴影
	GBuffer.CustomData.a = MaterialParameters.Thickness;
#elif MATERIAL_SHADINGMODEL_CLOTH
	GBuffer.ShadingModelID = SHADINGMODELID_CLOTH;
#elif MATERIAL_SHADINGMODEL_LOD_LIT
	GBuffer.ShadingModelID = SHADINGMODELID_LOD_LIT;
#elif MATERIAL_SHADINGMODEL_IMP_LIT
	GBuffer.ShadingModelID = SHADINGMODELID_IMP_LIT;	
#else
	// missing shading model
#endif

if(@_SKELETON)
	if(GBuffer.ShadingModelID == SHADINGMODELID_DEFAULT_LIT)
	GBuffer.ShadingModelID = SHADINGMODELID_DEFAULT_LIT_NO_GI;
	// derived from BaseColor, Metalness, Specular
	{
		GBuffer.SpecularColor = lerp(0.08 * GBuffer.Specular.xxx, GBuffer.BaseColor, GBuffer.Metallic);
		GBuffer.DiffuseColor = Diffuse_Lambert(GBuffer.BaseColor - GBuffer.BaseColor * GBuffer.Metallic);
	}

	//GBuffer.Emission += MaterialParameters.GILighting;
#if MATERIAL_SHADINGMODEL_SSS
	// 如果是sss皮肤则diffuse存储到GBUfferA，用于blur
	GBuffer.Emission += MaterialParameters.GILighting;
	GBuffer.GILighting = float3(0, 0, 0);
#else
	// 否则走对间接光预乘DiffuseColor，sunlightigng pass那边就可以直接被叠加了
	GBuffer.GILighting = MaterialParameters.GILighting;
#endif
}

#define PBRQualityLevel_None 	0
#define PBRQualityLevel_Normal 	1
#define PBRQualityLevel_High 	2

struct LightingData
{
	/*
		Though FalloffColor * Atten * NoL always means irradiance/iluminance, FalloffColor and Atten are defined in different ways.
		1, for point/spot lights
			FalloffColor is light intensity(i.e., point light = Φ / 4PI, spot light = Φ / [4PI*(1-cos(phi))].
			Atten is 1 / (1+SquareDistance), "1+" is added to prevent dividing by zero, maybe 0.01 is better!
		2, for sun light
			FalloffColor is projected irradiance/iluminance.
			Atten is always 1.
		3, for rect light
			FalloffColor is "radiant exitance" which emitted uniformly from the suface of light(e.g., Φ/A).
			Atten is "the transformed Formfactor"(without dot N) which project rect to the shading point.
			NoL is SphereIntegral.
	*/
	half3 FalloffColor;
	float Atten;
    //Normalized direction to light source
    float3 L;
    //Normalized view direction to eye
    float3 V;
	//Normalized world normal
    float3 N;
	//[-1,1]
	half _NoL;
	//[0,1] saturated
	half NoL;
    //[0,1]
    half InShadow;
	//[0,1,2]
	int PBRQualityLevel;
	//BTDF
	bool TranslucencyLighting;

	//Area Light Data
	bool 		IsSphericalLight;
	half 		RadiusTan;
	bool 		IsRectLight;
	bool		IsDiskLight;
	float3		Origin;
	float3x3	Axis;
    float2		Extent;
	bool		TwoSided;
};

//Li * brdf * cos(theta)
struct LightingResult
{
    half3 DiffuseLighting;//need * DiffuseColor
    half3 SpecularLighting;//final color
};

struct BxDFContext
{
	float NoV;
	float NoL;
	float VoL;
	float NoH;
	float VoH;
};

void Init(inout BxDFContext Context, in LightingData LitData)
{
	// Context.NoL = LitData._NoL;
	// Context.NoV = dot(LitData.N, LitData.V);
	// Context.VoL = dot(LitData.V, LitData.L);
	// float InvLenH = rsqrt( 2 + 2 * Context.VoL );
	// Context.NoH = saturate( ( Context.NoL + Context.NoV ) * InvLenH );
	// Context.VoH = saturate( InvLenH + InvLenH * Context.VoL );
	//Context.NoV = saturate(abs(Context.NoV) + 1e-5);

	// {HEX} do not use SphericalLight
	Context.NoL = LitData.NoL;
	Context.NoV = saturate(dot(LitData.N, LitData.V));
	Context.VoL = saturate(dot(LitData.V, LitData.L));
	float3 H = normalize(LitData.V + LitData.L);
	Context.NoH = saturate(dot(LitData.N, H));
	Context.VoH = saturate(dot(LitData.V, H));
	// {HEX} end
}
#endif//__FILE_DEFERRED_SHADING_COMMON_FXH__
