//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Effect.fxh>
///	@path	~/System/Shader/Include
///	@date	2020/09/09
///	@brief	Common calculations for effect material.

// Structs' Defination is in Effect_Common.fxh
// EFFECT_VS_GET_PARAMETERS for use VS_GetParameter() for Vertex Modify

#include "Lighting.fxh"

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.


VertexShadingOutput EffectUnifyVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);

	float4 LocalPosition = IN.Position;
	float3 LocalNormal = float3(0,1,0);
	[@_NORMAL]
	LocalNormal = (float3)Expand((half3)IN.Normal.xyz);
	[@_TANGENT]
	half3 Binormal = (float3)Expand((half3)IN.Binormal.xyz);
	[@_TANGENT]
	half3 Tangent = (float3)Expand((half3)IN.Tangent.xyz);
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, LocalPosition.xyz);
			LocalNormal = Matrix4x3TransformN(Mx, My, Mz, LocalNormal);
		
			[@_TANGENT]
			Tangent.xyz = Matrix4x3TransformN(Mx, My, Mz,IN.Tangent.xyz);
			[@_TANGENT]
			Binormal.xyz = Matrix4x3TransformN(Mx, My, Mz,IN.Binormal.xyz);
			
		}
	}

	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, LocalPosition.xyz), 1);
		OUT.WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, LocalNormal));
		OUT.TexCoord.xy = IN.TexCoord;
		OUT.EffectColor = IN.EffectColor;
		OUT.EffectParam = IN.EffectParam;
		// OUT.cTintColor1 = IN.cTintColor1;
		[@_TANGENT]
		OUT.WorldTangent.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, Tangent));
		[@_TANGENT]
		OUT.WorldBinormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, Binormal));
		
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(LocalPosition.xyz, 1)).xyz, 1);
		OUT.WorldNormal = normalize(MulMat3x4((float3x3)World, LocalNormal));
		OUT.TexCoord.xy = IN.TexCoord * SubUV.xy + SubUV.zw;

		[@_TANGENT]
		OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, Tangent));
		[@_TANGENT]
		OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, Binormal));
		
	}

	OUT.WorldPosition = WorldPosition.xyz;

	#ifdef EFFECT_VS_GET_PARAMETERS
	VS_GetParameters(IN,OUT);
	#endif

	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz,1));
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.HPosition.z -= CameraInfo.x * DepthBias / (DepthBias + OUT.HPosition.w);
	OUT.ViewDirection = normalize( CameraPos.xyz - OUT.WorldPosition.xyz );

	//fog
	OUT.TexCoord.w = FogTransmittance(OUT.WorldPosition.xyz - CameraPos.xyz);

	if (@_VERTEXCOLOR)
	{
		[@_VERTEXCOLOR]
		OUT.VertexColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.VertexColor = float4(1, 1, 1, 1);
	}

	[@_HDR]
	OUT.VertexColor.rgb *= HDRScale;

	return OUT;
}


void InitFogLightingEffect(/*in float2 stc,*/in VertexShadingOutput IN, in out float4 OUT)
{
	//issue : IN.DiffLighting.w is different to VolumetricFog-off
// 	if (@_FOG && @_VolumetricFog)
// 	{
// 		// float2 spos = IN.HPosition.xy;
// 		// [@_D3D9]
// 		// spos += 0.5;
// 		// float2 stc = spos * ScreenInfo.zw;
// 		float depth = IN.DiffLighting.w;
// 		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
// 		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
// 	}
	if (@_FOG && !@_VolumetricFog)
	{
		float4 fogFinalColor = float4(0,0,0,0);
		FogLightingAndScatter(IN.TexCoord.w, fogFinalColor.xyz, IN.ViewDirection, SunDirection);
		OUT.rgb = lerp(fogFinalColor.xyz * OUT.a, OUT.rgb, IN.TexCoord.w);
	}
}

void InitWorldNormal(in VertexShadingOutput IN,in out EffectShadingParameters materialParamters)
{

	float3 P = IN.WorldPosition;

	float3 N = IN.WorldNormal.xyz;

	half4 rst = tNormalMap.SampleBias(sNormalMap_Sampler, IN.TexCoord.xy, 0);
	half3 tangentNormal = rst.xyz;
	[@_DIRECTX]//always use bc5 for pc
	tangentNormal.z = 1;
	tangentNormal = (2 * tangentNormal) - 1;
	tangentNormal.xy *= 1;

	[@_TANGENT]
	N = normalize(IN.WorldTangent * tangentNormal.x + IN.WorldBinormal * tangentNormal.y + N * tangentNormal.z);

	materialParamters.WorldNormal = N;	
	
}

void InitBaseColorCommon(in VertexShadingOutput IN,in out EffectShadingParameters materialParamters)
{
	// BaseColor
	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	float4 baseColor = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	baseColor.rgb = DecodeGamma(baseColor.rgb);
	
	materialParamters.BaseColor = baseColor;	

	// Set Normal Here once , When @_TANGENT is false,Will not call InitWorldNormal()
	materialParamters.WorldNormal = IN.WorldNormal;
}

void SoftParticle(in VertexShadingOutput IN,in out EffectShadingParameters materialParamters, float4 screenCoord )
{
	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

	float depth = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc));
	depth *= CameraInfo.y;

	float fade = saturate((depth - IN.TexCoord.z) * InvFadeDistance);

	materialParamters.BaseColor.rgb = lerp(materialParamters.BaseColor.rgb,fadeColor.rgb,1-fade.xxx);
	materialParamters.BaseColor *= fade;
}

//////////////////////////////////////////////////////////////////////////
// technique Clip.

float4 EffectClipShading(in VertexShadingOutput IN) : SV_Target
{
	return float4(0, 0, 0, 0);
}
#ifndef USE_DEBUG_CLIPPING
#define USE_DEBUG_CLIPPING 1
#endif

#if USE_DEBUG_CLIPPING
technique Clip
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectClipShading();
}
#endif

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 EffectDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.05, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

#ifndef USE_DEBUG_TECHNIQUE
#define USE_DEBUG_TECHNIQUE 1
#endif

#if USE_DEBUG_TECHNIQUE
technique Debug
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectDebugShading();
}
#endif


void FetchPrecomputedLighingParticle(in out MaterialShadingParameters MaterialParameters)
{
	//Modified From PBR_Common:FetchPrecomputedLighing()
	//No LightMap For Particle
	//Only Calc Dymnamic GI Lighting as All PBR Material On Deferred
	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
	MaterialParameters.GILighting *=  MaterialParameters.AO;
}

