
//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;

	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	[@_NORMAL]
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
	[@_INSTANCED]
	float4		EffectColor		: TEXCOORD6;
	[@_INSTANCED]
	float4		EffectParam		: TEXCOORD7;
	// [@_INSTANCED]
	// float4		TintColor1		: TEXCOORD7;
};

//////////////////////////////////////////////////////////////////////////
// Intermediate results.

/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate and depth for surface.
	half3		WorldNormal		: TEXCOORDN;			//	World Normal
	half3		ViewDirection	: TEXCOORDN;			//	View direction
	float4		VertexColor		: TEXCOORDN;			//	Per-vertex color.	Used in HDR Scale...
	float3		WorldPosition   : TEXCOORDN;
	[@_INSTANCED]
	float4		EffectColor		: TEXCOORDN;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	[@_INSTANCED]
	float4		EffectParam		: TEXCOORDN;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	// [@_INSTANCED]
	// float4		cTintColor1		: TEXCOORDN;			//	Parameter By Life
	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;
};
void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{	
	OUT.HPosition = float4(0,0,0,0);
	OUT.TexCoord = float4(0,0,0,0);
	OUT.WorldNormal = half3(0,0,0);
	OUT.ViewDirection = half3(0,0,0);
	OUT.VertexColor = half4(0,0,0,0);
	OUT.WorldPosition = float3(0,0,0);
 
	[@_INSTANCED]
	OUT.EffectColor = float4(0,0,0,0);
	
	[@_INSTANCED]
	OUT.EffectParam = float4(0,0,0,0);

	// [@_INSTANCED]
	// OUT.cTintColor1 = float4(0,0,0,0);

	[@_TANGENT] 
	OUT.WorldTangent = half4(0,0,0,0);
	[@_TANGENT]
	OUT.WorldBinormal= half4(0,0,0,0);
}

struct EffectShadingParameters
{
    half4 BaseColor;
    half Metallic;
    half Roughness;
    half Specular;
    half3 WorldNormal;
	half Depth;
};

void InitEffectShadingParameters(in out EffectShadingParameters MaterialParamters)
{
    MaterialParamters.BaseColor = half4(1,1,1,1);
    MaterialParamters.Metallic = 0;
    MaterialParamters.Roughness = 1;
    MaterialParamters.Specular = 0.5;
    MaterialParamters.WorldNormal = half3(1,0,0);
	MaterialParamters.Depth = 0;
    // MaterialParamters.Alpha = 1;
}

//////////////////////////////////////////////////////////////////////////
// Utilities

float4 EncodePremultipliedAlphaAndBloom(const float4 IN, const float invEmission, const float fog, const float bloom)
{
	float4 OUT = IN;
	OUT.rgb *= OUT.a;
	OUT.a *= invEmission;
	OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, fog);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * bloom);
	return OUT;
}

float4 GetEffectColor(in const VertexShadingOutput IN)
{
	// EffectColor(ColorByLife)
	if (@_INSTANCED)
	{
		return IN.EffectColor * EffectColor;
	}
	else
	{
		return EffectColor;
	}
}

float4 GetEffectParam(in const VertexShadingOutput IN)
{
	// EffectParam(life, life * totalLife, 1 - emission, ControlByLife)
	if (@_INSTANCED)
	{
		return IN.EffectParam;
	}
	else
	{
		return EffectParam;
	}
}

// Note: GetParameterByLife will use VertexAssemble.TintColor1 
// But it won't be pass in struct when GetParameterByLife is off
// float3 GetEffectParamByLife(in VertexShadingOutput IN)
// {
// 	// cTintColor1(ParameterByLife.xyz,Unknow)
// 	if (@_INSTANCED)
// 	{
// 		return IN.TintColor1.xyz;
// 	}
// 	else
// 	{
// 		return cTintColor1.xyz;
// 	}	
// }

