#ifndef __FILE_HEXTAA_FXH__
#define __FILE_HEXTAA_FXH__

/*
	Config - TAA_HEX
	Low------------------
		bool DepthCull,				True
		bool ToneMapping,			False
		bool usingVBuffer,			True
		bool usingVarianceClip,		False
		bool fastMode,				True
		bool NearestVelocity,		False
		half Edge_Fix				0.25
		half Sample_Range			0.5
	
	medium---------------
		bool DepthCull,				True
		bool ToneMapping,			False
		bool usingVBuffer,			True
		bool usingVarianceClip,		True
		bool fastMode,				True
		bool NearestVelocity,		True
		half Edge_Fix				0.25
		half Sample_Range			0.5

	high-----------------
		bool DepthCull,				True
		bool ToneMapping,			True
		bool usingVBuffer,			True
		bool usingVarianceClip,		True
		bool fastMode,				false
		bool NearestVelocity,		True
		half Edge_Fix				0.25
		half Sample_Range			0.5

*/

float4 TAA_HEX(
	v2f IN,						//v2f
	bool DepthCull,				//Depth Cull , using LastFrame Depth for check failed pixel , expensive
	bool ToneMapping,			//Convert Color to YCoCg Space for Anti-Ghosting
	bool usingVBuffer,			//Using VBuffer for motion Objects' Reporjection
	bool usingVarianceClip,		//using Variance Clip
	bool fastMode,				//Fast Mode , less Samples , more Flicker 
	bool NearestVelocity,		//NearestVelocity , For Fix Edge's Aliasing when moving
	half Edge_Fix,				//Color Clamp Range Scale, default = 0.25
	half Sample_Range			//Sample Color/Depth Offset , Default = 1,Lower mean flicker but sharp
) : SV_Target
{
    bool IgnoreHistory = false;
	half2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	half2 lastScreenUV = IN.JitteredUV.xy;
	half DepthDiff = 0;
	half2 rcpcRTSizexy = rcp(cRTSize.xy) ;

	//Preset
	static const int2 kOffsets3x3[9] =
	{
		int2(-1, -1),
		int2( 0, -1),
		int2( 1, -1),
		int2(-1,  0),
		int2( 0,  0), // K
		int2( 1,  0),
		int2(-1,  1),
		int2( 0,  1),
		int2( 1,  1)
	};

	/*
		Sample Index For Different Count
		Kernal(4) is always at [0]
	*/
	static const int SampleIndexFull[9] 	= {4, 0, 1, 2, 3, 5, 6, 7, 8 };
	static const int SampleIndexCross[9] 	= {4, 0, 2, 6, 8, 0, 0, 0, 0 }; 
	static const int SampleIndexFast[9]		= {4, 0, 8, 0, 0, 0, 0, 0, 0 };

	static const half SampleWeightFull[9]	= {-0.1429,0.1429,0.1429,0.1429,0.1429,0.1429,0.1429,0.1429,0.1429};
	static const half SampleWeightCross[9]	= {-0.1429,0.2857,0.2857,0.2857,0.2857,0,0,0,0};
	static const half SampleWeightFast[9]	= {-0.1429,0.5714,0.5714,0,0,0,0,0,0};

	static const half FilterWeightFull[9]	= {0.4,0.075,0.075,0.075,0.075,0.075,0.075,0.075,0.075};
	static const half FilterWeightCross[9]	= {0.4,0.15,0.15,0.15,0.15,0,0,0,0};
	static const half FilterWeightFast[9]	= {0.4,0.3,0.3,0, 0, 0, 0, 0, 0};

	float curDepths[9];
	initFloat9(curDepths);
	float hisDepths[9];
	initFloat9(hisDepths);
	half4 colors[9];
	initHalf4c9(colors);

	int SampleIndices[9] = SampleIndexFull;
	half SampleWeight[9] = SampleWeightFull;
	half FilterWeight[9] = FilterWeightFull;

	int SampleCount = 1;
	if(fastMode)
	{
		SampleWeight = SampleWeightFast;
		SampleIndices = SampleIndexFast;
		FilterWeight = FilterWeightFast;
		SampleCount = 3;
	}
	else
	{
		SampleWeight = SampleWeightCross;
		SampleIndices = SampleIndexCross;
		FilterWeight = FilterWeightCross;
		SampleCount = 5;
	}


	//Sample Depth & Find Nearest Velocity Offset
	half2 VelocityOffset = half2(0,0);
	float3 worldPos = float3(0,0,0);
	curDepths[0] = tDepthBuffer.SampleLevel(sDepthBufferSampler,lastScreenUV ,0).x;
	float minDepth = curDepths[0];
	if(NearestVelocity || DepthCull)
	{
		[unroll]
		for(int icd = 1; icd < SampleCount; icd++)
		{
			int sampleIndex = SampleIndices[icd];
			curDepths[icd] = tDepthBuffer.SampleLevel(sDepthBufferSampler,lastScreenUV + rcpcRTSizexy * kOffsets3x3[sampleIndex],0).x;
			minDepth = min(curDepths[icd],minDepth);
			if(NearestVelocity) 
				VelocityOffset = curDepths[icd] < minDepth ? kOffsets3x3[sampleIndex] : VelocityOffset;
		}
	}

	worldPos = CalcHomogeneousPos(minDepth,( spos / cRTSize.zw)) + CameraPos.xyz;
	float4 lastPos = MulMat4x4(lastViewProjTex, float4(worldPos,1.0));
	float2 lastUV = (lastPos.xy / lastPos.w);


	//Sample Last Depth & Find Ghost Aera
	if(DepthCull)
	{
		float historyDepth = 0;
		half2 LastFrameJitteredUV = lastUV + IN.JitteredUV.zw;
		[unroll]
		for(int ihd = 0; ihd < SampleCount; ihd++)
		{
			int sampleIndex = SampleIndices[ihd];
			hisDepths[ihd] = tLastDepthMap.SampleLevel(sLastDepthSampler,LastFrameJitteredUV + Sample_Range * rcpcRTSizexy * kOffsets3x3[sampleIndex],0).x;
			historyDepth = max(historyDepth,hisDepths[ihd]);
		}
		DepthDiff = saturate(minDepth - historyDepth - 0.0001);
	}


	//Set Ignore History Flag for OutOfCamera & Depth Clip Result
	IgnoreHistory = false;
	IgnoreHistory = max(abs(lastUV.x - 0.5),abs(lastUV.y - 0.5)) > 0.5 ? true : IgnoreHistory;
	IgnoreHistory = step(DepthDiff, 0.0001) == 0 ? true : IgnoreHistory;


	//Sample Velocity Buffer & Calc UV
	half2 motion = half2(0,0);
	if(usingVBuffer)
	{
		motion = tVBuffer.SampleLevel( sVBufferSampler, IN.JitteredUV.xy + VelocityOffset , 0).xy;
		lastUV = (motion.x + motion.y == 0) ? lastUV : IN.TC.xy + motion.xy;
		IgnoreHistory = (motion.x + motion.y == 0) ? IgnoreHistory : false;
	}


	//Sample Screen Texture & Filter Color / Average Color 
	half4 fColor = half4(0,0,0,0);
	half4 average = half4(0,0,0,0);
	[unroll]
	for(int ic = 0; ic < SampleCount; ic++)
	{
		int sampleIndex = SampleIndices[ic];
		colors[ic] = tScreenMap.SampleLevel(sScreenSampler, lastScreenUV + Sample_Range * rcpcRTSizexy * kOffsets3x3[sampleIndex], 0); // issue
		if(ToneMapping) colors[ic] = RGB2YCoCg4(colors[ic]);
		average += colors[ic] * SampleWeight[ic];
		fColor += colors[ic] * FilterWeight[ic];
	}


	//Calc Motion info
	half2 fakeMotionInfo = (IN.TC - lastUV) * (minDepth) * 10.0 + motion * 5.0;
	half motionLength = length(fakeMotionInfo); 


	//Calc MinMax Neighboor
	half nudge = 0;
	if(ToneMapping)
    	nudge = lerp(4.0, 0.25, saturate(motionLength * 100.0)) * abs(average.x - colors[0].x);
	else
    	nudge = lerp(4.0, 0.25, saturate(motionLength * 100.0)) * abs(average.y - colors[0].y);

	half4 minimum = half4(10,10,10,10);
    half4 maximum = half4(0,0,0,0);
	half4 m1 = 0;
	half4 m2 = 0;
	if(usingVarianceClip)
	{
		[unroll]
		for(int iv = 0; iv <SampleCount; iv++)
		{
			half4 col = colors[iv];
			m1 += col;
			m2 += pow(col , 2);
		}
		
		m1 *= (1.0 / SampleCount);
		m2 *= (1.0 / SampleCount);

		half4 Variance = sqrt( abs(m2 - m1 * m1) );
		minimum = m1 - 1.25 * Variance - nudge * Edge_Fix;
		maximum = m1 + 1.25 * Variance + nudge * Edge_Fix;	

		minimum = min( minimum, colors[0] );
		maximum = max( maximum, colors[0] );
	}
	else
	{
		[unroll]
		for(int iv = 0; iv <SampleCount; iv++)
		{
			half4 col = colors[iv];
			maximum = max(col,maximum);
			minimum = min(col,minimum);
		}
		minimum -= nudge;
		maximum += nudge;
	}


	//Calc Blend Weight
	half weight = clamp(
		lerp(
				cLastFrameRate, 		//static
				cLastFrameRate2, 		//dynamic
				motionLength * 600.0 	//motion amplification 
			),
		cLastFrameRate2, 				//dynamic
		cLastFrameRate					//static
	); 
	

	//Sample History
  	half4 history = tLastScreenMap.SampleLevel(sLastScreenSample, lastUV, 0);	
	if(ToneMapping) history = RGB2YCoCg4(history);


	//Clip History Color & Fix Ghost
    history.rgb = clamp(history.rgb,minimum.xyz,maximum.xyz);
	if(IgnoreHistory)
		history.xyz = fColor.xyz;


	//Blend Current & History
    colors[0] = lerp(colors[0], history, weight);

	if(ToneMapping) colors[0] = YCoCg2RGB4(colors[0]);

	return colors[0];
}


#endif //__FILE_HEXTAA_FXH__
