#ifndef __FILE_LIGHTING_FXH__
#define __FILE_LIGHTING_FXH__

#include "VolumetricFogHelper.fxh"
#include "DeferredShadingCommon.fxh"
#define ESM_EXPONENT -30.0
texture2DArray tShadowMapArray : ShadowMapArray : layout(t12);
SamplerComparisonStateArray sShadowMapArraySampler : layout(s12)
{
	Texture = tShadowMapArray;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
};

//////////////////////////////////////////precomputed lighting result/////////////////////////////////////////
// getYlms_3
void getYlms_3(in half3 surfelNormal, out half c0, out half4 c14, out half4 c58)
{
	float x = surfelNormal.x, y = surfelNormal.y, z = surfelNormal.z;
	float P00=0.28209477663040161000f;
	c0 = P00;
	float P10=0.24430125951766968000f*2*z;
	float P11=-0.24430125951766968000f*2;
	float p1=x,n1=y;
	c14.x=P11*n1;
	c14.y=P10;
	c14.z=P11*p1;
	float z2=z*z;
	float p2=p1*x-n1*y,n2=n1*x+p1*y;
	float P20=0.15769578516483307000f*(-2.f+6.f*z2);
	float P21=-0.09104570746421814000f*12.f*z;
	float P22=0.04552285373210907000f*12.f;
	c14.w=P22*n2;
	c58.x=P21*n1;
	c58.y=P20;
	c58.z=P21*p1;
	c58.w=P22*p2;
}
#define K0_DIFF_Z0_P0 0.8862269427881898		//K0 * DIFF_Z0 * P0
#define K1_DIFF_Z1_P10 1.0233268352371807 		//K1 * DIFF_Z1 * P10
#define K1_DIFF_Z1_P11 -1.0233268352371807		//K1 * DIFF_Z1 * P11
#define K2_DIFF_Z2_P20_A -0.24770799551070435	//K2 * DIFF_Z2 * P20_A
#define K2_DIFF_Z2_P20_B 0.7431239865321131		//K2 * DIFF_Z2 * P20_B
#define K2_DIFF_Z2_P21 -0.8580856998394206		//K2 * DIFF_Z2 * P21
#define K2_DIFF_Z2_P22 0.4290428499197103		//K2 * DIFF_Z2 * P22
// f(w) = max(0, dot(N,w));
void calcIBLCoefficients(in half3 surfelNormal, out half c0, out half4 c14, out half4 c58)
{
	float x = surfelNormal.x, y = surfelNormal.y, z = surfelNormal.z;
	c0=K0_DIFF_Z0_P0;
	float P10=K1_DIFF_Z1_P10*z;
	float P11=K1_DIFF_Z1_P11;
	float p1=x,n1=y;
	c14.x=P11*n1;
	c14.y=P10;
	c14.z=P11*p1;
	float z2=z*z;
	float p2=p1*x-n1*y,n2=n1*x+p1*y;
	float P20=K2_DIFF_Z2_P20_A + K2_DIFF_Z2_P20_B*z2;
	float P21=K2_DIFF_Z2_P21*z;
	float P22=K2_DIFF_Z2_P22;
	c14.w=P22*n2;
	c58.x=P21*n1;
	c58.y=P20;
	c58.z=P21*p1;
	c58.w=P22*p2;
}
// getYlms_3(w)*PI
// void calcIBLCoefficients2(in half3 surfelNormal, out half c0, out half4 c14, out half4 c58)
// {
// 	float x = surfelNormal.x, y = surfelNormal.y, z = surfelNormal.z;
// 	float P00=0.8862268768654615;
// 	c0 = P00;
// 	float P10=1.5349900825729073*z;
// 	float P11=-1.5349900825729073;
// 	float p1=x,n1=y;
// 	c14.x=P11*n1;
// 	c14.y=P10;
// 	c14.z=P11*p1;
// 	float z2=z*z;
// 	float p2=p1*x-n1*y,n2=n1*x+p1*y;
// 	float P20=-0.9908318392196372 + 2.9724955176589116*z2;
// 	float P21=-3.2746465194388215*z;
// 	float P22=1.7161711523018273;
// 	c14.w=P22*n2;
// 	c58.x=P21*n1;
// 	c58.y=P20;
// 	c58.z=P21*p1;
// 	c58.w=P22*p2;
// }

#define K0_SKIN_RED_Z0 		3.0115 		//K0 * SKIN_RED_Z0
#define K1_SKIN_RED_Z1 		1.82022		//K1 * SKIN_RED_Z1;
#define K2_SKIN_RED_Z2 		0.500152	//K2 * SKIN_RED_Z2;
#define K0_SKIN_GREEN_Z0 	2.67854 	//K0 * SKIN_GREEN_Z0
#define K1_SKIN_GREEN_Z1 	1.76343 	//K1 * SKIN_GREEN_Z1;
#define K2_SKIN_GREEN_Z2 	0.616635	//K2 * SKIN_GREEN_Z2;
#define K0_SKIN_BLUE_Z0 	2.64012 	//K0 * SKIN_BLUE_Z0
#define K1_SKIN_BLUE_Z1 	1.75511		//K1 * SKIN_BLUE_Z1;
#define K2_SKIN_BLUE_Z2 	0.632751	//K2 * SKIN_BLUE_Z2;
 
void calcIBLCoefficientsForSkin(in half3 worldNormal, out half3 c0, out half4 c14_r, out half4 c14_g,
	out half4 c14_b, out half4 c58_r, out half4 c58_g, out half4 c58_b)
 {
	getYlms_3(worldNormal, c0.x, c14_r, c58_r);
	c0.y = c0.x * K0_SKIN_GREEN_Z0;
	c0.z = c0.x * K0_SKIN_BLUE_Z0;
	c0.x = c0.x * K0_SKIN_RED_Z0;

	c14_g.xyz = c14_r.xyz * K1_SKIN_GREEN_Z1;
	c14_b.xyz = c14_r.xyz * K1_SKIN_BLUE_Z1;
	c14_r.xyz = c14_r.xyz * K1_SKIN_RED_Z1;

	c14_g.w = c14_r.w * K2_SKIN_GREEN_Z2;
	c14_b.w = c14_r.w * K2_SKIN_BLUE_Z2;
	c14_r.w = c14_r.w * K2_SKIN_RED_Z2;

	c58_g = c58_r * K2_SKIN_GREEN_Z2;
	c58_b = c58_r * K2_SKIN_BLUE_Z2;
	c58_r = c58_r * K2_SKIN_RED_Z2;
 }
//cSHCoefficients[7]
//c[0] = {r0, g0, b0, InShadow}
//c[1] = {r1, r2, r3, r4}
//c[2] = {r5, r6, r7, r8}
//c[3] = {g1, g2, g3, g4}
//c[4] = {g5, g6, g7, g8}
//c[5] = {b1, b2, b3, b4}
//c[6] = {b5, b6, b7, b8}
half3 SunFogAndAmbientGI(in half3 worldNormal)
{
	half NoLSun = dot(SunDirection.xyz, worldNormal) * 0.5 + 0.5;
	half NoLSky = dot(half3(0,1,0), worldNormal) * 0.25 + 0.75;
	return NoLSun * max(half3(0.01,0.01,0.01), SunFogColor.rgb/max(1, SunFogColor.w)) + NoLSky * AmbientColor.rgb;
}
void DynamicGILight(in half3 worldNormal, out half3 GILighting, inout half InShadow)
{
	//InShadow = 0;
#if SD_ENABLE_DYNAMIC_GILIGHTING
	if (@_LOD_LEVEL <= LOD_Level_3)
	{
		half3 linearColor = half3(0,0,0);
	#if SD_SH_IBL
		#if MATERIAL_SHADINGMODEL_PREINTEGRATED_SKIN
			half3 c0;
			half4 c14_r, c14_g, c14_b;
			half4 c58_r, c58_g, c58_b;
			calcIBLCoefficientsForSkin(worldNormal, c0, c14_r, c14_g, c14_b, c58_r, c58_g, c58_b);
			linearColor.rgb = cSHCoefficients[0].xyz * c0.xyz;
			#if SD_SH_ORDER_3
				linearColor.r += dot(cSHCoefficients[1], c14_r) + dot(cSHCoefficients[2], c58_r);
				linearColor.g += dot(cSHCoefficients[3], c14_g) + dot(cSHCoefficients[4], c58_g);
				linearColor.b += dot(cSHCoefficients[5], c14_b) + dot(cSHCoefficients[6], c58_b);
			#else
				linearColor.r += dot(cSHCoefficients[1].xyz, c14_r.xyz);
				linearColor.g += dot(cSHCoefficients[3].xyz, c14_g.xyz);
				linearColor.b += dot(cSHCoefficients[5].xyz, c14_b.xyz);
			#endif
			linearColor.rgb = max(linearColor.rgb, 0);
		#else	
			half c0;
			half4 c14, c58;
			calcIBLCoefficients(worldNormal, c0, c14, c58);
			linearColor.rgb = cSHCoefficients[0].xyz * c0.rrr;
			#if SD_SH_ORDER_3
				linearColor.r += dot(cSHCoefficients[1], c14) + dot(cSHCoefficients[2], c58);
				linearColor.g += dot(cSHCoefficients[3], c14) + dot(cSHCoefficients[4], c58);
				linearColor.b += dot(cSHCoefficients[5], c14) + dot(cSHCoefficients[6], c58);
			#else
				linearColor.r += dot(cSHCoefficients[1].xyz, c14.xyz);
				linearColor.g += dot(cSHCoefficients[3].xyz, c14.xyz);
				linearColor.b += dot(cSHCoefficients[5].xyz, c14.xyz);
			#endif
			linearColor.rgb = max(linearColor.rgb, 0);
		#endif
	#else
		// /*if (@_ES20)
		// {
			half3 nSquared = worldNormal * worldNormal;
			bool3 isNegative = ( worldNormal < 0.0 );
			if (isNegative.x)
				linearColor += nSquared.x * cPointCloud[1].rgb;
			else
				linearColor += nSquared.x * cPointCloud[0].rgb;
			if (isNegative.y)
				linearColor += nSquared.y * cPointCloud[3].rgb;
			else
				linearColor += nSquared.y * cPointCloud[2].rgb;
			if (isNegative.z)
				linearColor += nSquared.z * cPointCloud[5].rgb;
			else
				linearColor += nSquared.z * cPointCloud[4].rgb;
		// }
		// else*/
		//{
			/*half3 nSquared = worldNormal * worldNormal;
			int3 isNegative = ( worldNormal < 0.0 );
			linearColor = nSquared.x * cPointCloud[isNegative.x].rgb +
				nSquared.y * cPointCloud[isNegative.y+2].rgb +
				nSquared.z * cPointCloud[isNegative.z+4].rgb;*/
			//GILighting from Atmosphere
			// linearColor = SunFogAndAmbientGI(worldNormal);
		//}
	#endif
		GILighting = linearColor;
	#if SD_SH_IBL
		InShadow = max(InShadow, cSHCoefficients[0].w);
	#else
		InShadow = max(InShadow, cPointCloud[0].w);
	#endif
	}
	// else
#endif
	// {
	// 	GILighting = half3(0,0,0);
	// }
}

void DecodeLightMap(in half4 lightMapRaw, in half3 lightMapScale, out half3 GILighting, out half InShadow) 
{
	InShadow = 0.0;
	half3 lightMapCol = lightMapRaw.rgb * lightMapRaw.rgb * lightMapScale;
	GILighting = lightMapCol;
	InShadow = lightMapRaw.a;
}
//IBL
half3 IBL_Specular(in GBufferData GBuffer, in half3 R, in half NoV)
{
	half3 OUT = half3(0, 0, 0);
	if (@_LOD_LEVEL <= LOD_Level_2)
	{
		half3 sampleEnvSpecular = half3(0,0,0);
		half MIP_ROUGHNESS = 0.17;
		half level = GBuffer.Roughness / MIP_ROUGHNESS;
		half fSign = R.z > 0;
		half fSign2 = fSign*2 - 1;
		R.xy /= R.z*fSign2 + 1;
		R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
		{
			half4 srcColor;
			srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy, level*RenderToMip);
			sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE);
		}
		if(GBuffer.ShadingModelID == SHADINGMODELID_HAIR)
			OUT = EnvBRDFApprox(GBuffer.SpecularColor, 0.8, NoV) * sampleEnvSpecular;
		else
			OUT = EnvBRDFApprox(GBuffer.SpecularColor, GBuffer.Roughness, NoV) * sampleEnvSpecular;

		if (GBuffer.ShadingModelID == SHADINGMODELID_CLEAR_COAT)
		{
			half ClearCoat = GBuffer.CustomData.x;
			half ClearCoatRoughness = GBuffer.CustomData.y;
			float VoH = NoV;
			float F = F_SchlickGray(0.04, VoH) * ClearCoat;
			OUT *= 1-F;

			level = ClearCoatRoughness / MIP_ROUGHNESS;
			half4 srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy, level*RenderToMip);
			sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE);
			OUT += F * sampleEnvSpecular;
		}
	}
	return OUT * ReflectionScale * GBuffer.AO;
}

//////////////////////////////////////////Make Light Data/////////////////////////////////////////
LightingData GetSunLightData(in half3 V, in GBufferData GBuffer, in int PBRQualityLevel)
{
	LightingData LitData;
	LitData.FalloffColor = SunColor.rgb;
	if(!@_HDR)
	{
	LitData.FalloffColor *= 2;	
	}

	
	LitData.Atten = 1;
	LitData.L = SunDirection.xyz;
	LitData.V = V;
	LitData.N = GBuffer.WorldNormal;
	LitData._NoL = dot(LitData.N, LitData.L);
	LitData.NoL = saturate(LitData._NoL);
	LitData.InShadow = GBuffer.InShadow;
	LitData.RadiusTan = 0;
	LitData.PBRQualityLevel = PBRQualityLevel;
	LitData.TranslucencyLighting = false;
	LitData.IsSphericalLight = true;
	LitData.IsRectLight = false;
	LitData.IsDiskLight = false;
	return LitData;
}

LightingData GetPointLightData(in float3 P, in GBufferData GBuffer, in float3 V, 
	in half InShadow, int PBRQualityLevel, in float4 LightParams[4])
{
	LightingData LitData;
	float3 L = LightParams[0].xyz - P;
	float D2 = dot(L, L);
	// float D = sqrt(D2);
	float D = D2;
	L /= D2;
	LitData.FalloffColor = LightParams[1].rgb * LightRangeMask(D, LightParams[0].w);
	LitData.Atten = CalcDistanceAttenuation(D2, LightParams);
	LitData.L = L;
	LitData.V = V;
	LitData.N = GBuffer.WorldNormal;
	LitData._NoL = dot(GBuffer.WorldNormal, L);
	LitData.NoL = saturate(LitData._NoL);
	LitData.InShadow = InShadow;
	LitData.RadiusTan = LightParams[3].y / D;
	LitData.PBRQualityLevel = PBRQualityLevel;
	LitData.TranslucencyLighting = false;
	LitData.IsSphericalLight = false;
	LitData.IsRectLight = false;
	LitData.IsDiskLight = false;
	return LitData;
}

LightingData GetSphericalLightData(in float3 P, in GBufferData GBuffer, in float3 V, 
	in half InShadow, int PBRQualityLevel, in float4 LightParams[4])
{
	LightingData LitData;
	float3 L = LightParams[0].xyz - P;
	float D2 = dot(L, L);
	// float D = sqrt(D2);
	float D = D2 * D2;
	L /= D;
	LitData.FalloffColor = LightParams[1].rgb * LightRangeMask(D2, LightParams[0].w);
	LitData.Atten = CalcDistanceAttenuation(D2, LightParams);
	LitData.L = L;
	LitData.V = V;
	LitData.N = GBuffer.WorldNormal;
	LitData._NoL = dot(GBuffer.WorldNormal, L);
	LitData.NoL = saturate(LitData._NoL);
	LitData.InShadow = InShadow;
	LitData.RadiusTan = LightParams[3].y / D;
	LitData.PBRQualityLevel = PBRQualityLevel;
	LitData.TranslucencyLighting = false;
	LitData.IsSphericalLight = true;
	LitData.IsRectLight = false;
	LitData.IsDiskLight = false;
	LitData.Atten = 2.0 * (1.0 - sqrt(max(0, D2 - Pow2(LightParams[3].y))) / D);
	return LitData;
}

LightingData GetSpotLightData(in float3 P, in GBufferData GBuffer, in float3 V, 
	in half InShadow, int PBRQualityLevel, in float4 LightParams[4])
{
	LightingData LitData;
	float3 L = LightParams[0].xyz - P;
	float D2 = dot(L, L);
	float D = sqrt(D2);
	L /= D;
	half DoL = dot( LightParams[2].xyz, -L );
	LitData.FalloffColor = LightParams[1].rgb * LightRangeMask(D2, LightParams[0].w);
	LitData.Atten = CalcDistanceAttenuation(D2, LightParams) * Square(saturate(DoL*LightParams[3].y + LightParams[3].z));
	LitData.L = L;
	LitData.V = V;
	LitData.N = GBuffer.WorldNormal;
	LitData._NoL = dot(GBuffer.WorldNormal, L);
	LitData.NoL = saturate(LitData._NoL);
	LitData.InShadow = InShadow;
	LitData.RadiusTan = 0;
	LitData.PBRQualityLevel = PBRQualityLevel;
	LitData.TranslucencyLighting = false;
	LitData.IsSphericalLight = false;
	LitData.IsRectLight = false;
	LitData.IsDiskLight = false;
	return LitData;
}

// Apply IES light profile texture
float2 ComputeLightProfileMultiplier(float3 ToLight, float3 LightDirection, float3 LightTangent, float3 LightBitangent)
{
	//float4x4 LightTransform = float4x4( float4(dd.xyz, 0), float4(LightBitangent.xyz, 0), float4(LightTangent.xyz, 0), float4(0, 0, 0, 1) );
	//float4x4 InvLightTransform = transpose(LightTransform);
	float4x4 InvLightTransform = float4x4(
		LightDirection.x, LightBitangent.x, LightTangent.x, 0,
		LightDirection.y, LightBitangent.y, LightTangent.y, 0,
		LightDirection.z, LightBitangent.z, LightTangent.z, 0,
		0, 0, 0, 1
	);
	float3 LocalToLight = mul(float4(ToLight.xyz, 0), InvLightTransform).xyz;

	// -1..1
	float DotProd = dot(ToLight, LightDirection);
	// -PI/2..PI/2
	float Angle = asin(DotProd);
	// 0..1
	float NormAngle = Angle / PI + 0.5f;

	float TangentAngle = atan2( -LocalToLight.z, -LocalToLight.y );
	float NormTangentAngle = TangentAngle / (PI * 2.f) + 0.5f;

	return float2(NormAngle, NormTangentAngle);
}

//LightPosRange, LightColorAtten
// struct alignas(16) CB_PointLight
// {
// 	Vector3 WorldPos;
// 	float InvSquareRange;//1/(Range*Range)

// 	Vector3 Color;
// 	float IsInverseSquaredFalloff;//true: 1; false 0

// 	Vector3 WorldDir;
// 	float IsPointLight;//true: 1; false: -1

//	float VolumetricScatteringScale;
// float RcpCosHalfThetaSubPhi;// 1 / [cos(0.5f*theta) - cos(0.5f*phi)]
// float CosHalfPHi;			// - cos(0.5f*phi) / [cos(0.5f*theta) - cos(0.5f*phi)]
// float CosHalfPhi2;
// };
#define PointLighting(P, N, LightData, DiffLit)\
{\
	float3 L = LightData[0].xyz - P;\
	float D2 = dot(L, L);\
	float D = sqrt(D2);\
	L /= D;\
	float NoL = saturate( dot(N, L) );\
	float Atten = CalcDistanceAttenuation(D2, LightData) * LightRangeMask(D2, LightData[0].w);\
	DiffLit += LightData[1].rgb * (NoL * Atten);\
}

half3 SpotLighting(in half3 P, in half3 N, in float4 LightData[4])
{
	half3 diffLit = half3(0, 0, 0);

	// Spot lighting.
	float3 L = LightData[0].xyz - P;
	float D2 = dot(L, L);
	float D = sqrt(D2);
	L /= D;
	float NoL = saturate( dot(N, L) );
	float DoL = dot( LightData[2].xyz, -L );
	
	float Atten = CalcDistanceAttenuation(D2, LightData) * LightRangeMask(D2, LightData[0].w);
	float Spot = Square(saturate(DoL*LightData[3].y + LightData[3].z));
#ifdef VOLUMETRIC_SHADING
	NoL = 1;
#endif	
	diffLit = LightData[1].rgb * (NoL * Atten * Spot);

	return diffLit;
}

float3 LightingVS(in float3 P, in float3 N)
{
	float3 diffLit = float3(0, 0, 0);

	// //lit_0
	// if( @_LitCount >= 1 && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) < 0 )
	// {
	// 	if(Lights0[2].w > 0)
	// 	{
	// 		PointLighting( P, N, Lights0, diffLit );
	// 	}
	// 	else
	// 	{
	// 		diffLit = SpotLighting( P, N, Lights0);
	// 	}
		
	// }
	// //lit_1
	// if( @_LitCount >= 2 && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) < 1 )
	// {
	// 	//PointLighting( P, N, Lights1, diffLit );
	// 	if(Lights0[2].w > 0)
	// 	{
	// 		PointLighting( P, N, Lights1, diffLit );
	// 	}
	// 	else
	// 	{
	// 		diffLit = SpotLighting( P, N, Lights1);
	// 	}
	// }
	// //lit_2
	// if( @_LitCount >= 3 )
	// {
	// 	//PointLighting( P, N, Lights2, diffLit );
	// 	if(Lights0[2].w > 0)
	// 	{
	// 		PointLighting( P, N, Lights2, diffLit );
	// 	}
	// 	else
	// 	{
	// 		diffLit = SpotLighting( P, N, Lights2);
	// 	}
	// }

	return diffLit;
}

#define FogLighting(T, color) lerp(FogColor.rgb, color, T)

#define SkyFogColorAndLighting(T, color, V, L)\
{\
	float VoL = saturate(dot(-V, L)); \
	float VoL2 = VoL * VoL;\
	float3 fogColor = VoL2 * VoL2 * SunFogColor.rgb + FogColor.rgb;\
	color = lerp(color, fogColor, T);\
}

#define SkyFogLightingAndScatter(color, V, L)\
{\
	float VoL = saturate(dot(-V, L)); \
	float3 single_r = FogColor.rgb * Rayleigh(1 - VoL);\
	float3 single_m = SunFogColor.rgb * Mie(VoL, 0.6);\
	color = single_r + single_m;\
}

#define FogColorAndLighting(T, color, V, L)\
{\
	float VoL = saturate(dot(-V, L)); \
	float VoL2 = VoL * VoL;\
	float S = (1 - pow(T, 0.5)) * VoL2 * VoL2;\
	color = lerp(SunFogColor.rgb * S + FogColor.rgb, color, T);\
}

#define FogLightingAndScatter(T, color, V, L)\
{\
	float VoL = saturate(dot(-V, L)); \
	float S = 1 - pow(T, 0.5);\
	float3 single_r = FogColor.rgb * Rayleigh(1 - VoL);\
	float3 single_m = SunFogColor.rgb * Mie(VoL, 0.6);\
	color = single_r + single_m * S;\
}

#define MipFogLighting(MipfogColor, T, color) lerp(MipfogColor, color, T)
#define InvertFogLighting(T, color) ((color - FogColor.rgb*(1-T)) / T)

#define SampleShadowMap(destTC, fDepth)\
{\
	fDepth = tShadowMap.SampleLevel(sShadowMapSampler, destTC, 0).x;\
}

#define irregular_pcf_high(pInLitTC, litZ, vInvShadowMapWH, shadowTest)\
{\
	shadowTest = 0;\
	static const float4 irreg_kernel_2d[4] = {\
		float4(-0.613392,0.617481,0.553514,-0.832698),\
		float4(0.806146,0.578845,-0.724784,-0.673879),\
		float4(0.037202,-0.049409,0.099155,0.986755),\
		float4(0.975463,-0.180090,-0.989929,-0.006012)\
	};\
	float2 filterRange = vInvShadowMapWH*2;\
	for(int i=0; i<4; i+=2)\
	{\
		float4 sampleDepth;\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[i+0].xy, sampleDepth.x);\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[i+0].zw, sampleDepth.y);\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[i+1].xy, sampleDepth.z);\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[i+1].zw, sampleDepth.w);\
		float4 InShadow = ( litZ.xxxx > sampleDepth );\
		shadowTest += dot( InShadow, float4(1,1,1,1) );\
	}\
	shadowTest /= 8;\
}

#define irregular_pcf_low(pInLitTC, litZ, vInvShadowMapWH, shadowTest)\
{\
	shadowTest = 0;\
	static const float4 irreg_kernel_2d[2] = {\
		float4(-0.613392,0.617481,0.553514,-0.832698),\
		float4(0.804315,0.560289,-0.720084,-0.677786)\
	};\
	float2 filterRange = vInvShadowMapWH;\
	{\
		float4 sampleDepth;\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[0].xy, sampleDepth.x);\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[0].zw, sampleDepth.y);\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[1].xy, sampleDepth.z);\
		SampleShadowMap(pInLitTC+filterRange*irreg_kernel_2d[1].zw, sampleDepth.w);\
		float4 InShadow = ( litZ.xxxx > sampleDepth );\
		shadowTest += dot( InShadow, float4(1,1,1,1) );\
	}\
	shadowTest /= 4;\
}

void pcf_4x4(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, out float inShadow)
{
	float2 TexelPos = pInLitTC / vInvShadowMapWH - 0.5;	// bias to be consistent with texture filtering hardware
	float2 Fraction = frac(TexelPos);
	float2 TexelCenter = floor(TexelPos) + 0.5;	// bias to get reliable texel center content
	float2 Sample00TexelCenter = TexelCenter - float2(1, 1);
	
	float4 Values0, Values1, Values2, Values3;// Valuesi in row i from left to right: x,y,z,w
	SampleShadowMap( (Sample00TexelCenter+float2(0,0))*vInvShadowMapWH, Values0.x );
	SampleShadowMap( (Sample00TexelCenter+float2(1,0))*vInvShadowMapWH, Values0.y );
	SampleShadowMap( (Sample00TexelCenter+float2(2,0))*vInvShadowMapWH, Values0.z );
	SampleShadowMap( (Sample00TexelCenter+float2(3,0))*vInvShadowMapWH, Values0.w );
	Values0 = step(Values0, litZ.xxxx);
	SampleShadowMap( (Sample00TexelCenter+float2(0,1))*vInvShadowMapWH, Values1.x );
	SampleShadowMap( (Sample00TexelCenter+float2(1,1))*vInvShadowMapWH, Values1.y );
	SampleShadowMap( (Sample00TexelCenter+float2(2,1))*vInvShadowMapWH, Values1.z );
	SampleShadowMap( (Sample00TexelCenter+float2(3,1))*vInvShadowMapWH, Values1.w );
	Values1 = step(Values1, litZ.xxxx);
	SampleShadowMap( (Sample00TexelCenter+float2(0,2))*vInvShadowMapWH, Values2.x );
	SampleShadowMap( (Sample00TexelCenter+float2(1,2))*vInvShadowMapWH, Values2.y );
	SampleShadowMap( (Sample00TexelCenter+float2(2,2))*vInvShadowMapWH, Values2.z );
	SampleShadowMap( (Sample00TexelCenter+float2(3,2))*vInvShadowMapWH, Values2.w );
	Values2 = step(Values2, litZ.xxxx);
	SampleShadowMap( (Sample00TexelCenter+float2(0,3))*vInvShadowMapWH, Values3.x );
	SampleShadowMap( (Sample00TexelCenter+float2(1,3))*vInvShadowMapWH, Values3.y );
	SampleShadowMap( (Sample00TexelCenter+float2(2,3))*vInvShadowMapWH, Values3.z );
	SampleShadowMap( (Sample00TexelCenter+float2(3,3))*vInvShadowMapWH, Values3.w );
	Values3 = step(Values3, litZ.xxxx);

	float2 VerticalLerp00 = lerp(float2(Values0.x, Values1.x), float2(Values0.y, Values1.y), Fraction.xx);
	float PCFResult00 = lerp(VerticalLerp00.x, VerticalLerp00.y, Fraction.y);
	float2 VerticalLerp10 = lerp(float2(Values0.y, Values1.y), float2(Values0.z, Values1.z), Fraction.xx);
	float PCFResult10 = lerp(VerticalLerp10.x, VerticalLerp10.y, Fraction.y);
	float2 VerticalLerp20 = lerp(float2(Values0.z, Values1.z), float2(Values0.w, Values1.w), Fraction.xx);
	float PCFResult20 = lerp(VerticalLerp20.x, VerticalLerp20.y, Fraction.y);

	float2 VerticalLerp01 = lerp(float2(Values1.x, Values2.x), float2(Values1.y, Values2.y), Fraction.xx);
	float PCFResult01 = lerp(VerticalLerp01.x, VerticalLerp01.y, Fraction.y);
	float2 VerticalLerp11 = lerp(float2(Values1.y, Values2.y), float2(Values1.z, Values2.z), Fraction.xx);
	float PCFResult11 = lerp(VerticalLerp11.x, VerticalLerp11.y, Fraction.y);
	float2 VerticalLerp21 = lerp(float2(Values1.z, Values2.z), float2(Values1.w, Values2.w), Fraction.xx);
	float PCFResult21 = lerp(VerticalLerp21.x, VerticalLerp21.y, Fraction.y);

	float2 VerticalLerp02 = lerp(float2(Values2.x, Values3.x), float2(Values2.y, Values3.y), Fraction.xx);
	float PCFResult02 = lerp(VerticalLerp02.x, VerticalLerp02.y, Fraction.y);
	float2 VerticalLerp12 = lerp(float2(Values2.y, Values3.y), float2(Values2.z, Values3.z), Fraction.xx);
	float PCFResult12 = lerp(VerticalLerp12.x, VerticalLerp12.y, Fraction.y);
	float2 VerticalLerp22 = lerp(float2(Values2.z, Values3.z), float2(Values2.w, Values3.w), Fraction.xx);
	float PCFResult22 = lerp(VerticalLerp22.x, VerticalLerp22.y, Fraction.y);

	inShadow = (PCFResult00 + PCFResult10 + PCFResult20 + PCFResult01 + PCFResult11 + PCFResult21 + PCFResult02 + PCFResult12 + PCFResult22) * .11111;
}

void hardware_pcf_4x4(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, out float inShadow)
{
	float2 TexelPos = pInLitTC / vInvShadowMapWH - 0.5;	// bias to be consistent with texture filtering hardware
	float2 Fraction = frac(TexelPos);
	float2 TexelCenter = floor(TexelPos) + 0.5;	// bias to get reliable texel center content

	float2 w0 = 2 - Fraction;
	float2 w1 = Fraction + 1;
	float2 u0 = 1.0 / w0;
	float2 u1 = Fraction / w1;

	float2 Pos0 = TexelCenter + float2(-1, -1) + float2(u0.x, u0.y);
	float s0 = tShadowMap.SampleCmpLevelZero(sShadowMapSampler, Pos0 * vInvShadowMapWH, litZ);
	float2 Pos1 = TexelCenter + float2( 1, -1) + float2(u1.x, u0.y);
	float s1 = tShadowMap.SampleCmpLevelZero(sShadowMapSampler, Pos1 * vInvShadowMapWH, litZ);
	float2 Pos2 = TexelCenter + float2(-1,  1) + float2(u0.x, u1.y);
	float s2 = tShadowMap.SampleCmpLevelZero(sShadowMapSampler, Pos2 * vInvShadowMapWH, litZ);
	float2 Pos3 = TexelCenter + float2( 1,  1) + float2(u1.x, u1.y);
	float s3 = tShadowMap.SampleCmpLevelZero(sShadowMapSampler, Pos3 * vInvShadowMapWH, litZ);

	inShadow = (s0 * w0.x * w0.y + s1 * w1.x * w0.y + s2 * w0.x * w1.y + s3 * w1.x * w1.y) * 0.1111111;
}

float hardware_pcf_array_4x4(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, int slice)
{
	float2 TexelPos = pInLitTC / vInvShadowMapWH - 0.5;	// bias to be consistent with texture filtering hardware
	float2 Fraction = frac(TexelPos);
	float2 TexelCenter = floor(TexelPos) + 0.5;	// bias to get reliable texel center content

	float2 w0 = 2 - Fraction;
	float2 w1 = Fraction + 1;
	float2 u0 = 1.0 / w0;
	float2 u1 = Fraction / w1;

	float2 Pos0 = TexelCenter + float2(-1, -1) + float2(u0.x, u0.y);
	float s0 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos0 * vInvShadowMapWH, slice), litZ);
	float2 Pos1 = TexelCenter + float2( 1, -1) + float2(u1.x, u0.y);
	float s1 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos1 * vInvShadowMapWH, slice), litZ);
	float2 Pos2 = TexelCenter + float2(-1,  1) + float2(u0.x, u1.y);
	float s2 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos2 * vInvShadowMapWH, slice), litZ);
	float2 Pos3 = TexelCenter + float2( 1,  1) + float2(u1.x, u1.y);
	float s3 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos3 * vInvShadowMapWH, slice), litZ);

	return (s0 * w0.x * w0.y + s1 * w1.x * w0.y + s2 * w0.x * w1.y + s3 * w1.x * w1.y) * 0.1111111;
}

float hardware_pcf_array_8x8(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, int slice)
{
	float2 TexelPos = pInLitTC / vInvShadowMapWH - 0.5;	// bias to be consistent with texture filtering hardware
	float2 Fraction = frac(TexelPos);
	float2 TexelCenter = floor(TexelPos) + 0.5;	// bias to get reliable texel center content

	float2 w0 = 2 - Fraction;
	float2 w1 = Fraction + 1;
	float2 u0 = 1.0 / w0;
	float2 u1 = Fraction / w1;

	float2 Pos0 = TexelCenter + float2(-1, -1) + float2(u0.x, u0.y);
	float s0 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos0 * vInvShadowMapWH, slice), litZ);
	float2 Pos1 = TexelCenter + float2( 1, -1) + float2(u1.x, u0.y);
	float s1 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos1 * vInvShadowMapWH, slice), litZ);
	float2 Pos2 = TexelCenter + float2(-1,  1) + float2(u0.x, u1.y);
	float s2 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos2 * vInvShadowMapWH, slice), litZ);
	float2 Pos3 = TexelCenter + float2( 1,  1) + float2(u1.x, u1.y);
	float s3 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos3 * vInvShadowMapWH, slice), litZ);

	float2 Pos4 = TexelCenter + float2(-2, -2) + float2(u0.x, u0.y);
	float s4 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos4 * vInvShadowMapWH, slice), litZ);
	float2 Pos5 = TexelCenter + float2( 2, -2) + float2(u1.x, u0.y);
	float s5 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos5 * vInvShadowMapWH, slice), litZ);
	float2 Pos6 = TexelCenter + float2(-2,  2) + float2(u0.x, u1.y);
	float s6 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos6 * vInvShadowMapWH, slice), litZ);
	float2 Pos7 = TexelCenter + float2( 2,  2) + float2(u1.x, u1.y);
	float s7 = tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(Pos7 * vInvShadowMapWH, slice), litZ); 

	return (s0 * w0.x * w0.y + s1 * w1.x * w0.y + s2 * w0.x * w1.y + s3 * w1.x * w1.y + s4 * w0.x * w0.y + s5 * w1.x * w0.y + s6 * w0.x * w1.y + s7 * w1.x * w1.y) * 0.05555555;
	//return (s0 + s1 + s2 + s3 + s4 + s5 + s6 + s7) / 8;
	//return (s0 * w0.x * w0.y + s1 * w1.x * w0.y + s2 * w0.x * w1.y + s3 * w1.x * w1.y) * 0.1111111;
}

void pcf_2x2(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, out float inShadow)
{
	inShadow = 0;
	float2 fc = frac(pInLitTC/vInvShadowMapWH - 0.5);
	float2 tc0 = pInLitTC - (fc-0.5)*vInvShadowMapWH;
	float2 tc1 = tc0 + float2(vInvShadowMapWH.x, 0);
	float2 tc2 = tc0 + float2(0, vInvShadowMapWH.y);
	float2 tc3 = tc1 + float2(0, vInvShadowMapWH.y);
	float4 sampleDepth;
	SampleShadowMap(tc0, sampleDepth.x);
	SampleShadowMap(tc1, sampleDepth.y);
	SampleShadowMap(tc2, sampleDepth.z);
	SampleShadowMap(tc3, sampleDepth.w);
	float4 InShadow = ( litZ.xxxx > sampleDepth );
	float inShadow01 = lerp(InShadow.x, InShadow.y, fc.x);
	float inShadow23 = lerp(InShadow.z, InShadow.w, fc.x);
	inShadow = lerp(inShadow01, inShadow23, fc.y);
}

void hardware_pcf_2x2(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, out float inShadow)
{
	inShadow = tShadowMap.SampleCmpLevelZero(sShadowMapSampler, pInLitTC, litZ);
}

float hardware_pcf_array_2x2(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, int slice)
{
	return tShadowMapArray.SampleCmpLevelZero(sShadowMapArraySampler, float3(pInLitTC, slice), litZ);
}

void GetDynamicShadow(float3 LitUVZ, float2 vInvShadowMapWH, out half InDynamicShadow, out half ShadowRate)
{
	float inShadow;
	float2 pInLitTC = LitUVZ.xy;
	float litZ = LitUVZ.z;
	half2 inRange = abs(pInLitTC - 0.5);
	ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
	if(@_LOD_LEVEL <= SD_SHADOW_HIGHQUALITY_LEVEL)
	{
		hardware_pcf_4x4(pInLitTC, litZ, vInvShadowMapWH, inShadow);
	}
	else
	{
		hardware_pcf_2x2(pInLitTC, litZ, vInvShadowMapWH, inShadow);
	}
	InDynamicShadow = inShadow;
}

float GetInShadowFromCSM(in bool IsHighQualityShadow, in float3 worldPos, in float3 worldNormal, 
	in float shadowBiasScale, out float4 P_LS)
{
	float4x4 cShadowViewProjTexs[3];
	cShadowViewProjTexs[0] = ShadowViewProjTexs0;
	cShadowViewProjTexs[1] = ShadowViewProjTexs1;
	cShadowViewProjTexs[2] = ShadowViewProjTexs2;
	float cInvShadowMapSize = CSMShadowSlopBiases[3];

	float sliceId = 2;
	P_LS = 0;

	[unroll]
	for(int i = 2; i >= 0; --i)
	{
		float flag = step(-0.1, CSMShadowBiases[i]);

		//light space uvz
		float4 temp_P_LS = mul(float4(worldPos+worldNormal*CSMShadowSlopBiases[i],1), cShadowViewProjTexs[i]);
		temp_P_LS.z -= CSMShadowBiases[i] * shadowBiasScale;//shadowBiasScale=clamp(sceneZ * CameraInfo.y * 0.1, 1, 3)
		temp_P_LS.xyz /= temp_P_LS.w;

		flag *= float(all(temp_P_LS.xyz > 0) && all(temp_P_LS.xyz < 1)); 
		P_LS += (temp_P_LS - P_LS) * flag;
		sliceId += (i-sliceId) * flag;
	}

	//Shadow
	float InShadow = 0;
	float tempShadow = 0;
	if (IsHighQualityShadow)
	{
		InShadow = hardware_pcf_array_4x4(P_LS.xy, P_LS.z, cInvShadowMapSize.xx, 2*sliceId);
		tempShadow = (sliceId==2)?0:hardware_pcf_array_4x4(P_LS.xy, P_LS.z, cInvShadowMapSize.xx, (2*sliceId) + 1);
	}
	else
	{
		InShadow = hardware_pcf_array_2x2(P_LS.xy, P_LS.z, cInvShadowMapSize.xx, (2*sliceId));
		tempShadow = (sliceId==2)?0:hardware_pcf_array_2x2(P_LS.xy, P_LS.z, cInvShadowMapSize.xx, (2*sliceId) + 1);
	}
	InShadow = max(tempShadow, InShadow);
	// half2 inRange = abs(P_LS.xy - 0.5);
	// P_LS.w = min(0.5-inRange.x, 0.5-inRange.y);
	// if (cShadowBiass[sliceId+1]>=0)
	// 	P_LS.w = 1;
	return InShadow;
}

float GetForwardInShadowFromCSM(in bool IsHighQualityShadow, in float3 litUVZ0, in float3 litUVZ1, 
	in float3 litUVZ2, out float ShadowRate)
{
	float3 P_LSS[3];
	P_LSS[0] = litUVZ0;
	P_LSS[1] = litUVZ1;
	P_LSS[2] = litUVZ2;
	float ShadowRateScale = 0;

	float sliceId = 2;
	float3 P_LS = 0;

	[unroll]
	for(int i = 2; i >= 0; --i)
	{
		float flag = step(-0.1, CSMShadowBiases[i]);
		float3 temp_P_LS = P_LSS[i];
		flag *= float(all(temp_P_LS.xyz > 0) && all(temp_P_LS.xyz < 1)); 
		P_LS += (temp_P_LS - P_LS) * flag;
		sliceId += (i-sliceId) * flag;
	}

	//Shadow
	float InShadow = 0;
	float tempShadow = 0;
	if (IsHighQualityShadow)
	{
		InShadow = hardware_pcf_array_4x4(P_LS.xy, P_LS.z, CSMShadowSlopBiases[3].xx, 2*sliceId);
		tempShadow = (sliceId==2)?0:hardware_pcf_array_4x4(P_LS.xy, P_LS.z, CSMShadowSlopBiases[3].xx, (2*sliceId) + 1);
	}
	else
	{
		InShadow = hardware_pcf_array_2x2(P_LS.xy, P_LS.z, CSMShadowSlopBiases[3].xx, (2*sliceId));
		tempShadow = (sliceId==2)?0:hardware_pcf_array_2x2(P_LS.xy, P_LS.z, CSMShadowSlopBiases[3].xx, (2*sliceId) + 1);
	}
	InShadow = max(tempShadow, InShadow);
	half2 inRange = abs(P_LS.xy - 0.5);
	ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
	if (CSMShadowBiases[int(sliceId+1)]>=0)
		ShadowRate = 1;
	return InShadow;
}

#endif //__FILE_LIGHTING_FXH__
