#ifndef __FILE_MESSIAHTAA_FXH__
#define __FILE_MESSIAHTAA_FXH__

#define bicubicSampleCatmullRom3x3(uv, texSize, tex, samp, result) \
{\
	float2 samplePos = uv * texSize;\
	float2 tc = floor(samplePos - 0.5) + 0.5;\
	float2 f = samplePos - tc;\
	float2 f2 = f*f;\
	float2 f3 = f2 * f;\
	float2 w0 = f2 - 0.5 * (f3 + f);\
	float2 w1 = 1.5 * f3 - 2.5 * f2 + 1.0;\
	float2 w3 = 0.5 * (f3 - f2);\
	float2 w2 = 1.0 - w0 - w1 - w3;\
	float2 w12 = w1 + w2;\
\
	float2 tc0 = (tc - 1) / texSize;\
	float2 tc12 = (tc + w2/w12) / texSize;\
	float2 tc3 = (tc + 2) / texSize;\
\
	result = \
		SampleColorMap(tex, samp, float2(tc0.x, tc0.y))   * (w0.x  * w0.y) +\
		SampleColorMap(tex, samp, float2(tc0.x, tc12.y))  * (w0.x  * w12.y) +\
		SampleColorMap(tex, samp, float2(tc0.x, tc3.y))   * (w0.x  * w3.y) +\
		SampleColorMap(tex, samp, float2(tc12.x, tc0.y))  * (w12.x * w0.y) +\
		SampleColorMap(tex, samp, float2(tc12.x, tc12.y)) * (w12.x * w12.y) +\
		SampleColorMap(tex, samp, float2(tc12.x, tc3.y))  * (w12.x * w3.y) +\
		SampleColorMap(tex, samp, float2(tc3.x, tc0.y))   * (w3.x  * w0.y) +\
		SampleColorMap(tex, samp, float2(tc3.x, tc12.y))  * (w3.x  * w12.y) +\
		SampleColorMap(tex, samp, float2(tc3.x, tc3.y))   * (w3.x  * w3.y);\
}

#define SampleTextureBilinearlyAndUnpack(T, S, C) (T.Sample(S, C.xy)*2 - 1)
float4 bicubicSampleCatmullRom2x2(float2 uv, float2 texSize)
{
	// Based on the standard Catmull-Rom spline: w1*C1+w2*C2+w3*C3+w4*C4, where
	// w1 = ((-0.5*f + 1.0)*f - 0.5)*f, w2 = (1.5*f - 2.5)*f*f + 1.0,
	// w3 = ((-1.5*f + 2.0)*f + 0.5)*f and w4 = (0.5*f - 0.5)*f*f with f as the
	// normalized interpolation position between C2 (at f=0) and C3 (at f=1).

	// half_f is a sort of sub-pixelquad fraction, -1 <= half_f < 1.
	float2 half_f     = 2.0 * frac(0.5 * uv * texSize - 0.25) - 1.0;
 
    // f is the regular sub-pixel fraction, 0 <= f < 1. This is equivalent to
    // fract(uv * texSize - 0.5), but based on half_f to prevent rounding issues.
    float2 f          = frac(half_f);
 
    float2 s1         = ( 0.5 * f - 0.5) * f;            // = w1 / (1 - f)
    float2 s12        = (-2.0 * f + 1.5) * f + 1.0;      // = (w2 - w1) / (1 - f)
    float2 s34        = ( 2.0 * f - 2.5) * f - 0.5;      // = (w4 - w3) / f
 
    // positions is equivalent to: (floor(uv * texSize - 0.5).xyxy + 0.5 +
    // vec4(-1.0 + w2 / (w2 - w1), 1.0 + w4 / (w4 - w3))) / texSize.xyxy.
    float4 positions  = float4((-f * s12 + s1      ) / (texSize * s12) + uv,
                           (-f * s34 + s1 + s34) / (texSize * s34) + uv);
 
    // Determine if the output needs to be sign-flipped. Equivalent to .x*.y of
    // (1.0 - 2.0 * floor(t - 2.0 * floor(0.5 * t))), where t is uv * texSize - 0.5.
    float sign_flip = half_f.x * half_f.y > 0.0 ? 1.0 : -1.0;
 
    float4 w          = float4(-f * s12 + s12, s34 * f); // = (w2 - w1, w4 - w3)
    float4 weights    = float4(w.xz * (w.y * sign_flip), w.xz * (w.w * sign_flip));
 
	return 	SampleTextureBilinearlyAndUnpack(tLastScreenMap, sLastScreenSample, positions.xy) * weights.x +
			SampleTextureBilinearlyAndUnpack(tLastScreenMap, sLastScreenSample, positions.zy) * weights.y +
			SampleTextureBilinearlyAndUnpack(tLastScreenMap, sLastScreenSample, positions.xw) * weights.z +
			SampleTextureBilinearlyAndUnpack(tLastScreenMap, sLastScreenSample, positions.zw) * weights.w;
}


float4 TAA_Messiah(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;

	half2 jitter2 = half2( cJitter.x / (cRTSize.x)  ,  cJitter.y / (cRTSize.y)) ;
	
	half2 lastScreenUV = IN.JitteredUV;

	float2 stc = spos / cRTSize.zw;
	float2 tcPosition = floor(stc * cRTSize.xy) + 0.5;
	float2 tc = lastScreenUV;
	float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, tc));
	float3 worldPos = CalcHomogeneousPos(depth,( spos / cRTSize.zw)  );
	worldPos += CameraPos.xyz;
	float4 c4;
	// if (TAALevel == 0 || @_TAAFullSize)
	//  	c4 = SampleColorMap(tScreenMap, sScreenSampler, stc + jitter2);
	// else
		bicubicSampleCatmullRom3x3(spos + jitter2, cRTSize.zw, tScreenMapLinear, sScreenLinearSampler, c4);
	float4 OUT = float4(c4.rgb, 1);
//	return float4(0,0,1,1);
//	return OUT;
	//motion vector
	float4 lastPos = MulMat4x4(lastViewProjTex, float4(worldPos,1.0));
	half2 lastUV = (lastPos.xy / lastPos.w);
	#if TSAA_DEBUG_DISABLE_REPROJECT
		lastUV = lastScreenUV;
	#endif
	half2 motion = half2(0,0);
	#if TSAA_USING_VELOCITY_BUFFER
		motion = SampleColorMap(tVBuffer, sVBufferSampler, lastScreenUV ).xy;
		if(motion.x != 0 || motion.y != 0)
		{
			lastUV = lastUV + motion / 2 ;
			half VBufferFlag = step(length(motion),0.0003);
			//	VBufferFlag = 1-smoothstep(0.0001,0.0005,motion);
			lastUV = lerp(motion+lastScreenUV,lastUV,VBufferFlag);
		}	
	#endif
	// float4 lastPos = MulMat4x4(lastViewProjTex, float4(worldPos,1));
	// float2 lastUV = lastPos.xy / lastPos.w;
	// lastUV += stc - tc;
	float lastFrameRate = cLastFrameRate;
// #if SD_TSAA_VBUFFER		
	// if (@_VelocityMapEnable)
	{
		float2 motionVector = tVBuffer.SampleLevel(sVBufferSampler, tc, 0).xy;
		//todo: use longest motion vector
		if (motionVector.x != 0 || motionVector.y != 0)
		{
			lastUV = stc + motionVector;
			// lastFrameRate = cLastFrameRate2;
			// OUT.a = 1.0;
		}
	}
// #endif
//	return float4(cLastFrameRate,0,0,1);
	// if (lastUV.x >= 0 && lastUV.x <= 1 && lastUV.y >= 0 && lastUV.y <= 1)//in screen
	// {
		//get border Pixels
		float4 c0 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2(ScreenInfo.z, 0));
		float4 c1 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2(-ScreenInfo.z, 0));
		float4 c2 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2(0, ScreenInfo.w));
		float4 c3 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2(0, -ScreenInfo.w));
		//float4 c4 = OUT.rgb;
		float4 c5, c6, c7, c8;
		if (TAALevel == 3)//ENABLE_SAMPLE3x3
		{
			c5 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2(-ScreenInfo.z, -ScreenInfo.w));
			c6 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2(-ScreenInfo.z,  ScreenInfo.w));
			c7 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2( ScreenInfo.z, -ScreenInfo.w));
			c8 = SampleColorMap(tScreenMap, sScreenSampler, tc + float2( ScreenInfo.z,  ScreenInfo.w));
		}

		//To YCoCg
		if (TAALevel >= 2)//#if ENABLE_YCOCG
		{
			c0.rgb = RGB2YCoCg(c0.rgb);
			c1.rgb = RGB2YCoCg(c1.rgb);
			c2.rgb = RGB2YCoCg(c2.rgb);
			c3.rgb = RGB2YCoCg(c3.rgb);
			c4.rgb = RGB2YCoCg(c4.rgb);
			if (TAALevel == 3)//ENABLE_SAMPLE3x3
			{
				c5.rgb = RGB2YCoCg(c5.rgb);
				c6.rgb = RGB2YCoCg(c6.rgb);
				c7.rgb = RGB2YCoCg(c7.rgb);
				c8.rgb = RGB2YCoCg(c8.rgb);	
			}		
		}

		//minmax
		float4 cMin = min(min(min(c0,c1), min(c2,c3)), c4);
		float4 cMax = max(max(max(c0,c1), max(c2,c3)), c4);
		if (TAALevel == 3)//ENABLE_SAMPLE3x3
		{
			cMin = min(min(min(c5,c6), min(c7,c8)), cMin);
			cMax = max(max(max(c5,c6), max(c7,c8)), cMax);	
		}

		//variance cliping
		if (TAALevel == 3)//ENABLE_SAMPLE3x3
		{
			float3 m1 = (c0 + c1 + c2 + c3 + c4 + c5 + c6 + c7 + c8).rgb;
			float3 m2 = (c0*c0 + c1*c1 + c2*c2 + c3*c3 + c4*c4 + c5*c5 + c6*c6 + c7*c7 + c8*c8).rgb;
			m1 /= 9.0;
			m2 /= 9.0;

			float3 sigma = sqrt(max(m2 - m1*m1, 0));
			cMin.rgb = max(cMin.rgb, m1 - sigma);
			cMax.rgb = min(cMax.rgb, m1 + sigma);
		}

		//history interpolation
		float4 history;
		if (TAALevel <= 0)
			history = SampleColorMap(tLastScreenMap, sLastScreenSample, lastUV);
		else if(TAALevel <= 2)
			history = bicubicSampleCatmullRom2x2(lastUV, ScreenInfo.xy);
		else
		{
			bicubicSampleCatmullRom3x3(lastUV, cRTSize.zw, tLastScreenMap, sLastScreenSample, history);
		}

		//clamp
		if (TAALevel >= 2)//#if ENABLE_YCOCG
		{
			history.rgb = RGB2YCoCg(history.rgb);
			history.rgb = clamp(history.rgb, cMin.rgb, cMax.rgb);
			history.rgb = YCoCg2RGB(history.rgb);
		}
		else
			history.rgb = clamp(history.rgb, cMin.rgb, cMax.rgb);
		//remove ghost
		// if (history.a > 0.9 && OUT.a == 0)
		// 	lastFrameRate *= 0.5;
		OUT.rgb = lerp(OUT.rgb, history.rgb, cLastFrameRate);
	
	// }
	return OUT;
}


#endif /*__FILE_MESSIAHTAA_FXH__*/
