#ifndef __SHADER_GRAPH_VARYING_FXH__
#define __SHADER_GRAPH_VARYING_FXH__

#ifndef VARYING_MASK
#define VARYING_MASK             0
#endif

#if (VARYING_MASK & VA_TEXCOORD) && !defined(VATexCoord)
#define VATexCoord(Arg)          Arg.TexCoord.xy
#define VSTexCoord               VATexCoord(OUT)
#define PSTexCoord               VATexCoord(IN)
#endif

#if (VARYING_MASK & VA_TEXCOORD2) && !defined(VATexCoord2)
#define VATexCoord2(Arg)         half2(0, 0)
#define VSTexCoord2              VATexCoord2(OUT)
#define PSTexCoord2              VATexCoord2(IN)
#endif

#if (VARYING_MASK & VA_TINT_COLOR) && !defined(VATintColor)
#define VATintColor(Arg)         Arg.TintColor
#define VSTintColor              VATintColor(OUT)
#define PSTintColor              VATintColor(IN)
#endif

#if (VARYING_MASK & VA_WORLD_NORMAL) && !defined(VAWorldNormal)
#define VAWorldNormal(Arg)       Arg.WorldNormal.xyz
#define VSWorldNormal            VAWorldNormal(OUT)
#define PSWorldNormal            VAWorldNormal(IN)
#endif

#if (VARYING_MASK & VA_WORLD_TANGENT) && !defined(VAWorldTangent)
float3 GetVAWorldTangent(VertexShadingOutput arg)
{
    if (@_TANGENT)
        return arg.WorldTangent.xyz;
    else
        return float3(0, 0, 0);
}
#define VAWorldTangent(Arg)      GetVAWorldTangent(Arg)
#define VSWorldTangent           VAWorldTangent(OUT)
#define PSWorldTangent           VAWorldTangent(IN)
#endif

#if (VARYING_MASK & VA_WORLD_BINORMAL) && !defined(VAWorldBinormal)
float3 GetVAWorldBinormal(VertexShadingOutput arg)
{
    if (@_TANGENT)
        return arg.WorldBinormal.xyz;
    else
        return float3(0, 0, 0);
}
#define VAWorldBinormal(Arg)     GetVAWorldBinormal(Arg)
#define VSWorldBinormal          VAWorldBinormal(OUT)
#define PSWorldBinormal          VAWorldBinormal(IN)
#endif

#if (VARYING_MASK & VA_WORLD_POSITION) && !defined(VAWorldPosition)
#define VAWorldPosition(Arg)     Arg.WorldPosition.xyz
#define VSWorldPosition          VAWorldPosition(OUT)
#define PSWorldPosition          VAWorldPosition(IN)
#endif

#if (VARYING_MASK & VA_VIEW_DIRECTION) && !defined(VAViewDirectoin)
#define VAViewDirectoin(Arg)     Arg.ViewDirection.xyz
#define VSViewDirection          VAViewDirectoin(OUT)
#define PSViewDirection          VAViewDirectoin(IN)
#endif

#if (VARYING_MASK & VA_LINEAR_DEPTH) && !defined(VALinearDepth)
#define VALinearDepth(Arg)       Arg.LinearZ
#define VSLinearDepth            VALinearDepth(OUT)
#define PSLinearDepth            VALinearDepth(IN)
#endif

#if (VARYING_MASK & VA_SCREEN_COORD) && !defined(VAScreenCoord)
#define VAScreenCoord(Arg)       Arg.HPosition
#define PSScreenCoord            VAScreenCoord(IN)
#endif

#endif /*__SHADER_GRAPH_VARYING_FXH__*/

