
// greatly reduces shadow mapping artifacts
float BiasedNDotL(float NDotLWithoutSaturate)
{
	return saturate(NDotLWithoutSaturate * 1.08f - 0.08f);
}
float Luminance(float3 LinearColor)
{
	return dot(LinearColor, float3(0.3, 0.59, 0.11));
}

float length2(float2 v)
{
	return dot(v, v);
}
float length2(float3 v)
{
	return dot(v, v);
}
float length2(float4 v)
{
	return dot(v, v);
}

// Clamp the base, so it's never <= 0.0f (INF/NaN).
float ClampedPow(float X, float Y)
{
	return pow(max(abs(X), 0.000001f), Y);
}
float2 ClampedPow(float2 X, float2 Y)
{
	return pow(max(abs(X), float2(0.000001f, 0.000001f)), Y);
}
float3 ClampedPow(float3 X, float3 Y)
{
	return pow(max(abs(X), float3(0.000001f, 0.000001f, 0.000001f)), Y);
}
float4 ClampedPow(float4 X, float4 Y)
{
	return pow(max(abs(X), float4(0.000001f, 0.000001f, 0.000001f, 0.000001f)), Y);
}

float4 ProcessMaterialColorTextureLookup(float4 TextureValue)
{
	return TextureValue;
}

float4 ProcessMaterialLinearColorTextureLookup(float4 TextureValue)
{
	return TextureValue;
}

float ProcessMaterialGreyscaleTextureLookup(float TextureValue)
{
	return TextureValue;
}

float ProcessMaterialLinearGreyscaleTextureLookup(float TextureValue)
{
	return TextureValue;
}

/** 
 * aligns the clip space position so that it can be used as a texture coordinate
 * to properly align in screen space
 */
float4 ScreenAlignedPosition( float4 ScreenPosition )
{
	return float4(ScreenPosition.xy / ScreenPosition.w * float2(0.5, -0.5) + 0.5, ScreenPosition.z/ScreenPosition.w,1);
}

/** 
 * Aligns the [0,1] UV to match the view within the backbuffer
 */
float2 ScreenAlignedUV( float2 UV )
{
	return (UV*float2(2,-2) + float2(-1,1))*float2(0.5, -0.5) + 0.5;
}

struct MaterialPixelParameters
{
	float2 TexCoords[2];
	// Interpolated vertex color, in linear color space
	float4 VertexColor;
	// Normalized world space normal
	float3 WorldNormal;
	// Normalized world space camera vector, which is from the pixel worldPos to the camera position.
	float3 CameraVector;
	// Post projection position, before the divide by W
	float4 ScreenPosition;
};

MaterialPixelParameters GetMaterialPixelParameters(CocosVertexOutput Input)
{
	MaterialPixelParameters Parameters;
	Parameters.TexCoords[0] = Input.TexCoord.xy;
	Parameters.TexCoords[1] = Input.TexCoord.zw;
	Parameters.WorldNormal = float3(0.0, 0.0, -1.0);
	Parameters.CameraVector = float3(0.0, 0.0, 1.0);
	Parameters.VertexColor = Input.Color;
	return Parameters;
}
