// {HEX} VXGI.
#ifndef _VXGI_COMMON_FXH_
#define _VXGI_COMMON_FXH_ 1

int SelectSlot(float3 N)
{
	float x = abs(N.x);
	float y = abs(N.y);
	float z = abs(N.z);
	float3 IsNegtive = (N > 0.0);
	if(x >= y && x >= z) return int(IsNegtive.x);
	if (y >= x && y >= z) return int(2 + IsNegtive.y);
	return int(4 + IsNegtive.z);
}

uint Float4ToUint(float4 Val)
{
	return  ((uint(clamp(Val.w * 255.0, 0, 255)) & uint(255)) << uint(24)) | 
			((uint(clamp(Val.z * 255.0, 0, 255)) & uint(255)) << uint(16)) | 
			((uint(clamp(Val.y * 255.0, 0, 255)) & uint(255)) << uint(8))  | 
			((uint(clamp(Val.x * 255.0, 0, 255)) & uint(255)));
}

float4 UintToFloat4(uint Val)
{
	return float4(
		float(Val & uint(255)) / 255.0,
		float(uint(Val >> uint(8))  & uint(255)) / 255.0,
		float(uint(Val >> uint(16)) & uint(255)) / 255.0,
		float(uint(Val >> uint(24)) & uint(255)) / 255.0
	);
}

int VXGITo1DCoord(int3 Index, int3 Dimension)
{
	return (Dimension.y * Index.z + Index.y) * Dimension.x + Index.x;
}

int3 VXGITo3DCoord(int Index, int3 Dimension)
{
	return int3(
		Index % Dimension.x,
		(Index / Dimension.x) % Dimension.y,
		Index / (Dimension.x * Dimension.y)
	);
}

int3 VXGIGetLocalCellIndex(float3 WorldPosition, float3 BoundsMin, float CellSize)
{
	return int3((WorldPosition - BoundsMin) / CellSize);
}

void VXGIGetGridRange(float CellSize, int3 Dimension, float3 Center, out int3 GridMin, out int3 GridMax)
{
	int3 Extend = Dimension / 2;
	int3 CenterIndex = int3(
		floor(Center.x / CellSize),
		floor(Center.y / CellSize),
		floor(Center.z / CellSize)
	);
	GridMin = CenterIndex - Extend;
	GridMax = CenterIndex + Extend - int3(1,1,1);
}

void VXGIGetGridBounds(float CellSize, int3 Dimension, float3 Center, out float3 BoundsMin, out float3 BoundsMax)
{
	int3 Extend = Dimension / 2;
	int3 CenterIndex = int3(
		floor(Center.x / CellSize),
		floor(Center.y / CellSize),
		floor(Center.z / CellSize)
	);
	BoundsMin = (CenterIndex - Extend) * CellSize;
	BoundsMax = (CenterIndex + Extend - int3(1,1,1)) * CellSize;
}

// Toroidal addressing used by clipmap.
int3 VXGIGetClipmapCoord(float3 WorldPosition, float CellSize, int3 Dimension, int Layer, int Slot)
{
	float3 ToroidalCoord = frac(WorldPosition / (Dimension * CellSize));
	int3 Coord = int3(ToroidalCoord * Dimension);
	Coord = int3(Coord.x * VXGIDirCount + Slot, Coord.y, Coord.z + Dimension.z * Layer);
	return Coord;
}

int3 VXGISlotsScale(int3 UV)
{
	return int3(UV.x * VXGIDirCount, UV.y, UV.z);
}

// LogicalAddress: logical address without slots scale. Layer: 0-2
int VXGIGetClipmapPhysicalAddress(int3 LogicalAddress, int Layer)
{
	int3 UV = int3(LogicalAddress.x * VXGIDirCount, LogicalAddress.y, LogicalAddress.z + VXGIColorDimension.z * Layer);
	return VXGITo1DCoord(UV, VXGIPhysicalDimension);
}

// LogicalAddress: logical address without slots scale. Layer: 1-3
int VXGIGetMipmapPhysicalAddress(int3 LogicalAddress, int Layer)
{
	int ClipmapOffset = VXGIColorDimension.x * VXGIDirCount * VXGIColorDimension.y * VXGIColorDimension.z * VXGIClipmapCascades;
	int MipmapOffset = 0;
	for (int i = 1; i < Layer; ++i)
	{
		int3 LayerDimension = int3(VXGIColorDimension.x >> i, VXGIColorDimension.y >> i, VXGIColorDimension.z >> i);
		MipmapOffset += LayerDimension.x * VXGIDirCount * LayerDimension.y * LayerDimension.z;
	}
	int3 CurPhysicalDimension = int3((VXGIColorDimension.x >> Layer) * VXGIDirCount, VXGIColorDimension.y >> Layer, VXGIColorDimension.z >> Layer);
	int LayerIndex = VXGITo1DCoord(VXGISlotsScale(LogicalAddress), CurPhysicalDimension);
	return ClipmapOffset + MipmapOffset + LayerIndex;
}

#endif // _VXGI_COMMON_FXH_
// {HEX} end
