// {HEX} VXGI.
#ifndef _VXGI_VOXELIZE_FXH_
#define _VXGI_VOXELIZE_FXH_ 1

#ifndef VoxelizeVS
#define VoxelizeVS PbrVertexShading
#endif

#ifndef VSOutput
#define VSOutput VertexShadingOutput
#endif

#include "Config.fxh"
#include "VXGI/VXGIConst.fxh"
#include "VXGI/VXGICommon.fxh"

RWTexture3D<uint> tVoxelColor;

#define VXGI_VOXELIZE_COLOR(CellSize, Dimension, Tex, Layer)\
{\
	float3 BoundsMin, BoundsMax;\
	VXGIGetGridBounds(CellSize, Dimension, VXGICurCenter, BoundsMin, BoundsMax);\
	if (all(WorldPosition >= BoundsMin) && all(WorldPosition <= BoundsMax))\
	{\
		float NoL = saturate(dot(N, SunDirection.xyz));\
		float3 Color = BaseColor.rgb * max(0, SunColor * NoL) + Emission.rgb;\
		int Face = SelectSlot(N);\
		int3 LayerCoord = VXGIGetClipmapCoord(WorldPosition, CellSize, Dimension, Layer, Face);\
		Tex[LayerCoord] = Float4ToUint(float4(Color, 1.0));\
	}\
}

#define VXGI_PREPARE_VOXELIZE()\
	float4 OUT = float4(0, 1, 0, 1);\
	half3 BaseColor = half3(1, 1, 1);\
	float3 N = float3(0, 1, 0);\
	half4 Emission = half4(0, 0, 0, 1);\
	half Alpha = 1.0;\
	MaterialShadingParameters MaterialParameters;\
	PS_GetParameters(IN, MaterialParameters);\
	BaseColor = MaterialParameters.BaseColor;\
	Emission.rgb = MaterialParameters.Emission;\
	Alpha = MaterialParameters.Alpha;\
	N = normalize(IN.WorldNormal);\
	float3 WorldPosition = IN.WorldPosition.xyz;

float4 VoxelizeLayer0PS(in VSOutput IN) : SV_Target
{
	VXGI_PREPARE_VOXELIZE();
	VXGI_VOXELIZE_COLOR(VXGIColorCellSizes.x, VXGIColorDimension, tVoxelColor, 0);
	return OUT;
}

float4 VoxelizeLayer1PS(in VSOutput IN) : SV_Target
{
	VXGI_PREPARE_VOXELIZE();
	VXGI_VOXELIZE_COLOR(VXGIColorCellSizes.y, VXGIColorDimension, tVoxelColor, 1);
	return OUT;
}

float4 VoxelizeLayer2PS(in VSOutput IN) : SV_Target
{
	VXGI_PREPARE_VOXELIZE();
	VXGI_VOXELIZE_COLOR(VXGIColorCellSizes.z, VXGIColorDimension, tVoxelColor, 2);
	return OUT;
}

technique VXGIVoxelizeLayer0
{
	VertexShader = VoxelizeVS();
	PixelShader = VoxelizeLayer0PS();
}

technique VXGIVoxelizeLayer1
{
	VertexShader = VoxelizeVS();
	PixelShader = VoxelizeLayer1PS();
}

technique VXGIVoxelizeLayer2
{
	VertexShader = VoxelizeVS();
	PixelShader = VoxelizeLayer2PS();
}
#endif // _VXGI_VOXELIZE_FXH_
// {HEX} end
