
texture2D tScatteringAndTransmittanceVolume : VolumetricFogVolume : layout(t15) = "Black";
sampler sScatteringAndTransmittanceVolume : layout(s15)
{
	Texture = tScatteringAndTransmittanceVolume;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

#define ZA VolumetricFogParam.x
#define ZB VolumetricFogParam.y
#define SliceSize VolumetricFogParam.zw

#define SLICE_PERROW (16)
#define SLICE_COUNT (SLICE_PERROW*SLICE_PERROW)
#define SLICE_UVSIZE (1.0/16.0)
#define GRID_SIZE float3(SliceSize, SLICE_COUNT)  
//sliceIdx in range [0,63]
float2 UVBeginOfSlice(float sliceIdx)
{
	float sliceY = floor(sliceIdx/SLICE_PERROW);
	float sliceX = sliceIdx - sliceY * SLICE_PERROW;
	return float2(sliceX*SLICE_UVSIZE, sliceY*SLICE_UVSIZE);
}
//sliceIdx must be an integer
float2 SliceUVTo2DUV(float2 inSliceUV, float sliceIdx)
{
	return inSliceUV * SLICE_UVSIZE + UVBeginOfSlice(sliceIdx);
}
//gridIdx.xy can be an float value, gridIdx.z must be an integer
float2 GridIdxTo2DUV(float3 gridIdx)
{
	float2 inSliceUV = (gridIdx.xy + 0.5) / GRID_SIZE.xy;
	return SliceUVTo2DUV(inSliceUV, gridIdx.z);
}
//zuv = (z^0.5 - N^0.5) / (F^0.5 - N^0.5)
//0:zn:0, 1:zf:64
float DepthToSliceIdx(float SceneDepth)
{
	float z = sqrt(SceneDepth);//TODO: NAN
	return saturate((z*ZA + ZB)) * SLICE_COUNT;
}
float SLiceIdxToDepth(float sliceIdx)
{
	float rsqz = (sliceIdx/SLICE_COUNT - ZB) / ZA;
	float z = rsqz * rsqz;
	return z;
}
//xy:sliceUV, z:sliceIdxUV, w-SceneDepth
float4 WorldPosToVolumeUV(float3 WorldPosition, float4x4 WorldToUV)
{
	float4 HPosition = mul(float4(WorldPosition, 1), WorldToUV);
	HPosition.xy /= HPosition.w;
	return float4(HPosition.xy, DepthToSliceIdx(HPosition.w)-0.5, HPosition.w);
}
#define SampleVolume(VolumeTexture, VolumeSampler, VolumeUV, VolumeData) {\
	float sliceA = floor(max(0,VolumeUV.z));\
	float sliceB = min(sliceA+1, SLICE_COUNT-1);\
	float sliceZA = SLiceIdxToDepth(sliceA+0.5);\
	float sliceZB = SLiceIdxToDepth(sliceB+0.5);\
	float lerpAB = saturate((VolumeUV.w - sliceZA) / max(EPSILON, sliceZB - sliceZA));\
	lerpAB = VolumeUV.z - sliceA;\
	float2 uvA = SliceUVTo2DUV(VolumeUV.xy, sliceA);\
	float2 uvB = SliceUVTo2DUV(VolumeUV.xy, sliceB);\
	float4 dataA = VolumeTexture.SampleLevel(VolumeSampler, uvA, 0);\
	float4 dataB = VolumeTexture.SampleLevel(VolumeSampler, uvB, 0);\
	VolumeData = lerp(dataA, dataB, lerpAB);\
}

float3 HPositionToGridIdx(float2 spos)
{
	float2 sliceXY = floor((spos.xy+0.5) / SliceSize);//avoiding floating point accuracy problem 
	float sliceIdx = sliceXY.y * SLICE_PERROW + sliceXY.x;
	float3 gridIdx;
	gridIdx.xy = spos - sliceXY * SliceSize;
	gridIdx.z = sliceIdx;
	return gridIdx;
}
float IsotropicPhase()
{
	return 1 / (4 * PI);
}
float HenyeyGreensteinPhase(float g, float CosTheta)
{
	return (1 - g * g) / (4 * PI * pow(1 + g * g - 2 * g * CosTheta, 1.5));
}
float SchlickPhase(float k, float CosTheta)
{
	float Inner = (1 + k * CosTheta);
	return (1 - k * k) / (4 * PI * Inner * Inner);
}
float RaleighPhase(float CosTheta)
{
	return 3.0f * (1.0f + CosTheta * CosTheta) / (16.0f * PI);
}
// Positive g = forward scattering
// Zero g = isotropic
// Negative g = backward scattering
float PhaseFunction(float g, float CosTheta)
{
	return IsotropicPhase();//todo: anisotropic phase function？
	return HenyeyGreensteinPhase(g, CosTheta);
}

float4 VolumetricFogLighting(in float2 stc, in float worldDepth)
{
	float2 inSliceUV = stc;
	float depth = worldDepth;
	float sliceIdxUV = DepthToSliceIdx(depth) - 0.5;
	float4 ScatteringAndTransmittance;
	inSliceUV.xy = clamp(inSliceUV.xy, 0.5/SliceSize, 1-0.5/SliceSize);//avoiding wrong bilinear filter between two slices.
	SampleVolume(tScatteringAndTransmittanceVolume, sScatteringAndTransmittanceVolume, float4(inSliceUV, sliceIdxUV, depth), ScatteringAndTransmittance);
	ScatteringAndTransmittance.w = lerp(FOG_MINT, 0.999, ScatteringAndTransmittance.w);
	return ScatteringAndTransmittance;
}


