//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<LocalHdr.fx>
///	@path	~/System/Shader/
///	@date	2014/11/05
///	@brief	Local HDR.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////
@_EnableAdaption
{
	Name = enableAdaption;
	Description = "use eye adaption";
	AffectedTex = tLumMap;
	Visible = true;
};
@_EnableBloom
{
	Name = enableBloom;
	Description = "use bloom";
	Visible = true;
};
@_HasBloomLayer2
{
	Name = hasBloomLayer2;
	Description = "has bloom layer 2";
	AffectedTex = tBloomMap2;
	Visible = true;
};
@_HasBloomLayer3
{
	Name = hasBloomLayer3;
	Description = "has bloom layer 3";
	AffectedTex = tBloomMap3;
	Visible = true;
};
@_HasBloomLayer4
{
	Name = hasBloomLayer4;
	Description = "has bloom layer 4";
	AffectedTex = tBloomMap4;
	Visible = true;
};
@_VignettingEnable
{
	Name = VignettingEnable;
	Description = "VignettingEnable";
	Visible = true;
};
@_EnableDetailEnhancement
{
	Name = EnableDetailEnhancement;
	Description = "Enable Detail Enhancement";
	Visible = true;
};
@_EnableChromaticAberration
{
	Name = enableChromaticAberration;
	Description = "Enable Chromatic Aberration";
	Visible = true;
};
@_OutputHDRDisplay
{
	Name = outputHDRDisplay;
	Description = "Output to HDR display";
	Visible = true;
};
@_HDRColorGrading
{
	Name = hdrColorGrading;
	Description = "Enable hdr color grading";
	AffectedTex = tColorGradingMap;
	Visible = true;
};
@_ColorCorrectionEnable
{
	Name = ColorCorrectionEnable;
	Description = "Color correction Enable";
	Visible = true;
}
@_RadarEnable
{
	Name = RadarEnable;
	Description = "RadarEnable";
	Visible = true;
}

///////////////// const buffer //////////////////
#define ElapsedTime 		cHDRParam1.x
#define EyeAdaptionLevel	cHDRParam1.y
#define EyeAdaptionScale	cHDRParam1.z
#define BrightThreshold		cHDRParam1.w

cbuffer Batch
{
	//day sky
	float4 cPixelSize					;	//texture (1/w, -1/w, 1/h,-1/h)
	float4 cHDRParam1					;
	float4 cChromaticAberrationParams	;
	float  cLocalRate					;
	float3 cBloomTint					;
	float  cAdaptedLum					;
	float  cVignettingDarkenss 	 		;
	float  cDetailEnhancementLevel  	;
	float  cExposureCompensation		;
	float  cSpeedUp						;
	float  cSpeedDown					;
	float  cMinBrightness				;
	float  cMaxBrightness				;	
	// white blance
	float4 cSaturation					;
	float4 cContrast					;
	float4 cGamma						;
	float4 cGain						;
	float4 cOffset						;
	float cWhiteTemp					;
	float cWhiteTint					;

};
cbuffer Shader
{
	float cRadarPercentage				;

};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tBaseMapPoint : layout(t0);
sampler sBasePointSampler : layout(s0)
{
	Texture = tBaseMapPoint;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBaseMapLinear : layout(t1);
sampler sBaseLinearSampler : layout(s1)
{
	Texture = tBaseMapLinear;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tLumMap : layout(t2);
sampler sLumSampler : layout(s2)
{
	Texture = tLumMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBloomMap0 : layout(t3);
sampler sBloomSampler0 : layout(s3)
{
	Texture = tBloomMap0;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBloomMap1 : layout(t4);
sampler sBloomSampler1 : layout(s4)
{
	Texture = tBloomMap1;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBloomMap2 : layout(t5);
sampler sBloomSampler2 : layout(s5)
{
	Texture = tBloomMap2;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBaseLayer : layout(t6);
sampler sBaseLayer : layout(s6)
{
	Texture = tBaseLayer;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBloomMap3 : layout(t7);
sampler sBloomSampler3 : layout(s7)
{
	Texture = tBloomMap3;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tBloomMap4 : layout(t8);
sampler sBloomSampler4 : layout(s8)
{
	Texture = tBloomMap4;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tColorGradingMap : layout(t9);
sampler sColorGradingSampler : layout(s9)
{
	Texture = tColorGradingMap;
	Filter = Linear;
	Address = Clamp;
};

//for HDRCalculateAdaptedLum
texture2D tLumMap0 : layout(t0);
sampler sLumSampler0 : layout(s0)
{
	Texture = tLumMap0;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tLumMap1 : layout(t1);
sampler sLumSampler1 : layout(s1)
{
	Texture = tLumMap1;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f_hdr
{
	float4 HPosition 	: SV_Position;
	float2 TC			: TEXCOORD0;
};

struct v2f_hdr_fp
{
	float4 HPosition 	: SV_Position;
	float2 TC			: TEXCOORD0;
	float2 ScreenPos	: TEXCOORD1;
};

/////////////////// HDRSampleLumInitial ////////////////////
v2f_hdr ScreenVS( a2v IN )
{
	v2f_hdr OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;

	return OUT;
}

#define USE_LUM_AVG 1//log avg need f16 rt
float4 HDRSampleLumInitialPS(v2f_hdr IN) : SV_Target
{
	half fRecipSampleCount = 0.25;
	half avLum = 0.;

	//cPixelSize(1/w, -1/w, 1/h,-1/h)
	//-1,-1
	half3 cTex = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.yw).rgb;
	half fLum = GetLuminance(cTex.rgb);

	//return float4(IN.TC.xy,1,fLum);

#if USE_LUM_AVG
	avLum += fLum;
#else
	avLum += log(fLum + 1e-6);
#endif
	//-1,1
	cTex = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.yz).rgb;
	fLum = GetLuminance(cTex.rgb);
#if USE_LUM_AVG
	avLum += fLum;
#else
	avLum += log(fLum + 1e-6);
#endif
	//1,-1
	cTex = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.xw).rgb;
	fLum = GetLuminance(cTex.rgb);
#if USE_LUM_AVG
	avLum += fLum;
#else
	avLum += log(fLum + 1e-6);
#endif
	//1,1
	cTex = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.xz).rgb;
	fLum = GetLuminance(cTex.rgb);
#if USE_LUM_AVG
	avLum += fLum;
#else
	avLum += log(fLum + 1e-6);
#endif
	
	avLum *= fRecipSampleCount;
	return avLum.rrrr;
}

technique HDRSampleLumInitial
{
	VertexShader = ScreenVS();
	PixelShader = HDRSampleLumInitialPS();
}

/////////////////// HDRSampleLumIterative ////////////////////
float4 HDRSampleLumIterativePS(v2f_hdr IN) : SV_Target
{
	half nRecipIter = 0.25;

	//cPixelSize(1/w, -1/w, 1/h,-1/h)
	//-1,-1
	half fLum = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.yw).r;
	half vResampleSum = fLum;
	//-1,1
	fLum = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.yz).r;
	vResampleSum += fLum;
	//1,-1
	fLum = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.xw).r;
	vResampleSum += fLum;
	//1,1
	fLum = tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + cPixelSize.xz).r;
	vResampleSum += fLum;

	vResampleSum *= nRecipIter;
	return vResampleSum.rrrr;
}

technique HDRSampleLumIterative
{
	VertexShader = ScreenVS();
	PixelShader = HDRSampleLumIterativePS();
}

/////////////////// HDRCalculateAdaptedLum ////////////////////


float4 HDRCalculateAdaptedLumPS(v2f_hdr IN) : SV_Target
{
	half2 vAdaptedLum = tLumMap0.Sample(sLumSampler0, IN.TC.xy).rg;
	half vCurrentLum = tLumMap1.Sample(sLumSampler1, IN.TC.xy).r;
	
	[@_ENCODEHDR]
	vAdaptedLum.r = R8G8ToHalf(vAdaptedLum.rg);

	half Diff = vCurrentLum - vAdaptedLum.r;
    half AdaptionSpeed = (Diff > 0) ? cSpeedUp : cSpeedDown;
    half Factor = 1.0 - exp2(-ElapsedTime * AdaptionSpeed);
	half vNewAdaptation = clamp(vAdaptedLum.r + Diff * Factor, 0.1, SD_SCENE_MAX_LUMINANCE);

	if(@_ENCODEHDR)
		return HalfToR8G8(vNewAdaptation).rgrg;
	else
		return vNewAdaptation.rrrr;
}

technique HDRCalculateAdaptedLum
{
	VertexShader = ScreenVS();
	PixelShader = HDRCalculateAdaptedLumPS();
}

/////////////////// HDRBrightPassFilter ////////////////////
float4 HDRBrightPassFilterPS(v2f_hdr IN) : SV_Target
{
	//half AdaptedLum = cAdaptedLum;
	half AdaptedLum = 1.0;
	if (@_EnableAdaption)
	{
		half2 _AdaptedLum = tLumMap.Sample(sLumSampler, IN.TC).rg;
		[@_ENCODEHDR]
		_AdaptedLum.r = R8G8ToHalf(_AdaptedLum.rg);
		AdaptedLum = DecodeHDR(_AdaptedLum.r);//linear encode
		AdaptedLum = clamp(cAdaptedLum * AdaptedLum.r * PI, cMinBrightness, cMaxBrightness);
	}
	// AdaptedLum*=EyeAdaptionScale;
	// half AdaptedLumDest = cExposureCompensation / (EyeAdaptionLevel + AdaptedLum);
	half AdaptedLumDest = 1.0 / AdaptedLum;

	float3 sum = float3(0, 0, 0);
	half A,B;
#define BrightPassSample(i, tc) \
	half4 S##i = DecodeHDR(tBaseMapLinear.Sample(sBaseLinearSampler, tc)); \
	A = max(0.0001, GetLuminance(S##i.rgb) * AdaptedLumDest); \
	B = max(A - BrightThreshold, 0.0); \
	S##i.rgb *= B / A; \
	sum += S##i.rgb;
	
	BrightPassSample(0, (IN.TC.xy + cPixelSize.yw));
	BrightPassSample(1, (IN.TC.xy + cPixelSize.yz));
	BrightPassSample(2, (IN.TC.xy + cPixelSize.xw));
	BrightPassSample(3, (IN.TC.xy + cPixelSize.xz));
#undef BrightPassSample

	if (@_ENCODEHDR)
		return float4(sum / (4.0*SD_SCENE_MAX_LUMINANCE), 1.0);
	else
		return float4(sum / 4.0, 1.0);
	
	// this way we check for NAN (illegal floats) as HLSL optimized the isnan() away
	// visible mostly on NVIDIA, if possible this should be removed and no shader should produce NAN
	//vSample.rgb = (vSample.rgb> 10000.0f)? half3(1, 1, 1): vSample.rgb;
}

technique HDRBrightPassFilter
{
	VertexShader = ScreenVS();
	PixelShader = HDRBrightPassFilterPS();
}

/////////////////// HDRBaseLayerFilter ////////////////////
float4 HDRBaseLayerFilterPS(v2f_hdr IN) : SV_Target
{
	half3 sumColor = DecodeHDR(tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + float2(-ScreenInfo.z, -ScreenInfo.w)).rgb);
	sumColor += DecodeHDR(tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + float2(-ScreenInfo.z, ScreenInfo.w)).rgb);
	sumColor += DecodeHDR(tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + float2(ScreenInfo.z, -ScreenInfo.w)).rgb);
	sumColor += DecodeHDR(tBaseMapLinear.Sample(sBaseLinearSampler, IN.TC.xy + float2(ScreenInfo.z, ScreenInfo.w)).rgb);
	float4 OUT = float4(sumColor.rgb/4.0, 1.0);
	EncodeHDR(OUT);
	return OUT;
}

technique HDRBaseLayerFilter
{
	VertexShader = ScreenVS();
	PixelShader = HDRBaseLayerFilterPS();
}

/////////////////// HDRFinalPass ////////////////////
float2 UVToScreenPos(float2 UV)
{
	[!@_GLSL]
	UV.y = 1.0 - UV.y;

	return (UV - 0.5) * 2.0;
}

float2 ScreenPosToUV(float2 ScreenPos)
{
	float2 UV = (ScreenPos * 0.5 + 0.5);

	[!@_GLSL]
	UV.y = 1.0 - UV.y;

	return UV;
}

v2f_hdr_fp ScreenFinalPassVS( a2v IN )
{
	v2f_hdr_fp OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;

	// Chromatic Aberration
	OUT.ScreenPos = UVToScreenPos(IN.BaseTC);
	return OUT;
}

half3 ApplyBlueShift(half3 HDRColor, half Luminance)
{
	half3 BlueTarget = half3(0.46, 0.46, 1) * Luminance;
	return lerp(HDRColor, BlueTarget, BlueShiftLevel);
}

float3 LinearToLog(float3 x)
{
	half3 y = log2(5.34 * x + 0.04) * 0.1 + 0.45;
	return saturate(y);
}

float3 LogToLinear(float3 x)
{
	return exp2(10 * x - 4.5) * 0.187267 - 0.0075;
}

half3 ColorLookupTable(float3 linearColor)
{
	// 33x33x33 lut, unwrapped as a 1089 * 33 2d texture
	#define LUTSize 33.0f

	float3 logColor = LinearToLog(linearColor);
	float3 UVW = logColor * ((LUTSize - 1) / LUTSize) + (0.5f / LUTSize);

	float IntW = floor( UVW.z * LUTSize - 0.5 );
	half FracW = UVW.z * LUTSize - 0.5 - IntW;

	float U = ( UVW.x + IntW ) / LUTSize;
	float V = UVW.y;

	half3 RG0 = tColorGradingMap.SampleLevel(sColorGradingSampler, float2(U, V), 0).rgb;
	half3 RG1 = tColorGradingMap.SampleLevel(sColorGradingSampler, float2(U + 1.0f / LUTSize, V), 0).rgb;
	return lerp(RG0, RG1, FracW);
}

// @param LinearColor in Rec.709 primaries.
half3 ColorCorrect( half3 LinearColor)
{
	half3 WorkingColor = SRGB2AP1(LinearColor);

	float3 AP1_RGB2Y = float3(0.2722287168,0.6740817658,0.0536895174);

	float Luma = dot( WorkingColor, AP1_RGB2Y);
	WorkingColor = max( 0, lerp( Luma.xxx, WorkingColor, cSaturation.xyz*cSaturation.w ) );
	WorkingColor = pow( WorkingColor * (1.0 / 0.18), cContrast.xyz*cContrast.w ) * 0.18;
	WorkingColor = pow( WorkingColor, 1.0 / (cGamma.xyz*cGamma.w) );
	WorkingColor = WorkingColor * (cGain.xyz * cGain.w) + (cOffset.xyz + cOffset.w);
	return AP12SRGB(WorkingColor);
}

// Accurate for 4000K < Temp < 25000K
// in: correlated color temperature
// out: CIE 1931 chromaticity
float2 D_IlluminantChromaticity( float Temp )
{
	// Correct for revision of Plank's law
	// This makes 6500 == D65
	Temp *= 1.4388 / 1.438;

	float x =	Temp <= 7000 ?
				0.244063 + ( 0.09911e3 + ( 2.9678e6 - 4.6070e9 / Temp ) / Temp ) / Temp :
				0.237040 + ( 0.24748e3 + ( 1.9018e6 - 2.0064e9 / Temp ) / Temp ) / Temp;
	
	float y = -3 * x*x + 2.87 * x - 0.275;

	return float2(x,y);
}

// Accurate for 1000K < Temp < 15000K
// [Krystek 1985, "An algorithm to calculate correlated colour temperature"]
float2 PlanckianLocusChromaticity( float Temp )
{
	float u = ( 0.860117757f + 1.54118254e-4f * Temp + 1.28641212e-7f * Temp*Temp ) / ( 1.0f + 8.42420235e-4f * Temp + 7.08145163e-7f * Temp*Temp );
	float v = ( 0.317398726f + 4.22806245e-5f * Temp + 4.20481691e-8f * Temp*Temp ) / ( 1.0f - 2.89741816e-5f * Temp + 1.61456053e-7f * Temp*Temp );

	float x = 3*u / ( 2*u - 8*v + 4 );
	float y = 2*v / ( 2*u - 8*v + 4 );

	return float2(x,y);
}

float2 PlanckianIsothermal( float Temp, float Tint )
{
	float u = ( 0.860117757f + 1.54118254e-4f * Temp + 1.28641212e-7f * Temp*Temp ) / ( 1.0f + 8.42420235e-4f * Temp + 7.08145163e-7f * Temp*Temp );
	float v = ( 0.317398726f + 4.22806245e-5f * Temp + 4.20481691e-8f * Temp*Temp ) / ( 1.0f - 2.89741816e-5f * Temp + 1.61456053e-7f * Temp*Temp );

	float ud = ( -1.13758118e9f - 1.91615621e6f * Temp - 1.53177f * Temp*Temp ) / Square( 1.41213984e6f + 1189.62f * Temp + Temp*Temp );
	float vd = (  1.97471536e9f - 705674.0f * Temp - 308.607f * Temp*Temp ) / Square( 6.19363586e6f - 179.456f * Temp + Temp*Temp );

	float2 uvd = normalize( float2( u, v ) );

	// Correlated color temperature is meaningful within +/- 0.05
	u += -uvd.y * Tint * 0.05;
	v +=  uvd.x * Tint * 0.05;
	
	float x = 3*u / ( 2*u - 8*v + 4 );
	float y = 2*v / ( 2*u - 8*v + 4 );

	return float2(x,y);
}

// Transformations between CIE XYZ tristimulus values and CIE x,y 
// chromaticity coordinates
float3 XYZ_2_xyY( float3 XYZ )
{  
	float3 xyY;
	float divisor = (XYZ[0] + XYZ[1] + XYZ[2]);
	if (divisor == 0.) divisor = 1e-10;
	xyY[0] = XYZ[0] / divisor;
	xyY[1] = XYZ[1] / divisor;  
	xyY[2] = XYZ[1];
  
	return xyY;
}

float3 xyY_2_XYZ( float3 xyY )
{
	float3 XYZ;
	XYZ[0] = xyY[0] * xyY[2] / max( xyY[1], 1e-10);
	XYZ[1] = xyY[2];  
	XYZ[2] = (1.0 - xyY[0] - xyY[1]) * xyY[2] / max( xyY[1], 1e-10);

	return XYZ;
}

float3x3 ChromaticAdaptation( float2 src_xy, float2 dst_xy )
{
	// Von Kries chromatic adaptation 
	static const float3x3 ConeResponse =
	{
		 0.8951,  0.2664, -0.1614,
		-0.7502,  1.7135,  0.0367,
		 0.0389, -0.0685,  1.0296
	};
	static const float3x3 InvConeResponse =
	{
		 0.9869929, -0.1470543,  0.1599627,
		 0.4323053,  0.5183603,  0.0492912,
		-0.0085287,  0.0400428,  0.9684867
	};

	float3 src_XYZ = xyY_2_XYZ( float3( src_xy, 1 ) );
	float3 dst_XYZ = xyY_2_XYZ( float3( dst_xy, 1 ) );

	float3 src_coneResp = mul( ConeResponse, src_XYZ );
	float3 dst_coneResp = mul( ConeResponse, dst_XYZ );

	return mul( InvConeResponse, mul( float3x3(dst_coneResp[0] / src_coneResp[0], 0.0, 0.0,0.0, dst_coneResp[1] / src_coneResp[1], 0.0,0.0, 0.0, dst_coneResp[2] / src_coneResp[2]), ConeResponse ) );
}

float3 WhiteBalance( float3 LinearColor )
{
	float2 SrcWhiteDaylight = D_IlluminantChromaticity( cWhiteTemp );
	float2 SrcWhitePlankian = PlanckianLocusChromaticity( cWhiteTemp );

	//float factor = step(cWhiteTemp, 4000);
	float2 SrcWhite = cWhiteTemp < 4000 ? SrcWhitePlankian : SrcWhiteDaylight;
	float2 D65White = float2( 0.31270,  0.32900 );

	{
		// Offset along isotherm
		float2 Isothermal = PlanckianIsothermal( cWhiteTemp, cWhiteTint ) - SrcWhitePlankian;
		SrcWhite += Isothermal;
	}

	static const float3x3 sRGB_2_XYZ_MAT =
	{
		0.4124564, 0.3575761, 0.1804375,
		0.2126729, 0.7151522, 0.0721750,
		0.0193339, 0.1191920, 0.9503041
	};

	static const float3x3 XYZ_2_sRGB_MAT =
	{
		3.2409699419, -1.5373831776, -0.4986107603,
		-0.9692436363,  1.8759675015,  0.0415550574,
		0.0556300797, -0.2039769589,  1.0569715142
	};

	float3x3 WhiteBalanceMat = ChromaticAdaptation( SrcWhite, D65White );
	WhiteBalanceMat = mul( XYZ_2_sRGB_MAT, mul( WhiteBalanceMat, sRGB_2_XYZ_MAT ) );

	return mul( WhiteBalanceMat, LinearColor );
}

half3 ToneMapping(half3 HDRColor, half3 BaseColor, half3 Bloom, half AdaptedLum, half Vignetting)
{
	//+- EV
	HDRColor *= cExposureCompensation;
	BaseColor *= cExposureCompensation;
	Bloom *= cExposureCompensation;

	if (@_OutputHDRDisplay)
	{
		// Exposure
		half AdaptedLumDest = 1.0 / saturate(EyeAdaptionLevel + AdaptedLum);

		// We have a MaxFALL of 180 nits. (1.04 in Messiah units)
		AdaptedLumDest = min(1.04 / AdaptedLum, AdaptedLumDest);

		// Output HDR color directly.
		HDRColor = Vignetting * AdaptedLumDest * (HDRColor + Bloom);
		HDRColor = max(HDRColor, 0);

		if (@_HDRColorGrading)
		{
			HDRColor = ColorLookupTable(HDRColor);
			HDRColor = LogToLinear(HDRColor);
		}

		EncodeHDR(HDRColor);
		return HDRColor;
	}
	else
	{
		// Exposure
		//half AdaptedLumDest = 1.0 / (EyeAdaptionLevel + AdaptedLum);
		half AdaptedLumDest = 1.0 / AdaptedLum;
		half3 Offset = HDRColor - BaseColor;
		HDRColor = Vignetting * AdaptedLumDest * (HDRColor + Bloom);
		//对齐UE
		half GrayColor = GetLuminance(HDRColor);
		HDRColor = lerp(half3(GrayColor,GrayColor,GrayColor), HDRColor, 0.9);

		if (@_ColorCorrectionEnable)
		{
			HDRColor = WhiteBalance(HDRColor);
			HDRColor = ColorCorrect(HDRColor);
		}

		// Tonemapping
		half3 LDRColor = ACESFilm(HDRColor);

		// Detail Enhance
		if (@_EnableDetailEnhancement)
		{
			Offset *= AdaptedLumDest * Vignetting;
			LDRColor = max(LDRColor*0.8, LDRColor + Offset * cDetailEnhancementLevel);
		}
		LDRColor = saturate(LDRColor);
		
		// Encode Gamma
		return EncodeGamma(LDRColor);
	}
}

float4 LocalHdrPS(v2f_hdr_fp IN) : SV_Target
{
	//half AdaptedLum = cAdaptedLum;
	half AdaptedLum = 1.0;
	if (@_EnableAdaption)
	{
		half2 _AdaptedLum = tLumMap.Sample(sLumSampler, IN.TC.xy).rg;
		[@_ENCODEHDR]
		_AdaptedLum.r = R8G8ToHalf(_AdaptedLum.rg);
		AdaptedLum = DecodeHDR(_AdaptedLum.r);//linear encode
		AdaptedLum = clamp(cAdaptedLum * AdaptedLum.r * PI, cMinBrightness, cMaxBrightness);
#if !USE_LUM_AVG
	AdaptedLum = exp(AdaptedLum);
#endif
	}
	//AdaptedLum*=EyeAdaptionScale;

	half3 Bloom = half3(0, 0, 0);
	if(@_EnableBloom)
	{
		Bloom = tBloomMap0.Sample(sBloomSampler0, IN.TC).rgb * 0.3 +
			tBloomMap1.Sample(sBloomSampler1, IN.TC).rgb * 0.4;
		[@_HasBloomLayer2]
		Bloom += tBloomMap2.Sample(sBloomSampler2, IN.TC).rgb * 0.5;
		[@_HasBloomLayer3]
		Bloom += tBloomMap3.Sample(sBloomSampler3, IN.TC).rgb * 0.5;
		[@_HasBloomLayer4]
		Bloom += tBloomMap4.Sample(sBloomSampler4, IN.TC).rgb * 0.6;
		Bloom *= cBloomTint;
		[@_ENCODEHDR]
		Bloom *= SD_SCENE_MAX_LUMINANCE;
	}

	half Vignetting = 1;
	if (@_VignettingEnable)
	{
		Vignetting = CalcVignetting(IN.TC.xy, cVignettingDarkenss);
	}

	// Detail Enhancement
	float4 HDRColor = float4(0,0,0,0);
	half3 BaseColor = half3(0, 0, 0);
	if(@_EnableChromaticAberration)
	{
		// Fringe
		float2 CAScale = cChromaticAberrationParams.rg;
		float StartOffset = cChromaticAberrationParams.z;

		float4 CAUV;
		CAUV = IN.ScreenPos.xyxy - sign(IN.ScreenPos).xyxy * saturate(abs(IN.ScreenPos) - StartOffset).xyxy * CAScale.rrgg;
		CAUV.xy = ScreenPosToUV(CAUV.xy);
		CAUV.zw = ScreenPosToUV(CAUV.zw);

		half4 SceneColor = tBaseMapPoint.Sample(sBasePointSampler, CAUV.xy).rgba;
		half SceneColorG = tBaseMapPoint.Sample(sBasePointSampler, CAUV.zw).g;
		half SceneColorB = tBaseMapPoint.Sample(sBasePointSampler, IN.TC).b;
		SceneColor.g = SceneColorG;
		SceneColor.b = SceneColorB;

		HDRColor = DecodeHDR(SceneColor).rgba;
		if (@_EnableDetailEnhancement)
		{
			// BaseColor = DecodeHDR(tBaseLayer.Sample(sBaseLayer, IN.TC)).rgb;
			SceneColor = tBaseLayer.Sample(sBaseLayer, CAUV.xy).rgba;
			SceneColorG = tBaseLayer.Sample(sBaseLayer, CAUV.zw).g;
			SceneColorB = tBaseLayer.Sample(sBaseLayer, IN.TC).b;
			SceneColor.g = SceneColorG;
			SceneColor.b = SceneColorB;
			BaseColor = DecodeHDR(SceneColor).rgba;
		}
		
	}
	else
	{
		HDRColor = DecodeHDR(tBaseMapPoint.Sample(sBasePointSampler, IN.TC)).rgba;
		if (@_EnableDetailEnhancement)
			BaseColor = DecodeHDR(tBaseLayer.Sample(sBaseLayer, IN.TC)).rgb;
	}
	
	half3 LDRColor = ToneMapping(HDRColor.rgb, BaseColor, Bloom, AdaptedLum.r, Vignetting);
		if (@_RadarEnable)
	{
		half3 Source = LDRColor;
		half3 Result = GetLuminance(Source);
		
		// Shift Color
		Result.r *= 0.5;
		Result.g *= 0.5;
		Result.b *= 0.5;
		
		// Darken corner.
		Result *= 1 - abs(0.5*0.8 - IN.TC.x*0.8) * abs(0.5*0.8 - IN.TC.y*0.8 ) * 5;

		// blend with sample.
		LDRColor = lerp(Source, Result, cRadarPercentage);
	}
	return float4(LDRColor, HDRColor.a);
}

technique LocalHdr
{
	VertexShader = ScreenFinalPassVS();
	PixelShader = LocalHdrPS();
}

