//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<HeightmapGen.fx>
///	@path	~/System/Shader/
///	@date	2018/07/26
///	@brief	Motin Blur.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////
@_LastMapEnable
{
	Name = LastMapEnable;
	Default = false;
	Visible = false;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 cLastMin				;
	float4 cLastMax				;
	float4 cCurrMin				;
	float4 cCurrMax				;
	float cAccumulation			;
	float cPixelSize			;
	float cMaxStep				;
};

///////////////// samplers //////////////////
texture2D tLastHeightmap : layout(t0);
sampler sLastHeightSampler : layout(s0)
{
	Texture = tLastHeightmap;
	Filter = Point;
	Address = Clamp;
};

texture2D tCurrHeightmap : layout(t1);
sampler sCurrHeightSampler : layout(s1)
{
	Texture = tCurrHeightmap;
	Filter = Linear;
	Address = Clamp;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};


/////////////////// ScreenVS ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

[target("R32F")]
float4 MergeHeightmapPS(v2f IN) : SV_Target
{
	float2 currUV = IN.TC;
	float currDepth = DecodeDepth(tCurrHeightmap.Sample(sCurrHeightSampler, currUV));
	float3 currLerp = float3(currUV.x, currDepth, 1-currUV.y);

	[@_GLSL]
	currLerp = float3(currUV.x, currDepth, currUV.y);
	float3 currPos = lerp(cCurrMin.xyz, cCurrMax.xyz, currLerp);
	if (@_LastMapEnable)
	{
		float2 lastUV = (currPos.xz - cLastMin.xz) / (cLastMax.xz - cLastMin.xz);
		if (lastUV.x >= 0 && lastUV.x <= 1 && lastUV.y >= 0 && lastUV.y <= 1)
		{
			if (!@_GLSL)
				lastUV.y = 1 - lastUV.y;
			float lastDepth = DecodeDepth(tLastHeightmap.Sample(sLastHeightSampler, lastUV));
			//smooth depth
			if (currDepth > 0.99)
			{
				float4 depth;
				depth.x = DecodeDepth(tLastHeightmap.Sample(sLastHeightSampler, lastUV + float2(-cPixelSize, 0))); 
				depth.y = DecodeDepth(tLastHeightmap.Sample(sLastHeightSampler, lastUV + float2( cPixelSize, 0))); 
				depth.z = DecodeDepth(tLastHeightmap.Sample(sLastHeightSampler, lastUV + float2( 0, -cPixelSize))); 
				depth.w = DecodeDepth(tLastHeightmap.Sample(sLastHeightSampler, lastUV + float2( 0,  cPixelSize)));
				float minDepth = min(min(depth.x, depth.y), min(depth.z, depth.w));
				lastDepth = min(minDepth + cMaxStep, lastDepth);
			}
			float lastY = lerp(cLastMin.y, cLastMax.y, lastDepth);
			currPos.y = min(currPos.y, lastY);
		}
	}

	currPos.y += cAccumulation;
	currDepth = saturate((currPos.y - cCurrMin.y) / (cCurrMax.y - cCurrMin.y));
	return EncodeDepth(currDepth);
}

technique MergeHeightmapTech
{
	VertexShader = ScreenVS();
	PixelShader = MergeHeightmapPS();
}

