//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<MotionBlur.fx>
///	@path	~/System/Shader/
///	@date	2015/10/12
///	@brief	Motin Blur.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////


///////////////// const buffer //////////////////
cbuffer Batch
{
	float cBlurPercent 			;
	float2 cBlurCenterUV		;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tBaseMap : layout(t0);
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/////////////////// HDRSampleLumInitial ////////////////////
v2f ScreenVS(a2v IN)
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

#define MOTION_SAMPLES 16
float4 MotionBlurPS(v2f IN) : SV_Target
{
	float2 Center = float2(0.5, 0.5);
	float2 Motiondir = (IN.TC - Center);
	Motiondir = normalize(Motiondir) * dot(Motiondir, Motiondir);
	Motiondir *= 2.0 * cBlurPercent / MOTION_SAMPLES;

	float3 Result = float3(0, 0, 0);
	float2 itrUV = IN.TC;
	for (int i = 0; i < MOTION_SAMPLES; ++i)
	{
		itrUV += Motiondir;
		float3 Source = SampleColorMap(tBaseMap, sBaseSampler, itrUV).rgb;
		Result += Source;
	}

	return float4(Result/MOTION_SAMPLES, 1.0);
}

technique MotionBlurTech
{
	VertexShader = ScreenVS();
	PixelShader = MotionBlurPS();
}
