//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzzhengyi@corp.netease.com
//	Author: Bleston
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<NaviMap.fx>
///	@path	~/System/Shader/
///	@date	2017/04/11
///	@brief	Render navigate map

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 			;
};

cbuffer Shader
{
	float3 cBaseColor				 = { 1, 1, 1 };
	float cAlpha					 = 0.5;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
};


//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
};


//////////////////////////////////Functions////////////////////////////////////////


//////////////////////////////////Shading////////////////////////////////////////

VertexShadingOutput NaviMapVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1.0)).xyz, 1.0);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	//OUT.HPosition = float4(IN.Position.xyz, 1.0);
	//OUT.HPosition.z += OUT.HPosition.w * 0.01;

	return OUT;
}

float4 NaviMapPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);

	OUT.rgb = cBaseColor.xyz;
	OUT.a = EncodeTransparentBloomFactor(cAlpha, 0);

	if (@_TRANSPARENT)
	{
		OUT.rgb = saturate(OUT.rgb) * OUT.a;
	}

	return OUT;
}

technique Shading
{
	VertexShader = NaviMapVertexShading();
	PixelShader = NaviMapPixelShading();
}
