//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<OldMovie.fx>
///	@path	~/System/Shader/
///	@date	2016/05/16
///	@brief	Old Movie effect

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// const buffer //////////////////
cbuffer Batch
{
	float cGrayPercent 			;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tBaseMap : layout(t0);
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/////////////////// HDRSampleLumInitial ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

// 旧化噪点

float4 OldMoviePS(v2f IN) : SV_Target
{
	half3 Source = SampleColorMap(tBaseMap, sBaseSampler, IN.TC.xy).rgb;
	half3 Result = GetLuminance(Source);
	
	// Shift Color
	Result.r += 0.25;
	Result.g += 0.15;
	
	//倍化从4变成5从而扩大边框黑化
	Result *= 1 - abs(0.5*0.8 - IN.TC.x*0.8) * abs(0.5*0.8 - IN.TC.y*0.8 )*5;

	// blend with sample.
	Result = lerp(Source, Result, cGrayPercent);

	return float4(Result, 1.0);
}

technique OldMovieTech
{
	VertexShader = ScreenVS();
	PixelShader = OldMoviePS();
}


