//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Outline.fx>
///	@path	~/System/Shader/
///	@date	2015/07/30
///	@brief	Outline effect.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////

cbuffer Batch
{
	float4 TexelParams0						;
	float4 TexelParams1						;
	float4 OutlineColor						;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D StencilMap : layout(t0);
sampler StencilMap_Sampler : layout(s0)
{
	Texture = StencilMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D BlurredMap : layout(t1);
sampler BlurredMap_Sampler : layout(s1)
{
	Texture = BlurredMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f_outline
{
	float4 HPosition : SV_Position;
	float2 Pos : TEXCOORD0;
	float2 TC  : TEXCOORD1;
};

void MinMax(inout float2 var, float value)
{
	var.x = min(var.x, value);
	var.y = max(var.y, value);
}

v2f_outline OutlineVS( a2v IN )
{
	v2f_outline OUT;
	float3 vPos = IN.Position;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, EPSILON, 1.0);
	OUT.Pos = (OUT.HPosition.xy + 1) * ScreenInfo.xy;
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 OutlinePS(v2f_outline IN) : SV_Target
{
	float4 s0 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy);
	float4 s1 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams0.xy);
	float4 s2 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams0.zw);
	float4 s3 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams1.xy);
	float4 s4 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams1.zw);

	float2 BorderMinMax = float2(2, 0);
	MinMax(BorderMinMax, s0.r);
	MinMax(BorderMinMax, s1.r);
	MinMax(BorderMinMax, s2.r);
	MinMax(BorderMinMax, s3.r);
	MinMax(BorderMinMax, s4.r);

	float2 HiddenMinMax = float2(2, 0);

	MinMax(HiddenMinMax, s0.a);
	MinMax(HiddenMinMax, s1.a);
	MinMax(HiddenMinMax, s2.a);
	MinMax(HiddenMinMax, s3.a);
	MinMax(HiddenMinMax, s4.a);

	// Border.
	if (BorderMinMax.x != BorderMinMax.y)
	{
		float clipValue = min(1, GetRandom(IN.HPosition.xy) - 0.5) * ( 1- HiddenMinMax.y );
		clip(clipValue);
		return float4(OutlineColor.rgb, 1.0);
	}
	else
	{
		clip(1);
		return float4(OutlineColor.rgb, OutlineColor.a);
	}
}

technique Outline
{
	VertexShader = OutlineVS();
	PixelShader = OutlinePS();
}

float4 OutlineMixPS(v2f_outline IN) : SV_Target
{
	float4 s0 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy);
	float4 s1 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams0.xy);
	float4 s2 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams0.zw);
	float4 s3 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams1.xy);
	float4 s4 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy + TexelParams1.zw);

	float2 BorderMinMax = float2(2, 0);

	MinMax(BorderMinMax, s0.r);
	MinMax(BorderMinMax, s1.r);
	MinMax(BorderMinMax, s2.r);
	MinMax(BorderMinMax, s3.r);
	MinMax(BorderMinMax, s4.r);

	float2 HiddenMinMax = float2(2, 0);

	MinMax(HiddenMinMax, s0.a);
	MinMax(HiddenMinMax, s1.a);
	MinMax(HiddenMinMax, s2.a);
	MinMax(HiddenMinMax, s3.a);
	MinMax(HiddenMinMax, s4.a);

	// Border.
	if (BorderMinMax.x != BorderMinMax.y)
	{
		float clipValue = min(1, GetRandom(IN.HPosition.xy) - 0.5) * ( 1- HiddenMinMax.y );
		clip(clipValue);

		return float4(OutlineColor.rgb, 1.0);
	}
	else
	{
		// Outer.
		MinMax(BorderMinMax, s0.r);
		float clipValue = min(1, GetRandom(IN.HPosition.xy) - (1 - (0.5 * OutlineColor.a))) * ( 1- HiddenMinMax.y );

		clip(BorderMinMax.y == 0 ? -1.0 : 1.0);

		// clip(clipValue);
		float2 lineps = ceil(frac(IN.TC.xy*100-GameTime)-0.7) ; 
		// Inner.
		return OutlineColor.rgba+lineps.y*0.2;
	}
}

technique OutlineMix
{
	VertexShader = OutlineVS();
	PixelShader = OutlinePS();
}

float4 OutlineBlurredPS(v2f_outline IN) : SV_Target
{
	float4 s0 = StencilMap.Sample(StencilMap_Sampler, IN.TC.xy);
	float4 s1 = BlurredMap.Sample(BlurredMap_Sampler, IN.TC.xy) * 2;
	s1.a *= saturate(1 - 16 * s0.a);
	return s1;
}

technique OutlineBlurred
{
	VertexShader = OutlineVS();
	PixelShader = OutlineBlurredPS();
};
