//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzzhengyi@corp.netease.com
//	Author: Bleston
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<PhysicsCollision.fx>
///	@path	~/System/Shader/
///	@date	2019/02/27
///	@brief	Render physics collision

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 			;
};

cbuffer Shader
{
	float3 cBaseColor				 = { 1, 1, 1 };
	float cAlpha					 = 0.5;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate
};


//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
};


//////////////////////////////////Functions////////////////////////////////////////


//////////////////////////////////Shading////////////////////////////////////////

VertexShadingOutput PhysicsCollisionVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1.0)).xyz, 1.0);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.HPosition.z = OUT.HPosition.z * 1.01;
	OUT.TexCoord = float4(0, 0, 0, 0);

	return OUT;
}

float4 PhysicsCollisionPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);

	OUT.rgb = cBaseColor.xyz;
	OUT.a = EncodeTransparentBloomFactor(cAlpha, 0);

	if (@_TRANSPARENT)
	{
		OUT.rgb = saturate(OUT.rgb) * OUT.a;
	}

	return OUT;
}

[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].TexCoord = float4(1, 0, 0, 0);
	OUT.Append(IN[0]);
	IN[1].TexCoord = float4(0, 1, 0, 0);
	OUT.Append(IN[1]);
	IN[2].TexCoord = float4(0, 0, 1, 0);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 PhysicsCollisionWireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.TexCoord.xyz);
	float3 distance = IN.TexCoord.xyz / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(saturate(cBaseColor.xyz) * a, EncodeTransparentBloomFactor(a, 0));
}

technique Shading
{
	VertexShader = PhysicsCollisionVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = PhysicsCollisionWireframePixelShading();
}
