//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<SSAO.fx>
///	@path	~/System/Shader/
///	@date	2018/05/28
///	@brief	SSAO.
//

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////
@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};

@_IsFirstFrame
{
	Name = IsFirstFrame;
	Default = true;
	Visible = false;
};

@_ISRGBA8
{
	Name = ISRGBA8;
	Default = true;
	Visible = false;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX				;
	float4 vWBasisY				;
	float4 vWBasisZ				;
	float4x4 lastViewProjTex	;
	float4 cRTSize				;			//xy-ao rt size, zw-pingpong rt size
	float2 cPixelSize			;		//1/ao rt size
	float cLastFrameRateStatic	 = 0.9;
	float cLastFrameRateDynamic	 = 0.8;
	float AORangeStatic			 = 1.0;
	float AOScaleStatic			 = 1;
	float AONearStatic			 = 0.01;
	float AOFarStatic			 = 0.5;
	float AORangeDynamic		 = 1.0;
	float AOScaleDynamic		 = 1;
	float AONearDynamic			 = 0.01;
	float AOFarDynamic			 = 0.5;
	float2 cClipToView			 = {0, 0};//tan(0.5*fovy)*w/h, tan(0.5*fovy)
	float2 cViewToClip 			 = {0, 0};//cot(0.5*fovy)*h/w, cot(0.5*fovy)
	float AOJitter				 = 1;
	float AOBase				 = 0.2;
	float AOBiasStatic			 = 0.02;
	float AOBiasDynamic			 = 0.02;
};

///////////////// samplers //////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tDepthBufferLinear : SceneDepth : layout(t1);
sampler sDepthBufferSamplerLinear : layout(s1)
{
	Texture = tDepthBufferLinear;
	Filter = Linear;
	Address = Clamp;
};

texture2D tRandom3D : layout(t2);
sampler sRandom3D : layout(s2)
{
	Texture = tRandom3D;
	Filter = Point;
	Address = Wrap;
};

texture2D tLastAOMap : layout(t3);
sampler sLastAOSampler : layout(s3)
{
	Texture = tLastAOMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tCurrAOMap : layout(t4);
sampler sCurrAOSampler : layout(s4)
{
	Texture = tCurrAOMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tVBuffer : VelocityMap : layout(t5);
sampler sVBufferSampler : layout(s5)
{
	Texture = tVBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tGBufferNormal : GBufferMapB : layout(t6);
sampler sGBufferNormal : layout(s6)
{
	Texture = tGBufferNormal;
	Filter = Linear;
	Address = Clamp;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////////////////functions///////////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}
float3 uvToViewPosWithDepth(float2 uv, float depth)
{
	float2 vPosInClip = uv * 2 - 1;
	[!@_GLSL]
	vPosInClip.y = -vPosInClip.y;
	float3 viewPos;
	viewPos.z = depth * CameraInfo.y;
	viewPos.xy = vPosInClip * cClipToView * viewPos.z;
	return viewPos;
}

float3 uvToViewPosLinear(float2 uv)
{
	float depth = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, uv));
	return uvToViewPosWithDepth(uv, depth);
}

float4 EncodeAODepth(float ao, float depth)
{
	// if(@_GLSL)
	// 	return float4(FloatToR8G8(ao), FloatToR8G8(depth));
	// else
	// 	return float4(ao, depth, 1, 1);
	return float4(ao, depth, 1, 1);
}

float2 DecodeAODepth(float4 raw)
{
	// if(@_GLSL)
	// 	return float2(R8G8ToFloat(raw.xy), R8G8ToFloat(raw.zw));
	// else
	return raw.xy;
}

float2 HDecodeAODepth(float4 raw)
{
	if (@_ISRGBA8 || @_GLSL)
		return float2(R8G8ToFloat(raw.xy), R8G8ToFloat(raw.zw));
	else
		return raw.xy;
}

float3 rebuildNormalFromDepth(float3 origPos, float2 uv)
{
	float3 posXN = uvToViewPosLinear(uv + float2(-ScreenInfo.z, 0));		//-x 
	float3 posX = uvToViewPosLinear(uv + float2(ScreenInfo.z, 0));		//+x

	float3 posYN = uvToViewPosLinear(uv + float2(0, ScreenInfo.w));		//-y
	float3 posY = uvToViewPosLinear(uv + float2(0, -ScreenInfo.w));		//+y
	if (@_GLSL)
	{
		float3 temp = posYN;
		posYN = posY;
		posY = temp;
	}

	float3 _ddx1 = posX-origPos;
	float3 _ddx2 = origPos - posXN;
	float3 _ddx = lerp(_ddx1, _ddx2, step(abs(_ddx2.z),abs(_ddx1.z)));

	float3 _ddy1 = posY-origPos;
	float3 _ddy2 = origPos - posYN;
	float3 _ddy = lerp(_ddy1, _ddy2, step(abs(_ddy2.z),abs(_ddy1.z)));

	float3 N = -normalize(cross(_ddx, _ddy));
	return N;
}

/////////////////// technique CalcAO ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.HPosition.z = OUT.HPosition.w;

	OUT.TC = IN.BaseTC;
	return OUT;
}

#define SSAO_SAMPS 8
float4 CalcAOPS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 tc = spos / cRTSize.xy;
	float depth = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, tc));
	float3 worldPos = CalcHomogeneousPos(depth, tc);
	worldPos += CameraPos.xyz;

	//origPos
	float2 uv = tc;
	float3 origPos = uvToViewPosWithDepth(uv, depth);
	float3 N = rebuildNormalFromDepth(origPos, uv);
	//if (@_DEFERRED_SHADING)????
	//	N = normalize(DecodeNormal(tGBufferNormal.Sample(sGBufferNormal, uv).xyz));

	//select ao parameters
	float AORange = AORangeStatic;
	float AONear = AONearStatic;
	float AOFar = AOFarStatic;
	float AOScale = AOScaleStatic;
	float AOBias = AOBiasStatic;
	if (@_VelocityMapEnable)
	{
		float2 motionVector = tVBuffer.SampleLevel(sVBufferSampler, tc, 0).xy;
		if (motionVector.x != 0 || motionVector.y != 0)
		{
			AORange = AORangeDynamic;
			AONear = AONearDynamic;
			AOFar = AOFarDynamic;
			AOScale = AOScaleDynamic;
			AOBias = AOBiasDynamic;
		}
	}
	
	//jitterDir
	float2 jitterUV = frac(float2(worldPos.z+worldPos.y, worldPos.x+worldPos.y)*1000) + float2(AOJitter, 0);
	float3 jitterDir = tRandom3D.Sample(sRandom3D, jitterUV).xyz;

	//calcAO
	static const float3 sample_sphere[16] = {
		float3( 0.5381, 0.1856,-0.4319), float3( 0.1379, 0.2486, 0.4430),
		float3( 0.3371, 0.5679,-0.0057), float3(-0.6999,-0.0451,-0.0019),
		float3( 0.0689,-0.1598,-0.8547), float3( 0.0560, 0.0069,-0.1843),
		float3(-0.0146, 0.1402, 0.0762), float3( 0.0100,-0.1924,-0.0344),
		float3(-0.3577,-0.5301,-0.4358), float3(-0.3169, 0.1063, 0.0158),
		float3( 0.0103,-0.5869, 0.0046), float3(-0.0897,-0.4940, 0.3287),
		float3( 0.7119,-0.0154,-0.0918), float3(-0.0533, 0.0596,-0.5411),
		float3( 0.0352,-0.0631, 0.5460), float3(-0.4776, 0.2847,-0.0271)

		// float3(0.07702117127471324, 0.1156958976381993, -0.03177204250704574),
		// float3(-0.21237609978985772, 0.06215192874276373, 0.1515485556249499),
		// float3(-0.02838386911487637, 0.23060063859350782, -0.09014794258434064),
		// float3(-0.32956354015103634, 0.5865088323431984, -0.42363226996153763),
		// float3(-0.10862027253205746, 0.038085763112080014, -0.017649267597608286),
		// float3(-0.29674128423203233, 0.25647650817097284, 0.49427388987163345),
		// float3(-0.1654924749378402, 0.466877819293457, 0.3636100071836359),
		// float3(0.27692445362602053, 0.655713532752049, -0.44266086462935456)
	};
	float occlusion = 0.0;
	float range =  AORange;
	int nSamples = (int)(SSAO_SAMPS * (1 - depth));
	for(int i=0; i < nSamples; i+=1)
	{
		float3 ray = reflect(sample_sphere[i] * range, jitterDir);
		ray *= sign(dot(ray,N));
		float3 samplePos = origPos + ray;
		float2 sampleUV; 
		if (@_GLSL)
			sampleUV = (samplePos.xy * cViewToClip / samplePos.z) * 0.5 + 0.5;
		else
			sampleUV = (samplePos.xy * cViewToClip / samplePos.z) * float2(0.5, -0.5) + float2(0.5, 0.5);
		float occluZ = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, sampleUV)) * CameraInfo.y;
		float isOcclu = step(occluZ+AOBias, samplePos.z);
		float diff = abs(origPos.z - occluZ);
		float t = isOcclu * (1.0-smoothstep(AONear, AOFar, diff));
		occlusion += t;
	}
	float ao = saturate(AOScale * occlusion / (float)nSamples);
	if (@_ISRGBA8)
		return float4(FloatToR8G8(ao), FloatToR8G8(depth));
	else
		return EncodeAODepth(ao, depth);
}

technique CalcAO
{
	VertexShader = ScreenVS();
	PixelShader = CalcAOPS();
}

/////////////////// technique TSSAO //////////////////////
float4 TSSAOPS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 tc = spos / cRTSize.xy;
	float2 currAODepth = HDecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, tc));
	float currAO = currAODepth.x;
	float depth = currAODepth.y;
	if (@_ISRGBA8)
		depth = DecodeDepth(tDepthBufferLinear.Sample(sDepthBufferSamplerLinear, tc));
	float3 worldPos = CalcHomogeneousPos(depth, tc);
	worldPos += CameraPos.xyz;
	float2 uv = tc;

	//motion vector
	float4 lastPos = MulMat4x4(lastViewProjTex, float4(worldPos,1));
	float2 lastUV = lastPos.xy / lastPos.w;
	float lastFrameRate = cLastFrameRateStatic;

	if (@_VelocityMapEnable)
	{
		float2 motionVector = tVBuffer.SampleLevel(sVBufferSampler, tc, 0).xy;
		if (motionVector.x != 0 || motionVector.y != 0)
		{
			lastUV = tc + motionVector;
			lastFrameRate = cLastFrameRateDynamic;
		}
	}
	
	//lerp
	float InLastFrameScreen = lastUV.x >= 0 && lastUV.x <= 1 && lastUV.y >= 0 && lastUV.y <= 1;
	if (@_IsFirstFrame)
		InLastFrameScreen = 0;
	float2 lastAODepth = DecodeAODepth(tLastAOMap.Sample(sLastAOSampler, lastUV));
	InLastFrameScreen *= step(abs(1- lastPos.w/(lastAODepth.y*CameraInfo.y)), 0.005);
	uv += float2(0.5, 0.5) * cPixelSize;
	float ao0 = HDecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, uv + float2(-cPixelSize.x, -cPixelSize.y))).x;
	float ao1 = HDecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, uv + float2(-cPixelSize.x,  cPixelSize.y))).x;
	float ao2 = HDecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, uv + float2( cPixelSize.x,  cPixelSize.y))).x;
	float ao3 = HDecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, uv + float2( cPixelSize.x,  cPixelSize.y))).x;
	float averAO = (currAO + ao0 + ao1 + ao2 + ao3) / 5.0;
	float minAO = averAO * 0.25;//min(currAO, min(min(ao0, ao1), min(ao2, ao3)));
	float maxAO = averAO * 1.5;//max(currAO, max(max(ao0, ao1), max(ao2, ao3))) * 1.5;
	float ao = lerp(currAO, clamp(lastAODepth.x, minAO, maxAO), InLastFrameScreen*lastFrameRate);
	return EncodeAODepth(ao, depth);
}

technique TSSAO
{
	VertexShader = ScreenVS();
	PixelShader = TSSAOPS();
}

/////////////////// technique MergeAO ////////////////////
float4 MergeAOPS(v2f IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	//get border Pixels
	stc += float2(0.5, 0.5) * cPixelSize;
	float ao0 = DecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, stc + float2(-cPixelSize.x, -cPixelSize.y))).x;
	float ao1 = DecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, stc + float2(-cPixelSize.x,  cPixelSize.y))).x;
	float ao2 = DecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, stc + float2( cPixelSize.x,  cPixelSize.y))).x;
	float ao3 = DecodeAODepth(tCurrAOMap.Sample(sCurrAOSampler, stc + float2( cPixelSize.x,  cPixelSize.y))).x;
	float ao = (2*ao0 + ao1 + ao2 + ao3) / 5.0;
	// float ao  = DecodeDepth(tCurrAOMap.Sample(sCurrAOSampler, stc));
	ao = max(0, ao - AOBase);
	//return float4(1-ao, 1-ao, 1-ao, 1);
	return float4(0, 0, 0, ao);
}

technique MergeAO
{
	VertexShader = ScreenVS();
	PixelShader = MergeAOPS();
}






