//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Deferred.fx>
///	@path	~/System/Shader/
///	@date	2019/08/29
///	@brief	Deferred rendering.

#pragma shader draw
#define IS_DEFERRED_LIGHTTING 1

#include "Constants.fxh"
#include "Common.fxh"
#include "BRDF.fxh"

#define HIGH_QUALITY
#ifdef HIGH_QUALITY
	#define MAX_STEPS		32
	#define STRIDE			24
	#define SSR_THICKNESS 	0.25 
	#define STOP_STRIDE_LOD	1
#else
	#define MAX_STEPS		16
	#define STRIDE			48
	#define SSR_THICKNESS	0.5
	#define STOP_STRIDE_LOD	3
#endif
//////////////////////////////////Properties////////////////////////////////////////
@_BlurX
{
	Name = BlurX;
	Default = false;
	Visible = false;
}
@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};
@_EnableSSGI
{
	Name = EnableSSGI;
	Description = "Enable SSGI";  
	Default = false;
};
///////////////// const buffer ////////////////// 
cbuffer Batch
{
	float4x4 currViewToLastTex		;
	float4x3 View 		: View		;
	float4 cWBasisX		: WBasisX	;
	float4 cWBasisY		: WBasisY	;
	float4 cWBasisZ		: WBasisZ	;
	float4x4 projectToPixelMatrix	;
	float cCtanHalfFov				;	//proj z
	float2 cDepthMapSize			;
	float cSSRBRDFBias				;
	float2 cTraceMapPixelSize		;
	float2 cPixelSize				;
	float cRandomOffset 			;
	float cResolveRange				;
	float cBilaterialRange			;
};

texture2D tGBufferB : GBufferMapB : layout(t0);
sampler sGBufferB : layout(s0)
{
	Texture = tGBufferB;
	Filter = Point;
	Address = Clamp;
};

texture2D tGBufferC : GBufferMapC : layout(t1);
sampler sGBufferC : layout(s1)
{
	Texture = tGBufferC;
	Filter = Point;
	Address = Clamp;
};

texture2D tGBufferE : GBufferMapE : layout(t2);
sampler sGBufferE : layout(s2)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};

texture2D tDepthMap : SceneDepth : layout(t3);
sampler sDepthSampler : layout(s3)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tEnvMap : EnvMap : layout(t4) = "White";
sampler sEnvSampler : layout(s4)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tLastFrameScreen : layout(t5) = "White";
sampler sLastFrameScreen : layout(s5)
{
	Texture = tLastFrameScreen;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tUniformRandom : layout(t6) = "White";
sampler sUniformRandom : layout(s6)
{
	Texture = tUniformRandom;
	Filter = Point;
	Address = Wrap;
	SRGB = false; 
};

texture2D tRandom2D : layout(t6);
sampler sRandom2D : layout(s6)
{
	Texture = tRandom2D;
	Filter = Point;
	Address = Wrap;
};

texture2D tTraceSamplesMap : layout(t7) = "White";
sampler sTraceSamplesSampler : layout(s7)
{
	Texture = tTraceSamplesMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false; 
};

texture2D tGBufferF : GBufferMapF : layout(t8);
sampler sGBufferF : layout(s8)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};

texture2D tCurrRT : layout(t0) = "White";
sampler sCurrRT : layout(s0)
{
	Texture = tCurrRT;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tHistoryRT : layout(t1) = "White";
sampler sHistoryRT : layout(s1)
{
	Texture = tHistoryRT;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tVBuffer : VelocityMap : layout(t4);
sampler sVBufferSampler : layout(s4)
{
	Texture = tVBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D tGIAO : layout(t9);
sampler sGIAO : layout(s9)
{
	Texture = tGIAO;
	Filter = Linear;
	Address = Clamp;
};
////////////////////////////////////Lighting Function////////////////////////////////////
#include "Lighting.fxh"
#include "ShadingModel.fxh"

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/// light volume vertex to vs
struct VertexAssembleLightVolume
{
	float3		Position		: POSITION;
};

// vertex output to pixel shader.
struct VertexOutputLightVolume
{
	float4		HPosition		: SV_Position;
};

///////////////// helper functions //////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}

float GetCameraZ(float2 uv)
{
	return -DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, uv, 0)) * CameraInfo.y;
}

float3 UVToViewPos(float2 uv, float cameraZ)
{
	[!@_GLSL]
	uv.y = 1.0 - uv.y;

	float3 viewPos = float3(uv*2-1, -cCtanHalfFov) * (cameraZ/(-cCtanHalfFov));
	viewPos.x *= ScreenInfo.x / ScreenInfo.y;
	return viewPos;
}

float SSR_BRDF(float3 V, float3 L, float3 N, float Roughness)
{
	float3 H = normalize(L + V);
	float NoH = max(dot(N, H), 0);
	float D = D_GGX(max(MIN_ROUGHNESS, Roughness), NoH);
	float G = G_Implicit();
	return max(0, D * G);
}

float SSR_MissBRDF(float Roughness)
{
	float D = D_GGX(max(MIN_ROUGHNESS, Roughness), 1);
	float G = G_Implicit();
	return max(0, D * G);
}

float GetScreenFadeBord(float2 pos, float value)
{
    float borderDist = min(1 - max(pos.x, pos.y), min(pos.x, pos.y));
    return saturate(borderDist / value);
}

/**
    \param csOrigin Camera-space ray origin, which must be 
    within the view volume and must have z < -0.01 and project within the valid screen rectangle
    \param csDirection Unit length camera-space ray direction
    \param projectToPixelMatrix A projection matrix that maps to pixel coordinates (not [-1, +1] normalized device coordinates)
    \param csZThickness Camera space thickness to ascribe to each pixel in the depth buffer
    \param nearPlaneZ Negative number
    \param stride Step in horizontal or vertical pixels between samples. This is a float
     because integer math is slow on GPUs, but should be set to an integer >= 1
    \param jitterFraction  Number between 0 and 1 for how far to bump the ray in stride units
      to conceal banding artifacts
    \param maxSteps Maximum number of iterations. Higher gives better images but may be slow
    \param maxRayTraceDistance Maximum camera-space distance to trace before returning a miss
    \param hitPixel Pixel coordinates of the first intersection with the scene
 */
bool traceScreenSpaceRay1(
	float3          csOrigin, // ViewPos
    float3         	csDirection, // R
    float           csZThickness,
    float           nearPlaneZ,
    float			stride,
    float           jitterFraction,
    float           maxSteps,
    in float        maxRayTraceDistance,
    out float2      hitPixel,
	out float		hitZ) 
{
    
    // Clip ray to a near plane in 3D (doesn't have to be *the* near plane, although that would be a good idea)
    float rayLength = ((csOrigin.z + csDirection.z * maxRayTraceDistance) > nearPlaneZ) ?
                        (nearPlaneZ - csOrigin.z) / csDirection.z :
                        maxRayTraceDistance;
	float3 csEndPoint = csDirection * rayLength + csOrigin;

    // Project into screen space
    float4 H0 = mul(float4(csOrigin, 1.0), projectToPixelMatrix);
    float4 H1 = mul(float4(csEndPoint, 1.0), projectToPixelMatrix);

    // There are a lot of divisions by w that can be turned into multiplications
    // at some minor precision loss...and we need to interpolate these 1/w values
    // anyway.
    //
    // Because the caller was required to clip to the near plane,
    // this homogeneous division (projecting from 4D to 2D) is guaranteed 
    // to succeed. 
    float k0 = 1.0 / H0.w;
    float k1 = 1.0 / H1.w;

    // Switch the original points to values that interpolate linearly in 2D
    float3 Q0 = csOrigin * k0; 
    float3 Q1 = csEndPoint * k1;

	// Screen-space endpoints
    float2 P0 = H0.xy * k0;
    float2 P1 = H1.xy * k1;

    // [Optional clipping to frustum sides here]
#if 1
    float yMax = cDepthMapSize.y - 0.5;
    float yMin = 0.5;
    float xMax = cDepthMapSize.x - 0.5;
    float xMin = 0.5;
    float alpha = 0;

    if (P1.y > yMax || P1.y < yMin)
    {
        float yClip = (P1.y > yMax) ? yMax : yMin;
        float yAlpha = (P1.y - yClip) / (P1.y - P0.y);
        alpha = yAlpha;
    }
    if (P1.x > xMax || P1.x < xMin)
    {
        float xClip = (P1.x > xMax) ? xMax : xMin;
        float xAlpha = (P1.x - xClip) / (P1.x - P0.x);
        alpha = max(alpha, xAlpha);
    }

    P1 = lerp(P1, P0, alpha);
    k1 = lerp(k1, k0, alpha);
    Q1 = lerp(Q1, Q0, alpha);
#endif

    // Initialize to off screen
    hitPixel = float2(-1.0, -1.0);

    // If the line is degenerate, make it cover at least one pixel
    // to avoid handling zero-pixel extent as a special case later
    P1 += (distanceSquared(P0, P1) < 0.0001) ? float2(0.01,0.01) : float2(0,0);

    float2 delta = P1 - P0;

    // Permute so that the primary iteration is in x to reduce
    // large branches later
    bool permute = false;
	if (abs(delta.x) < abs(delta.y)) {
		// More-vertical line. Create a permutation that swaps x and y in the output
		permute = true;

        // Directly swizzle the inputs
		delta = delta.yx;
		P1 = P1.yx;
		P0 = P0.yx;
	}
    
	// From now on, "x" is the primary iteration direction and "y" is the secondary one
    float stepDirection = sign(delta.x);
    float invdx = stepDirection / delta.x;
    float2 dP = float2(stepDirection, invdx * delta.y);

    // Track the derivatives of Q and k
    float3 dQ = (Q1 - Q0) * invdx;
    float  dk = (k1 - k0) * invdx;

    // Scale derivatives by the desired pixel stride
	dP *= stride; dQ *= stride; dk *= stride;

    // Offset the starting values by the jitter fraction
	P0 += dP * jitterFraction; Q0 += dQ * jitterFraction; k0 += dk * jitterFraction;

	// Slide P from P0 to P1, (now-homogeneous) Q from Q0 to Q1, and k from k0 to k1
    float3 Q = Q0;
    float  k = k0;

	// We track the ray depth at +/- 1/2 pixel to treat pixels as clip-space solid 
	// voxels. Because the depth at -1/2 for a given pixel will be the same as at 
	// +1/2 for the previous iteration, we actually only have to compute one value 
	// per iteration.
	float prevZMaxEstimate = csOrigin.z;
    float stepCount = 0.0;
    float rayZMax = prevZMaxEstimate, rayZMin = prevZMaxEstimate;
    hitZ = rayZMax + 1e4;

    // P1.x is never modified after this point, so pre-scale it by 
    // the step direction for a signed comparison
    float end = P1.x * stepDirection;

    // We only advance the z field of Q in the inner loop, since
    // Q.xy is never used until after the loop terminates.
	int strideLod = 0;
	float thickness = csZThickness * min(4,exp2(STOP_STRIDE_LOD - strideLod));
	//[unroll]
	for (float2 P = P0; ((P.x * stepDirection) <= end) && (stepCount < maxSteps) &&
        (hitZ != 0.0);
        stepCount += 1.0) 
	{
		hitPixel = permute ? P.yx : P;

        // The depth range that the ray covers within this loop
        // iteration.  Assume that the ray is moving in increasing z
        // and swap if backwards.  Because one end of the interval is
        // shared between adjacent iterations, we track the previous
        // value and then swap as needed to ensure correct ordering
        rayZMin = prevZMaxEstimate;

        // Compute the value at 1/2 pixel into the future
        rayZMax = (dQ.z * 0.5 + Q.z) / (dk * 0.5 + k);
		float _prevZMaxEstimate = rayZMax;
        if (rayZMin > rayZMax) 
		{
			float temp = rayZMin;
			rayZMin = rayZMax; 
			rayZMax = temp;
		}

		// Camera-space z of the background
		hitZ = GetCameraZ(hitPixel / cDepthMapSize.xy);
		if ((rayZMax >= hitZ - csZThickness) && (rayZMin <= hitZ))
		{
			return true;
		}
		else if ((rayZMax >= hitZ - thickness) && (rayZMin <= hitZ))
		{
			if (strideLod < STOP_STRIDE_LOD)
			{
				++strideLod;
				thickness = max(csZThickness, thickness*0.5);
				dP *= 0.5; dQ *= 0.5; k *= 0.5;
			}
		}
		else
		{
			prevZMaxEstimate = _prevZMaxEstimate;
			P += dP; Q.z += dQ.z; k += dk;
		}
    } // pixel on ray
    return false;
}

///////////////////////////////////////// Trace Pass /////////////////////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

//rg-hit uv, b-hit depth, a-pdf
float4 TraceBRDFPS(v2f IN) : SV_Target
{	
	//1, Gen Ray Data
	float2 stc = IN.TC;
	float cameraZ = GetCameraZ(stc);
	if (-cameraZ > 0.8*CameraInfo.y)//todo: only trace nearby pixels
		return float4(-1,-1, 0, -1);
	float3 ViewPos = UVToViewPos(stc, cameraZ);
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Sample(sGBufferB, stc), 
		float4(1,1,1,1),
		float4(1,1,1,1),
		tGBufferE.Sample(sGBufferE, stc),
		tGBufferF.Sample(sGBufferF, stc));
	if(GBuffer.Roughness > 0.2)
	{
		return float4(-1,-1, 0, -1);
	}
	float3 ViewNormal = normalize(mul(GBuffer.WorldNormal, (float3x3)View));

	//2, Monte Carlo Raytrace
	float2 E = tUniformRandom.Sample(sUniformRandom, (stc)*(ScreenInfo.xy/128)).xy;
	E.y = lerp(E.y, 0.0, cSSRBRDFBias);
	float3 H = ImportanceSampleGGX(E, GBuffer.Roughness, ViewNormal);
	float3 NV = normalize(ViewPos);
	float3 R = reflect(NV,H.xyz);
	if (dot(R,ViewNormal) < 0)//under surface rays
		R = reflect(NV, ViewNormal);
	ViewPos += NV * -0.1;
	ViewPos += R * 0.1;

	float2 hitPixel; float hitZ; 
	float pdf = 1;//H.w;

	float stride = lerp(0.125*STRIDE, STRIDE, saturate(1+cameraZ/100));
	bool isHit = traceScreenSpaceRay1(
		ViewPos, 
		R, 
		SSR_THICKNESS,				//csZThickness
		-CameraInfo.x, 				//nearPlaneZ
		stride,						//stride
		0.75*(E.x+E.y),				//frac((spos.x+spos.y)*0.5), 	//jitterFraction
		MAX_STEPS,
		500, 						//maxRayTraceDistance
		hitPixel, 
		hitZ);

	if (isHit)
	{
		return float4(hitPixel/cDepthMapSize.xy, HalfToR8G8(-hitZ*CameraInfo.z));
	}
	else
	{
		float3 farEnd = R * 100000 + ViewPos;
		float4 farEndClip = mul(float4(farEnd, 1.0), projectToPixelMatrix);
		farEndClip.xy /= farEndClip.w;
		if(farEnd.z < -CameraInfo.x && farEndClip.x > 0 && farEndClip.x < cDepthMapSize.x && farEndClip.y > 0 && farEndClip.y < cDepthMapSize.y)
		{
			return float4(farEndClip.xy/cDepthMapSize.xy, 1, 0);
		}
		else
		{
			return float4(hitPixel/cDepthMapSize.xy, 0, 0);
		}
	}
}

technique TraceBRDFPass
{
	VertexShader = ScreenVS();
	PixelShader = TraceBRDFPS();
}

///////////////////////////////////////// Resolve Pass //////////////////////////////////////
float4 ResolvePS(v2f IN) : SV_Target
{
	//1, gbuffer data
	float2 stc = IN.TC;
	float cameraZ = GetCameraZ(stc);
	float3 ViewPos = UVToViewPos(stc, cameraZ);
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Sample(sGBufferB, stc), 
		tGBufferC.Sample(sGBufferC, stc),
		float4(1,1,1,1),
		tGBufferE.Sample(sGBufferE, stc),
		tGBufferF.Sample(sGBufferF, stc));
	float3 ViewNormal = normalize(mul(GBuffer.WorldNormal, (float3x3)View));
	float3 V = normalize(-ViewPos);
	float NoV = saturate(dot(ViewNormal, V));
	float ao = 1;
	if(@_EnableSSGI)
	{	
		ao = tGIAO.Sample(sGIAO, stc).a;
	}
	//2,EnvColor 
	float3 EnvColor = float3(0,0,0);
	float sceneZ = cameraZ * -CameraInfo.z;
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;
	{
		float3 _V = normalize(CameraPos.xyz - WorldPos);
		float3 R = reflect(-_V, GBuffer.WorldNormal);
		float MIP_ROUGHNESS = 0.17;
		float level = GBuffer.Roughness / MIP_ROUGHNESS;
		float fSign = R.z > 0;
		float fSign2 = fSign*2 - 1;
		R.xy /= R.z*fSign2 + 1;
		R.xy = R.xy * float2(0.25, -0.25) + (0.25+0.5*fSign);
		{
			float4 srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy, level);
			EnvColor = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) * ao * ReflectionScale;
		}
	}
	if(GBuffer.Roughness > 0.2)
	{
		return float4(EnvColor * EnvBRDFApprox(GBuffer.SpecularColor, GBuffer.Roughness, NoV), -1);
	}
	//2, resolve
	#define RESOLVE_SAMPLES 8
	float3 ReflectionColor = float3(0, 0, 0);
	float SumWeight = EPSILON;
	float HitZ = 0;
	static const float2 irreg_kernel_2d[8] = {
		float2(-0.613392,0.617481),
		float2(0.553514,-0.832698),
		float2(0.806146,0.578845),
		float2(-0.724784,-0.673879),
		float2(0.037202,-0.049409), 
		float2(0.099155,0.986755),
		float2(0.975463,-0.180090),
		float2(-0.989929,-0.006012)
	};
	float2 jitterUV = (stc + cRandomOffset) * ScreenInfo.xy / 128;
	//jitterUV = float2(WorldPos.x+WorldPos.y*2, WorldPos.z+WorldPos.y);
	float2 BlueNoise = tRandom2D.Sample(sRandom2D, (jitterUV)).xy * 2 - 1;
	float ResolveRange = min(cResolveRange, lerp(2, cResolveRange*2, NoV * GBuffer.Roughness)); 
	BlueNoise *= cTraceMapPixelSize * ResolveRange;
	float2x2 OffsetRotationMatrix = float2x2(BlueNoise.x, BlueNoise.y, -BlueNoise.y, -BlueNoise.x);
	float MissWeight = SSR_MissBRDF(GBuffer.Roughness);
	for (int i = 0; i < RESOLVE_SAMPLES; ++i)
	{
		float2 Offset_UV = mul(irreg_kernel_2d[i], OffsetRotationMatrix);
		float4 HitUVZ = tTraceSamplesMap.Sample(sTraceSamplesSampler, stc + Offset_UV);
		HitUVZ.z = R8G8ToHalf(HitUVZ.zw);
		float3 hitColor = EnvColor;
		float Weight = MissWeight;
		if (HitUVZ.z > 0)
		{
			HitUVZ.z *= -CameraInfo.y;
			float3 HitViewPos = UVToViewPos(HitUVZ.xy, HitUVZ.z);
			float3 L = (HitViewPos - ViewPos);
			float TraceDistance = sqrt(dot(L, L));
			L /= TraceDistance;
			Weight = SSR_BRDF(V, L, ViewNormal, GBuffer.Roughness);// / max(1e-5, HitUVZ.a);
			float2 HitUV = HitUVZ.xy;
			float4 lastPosH = mul(float4(HitViewPos,1), currViewToLastTex);
			HitUV = lastPosH.xy / lastPosH.w;
			hitColor = lerp(EnvColor, tLastFrameScreen.SampleLevel(sLastFrameScreen, HitUV, 0).rgb,
				GetScreenFadeBord(HitUV, 0.00225));
		}
		else
		{
			HitUVZ.z = -CameraInfo.y; 
		}
		ReflectionColor.rgb += hitColor  * Weight;
		SumWeight += Weight;
		HitZ += HitUVZ.z;
	}
	ReflectionColor /= SumWeight;
	HitZ /= RESOLVE_SAMPLES;

	return float4(EnvBRDFApprox(GBuffer.SpecularColor, GBuffer.Roughness, NoV) * ReflectionColor, HitZ);
}

technique ResolvePass
{
	VertexShader = ScreenVS();
	PixelShader = ResolvePS();
}

///////////////////////////////////////// Temporal Filter Pass //////////////////////////////////////
float4 TemporalFilterPS(v2f IN) : SV_Target
{
	//Last UV
	float2 stc = IN.TC;
	float cameraZ = GetCameraZ(stc);
	float3 ViewPos = UVToViewPos(stc, cameraZ);
	float4 lastPosH = mul(float4(ViewPos,1), currViewToLastTex);
	float2 lastUV = lastPosH.xy / lastPosH.w;
	float currPosInLastZ = -lastPosH.w;
	float2 Depth_Velocity = lastUV - stc;
	float3 WorldNormal 	= DecodeNormal(tGBufferE.Sample(sGBufferE, stc).xyz);
	WorldNormal 		= normalize(WorldNormal);
	float4 c4 = tCurrRT.Sample(sCurrRT, stc);
	{
		float3 ViewPos2 = UVToViewPos(stc, c4.w);
		float4 lastPosH2 = mul(float4(ViewPos2,1), currViewToLastTex);
		float2 lastUV2 = lastPosH2.xy / lastPosH2.w;
		float2 Ray_Velocity = lastUV2 - stc;
		float2 Velocity = lerp(Depth_Velocity, Ray_Velocity, WorldNormal.y);
		lastUV = stc + Velocity;
	}
	c4.w = cameraZ;

	if (@_VelocityMapEnable)
	{
		float2 motionVector = tVBuffer.SampleLevel(sVBufferSampler, stc, 0).xy;
		//todo: use longest motion vector
		if (motionVector.x != 0 || motionVector.y != 0)
		{
			lastUV = stc + motionVector;
		}
	}
	
	if (lastUV.x >= 0 && lastUV.x <= 1 && lastUV.y >= 0 && lastUV.y <= 1)//in screen
	{
		//get border Pixels
		float3 c0 = tCurrRT.Sample(sCurrRT, stc + float2(ScreenInfo.z, 0)).xyz;
		float3 c1 = tCurrRT.Sample(sCurrRT, stc + float2(-ScreenInfo.z, 0)).xyz;
		float3 c2 = tCurrRT.Sample(sCurrRT, stc + float2(0, ScreenInfo.w)).xyz;
		float3 c3 = tCurrRT.Sample(sCurrRT, stc + float2(0, -ScreenInfo.w)).xyz;
		//float3 c5 = tCurrRT.Sample(sCurrRT, stc + float2(-ScreenInfo.z, -ScreenInfo.w)).xyz;
		//float3 c6 = tCurrRT.Sample(sCurrRT, stc + float2(+ScreenInfo.z, -ScreenInfo.w)).xyz;
		//float3 c7 = tCurrRT.Sample(sCurrRT, stc + float2(-ScreenInfo.z, -ScreenInfo.w)).xyz;
		//float3 c8 = tCurrRT.Sample(sCurrRT, stc + float2(+ScreenInfo.z, +ScreenInfo.w)).xyz;

		//minmax
		float3 cMin = min(min(min(c0,c1), min(c2,c3)), c4.xyz);
		float3 cMax = max(max(max(c0,c1), max(c2,c3)), c4.xyz);
		//cMin = min(min(min(c5,c6), min(c7,c8)), cMin);
		//cMax = max(max(max(c5,c6), max(c7,c8)), cMax); 

		float scale = 1;
		float3 center = (cMin + cMax) * 0.5f;
		cMin = (cMin - center) * scale + center;
		cMax = (cMax - center) * scale + center;

		float4 history = tHistoryRT.Sample(sHistoryRT, lastUV);
		history.rgb = clamp(history.rgb, cMin, cMax);
		float lastFrameRate = 0.9;
		c4.rgb = lerp(c4.rgb, history.rgb, lastFrameRate);
	}
	return c4;
}

technique TemporalFilterPass
{
	VertexShader = ScreenVS();
	PixelShader = TemporalFilterPS();
}

///////////////////////////////////////// Bilaterial Filter Pass //////////////////////////////////////
float4 BilaterialFilterPS(v2f IN) : SV_Target
{
	static const float s_offsets[5] = {-3.5, -1.5, 1.5, 3.5, 0};
	static const float s_weights[5] = {1.0/16, 1.0/4, 1.0/4, 1.0/16, 3.0/8};
	float c_phi0 = 3.3f;
	float n_phi0 = 1E-2f; // <- normals are VERY important
	float p_phi0 = 5.5f;
	
	float c_phi = 10 * c_phi0;
	float n_phi = n_phi0;
	float p_phi = p_phi0;
	float stepWidth = 1;

	float4 CurrColor = tCurrRT.Sample(sCurrRT, IN.TC);
	//return CurrColor;
	float3 CurrNormal = normalize(DecodeNormal(tGBufferE.Sample(sGBufferE, IN.TC).xyz));
	float SumWeight = s_weights[4];
	float3 SumColor = CurrColor.rgb * s_weights[4];
	for (int i = 0; i < 4; ++i)
	{
		float2 tempTC = IN.TC;
		if (@_BlurX)
			tempTC.x += s_offsets[i] * ScreenInfo.z * stepWidth; 
		else
			tempTC.y += s_offsets[i] * ScreenInfo.w * stepWidth;
		//uv = min(uv, g_pixelSize.zw - 0.5f*g_pixelSize.xy);
		float4 ctmp = tCurrRT.Sample(sCurrRT, tempTC);
		float3 t = CurrColor.rgb - ctmp.rgb;
		float c_dist = dot(t, t);
		//float c_w = exp(-c_dist / c_phi);
		float3 ntmp = normalize(DecodeNormal(tGBufferE.Sample(sGBufferE, tempTC).xyz));
		t = CurrNormal.xyz - ntmp.xyz;
		float n_dist = max(dot(t, t) / (stepWidth*stepWidth), 0.0);
		//float n_w = exp(-n_dist / n_phi);
		float p_dist = Square(CurrColor.w - ctmp.w);
		//float p_w = exp(-p_dist / p_phi);
		float weight = exp(-c_dist/c_phi - n_dist/n_phi - p_dist/p_phi); //c_w * n_w * p_w;
		SumColor += ctmp.rgb * weight;
		SumWeight += weight;
	}
	SumColor /= SumWeight; 
	
	if (@_BlurX)
		return float4(SumColor, CurrColor.w);
	else
		return float4(SumColor, 0);
}
technique BilaterialFilterPass
{
	VertexShader = ScreenVS();
	PixelShader = BilaterialFilterPS();
}

///////////////////////////////////////// Copy Frame Pass //////////////////////////////////////
float4 CopyFramePS(v2f IN) : SV_Target
{
	float4 OUT = tLastFrameScreen.Sample(sLastFrameScreen, IN.TC);
	OUT.rgb = min(SD_SCENE_MAX_LUMINANCE, OUT.rgb);
	return OUT;

}
technique CopyFramePass
{
	VertexShader = ScreenVS();
	PixelShader = CopyFramePS();
}


