// {HEX} GI
#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "DeferredShadingCommon.fxh"
@_PreGIEnable
{ 
	Name = PreGIEnable;
	Description = "Do we have Pre GI?";  
	Default = false;
};
@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};
@_BlurX
{
	Name = BlurX;
	Default = false;
	Visible = false;
}
@_EnableAtrousAnimatorFilter
{
	Name = EnableAtrousAnimatorFilter;
	Default = true;
	Visible = false;
}
///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX				: WBasisX ;
	float4 vWBasisY				: WBasisY ;
	float4 vWBasisZ				: WBasisZ ;
	float2 cPixelSize			;
	float2 cRadius				;
	float4x4 cLastViewProjTex	;
	float4 preWBasisX			;
	float4 preWBasisY			;
	float4 preWBasisZ			;
	float3 preCameraPos			;
};

texture2D tCurrGIMap : layout(t0);
sampler sCurrGISampler : layout(s0)
{
	Texture = tCurrGIMap;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};

texture2D tGBufferE : layout(t1);
sampler sGBufferE : layout(s1)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};
texture2D tPreGIMap : layout(t2);
sampler sPreGISampler : layout(s2)
{
	Texture = tPreGIMap;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};
texture2D tVBuffer : VelocityMap : layout(t3);
sampler sVBufferSampler : layout(s3)
{
	Texture = tVBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D tSceneDepthBuffer : SceneDepth : layout(t4);
sampler sSceneDepthBufferSampler : layout(s4)
{
	Texture = tSceneDepthBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D tCurrRT : layout(t5);
sampler sCurrRT : layout(s5)
{
	Texture = tCurrRT;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};
texture2D tPreDepth : layout(t6);
sampler sPreDepth : layout(s6)
{
	Texture = tPreDepth;
	Filter = Point;
	Address = Clamp;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}

float3 GetPreWorldPosition(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = preWBasisZ.xyz + (preWBasisX.xyz * screenTC.x) + (preWBasisY.xyz * screenTC.y);
	return preCameraPos + pos * depth;
}

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.HPosition.z = OUT.HPosition.w;

	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 FilterPS(v2f IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * cPixelSize;
	half4 currentSSGI = tCurrGIMap.Sample(sCurrGISampler, stc);
	if(!@_PreGIEnable)
	{
		return currentSSGI;
	}
	const int2 sampleOffset[8] = {int2(-1.0, -1.0), int2(0.0, -1.0), int2(1.0, -1.0), int2(-1.0, 0.0), int2(1.0, 0.0), int2(-1.0, 1.0), int2(0.0, 1.0), int2(1.0, 1.0)};
	half4 sampleGIs[8];
	[unroll]
	for(int i = 0; i < 8; ++i)
	{
		sampleGIs[i] = tCurrGIMap.Sample(sCurrGISampler, stc + sampleOffset[i]*cPixelSize);
	}
	float4 m1 = currentSSGI;
	float4 m2 = currentSSGI * currentSSGI;
	[unroll]
	for(int j = 0; j < 8; ++j) {
		m1 += sampleGIs[j];
		m2 += sampleGIs[j] * sampleGIs[j];
	}
	float4 mean = m1 / 9.0;
	float4 stddev = sqrt((m2 / 9.0) - mean * mean);
	float aabb = cRadius.x;
	float4 minGI = min(mean - aabb * stddev, currentSSGI);
	float4 maxGI = max(mean + aabb * stddev, currentSSGI);

	float depth = DecodeDepth(tSceneDepthBuffer.Sample(sSceneDepthBufferSampler, stc));
	float3 worldPos = CalcHomogeneousPos(depth, stc);
	worldPos += CameraPos.xyz;
	float4 lastPos = MulMat4x4(cLastViewProjTex, float4(worldPos, 1.0));
	float2 lastUV = (lastPos.xy / lastPos.w);
	float3 preWorldPos = GetPreWorldPosition(DecodeDepth(tPreDepth.Sample(sPreDepth, lastUV)), lastUV);

	half2 inRange = abs(lastUV - 0.5);
	half preGIRate = step(max(inRange.x, inRange.y), 0.5);
	half4 preGI = tPreGIMap.Sample(sPreGISampler, lastUV);
	int preFilterFlag;
	int currentFilterFlag;
	DecodeFilter(preGI.a, preFilterFlag);
	DecodeFilter(currentSSGI.a, currentFilterFlag);
	float scale = 0;
	if (@_VelocityMapEnable)
	{
		float2 motionVector = abs(tVBuffer.SampleLevel(sVBufferSampler, stc, 0).xy);
		if(motionVector.x+motionVector.y > EPSILON)
		{
			scale = 1;
		}
	}
	float diffDepth = max(max(saturate(preFilterFlag + currentFilterFlag), scale), 1 - length(worldPos - preWorldPos) * 3 / (depth * CameraInfo.y)); // really trick
	
	preGI = clamp(preGI, minGI, maxGI);
	half4 finalGI = lerp(currentSSGI, preGI, saturate(cRadius.y)* preGIRate * diffDepth);
	finalGI.rgb *= step(0.8 * saturate(1 - 50 * depth), diffDepth); // trick for skin
	finalGI.a = EncodeShadowAndFilter(finalGI.a, currentFilterFlag);
	return finalGI;
}
technique Filter
{
	VertexShader = ScreenVS();
	PixelShader = FilterPS();
}

float4 ATrousFilterPassPS(v2f IN) : SV_Target
{
	static const float s_offsets[5] = {-3.5, -1.5, 1.5, 3.5, 0};
	static const float s_weights[5] = {1.0/16, 1.0/4, 1.0/4, 1.0/16, 3.0/8}; //filter kernel about a trous wavelet decomposition
	float c_phi0 = 3.3f;
	float n_phi0 = 1E-2f; // <- normals are VERY important
	float p_phi0 = 0.8f;
	
	float c_phi = 10 * c_phi0;
	float n_phi = n_phi0;
	float p_phi = p_phi0;
	float stepWidth = 2;

	float4 currColor = tCurrRT.Sample(sCurrRT, IN.TC);
	int filterFlag = NO_NEED_FILTER;
	if(@_EnableAtrousAnimatorFilter)
	{
		DecodeFilter(currColor.a, filterFlag);
	}
	float filterWeight = saturate(1 - filterFlag);
	float depth = DecodeDepth(tSceneDepthBuffer.SampleLevel(sSceneDepthBufferSampler, IN.TC, 0));
	float3 currNormal = normalize(DecodeNormal(tGBufferE.Sample(sGBufferE, IN.TC).xyz));
	float sumWeight = s_weights[4];
	float4 sumColor = currColor * s_weights[4];
	for (int i = 0; i < 4; ++i)
	{
		float2 tempTC = IN.TC;
		if (@_BlurX)
			tempTC.x += s_offsets[i] * ScreenInfo.z * stepWidth; 
		else
			tempTC.y += s_offsets[i] * ScreenInfo.w * stepWidth;
		float4 ctmp = tCurrRT.Sample(sCurrRT, tempTC);
		float depthTmp = DecodeDepth(tSceneDepthBuffer.SampleLevel(sSceneDepthBufferSampler, tempTC, 0));
		float3 t = currColor.rgb - ctmp.rgb;
		float c_dist = dot(t, t);
		float3 ntmp = normalize(DecodeNormal(tGBufferE.Sample(sGBufferE, tempTC).xyz));
		t = currNormal.xyz - ntmp.xyz;
		float n_dist = max(dot(t, t) / (stepWidth*stepWidth), 0.0);
		
		float p_dist = Square((depth - depthTmp) * CameraInfo.y);
		float weight = exp(filterWeight*(-c_dist/c_phi - n_dist/n_phi) - p_dist/p_phi);
		sumColor += ctmp * weight;
		sumWeight += weight;
	}
	sumColor /= sumWeight;
	sumColor.a = EncodeShadowAndFilter(sumColor.a, filterFlag);
	return sumColor;
}

technique ATrousFilterPass
{
	VertexShader = ScreenVS();
	PixelShader = ATrousFilterPassPS();
}
// {HEX} end
