//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Effect")

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect shading model.
cbuffer Batch
{
	float4x3 World 				: World;
	float4 EffectColor			: Color;
	float4 EffectParam			: Param;
	float4 SubUV				: Param2;
};

cbuffer Shader
{
	float4 ModulateColor		= { 1, 1, 1, 1 };
	float  BloomScale			= 0.5;
	float  FogStrength			= 0;
	float  DepthBias			= 0;
	float  HDRScale				= 2.0;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Common effect pixel shader.
float4 EffectPixelShading(in EffectVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(1.0, 0.0, 0.0, 1.0);

	// Customize your pixel shader herer.

	if (@_TRANSPARENT)
	{
		float invEmission = GetEffectParam(IN).z;
		OUT = EncodePremultipliedAlphaAndBloom(OUT, invEmission, IN.TexCoord.w, BloomScale);
	}
	else
	{
		OUT.a = EncodeSolidBloomFactor(GetLuminance(OUT.rgb) * BloomScale);
	}

	if(@_HDR)
	{
		EncodeHDRParticle(OUT);
	}
	return OUT;
}

technique Shading
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}
