//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("PBR")

#define VARYING_MASK 0xFFFF

/*{ShadingMode}*/

/*{DefineUsage}*/

/*{VirtualLit}*/

/*{DOUBLE_REFLECTCOLOR}*/

#define OPACITY_MASK_VALUE(IN, ALPHA) { \
/*{OPACITY_MASK_VALUE}*/ \
}

#include "Constants.fxh"
#include "Common.fxh"
#include "ShaderGraphDefinitions.fxh"

//////////////////////////////////Spec. for Baking////////////////////////////////////////
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};
@_IsFoliageBake
{
	Name = IsFoliageBake;
	Description = "is foliage bake?";
	Visible = true;
};

//////////////////////////////////Properties////////////////////////////////////////
/*{GlobalParameters}*/

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 					;
	float4 cTintColor1				: TintColor1				; 
	float4 cTintColor2				: TintColor2				; 
	float4 cTintColor3				: TintColor3				; 

	float4 cShadowBias				: ShadowBias				; 
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
	float4 cPlaceHold											;
#endif
	float4 cParamter				: Param						;
	float4 cParamter2				: Param2					;
	float4 cVirtualLitDir			: VirtualLitDir				;
	float4 cVirtualLitColor			: VirtualLitColor			;
	float4x4 WorldViewProj			: WorldViewProj				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex		;
	float4 PlayerPos				: PlayerPos					;
/*{BatchParameters}*/
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_UV3UV4]
	float4		TexCoord3		: TEXCOORD2;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
	[@_INSTANCED]
	float4		LightMapScale	: TEXCOORD6;
	[@_INSTANCED]
	float4		LightMapTransform: TEXCOORD7;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	
	float4 		DiffLighting 	: TEXCOORDN;			//  PerVertex Diffuse Lighting.
 	
	[@_TANGENT] 
	float4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	float4		WorldBinormal	: TEXCOORDN;

	[@_CastSunShadow]
	float4 		LitUVZ			: TEXCOORDN;

	[@_CastSunShadow]
	float4 		LitUVZ1			: TEXCOORDN;

	[@_CastSunShadow]
	float4 		LitUVZ2			: TEXCOORDN;

	// Shader special output
	float4		TintColor 		: COLOR;				//	TintColor

	[@_INSTANCED]
	float4		LightMapScale	: TEXCOORDN;			//	Light map scale

	[@_UV3UV4]
	float4		MultiUVs		: TEXCOORDN;

/*{costumized uvs}*/
};

//////////////////////////////////Common//////////////////////////////////////
#define VATexCoord2(Arg)		Arg.TexCoord.zw
#define VSTexCoord2				VATexCoord2(OUT)
#define PSTexCoord2				VATexCoord2(IN)

#define VALinearDepth(Arg)		Arg.WorldPosition.w
#define VSLinearDepth			VALinearDepth(OUT)
#define PSLinearDepth			VALinearDepth(IN)

float4 GetVSTintColor(in VertexAssemble IN)
{

	if (@_VERTEXCOLOR)
		return GetVertexColor(IN.Color);
	else
		return float4(1, 1, 1, 1);
}
#define VATintColor(Arg)         
#define VSTintColor             GetVSTintColor(IN)
#define PSTintColor             IN.TintColor

float4 GetVAMultiUVS(in VertexShadingOutput arg)
{
    if (@_UV3UV4)
        return arg.MultiUVs;
    else
        return float4(0, 0, 0, 0);
}
#define VAMultiUVS(Arg)		 GetVAMultiUVS(Arg)
#define VSMultiUVS           VAMultiUVS(OUT)
#define PSMultiUVS           VAMultiUVS(IN)

#include "ShaderGraphVarying.fxh"

//////////////////////////////////Util////////////////////////////////////////

float4 GetDefaultTintColor(in VertexAssemble IN)
{
	float4 tintColor;
	if (@_VERTEXCOLOR && !@_AnisotropicEnable)
	{
		half3 mask = GetVertexColor(IN.Color) > 0.1;
		tintColor = mask.r*cTintColor1 + mask.g*cTintColor2 + mask.b*cTintColor3;
		tintColor += 1 - (half)(mask.r + mask.g + mask.b > 0.001);
	}
	else if (@_VERTEXCOLOR && @_AnisotropicEnable)
		tintColor = GetVertexColor(IN.Color);
	else
	{
		tintColor = float4(1, 1, 1, 1);
	}
	return tintColor;
}


//////////////////////////////////VS////////////////////////////////////////

#define SHADOW_BIAS (cShadowBias.x)
#define SHADOW_SLOPEBIAS (cShadowBias.y)
#if SD_FORWARD_CSM
	#define SHADOW_SELFRATE (frac(cShadowBias.z))
#else
	#define SHADOW_SELFRATE (cShadowBias.z)
#endif
#define SHADOW_PIXELSIZE (cShadowBias.w)
#define LOD_RATE (cParamter.w)
#define SHADOW_BIAS2 (SHADOW_BIAS * floor(cShadowBias.z))
#define SHADOW_SLOPEBIAS2 (SHADOW_SLOPEBIAS * floor(cShadowBias.z))

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;

	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	[@_HLSL]
	INTM = (InterMediateVertex)0;

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		uint4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			if(@_TANGENT && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
			}
		}
	}
}

// Get 3x4 matrix from skeleton data.
void GetSkeletonMatrix4x3Tex( int idx, out float4 Mx, out float4 My, out float4 Mz )
{
	Mx = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+0.5)*cParamter.z, 0), 0);
	My = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+1.5)*cParamter.z, 0), 0);
	Mz = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+2.5)*cParamter.z, 0), 0);
}

void VertexUnifySkinningTex(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3Tex(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3Tex(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3Tex(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3Tex(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			if(@_TANGENT && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
			}
		}
	}
}

void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	[@_UV3UV4]
	OUT.MultiUVs = IN.TexCoord3;
	
/*{VertexShader}*/

	/*{TintColor}*/

	/*{WorldPositionOffset}*/

	// linear z will be calced in BasePBRPixelShader.fxh
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;

	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));

	//{DepthBias} OUT.HPosition.z -= CameraInfo.x * /*{DepthBias}*/ / (/*{DepthBias}*/ + OUT.HPosition.w);

}

//////////////////////////////////HS////////////////////////////////////////
float4 HS_GetTessellationFactor(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2)
{
/*{HullShader}*/
    
	return /*{HS_FACTOR}*/;
}

//////////////////////////////////DS////////////////////////////////////////
float3 Project(float3 p, float3 c, float3 n)
{
	return p - dot(p - c, n) * n;
}
void DS_DIY_OUTPUT(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2, float3 domain, inout VertexShadingOutput IN)
{
	/*{DS_Default_OUTPUT}
	float3 c0 = Project(IN.WorldPosition.xyz, p0.WorldPosition.xyz, normalize(p0.WorldNormal.xyz));
	float3 c1 = Project(IN.WorldPosition.xyz, p1.WorldPosition.xyz, normalize(p1.WorldNormal.xyz));
	float3 c2 = Project(IN.WorldPosition.xyz, p2.WorldPosition.xyz, normalize(p2.WorldNormal.xyz));
	float3 pos = c0 * domain.x + c1 * domain.y + c2 * domain.z;
	IN.WorldPosition.xyz = lerp(IN.WorldPosition.xyz, pos, 0.5);
	{DS_Default_OUTPUT}*/
	
/*{DomainShader}*/
    /*{DS_OUTPUT}*/
}

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);

/*{PixelShader}*/

	/*{BaseColor}*/
	/*{Metallic}*/
	/*{Roughness}*/
	/*{Specular}*/
	/*{Emission}*/
	/*{AO}*/
	/*{Alpha}*/

	/*{SubsurfaceColor}*/
	/*{Opacity}*/
	/*{Thickness}*/
	/*{CustomData}*/
}
#include "PBR_PS.fxh"
