//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

// Shader template for shader graph thumbnail rendering.
#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

#define VARYING_MASK 0xFFFF
#include "ShaderGraphDefinitions.fxh"

/*{GlobalParameters}*/

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World;
/*{BatchParameters}*/
};

struct VertexShadingOutput
{
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4 		TintColor		: COLOR;				//  Vertex Color.
	float2		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.
	float2		TexCoord2		: TEXCOORDN;			//	Texture coordinate for surface.
	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.
	float4		WorldTangent	: TEXCOORDN;			//	Output tangents in world space for normal mapping.
	float4		WorldBinormal	: TEXCOORDN;			//	Output binormal in world space for normal mapping.
	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
/*{costumized uvs}*/
};

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Color 			: COLOR;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;
	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;
};

//////////////////////////////////////////////////////////////////////////

#define VATexCoord2(Arg)		Arg.TexCoord2
#define VSTexCoord2				VATexCoord2(OUT)
#define PSTexCoord2				VATexCoord2(IN)

#define VALinearDepth(Arg)		Arg.WorldPosition.w
#define VSLinearDepth			VALinearDepth(OUT)
#define PSLinearDepth			VALinearDepth(IN)

#include "ShaderGraphVarying.fxh"

//////////////////////////////////////////////////////////////////////////
VertexShadingOutput ThumbnailVS(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	OUT = (VertexShadingOutput)0;

/*{VertexShader}*/

	OUT.HPosition = MulMat4x4(ViewProj, IN.Position);
	OUT.TintColor = GetVertexColor(IN.Color);
	OUT.TexCoord = IN.TexCoord;
	OUT.TexCoord2 = IN.TexCoord;

	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, (float3)Expand((half3)IN.Normal.xyz)));
	OUT.WorldNormal.w = 0;

	OUT.WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, OUT.HPosition.w);

	OUT.ViewDirection = (CameraPos.xyz - OUT.WorldPosition.xyz);

	if(@_TANGENT)
	{
		OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, normalize(IN.Tangent.xyz)));
		OUT.WorldTangent.w = 0;
		OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, normalize(IN.Binormal.xyz)));	
		OUT.WorldBinormal.w = 0;
	}

	return OUT;
}

float4 ThumbnailPS(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = 0;

	IN.ViewDirection = normalize(IN.ViewDirection);

/*{PixelShader}*/
	return OUT;
}

technique Shading
{
	VertexShader = ThumbnailVS();
	PixelShader = ThumbnailPS();
}
