//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Shadow.fx>
///	@path	~/System/Shader/
///	@date	2016/05/24
///	@brief	shadow things

#pragma shader draw
#pragma shader predict("@_INSTANCED", "false")

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_CameraInVolume
{ 
	Name = CameraInVolume;
	Default = true;
};

@_PerspectiveShadowMap
{ 
	Name = PerspectiveShadowMap;
	Default = false;
};

@_HighQualityShadow
{
	Name = HighQualityShadow;
	Default = false;
};

@_HasSpotLightCookie
{
	Name = HasSpotLightCookie;
	Description = "Does this spot light has cookie?";  
	Visible = true;
};

@_IsSpotLightVolume
{
	Name = IsSpotLightVolume;
	Description = "Is spot light volume";
	Default = false;
	Visible = true;
};
///////////////// const buffer //////////////////
cbuffer Batch
{
	float4x3 cWorld 				;

	float4 cWBasisX		: WBasisX 	;
	float4 cWBasisY		: WBasisY	;
	float4 cWBasisZ		: WBasisZ	;

	float4x4 cShadowViewProjTex		;
	float cFilterRadius				;
	float cInvShadowMapSize			;
	float cFarPlane					;
	float cCookieRepeat				 = 5;
	float cShadowBias				;
	float cShadowSlopBias			;
	float  cLightYScaleBelow1		; //Range / [sqrt(2) * YScale]
	int   cShadowSliceS				;
	int   cShadowSliceD				;
	int   cShadowSliceSize			;
};

[inputColor(0)]
texture2D tGBufferA : layout(t0);
sampler sGBufferA : layout(s0)
{
	Texture = tGBufferA;
	Filter = Point;
	Address = Clamp;
};

[inputColor(1)]
texture2D tGBufferB : GBufferMapB : layout(t1);
sampler sGBufferB : layout(s1)
{
	Texture = tGBufferB;
	Filter = Point;
	Address = Clamp;
};

[inputColor(2)]
texture2D tGBufferC : GBufferMapC : layout(t2);
sampler sGBufferC : layout(s2)
{
	Texture = tGBufferC;
	Filter = Point;
	Address = Clamp;
};

[inputColor(3)]
texture2D tGBufferD : GBufferMapD : layout(t3);
sampler sGBufferD : layout(s3)
{
	Texture = tGBufferD;
	Filter = Point;
	Address = Clamp;
};

[inputColor(4)]
texture2D tGBufferE : GBufferMapE : layout(t4);
sampler sGBufferE : layout(s4)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};

//the writting RT
[inputColor(5)]
texture2D tGBufferF : GBufferMapF : layout(t5);
sampler sGBufferF : layout(s5)
{
 	Texture = tGBufferF;
 	Filter = Point;
 	Address = Clamp;
};

[inputColor(6)]
texture2D tDepthMap : SceneDepth : layout(t6);
sampler sDepthSampler : layout(s6)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tShadowMap : layout(t7);
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
};

texture2D tShadowMapS : layout(t0);
sampler sShadowSamplerS : layout(s0)
{
	Texture = tShadowMapS;
	Filter = Point;
	Address = Clamp;
};
texture2D tShadowMapD : layout(t1);
sampler sShadowSamplerD : layout(s1)
{
	Texture = tShadowMapD;
	Filter = Point;
	Address = Clamp;
};

texture2D tLightCookie : layout(t8) = "White";
sampler sLightCookie : layout(s8)
{
	Texture = tLightCookie;
	Filter = Linear;
	Address = Wrap;
	SRGB = false; 
};

///////////////// io signatures //////////////////
/// light volume vertex to vs
struct VertexAssembleShadowVolume
{
	float3		Position		: POSITION;
};

// vertex output to pixel shader.
struct VertexOutputShadowVolume
{
	float4		HPosition		: SV_Position;
};

///////////////// helper functions //////////////////
#include "Lighting.fxh"

float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}
#define DecodeGBuffer(_GBufferBaseColor, _GBufferNormal, BaseColor, Roughness, N, Metallic, SSS, InShadow)\
{\
	BaseColor = _GBufferBaseColor.rgb;\
	DecodeFloatAndBit(_GBufferBaseColor.a, Roughness, SSS);\
	N = normalize(DecodeNormal(_GBufferNormal.rgb));\
	DecodeFloatAB(_GBufferNormal.a, Metallic, InShadow);\
}

////////////////////////////////////LightVolumeStencil////////////////////////////////////
VertexOutputShadowVolume ShadowVolumeVS( VertexAssembleShadowVolume IN )
{
	VertexOutputShadowVolume OUT;

	float3 pos = IN.Position;
	if (@_IsSpotLightVolume)
	{
		if (pos.y < -0.9)
		{
			float oldR2 = 2 - pos.y*pos.y;
			pos.y = -1 + (pos.y+1) * (cLightYScaleBelow1 - 1) / (sqrt(2) - 1);
			float newR2 = (cLightYScaleBelow1*cLightYScaleBelow1-pos.y*pos.y);
			if (oldR2 > 0.01)
				pos.xz *= sqrt(newR2 / oldR2);
		}
	}

	// Transform position.
	float4 worldPosition = float4(MulMat3x4(cWorld, float4(pos, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, worldPosition);
	OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);
	
	return OUT;
}

float4 ShadowVolumeStencilPS( VertexOutputShadowVolume IN ) : SV_Target
{
	return float4(0, 0, 1, 1);
}

technique ShadowVolumeStencil
{
	VertexShader = ShadowVolumeVS();
	PixelShader = ShadowVolumeStencilPS();
}

////////////////////////////////ShadowMask////////////////////////////////////////
float4 ShadowMaskPS( VertexOutputShadowVolume IN ) : SV_Target
{
	//1, GBufferData
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Fetch(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 WorldPos = CameraPos.xyz + D;
	GBufferData GBuffer = DecodeGBufferData(
		tGBufferB.Fetch(sGBufferB, stc), 
		tGBufferC.Fetch(sGBufferC, stc),
		tGBufferD.Fetch(sGBufferD, stc),
		tGBufferE.Fetch(sGBufferE, stc),
		tGBufferF.Fetch(sGBufferF, stc));

	//light space uvz
	//todo: face to normal?
	float4 P_LS = mul(float4(WorldPos+GBuffer.WorldNormal*cShadowSlopBias,1), cShadowViewProjTex);

	if(@_PerspectiveShadowMap)
	{
		half NoL = saturate(dot(SunDirection.xyz, GBuffer.WorldNormal));
		P_LS.z -= cShadowBias;
		P_LS.xyz /= P_LS.w;
		//P_LS.z = P_LS.w / cFarPlane;
	}
	else//sun light
	{
		half NoL = saturate(dot(SunDirection.xyz, GBuffer.WorldNormal));
		//hard code scale avoiding artifact(mathematical precision of WorldPos & P_LS)
		P_LS.z -= cShadowBias * clamp(sceneZ * CameraInfo.y * 0.1, 1, 3);
		P_LS.xyz /= P_LS.w;
	}
	//do not need z clamping here, stencil test has already clipped out of range pixels for us.
	//P_LS.z = min(1.0-10*EPSILON, P_LS.z);

	//clip outrange 
	if (!@_CameraInVolume)
	{
		clip(0.5 - abs(P_LS.xyz - 0.5));
	}
	
	//Shadow
	float InShadow = 0;
	if (@_HighQualityShadow)
	{
		hardware_pcf_4x4(P_LS.xy, P_LS.z, cInvShadowMapSize.xx, InShadow);
	}
	else
	{
		hardware_pcf_2x2(P_LS.xy, P_LS.z, cInvShadowMapSize.xx, InShadow);
	}

	//Light cookie
	float Shadow = 1 - InShadow;
	if (@_HasSpotLightCookie)
	{
		[@_GLSL]
		P_LS.y = 1 - P_LS.y;
		
		Shadow *= tLightCookie.Sample(sLightCookie, P_LS.xy * cCookieRepeat).r;
	}
	InShadow = 1 - Shadow;
	return float4(InShadow, 1, 1, 1);
}

technique ShadowMask
{
	VertexShader = ShadowVolumeVS();
	PixelShader = ShadowMaskPS();
}
////////////////////////////////MergeShadowMap////////////////////////////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC		 : TEXCOORD0;
};

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, EPSILON, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

[target("X16UNORM")]
float4 MergeShadowMapPS(v2f IN) : SV_Target
{
	float depthS = tShadowMapS.Sample(sShadowSamplerS, IN.TC).r;
	float depthD = tShadowMapD.Sample(sShadowSamplerD, IN.TC).r;
	float depth = min(depthS, depthD);
	return float4(depth,depth,depth,depth);
}

technique MergeShadowMap
{
	VertexShader = ScreenVS();
	PixelShader = MergeShadowMapPS();
}

////////////////////////////////Separate Shadow Pass////////////////////////////////////////
float4 SeparateShadowPS( VertexOutputShadowVolume IN ) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	
	[@_D3D9]  
	spos += 0.5;
	
	//P	
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.Sample(sDepthSampler, stc));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	float3 P = CameraPos.xyz + D;
	
	//light space uvz
	float4 P_LS = mul(float4(P,1), cShadowViewProjTex);
	P_LS.xyz /= P_LS.w;

	// if(@_PerspectiveShadowMap)
	// {
	// 	P_LS.z = P_LS.w / cFarPlane;
	// }

	// //clip outrange 
	// if (!@_CameraInVolume)
	// {
	// 	clip(0.5 - abs(P_LS.xyz - 0.5));
	// }
	
	//Shadow
	float InShadow = 0;
	if (@_HighQualityShadow)
	{
		hardware_pcf_4x4(P_LS.xy, P_LS.z, cInvShadowMapSize, InShadow);
	}
	else
	{
		hardware_pcf_2x2(P_LS.xy, P_LS.z, cInvShadowMapSize, InShadow);
	}

	//float Shadow = 1 - InShadow;
	return float4(lerp(float3(1,1,1), ShadowColor.rgb*0.6, InShadow * SHADOW_RATE), 1);
}

technique SeparateShadow
{
	VertexShader = ShadowVolumeVS();
	PixelShader = SeparateShadowPS();
}

