//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzlixianying@corp.netease.com
//	Author: sinelee
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Stylization.fx>
///	@path	~/System/Shader/
///	@date	2013/04/22
///	@brief	Stylized filters.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Stylized filters.

cbuffer Batch
{
	float4 AdjustHsvParams0	;
};

// Source target.
texture2D SourceLinear0 : layout(t0);
sampler SourceLinear0_Sampler : layout(s0)
{
	Texture = SourceLinear0;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

/// Pixel output.
struct PixelOutput
{
	float4		Color			: SV_Target;
};

VertexScreenOutput ScreenVS( VertexAssembleScreen IN )
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1);
	OUT.TC = IN.BaseTC;
	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Adjustment in HSV color space.

PixelOutput AdjustHsvPS( VertexScreenOutput IN )
{
	PixelOutput OUT;
	half4 c0 = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC);

	half4 hsvl = RGB2HSVL(c0.rgb);
	// Adjust hue.
	hsvl.x += AdjustHsvParams0.x;
	// sinelee: todo: apply adobe ps's algorithm for manipulating S/V channels
	// Adjust saturation.
	hsvl.y += AdjustHsvParams0.y;
	// Adjust brightness.
	hsvl.z += AdjustHsvParams0.z;
	c0.rgb = HSV2RGB(hsvl.xyz);

	OUT.Color = float4(c0.rgb, 1.0);
	return OUT;
}

technique AdjustHsv
{
	VertexShader = ScreenVS();
	PixelShader = AdjustHsvPS();
}
