//	--------------------------------------------------------------------
//	Copyright(C) 2006~2019 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liusizhuang@corp.netease.com
//	Author: liusizhuang
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<TSAA.fx>
///	@path	~/System/Shader/
///	@date	2019/11/--

//@_TAAQualityLevel		 Sample			DepthCull
//		0				(6 samps)		   Off
//		1				(9 samps)		   1+3
//		2				(11 samps)		   3+3
//		3				(15 samps)		   5+5

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////

#define TAALevel (@_TAAQualityLevel1*2+@_TAAQualityLevel0)


@_TAAQualityLevel0
{
	Name = TAAQualityLevel0;
	Default = true;
	Visible = false;
};

@_TAAQualityLevel1
{
	Name = TAAQualityLevel1;
	Default = true;
	Visible = false;
};

@_TAAFullSize
{
	Name = TAAFullSize;
	Default = false;
	Visible = false;
};

// #if SD_TSAA_VBUFFER
@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};
// #endif

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX				;
	float4 vWBasisY				;
	float4 vWBasisZ				;
	float4x4 lastViewProjTex	;
	float4 cRTSize				;
	float cLastFrameRate 		;
	float cLastFrameRate2		;
	float2 padding				;
	float4 cJitter				;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tScreenMap : layout(t0);
sampler sScreenSampler : layout(s0)
{
	Texture = tScreenMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
	Precision = Medium;
};

texture2D tScreenMapLinear : layout(t1);
sampler sScreenLinearSampler : layout(s1)
{
	Texture = tScreenMapLinear;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
	Precision = Medium;
};

// Backbuffer target.
texture2D tLastScreenMap : layout(t2);
sampler sLastScreenSample : layout(s2)
{
	Texture = tLastScreenMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
	Precision = Medium;
};

texture2D tDepthBuffer : SceneDepth : layout(t3);
sampler sDepthBufferSampler : layout(s3)
{
	Texture = tDepthBuffer;
	Filter = Linear;
	Address = Clamp;
};
texture2D tLastDepthMap : layout(t4);
sampler sLastDepthSampler : layout(s4)
{
	Texture = tLastDepthMap;
	Filter = Linear;
	Address = Clamp;
};
// #if SD_TSAA_VBUFFER
texture2D tVBuffer : VelocityMap : layout(t5);
sampler sVBufferSampler : layout(s5)
{
	Texture = tVBuffer;
	Filter = Linear;
	Address = Clamp;
};
// #endif

///////////////////////// io signatures ///////////////////////
struct a2v 
{
	float4 Position 	: POSITION; //in object space
	float2 BaseTC   	: TEXCOORD0;
};

struct v2f
{
	float4 HPosition 		: SV_Position;
	float2 TC		 		: TEXCOORDN;
	float4 JitteredUV		: TEXCOORDN;
	// float2 LastJitteredUV	: TEXCOORDN;
};


/////////////////////////Common Functions///////////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}

float Luminance(float4 linearRgb)
{
    return dot(linearRgb.xyz, float3(0.2126729, 0.7151522, 0.0721750));
}

float3 RGB2YCoCg(float3 rgb)
{
	float R  = rgb.r; float G = rgb.g; float B = rgb.b;
	float T = 0.25*R + 0.25*B;
	float Y  = 0.5*G + T;
	float Cg = 0.5*G - T;
	float Co = 0.5*(R-B);
	return float3(Y, Cg, Co);
}

float3 YCoCg2RGB(float3 YCoCg)
{
	float Y  = YCoCg.x; float Cg = YCoCg.y; float Co = YCoCg.z;
	float t = Y - Cg;
	float R = t + Co;
	float G = Y + Cg;
	float B = t - Co;

	return float3(R, G, B);
}

float4 RGB2YCoCg4(float4 rgb)
{
	float R  = rgb.r; float G = rgb.g; float B = rgb.b;
	float T = 0.25*R + 0.25*B;
	float Y  = 0.5*G + T;
	// Y = sqrt(Y);
	float Cg = 0.5*G - T;
	float Co = 0.5*(R-B);
	return float4(Y, Cg, Co, 1);
}

float4 YCoCg2RGB4(float4 YCoCg)
{
	float Y  = YCoCg.x; float Cg = YCoCg.y; float Co = YCoCg.z;
	// Y = Y*(Y);
	float t = Y - Cg;
	float R = t + Co;
	float G = Y + Cg;
	float B = t - Co;

	return float4(R, G, B, 1);
}

#define initHalf4c9(colors)\
{\
	colors[0] = half4(0,0,0,0);\
	colors[1] = half4(0,0,0,0);\
	colors[2] = half4(0,0,0,0);\
	colors[3] = half4(0,0,0,0);\
	colors[4] = half4(0,0,0,0);\
	colors[5] = half4(0,0,0,0);\
	colors[6] = half4(0,0,0,0);\
	colors[7] = half4(0,0,0,0);\
	colors[8] = half4(0,0,0,0);\
}

#define initFloat9(va)\
{\
	va[0] = 0;\
	va[1] = 0;\
	va[2] = 0;\
	va[3] = 0;\
	va[4] = 0;\
	va[5] = 0;\
	va[6] = 0;\
	va[7] = 0;\
	va[8] = 0;\
}

/////////////////// include Different TAA /////////////////////

// TSAA By Messiah Source Version
#include "MessiahTAA.fxh"

// TSAA By HEX
#include "HexTAA.fxh"

// FXAA By Nvidia
#define FXAA_PC 1
#define FXAA_GREEN_AS_LUMA 1
#define FXAA_QUALITY__PRESET 10
#define FXAA_QUALITY_SUBPIX 1.0
#define FXAA_QUALITY_EDGE_THRESHOLD 0.063
#define FXAA_QUALITY_EDGE_THRESHOLD_MIN 0.0312
#define FxaaTexture tScreenMap
#define FxaaSampler sScreenSampler
#include "FXAA.fxh"

////////////////////// Vertex Shading //////////////////////
float CatmullRom(float x)
{
	float ax = abs(x);
	if (ax > 1.0f)
		return ((-0.5f * ax + 2.5f) * ax - 4.0f) *ax + 2.0f;
	else
		return (1.5f * ax - 2.5f) * ax*ax + 1.0f;
}

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.HPosition.z = OUT.HPosition.w;

	OUT.TC.xy = IN.BaseTC.xy;

	half2 spos = OUT.HPosition.xy;
	[@_D3D9]
	spos += 0.5;

	half4 jitter = cJitter / cRTSize.xyxy;

	//OUT.JitteredUV.xy__ -> JitterFixedUV
	//OUT.JitteredUV.__zw -> LastJitterOFfset
	OUT.JitteredUV.xy = OUT.TC.xy; 
	OUT.JitteredUV.xy += jitter.xy;
	OUT.JitteredUV.zw = jitter.zw;

	return OUT;
}

////////////////////// Pixel Shading  //////////////////////
float4 TAAPS_Main(v2f IN) : SV_Target
{
	if(TAALevel == 0)
		return FxaaPixelShader(
			IN.JitteredUV.xy,           // ScreenUV : Jitter Fixed UV
			0.0,                        // fxaaConsolePosPos (unused)
			0.0,                    	// tex (unused) <- SampleTexture is Define in next Level define
			0.0,                    	// fxaaConsole360TexExpBiasNegOne (unused)
			0.0,                    	// fxaaConsole360TexExpBiasNegTwo (unused)
			1.0/cRTSize.xy,      		// fxaaQualityRcpFrame : Pre-pixel size
			0.0,                        // fxaaConsoleRcpFrameOpt (unused)
			0.0,                        // fxaaConsoleRcpFrameOpt2 (unused)
			0.0,                        // fxaaConsole360RcpFrameOpt2 (unused)
			FXAA_QUALITY_SUBPIX,
			FXAA_QUALITY_EDGE_THRESHOLD,
			FXAA_QUALITY_EDGE_THRESHOLD_MIN,
			0.0,                        // fxaaConsoleEdgeSharpness (unused)
			0.0,                        // fxaaConsoleEdgeThreshold (unused)
			0.0,                        // fxaaConsoleEdgeThresholdMin (unused)
			0.0                         // fxaaConsole360ConstDir (unused)
		);
	else if(TAALevel == 1)
		return TAA_HEX(
			IN,							//v2f
			false,						//is Depth Cull Enable
			false,						//is YCoCg Tone Mapping Enable
			true,						//using VBuffer
			false,						//using Variance Clip
			true,						//fast mode
			true,						//Nearest Velocity
			0.25,						//Edge Blur Enhance
			0.5							//Sample Range Scale
		);
	else if(TAALevel == 2)
		return TAA_HEX(
			IN,							//v2f
			true,						//is Depth Cull Enable
			false,						//is YCoCg Tone Mapping Enable
			true,						//using VBuffer
			true,						//using Variance Clip
			true,						//fast mode
			true,						//Nearest Velocity
			0.25,						//Edge Blur Enhance
			0.5							//Sample Range Scale
		);
	else if(TAALevel == 3)
		return TAA_HEX(
			IN,							//v2f
			true,						//is Depth Cull Enable
			true,						//is YCoCg Tone Mapping Enable
			true,						//using VBuffer
			true,						//using Variance Clip
			false,						//fast mode
			true,						//Nearest Velocity
			0.25,						//Edge Blur Enhance
			0.5							//Sample Range Scale
		);
	else
		return SampleColorMap(tScreenMap, sScreenSampler, IN.JitteredUV.xy );
}

technique TSAATech
{
	VertexShader = ScreenVS();
	PixelShader = TAAPS_Main();
}


