//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzlixianying@corp.netease.com
//	Author: sinelee 
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<TerrainMask.fx>
///	@path	~/System/Shader/
///	@date	2020/06/02
///	@brief	Terrain material.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			0
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define NO_SHADOW_CLIPPING
#define USE_REFINED_BLEND		1
#define IS_TERRAIN 				1

#define half  float
#define half2 float2
#define half3 float3
#define half4 float4

#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "false")
#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "Enable reflection?";
	Default = false;
	Visible = false;
}
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable image based lighting?";
	Default = false;
	Visible = false;
}
@_EnableCliff
{
	Name = EnableCliff;
	Description = "Enable cliff?";
	Default = false;
	Visible = false;
}
@_EnableBlendSharpness
{
	Name = EnableBlendSharpness;
	Description = "Enable blend sharpness?";
	Default = false;
	Visible = false;
}
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have light map?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = false;
}
@_BlockMapEnable
{
	Name = BlockMapEnable;
	Description = "Do we have block map?";
	AffectedTex = tBlockMap;
	Visible = false;
}
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have normal map?";
	AffectedTex = tNormalMap;
	Fallback = true;
	Visible = false;
}
@_MixtureMapEnable
{
	Name = MixtureMapEnable;
	Description = "Do we have mixture map?";
	AffectedTex = tMixtureMap;
	Fallback = true;
	Visible = false;
}
@_MaskMapEnable
{
	Name = MaskMapEnable;
	Description = "Do we have mask map?";
	AffectedTex = tMaskMap;
	Visible = false;
}

cbuffer Shader
{
	float cAliasingFactor			 = 0;   //抗锯齿因子
	float cEnvStrength				 = 1;   //环境强度[0:8]
	float cRoughnessLow				 = 0;   //粗糙度下界
	float cRoughnessHigh			 = 1;   //粗糙度上界

	float cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cBlendSharpness			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;

	float4 cBlockScale				 = { 1, 0, 0, 0 };
	float4 cBlockShrinkage			 = { 0, 0, 1, 1 };

	float4 cBlendMapSize			 = { 1, 1, 0, 0 };

	float4 cLightMapScale			 = { 1, 1, 1, 1 };
	float4 cLightMapUVTransform		 = { 1, 1, 0, 0 };
	float4 cMaskMapUVTransform		 = { 1, 1, 0, 0 };
};

texture2D tIndexMap : layout(t0);
sampler sIndexMapSampler : layout(s0)
{
	Texture = tIndexMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlendMap : layout(t1);
sampler sBlendMapSampler : layout(s1)
{
	Texture = tBlendMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlockMap : layout(t2);
sampler sBlockMapSampler : layout(s2)
{
	Texture = tBlockMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tMixtureMap : layout(t3) = "Reflection";
sampler sMixtureMapSampler : layout(s3)
{
	Texture = tMixtureMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalMapSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tEnvMap : EnvMap : layout(t6);
sampler sEnvSampler : layout(s6)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tShadowMap : ShadowMap : layout(t7);
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tMaskMap : layout(t8);
sampler sMaskMapSampler : layout(s8)
{
	Texture = tMaskMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tRainMap : RainMap : layout(t9) = "NormalMap";
sampler sRainSampler : layout(s9)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	OUT.ViewDirection.w = 0;

	// World * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));

	if (@_LOD_LEVEL <= LOD_Level_1)
	{
		if (@_EnableCliff)
		{
			half up = step(OUT.WorldNormal.x, OUT.WorldNormal.y) * step(OUT.WorldNormal.z, OUT.WorldNormal.y);
			OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, half3(0, 1 - up, up)));
		}
		else
		{
			OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, half3(0, 0, 1)));
		}
		OUT.WorldBinormal.xyz = cross(OUT.WorldNormal.xyz, OUT.WorldTangent.xyz);
	}
}

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);

	// mask.
	half3 diff = half3(0,0,0);
	if (@_MaskMapEnable)
	{
		float2 tc = BaseMapUV(IN.TexCoord).yx * cMaskMapUVTransform.xy + cMaskMapUVTransform.zw;
		diff = tMaskMap.Sample(sMaskMapSampler, tc).rgb;
	}
	// shading.
	MaterialParameters.BaseColor = DecodeGamma(diff);
	MaterialParameters.Metallic = 0;
	MaterialParameters.Specular = 0.5;
	MaterialParameters.Roughness = 1 - cSmoothness;
	MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
	MaterialParameters.Alpha = 1.0;
}
#include "PBR_PS.fxh"
