//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzwangyingqiao@corp.netease.com
//	Author: gzwangyingqiao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<TextureDisplay.fx>
///	@path	~/System/Shader/
///	@date	2019/12/09
///	@brief	Display a texture

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tColorMap : layout(t0) = "White";
sampler sColorSampler : layout(s0)
{
	Texture = tColorMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

cbuffer Batch
{
	float4 UVTransform					 = { 1.0f, 0.0f, 0.0f, 1.0f };
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/////////////////// HDRSampleLumInitial ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;

	if (@_VULKAN)
	{
		float2 t = (OUT.TC - 0.5) * 2.0;
		OUT.TC.x = dot(t, UVTransform.xy);
		OUT.TC.y = dot(t, UVTransform.zw);
		OUT.TC = OUT.TC * 0.5 + 0.5;
	}

	return OUT;
}

float4 TextureDisplayPS(v2f IN) : SV_Target
{
	half3 Source = SampleColorMap(tColorMap, sColorSampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

technique TextureDisplayTech
{
	VertexShader = ScreenVS();
	PixelShader = TextureDisplayPS();
}


