//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<PostProcess.fx>
///	@path	~/System/Shader/
///	@date	2013/04/24
///	@brief	Post process utils.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "Noise.fxh"

@_OutputHDRDisplay
{
	Name = outputHDRDisplay;
	Description = "output to HDR display";
	AffectedTex = tUIMap;
	// Predict = true;
	Visible = true;
}

@_LinearFilter
{
	Name = useLinearFilter;
	Description = "use linear filter";
	Visible = true;
	Predict = true;
	Default = false;
};

@_FullSize
{
	Name = FullSize;
	Visible = true;
	Predict = false;
	Default = false;

};

@_ColorGradingEnable
{
	Name = ColorGradingEnable;
	Description = "Color Grading Enable";
	Visible = true;
	Predict = false;
	Default = false;
}

@_ColorCorrectionEnable
{	
	Name = ColorCorrectionEnable;
	Description = "Color correction Enable";
	Default = false;
	Visible = true;
	Predict = false;
	
	
}

@_GrayEnable
{
	Name = GrayEnable;
	Description = "GrayEnable";
	Visible = false;
	Predict = false;
	Default = false;
}

@_MotionBlurEnable
{
	Name = MotionBlurEnable;
	Description = "MotionBlurEnable";
	Visible = false;
	Predict = false;
	Default = false;
}

// {HEX}
@_VignettingEnable
{
	Name = VignettingEnable;
	Description = "VignettingEnable";
	Visible = true;
	Predict = true;
}
// {HEX} end

@_DiyMapEnable
{
	Name = DiyMapEnable;
	AffectedTex = tDiyMap;
	Description = "Diy Texture Enable";
	Visible = true;	
	Default = false;
	Predict = false;
}

@_DiyLocalBlurEnable
{
	Name = DiyLocalBlurEnable;
	Description = "use local blur";
	Visible = true;
	Default = false;
	Predict = false;

}
@_GlitchyEnable
{
	Name = GlitchyEnable;
	Description = "use glitchy";
	Visible = true;
	Default = false;
	Predict = false;
}

@_RainDropEnable
{
	Name = RainDropEnable;
	Description = "use glitchy";
	Visible = true;
	Default = false;
	Predict = false;
}

@_DiySickBlurEnable
{
	Name = DiySickBlurEnable;
	Description = "sick blur Enable";
	Visible = true;
	Predict = false;
	Default = false;
}

@_TrembleEnable
{
	Name = TrembleEnable;
	Description = "TrembleEnable";
	Visible = true;
}
@_QuiverEnable
{
	Name = QuiverEnable;
	Description = "QuiverEnable";
	Visible = true;
}
@_DamageEnable
{
	Name = DamageEnable;
	Description = "DamageEnable";
	Visible = true;
}
@_FlashEnable
{
	Name = FlashEnable;
	Description = "FlashEnable";
	Visible = true;
}
@_bloody
{
	Name = bloody;
	Description = "bloodyEnable";
	Visible = true;
}

//////////////////////////////////////////////////////////////////////////
// Post process utils.

cbuffer Batch
{
	float cLocalRate			;
	float3 cMaskColor			;
	
	float cGrayPercent	 		;
	float cOldMoviePercent 		;
	float2 cBlurCenterUV		;
	
	float cBlurPercent 			;
	float cBlurIntensity 		 = 1.0f;
	float cDiyLocalBlurPercent	;
	float cWhiteTemp			;
	
	float3 cWhiteTint			;
	float cPlaceHold2			;	
	
	float4 cSaturation			;
	float4 cContrast			;
	float4 cGamma				;
	float4 cGain				;
	float4 cOffset				;
	float4 UVTransform			 = { 1.0f, 0.0f, 0.0f, 1.0f };
	float4 cPlayerScreenRect	;
	float4 cDiyLocalBlurParam	 = {1, 0, 1, 1};
	
	float cIsMainViewport		 = 1;
	float cDistortionIntensity		 = 1;
	float cClipIntensity			 = 1;
	float cColorShiftIntensity		 = 1;
	
	float2 cGlitchTextureTileCount  ;
	float cIntensity				 = 0.5;
	float cTextureAnimIntensity		 = 1;
	
	float4 cDiySickBlurParam		 = {1, 1, 1, 1};
	
	float3 cDiySickBlurColor		 = {1, 0.7, 0.9};
	float cDiySickBlurPercent		;

	// {HEX} Vignetting
	float3 cVignettingColor			= {0.0f, 0.0f, 0.0f};
	float cVignettingIntensity      = 1;
	
	float cVignettingPercent        ;
	float cDropIntensity			 = 0.5;
	float cDamageRate				 = 0.0;
	float cHitRate					 = 0.0;
	
	float cflashRate				 = 0.0;
	float cTrembleRate				 = 0.0;
	float cQuiverIntensity			 = 0.0;
	float cBloodRate				 = 0.5;
	float cScreenCut				 = 1.0;
	float cPadding					 = 0.0;
};

// Source target.
texture2D SourcePoint0 : layout(t0);
sampler SourcePoint0_Sampler : layout(s0)
{
	Texture = SourcePoint0;
	Filter = Point;
	Address = Clamp;
};
texture2D SourceLinear0 : layout(t1);
sampler SourceLinear0_Sampler : layout(s1)
{
	Texture = SourceLinear0;
	Filter = Linear;
	Address = Wrap;
};
texture2D tColorGradingMap : layout(t2);
sampler sColorGradingSampler : layout(s2)
{
	Texture = tColorGradingMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
// ColorGradingMerge
texture2D tColorGradingMapA : layout(t0);
sampler sColorGradingSamplerA : layout(s0)
{
	Texture = tColorGradingMapA;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tColorGradingMapB : layout(t1);
sampler sColorGradingSamplerB : layout(s1)
{
	Texture = tColorGradingMapB;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tDiyMap : layout(t3);
sampler sDiySampler : layout(s3)
{
	Texture = tDiyMap;
	Filter = Linear;
	Address = Clamp;
};

//HDR
texture2D tLumMap : layout(t4);
sampler sLumSampler : layout(s4)
{
	Texture = tLumMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tDistortionMap : layout(t8);
sampler sDistortionSampler : layout(s8)
{
	Texture = tDistortionMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tUIMap : layout(t9);
sampler sUIMapSampler : layout(s9)
{
	Texture = tUIMap;
	Filter = Linear;
	Address = Clamp;
};
texture2D GlitchyTexture : layout(t10);
sampler sGlitchySampler : layout(s10)
{
	Texture = GlitchyTexture;
	Filter = Linear;
	Address = Mirror;
	MaxMipLevel = 0;
};
texture2D RainDroppTexture  : layout(t11);
sampler sRainDroppSampler : layout(s11)
{
	Texture = RainDroppTexture;
	Filter = Linear;
	Address = Mirror;
	MaxMipLevel = 0;
};
texture2D RainDroppFlowMap  : layout(t12);
sampler sRainFlowSampler : layout(s12)
{
	Texture = RainDroppFlowMap;
	Filter = Linear;
	Address = Mirror;
	MaxMipLevel = 0;
};
texture2D DamageTexture  : layout(t13);
sampler sDamageSampler : layout(s13)
{
	Texture = DamageTexture ;
	Filter = Linear;
	Address = Mirror;
	MaxMipLevel = 0;
};
texture2D tSSGIMap : layout(t5) = "White";
sampler sSSGISampler : layout(s5)
{
	Texture = tSSGIMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlurredTexture : layout(t6);
sampler sBlurredTexture : layout(s6)
{
	Texture = tBlurredTexture;
	Filter = Linear;
	Address = Clamp;
};
//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

/// Pixel output.
struct PixelOutput
{
	float4		Color			: SV_Target;
};
#define S(a, b, t) smoothstep(a, b, t) 
float Saw(float b,float t)
{
    return S(0, b, t)* S(1, b, t);
}

float3 N13(float p) 
{
  float3 p3 = frac(float3(p, p, p) * float3(0.1031, 0.11369, 0.13787));
  //return float3(frac(p*0.1031),0,0);
  p3 += dot(p3, p3.yzx + 19.19);
  return frac(float3((p3.x + p3.y)*p3.z, (p3.x + p3.z)*p3.y, (p3.y + p3.z)*p3.x));
}

 float4 N14(float t) 
{
	return frac(sin(t*float4(123., 1024., 1456., 264.))*float4(6547., 345., 8799., 1564.));
} 

float N(float t) 
{
    return frac(sin(t*12345.564)*7658.76);
}

float StaticDrops(float2 Suv, float t)
{
  Suv *= 20;  //slice 10 is count.
  
  float2 id = floor(Suv); // 1 to 10
  //return id.x;
  Suv = frac(Suv) - 0.5;  // uv is -0.5 to 0.5 of sub region
  
  float3 n = N13(id.x*107.45 + id.y*3543.654); //random value

  float2 p = (n.xy - 0.5-(1-1))*0.1; 
  
  float d = length(Suv - p);
  //test
  //return p.x;pc
  float fade = Saw(0.025, frac(t + n.z))*0.5*cBloodRate; //Calculate transparency disappearance and generation
  
  float c = S(0.6*(n.x+0.2), 0, d)*frac(n.z*10)*fade; //0.3 is size of raindrop
  return c;
}

float DropLayer2(float2 Suv, float t) 
{
      Suv *= 1;
    float2 UV = Suv;
	t*=0.1;
    //return Suv.y;
    Suv.y += t*0.75;
    float2 a = float2(6, 2);
    a = float2(2, 1); //size
    float2 grid = a*2;
    
    float2 id = floor(Suv*grid);
    
    float colShift = N(id.x);  
    Suv.y += colShift;
    
    id = floor(Suv*grid);
    
    float3 n = N13(id.x*35.2+id.y*2376);
    float2 st = frac(Suv*grid) - float2(0.5, 0);
      
    float x = n.x-0.5;  
    float y = UV.y*20;
    
    float wiggle = sin(y+sin(y));
    //return float2(wiggle,0);
    x += wiggle*(.5-abs(x))*(n.z-.5);
    x *= 0.7;
    float ti = frac(t+n.z);
    y = (Saw(.85, ti)-.5)*.9+.5;
    float2 p = float2(x, y);
    
    float d = length((st-p)*a.yx);
    
    float mainDrop = S(0.25,0.0, d); //maindrop size
    
    float r = sqrt(S(1.0, y, st.y));
    float cd = abs(st.x-x);
    //float trail = S(0.23*r, 0.15*r*r, cd);
    float trailFront = S(-0.02, 0.02, st.y-y);

    y = UV.y;
    float trail2 = S(.2*r, .0, cd);
    float droplets = max(0, (sin(y*(1.-y)*120.)-st.y))*trail2*trailFront*n.z;
 
    y = frac(y*10)+(st.y-0.5);
    float dd = length(st-float2(x, y*0.7));

    droplets = S(0.2, 0., dd);     //droplets size  
    
    float m = mainDrop+droplets*r*trailFront;
    //m = droplets;
    //m += st.x>a.y*.45 || st.y>a.x*.165 ? 1.2 : 0.;
  
    //m = droplets;
    //m += st.x>a.y*.45 || st.y>a.x*.165 ? 1.2 : 0.;
    return m*m;
}

float Drops(float2 Suv, float t) 
{
    float s = StaticDrops(Suv*0.2, t)*4; 
   
    float m1 = 0;
if(@_LOD_LEVEL == LOD_Level_1)
{
m1 = DropLayer2(Suv, t);
}
	 
	
	 
    //return m1;  
    //s = 0;
    //m1 = 0;
    float c = s;
   
    c = S(0.2, 1.5, c)*1;
	c*=c; //0.3 is size of raindrop   
    return c+m1;
}
// @param InLDRColor in gamma space, has to be in 0..1 range
half3 ColorLookupTable(half3 InLDRColor)
{
	// Requires a volume texture 16x16x16 unwrapped in a 2d texture 256x16.
	half2 Offset0 = half2( 0.5 / 256.0, 0.5 / 16.0);
	half2 Offset1 = half2(16.5 / 256.0, 0.5 / 16.0);
	float Scale = 15.0 / 16.0; 

	// Also consider blur value in the blur buffer written by translucency.
	half IntB = floor(InLDRColor.b * 15.0);
	half FracB = InLDRColor.b * 15.0 - IntB;

	float U = (IntB + InLDRColor.r * Scale) / 16.0;
	float V = InLDRColor.g * Scale;

	half3 RG0 = tColorGradingMap.SampleLevel(sColorGradingSampler, Offset0 + float2(U, V), 0).rgb;
	half3 RG1 = tColorGradingMap.SampleLevel(sColorGradingSampler, Offset1 + float2(U, V), 0).rgb;

	return lerp(RG0, RG1, FracB);
}

// @param LinearColor in Rec.709 primaries.
half3 ColorCorrect( half3 LinearColor)
{
	half3 WorkingColor = SRGB2AP1(LinearColor);

	float3 AP1_RGB2Y = float3(0.2722287168,0.6740817658,0.0536895174);

	float Luma = dot( WorkingColor, AP1_RGB2Y);
	WorkingColor = max( 0, lerp( Luma.rrr, WorkingColor, cSaturation.xyz*cSaturation.w ) );
	WorkingColor = pow( WorkingColor * (1.0 / 0.18), cContrast.xyz*cContrast.w ) * 0.18;
	WorkingColor = pow( WorkingColor, 1.0 / (cGamma.xyz*cGamma.w) );
	WorkingColor = WorkingColor * (cGain.xyz * cGain.w) + (cOffset.xyz + cOffset.w);
	return AP12SRGB(WorkingColor);
}

// Accurate for 4000K < Temp < 25000K
// in: correlated color temperature
// out: CIE 1931 chromaticity
float2 D_IlluminantChromaticity( float Temp )
{
	// Correct for revision of Plank's law
	// This makes 6500 == D65
	Temp *= 1.4388 / 1.438;

	float x =	Temp <= 7000 ?
				0.244063 + ( 0.09911e3 + ( 2.9678e6 - 4.6070e9 / Temp ) / Temp ) / Temp :
				0.237040 + ( 0.24748e3 + ( 1.9018e6 - 2.0064e9 / Temp ) / Temp ) / Temp;
	
	float y = -3 * x*x + 2.87 * x - 0.275;

	return float2(x,y);
}

// Accurate for 1000K < Temp < 15000K
// [Krystek 1985, "An algorithm to calculate correlated colour temperature"]
float2 PlanckianLocusChromaticity( float Temp )
{
	float u = ( 0.860117757f + 1.54118254e-4f * Temp + 1.28641212e-7f * Temp*Temp ) / ( 1.0f + 8.42420235e-4f * Temp + 7.08145163e-7f * Temp*Temp );
	float v = ( 0.317398726f + 4.22806245e-5f * Temp + 4.20481691e-8f * Temp*Temp ) / ( 1.0f - 2.89741816e-5f * Temp + 1.61456053e-7f * Temp*Temp );

	float x = 3*u / ( 2*u - 8*v + 4 );
	float y = 2*v / ( 2*u - 8*v + 4 );

	return float2(x,y);
}

float2 PlanckianIsothermal( float Temp, float Tint )
{
	float u = ( 0.860117757f + 1.54118254e-4f * Temp + 1.28641212e-7f * Temp*Temp ) / ( 1.0f + 8.42420235e-4f * Temp + 7.08145163e-7f * Temp*Temp );
	float v = ( 0.317398726f + 4.22806245e-5f * Temp + 4.20481691e-8f * Temp*Temp ) / ( 1.0f - 2.89741816e-5f * Temp + 1.61456053e-7f * Temp*Temp );

	float ud = ( -1.13758118e9f - 1.91615621e6f * Temp - 1.53177f * Temp*Temp ) / Square( 1.41213984e6f + 1189.62f * Temp + Temp*Temp );
	float vd = (  1.97471536e9f - 705674.0f * Temp - 308.607f * Temp*Temp ) / Square( 6.19363586e6f - 179.456f * Temp + Temp*Temp );

	float2 uvd = normalize( float2( u, v ) );

	// Correlated color temperature is meaningful within +/- 0.05
	u += -uvd.y * Tint * 0.05;
	v +=  uvd.x * Tint * 0.05;
	
	float x = 3*u / ( 2*u - 8*v + 4 );
	float y = 2*v / ( 2*u - 8*v + 4 );

	return float2(x,y);
}

// Transformations between CIE XYZ tristimulus values and CIE x,y 
// chromaticity coordinates
float3 XYZ_2_xyY( float3 XYZ )
{  
	float3 xyY;
	float divisor = (XYZ[0] + XYZ[1] + XYZ[2]);
	if (divisor == 0.) divisor = 1e-10;
	xyY[0] = XYZ[0] / divisor;
	xyY[1] = XYZ[1] / divisor;  
	xyY[2] = XYZ[1];
  
	return xyY;
}

float3 xyY_2_XYZ( float3 xyY )
{
	float3 XYZ;
	XYZ[0] = xyY[0] * xyY[2] / max( xyY[1], 1e-10);
	XYZ[1] = xyY[2];  
	XYZ[2] = (1.0 - xyY[0] - xyY[1]) * xyY[2] / max( xyY[1], 1e-10);

	return XYZ;
}

float3x3 ChromaticAdaptation( float2 src_xy, float2 dst_xy )
{
	// Von Kries chromatic adaptation 
	static const float3x3 ConeResponse =
	{
		 0.8951,  0.2664, -0.1614,
		-0.7502,  1.7135,  0.0367,
		 0.0389, -0.0685,  1.0296
	};
	static const float3x3 InvConeResponse =
	{
		 0.9869929, -0.1470543,  0.1599627,
		 0.4323053,  0.5183603,  0.0492912,
		-0.0085287,  0.0400428,  0.9684867
	};

	float3 src_XYZ = xyY_2_XYZ( float3( src_xy, 1 ) );
	float3 dst_XYZ = xyY_2_XYZ( float3( dst_xy, 1 ) );

	float3 src_coneResp = mul( ConeResponse, src_XYZ );
	float3 dst_coneResp = mul( ConeResponse, dst_XYZ );

	return mul( InvConeResponse, mul( float3x3(dst_coneResp[0] / src_coneResp[0], 0.0, 0.0,0.0, dst_coneResp[1] / src_coneResp[1], 0.0,0.0, 0.0, dst_coneResp[2] / src_coneResp[2]), ConeResponse ) );
}

half4 RadiusBlur(half2 UV ,half BlurRadius,half Iteriate )
{
	float2 blurVector = (half2(0.5,0.5) - UV) * BlurRadius;
	
	half4 acumulateColor = half4(0, 0, 0, 0);
	
	[unroll(30)]
	for (int j = 0; j < Iteriate; j ++)
	{
		acumulateColor += SampleColorMap(SourceLinear0, SourceLinear0_Sampler,UV);
		UV += blurVector;
	}
	
	return acumulateColor /(Iteriate+1);
}

float3 WhiteBalance( float3 LinearColor )
{
	float2 SrcWhiteDaylight = D_IlluminantChromaticity( cWhiteTemp );
	float2 SrcWhitePlankian = PlanckianLocusChromaticity( cWhiteTemp );

	//float factor = step(cWhiteTemp, 4000);
	float2 SrcWhite = cWhiteTemp < 4000 ? SrcWhitePlankian : SrcWhiteDaylight;
	float2 D65White = float2( 0.31270,  0.32900 );

	{
		// Offset along isotherm
		float2 Isothermal = PlanckianIsothermal( cWhiteTemp, cWhiteTint.x) - SrcWhitePlankian;
		SrcWhite += Isothermal;
	}

	static const float3x3 sRGB_2_XYZ_MAT =
	{
		0.4124564, 0.3575761, 0.1804375,
		0.2126729, 0.7151522, 0.0721750,
		0.0193339, 0.1191920, 0.9503041
	};

	static const float3x3 XYZ_2_sRGB_MAT =
	{
		3.2409699419, -1.5373831776, -0.4986107603,
		-0.9692436363,  1.8759675015,  0.0415550574,
		0.0556300797, -0.2039769589,  1.0569715142
	};

	float3x3 WhiteBalanceMat = ChromaticAdaptation( SrcWhite, D65White );
	WhiteBalanceMat = mul( XYZ_2_sRGB_MAT, mul( WhiteBalanceMat, sRGB_2_XYZ_MAT ) );

	return mul( WhiteBalanceMat, LinearColor );
}

// Glitchy for sizhuang
float NoiseGen2(float2 UV,float2 Size,float ChangeRate,float time)
{
	float AcTime = floor(time / ChangeRate) * ChangeRate;
	UV = float2(floor(UV.x * Size.x ) / Size.x , floor(UV.y * Size.y ) / Size.y);
	UV += float2(AcTime , AcTime / 2 );
	float2 ScreenPos = UV * ScreenInfo.xy;
	float rst = noise(ScreenPos.x + ScreenPos.y);
	return rst;
}
float NoiseGen5(float2 UV,float2 Size,float ChangeRate,float Power,float time)
{
	float AcTime = floor(time / ChangeRate) * ChangeRate;
	UV = float2(floor(UV.x * Size.x ) / Size.x , floor(UV.y * Size.y ) / Size.y);
	float CenterFix = saturate(Power * pow(length(UV - float2(0.5,0.5)),Power));
	UV += float2(AcTime, AcTime / 2);
	float2 ScreenPos = UV * ScreenInfo.xy;
	float rst = noise(ScreenPos.x + ScreenPos.y);
	return rst*CenterFix;
}
float NoiseGenSin2(float2 UV,float4 Size,float Speed,float time)
{
	float SinRst1 = saturate(sin( Size.x * UV.y + floor(Speed * time * Size.y)/Size.y));
	SinRst1 -= saturate(sin( (Size.x * UV.y+Size.w*PI) + floor(Speed * time * Size.y)/Size.y));
	SinRst1 = floor(SinRst1 * Size.z) / Size.z;
	return saturate(SinRst1);
}
float2 NoiseGenUV(float2 UV,float2 Size,float ChangeRate,float Power,float time)
{
	float2 textureTileSize = cGlitchTextureTileCount;
	UV = frac(UV*Size);
	float rst = noise(floor(time/ChangeRate)*ChangeRate);
	UV = UV/textureTileSize + float2(floor(rst*textureTileSize.x)/textureTileSize.x,floor(rand(rst)*textureTileSize.y)/textureTileSize.y);
	return UV;
}

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;

	if (@_VULKAN)
	{
		if (cIsMainViewport > 0.9)
		{
			float2 t = (OUT.TC - 0.5) * 2.0;
			OUT.TC.x = dot(t, UVTransform.xy);
			OUT.TC.y = dot(t, UVTransform.zw);
			OUT.TC = OUT.TC * 0.5 + 0.5;
		}
	}
	return OUT;
}

/////////////////// ColorGradingMerge ////////////////////
float4 ColorGradingMergePS(VertexScreenOutput IN) : SV_Target
{
	half3 A = tColorGradingMapA.Sample(sColorGradingSamplerA, IN.TC).rgb;
	half3 B = tColorGradingMapB.Sample(sColorGradingSamplerB, IN.TC).rgb;
	return lerp(A, B, cLocalRate).rgbb;
}

technique ColorGradingMerge
{
	VertexShader = ScreenVS();
	PixelShader = ColorGradingMergePS();
}

float3 Rec709ToRec2020(float3 v)
{
	return float3(dot(float3(0.6274038959, 0.3292830384, 0.0433130657), v),
				  dot(float3(0.0690972894, 0.9195403951, 0.0113623156), v),
				  dot(float3(0.0163914389, 0.0880133079, 0.8955952532), v));
}

float3 ApplyREC2084Curve(float3 L)
{
	float m1 = 2610.0 / 4096.0 / 4;
	float m2 = 2523.0 / 4096.0 * 128;
	float c1 = 3424.0 / 4096.0;
	float c2 = 2413.0 / 4096.0 * 32;
	float c3 = 2392.0 / 4096.0 * 32;
	float3 Lp = pow(L, m1);
	return pow((c1 + c2 * Lp) / (1 + c3 * Lp), m2);
}

//////////////////////////////////////////////////////////////////////////
// Texture to screen
#define MOTION_SAMPLES 16
PixelOutput TextureToScreenPS(in VertexScreenOutput IN)
{
	PixelOutput OUT;
	
	if (@_MotionBlurEnable)
	{
		float2 Center = cBlurCenterUV;
		float2 Motiondir = (IN.TC - Center);
		float MotiondirMask = saturate(dot(Motiondir, Motiondir));
		//Motiondir = normalize(Motiondir) * dot(Motiondir, Motiondir);
		MotiondirMask *= cBlurPercent / MOTION_SAMPLES;
		float3 Scr = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC).rgb;

		float3 Result = float3(0, 0, 0);
		//float2 itrUV = IN.TC;
		for (int i = 0; i < MOTION_SAMPLES; ++i)
		{
			//itrUV += Motiondir;
			float3 Source = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, Motiondir*(1- i*MotiondirMask)+Center).rgb;
			Result += Source;
		}
		Result = Result/MOTION_SAMPLES;
		Result = lerp(Scr, Result, cBlurIntensity);
		OUT.Color = float4(Result, 1.0);
	}
	else if (@_FullSize)
	{
		OUT.Color = SampleColorMap(SourcePoint0, SourcePoint0_Sampler, IN.TC);
	}
	else if (@_LinearFilter)
	{
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC);
	}
	else
	{
#if 1	// Modified bilinear interpolation
		float2 p = IN.TC * ScreenInfo.xy + 0.5;
		float2 i = floor(p);
		float2 f = p - i;

		//f = f * f * (3 - 2 * f);						// 3rd order
		f = f * f * f * (f * (f * 6 - 15) + 10);		// 5th order

		p = i + f;
		float2 tc = (p - 0.5) * ScreenInfo.zw;
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, tc);

#else	// Bicubic interpolation
		float2 uv = IN.TC * ScreenInfo.xy;
		float2 texelCenter = floor(uv - 0.5) + 0.5;
		float2 fracOffset = uv - texelCenter;
		float2 fracOffset_x2 = fracOffset * fracOffset;
		float2 fracOffset_x3 = fracOffset * fracOffset_x2;
	
		// Calculate the filter weights (B-Spline weighting function)
		float2 weight0 = fracOffset_x2 - 0.5 * (fracOffset_x3 + fracOffset);
		float2 weight1 = 1.5 * fracOffset_x3 - 2.5 * fracOffset_x2 + 1.0;
		float2 weight3 = 0.5 * (fracOffset_x3 - fracOffset_x2);
		float2 weight2 = 1.0 - weight0 - weight1 - weight3;
	
		// Calculate the texture coordinates
		float2 scalingFactor0 = weight0 + weight1;
		float2 scalingFactor1 = weight2 + weight3;
	
		float2 f0 = weight1 / (weight0 + weight1);
		float2 f1 = weight3 / (weight2 + weight3);
	
		float2 texCoord0 = texelCenter - 1.0 + f0;
		float2 texCoord1 = texelCenter + 1.0 + f1;
	
		texCoord0 *= ScreenInfo.zw;
		texCoord1 *= ScreenInfo.zw;
	
		// Sample the texture (4 taps)
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord0.x, texCoord0.y)) * (scalingFactor0.x * scalingFactor0.y)
			      + SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord1.x, texCoord0.y)) * (scalingFactor1.x * scalingFactor0.y)
			      + SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord0.x, texCoord1.y)) * (scalingFactor0.x * scalingFactor1.y)
			      + SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord1.x, texCoord1.y)) * (scalingFactor1.x * scalingFactor1.y);
#endif
	}

	if(@_RainDropEnable)
	{
		float2 st = IN.TC;
		float2 Suv = float2((st.x*2-1),(1- st.y)-0.5); //x is -1 to 1
		Suv.y +=GameTime*0.01;
		float mask = length(IN.TC-0.5);
		mask *=mask*3;
		float2 Luv = st.xy;
		float t = GameTime*0.2;  //0.1 is deop speed
		float2 e = float2(0.002,0.0);
		//expensive normals
		float c = Drops(Suv, t);
		//return float4(c,0,0,1);
		float cx = Drops(Suv+e.xy, t);
		float cy = Drops(Suv+e.yx, t);
		float2 n = float2(cx-c, cy-c); 
		IN.TC = Luv+n;  
		OUT.Color = lerp(OUT.Color,SampleColorMap(tBlurredTexture, sBlurredTexture,IN.TC)+c*0.2,saturate(c*mask)) ;
		// OUT.Color = mask.xxxx;
	}
	if(@_bloody)
	{
		float2 st = IN.TC;
		float2 Suv = float2((st.x*2-1),(1- st.y)-2.7); //x is -1 to 1
		Suv.y +=GameTime*0.01;
		float mask = length(IN.TC-0.5);
		mask *=mask*5*mask;
		float t = GameTime*0.2;  //0.1 is deop speed
		
		//expensive normals
		float c = Drops(Suv, t);
		//return float4(c,0,0,1);
	
		
		OUT.Color.rgb = lerp(OUT.Color.rgb,OUT.Color.rgb*half3(1,0,0),saturate(c*mask)) ;
		// OUT.Color = mask.xxxx;
	}

	if(@_DiyLocalBlurEnable)
	{
		float blurPosX = (cPlayerScreenRect.x + cPlayerScreenRect.z) * 0.5;
		float blurPosY = (cPlayerScreenRect.y + cPlayerScreenRect.w) * 0.5;
		float rangeWidth = (cPlayerScreenRect.z - cPlayerScreenRect.x) * 0.5 * cDiyLocalBlurParam.z;
		float rangeHeight = (cPlayerScreenRect.w - cPlayerScreenRect.y) * 0.5 * cDiyLocalBlurParam.w;
		float3 result = float3(0, 0, 0);
		float2 itrUV = IN.TC;

		if(abs(itrUV.x - blurPosX) < rangeWidth && abs(itrUV.y - blurPosY) < rangeHeight)
		{
			float2 motionDir =  cDiyLocalBlurParam.xy;
			float x = IN.TC.x - blurPosX;
			float y = IN.TC.y - blurPosY;
			motionDir *= 0.02 * cDiyLocalBlurPercent / MOTION_SAMPLES;
			[unroll]
			for (int i = 0; i < MOTION_SAMPLES; ++i)
			{
				itrUV += motionDir;
				float3 Source = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, itrUV).rgb;
				result += Source;
			}
			float factor = saturate((x + rangeWidth) * (rangeWidth - x) * (rangeHeight - y) * (rangeHeight + y) / (rangeWidth * rangeWidth * rangeHeight * rangeHeight));
			OUT.Color.rgb = lerp(OUT.Color.rgb, result / MOTION_SAMPLES, factor);
		}
	}

	// {HEX} Vignetting
	if (@_VignettingEnable)
	{
		half2 offset = IN.TC.xy - half2(0.5, 0.5);
		half Vignetting = max(0, cVignettingIntensity * dot(offset, offset));
		//Vignetting *= Vignetting;
		half percent = lerp(0, Vignetting, cVignettingPercent);
		float3 Result = lerp(OUT.Color.rgb, cVignettingColor, saturate(percent));
		OUT.Color.rgb = Result;
	}

	// {HEX} end
	if(@_DiySickBlurEnable)
	{
		float2 center = float2(0.5, 0.5);
		float2 outdir = (IN.TC - center);
		float distortion_degree = saturate(dot(outdir, outdir)) * cDiySickBlurParam.x * cDiySickBlurPercent;
		float degree = (1 - smoothstep(0.6, 0.8, 1 - length(outdir))) * cDiySickBlurParam.y;
		float2 new_center = outdir * (1 - distortion_degree) + center;

		float r = SampleColorMap(tBlurredTexture, sBlurredTexture, new_center + float2(0.01, 0) * cDiySickBlurParam.z).r;
		float g = SampleColorMap(tBlurredTexture, sBlurredTexture, new_center).g;
		float b = SampleColorMap(tBlurredTexture, sBlurredTexture, new_center + float2(0, 0.01) * cDiySickBlurParam.z).b;
		
		float3 origin = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, new_center).rgb;

		OUT.Color.rgb = lerp(origin, lerp(origin, float3(r, g, b), degree) * cDiySickBlurColor.rgb, cDiySickBlurPercent);
	}

	[@_OutputHDRDisplay]
	OUT.Color.rgb = DecodeHDR(OUT.Color.rgb);

	// <-- SDR Display: gamma space, HDR Display: linear space.

	if (@_ColorCorrectionEnable)
	{
		// TODO: color correction/grading should be applied in scene-referrd space (before tone mapping).
		[!@_OutputHDRDisplay]
		OUT.Color.rgb = DecodeGamma(OUT.Color.rgb);

		OUT.Color.rgb = WhiteBalance(OUT.Color.rgb);
		OUT.Color.rgb = ColorCorrect(OUT.Color.rgb);

		[!@_OutputHDRDisplay]
		OUT.Color.rgb = EncodeGamma(OUT.Color.rgb);
	}

	// <-- SDR Display: gamma space, HDR Display: linear space.

	if (@_OutputHDRDisplay)
	{
		OUT.Color.rgb = EncodeGamma(ACESFilm(OUT.Color.rgb));
	}

	// <-- SDR Display: gamma space, HDR Display: gamma space.

	if (@_ColorGradingEnable)
	{
		OUT.Color.rgb = ColorLookupTable(saturate(OUT.Color.rgb));
	}

	if (@_GrayEnable)
	{
		half d = GetLuminance(OUT.Color.rgb);
		half dd = d * d;
		half3 Result = (cMaskColor*(d-dd)+dd);	
		OUT.Color.rgb = lerp(OUT.Color.rgb, Result, cGrayPercent);
	}


	if (@_OutputHDRDisplay)
	{
		// TODO: fold gamma encoding/decoding into fitted curve.
		half3 sceneColor = ACESFilmInverse(DecodeGamma(OUT.Color.rgb));
		sceneColor = ACES_Tone_1000nits(sceneColor);

		half uiMaxBrightness = 203;	// Rec.2010
		half4 uiColorAlpha = tUIMap.Sample(sUIMapSampler, IN.TC);
		half3 uiColor = pow(uiColorAlpha.rgb, 2.4) * uiMaxBrightness;
		sceneColor = lerp(sceneColor, uiColor, uiColorAlpha.a);
		half3 sceneColorBt2020 = Rec709ToRec2020(sceneColor);
		OUT.Color.rgb = ApplyREC2084Curve(sceneColorBt2020 / 10000.0);
	}

	// <-- SDR Display: sRGB space, HDR Display: BT.2020 PQ space.

	if(@_GlitchyEnable)
	{
		float Controller = 1-cIntensity;
		const float TextureAnimPlayDurationDefault = 8.0;
		//Distortion Layer
		float DistortionAera01 = NoiseGen5(IN.TC, float2(8,16), 0.05, 2, GameTime);
		float2 DistortionOffset = (1-Controller) * step(Controller/2,DistortionAera01) * 0.25 * DistortionAera01 * float2((DistortionAera01),rand(DistortionAera01-1));
		float NoiseGenSinAera4 = (1-Controller) * NoiseGenSin2(IN.TC, float4(5,0.5,2,0.4), 30, GameTime);
		DistortionOffset += 0.25 * NoiseGenSinAera4 * float2(rand(NoiseGenSinAera4),rand(NoiseGenSinAera4-1));
		float4 Rst = SampleColorMap(SourcePoint0, SourcePoint0_Sampler, IN.TC + (cDistortionIntensity*1.05-0.1) * DistortionOffset);

		float2 TextureAera = NoiseGenUV(IN.TC + DistortionOffset, float2(1,1), 0.17, 2, GameTime);
		float4 TRst = SampleColorMap(GlitchyTexture, sGlitchySampler, TextureAera.xy);
		Rst += TRst * cTextureAnimIntensity * saturate(TextureAnimPlayDurationDefault*(1-Controller));

		//Color Shift Layer
		float ColorShiftAera02 = NoiseGen5(IN.TC, float2(10,18), 0.3, 1, GameTime);
		ColorShiftAera02 = min(ColorShiftAera02, NoiseGen2(IN.TC, float2(19,38), 0.1, GameTime));
		Rst = lerp(Rst,saturate((Rst - 0.55) * 10.0 - 2.35 ) ,step(Controller,ColorShiftAera02 * cColorShiftIntensity));

		//Clip Layer
		float ClipAera01 = NoiseGen5(IN.TC, float2(10,16), 0.035, 2, GameTime);
		ClipAera01 = min(ClipAera01, NoiseGen2(IN.TC, float2(16,28), 0.3, GameTime));
		float ClipValue = step(Controller, ClipAera01); 

		OUT.Color = lerp(Rst,float4(0,0,0,1),step(Controller,ClipValue * cClipIntensity)); 

		return OUT;
	}

	if(@_TrembleEnable)
	{
	half2 Tremble = (abs(sin(GameTime*2))+0.5)*0.012*cTrembleRate ;
	OUT.Color.rgb/=2;
	OUT.Color.rgb+=  RadiusBlur(IN.TC,Tremble ,2);
	}

	if(@_DamageEnable) 
	{
		half4 blood = SampleColorMap(DamageTexture, sDamageSampler, IN.TC);
		half damage  = cDamageRate;
		half3 darkConer =saturate((1-distance(IN.TC,half2(0.5,0.5))*damage*0.1)*1.3 ) ;
		// darkConer *= darkConer	;
		darkConer = lerp(half3(1.0,0.0,0),1,darkConer.r);
		//damage range from  0-10;
		float timeTricker =saturate( abs(sin(GameTime*2.3))) ;	
		
		blood*=0.4*damage ;
		blood.g *=0.5;
		blood.b *=0.25;
		blood.rgb =1-blood.rgb;
		half alpha = saturate(blood.r*blood.g*blood.b) ;
		half3 bloodColor = half3(1.0,0.0,0);
		half2 distorbUV = IN.TC+alpha.rr*half2(0.2,0.1);
		half3 distorbScreen = SampleColorMap(SourceLinear0, SourceLinear0_Sampler,distorbUV);
		bloodColor = bloodColor*distorbScreen;	
	
	
		OUT.Color.rgb =lerp(bloodColor , OUT.Color.rgb, alpha*darkConer) ;

	
	}


	// OUT.Color = SourcePoint0.SampleLevel(SourcePoint0_Sampler, TC.rg,8)+TC.b*0.02*timePreiod;
	if(@_FlashEnable)
	{
		half flash =cflashRate;
		OUT.Color.rgb = OUT.Color.rgb + flash.rrr;
	}
	

	
		// float2 Center = float2((IN.TC.x+0.2),(IN.TC.y-0.0));	
	if(@_QuiverEnable)
	{
		float2 Center = float2(0.0,0.0);	
		float3 Result = float3(0, 0, 0);
		//float2 itrUV = IN.TC;
		for (int i = 0; i < 3; ++i)
		{	

			Center += float2(0.01,0.0)*cQuiverIntensity;
			
			//itrUV += Motiondir;
			float3 Source = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC+Center).rgb;
			Result += Source *sin(IN.TC.x*3.14);
		}
		Result = Result/3;
		Result = OUT.Color.rgb*0.6+Result*0.4;
		
		OUT.Color = float4(Result, 1.0);

	}

	/*if(@_GIMap)
	{
		float4 gi = tSSGIMap.Sample(sSSGISampler, IN.TC);
		OUT.Color.rgb = float3(gi.a, gi.a, gi.a);
		//OUT.Color.rgb = gi.rgb;
		return OUT;
	}*/
	/*if(@_EnableDebugMap)
	{
		float2 uv = IN.TC * 4;
		uv.y *= (ScreenInfo.y / ScreenInfo.x);
		if(uv.x < 1 && uv.y < 1)
		{
			OUT.Color.rgb = tDebugMap.Sample(sDebugMapSampler, uv).rgb;
			return OUT;
		}		
	}*/
	// wenyao: important!!! for mx4 & mac (xqiao)
	OUT.Color.a = saturate(OUT.Color.a + cIsMainViewport);
	OUT.Color.rgb*=cScreenCut;
	return OUT;
}

technique FinalPass
{
	VertexShader = ScreenVS();
	PixelShader = TextureToScreenPS();
}



