//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<ArtText.fx>
///	@path	~/Shaders/UI/
///	@date	2015/06/02
///	@brief	Art text.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
};

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	Default = false;
};

@_DyeingEnable
{
	Name = DyeingEnable;
	Description = "Is the DiffuseTexture dyeing?";
	Default = true;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORDN;
	float2		TexCoord2				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		Color2					: TEXCOORDN;
	float2		TexCoord				: TEXCOORDN;
};


////////////////////////////////////////////////////////////////////////
// Art text draw.

VertexOutput ArtTextDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);

	OUT.Color = GetVertexColor(IN.Color);
	OUT.Color2 = IN.TexCoord;
	OUT.TexCoord = IN.TexCoord2;

	return OUT;
}

float4 ArtTextDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse;
	if (@_MaskTextureEnable)
	{
		float4 c1 = IN.Color;
		float4 c2 = IN.Color2;
		diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord);
		if (@_D3D11 || @_D3D12 || @_VULKAN || @_METAL)
			diffuse.rgba = diffuse.rrrg;
		else
			diffuse.rgba = diffuse.rrra;

		if (length(c1.rgb - c2.rgb) > 0.01)
		{
			float alpha = diffuse.r * c1.a;
			float blend = diffuse.a / max(diffuse.r, 0.004);
			diffuse.rgb = c1.rgb * blend + c2.rgb * ((1.0f - blend) * (1.f - diffuse.a));
			diffuse.a = alpha;
		}
		else
		{
			diffuse.a = diffuse.r * c1.a;
			diffuse.rgb = c1.rgb;
		}
	}
	else
	{
		diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord);

		if (@_DyeingEnable)
		{
			diffuse.rgb = Dyeing(diffuse.rgb, IN.Color.g);
		}
		else
		{
			diffuse.rgb *= IN.Color.rgb;
		}
		diffuse.a *= IN.Color.a;
	}

	return diffuse;
}

technique Shading
{
	VertexShader = ArtTextDrawVS();
	PixelShader = ArtTextDrawPS();
}
