//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<BoardElement.fx>
///	@path	~/Shaders/UI/
///	@date	2015/06/04
///	@brief	Board element.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh" 

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4 DyeColor						 = { 0.0f, 0.0f, 0.0f, 1.0f };
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float2		TexCoord				: TEXCOORD0;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float2		TexCoord				: TEXCOORD0;
};


////////////////////////////////////////////////////////////////////////
// Board image draw.

VertexOutput BoardImageDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;
	
	return OUT;
}

float4 BoardImageDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord);

	diffuse.rgb = saturate(diffuse.rgb + DyeColor.rgb);
	diffuse.a *= DyeColor.a;

	return diffuse;
}


technique Shading
{
	VertexShader = BoardImageDrawVS();
	PixelShader = BoardImageDrawPS();
}
