//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<BoardStrokeText.fx>
///	@path	~/Shaders/UI/
///	@date	2015/06/10
///	@brief	Board element.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4 DyeColor						 = { 1.0f, 1.0f, 1.0f, 1.0f };
	float4 StrokeColor					 = { 0.0f, 0.0f, 0.0f, 1.0f };
};

texture2D TextTexture				: layout(t0) = "Transparent";
sampler_state TextTexture_Sampler	: layout(s0)
{
	Texture = TextTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float2		TexCoord				: TEXCOORD0;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float2		TexCoord				: TEXCOORD0;
};


////////////////////////////////////////////////////////////////////////
// Board text draw.

VertexOutput BoardStrokeTextDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;
	
	return OUT;
}

float4 BoardStrokeTextDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(TextTexture, TextTexture_Sampler, IN.TexCoord);

	if (length(DyeColor.rgb - StrokeColor.rgb) > 0.01)
	{
		float alpha = diffuse.r * DyeColor.a;
		float blend = diffuse.a / max(diffuse.r, 0.004);
		diffuse.rgb = DyeColor.rgb * blend + StrokeColor.rgb * (1.0f - diffuse.a);
		diffuse.a = alpha;
	}
	else
	{
		diffuse.rgb = DyeColor.rgb;
	}

	return diffuse;
}

technique Shading
{
	VertexShader = BoardStrokeTextDrawVS();
	PixelShader = BoardStrokeTextDrawPS();
}
