//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<SpineElement.fx>
///	@path	~/Shaders/UI/
///	@date	2019/08/29
///	@brief	Spine element.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_SpineDissolving
{
	Name = SpineDissolving;
	Description = "SpineDissolving";
	Default = false;
	Visible = true;
}

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

@_DepthWriteEnable
{
	Name = DepthWriteEnable;
	Description = "Do we have DepthWrite?";
	Default = false;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float4		Normal 					: NORMAL;
	float2		TexCoord				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		DarkColor				: TEXCOORDN;
	float2		TexCoord				: TEXCOORDN;
};

////////////////////////////////////////////////////////////////////////
// Spine element draw.

VertexOutput SpineElementDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(ViewProj, float4(IN.Position.xyz, 1.0));
	if (@_DepthWriteEnable)
	{
		OUT.HPosition.z -= 0.001;
	}
	OUT.TexCoord = IN.TexCoord;
	OUT.Color = GetVertexColor(IN.Color);
	OUT.DarkColor = IN.Normal;

	return OUT;
}

float4 SpineElementDrawPS(in VertexOutput IN) : SV_Target
{
	if (@_DepthWriteEnable)
	{
		float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord);
		clip(diffuse.a - 0.01);
		return float4(0, 0, 0, 0);
	}
	else
	{
		float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord);

		float4 result = diffuse * IN.Color;
		if (@_SpineDissolving) 
		{
			float clipValue = result.a - GetRandom(IN.HPosition.xy);
			clip(clipValue);
			result.a = diffuse.a;
		}
		result.rgb += IN.DarkColor.rgb;
		return result;
	}
}

technique Shading
{
	VertexShader = SpineElementDrawVS();
	PixelShader = SpineElementDrawPS();
}
