//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIBaseElement.fx>
///	@path	~/Shader/UI/
///	@date	2017/02/27
///	@brief	Translucent base UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
};

@_Norm2Enable
{
	Name = Norm2Enable;
	Description = "Is norm2 enable?";
	Default = true;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent base UI draw.

VertexOutput UIBaseElementDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.zw = IN.TexCoord2;
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UIBaseElementDrawPS(in VertexOutput IN) : SV_Target
{
	float alpha = 1.f;
	if (@_Norm2Enable)
	{
		alpha = smoothstep(-IN.TexCoord.z, IN.TexCoord.z, 1.0 - IN.TexCoord.z - length(IN.TexCoord.xy)) * max(4.0f * smoothstep(-IN.TexCoord.z, IN.TexCoord.z, length(IN.TexCoord.xy) + IN.TexCoord.z - 1.0), IN.TexCoord.w);
	}
	else
	{
		alpha = smoothstep(-IN.TexCoord.z, IN.TexCoord.z, 1.0 - IN.TexCoord.z - abs(IN.TexCoord.x)) * smoothstep(-IN.TexCoord.w, IN.TexCoord.w, 1.0 - IN.TexCoord.w - abs(IN.TexCoord.y));
	}

	float4 diffuse = IN.Color;
	diffuse.a *= alpha;

	return diffuse;
}

technique Shading
{
	VertexShader = UIBaseElementDrawVS();
	PixelShader = UIBaseElementDrawPS();
}
