//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact :
//	Author: xqiao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UCameraImage.fx>
///	@path	~/Shader/UI/
///	@date	2019/12/30
///	@brief	Camera image

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	AffectedTex = MaskTexture;
	Default = false;
};

@_GrayingEnable
{
	Name = GrayingEnable;
	Description = "Is the result graying?";
	Default = false;
};

cbuffer Batch
{
	float4x4 WorldViewProj		 		: layout(c0, vs[VsBeg + 0]);
	float4 UVTransform					: layout(c4, vs[VsBeg + 4]) = { 1.0f, 0.0f, 0.0f, 1.0f };
	float4x3 ColorConversion			: layout(c5, ps[PsBeg + 0]);
	float EffectParam					: layout(c8.x, ps[PsBeg + 3].x) = 0.0;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture					: layout(t1) = "White";
sampler_state MaskTexture_Sampler		: layout(s1)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D AuxiliaryTexture				: layout(t2) = "Transparent";
sampler_state AuxiliaryTexture_Sampler	: layout(s2)
{
	Texture = AuxiliaryTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent base UI draw.

VertexOutput UICameraDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	
	OUT.TexCoord.xy = IN.TexCoord;

	{
		float2 t = (OUT.TexCoord.xy - 0.5) * 2.0;
		OUT.TexCoord.x = dot(t, UVTransform.xy);
		OUT.TexCoord.y = dot(t, UVTransform.zw);
		OUT.TexCoord.xy = OUT.TexCoord.xy * 0.5 + 0.5;
	}
	
	if (@_MaskTextureEnable)
	{
		OUT.TexCoord.zw = IN.TexCoord2;
	}

	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UICameraDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse;
	if (@_METAL)
	{
		float3 yuv;
		yuv.x = DiffuseTexture.Sample(DiffuseTexture_Sampler, IN.TexCoord.xy).r;
		yuv.yz = AuxiliaryTexture.Sample(AuxiliaryTexture_Sampler, IN.TexCoord.xy).rg - float2(0.5, 0.5);
		diffuse = float4(MulMat3x4(ColorConversion, float4(yuv, 0.0)).xyz, 1);
	}
	else
	{
		diffuse = DiffuseTexture.Sample(DiffuseTexture_Sampler, IN.TexCoord.xy);
	}

	if (@_MaskTextureEnable)
	{
		diffuse *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
	}

	diffuse *= IN.Color.a;

	if (@_GrayingEnable)
	{
		diffuse.rgb = Graying(diffuse.rgb, EffectParam);
	}
	
	return diffuse;
}

technique Shading
{
	VertexShader = UICameraDrawVS();
	PixelShader = UICameraDrawPS();
}
