//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIClipImage.fx>
///	@path	~/Shader/UI/
///	@date	2017/06/29
///	@brief	Translucent clip UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4 ClipParam					 = { 0.0f, 1.0f, 0.0f, 0.0f };
};

@_DiffuseTextureEnable
{
	Name = DiffuseTextureEnable;
	Description = "Is diffuse map enabled?";
	Default = true;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float2		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UIClipDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;
	OUT.Color = IN.Color;

	return OUT;
}

float4 UIClipDrawPS(in VertexOutput IN) : SV_Target
{
	float alpha = IN.Color.a;
	if (@_DiffuseTextureEnable)
	{
		alpha *= SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord).a;
	}

	clip((alpha - ClipParam.x) * ClipParam.y);

	return float4(0, 0, 0, 0);
}

technique Shading
{
	VertexShader = UIClipDrawVS();
	PixelShader = UIClipDrawPS();
}
