//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UICustomMotion.fx>
///	@path	~/Shader/UI/
///	@date	2019/01/08
///	@brief	Translucent circular UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_Blur
{
	Name = Blur;
	Description = "Is the result blur?";
	Default = false;
}

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4	 BlurParam					 = { 0.0f, 0.0f, 0.0f, 0.0f };
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UICustomMotionDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.zw = IN.TexCoord2;
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UICustomMotionDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = float4(0, 0, 0, 0);
	if (@_Blur)
	{
		[unroll]for (int i = 0; i < 16; i += 1)
		{
			diffuse += SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + i * BlurParam.xy);
		}
		diffuse *= 0.0625;
	}
	else
	{
		diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	}
	
	diffuse *= IN.Color;

	return diffuse;
}

technique Shading
{
	VertexShader = UICustomMotionDrawVS();
	PixelShader = UICustomMotionDrawPS();
}
