//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gsfn2281@corp.netease.com
//	Author: Shunfeng Guo
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIBaseElement.fx>
///	@path	~/Shader/UI/
///	@date	2020/04/28
///	@brief	ImGui.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float2		TexCoord				: TEXCOORD0;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};


////////////////////////////////////////////////////////////////////////

VertexOutput UIImGuiVS(in VertexAssemble IN)
{
	VertexOutput OUT;
	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.Color = GetVertexColor(IN.Color);
	if (@_VULKAN)
	{
		float y = OUT.HPosition.y;
		OUT.HPosition.y = OUT.HPosition.x;
		OUT.HPosition.x = 1 - y;
	}
	return OUT;
}

float4 UIImGuiPS(in VertexOutput IN) : SV_Target
{

	float4 out_color = IN.Color * SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	return out_color;
}

technique Shading
{
	VertexShader = UIImGuiVS();
	PixelShader = UIImGuiPS();
}
