//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UILive2D.fx>
///	@path	~/Shader/UI/
///	@date	2020/02/17
///	@brief	Translucent live UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
};

@_HighlightEnable
{
	Name = HighlightEnable;
	Description = "Is highlight enabled?";
	Default = false;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float2		TexCoord				: TEXCOORDN;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UILive2DDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UILive2DDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	if (@_HighlightEnable)
	{
		diffuse.rgb *= diffuse.a;
	}
	diffuse *= IN.Color;
	return diffuse;
}

technique Shading
{
	VertexShader = UILive2DDrawVS();
	PixelShader = UILive2DDrawPS();
}

VertexOutput UILive2DClipDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;

	return OUT;
}

float4 UILive2DClipDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	clip(diffuse.a - 0.004);
	return float4(0, 0, 0, 1);
}

technique Clip
{
	VertexShader = UILive2DClipDrawVS();
	PixelShader = UILive2DClipDrawPS();
}
