//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIParticle.fx>
///	@path	~/Shader/UI/
///	@date	2015/10/27
///	@brief	UI Particle.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4   UVTransform				 = { 1.0f, 0.0f, 0.0f, 1.0f };
};

@_HighlightEnable
{
	Name = HighlightEnable;
	Description = "Is highlight enabled?";
	Default = false;
};

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Do we have DepthMap?";
	Visible = false;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tDepthMap 	: SceneDepth	: layout(t2);
sampler sDepthSampler 					: layout(s2)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	[@_DepthMapEnable]
	float4		ScreenCoord				: TEXCOORD1;
};

////////////////////////////////////////////////////////////////////////
// UI particle draw.

VertexOutput UIParticleDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;
	OUT.Color = GetVertexColor(IN.Color);

	if (@_DepthMapEnable)
	{
		float2 screenUV = OUT.HPosition.xy / (2.0 * OUT.HPosition.w) + 0.5;
		[!@_GLSL]
		screenUV.y = 1.0 - screenUV.y;

		if (@_VULKAN)
		{
			float2 t = (screenUV - 0.5) * 2.0;
			screenUV.x = dot(t, UVTransform.xy);
			screenUV.y = dot(t, UVTransform.zw);
			screenUV = screenUV * 0.5 + 0.5;
		}

		float near = CameraInfo.x;
		float far = CameraInfo.y;
		float tmp = IN.Position.z + near + far;
		float tmp2 = near * far;

		OUT.ScreenCoord.xy = screenUV;
		OUT.ScreenCoord.z = (tmp2 / tmp - BOARD_DEPTH_BIAS) / far;
		OUT.ScreenCoord.w = near / tmp;
	}

	return OUT;
}

float4 UIParticleDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord);
	
	if (!@_HighlightEnable)
	{
		diffuse.rgb *= diffuse.a;
	}

	diffuse *= IN.Color;

	if (@_HighlightEnable)
	{
		diffuse.rgb *= diffuse.a;
	}

	if (@_DepthMapEnable)
	{
		float sDepth = DecodeDepth(tDepthMap.Sample(sDepthSampler, IN.ScreenCoord.xy));
		diffuse *= smoothstep(IN.ScreenCoord.z, IN.ScreenCoord.w, sDepth);
	}
	
	return diffuse;
}

technique Shading
{
	VertexShader = UIParticleDrawVS();
	PixelShader = UIParticleDrawPS();
}
