//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIProgressTimer.fx>
///	@path	~/Shader/UI/
///	@date	2019/01/24
///	@brief	Translucent circular UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_IsBar
{
	Name = IsBar;
	Description = "Is the Bar?";
	Default = false;
}

@_IsReverse
{
	Name = IsReverse;
	Description = "Is the reverse?";
	Default = false;
}

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4	 ProgressTimerParam			 = { 0.0f, 0.0f, 0.0f, 0.0f };
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

#define D_PI 3.1415926535897932384626433832795

VertexOutput UIProgressTimerDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.zw = IN.TexCoord2;
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UIProgressTimerDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	if (@_IsBar)
	{
		float2 dis = abs(ProgressTimerParam.xy - IN.TexCoord.zw);
		float alpha = (dis.x > ProgressTimerParam.z || dis.y > ProgressTimerParam.w) ? 0.f : 1.f;
		if (@_IsReverse)
		{
			alpha = 1 - alpha;
		}
		diffuse.a *= alpha;
	}
	else
	{
		float2 dir = IN.TexCoord.zw - ProgressTimerParam.xy;
		float theta = atan2(dir.y, dir.x);
		if (theta > D_PI * 0.5)
			theta = theta - D_PI * 2;
		theta += 1.5 * D_PI;
		if (!@_IsReverse)
		{
			theta = D_PI * 2 - theta;
		}
		float alpha = (theta > ProgressTimerParam.z) ? 0.f : 1.f;
		diffuse.a *= alpha;
	}

	diffuse *= IN.Color;
	return diffuse;
}

technique Shading
{
	VertexShader = UIProgressTimerDrawVS();
	PixelShader = UIProgressTimerDrawPS();
}
