//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIPuppet.fx>
///	@path	~/Shader/UI/
///	@date	2018/05/16
///	@brief	Translucent puppet UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
};

@_HighlightEnable
{
	Name = HighlightEnable;
	Description = "Is highlight enabled?";
	Default = false;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float4		Normal 					: NORMAL;
	float2		TexCoord				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		DarkColor				: TEXCOORDN;
	float2		TexCoord				: TEXCOORDN;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UIPuppetDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;
	OUT.Color = GetVertexColor(IN.Color);
	OUT.DarkColor = IN.Normal;

	return OUT;
}

float4 UIPuppetDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	if (@_HighlightEnable)
	{
		diffuse.rgb *= diffuse.a;
	}
	diffuse *= IN.Color;
	diffuse.rgb += IN.DarkColor.rgb;

	return diffuse;
}

technique Shading
{
	VertexShader = UIPuppetDrawVS();
	PixelShader = UIPuppetDrawPS();
}

VertexOutput UIPuppetClipDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);

	return OUT;
}

float4 UIPuppetClipDrawPS(in VertexOutput IN) : SV_Target
{
	return float4(0, 0, 0, 1);
}

technique Clip
{
	VertexShader = UIPuppetClipDrawVS();
	PixelShader = UIPuppetClipDrawPS();
}
