//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIBaseImage.fx>
///	@path	~/Shader/UI/
///	@date	2015/01/16
///	@brief	Translucent base UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	Default = false;
};

@_BackgroundEnable
{
	Name = BackgroundEnable;
	Description = "Is background render enable?";
	Default = false;
};

@_GrayingEnable
{
	Name = GrayingEnable;
	Description = "Is the result graying?";
	Default = false;
};

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Do we have DepthMap?";
	Visible = false;
};

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float EffectParam					 = 0.0;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture					: layout(t1) = "White";
sampler_state MaskTexture_Sampler		: layout(s1, s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent base UI draw.

VertexOutput UIShowDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	
	[@_D3D9]
		OUT.HPosition.xy += ScreenInfo.zw * OUT.HPosition.w;

	OUT.TexCoord.xy = IN.TexCoord;
	
	[@_GLSL]
		OUT.TexCoord.y = 1 - OUT.TexCoord.y;
	
	[@_MaskTextureEnable]
		OUT.TexCoord.zw = IN.TexCoord2;
	
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UIShowDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);

	float addAlpha = (diffuse.r * 0.325 + diffuse.g * 0.375 + diffuse.b * 0.275);
	diffuse.a = min(1, max(diffuse.a, addAlpha));

	[@_BackgroundEnable]
		diffuse.a = 1.0;

	if (@_MaskTextureEnable)
	{
		diffuse *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
	}
	
	diffuse *= IN.Color.a;

	if (@_GrayingEnable)
	{
		diffuse.rgb = Graying(diffuse.rgb, EffectParam);
	}

	return diffuse;
}

technique Shading
{
	VertexShader = UIShowDrawVS();
	PixelShader = UIShowDrawPS();
}
