//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIText.fx>
///	@path	~/Shader/UI/
///	@date	2015/02/12
///	@brief	UI text.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4   UVTransform				 = { 1.0f, 0.0f, 0.0f, 1.0f };
};

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	Default = false;
};

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Do we have DepthMap?";
	Visible = false;
};

texture2D Text_Texture				: layout(t0) = "Transparent";
sampler_state Text_Sampler			: layout(s0)
{
	Texture = Text_Texture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture				: layout(t1) = "White";
sampler_state MaskTexture_Sampler	: layout(s1, s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tDepthMap 	: SceneDepth	: layout(t3);
sampler sDepthSampler 					: layout(s3)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

struct VertexScreenInput
{
	float4		Position 	: POSITION;
	float4		Color 		: COLOR;
	float2		TexCoord	: TEXCOORD0;
	float2		TexCoord2	: TEXCOORD1;
};

struct VertexScreenOutput
{
	float4		HPosition	: SV_Position;
	float4		Color		: COLOR;
	float4		TexCoord	: TEXCOORD0;
	[@_DepthMapEnable]
	float4		ScreenCoord	: TEXCOORD1;
};

////////////////////////////////////////////////////////////////////////
// System text draw.

VertexScreenOutput TextDrawVS(in VertexScreenInput IN)
{
	VertexScreenOutput OUT;

	[@_HLSL]
	OUT = (VertexScreenOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	[@_D3D9]
		OUT.HPosition.xy += ScreenInfo.zw * OUT.HPosition.w;

	OUT.TexCoord.xy = IN.TexCoord;
	[@_MaskTextureEnable]
		OUT.TexCoord.zw = IN.TexCoord2;

	OUT.Color = GetVertexColor(IN.Color);

	if (@_DepthMapEnable)
	{
		float2 screenUV = OUT.HPosition.xy / (2.0 * OUT.HPosition.w) + 0.5;
		[!@_GLSL]
		screenUV.y = 1.0 - screenUV.y;

		if (@_VULKAN)
		{
			float2 t = (screenUV - 0.5) * 2.0;
			screenUV.x = dot(t, UVTransform.xy);
			screenUV.y = dot(t, UVTransform.zw);
			screenUV = screenUV * 0.5 + 0.5;
		}

		float near = CameraInfo.x;
		float far = CameraInfo.y;
		float tmp = IN.Position.z + near + far;
		float tmp2 = near * far;

		OUT.ScreenCoord.xy = screenUV;
		OUT.ScreenCoord.z = (tmp2 / tmp - BOARD_DEPTH_BIAS) / far;
		OUT.ScreenCoord.w = near / tmp;
	}

	return OUT;
}

float4 TextDrawPS(in VertexScreenOutput IN) : SV_Target
{
	float4 result = SampleColorMap(Text_Texture, Text_Sampler, IN.TexCoord.xy);

	if (@_MaskTextureEnable)
	{
		result.a *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
	}

	if (@_DepthMapEnable)
	{
		float sDepth = DecodeDepth(tDepthMap.Sample(sDepthSampler, IN.ScreenCoord.xy));
		result *= smoothstep(IN.ScreenCoord.z, IN.ScreenCoord.w, sDepth);
	}

	result *= IN.Color;

	if (@_DepthMapEnable)
	{
		float sDepth = DecodeDepth(tDepthMap.Sample(sDepthSampler, IN.ScreenCoord.xy));
		result *= smoothstep(IN.ScreenCoord.z, IN.ScreenCoord.w, sDepth);
	}
	
	return result;
}

technique Shading
{
	VertexShader = TextDrawVS();
	PixelShader = TextDrawPS();
}
