//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UITextAtlas.fx>
///	@path	~/Shaders/UI/
///	@date	2018/01/23
///	@brief	UITextAtlas element.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4	WorldViewProj		 	;
	float4		EffectColor				 = { 1.0f, 1.0f, 1.0f, 1.0f };
};

@_EffectEnable
{
	Name = EffectEnable;
	Description = "Is effect enabled?";
	Default = false;
};

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	Default = false;
};

texture2D Text_Texture				: layout(t0) = "Transparent";
sampler_state Text_Sampler			: layout(s0)
{
	Texture = Text_Texture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture				: layout(t1) = "White";
sampler_state MaskTexture_Sampler	: layout(s1, s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexScreenInput
{
	float4		Position 	: POSITION;
	float4		Color 		: COLOR;
	float2		TexCoord	: TEXCOORD0;
	float2		TexCoord2	: TEXCOORD1;
};

struct VertexScreenOutput
{
	float4		HPosition	: SV_Position;
	float4		Color		: COLOR;
	float4		TexCoord	: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// System text draw.

VertexScreenOutput TextAtlasDrawVS(in VertexScreenInput IN)
{
	VertexScreenOutput OUT;

	[@_HLSL]
	OUT = (VertexScreenOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	[@_MaskTextureEnable]
		OUT.TexCoord.zw = IN.TexCoord2;

	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 TextAtlasDrawPS(in VertexScreenOutput IN) : SV_Target
{
	float4 result = SampleColorMap(Text_Texture, Text_Sampler, IN.TexCoord.xy);

	if (@_D3D11 || @_D3D12 || @_VULKAN || @_METAL)
		result.rgba = result.rrrg;
	else
		result.rgba = result.rrra;

	float4 c1 = IN.Color;
	if (@_EffectEnable)
	{
		float alpha = result.r * c1.a;
		float blend = result.a / max(result.r, 0.004);
		result.rgb = c1.rgb * blend + EffectColor.rgb * ((1.0f - blend) * (1.f - result.a));
		result.a = alpha;
	}
	else
	{
		result.rgb = c1.rgb;
		result.a *= c1.a;
	}

	if (@_MaskTextureEnable)
	{
		result.a *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
	}

	return result;
}

technique Shading
{
	VertexShader = TextAtlasDrawVS();
	PixelShader = TextAtlasDrawPS();
}
