//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : lnn4047@corp.netease.com
//	Author: liuning
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<VXGIClearVoxel.fx>
///	@path	~/System/Shader/
///	@date	2020/09/18
///	@brief	Clear outdated voxel data.

// {HEX} VXGI.
#pragma shader draw

#include "Config.fxh"

#if SD_ENABLE_VXGI

#include "Constants.fxh"
#include "Common.fxh"
#include "VXGI/VXGIConst.fxh"
#include "VXGI/VXGICommon.fxh"

@_IsFirstFrame
{ 
	Name = IsFirstFrame;
	Description = "In the first frame, history information is invalid";
	Default = false;
};
@_ForceFullClear
{
	Name = ForceFullClear;
	Description = "Incremental or full cleanup? Full cleanup will clean all voxel data";
	Default = false;
}

RWTexture3D<uint> tVoxelColor;

////////////////////////// Clear Voxel Data ////////////////////////////////////////////

#define CLEAR_VOXEL(PreGridMin, CurGridMin, CurGridMax, Dimension, CellSize, Tex, Layer)\
{\
	int3 CellIndex = PreGridMin + DispatchID;\
	if (any(CellIndex < CurGridMin) || any(CellIndex > CurGridMax))\
	{\
		float3 WorldPosition = (CellIndex + float3(0.5, 0.5, 0.5)) * CellSize;\
		int3 LayerCoord = VXGIGetClipmapCoord(WorldPosition, CellSize, Dimension, Layer, 0);\
		Tex[LayerCoord + int3(0, 0, 0)] = 0;\
		Tex[LayerCoord + int3(1, 0, 0)] = 0;\
		Tex[LayerCoord + int3(2, 0, 0)] = 0;\
		Tex[LayerCoord + int3(3, 0, 0)] = 0;\
		Tex[LayerCoord + int3(4, 0, 0)] = 0;\
		Tex[LayerCoord + int3(5, 0, 0)] = 0;\
	}\
}

#define CLEAR_VOXEL_WITH_HISTORY(CellSize, Dimension, Tex, Layer)\
{\
	int3 PreGridMin, PreGridMax;\
	VXGIGetGridRange(CellSize, Dimension, VXGIPreCenter, PreGridMin, PreGridMax);\
	int3 CurGridMin, CurGridMax;\
	VXGIGetGridRange(CellSize, Dimension, VXGICurCenter, CurGridMin, CurGridMax);\
	CLEAR_VOXEL(PreGridMin, CurGridMin, CurGridMax, Dimension, CellSize, Tex, Layer);\
}

#define CLEAR_VOXEL_WITHOUT_HISTORY(Dimension, Tex, Layer)\
{\
	int3 LayerCoord = int3(DispatchID.x * VXGIDirCount, DispatchID.y, DispatchID.z + Dimension.z * Layer);\
	Tex[LayerCoord + int3(0, 0, 0)] = 0;\
	Tex[LayerCoord + int3(1, 0, 0)] = 0;\
	Tex[LayerCoord + int3(2, 0, 0)] = 0;\
	Tex[LayerCoord + int3(3, 0, 0)] = 0;\
	Tex[LayerCoord + int3(4, 0, 0)] = 0;\
	Tex[LayerCoord + int3(5, 0, 0)] = 0;\
}

[numthreads(4, 4, 4)]
void ClearVoxelCS(int3 GroupID : SV_GroupThreadID, int3 DispatchID : SV_DispatchThreadID)
{
	if (@_IsFirstFrame || @_ForceFullClear)
	{
		CLEAR_VOXEL_WITHOUT_HISTORY(VXGIColorDimension, tVoxelColor, 0);
		CLEAR_VOXEL_WITHOUT_HISTORY(VXGIColorDimension, tVoxelColor, 1);
		CLEAR_VOXEL_WITHOUT_HISTORY(VXGIColorDimension, tVoxelColor, 2);
	}
	else
	{
		CLEAR_VOXEL_WITH_HISTORY(VXGIColorCellSizes.x, VXGIColorDimension, tVoxelColor, 0);
		CLEAR_VOXEL_WITH_HISTORY(VXGIColorCellSizes.y, VXGIColorDimension, tVoxelColor, 1);
		CLEAR_VOXEL_WITH_HISTORY(VXGIColorCellSizes.z, VXGIColorDimension, tVoxelColor, 2);
	}
}

technique ComputeClearVoxel
{
	ComputeShader = ClearVoxelCS();
}

#else

float4 UnusedVertexShading(float4 position: POSITION) : SV_Position
{
	return float4(0, 0, 0, 1);
}

float4 UnusedPixelShading(float4 position: SV_Position) : SV_Target
{
	return float4(0, 0, 0, 1);
}

technique Unused
{
	VertexShader = UnusedVertexShading();
	PixelShader = UnusedPixelShading();
}

#endif
// {HEX} end
