//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : szn3878@corp.netease.com
//	Author: Si Zheng
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Vignetting.fx>
///	@path	~/System/Shader/
///	@date	2016/08/23
///	@brief	Screen Vignetting Effect

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// const buffer //////////////////
cbuffer Batch
{
	float cVignettingPercent 			;
	float cDarkenss						;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tBaseMap : layout(t0);
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

/////////////////// HDRSampleLumInitial ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 VignettingPS(v2f IN) : SV_Target
{
	half3 Source = SampleColorMap(tBaseMap, sBaseSampler, IN.TC.xy).rgb;
	
	//边沿黑化
	half sideDarkness = distance(IN.TC.xy,half2(0.5,0.5));

	half3 Result = Source * lerp( 1, saturate(1 - sideDarkness*cDarkenss), cVignettingPercent);

	return float4(Result, 1.0);
}

technique VignettingTech
{
	VertexShader = ScreenVS();
	PixelShader = VignettingPS();
}


