//	--------------------------------------------------------------------
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<VolumetricCloud.fx>
///	@path	~/System/Shader/
///	@date	2019/01/23
///	@brief

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"
#include "Noise.fxh"

#define TEX3D_SIZEA (64)
#define TEX3D_SIZEB (32)

//const value for how many type in LUT
#define RAYMARCHING_CLOUD_TYPE_COUNT 					32

//Remove Cloud at specific distance
#define RAYMARCHING_NEARCLIP_ENABLE 					true
#define RAYMARCHING_NEARCLIP_ENABLE_FAR_DISTANCE 		40.0
#define RAYMARCHING_NEARCLIP_ENABLE_NEAR_DISTANCE 		20.0

//using a const Raymarching result's density as standard, then density won't be affect by rayMarching steps
#define RAYMARCHING_DENSITY_STANDARD 					512

//Lerp last raymarching step's result to remove stripe
#define RAYMARCHING_NEAR_DEPTH_FIX 						true

//Fake Scale for reducing total steps of raymarching
#define RAYMARCHING_STEP_SCALE							true

//Fake fix for reducing height of further Cloud
#define RAYMARCHING_HORLIZON_FIX 						1000000.0

//Extend Depth For fill empty aera
#define RAYMARCHING_ENABLE_DEPTH_EXTEND 				true

//Raymaching Iterations of each step
#define RAYMARCHING_LIGHT_ITERATIONS  					6
#define RAYMARCHING_LIGHT_DENSITY_STANDARD				6
#define RCP_RAYMARCHING_LIGHT_ITERATIONS 				(1.0 / RAYMARCHING_LIGHT_ITERATIONS);

//Return Result of Upscale Directly
#define COMPOSITE_ENABLE 								true

//Soft Depth Test Range
#define COMPOSITE_SOTE_RANGE 							500

//Return Result of RayMarching Directly
#define UPSCALE_ENABLE 									true

//Temporal Lerp : How much current pixel in RayMarching Resule affect Final Result
//When using TSAA , Larger Value May Get Blur
#define UPSCALE_TEMPORAL_STATIC_BLEND 					0.65

//Disable Color Clamp For Debug
#define UPSCALE_TEMPORAL_COLOR_CLAMP 					true

//Scale Between RayMarching RT and Upsacle RT
#define UPSCALE_DITHER_SIZE 							2

//debug / dev
#define DEBUG_ENABLE_FOG false
#define DEBUG_FOG_MODE false
#define DEBUG_VISUAL_MODE 0
#define DEBUG_VISUAL_LOOP_DEFAULT 1.0

///////////////// samplers //////////////////
texture2D tScatteringAndAlpha : layout(t0,t1);
sampler sScatteringAndAlpha : layout(s0)
{
	Texture = tScatteringAndAlpha;
	Filter = Point;
	Address = Clamp;
};
sampler sScatteringAndAlphaLinear : layout(s1)
{
	Texture = tScatteringAndAlpha;
	Filter = Linear;
	Address = Clamp;
};

texture2D tDepthBuffer : SceneDepth: layout(t2);
sampler sDepthBufferSampler : layout(s2)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};
sampler sDepthBufferLinearSampler : layout(s3)
{
	Texture = tDepthBuffer;
	Filter = Linear;
	Address = Clamp;
};

//ADD
texture2D tScreenTexture : SceneColorOpaque : layout(t4);
sampler sScreenTextureSampler : layout(s4)
{
	Texture = tScreenTexture;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};

texture2D tWeightDepth : layout(t5);
sampler sWeightDepth : layout(s5)
{
	Texture = tWeightDepth;
	Filter = Linear;
	Address = Clamp;
	Precision = Medium;
};

texture2D tLastCloudRT : layout(t6);
sampler sLastCloudRT : layout(s6)
{
	Texture = tLastCloudRT;
	Filter = Linear;
	Address = Clamp;
};



texture2D tBlueNoise : layout(t7);
sampler sBlueNoise : layout(s7)
{
	Texture = tBlueNoise;
	Filter = Point;
	Address = Wrap;
};


texture2D tHeightLUT : layout(t8);
sampler sHeightLUT : layout(s8)
{
	Texture = tHeightLUT;
	Filter = Linear;
	Address = Wrap;
};

texture2D tCloud3DNoiseTextureA : layout(t9);
sampler sCloud3DNoiseTextureA : layout(s9)
{
	Texture = tCloud3DNoiseTextureA;
	Filter = Linear;
	Address = Wrap;
};

texture2D tCloud3DNoiseTextureB : layout(t10);
sampler sCloud3DNoiseTextureB : layout(s10)
{
	Texture = tCloud3DNoiseTextureB;
	Filter = Linear;
	Address = Wrap;
};

texture2D tWeatherData : layout(t11);
sampler sWeatherData : layout(s11)
{
	Texture = tWeatherData;
	Filter = Linear;
	Address = Wrap;
};


texture2D tGodRaySetupMap : layout(t0);
sampler sGodRaySetupMap : layout(s0)
{
	Texture = tGodRaySetupMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tLastGodRayMap : layout(t1);
sampler sLastGodRayMap : layout(s1)
{
	Texture = tLastGodRayMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tGodRayFinalMap : layout(t0);
sampler sGodRayFinalMap : layout(s0)
{
	Texture = tGodRayFinalMap;
	Filter = Linear;
	Address = Clamp;
};



///////////////// io signatures //////////////////
struct a2v
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

struct VCPS
{
	float4 Color : SV_Target0;
	float2 Weight : SV_Target1;
};

struct VertexShadingOutput
{
	float4 HPosition : SV_Position;
	float2 UV : TEXCOORDN;
	float4 CurrentSubPixel : TEXCOORDN;
	// float2 TraceMinMax : TEXCOORDN;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 cWBasisX						: WBasisX;
	float4 cWBasisY						: WBasisY;
	float4 cWBasisZ						: WBasisZ;
	float2 cCloudHeightMinMax           ;
	float cNoiseFreqScaleA 				;
	float cNoiseFreqScaleB				;			
	float cPhaseG 						;
	float4 cCoverage					;			
	float3 cCloudAmbient				;
	float cCutoff 						;			
	float3 cWindDir						;
	float cWindSpeed 					;			
	float3 cSunColor					;
	float cMaxTraceDistance 			;			
	float cCloudTypeOffset 				;
	float cCloudDarkness 				;

	float cGodRayRange 					;
	float2 cSunPos						;
	float cMaxBlurDistance 				;
	float cAspectRatio 					;
	float2 cInvDepthSize				;
	float3 cGodRayColor					;
	float cDitherSize					;

	float4x4 lastViewProjTex			;

	//Remove
	float cSilverIntensity 				;
	float cSilverSpread 				;
	float cCloudDensityScale 			;
	float cFadeDistance 				;
	float2 cCoverageUVOffset			;

	// ADD
	float2 cCloudLightingPhaseControl	;
	float cCloudSunIntensity			;
	float cCloudLightingStepScale		;
	float2 cCloudShapeControl			;
    float cCloudShapeSoftness			;
	float cCloudRayMarchingCount		;
};

///////////////// pre defines & helper functions //////////////////

half GetRelativeHeightFractionForPoint(float3 inPosition)
{
	half HorizonFix = (length(inPosition.xyz - CameraPos.xyz) / (RAYMARCHING_HORLIZON_FIX * cMaxTraceDistance));
	//HorizonFix = 0;
	half height_fraction = (inPosition.y - cCloudHeightMinMax.x) / (cCloudHeightMinMax.y - cCloudHeightMinMax.x);
	return saturate(height_fraction + HorizonFix);
}

half rescale(half vMin, half vMax, half v)
{
	return ((v - vMin)) / (saturate(vMax - vMin));
}

half SampleLowFrequencyNoises(float3 uvw, half mipLevel)
{
	uvw *= 0.0003 * cNoiseFreqScaleA;
	uvw.yz = frac(uvw.yz);//frac(uvw.xyz) lead seam
	float sliceIdx = uvw.z * TEX3D_SIZEA - 0.5;
	float sliceA = floor(sliceIdx);
	float lerpAB = sliceIdx - sliceA;
	sliceA += step(sliceA, -0.1) * TEX3D_SIZEA;
	float sliceB = sliceA + 1;
	sliceB -= step(TEX3D_SIZEA - 0.5, sliceB) * TEX3D_SIZEA;

	float2 uvA = float2((sliceA + uvw.x) / TEX3D_SIZEA, uvw.y);
	float2 uvB = float2((sliceB + uvw.x) / TEX3D_SIZEA, uvw.y);
	half dataA = tCloud3DNoiseTextureA.SampleLevel(sCloud3DNoiseTextureA, uvA, mipLevel).r;
	half dataB = tCloud3DNoiseTextureA.SampleLevel(sCloud3DNoiseTextureA, uvB, mipLevel).r;
	return lerp(dataA, dataB, lerpAB);
}

half3 SampleWeather(float3 pos)
{
	//float3 weatherMoveDir = float3(0, 0, 1);
	//pos += weatherMoveDir * cWindSpeed * GameTime;
	float2 uv = (pos.xz ) / cCoverage.xy;
	uv += (cCoverage.xy + cCoverage.zw) / cCoverage.xy;
	half3 weatherData = tWeatherData.SampleLevel(sWeatherData, uv, 0).rgb;
	return weatherData;
}

// float powder(float sampleDensity, float lightDotEye)
// {
// 	float powd = 1.0f - exp(-sampleDensity * 2.0f);
// 	return lerp(1.0f, powd, saturate((-lightDotEye * 0.5f) + 0.5f)); // [-1,1]->[0,1]
// }

// float randomC(float2 st) {
// 	st = floor(st * sin(st.x));
// 	return frac(sin(dot(st.xy, float2(12.9898, 78.233))) * 43758.5453123);
// }

half PhaseFunctionOrigami(half g, half CosTheta, half Ext)
{
	half w1 = cCloudLightingPhaseControl.x * 50.0;
	half rst = w1 * Ext * henyeyGreenstein(CosTheta, 0.95 * g);

	half w2 = cCloudLightingPhaseControl.y * 50.0;
	rst += w2 * Ext * henyeyGreenstein(CosTheta, 0.1 * g);

	//	float w0 = LightingControl.y * 0.2;
	//rst += Ext * 0.1 * w0 * henyeyGreenstein(CosTheta , g);

	return rst;
}

float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}

half lightEnergy2(float lightDotEye, float densitySample)
{
	/*0611M_Remove unusual Const*/
	float rst = PhaseFunctionOrigami(cPhaseG, lightDotEye, 1.0);			// Get Phase Function
	rst *= saturate(exp(-(cCloudDarkness - 0.5) * 5 * densitySample * 1.0)); 	// Beer Lambert's Law
	return rst;
}

/////////////////// MediaSetupByGlobalFog ///////////////
v2f ScreenVS(a2v IN)
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.HPosition.z = (HALF_EPSILON * 5) * OUT.HPosition.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.
	OUT.TC = IN.BaseTC;
	return OUT;
}

///////////////////////////////VolumetricCloudCompositePS/////////////////////////////////////
float4 VolumetricCloudCompositePS(VertexShadingOutput IN) : SV_Target
{
	// return float4(0,0,0,1);
	float2 UV = IN.UV;
	half4 Cloud =  tScatteringAndAlpha.Sample(sScatteringAndAlphaLinear, UV);

	if (COMPOSITE_ENABLE == false)
		return tScatteringAndAlpha.Sample(sScatteringAndAlpha, UV);

	half4 WeightDepth = tWeightDepth.Sample(sWeightDepth, UV);
	half4 WeightDepth2 = tWeightDepth.Sample(sWeightDepth , UV + half2(1,1) * ScreenInfo.zw);	
	half4 WeightDepth3 = tWeightDepth.Sample(sWeightDepth , UV + half2(-1,-1) * ScreenInfo.zw);
	WeightDepth = max(WeightDepth,max(WeightDepth2,WeightDepth3));

	float minDepth = WeightDepth.x - WeightDepth.y;

	float scene_depth  = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, UV));
	half SoftRangeInDepth = (scene_depth * COMPOSITE_SOTE_RANGE - CameraInfo.x) / CameraInfo.y ;
	scene_depth = scene_depth > (1 - HALF_EPSILON) ? (999999 - CameraInfo.x) / CameraInfo.y : scene_depth;

	half nearAlphaFix = smoothstep(
		RAYMARCHING_NEARCLIP_ENABLE_FAR_DISTANCE/(CameraInfo.y-CameraInfo.x),
		(2*RAYMARCHING_NEARCLIP_ENABLE_FAR_DISTANCE-RAYMARCHING_NEARCLIP_ENABLE_NEAR_DISTANCE)/(CameraInfo.y-CameraInfo.x),
		scene_depth
	);

	half farAlphaFix = smoothstep(
		saturate(minDepth - SoftRangeInDepth),
		saturate(minDepth + SoftRangeInDepth),
		scene_depth
	);

	return Cloud * nearAlphaFix * farAlphaFix;
}

float4 VolumetricCloudUpscalePS(VertexShadingOutput IN) : SV_Target
{
	// return float4(0,0,0,1);
// UV & Pixel Calc - Reconstruct Position for both Current Pixel and History Pixel 
	const half DitherSize = UPSCALE_DITHER_SIZE;
	float2 uv = IN.UV;
	half2 Addition = IN.CurrentSubPixel.zw * step(1-IN.UV , 0.5);
	half2 PixelIndex = fmod(floor((IN.UV + Addition * ScreenInfo.zw) * ScreenInfo.xy),DitherSize);
	half2 DiffAC = PixelIndex - IN.CurrentSubPixel.xy * DitherSize;
	half ActivedPixel = step(length(DiffAC),0.02);
	half2 PixelSizeQ = ScreenInfo.zw * 2;

	if(UPSCALE_ENABLE == false)
		return tScatteringAndAlpha.Sample(sScatteringAndAlpha,uv);

	half4 WeightDepth = (tWeightDepth.Sample(sWeightDepth,uv));//1/4
	half4 WeightDepth2 = tWeightDepth.Sample(sWeightDepth , uv + half2(2,2) * ScreenInfo.zw);	
	half4 WeightDepth3 = tWeightDepth.Sample(sWeightDepth , uv + half2(-2,-2) * ScreenInfo.zw);

	half2 WeightDepthMax = min(WeightDepth.xy,min(WeightDepth2.xy, WeightDepth3.xy));
	half minDepth = WeightDepthMax.x - WeightDepthMax.y;

	float3 worldPos = CalcHomogeneousPos( WeightDepth.x ,uv);
	worldPos += CameraPos.xyz;
	float4 LastPos = MulMat4x4(lastViewProjTex, float4(worldPos.xyz,1.0));
	float2 LastUV = (LastPos.xy / LastPos.w);
	
	half4 OUT11 = tScatteringAndAlpha.Sample(sScatteringAndAlphaLinear, uv + half2(1,0) * PixelSizeQ);//1/4
	half4 OUT12 = tScatteringAndAlpha.Sample(sScatteringAndAlphaLinear, uv + half2(-1,0)* PixelSizeQ);//1/4
	half4 OUT13 = tScatteringAndAlpha.Sample(sScatteringAndAlphaLinear, uv + half2(0,1) * PixelSizeQ);//1/4
	half4 OUT14 = tScatteringAndAlpha.Sample(sScatteringAndAlphaLinear, uv + half2(0,-1)* PixelSizeQ);//1/4

	half4 OUT = tScatteringAndAlpha.Sample(sScatteringAndAlpha,uv);

	float scene_depth = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, IN.UV , 0));

	half4 OUT2 = tLastCloudRT.Sample(sLastCloudRT, LastUV);//1/2   History

	half alphaDiff = saturate(OUT2.a - OUT.a);
	half depthDiff = saturate(WeightDepth.x + WeightDepth.y - scene_depth);
	half C = saturate(( -minDepth + scene_depth )*5.0);
	half D = scene_depth > 1 - HALF_EPSILON ? C : 10.0;

	half4 AMin2 = min(OUT,min(OUT11,min(OUT13,min(OUT14,OUT12))));
	half4 AMax2 = max(OUT,max(OUT11,max(OUT13,max(OUT14,OUT12))));

	if(UPSCALE_TEMPORAL_COLOR_CLAMP == true)
		 OUT2 = clamp(OUT2,AMin2,AMax2);
	
	half2 FailedPixelOutCamera = 1-step(abs(LastUV - saturate(LastUV)),0);

	half weight = 0;

	weight = ActivedPixel*(1-UPSCALE_TEMPORAL_STATIC_BLEND);						//Basic	
	weight = saturate(weight + (depthDiff * alphaDiff * D));						//RayMarching with Extend Depth's fix
	weight = saturate(weight + FailedPixelOutCamera.x + FailedPixelOutCamera.y);	//Out Of Camera 's fix

	return lerp(OUT2,OUT,saturate(weight));											//history Lerp
}
//========================================================================================================================================

half SampleCloudDensity(float3 p,half4 CloudInfo, float height_fraction, half3 weather_data, bool doCheaply, float3 ViewDir, bool isLightingSample)
{
	//Calc Wind's Affection on position for Sample
	half cloud_top_offset = 500.0;
	p += saturate(cWindDir) * GameTime * cWindSpeed;
	p += height_fraction * cWindDir * cloud_top_offset;

	//Sample Weather
	weather_data = saturate(weather_data - cCutoff + 0.5);

	//Rescale WeatherMap's Rnage
	half base_cloud = smoothstep(-cCloudShapeControl.x + CloudInfo.y, -cCloudShapeControl.y + CloudInfo.y, weather_data.x);

	//Here Shape Control When doCheaply is true 
	if (doCheaply)
		return rescale(0.1 , CloudInfo.w + 1, base_cloud);

	//Displacement Texture , Enable Only for PC
	half3 d3 = 0.5;
	// if(IS_WINDOWS)
	// {	
		//now displacement texture is packed in WeatherData
		half3 vp = p * 0.0005 * cNoiseFreqScaleB;
		half3 d1 = tWeatherData.SampleLevel(sWeatherData, vp.xz, 0).yzw;
		half3 d2 = tWeatherData.SampleLevel(sWeatherData, vp.zy, 0).yzw;
		// float3 d1 = tCloud3DNoiseTextureB.SampleLevel(sCloud3DNoiseTextureB, vp.xz, 0).rgb;
		// float3 d2 = tCloud3DNoiseTextureB.SampleLevel(sCloud3DNoiseTextureB, vp.zy, 0).rgb;
		d3 = (d1 + d2) * 2 + 2;
	// }

	half3 noiseD = (SampleLowFrequencyNoises(p + d3 * 250, 0)-0.5) * cCloudShapeSoftness;
	return rescale(noiseD.x , CloudInfo.w + 1, base_cloud);
	
}

half sampleCloudDensityAlongConeArcApple(float3 startPos , half lightDotEye, half originalDensity, bool doCheaply, float3 viewDir)
{
	// random vectors on the unit sphere
	static const float3 RANDOM_VECTORS[6] =
	{
		float3(0.38051305,  0.92453449, -0.02111345),
		float3(-0.50625799, -0.03590792, -0.86163418),
		float3(-0.32509218, -0.94557439,  0.01428793),
		float3(0.09026238, -0.27376545,  0.95755165),
		float3(0.28128598,  0.42443639, -0.86065785),
		float3(-0.16852403,  0.14748697,  0.97460106)
	};
	half sampleCount = min(6,floor(cCloudRayMarchingCount / 16.0));
	half densityFixup =  RAYMARCHING_LIGHT_DENSITY_STANDARD / sampleCount;

	float stepSize = (cCloudHeightMinMax.y - cCloudHeightMinMax.x) * /*constStepSize*/ 0.01 * cCloudLightingStepScale;
	float3 lightStep = stepSize * SunDirection.xyz;
	float3 pos = startPos;
	half densityAlongCone = 0.0;
	half3 weatherData = 0.0;
	float3 conePos = float3(0,0,0);
	half heightFrac = 0.0;
	half2 LUTuv = half2(0,0);
	half4 CloudInfo = float4(0,0,0,0);
	half cloudDensity = 0;

	for (int i = 0; i < sampleCount; ++i)
	{
		conePos = pos + /*coneRadius*/ 1.0  * RANDOM_VECTORS[i] * float(i);
		weatherData = SampleWeather(conePos);

		if (weatherData.x > HALF_EPSILON)
		{
			heightFrac = GetRelativeHeightFractionForPoint(conePos);
			LUTuv = half2((half)/*int_cloudType*/ ((cCloudTypeOffset + 0.5) * 36) / RAYMARCHING_CLOUD_TYPE_COUNT, 1 - heightFrac);
			CloudInfo = tHeightLUT.SampleLevel(sHeightLUT, LUTuv, 0);
			cloudDensity = SampleCloudDensity(conePos, CloudInfo, heightFrac, weatherData, true, viewDir, true);
			cloudDensity *= densityFixup;
			densityAlongCone += cloudDensity;
		}

		pos += lightStep * 1.5 * densityFixup;
	}

	return lightEnergy2(lightDotEye, densityAlongCone);
}

VCPS traceClouds(half BlueNoise,float3 viewDir, float3 startPos, half traceLen, float Depth, float DepthMax, float DepthMin)
{
	//FOR DEBUG
	float DebugDensityVisualCount = 0.0;

	VCPS OUT;

	half sampleCount = cCloudRayMarchingCount;
	float stepSize = cMaxTraceDistance / cCloudRayMarchingCount ;
	float3 posStep = stepSize * viewDir;
	half lightDotEye = dot(SunDirection.xyz, viewDir);
	float3 pos = startPos;
	half3 weatherData = 0.0;
	half4 result = 0.0;
	half DensityFixup = (float)RAYMARCHING_DENSITY_STANDARD / (float)cCloudRayMarchingCount;
	half EndDis2 = 0;
	half dis2 = 0;
	half minLen = cMaxTraceDistance;
	half maxLen = 0;
	half cloudDensity = 0;

	//Fix Trace Distance using Depth
	EndDis2 = DepthMax * CameraInfo.y + CameraInfo.x;
	EndDis2 = DepthMax >= 1 - HALF_EPSILON ? traceLen : EndDis2;

	//Offset To Cloud Layer
	if(RAYMARCHING_NEARCLIP_ENABLE == true)
	{
		pos += viewDir * RAYMARCHING_NEARCLIP_ENABLE_NEAR_DISTANCE;
		dis2 += RAYMARCHING_NEARCLIP_ENABLE_NEAR_DISTANCE;
		maxLen = RAYMARCHING_NEARCLIP_ENABLE_NEAR_DISTANCE;
	}

	// Random Origin Offset
	// float r = frac(randomC(frac(GameTime + 7106.885) + 7106.885 * UV.xy));
	// pos+= r * viewDir * posStep;
	// dis2+= r * posStep;


	// [ Under Dev ] Post Search <- useless when RayMarching Lowest Step Counts is lower then 64
	// float3 posStepA = viewDir * cMaxTraceDistance / 32;
	// for (int c = 0; c < 32; ++c)
	// {
	// 	weatherData = SampleWeather(pos);
	// 	if (weatherData.r >= HALF_EPSILON)
	// 	{
	// 		cloudDensity = SampleCloudDensity(pos, CloudInfo, heightFrac, weatherData, true, viewDir, true);
	// 		if (cloudDensity >= HALF_EPSILON)
	// 		{	
	// 			pos -= 4 * posStepA * (2-viewDir.y);
	// 			break;
	// 		}
	// 	}
	// 	heightFrac = GetRelativeHeightFractionForPoint(pos);
	// 	LUTuv = half2((half)/*int_cloudType*/ ((cCloudTypeOffset + 0.5) * 36) / RAYMARCHING_CLOUD_TYPE_COUNT, 1.0 - heightFrac);
	// 	CloudInfo = tHeightLUT.SampleLevel(sHeightLUT, LUTuv, 0);
	// 	pos += posStepA;
	// }

	for (int i = 0; i < sampleCount; ++i)
	{
		maxLen = max(maxLen,dis2);
		weatherData = SampleWeather(pos);
 		cloudDensity = 0;
		if (weatherData.r >= HALF_EPSILON)
		{

			half heightFrac = GetRelativeHeightFractionForPoint(pos);
			half2 LUTuv = half2((half)/*int_cloudType*/ ((cCloudTypeOffset + 0.5) * 36) / RAYMARCHING_CLOUD_TYPE_COUNT, 1.0 - heightFrac);
			half4 CloudInfo = tHeightLUT.SampleLevel(sHeightLUT, LUTuv, 0);

			cloudDensity = SampleCloudDensity(pos, CloudInfo, heightFrac, weatherData, false, viewDir, true);
			cloudDensity *= DensityFixup;

			if (cloudDensity >= HALF_EPSILON)
			{		
				if(RAYMARCHING_NEARCLIP_ENABLE == true)
					cloudDensity = saturate(cloudDensity * smoothstep(RAYMARCHING_NEARCLIP_ENABLE_NEAR_DISTANCE,RAYMARCHING_NEARCLIP_ENABLE_FAR_DISTANCE,dis2) );
				
				if(RAYMARCHING_NEAR_DEPTH_FIX == true)
					cloudDensity = lerp(0,cloudDensity, saturate((EndDis2 - dis2) / stepSize));

				minLen = min(minLen,dis2);

				half lightDensity = sampleCloudDensityAlongConeArcApple(pos , lightDotEye, cloudDensity, result.a > 0.7, viewDir);
				float4 source = half4((cSunColor.rgb * ThunderRate * lightDensity * cCloudSunIntensity * 5.0) + cCloudAmbient.rgb, cloudDensity);
				source.rgb *= source.a;
				result = clamp((1.0 - result.a) * source + result, 0, 16.0);

				if (result.a >= 1.0) 
				{
					break;
				}

				//FOR DEBUG
				if(DEBUG_VISUAL_MODE == 2)
					DebugDensityVisualCount += 5.0;
			}
		}

		if (dis2 > EndDis2)
		{						
			minLen = min(minLen,dis2 - stepSize);
			break;
		}

		pos += posStep;
		dis2 += stepSize;

		//FOR DEBUG
		DebugDensityVisualCount += 1.0;

	}

	float2 depthMinMax = (float2(minLen,maxLen) - CameraInfo.xx) / (CameraInfo.y - CameraInfo.x);
	float center = depthMinMax.y + depthMinMax.x;
	float extend = depthMinMax.y - depthMinMax.x;
	OUT.Weight = float2(center,extend) * 0.5;

	if(@_FOG && !@_VolumetricFog)
	{
		if(!@_DEFERRED_SHADING)
		{
			float T = FogTransmittance(CameraPos.xyz + viewDir * CameraInfo.y);
			float3 fogFinalColor = float3(0,0,0);
			FogLightingAndScatter(T,fogFinalColor, -viewDir, SunDirection.xyz);
			float Alpha = T;
			result.xyz = lerp(fogFinalColor * result.a , result.rgb, Alpha);
		}
	}

//*/*/*/*/*/*/*/*/*/*
	if (DebugDensityVisualCount > 1.0 && DEBUG_VISUAL_MODE != 0)
	{
		// if (DEBUG_VISUAL_MODE == 1)
		// {
			DebugDensityVisualCount = DebugDensityVisualCount / ((float)cCloudRayMarchingCount * DEBUG_VISUAL_LOOP_DEFAULT);
			if (DebugDensityVisualCount > 1.0)
			{
				OUT.Color = float4(1, 0, 1, 1);
				return OUT;
			}
			OUT.Color = lerp(float4(0, 1, 0, 1), float4(1, 0, 0, 1), DebugDensityVisualCount);
		// }

		// if (DEBUG_VISUAL_MODE == 2)
		// {
		// 	//	return lerp(float4(1,0,0,1),float4(0,0,1,1),abs(viewDir.y));
		// 	OUT.Color = float4((float)traceLen / (float)DEBUG_traceSrc, 0, 0, 1);
		// 	OUT.Color = float4((float)sampleCount / (float)cCloudRayMarchingCount, 0, 0, 1);
		// }
		return OUT;
	}
//*/*/*/*/*/*/*/*/*/*/*

	OUT.Color = result;
	return OUT;
}

VCPS CalcCloudColorArcCake(VertexShadingOutput IN)
{
	float2 JitteredUV = IN.UV + IN.CurrentSubPixel.xy * ScreenInfo.zw ;
	
	float3 viewDir = normalize(CalcHomogeneousPos(1, JitteredUV));

	float depth = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV, 0));
	float d1,d2,d3,d4;
	d1 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(0,1) * 0.25 * ScreenInfo.zw, 0));
	d2 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(1,0) * 0.25 * ScreenInfo.zw, 0));
	d3 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(0,-1) * 0.25 * ScreenInfo.zw, 0));
	d4 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(-1,0) * 0.25 * ScreenInfo.zw, 0));
	float maxInPixelDepth = max(d1,max(d2,max(d3,d4)));

	if(RAYMARCHING_ENABLE_DEPTH_EXTEND)
	{
		float d5 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(-5,5) * 0.25 * ScreenInfo.zw, 0));
		float d6 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(5,-5) * 0.25 * ScreenInfo.zw, 0));
		float d7 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(5,5) * 0.25 * ScreenInfo.zw, 0));
		float d8 = DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, JitteredUV + float2(-5,-5) * 0.25 * ScreenInfo.zw, 0));
		maxInPixelDepth = max(maxInPixelDepth,max(d5,max(d6,max(d7,d8))));
	}

	//Construct RayMarching Aera
	float2 traceLen = ( cCloudHeightMinMax.xy - CameraPos.yy ) / viewDir.y;
	if(CameraPos.y < cCloudHeightMinMax.x)
		traceLen.xy = traceLen.xy;
	else if(CameraPos.y > cCloudHeightMinMax.y)
		traceLen.xy = traceLen.yx;
	else
		traceLen.xy = float2(0,max(traceLen.x,traceLen.y));
	traceLen.y = min(traceLen.y,cMaxTraceDistance);

	VCPS rst = traceClouds(IN.CurrentSubPixel.x, viewDir, CameraPos.xyz + viewDir * traceLen.x, traceLen.y - traceLen.x, depth, maxInPixelDepth, maxInPixelDepth);
	return rst;
}

VertexShadingOutput VolumetricCloudVS(a2v IN)
{
	VertexShadingOutput OUT;
	OUT.CurrentSubPixel = float4(0,0,0,0);

	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.HPosition.z = (HALF_EPSILON * 5) * OUT.HPosition.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.
	OUT.UV = IN.BaseTC;

	float CurrentIndex = floor(UPSCALE_DITHER_SIZE * UPSCALE_DITHER_SIZE * frac(GameTime*8.71828) );
	float2 AnimatedUV = float2(0,0);
	AnimatedUV.x = fmod(CurrentIndex , UPSCALE_DITHER_SIZE);
	AnimatedUV.y = floor(CurrentIndex / UPSCALE_DITHER_SIZE);
	float2 OffsetUV = AnimatedUV / float2(UPSCALE_DITHER_SIZE,UPSCALE_DITHER_SIZE);
	float2 Offset = tBlueNoise.SampleLevel(sBlueNoise,OffsetUV,0).xy;
	OUT.CurrentSubPixel.xy = Offset;
	OUT.CurrentSubPixel.zw = fmod(ScreenInfo.xy,float2(2,2));

	return OUT;
}


VCPS VolumetricCloudPS(VertexShadingOutput IN)
{
	return CalcCloudColorArcCake(IN);
}
technique VolumetricCloudApplyToScreen
{
	VertexShader = VolumetricCloudVS();
	PixelShader = VolumetricCloudCompositePS();
}

technique VolumetricCloudRayMarching
{
	VertexShader = VolumetricCloudVS();
	PixelShader = VolumetricCloudPS();
}

technique VolumetricCloudTemporalUpscale
{
	VertexShader = VolumetricCloudVS();
	PixelShader = VolumetricCloudUpscalePS();
}

///////////////////////////VolumetricCloudRayMarchingCS/////////////////////////////
RWTexture2D<float4> TextureOut		: layout(u0);
RWTexture2D<float2> DepthOut		: layout(u1);

[numthreads(8, 8, 1)]
void VolumetricCloudCS(int3 GlobalID : SV_DispatchThreadID)
{
	if (GlobalID.x < int(ScreenInfo.x) && GlobalID.y < int(ScreenInfo.y))
	{
		//Same As VertexShading
		float2 stc = float2(GlobalID.xy + 0.5) * ScreenInfo.zw;
		VertexShadingOutput IN;
		IN.UV = stc;
		IN.HPosition = float4(0,0,0,0);

		float CurrentIndex = floor(UPSCALE_DITHER_SIZE * UPSCALE_DITHER_SIZE * frac(GameTime*8.71828) );
		float2 AnimatedUV = float2(0,0);
		AnimatedUV.x = fmod(CurrentIndex , UPSCALE_DITHER_SIZE);
		AnimatedUV.y = floor(CurrentIndex / UPSCALE_DITHER_SIZE);
		float2 OffsetUV = AnimatedUV / float2(UPSCALE_DITHER_SIZE,UPSCALE_DITHER_SIZE);
		float2 Offset = tBlueNoise.SampleLevel(sBlueNoise,OffsetUV,0).xy;
		IN.CurrentSubPixel.xy = Offset;
		IN.CurrentSubPixel.zw = fmod(ScreenInfo.xy,float2(2,2));

		VCPS OUT = CalcCloudColorArcCake(IN);

		TextureOut.Store(GlobalID.xy, OUT.Color);
		DepthOut.Store(GlobalID.xy, float4(OUT.Weight,0,0));
	}
}

technique VolumetricCloudRayMarchingCS
{
	ComputeShader = VolumetricCloudCS();
}

/////////////////////////////GodRay Setup/////////////////////////
//cGodRayRange
//cSunPos
float4 GodRaySetupPS(v2f IN) : SV_Target
{
	float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferLinearSampler, IN.TC));
	if (depth < 1 - HALF_EPSILON)
		return float4(0, 0, 0, 0);

	float Alpha = tScatteringAndAlpha.Sample(sScatteringAndAlpha, IN.TC).a;
	float2 blurDir = cSunPos - IN.TC;
	blurDir.x *= cAspectRatio;
	float t = smoothstep(cGodRayRange, 1, 1 - length(blurDir));
	float3 GodRayColor = cGodRayColor.rgb * (1 - Alpha) * t * t;
	return float4(GodRayColor, 1);
}

technique VolumetricCloudGodRaySetup
{
	VertexShader = ScreenVS();
	PixelShader = GodRaySetupPS();
}

/////////////////////////////GodRay Blur/////////////////////////
#define FetchGodRayBlur(begTC){\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC).rgb;\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 1.0/8 * blurDir).rgb * Square(7.0/8);\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 2.0/8 * blurDir).rgb * Square(6.0/8);\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 3.0/8 * blurDir).rgb * Square(5.0/8);\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 4.0/8 * blurDir).rgb * Square(4.0/8);\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 5.0/8 * blurDir).rgb * Square(3.0/8);\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 6.0/8 * blurDir).rgb * Square(2.0/8);\
	blurValue += tGodRaySetupMap.Sample(sGodRaySetupMap, begTC + 7.0/8 * blurDir).rgb * Square(1.0/8);\
}
float4 GodRayBlurPS(v2f IN) : SV_Target
{
	float2 currPos = IN.TC;
	float2 blurDir = cSunPos - IN.TC;
	blurDir.x *= cAspectRatio;
	float blurDistance = length(blurDir);
	blurDir /= blurDistance;
	blurDistance = min(blurDistance, cMaxBlurDistance);
	blurDir *= blurDistance;
	blurDir.x /= cAspectRatio;
	float3 blurValue = float3(0, 0, 0);

	//TAA
	// float2 spos = IN.HPosition.xy;
	// [@_D3D9]
	// spos += 0.5;
	// float2 stc = spos * ScreenInfo.zw;
	// float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc));
	// float3 worldPos = CameraPos.xyz + CalcHomogeneousPos(depth, stc);
	// float4 lastPos = MulMat4x4(cLastViewProjTex, float4(worldPos,1));
	// float2 lastUV = lastPos.xy / lastPos.w;
	// if (lastUV.x > 0 && lastUV.x < 1 && lastUV.y > 0 && lastUV.y < 1)
	// {
	// 	float2 begTC = IN.TC + cJitter/8 * blurDir;
	// 	FetchGodRayBlur(begTC);
	// 	blurValue /= 8;
	// 	float3 history = tLastGodRayMap.Sample(sLastGodRayMap, lastUV).rgb;
	// 	blurValue = lerp(blurValue, history, 0.95);
	// }
	// else//calc all pixels....
	{
		float2 begTC = IN.TC + 0.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 1.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 2.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 3.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 4.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 5.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 6.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		begTC = IN.TC + 7.5 / 64 * blurDir; FetchGodRayBlur(begTC);
		blurValue /= 64;
	}
	return float4(blurValue, 0);
}

technique VolumetricCloudGodRayBlur
{
	VertexShader = ScreenVS();
	PixelShader = GodRayBlurPS();
}


float4 GodRayFinalPS(v2f IN) : SV_Target
{
	float3 blendColor = tGodRayFinalMap.Sample(sGodRayFinalMap, IN.TC).rgb;
	return float4(blendColor, 0);
}

technique VolumetricCloudGodRayFinal
{
	VertexShader = ScreenVS();
	PixelShader = GodRayFinalPS();
}
