//	--------------------------------------------------------------------
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<VolumetricFog.fx>
///	@path	~/System/Shader/
///	@date	2019/01/21
///	@brief

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "Noise.fxh"

#define VOLUMETRIC_SHADING

///////////////// properties //////////////////
@_EnableVolumetricShadow
{
	Name = EnableVolumetricShadow;
	Description = "Apply volumetic shadow";
	Fallback = false;
	Default = false;
}

@_EnableSpotShadow
{
	Name = EnableSpotShadow;
	Description = "Apply spot shadow";
	Fallback = false;
	Default = false;
}

@_IsFirstFrame
{
	Name = IsFirstFrame;
	Description = "Is First Frame";
	Default = false;
}

@_LightsCount0
{
	Name = LightsCount0;
	Description = "Lights Count 0";
	Fallback = false;
	Default = false;
}

@_LightsCount1
{
	Name = LightsCount1;
	Description = "Lights Count 1";
	Fallback = false;
	Default = false;
}

@_LightsCount2
{
	Name = LightsCount2;
	Description = "Lights Count 2";
	Fallback = false;
	Default = false;
}

@_EnableDetailFog
{
	Name = EnableDetailFog;
	Description = "Enable Detail Fog";
	Fallback = false;
	Default = false;
}

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////// samplers //////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tHistoryMedia : layout(t1);
sampler sHistoryMedia : layout(s1)
{
	Texture = tHistoryMedia;
	Filter = Linear;
	Address = Clamp;
};

texture2D tHistoryRT : layout(t2);
sampler sHistoryRT : layout(s2)
{
	Texture = tHistoryRT;
	Filter = Linear;
	Address = Clamp;
};

texture2D tESM : layout(t3);
sampler sESM : layout(s3)
{
	Texture = tESM;
	Filter = Linear;
	Address = Clamp;
};

texture2D tLastStep : layout(t0);
sampler sLastStep : layout(s0)
{
	Texture = tLastStep;
	Filter = Point;
	Address = Clamp;
};
texture2D tShadowMap : layout(t4);
SamplerComparisonState sShadowMapSampler : layout(s4)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
// texture2DArray tShadowMapArray : layout(t9);
// SamplerComparisonStateArray sShadowMapArraySampler : layout(s9)
// {
// 	Texture = tShadowMapArray;
// 	Filter = Linear;
// 	Address = Clamp;
// 	ComparisonFunc = Greater;
// };

texture2D tWeatherData : layout(t5);
sampler sWeatherData : layout(s5)
{
	Texture = tWeatherData;
	Filter = Linear;
	Address = Clamp;
};

///////////////// const buffer //////////////////
#define JITTER_COUNT 16
#define MAX_LIGHTS_COUNT 7
cbuffer Batch
{
	float4 vWBasisX							;
	float4 vWBasisY							;
	float4 vWBasisZ							;
	float3 cGlobalFogAlbedo					;
	float cGlobalFogPhaseG					;
	float cFogBaseHeight					;
	float cGlobalFogDensity					;		// cGlobalFogDensity = GlobalFogExtinctionScale * FOG_DENSITY
	float cVolumetricFogDistance			;
	float cFogK								;		// cFogK = 5.0/FOG_HEIGHT
	float4x4 cShadowViewProjTex				;
	float3 cVolumetricEmissive				;
	float cPlaceHold0						;
	float4x4 lastViewProjTex				;
	float4 vFogInfo				: FogInfo	;
	float cVolumetricFogZNear				;
	float cIntegrationStep					;
	float cCheckerboardSampleBias			;
	float cInvShadowMapSize					;
	float4 cCoverage						;
	float2 cFogHeightMinMax           		;
	float cSpotInvShadowMapSize				;
	float cCoverageCutoff					;
	float3 cDetailFogAlbedoA				;
	float cPlaceHold1 						;
	float3 cDetailFogAlbedoB				;
	float cPlaceHold2 						;
	float4x4 cSpotShadowViewProjTex			;
	float4 cJitterOffset[JITTER_COUNT]		;
	float4 cLightsData[4*MAX_LIGHTS_COUNT]	;
};

///////////////// pre defines & helper functions //////////////////
#include "Lighting.fxh"
int GetLightsCount()
{
	int count = 0;
	[@_LightsCount2]
	count += 4;
	
	[@_LightsCount1]
	count +=2;
	
	[@_LightsCount0]
	count += 1;

	return count;
}
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}
//celloffset
float3 GridIdxToWorldPos(float3 gridIdx, float3 celloffset, out float SceneDepth)
{
	float2 CellScreenPos = gridIdx.xy + celloffset.xy;
	SceneDepth = SLiceIdxToDepth(gridIdx.z + celloffset.z);
	float3 cameraToCell = CalcHomogeneousPos(SceneDepth*CameraInfo.z, CellScreenPos / GRID_SIZE.xy);
	float3 cellWorldPos = CameraPos.xyz + cameraToCell;
	return cellWorldPos;
}
// cGlobalFogDensity = GlobalFogExtinctionScale * FOG_DENSITY
// cFogK = 5.0/FOG_HEIGHT
float GlobalFogExtinction(in float3 worldPos)
{
	return cGlobalFogDensity * exp(-cFogK*max(0, worldPos.y - cFogBaseHeight));
}
/////////////////// MediaSetupByGlobalFog ///////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}
// Utility function that maps a value from one range to another.
float Remap(float original_value, float original_min, float original_max, float new_min, float new_max)
{
    return new_min + (saturate((original_value - original_min) / (original_max - original_min)) * (new_max - new_min));
}
//weather_data.r fogHeightBegin
//weather_data.g fogHeightLen
//weather_data.b extinction
//weather_data.a fog color selection
float4 SampleWeather(float3 pos)
{
	float2 uv = (pos.xz-cCoverage.xy) / cCoverage.zw;
	return tWeatherData.SampleLevel(sWeatherData, uv, 0).rgba;
}
float GetExtinction(float4 weatherData, float yPos)
{
	float y = (yPos - cFogHeightMinMax.x) / (cFogHeightMinMax.y - cFogHeightMinMax.x);
	y = saturate(y);
	float bh = min(weatherData.r, 1-EPSILON);
	float lenh = min(weatherData.g, 1-bh);

	return smoothstep(cCoverageCutoff, 1.0, weatherData.b) * Remap(y, bh, bh+0.25*lenh, 0, 1) * Remap(y, bh+0.75*lenh, bh+lenh, 1, 0);
}
float3 GetAlbedo(float4 weatherData)
{
	return lerp(cDetailFogAlbedoA, cDetailFogAlbedoB, weatherData.a);
}
float4 GetMediaScatteringAndAbsorption(float3 worldPos)
{
	float Extinction = GlobalFogExtinction(worldPos);
	float3 Scattering = cGlobalFogAlbedo * Extinction;
	float Absorption = max(Extinction - GetLuminance(Scattering), 0);

	//sea fog
	if (@_EnableDetailFog)
	{
		float4 WeatherData = SampleWeather(worldPos);
		float DetailFogExtinction = GetExtinction(WeatherData, worldPos.y);// * saturate(cnoise(worldPos * 0.1));
		float3 DetailFogAlbedo = GetAlbedo(WeatherData);
		float3 DetailFogScattering = DetailFogAlbedo * DetailFogExtinction;
		float DetailFogAbsorption = max(DetailFogExtinction - GetLuminance(DetailFogScattering), 0);
		Scattering += DetailFogScattering;
		Absorption += DetailFogAbsorption;
	}

	return float4(Scattering, Absorption);
}

//SliceSize
//ZA,ZB
//Volume_vWBasisXYZ
//cFogBaseHeight,cGlobalFogDensity,cGlobalFogAlbedo,cFogK
//lastViewProjTex,cVolumetricFogZNear,cVolumetricFogDistance,
float4 MediaSetupByGlobalFogPS(v2f IN) : SV_Target
{
	//calc worldpos
	float2 spos = IN.HPosition.xy;
	[!@_D3D9]
	spos -= 0.5;
	float3 gridIdx = HPositionToGridIdx(spos);
	float sceneDepth;
	float3 worldPos = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), sceneDepth);

	//calc samples
	float4 lastVolumeUV = WorldPosToVolumeUV(worldPos, lastViewProjTex);
	int NumSuperSamples = 1;
	if (@_IsFirstFrame)
	{
		NumSuperSamples += (JITTER_COUNT-1);
	}
	else
	{
		float3 cmp1 = step(lastVolumeUV.xyw, float3(0, 0, 2*cVolumetricFogZNear));
		float3 cmp2 = step(float3(1, 1, cVolumetricFogDistance), lastVolumeUV.xyw);
		float cmp3 = step(0.01, dot(cmp1+cmp2, float3(1,1,1)));
		NumSuperSamples += (JITTER_COUNT-1) * (int)cmp3;
	}

	//calc curr frame
	float4 Curr = float4(0, 0, 0, 0);
	for (int SampleIndex = 0; SampleIndex < NumSuperSamples; ++SampleIndex)
	{
		float3 _worldPos = GridIdxToWorldPos(gridIdx, cJitterOffset[SampleIndex].xyz, sceneDepth);
		Curr += GetMediaScatteringAndAbsorption(_worldPos);
	}
	Curr /= NumSuperSamples;

	if (@_IsFirstFrame)
	{
		return Curr;
	}
	else
	{
		//temporal filter
		float historyWeight = 0.95 * float(NumSuperSamples == 1);
		float4 History;
		lastVolumeUV.xy = clamp(lastVolumeUV.xy, 0.5/SliceSize, 1-0.5/SliceSize);//avoiding wrong bilinear filter between two slices.
		SampleVolume(tHistoryMedia, sHistoryMedia, lastVolumeUV, History);
		return lerp(Curr, History, historyWeight);
	}
}
technique MediaSetupByGlobalFog
{
	VertexShader = ScreenVS();
	PixelShader = MediaSetupByGlobalFogPS();
}

/////////////////// InjectSunLighting ///////////////
float VolumetricShadow(in float3 worldPos)
{
	if (@_EnableVolumetricShadow)
	{
#if SD_FORWARD_CSM
		if (1)
#else
		if (@_DEFERRED_SHADING)
#endif
		{
			float shadowBiasScale = 1;
			float4 P_LS = 0;
			float InShadow = GetInShadowFromCSM(false, worldPos, float3(0,1,0), 1, P_LS);
			return 1-InShadow;
		}
		else
		{
			//light space uvz
			float4 P_LS = mul(float4(worldPos,1), cShadowViewProjTex);
			P_LS.xyz /= P_LS.w;

			if (dot(step(P_LS.xyz, float3(1,1,1)), step(float3(0,0,0), P_LS.xyz)) < 2.99)
				return 1;

			float2 InvShadowMapSizeXY = cInvShadowMapSize.xx;
#if SD_FORWARD_CSM
			P_LS.x = 0.5 + 0.5*P_LS.x;
			InvShadowMapSizeXY.x *= 0.5;
#endif
			float InShadow = 0;
			hardware_pcf_2x2(P_LS.xy, P_LS.z, InvShadowMapSizeXY, InShadow);
			return 1-InShadow;
		}
	}
	return 1;
}
//SliceSize
//ZA,ZB
//Volume_vWBasisXYZ
//lastViewProjTex, cVolumetricFogZNear, cVolumetricFogDistance, cGlobalFogPhaseG, cVolumetricEmissive
//cJitterOffset
//@_EnableVolumetricShadow,cShadowViewProjTex,tShadowMap
float4 InjectSunLightingPS(v2f IN) : SV_Target
{
	//calc worldpos
	float2 spos = IN.HPosition.xy;
	[!@_D3D9]
	spos -= 0.5;
	float3 gridIdx = HPositionToGridIdx(spos);
	float sceneDepth;
	float3 worldPos = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), sceneDepth);

	//calc samples
	float4 lastVolumeUV = WorldPosToVolumeUV(worldPos, lastViewProjTex);
	int NumSuperSamples = 1;
	if (@_IsFirstFrame)
	{
		NumSuperSamples += (JITTER_COUNT-1);
	}
	else
	{
		float3 cmp1 = step(lastVolumeUV.xyw, float3(0, 0, 2*cVolumetricFogZNear));
		float3 cmp2 = step(float3(1, 1, cVolumetricFogDistance), lastVolumeUV.xyw);
		float cmp3 = step(0.01, dot(cmp1+cmp2, float3(1,1,1)));
		NumSuperSamples += (JITTER_COUNT-1) * (int)cmp3;
	}
	
	//light scattering
	float3 LightScattering = float3(0, 0, 0);
	[loop]
	for (int SampleIndex = 0; SampleIndex < NumSuperSamples; ++SampleIndex)
	{
		float _sceneDepth;
		float3 _offset = cJitterOffset[SampleIndex].xyz;
		float3 _worldPos = GridIdxToWorldPos(gridIdx, _offset, _sceneDepth);
		float3 _WorldToCamera = normalize(CameraPos.xyz - _worldPos);
		//todo indirect lighting volume
		LightScattering += SunColor.rgb * VolumetricShadow(_worldPos) * PhaseFunction(cGlobalFogPhaseG, dot(-SunDirection.xyz, _WorldToCamera)) + cVolumetricEmissive;
		[unroll] for (int i = 0; i < GetLightsCount(); ++i)
		{
			int beg = 4 * i;
			float3 L = cLightsData[beg+0].xyz - _worldPos;
			half D2 = dot(L, L);
			half RangeMask = D2 * cLightsData[beg+0].w;
			if (RangeMask < 1)//in range
			{
				half D = sqrt(D2);
				L /= D;
				half DoL = dot( cLightsData[beg+2].xyz, -L );
				half Atten = 1 / (D2*cLightsData[beg+1].w+1) * Square(saturate(1 - Square(RangeMask)));
				[branch]if (cLightsData[beg+2].w < 0)//Point[1], Spot[-1]
				{
					Atten *= Square(saturate(DoL*cLightsData[beg+3].y + cLightsData[beg+3].z));
					if (@_EnableSpotShadow)
					{
						if (i==0)
						{
							//light space uvz
							float4 P_LS = mul(float4(_worldPos,1), cSpotShadowViewProjTex);
							P_LS.xyz /= P_LS.w;
							if (dot(step(P_LS.xyz, float3(1,1,1)), step(float3(0,0,0), P_LS.xyz)) > 2.99)
							{
								float2 InvShadowMapSizeXY = cSpotInvShadowMapSize.xx;
								float InShadow = 0;
								hardware_pcf_2x2(P_LS.xy, P_LS.z, InvShadowMapSizeXY, InShadow);
								Atten *= saturate(1 - InShadow);
							}
						}
					}
				}
				LightScattering += cLightsData[beg+3].r * cLightsData[beg+1].rgb * Atten * PhaseFunction(cGlobalFogPhaseG, dot(-L, _WorldToCamera));
			}
		}
	}
	LightScattering /= NumSuperSamples;
	
	//Curr
	float2 inSliceUV = min(gridIdx.xy + 1, GRID_SIZE.xy-0.5) / GRID_SIZE.xy;
	float2 uv = SliceUVTo2DUV(inSliceUV, gridIdx.z); 
	float4 MediaScatteringAndAbsorption = tHistoryMedia.Sample(sHistoryMedia, uv);
	float Extinction = MediaScatteringAndAbsorption.w + GetLuminance(MediaScatteringAndAbsorption.xyz);
	float4 Curr = float4(LightScattering*MediaScatteringAndAbsorption.xyz, Extinction);
	
	//temporal filter
	float historyWeight = 0.98 * float(NumSuperSamples == 1);
	float4 History;
	lastVolumeUV.xy = clamp(lastVolumeUV.xy, 0.5/SliceSize, 1-0.5/SliceSize);//avoiding wrong bilinear filter between two slices.
	SampleVolume(tHistoryRT, sHistoryRT, lastVolumeUV, History);
	return lerp(Curr, History, historyWeight);
}
technique InjectSunLighting
{
	VertexShader = ScreenVS();
	PixelShader = InjectSunLightingPS();
}

/////////////////// FirstIntegration ///////////////
//CurrRT -> TempRT
//SliceSize
//ZA,ZB
float4 FirstIntegrationPS(v2f IN) : SV_Target
{
	//calc gridIdx
	float2 spos = IN.HPosition.xy;
	[!@_D3D9]
	spos -= 0.5;
	float3 gridIdx = HPositionToGridIdx(spos);
	
	//curr value
	float2 inSliceUV = (gridIdx.xy + 0.5) / GRID_SIZE.xy;
	float2 currUV = SliceUVTo2DUV(inSliceUV, gridIdx.z);
	float4 currScatteringAndExtinction = tLastStep.Sample(sLastStep, currUV);
	float currDepth;
	float3 currPos = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), currDepth);

	//pre1
	gridIdx.z -= 1;
	float3 prePos1 = CameraPos.xyz;
	float4 preScatteringAndExtinction1 = float4(0, 0, 0, 0);
	if (gridIdx.z > -0.001)
	{
		float preDepth1 = 0;
		float2 preUV1 = SliceUVTo2DUV(inSliceUV, gridIdx.z);
		preScatteringAndExtinction1 = tLastStep.Sample(sLastStep, preUV1);
		prePos1 = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), preDepth1);
	}

	//pre2
	gridIdx.z -= 1;
	float3 prePos2 = CameraPos.xyz;
	float4 preScatteringAndExtinction2 = float4(0, 0, 0, 0);
	if (gridIdx.z > -0.001)
	{
		float preDepth2 = 0;
		float2 preUV2 = SliceUVTo2DUV(inSliceUV, gridIdx.z);
		preScatteringAndExtinction2 = tLastStep.Sample(sLastStep, preUV2);
		prePos2 = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), preDepth2);
	}

	//pre3
	gridIdx.z -= 1;
	float3 prePos3 = CameraPos.xyz;
	float4 preScatteringAndExtinction3 = float4(0, 0, 0, 0);
	if (gridIdx.z > -0.001)
	{
		float preDepth3 = 0;
		float2 preUV3 = SliceUVTo2DUV(inSliceUV, gridIdx.z);
		preScatteringAndExtinction3 = tLastStep.Sample(sLastStep, preUV3);
		prePos3 = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), preDepth3);
	}

	//pre4
	gridIdx.z -= 1;
	float3 prePos4 = CameraPos.xyz;
	if (gridIdx.z > -0.001)
	{
		float preDepth4 = 0;
		prePos4 = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), preDepth4);
	}

	float stepLen1 = length(currPos - prePos1);
	float stepLen2 = length(prePos1 - prePos2);
	float stepLen3 = length(prePos2 - prePos3);
	float stepLen4 = length(prePos3 - prePos4);
	float T1 = exp(-currScatteringAndExtinction.w * stepLen1);
	float T2 = exp(-preScatteringAndExtinction1.w * stepLen2);
	float T3 = exp(-preScatteringAndExtinction2.w * stepLen3);
	float T4 = exp(-preScatteringAndExtinction3.w * stepLen4);
	float3 Scattering = currScatteringAndExtinction.xyz * stepLen1;
	Scattering = Scattering * T2 + preScatteringAndExtinction1.xyz * stepLen2;
	Scattering = Scattering * T3 + preScatteringAndExtinction2.xyz * stepLen3;
	Scattering = Scattering * T4 + preScatteringAndExtinction3.xyz * stepLen4;
	float Transmittance = (T1 * T2) * (T3 * T4);
	return float4(Scattering, Transmittance);
}

technique FirstIntegration
{
	VertexShader = ScreenVS();
	PixelShader = FirstIntegrationPS();
}

/////////////////// StepIntegration ///////////////
//TempRT->HistoryRT
//SliceSize
//ZA,ZB
//cIntegrationStep
float4 StepIntegrationPS(v2f IN) : SV_Target
{
	//calc gridIdx
	float2 spos = IN.HPosition.xy;
	[!@_D3D9]
	spos -= 0.5;
	float3 gridIdx = HPositionToGridIdx(spos);
	
	//curr value
	float2 inSliceUV = (gridIdx.xy + 0.5) / GRID_SIZE.xy;
	float2 currUV = SliceUVTo2DUV(inSliceUV, gridIdx.z);
	float4 currScatteringAndTransmittance = tLastStep.Sample(sLastStep, currUV);

	//pre1
	gridIdx.z -= cIntegrationStep;
	float4 preScatteringAndTransmittance1 = float4(0, 0, 0, 1);
	if (gridIdx.z >= -0.001)
	{
		float2 preUV = SliceUVTo2DUV(inSliceUV, gridIdx.z);
		preScatteringAndTransmittance1 = tLastStep.Sample(sLastStep, preUV);
	}

	//pre2
	gridIdx.z -= cIntegrationStep;
	float4 preScatteringAndTransmittance2 = float4(0, 0, 0, 1);
	if (gridIdx.z >= -0.001)
	{
		float2 preUV = SliceUVTo2DUV(inSliceUV, gridIdx.z);
		preScatteringAndTransmittance2 = tLastStep.Sample(sLastStep, preUV);
	}
	
	//pre3
	gridIdx.z -= cIntegrationStep;
	float4 preScatteringAndTransmittance3 = float4(0, 0, 0, 1);
	if (gridIdx.z >= -0.001)
	{
		float2 preUV = SliceUVTo2DUV(inSliceUV, gridIdx.z);
		preScatteringAndTransmittance3 = tLastStep.Sample(sLastStep, preUV);
	}

	float3 Scattering = currScatteringAndTransmittance.xyz;
	Scattering = Scattering * preScatteringAndTransmittance1.w + preScatteringAndTransmittance1.xyz;
	Scattering = Scattering * preScatteringAndTransmittance2.w + preScatteringAndTransmittance2.xyz;
	Scattering = Scattering * preScatteringAndTransmittance3.w + preScatteringAndTransmittance3.xyz;
	float Transmittance = (currScatteringAndTransmittance.w * preScatteringAndTransmittance1.w)
		* (preScatteringAndTransmittance2.w * preScatteringAndTransmittance3.w);
	return float4(Scattering, Transmittance);
}

technique StepIntegration
{
	VertexShader = ScreenVS();
	PixelShader = StepIntegrationPS();
}

RWTexture2D<float4> TextureVolumeExtinction			: layout(u0);
RWTexture2D<float4> TextureVolumeTransmittance		: layout(u1);

/////////////////////// CalcExtinction //////////////////
// [numthreads(8, 8, 1)]
// void ScatteringExtinctionSetupCS(int3 GlobalID : SV_DispatchThreadID)
// {
// 	if (GlobalID.x < int(ScreenInfo.x + 0.5) && GlobalID.y < int(ScreenInfo.y + 0.5))
// 	{
// 		float3 gridIdx = HPositionToGridIdx(float2(GlobalID.x, GlobalID.y));
// 		float sceneDepth;
// 		float3 worldPos = GridIdxToWorldPos(gridIdx, float3(0.5, 0.5, 0.5), sceneDepth);

// 		float4 MediaScatteringAndAbsorption = GetMediaScatteringAndAbsorption(worldPos);
// 		float3 LightScattering = GetLightScattering(gridIdx, worldPos);

// 		float Extinction = MediaScatteringAndAbsorption.w + GetLuminance(MediaScatteringAndAbsorption.xyz);
// 		float4 curr = float4(LightScattering*MediaScatteringAndAbsorption.xyz, Extinction);
// 		float4 ScatteringAndExtinction = TemporalFilter(curr, worldPos);
// 		TextureVolumeExtinction.Store(GlobalID.xy, ScatteringAndExtinction);
// 	}
// }

// technique ScatteringExtinctionSetup
// {
// 	ComputeShader = ScatteringExtinctionSetupCS();
// }

/////////////////// IntegrationTransmittance ///////////////
[numthreads(8, 8, 1)]
void IntegrationCS(int3 GlobalID : SV_DispatchThreadID)
{
	int2 SliceSizeInt = int2(SliceSize.x + 0.1, SliceSize.y + 0.1);
	if (GlobalID.x < SliceSizeInt.x && GlobalID.y < SliceSizeInt.y)
	{
		float4 currScatteringAndTransmittance = float4(0, 0, 0, 1);
		float3 currPos = CameraPos.xyz;

		int3 gridIdx = int3(GlobalID.x, GlobalID.y, 0);
		for (int z = 0; z < SLICE_COUNT; ++z)
		{
			gridIdx.z = z;
			int row = gridIdx.z / SLICE_PERROW;
			int column = gridIdx.z - row * SLICE_PERROW;
			int2 uv = gridIdx.xy + int2(column, row) * SliceSizeInt;

			float2 inSliceUV = (float2(gridIdx.xy) + 0.5) / GRID_SIZE.xy;
			float2 currUV = SliceUVTo2DUV(inSliceUV, float(gridIdx.z));
			float4 nextScatteringAndExtinction = tLastStep.SampleLevel(sLastStep, currUV, 0);

			float nextDepth;
			float3 nextPos = GridIdxToWorldPos(float3(gridIdx), float3(0.5, 0.5, 0.5), nextDepth);
			float stepLen = length(currPos - nextPos);

			float3 Scattering = currScatteringAndTransmittance.rgb + nextScatteringAndExtinction.rgb * stepLen * currScatteringAndTransmittance.a;
			float Transmittance = exp(-nextScatteringAndExtinction.a * stepLen) * currScatteringAndTransmittance.a;

			currScatteringAndTransmittance = float4(Scattering, Transmittance);
			currPos = nextPos;
			TextureVolumeTransmittance.Store(uv, currScatteringAndTransmittance);
		}
	}
}

technique Integration
{
	ComputeShader = IntegrationCS();
}

/////////////////// VolumetricFogOpaque ////////////////////
float4 VolumetricFogOpaquePS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;

	if (@_VolumetricFog)
	{
		float2 stc = spos * ScreenInfo.zw;
		float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc)) * CameraInfo.y;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		return float4(ScatteringAndTransmittance.xyz, 1 - ScatteringAndTransmittance.w);
	}
	else
	{
		float2 stc = spos * ScreenInfo.zw;
		float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc));
		float3 cameraToWorld = CalcHomogeneousPos(depth, stc);
		float ViewDistance = length(cameraToWorld);
		float3 ViewDirection = cameraToWorld / ViewDistance;
		float T = FogTransmittance(cameraToWorld);
		float3 fogFinalColor = float3(0,0,0);
		FogLightingAndScatter(T, fogFinalColor, -ViewDirection, SunDirection.xyz);
		float Alpha = 1 - T;
		return float4(fogFinalColor * Alpha, Alpha);
	}
}
technique VolumetricFogOpaque
{
	VertexShader = ScreenVS();
	PixelShader = VolumetricFogOpaquePS();
}

