//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : huanghailiang@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<PostProcess.fx>
///	@path	~/System/Shader/
///	@date	2020/07/16
///	@brief	Post process utils.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

// Source target.
texture2D tBaseMap : layout(t0) = "Transparent";
sampler sBaseMap_Sampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tMakeup : layout(t1) = "Gray";
sampler sMakeupSampler : layout(s1)
{
	Texture = tMakeup;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

cbuffer Shader
{
	float4 srcUVTransform		=	{0,0,1,1};
	float4 dstUVTransform		=	{ 0,0,1,1};
	float4 blendColor			=	{1,1,1,1};
};

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 TargetBakePS(in VertexScreenOutput IN) : SV_Target
{
	float4 OUT = float4(0,0,0,0);

	OUT = tBaseMap.Sample(sBaseMap_Sampler, IN.TC.xy);
	if (IN.TC.x > srcUVTransform.x && IN.TC.x< srcUVTransform.x + srcUVTransform.z && IN.TC.y> srcUVTransform.y && IN.TC.y < srcUVTransform.y + srcUVTransform.w)
	{
		float2 uv = (IN.TC - srcUVTransform.xy) / srcUVTransform.zw *dstUVTransform.zw + dstUVTransform.xy;
		half4 makeup = tMakeup.Sample(sMakeupSampler, uv);

		OUT.rgb = OUT.rgb*(1.0 - makeup.a) + blendColor.rgb* makeup.rgb* makeup.a;
	}
	return OUT;
}

technique GenTexture
{
	VertexShader = ScreenVS();
	PixelShader = TargetBakePS();
}
