//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzwangyingqiao@corp.netease.com
//	Author: gzwangyingqiao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Fluid.fx>
///	@path	~/System/Shader/
///	@date	2020/02/21
///	@brief	Fluid Simulate

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// samplers //////////////////
texture2D tSourceOneMap : layout(t0);
sampler sSourceOneSampler : layout(s0)
{
	Texture = tSourceOneMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tSourceTwoMap : layout(t1);
sampler sSourceTwoSampler : layout(s1)
{
	Texture = tSourceTwoMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float 	BoundaryScale		 			: layout(c0.x, 	ps[PsBeg + 0].x);
	float 	Dissipation				 		: layout(c0.y, 	ps[PsBeg + 0].y);
	float 	PoissonSolverAlpah				: layout(c0.z, 	ps[PsBeg + 0].z);
	float 	PoissonSolverRBeta				: layout(c0.w, 	ps[PsBeg + 0].w);
	float 	ReciprocalDx					: layout(c1.x, 	ps[PsBeg + 1].x);
	float 	HalfReciprocalDx				: layout(c1.y, 	ps[PsBeg + 1].y);
	float 	TimeStep						: layout(c1.z, 	ps[PsBeg + 1].z);
	float 	ImpulseRadius					: layout(c1.w, 	ps[PsBeg + 1].w);
	float2	WindowDims 						: layout(c2.x, 	ps[PsBeg + 2].x);
	float2	ImpulsePosition					: layout(c2.z, 	ps[PsBeg + 2].z);
	float 	DxScale							: layout(c3.x,	ps[PsBeg + 3].x);
	float3	ImpulseColor					: layout(c3.y, 	ps[PsBeg + 3].y);
};

/////////////////// util   ////////////////////
float4 sourceOneMapBilerp(float2 s)
{
	float4 st;
	st.xy = floor(s - 0.5) + 0.5;
	st.zw = st.xy + 1;
  
	float2 t = s - st.xy; //interpolating factors 
    
    float2 tc = float2(st.x / WindowDims.x, st.y / WindowDims.y);
	float4 tex11 = SampleColorMap(tSourceOneMap, sSourceOneSampler, tc);

	tc = float2(st.z / WindowDims.x, st.y / WindowDims.y);
	float4 tex21 = SampleColorMap(tSourceOneMap, sSourceOneSampler, tc);

	tc = float2(st.x / WindowDims.x, st.w / WindowDims.y);
	float4 tex12 = SampleColorMap(tSourceOneMap, sSourceOneSampler, tc);

	tc = float2(st.z / WindowDims.x, st.w / WindowDims.y);
	float4 tex22 = SampleColorMap(tSourceOneMap, sSourceOneSampler, tc);

	// bilinear interpolation
	return lerp(lerp(tex11, tex21, t.x), lerp(tex12, tex22, t.x), t.y);
}

float4 sourceTwoMapBilerp(float2 s)
{
	float4 st;
	st.xy = floor(s - 0.5) + 0.5;
	st.zw = st.xy + 1;
  
	float2 t = s - st.xy; //interpolating factors 
    
    float2 tc = float2(st.x / WindowDims.x, st.y / WindowDims.y);
	float4 tex11 = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, tc);

	tc = float2(st.z / WindowDims.x, st.y / WindowDims.y);
	float4 tex21 = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, tc);

	tc = float2(st.x / WindowDims.x, st.w / WindowDims.y);
	float4 tex12 = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, tc);

	tc = float2(st.z / WindowDims.x, st.w / WindowDims.y);
	float4 tex22 = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, tc);

	// bilinear interpolation
	return lerp(lerp(tex11, tex21, t.x), lerp(tex12, tex22, t.x), t.y);
}

float2 texNeighborsL(float2 s)
{
	return float2(s.x - 1 / WindowDims.x, s.y);
}

float2 texNeighborsR(float2 s)
{
	return float2(s.x + 1 / WindowDims.x, s.y);
}

float2 texNeighborsB(float2 s)
{
	return float2(s.x, s.y - 1 / WindowDims.y);
}

float2 texNeighborsU(float2 s)
{
	return float2(s.x, s.y + 1 / WindowDims.y);
}

/////////////////// SlabOp ////////////////////
struct SlabOpInput 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct SlabOpOutput
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

SlabOpOutput SlabOpVS( SlabOpInput IN )
{
	SlabOpOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;

	[@_GLSL]
	OUT.TC.y = 1 - OUT.TC.y;
	return OUT;
}

float4 SlobOpPS(SlabOpOutput IN) : SV_Target
{
	return float4(1.0, 0.0, 0.0, 1.0);
}

float4 SlabCopyResultPS(SlabOpOutput IN) : SV_Target
{
	float3 Source = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

float4 AdvectVelocityComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is the velocity field.
	float rdx = ReciprocalDx;
	float2 u = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy).rg;
	float2 coords = IN.TC.xy * WindowDims;
	float2 pos = coords - TimeStep * rdx * u;

	return float4(Dissipation * sourceOneMapBilerp(pos).xy, 0.0, 1.0);
}

float4 AdvectComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is the velocity field.
	// SourceTwoMap is the scalar field.
	float rdx = ReciprocalDx;
	float2 u = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy).rg;
	float2 coords = IN.TC.xy * WindowDims;
	float2 pos = coords - TimeStep * rdx * u;

	return float4(Dissipation * sourceTwoMapBilerp(pos).xyz, 1.0);
}

float4 SplatComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is the field to add splat.
	float2 pulse = ImpulsePosition;
	[@_GLSL]
	pulse.y = 1 - pulse.y;

	float2 pos = WindowDims * pulse - WindowDims * IN.TC.xy;
	float rad = WindowDims.x * ImpulseRadius;
	float4 color = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy);
	float4 impulse = float4(ImpulseColor, 1.0);
	float v = - 1.0 * dot(pos, pos) / rad;
	float4 res = color + impulse * exp(v);
	return float4(res.xyz, 1.0);
}

float4 JacobiComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is x.
	// SourceTwoMap is b.
	float xL = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsL(IN.TC.xy)).r;
	float xR = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsR(IN.TC.xy)).r;
	float xB = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsB(IN.TC.xy)).r;
	float xT = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsU(IN.TC.xy)).r;

	float bC = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, IN.TC.xy).r;

	float v = (xL + xR + xB + xT + PoissonSolverAlpah * bC) * PoissonSolverRBeta;
	return float4(v, 0.0, 0.0, 1.0);
}

float4 VorticityComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is velocity.
	float4 uL = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsL(IN.TC.xy));
	float4 uR = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsR(IN.TC.xy));
	float4 uB = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsB(IN.TC.xy));
	float4 uT = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsU(IN.TC.xy));

	float v = HalfReciprocalDx * ((uR.y - uL.y) - (uT.x - uB.x));
	return float4(v, v, v, 1.0);
}

float4 VorticityForceComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is vorticity.
	// SourceTwoMap is velocity.
	float vL = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsL(IN.TC.xy)).r;
	float vR = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsR(IN.TC.xy)).r;
	float vB = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsB(IN.TC.xy)).r;
	float vT = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsU(IN.TC.xy)).r;

	float vC = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy).r;
	float2 force = float2(abs(vT) - abs(vB), abs(vR) - abs(vL)) * HalfReciprocalDx;

	// safe normalize
	float magSqrt = max(dot(force, force), 2.4414e-4);
	force = force * rsqrt(magSqrt);

	float ratio = DxScale * vC;
	force = float2(force.x * ratio, -1.0 * force.y * ratio);

	float2 velocity = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, IN.TC.xy).rg;
	velocity = velocity + TimeStep * force;
	return float4(velocity, 0.0, 1.0);
}

float4 DivergenceComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is w.
	float4 vL = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsL(IN.TC.xy));
	float4 vR = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsR(IN.TC.xy));
	float4 vB = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsB(IN.TC.xy));
	float4 vT = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsU(IN.TC.xy));

	float v = HalfReciprocalDx * (vR.x - vL.x + vT.y - vB.y);
	return float4(v, 0.0, 0.0, 1.0);
}

float4 GradientComputePS(SlabOpOutput IN) : SV_Target
{
	// SourceOneMap is pressure.
	// SourceTwoMap is velocity.

	float pL = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsL(IN.TC.xy)).r;
	float pR = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsR(IN.TC.xy)).r;
	float pB = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsB(IN.TC.xy)).r;
	float pT = SampleColorMap(tSourceOneMap, sSourceOneSampler, texNeighborsU(IN.TC.xy)).r;

	float2 grad = float2(pR - pL, pT - pB) * HalfReciprocalDx;

	float2 v = SampleColorMap(tSourceTwoMap, sSourceTwoSampler, IN.TC.xy).rg;
	return float4(v.xy - grad, 0.0, 1.0);
}

technique SlabOp
{
	VertexShader = SlabOpVS();
	PixelShader = SlobOpPS();
}

technique Advect
{
	VertexShader = SlabOpVS();
	PixelShader = AdvectComputePS();
}

technique AdvectVelocity
{
	VertexShader = SlabOpVS();
	PixelShader = AdvectVelocityComputePS();
}

technique AddImpulse
{
	VertexShader = SlabOpVS();
	PixelShader = SplatComputePS();
}

technique PoissonSolver
{
	VertexShader = SlabOpVS();
	PixelShader = JacobiComputePS();
}

technique Vorticity
{
	VertexShader = SlabOpVS();
	PixelShader = VorticityComputePS();
}

technique VorticityForce
{
	VertexShader = SlabOpVS();
	PixelShader = VorticityForceComputePS();
}

technique Divergence
{
	VertexShader = SlabOpVS();
	PixelShader = DivergenceComputePS();
}

technique Subtract
{
	VertexShader = SlabOpVS();
	PixelShader = GradientComputePS();
}

technique SlabCopyResult
{
	VertexShader = SlabOpVS();
	PixelShader = SlabCopyResultPS();
}

/////////////////// Boundary ////////////////////
// Use SourceOneMap 
struct BoundaryInput 
{
	float4		Position 	: POSITION; //in object space
	float4		Color 		: COLOR;
	float2		BaseTC		: TEXCOORD0;
	float2		TexCoord2	: TEXCOORD1;
};

struct BoundaryOutput
{
	float4 HPosition : SV_Position;
	float4 TC	: TEXCOORD0;
};

BoundaryOutput BoundaryVS( BoundaryInput IN )
{
	BoundaryOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC.xy = IN.BaseTC;

	[@_GLSL]
	OUT.TC.y = 1 - OUT.TC.y;

	OUT.TC.zw = IN.TexCoord2;

	[@_GLSL]
	OUT.TC.w = - OUT.TC.w;

	return OUT;
}

float4 BoundaryPS(BoundaryOutput IN) : SV_Target
{
	return float4(0.0, 1.0, 0.0, 1.0);
}

float4 BoundaryCopyResultPS(BoundaryOutput IN) : SV_Target
{
	float3 Source = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

float4 BoundaryComputePS(BoundaryOutput IN) : SV_Target
{
	float3 Source = SampleColorMap(tSourceOneMap, sSourceOneSampler, IN.TC.xy + IN.TC.zw).rgb;
	float4 result = float4(Source, 1.0);
	return result * BoundaryScale;
}

technique BoundaryCondition
{
	VertexShader = BoundaryVS();
	PixelShader = BoundaryPS();
}

technique BCCompute
{
	VertexShader = BoundaryVS();
	PixelShader = BoundaryComputePS();
}

technique BCCopyResult
{
	VertexShader = BoundaryVS();
	PixelShader = BoundaryCopyResultPS();
}
