//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzwangyingqiao@corp.netease.com
//	Author: gzwangyingqiao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<FluidDisaplay.fx>
///	@path	~/System/Shader/
///	@date	2020/02/22
///	@brief	Fluid Display

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// samplers //////////////////
texture2D tVelocityMap : FluidVelocity: layout(t0);
sampler sVelocitySampler : layout(s0)
{
	Texture = tVelocityMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tInkMap : FluidInk: layout(t1);
sampler sInkSampler : layout(s1)
{
	Texture = tInkMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tPressureMap : FluidPressure: layout(t2);
sampler sPressureSampler : layout(s2)
{
	Texture = tPressureMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tVorticityMap : FluidVorticity: layout(t3);
sampler sVorticitySampler : layout(s3)
{
	Texture = tVorticityMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

/////////////////// Display ////////////////////
struct DisplayInput 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct DisplayOutput
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

DisplayOutput DisplayVS( DisplayInput IN )
{
	DisplayOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;

	[@_GLSL]
	OUT.TC.y = 1 - OUT.TC.y;
	return OUT;
}

float4 DisplayVelocityPS(DisplayOutput IN) : SV_Target
{
	half3 Source = SampleColorMap(tVelocityMap, sVelocitySampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

float4 DisplayInkPS(DisplayOutput IN) : SV_Target
{
	half3 Source = SampleColorMap(tInkMap, sInkSampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

float4 DisplayPressurePS(DisplayOutput IN) : SV_Target
{
	half3 Source = SampleColorMap(tPressureMap, sPressureSampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

float4 DisplayVorticityPS(DisplayOutput IN) : SV_Target
{
	half3 Source = SampleColorMap(tVorticityMap, sVorticitySampler, IN.TC.xy).rgb;
	return float4(Source, 1.0);
}

technique FluidDisplayVelocity
{
	VertexShader = DisplayVS();
	PixelShader = DisplayVelocityPS();
}

technique FluidDisplayInk
{
	VertexShader = DisplayVS();
	PixelShader = DisplayInkPS();
}

technique FluidDisplayPressure
{
	VertexShader = DisplayVS();
	PixelShader = DisplayPressurePS();
}

technique FluidDisplayVorticity
{
	VertexShader = DisplayVS();
	PixelShader = DisplayVorticityPS();
}
