#pragma shader draw
#pragma shader predict("@_INSTANCED", "false")

#include "Constants.fxh"
#include "Common.fxh"
#include "DeferredShadingCommon.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_CameraInVolume
{ 
	Name = CameraInVolume;
	Default = true;
};

@_PerspectiveShadowMap
{ 
	Name = PerspectiveShadowMap;
	Default = false;
};

@_HighQualityShadow
{
	Name = HighQualityShadow;
	Default = false;
};
@_MinShadowMask
{
	Name = MinShadowMask;
	Default = false;
};
@_NeedDither
{
	Name = NeedDither;
	Default = false;
};
@_UseShadowMerge
{
	Name = UseShadowMerge;
	Default = false;
};
@_UseContactShadow
{
	Name = UseContactShadow;
	Default = false;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4x3 cWorld 				;

	float4 cWBasisX					: WBasisX ;
	float4 cWBasisY					: WBasisY ;
	float4 cWBasisZ					: WBasisZ ;

	float4x4 cShadowViewProjTex		;
	float cFilterRadius				;
	float cInvShadowMapSize			;
	float cFarPlane					;
	float cOffset					 = 0;
	float cLerpAO					 = 0;
	float cPixelSize				 = 2;
	float cShadowMapSize			;
	float cFilterRadius1			;
	float4x4 cShadowViewProjTex1	;
	float4 cPosParam				;
	float4 cRTAndHZUVScale			;
	float4x3 cWorldToView			: View ;
	float4x4 cLastViewProjTex		;
	float4x4 cLastProjTex			;
	float4x4 cShadowViewProj		;
	float cRayMarchingLength		;
	float  cFrameIndex				;
};

texture2D tDepthMap : SceneDepth : layout(t0);
sampler sDepthSampler : layout(s0)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tPreDepthMap : layout(t1);
sampler sPreDepthSampler : layout(s1)
{
	Texture = tPreDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tShadowMap : layout(t2);
sampler sShadowSampler : layout(s2)
{
	Texture = tShadowMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tShadowMapS : layout(t3);
sampler sShadowSamplerS : layout(s3)
{
	Texture = tShadowMapS;
	Filter = Point;
	Address = Clamp;
};
texture2D tShadowMapD : layout(t4);
sampler sShadowSamplerD : layout(s4)
{
	Texture = tShadowMapD;
	Filter = Point;
	Address = Clamp;
};
texture2D tBaseMask : layout(t5);
sampler sBaseMaskSampler : layout(s5)
{
	Texture = tBaseMask;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
texture2D tShadowMaskSF : layout(t6);
sampler sShadowMaskSFSampler : layout(s6)
{
	Texture = tShadowMaskSF;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};
[inputColor(6)]
texture2D tDepthDeferredMap : SceneDepth : layout(t7);
sampler sDepthDeferredSampler : layout(s7)
{
	Texture = tDepthDeferredMap;
	Filter = Point;
	Address = Clamp;
};
[inputColor(1)]
texture2D tGBufferB : GBufferMapB : layout(t8);
sampler sGBufferB : layout(s8)
{
	Texture = tGBufferB;
	Filter = Point;
	Address = Clamp;
};
[inputColor(3)]
texture2D tGBufferD : GBufferMapD : layout(t9);
sampler sGBufferD : layout(s9)
{
	Texture = tGBufferD;
	Filter = Point;
	Address = Clamp;
};

///////////////// io signatures //////////////////
/// light volume vertex to vs
struct VertexAssembleShadowVolume
{
	float3		Position		: POSITION;
};

// vertex output to pixel shader.
struct VertexOutputShadowVolume
{
	float4		HPosition		: SV_Position;
};

#define SampleShadowMap(destTC)	DecodeDepth(tShadowMap.Sample(sShadowSampler, destTC))

float4 CalculateOcclusion(float4 ShadowmapDepth, float4 SceneDepth, float TransitionScale)
{
	return saturate((ShadowmapDepth - SceneDepth) * TransitionScale + 1);
}
float3 CalculateOcclusion(float3 ShadowmapDepth, float3 SceneDepth, float TransitionScale)
{
	return saturate((ShadowmapDepth - SceneDepth) * TransitionScale + 1);
}
float CalculateOcclusion(float ShadowmapDepth, float SceneDepth, float TransitionScale)
{
	return saturate((ShadowmapDepth - SceneDepth) * TransitionScale + 1);
}
#define native_pcf3(pInLitTC, litZ, vInvShadowMapWH, vShadowMapWH, shadowTest, filterRadius)\
{\
	shadowTest = 0;\
	float2 filterRange = vInvShadowMapWH; \
	float2 texelPos = pInLitTC.xy * vShadowMapWH - 0.5;\
	float2 fraction = frac(texelPos);\
	float2 texelCenter = floor(texelPos) + 0.5;\
	float2 sample00TexelCenter = texelCenter;\
	float3 sampleDepth1, sampleDepth2, sampleDepth3; \
	sampleDepth1.x = SampleShadowMap((sample00TexelCenter+float2(0,0))*filterRange); \
	sampleDepth1.y = SampleShadowMap((sample00TexelCenter+float2(1,0))*filterRange); \
	sampleDepth1.z = SampleShadowMap((sample00TexelCenter+float2(2,0))*filterRange); \
	sampleDepth2.x = SampleShadowMap((sample00TexelCenter+float2(0,1))*filterRange); \
	sampleDepth2.y = SampleShadowMap((sample00TexelCenter+float2(1,1))*filterRange); \
	sampleDepth2.z = SampleShadowMap((sample00TexelCenter+float2(2,1))*filterRange); \
	sampleDepth3.x = SampleShadowMap((sample00TexelCenter+float2(0,2))*filterRange); \
	sampleDepth3.y = SampleShadowMap((sample00TexelCenter+float2(1,2))*filterRange); \
	sampleDepth3.z = SampleShadowMap((sample00TexelCenter+float2(2,2))*filterRange); \
	float3 InShadow1 = CalculateOcclusion(sampleDepth1, litZ.xxx, filterRadius);\
	float3 InShadow2 = CalculateOcclusion(sampleDepth2, litZ.xxx, filterRadius);\
	float3 InShadow3 = CalculateOcclusion(sampleDepth3, litZ.xxx, filterRadius);\
	float3 v2Y1 = lerp(float3(InShadow1.x, InShadow2.x, InShadow3.x), float3(InShadow1.y, InShadow2.y, InShadow3.y), fraction.xxx); \
	float3 v2Y2 = lerp(float3(InShadow1.y, InShadow2.y, InShadow3.y), float3(InShadow1.z, InShadow2.z, InShadow3.z), fraction.xxx); \
	shadowTest += lerp(v2Y1.x, v2Y1.y, fraction.y);\
	shadowTest += lerp(v2Y1.y, v2Y1.z, fraction.y);\
	shadowTest += lerp(v2Y2.x, v2Y2.y, fraction.y);\
	shadowTest += lerp(v2Y2.y, v2Y2.z, fraction.y);\
	shadowTest = shadowTest * 0.25;\
}

#define native_pcf4(pInLitTC, litZ, vInvShadowMapWH, vShadowMapWH, shadowTest, filterRadius)\
{\
	shadowTest = 0;\
	float2 filterRange = vInvShadowMapWH;\
	float2 texelPos = pInLitTC.xy * vShadowMapWH - 0.5;\
	float2 Fraction = frac(texelPos);\
	float2 texelCenter = floor(texelPos) + 0.5;\
	float2 sample00TexelCenter = texelCenter - float2(1, 1);\
	float4 sampleDepth1, sampleDepth2, sampleDepth3, sampleDepth4; \
	float flag = 1;\
	sampleDepth1.x = SampleShadowMap((sample00TexelCenter+float2(0,0) * flag)*filterRange); \
	sampleDepth1.y = SampleShadowMap((sample00TexelCenter+float2(1,0) * flag)*filterRange); \
	sampleDepth1.z = SampleShadowMap((sample00TexelCenter+float2(2,0) * flag)*filterRange); \
	sampleDepth1.w = SampleShadowMap((sample00TexelCenter+float2(3,0) * flag)*filterRange); \
	sampleDepth2.x = SampleShadowMap((sample00TexelCenter+float2(0,1) * flag)*filterRange); \
	sampleDepth2.y = SampleShadowMap((sample00TexelCenter+float2(1,1) * flag)*filterRange); \
	sampleDepth2.z = SampleShadowMap((sample00TexelCenter+float2(2,1) * flag)*filterRange); \
	sampleDepth2.w = SampleShadowMap((sample00TexelCenter+float2(3,1) * flag)*filterRange); \
	sampleDepth3.x = SampleShadowMap((sample00TexelCenter+float2(0,2) * flag)*filterRange); \
	sampleDepth3.y = SampleShadowMap((sample00TexelCenter+float2(1,2) * flag)*filterRange); \
	sampleDepth3.z = SampleShadowMap((sample00TexelCenter+float2(2,2) * flag)*filterRange); \
	sampleDepth3.w = SampleShadowMap((sample00TexelCenter+float2(3,2) * flag)*filterRange); \
	sampleDepth4.x = SampleShadowMap((sample00TexelCenter+float2(0,3) * flag)*filterRange); \
	sampleDepth4.y = SampleShadowMap((sample00TexelCenter+float2(1,3) * flag)*filterRange); \
	sampleDepth4.z = SampleShadowMap((sample00TexelCenter+float2(2,3) * flag)*filterRange); \
	sampleDepth4.w = SampleShadowMap((sample00TexelCenter+float2(3,3) * flag)*filterRange); \
	float4 Values0 = CalculateOcclusion(sampleDepth1, litZ.xxxx, filterRadius);\
	float4 Values1 = CalculateOcclusion(sampleDepth2, litZ.xxxx, filterRadius);\
	float4 Values2 = CalculateOcclusion(sampleDepth3, litZ.xxxx, filterRadius);\
	float4 Values3 = CalculateOcclusion(sampleDepth4, litZ.xxxx, filterRadius);\
	float2 VerticalLerp00 = lerp(float2(Values0.x, Values1.x), float2(Values0.y, Values1.y), Fraction.xx);\
	float PCFResult00 = lerp(VerticalLerp00.x, VerticalLerp00.y, Fraction.y);\
	float2 VerticalLerp10 = lerp(float2(Values0.y, Values1.y), float2(Values0.z, Values1.z), Fraction.xx);\
	float PCFResult10 = lerp(VerticalLerp10.x, VerticalLerp10.y, Fraction.y);\
	float2 VerticalLerp20 = lerp(float2(Values0.z, Values1.z), float2(Values0.w, Values1.w), Fraction.xx);\
	float PCFResult20 = lerp(VerticalLerp20.x, VerticalLerp20.y, Fraction.y);\
	float2 VerticalLerp01 = lerp(float2(Values1.x, Values2.x), float2(Values1.y, Values2.y), Fraction.xx);\
	float PCFResult01 = lerp(VerticalLerp01.x, VerticalLerp01.y, Fraction.y);\
	float2 VerticalLerp11 = lerp(float2(Values1.y, Values2.y), float2(Values1.z, Values2.z), Fraction.xx);\
	float PCFResult11 = lerp(VerticalLerp11.x, VerticalLerp11.y, Fraction.y);\
	float2 VerticalLerp21 = lerp(float2(Values1.z, Values2.z), float2(Values1.w, Values2.w), Fraction.xx);\
	float PCFResult21 = lerp(VerticalLerp21.x, VerticalLerp21.y, Fraction.y);\
	float2 VerticalLerp02 = lerp(float2(Values2.x, Values3.x), float2(Values2.y, Values3.y), Fraction.xx);\
	float PCFResult02 = lerp(VerticalLerp02.x, VerticalLerp02.y, Fraction.y);\
	float2 VerticalLerp12 = lerp(float2(Values2.y, Values3.y), float2(Values2.z, Values3.z), Fraction.xx);\
	float PCFResult12 = lerp(VerticalLerp12.x, VerticalLerp12.y, Fraction.y);\
	float2 VerticalLerp22 = lerp(float2(Values2.z, Values3.z), float2(Values2.w, Values3.w), Fraction.xx);\
	float PCFResult22 = lerp(VerticalLerp22.x, VerticalLerp22.y, Fraction.y);\
	shadowTest = (PCFResult00 + PCFResult10 + PCFResult20 + PCFResult01 + PCFResult11 + PCFResult21 + PCFResult02 + PCFResult12 + PCFResult22) * 0.11111;\
}
#define SampleShadowMap1_1(destTC)	DecodeDepth(tShadowMaskSF.Sample(sShadowMaskSFSampler, destTC)) // static first
half3 CalculateOcclusion1_1(float3 ShadowmapDepth, float3 SceneDepth)
{
	float TransitionScale = cFilterRadius1; //UserData[1].w;
	return saturate(half3((ShadowmapDepth - SceneDepth) * TransitionScale) + 1);
}
half4 CalculateOcclusion1_1(float4 ShadowmapDepth, float4 SceneDepth)
{
	float TransitionScale = cFilterRadius1; //UserData[1].w;
	return saturate(half4((ShadowmapDepth - SceneDepth) * TransitionScale) + 1);
}
void native_pcf3_1_1(float2 pInLitTC, float litZ, float2 vInvShadowMapWH, float2 vShadowMapWH, inout half shadowTest)
{
	shadowTest = 0;
	float2 filterRange = vInvShadowMapWH; 
	float2 texelPos = pInLitTC.xy * vShadowMapWH - 0.5;
	half2 fraction = frac(texelPos);
	float2 texelCenter = floor(texelPos) + 0.5;
	float2 sample00TexelCenterUV = texelCenter * filterRange;
	float3 sampleDepth1, sampleDepth2, sampleDepth3; 
	sampleDepth1.x = SampleShadowMap1_1(float2(0,0)*filterRange + sample00TexelCenterUV);
	sampleDepth1.y = SampleShadowMap1_1(float2(1,0)*filterRange + sample00TexelCenterUV);
	sampleDepth1.z = SampleShadowMap1_1(float2(2,0)*filterRange + sample00TexelCenterUV);
	sampleDepth2.x = SampleShadowMap1_1(float2(0,1)*filterRange + sample00TexelCenterUV);
	sampleDepth2.y = SampleShadowMap1_1(float2(1,1)*filterRange + sample00TexelCenterUV);
	sampleDepth2.z = SampleShadowMap1_1(float2(2,1)*filterRange + sample00TexelCenterUV);
	sampleDepth3.x = SampleShadowMap1_1(float2(0,2)*filterRange + sample00TexelCenterUV);
	sampleDepth3.y = SampleShadowMap1_1(float2(1,2)*filterRange + sample00TexelCenterUV);
	sampleDepth3.z = SampleShadowMap1_1(float2(2,2)*filterRange + sample00TexelCenterUV);
	half3 InShadow1 = CalculateOcclusion1_1(sampleDepth1, litZ.xxx);
	half3 InShadow2 = CalculateOcclusion1_1(sampleDepth2, litZ.xxx);
	half3 InShadow3 = CalculateOcclusion1_1(sampleDepth3, litZ.xxx);
	half3 v2Y1 = lerp(half3(InShadow1.x, InShadow2.x, InShadow3.x), half3(InShadow1.y, InShadow2.y, InShadow3.y), fraction.xxx);
	half3 v2Y2 = lerp(half3(InShadow1.y, InShadow2.y, InShadow3.y), half3(InShadow1.z, InShadow2.z, InShadow3.z), fraction.xxx);
	shadowTest += lerp(v2Y1.x, v2Y1.y, fraction.y);
	shadowTest += lerp(v2Y1.y, v2Y1.z, fraction.y);
	shadowTest += lerp(v2Y2.x, v2Y2.y, fraction.y);
	shadowTest += lerp(v2Y2.y, v2Y2.z, fraction.y);
	shadowTest = shadowTest * 0.25;
}
#define native_pcf4_1_1(pInLitTC, litZ, vInvShadowMapWH, vShadowMapWH, shadowTest)\
{\
	shadowTest = 0;\
	float2 filterRange = vInvShadowMapWH;\
	float2 texelPos = pInLitTC.xy * vShadowMapWH - 0.5;\
	float2 Fraction = frac(texelPos);\
	float2 texelCenter = floor(texelPos) + 0.5;\
	float2 sample00TexelCenter = texelCenter - float2(1, 1);\
	float4 sampleDepth1, sampleDepth2, sampleDepth3, sampleDepth4; \
	float flag = 1;\
	sampleDepth1.x = SampleShadowMap1_1((sample00TexelCenter+float2(0,0) * flag)*filterRange); \
	sampleDepth1.y = SampleShadowMap1_1((sample00TexelCenter+float2(1,0) * flag)*filterRange); \
	sampleDepth1.z = SampleShadowMap1_1((sample00TexelCenter+float2(2,0) * flag)*filterRange); \
	sampleDepth1.w = SampleShadowMap1_1((sample00TexelCenter+float2(3,0) * flag)*filterRange); \
	sampleDepth2.x = SampleShadowMap1_1((sample00TexelCenter+float2(0,1) * flag)*filterRange); \
	sampleDepth2.y = SampleShadowMap1_1((sample00TexelCenter+float2(1,1) * flag)*filterRange); \
	sampleDepth2.z = SampleShadowMap1_1((sample00TexelCenter+float2(2,1) * flag)*filterRange); \
	sampleDepth2.w = SampleShadowMap1_1((sample00TexelCenter+float2(3,1) * flag)*filterRange); \
	sampleDepth3.x = SampleShadowMap1_1((sample00TexelCenter+float2(0,2) * flag)*filterRange); \
	sampleDepth3.y = SampleShadowMap1_1((sample00TexelCenter+float2(1,2) * flag)*filterRange); \
	sampleDepth3.z = SampleShadowMap1_1((sample00TexelCenter+float2(2,2) * flag)*filterRange); \
	sampleDepth3.w = SampleShadowMap1_1((sample00TexelCenter+float2(3,2) * flag)*filterRange); \
	sampleDepth4.x = SampleShadowMap1_1((sample00TexelCenter+float2(0,3) * flag)*filterRange); \
	sampleDepth4.y = SampleShadowMap1_1((sample00TexelCenter+float2(1,3) * flag)*filterRange); \
	sampleDepth4.z = SampleShadowMap1_1((sample00TexelCenter+float2(2,3) * flag)*filterRange); \
	sampleDepth4.w = SampleShadowMap1_1((sample00TexelCenter+float2(3,3) * flag)*filterRange); \
	float4 Values0 = CalculateOcclusion1_1(sampleDepth1, litZ.xxxx);\
	float4 Values1 = CalculateOcclusion1_1(sampleDepth2, litZ.xxxx);\
	float4 Values2 = CalculateOcclusion1_1(sampleDepth3, litZ.xxxx);\
	float4 Values3 = CalculateOcclusion1_1(sampleDepth4, litZ.xxxx);\
	float2 VerticalLerp00 = lerp(float2(Values0.x, Values1.x), float2(Values0.y, Values1.y), Fraction.xx);\
	float PCFResult00 = lerp(VerticalLerp00.x, VerticalLerp00.y, Fraction.y);\
	float2 VerticalLerp10 = lerp(float2(Values0.y, Values1.y), float2(Values0.z, Values1.z), Fraction.xx);\
	float PCFResult10 = lerp(VerticalLerp10.x, VerticalLerp10.y, Fraction.y);\
	float2 VerticalLerp20 = lerp(float2(Values0.z, Values1.z), float2(Values0.w, Values1.w), Fraction.xx);\
	float PCFResult20 = lerp(VerticalLerp20.x, VerticalLerp20.y, Fraction.y);\
	float2 VerticalLerp01 = lerp(float2(Values1.x, Values2.x), float2(Values1.y, Values2.y), Fraction.xx);\
	float PCFResult01 = lerp(VerticalLerp01.x, VerticalLerp01.y, Fraction.y);\
	float2 VerticalLerp11 = lerp(float2(Values1.y, Values2.y), float2(Values1.z, Values2.z), Fraction.xx);\
	float PCFResult11 = lerp(VerticalLerp11.x, VerticalLerp11.y, Fraction.y);\
	float2 VerticalLerp21 = lerp(float2(Values1.z, Values2.z), float2(Values1.w, Values2.w), Fraction.xx);\
	float PCFResult21 = lerp(VerticalLerp21.x, VerticalLerp21.y, Fraction.y);\
	float2 VerticalLerp02 = lerp(float2(Values2.x, Values3.x), float2(Values2.y, Values3.y), Fraction.xx);\
	float PCFResult02 = lerp(VerticalLerp02.x, VerticalLerp02.y, Fraction.y);\
	float2 VerticalLerp12 = lerp(float2(Values2.y, Values3.y), float2(Values2.z, Values3.z), Fraction.xx);\
	float PCFResult12 = lerp(VerticalLerp12.x, VerticalLerp12.y, Fraction.y);\
	float2 VerticalLerp22 = lerp(float2(Values2.z, Values3.z), float2(Values2.w, Values3.w), Fraction.xx);\
	float PCFResult22 = lerp(VerticalLerp22.x, VerticalLerp22.y, Fraction.y);\
	shadowTest = (PCFResult00 + PCFResult10 + PCFResult20 + PCFResult01 + PCFResult11 + PCFResult21 + PCFResult02 + PCFResult12 + PCFResult22) * 0.11111;\
}
float dither(float2 vPos)
{
	float randomv = frac(dot(float2(276.0,273.0)/71, vPos.xy + GameTime % 16));
	return frac(52.9829189 * (frac(dot(vPos, float2(0.06711056, 0.00583715)))+randomv));
}

float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}


////////////////////////////////////LightVolumeStencil////////////////////////////////////
VertexOutputShadowVolume ShadowVolumeVS( VertexAssembleShadowVolume IN )
{
	VertexOutputShadowVolume OUT;

	// Transform position.
	float4 worldPosition = float4(MulMat3x4(cWorld, float4(IN.Position*cPosParam.w, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, worldPosition);
	OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);//disable farplane culling
	return OUT;
}
half ContactShadowCalc(float3 worldPos, float randomOffset)
{
	float4 rayStartClip = MulMat4x4(cLastViewProjTex, float4(worldPos, 1.0));
	float k = 1.0 / rayStartClip.w;
	float3 rayStartScreen = rayStartClip.xyz * k; // go device depth

	float4 rayEndClip = MulMat4x4(cLastViewProjTex, float4(worldPos + SunDirection.xyz * cRayMarchingLength, 1.0));
	float3 rayEndScreen;
	k = 1.0 / rayEndClip.w;
	rayEndScreen = rayEndClip.xyz * k;

	float3 rayStepScreen = rayEndScreen - rayStartScreen;
	if(rayEndScreen.z < 0)
	{
		return 1;
	}

	float4 rayDepthClip	= rayStartClip + MulMat4x4(cLastProjTex, float4( 0, 0, cRayMarchingLength, 0 ));
	float3 rayDepthScreen = rayDepthClip.xyz / rayDepthClip.w;
	const float numStep = 8;
	const float stepOffset = 1.0 / numStep;
	const float compareTolerance = abs(rayDepthScreen.z - rayStartScreen.z) * stepOffset * 2; // more far more tolerance
	float sampleTime = stepOffset * randomOffset + stepOffset;
	float firstHitTime = -1.0;

	[loop]
	for (int i = 1; i <= numStep; ++i)
	{
		float3 sampleUVZ = rayStartScreen + rayStepScreen * sampleTime;
		float sampleDepth = GetDeviceZ(DecodeDepth(tPreDepthMap.SampleLevel(sPreDepthSampler, sampleUVZ.xy, 0)));
		float depthDiff = sampleUVZ.z - sampleDepth + 0.0001; // trick for bias
		bool hit = abs(depthDiff + compareTolerance) < compareTolerance;
		firstHitTime = (hit && firstHitTime < 0.0) ? sampleTime : firstHitTime;
		sampleTime += stepOffset;
	}
	float shadow = firstHitTime > 0.0 ? 1.0 : 0.0;

	float2 vignette = max(6.0 * abs(2 * (rayStartScreen.xy + rayStepScreen.xy * firstHitTime) - 1) - 5.0, 0.0); // solve out of screen
	shadow *= saturate(1.0 - dot(vignette, vignette));
	return 1 - shadow;
}

////////////////////////////////ShadowMask////////////////////////////////////////
float4 ShadowMaskPS( VertexOutputShadowVolume IN )
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	//P	
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = 1;
	if(@_DEFERRED_SHADING)
	{
		sceneZ = DecodeDepth(tDepthDeferredMap.Fetch(sDepthDeferredSampler, stc));
	}else
	{
		sceneZ = DecodeDepth(tDepthMap.Sample(sDepthSampler, stc));
	}
	
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	
	float3 P = CameraPos.xyz + D;

	int shadingModelID;
	half3 offset = half3(0, 0, 0);
	int filterFlag = NO_NEED_FILTER;
	if(@_DEFERRED_SHADING)
	{
		shadingModelID = DecodeShadingModelID(tGBufferB.Fetch(sGBufferB, stc).a);
		if(shadingModelID == SHADINGMODELID_TWOSIDED_FOLIAGE)
		{
			offset = DecodeWorldOffset(tGBufferD.Fetch(sGBufferD, stc).rgb);
			filterFlag = NEED_FILTER;
		}
	}else
	{
		shadingModelID = DecodeShadingModelID(tGBufferB.Sample(sGBufferB,stc).a);
		if(shadingModelID == SHADINGMODELID_TWOSIDED_FOLIAGE)
		{
			offset = DecodeWorldOffset(tGBufferD.Sample(sGBufferD, stc).rgb);
			filterFlag = NEED_FILTER;
		}
	}
	float3 noOffsetP = P;
	P -= offset;
	float4 modifyP = float4(P, 1);
	
	float4 P_LS = mul(modifyP, cShadowViewProjTex);
	
	P_LS.xyz /= P_LS.w;
	
	if (!@_CameraInVolume)
	{
		clip(0.5 - abs(P_LS.xyz - 0.5));
	}
	float inShadow = 0;
	float litZ = min(1.0 - 10 * EPSILON, P_LS.z);
	
	if(@_UseShadowMerge)
	{
		float4 P_LS1 = mul(modifyP, cShadowViewProjTex1);
		P_LS1.xyz /= P_LS1.w;
		float litZ1 = min(1.0 - 10 * EPSILON, P_LS1.z);
		half InShadowS = 0;
		if(@_D3D11)
		{
			native_pcf4(P_LS.xy, litZ, half2(cInvShadowMapSize, cInvShadowMapSize), half2(cShadowMapSize, cShadowMapSize), inShadow, cFilterRadius);
			native_pcf4_1_1(P_LS1.xy, litZ1, half2(cInvShadowMapSize, cInvShadowMapSize), half2(cShadowMapSize, cShadowMapSize), InShadowS);
		}
		else
		{
			native_pcf3(P_LS.xy, litZ, half2(cInvShadowMapSize, cInvShadowMapSize), half2(cShadowMapSize, cShadowMapSize), inShadow, cFilterRadius);
			native_pcf3_1_1(P_LS1.xy, litZ1, half2(cInvShadowMapSize, cInvShadowMapSize), half2(cShadowMapSize, cShadowMapSize), InShadowS);
		}
		inShadow = min(inShadow, InShadowS);
	}else
	{
		//float dist = sqrt(D.x*D.x + D.z*D.z);
		float dist = length(D);
		float fixFilterRadius = lerp(cFilterRadius, cPosParam.z, max(0,min(1,(dist - cPosParam.x)/cPosParam.y)));
		if(@_D3D11)
		{
			native_pcf4(P_LS.xy, litZ, half2(cInvShadowMapSize, cInvShadowMapSize), half2(cShadowMapSize, cShadowMapSize), inShadow, fixFilterRadius);
		}
		else
		{
			native_pcf3(P_LS.xy, litZ, half2(cInvShadowMapSize, cInvShadowMapSize), half2(cShadowMapSize, cShadowMapSize), inShadow, fixFilterRadius);
		}
	}
	if(@_UseContactShadow)
	{
		float randomOffset = InterleavedGradientNoise(spos, cFrameIndex) - 0.5;
		inShadow = inShadow * ContactShadowCalc(noOffsetP, randomOffset);
		//inShadow = ContactShadowCalc(P, randomOffset);
	}
	return float4(EncodeShadowAndFilter(1 - inShadow, filterFlag), 1, 0, 0);
}

float4 ForwardShadowMaskPS( VertexOutputShadowVolume IN ) : SV_Target
{
	return ShadowMaskPS(IN);
}
// shadowmap merge
technique ForwardShadowMask
{
	VertexShader = ShadowVolumeVS();
	PixelShader = ForwardShadowMaskPS();
}

////////////////////////////////MergeShadowMap////////////////////////////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC		 : TEXCOORD0;
};

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, EPSILON, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float MergeShadowMapPS(v2f IN) : SV_Depth
{
	float depthS = DecodeDepth(tShadowMapS.Sample(sShadowSamplerS, IN.TC));
	float depthD = DecodeDepth(tShadowMapD.Sample(sShadowSamplerD, IN.TC));
	float depth = min(depthS, depthD);
	return depth;
}

technique ForwardMergeShadowMap
{
	VertexShader = ScreenVS();
	PixelShader = MergeShadowMapPS();
}

float4 DeferredShadowMaskPS( VertexOutputShadowVolume IN ) : SV_Target4
{
	float4 color = ShadowMaskPS(IN);
	return float4(0, 0, 0, color.r);
}

technique DeferredShadowMask
{
	VertexShader = ShadowVolumeVS();
	PixelShader = DeferredShadowMaskPS();
}

VertexOutputShadowVolume ShadowTileVS( VertexAssembleShadowVolume IN )
{
	VertexOutputShadowVolume OUT;

	// Transform position.
	float4 worldPosition = float4(MulMat3x4(cWorld, float4(IN.Position, 1)).xyz, 1);
	OUT.HPosition = mul(worldPosition, cShadowViewProj);
	OUT.HPosition.z = 1 - EPSILON;
	return OUT;
}

////////////////////////////////ShadowMask////////////////////////////////////////
float4 ShadowTilePS( VertexOutputShadowVolume IN )  : SV_Target
{
	return float4(0, 0, 0, 0);
}
// shadowmap merge
technique ShadowTileRender
{
	VertexShader = ShadowTileVS();
	PixelShader = ShadowTilePS();
}
