#pragma shader material
#define USE_LIGHTMAP			0
#define USE_CLIPPING			0
#define USE_WIREFRAME			0
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			0

#include "Constants.fxh"
#include "Common.fxh"
//#include "BRDF.fxh"

#pragma shader predict("@_INSTANCED", "false")

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor				 < Widget = "LinearColor3"; > = {0.1, 0.1, 0.1};
	float  cAlpha					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;
	float  cFresnelPower			 = 1;
	float  cFresnelAlpha			 = 1;
};

//////////////////////////////////Input Texture////////////////////////////////////////
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	Fallback = "@_TANGENT";
	Visible = false;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct GhostVertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
//#include "Lighting.fxh" 
void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out GhostVertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (GhostVertexShadingOutput)0;

	// Position
	if (@_INSTANCED)
	{
		OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	// TexCoord
	OUT.TexCoord = IN.TexCoord.xyxy;
	[@_ES20] //es2 bug sample lightmap with .zw leads to mosaic
	OUT.TexCoord.xyzw = OUT.TexCoord.zwxy;

	// Normal
	if (@_INSTANCED)
	{
		OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
	}
	else
	{
		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	}
	OUT.WorldNormal.w = 0;

	// ViewDirection
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;
	OUT.ViewDirection.w = 0;
	
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
}

GhostVertexShadingOutput GhostModelVertexShading(in VertexAssemble IN)
{
	GhostVertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);
	return OUT;
}

float4 GhostModelPixelShading(in GhostVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(cBaseColor, cAlpha);
	float fresnel = pow(saturate(1 - abs(dot(normalize(IN.WorldNormal.xyz), normalize(IN.ViewDirection.xyz)))), cFresnelPower);
	//OUT.a = lerp(OUT.a, cFresnelAlpha, fresnel);
	OUT.a *= fresnel;

	half bloomScale = 0;
	
	if (@_TRANSPARENT)
	{
		if (@_HDR)
			OUT.rgb = OUT.rgb * OUT.a;
		else
			OUT.rgb = saturate(OUT.rgb) * OUT.a;
		OUT.a = EncodeTransparentBloomFactor(OUT.a, bloomScale);
	}
	else
	{
		OUT.a = EncodeSolidBloomFactor(bloomScale + SunColor.w);
	}

	return OUT;
}

technique Shading
{
	VertexShader = GhostModelVertexShading();
	PixelShader = GhostModelPixelShading();
}

float4 ClipGhostModelPixelShading(in GhostVertexShadingOutput IN) : SV_Target
{
	return float4(1, 0, 0, 0);
}

technique Clip
{
	VertexShader = GhostModelVertexShading();
	PixelShader = ClipGhostModelPixelShading();
}
