
#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4x4 WorldViewProj	: WorldViewProj		;
	float4 InnerLineColor						;
	float4 OuterLineColor						;
	float4 RectLineColor						;
	float GridLineWidth							;
};


///////////////// io signatures //////////////////
struct a2v
{
	float4 Position 	: POSITION;
	float4 Color		: COLOR;
	float2 UV			: TEXCOORD;
};

struct v2f
{
	float4 Position : SV_Position;
	float4 Color	: COLOR;
	float2 UV		: TEXCOORD;
};

struct a2v_rect
{
	float4 Position 	: POSITION;
};

struct v2f_rect
{
	float4 Position : SV_Position;
};

struct v2f_triangle
{
	float4 Position : SV_Position;
	float4 Color	: COLOR;
};

///////////////// general pass //////////////////
v2f GridHelperVS(a2v IN)
{
	v2f OUT;

	[@_HLSL]
	OUT = (v2f)0;

	OUT.Position = MulMat4x4(WorldViewProj, float4(IN.Position.xyz,1.0));
	OUT.Color = IN.Color;
	OUT.UV = IN.Position.xz;

	return OUT;
}

v2f_rect GridHelperRectVS(a2v_rect IN)
{
	v2f_rect OUT;

	[@_HLSL]
	OUT = (v2f_rect)0;

	OUT.Position = MulMat4x4(WorldViewProj, float4(IN.Position.xyz,1.0));

	return OUT;
}

v2f_triangle GridHelperTriangleVS(a2v IN)
{
	v2f_triangle OUT;

	[@_HLSL]
	OUT = (v2f_triangle)0;

	OUT.Position = MulMat4x4(WorldViewProj, float4(IN.Position.xyz,1.0));
	OUT.Color = IN.Color;

	return OUT;
}

float4 GridHelperPS(v2f IN) : SV_Target
{
	float2 frac_uv = frac(IN.UV);

	float4 OUT = IN.Color;

	// at least two (consider screen scale) pixel width
	float GridLineWidth_x = max(GridLineWidth, 2 * sqrt(pow(ddx(IN.UV.x), 2) + pow(ddy(IN.UV.x), 2)));
	float GridLineWidth_y = max(GridLineWidth, 2 * sqrt(pow(ddx(IN.UV.y), 2) + pow(ddy(IN.UV.y), 2)));

	if(frac_uv.x < GridLineWidth_x / 2 || frac_uv.x > 1 - GridLineWidth_x / 2 || frac_uv.y < GridLineWidth_y / 2 || frac_uv.y > 1 - GridLineWidth_y / 2)
	{
		OUT = OuterLineColor;
	}
	if(abs(frac_uv.x - 0.5) < GridLineWidth_x / 2 && frac(frac_uv.y * min(10, 0.2 / GridLineWidth_y)) < 0.5)
	{
		OUT = InnerLineColor;
	}
	if(abs(frac_uv.y - 0.5) < GridLineWidth_y / 2 && frac(frac_uv.x * min(10, 0.2 / GridLineWidth_x)) < 0.5)
	{
		OUT = InnerLineColor;
	}

	OUT.a = OUT.a / (max(GridLineWidth_x, GridLineWidth_y) / GridLineWidth);

	OUT.rgb *= OUT.a;
	return OUT;
}

float4 GridHelperRectPS(v2f_rect IN) : SV_Target
{
	float4 OUT = RectLineColor;
	OUT.rgb *= OUT.a;
	return OUT;
}

float4 GridHelperTrianglePS(v2f_triangle IN) : SV_Target
{
	float4 OUT = IN.Color;
	OUT.rgb *= OUT.a;
	return OUT;
}

///////////////// techniques //////////////////
technique GridHelper
{
	VertexShader = GridHelperVS();
	PixelShader = GridHelperPS();
}

technique GridHelperRect
{
	VertexShader = GridHelperRectVS();
	PixelShader = GridHelperRectPS();
}

technique GridHelperTriangle
{
	VertexShader = GridHelperTriangleVS();
	PixelShader = GridHelperTrianglePS();
}
///////////////// eof //////////////////
