#pragma shader material
#define MATERIAL_SHADINGMODEL_IMP_LIT 		1
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow)
#define DEFERRED_PRE_AO						1 // 是否走ao

#include "Constants.fxh"
#include "Common.fxh"
#include "FastMath.fxh"
#include "BRDF.fxh"
#include "DeferredShadingCommon.fxh"

//////////////////////////////////Properties////////////////////////////////////////

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;
	// float3 cAdjustCenter			 = {0, 0, 0};
	// float2 cScale 					 = {1, 1};
};
//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
	Precision = Medium;
};
sampler sBaseSamplerClip : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Point;
	Address = FromTexture;
	SRGB = false;
	Precision = Medium;
};

texture2D tDirLightMap : layout(t2) = "White";
sampler sDirLightMapSampler : layout(s2)
{
	Texture = tDirLightMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
	Precision = Medium;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct LODVertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.

//	float4 		WorldPivot	 	: TEXCOORDN;			//  WorldPivot

	float 		LinearZ			: TEXCOORDN;

	float 		TexFrameWeight  : TEXCOORDN;

	[@_INSTANCED]
    float4 Transform0           :TEXCOORDN;
    [@_INSTANCED]
    float4 Transform1           :TEXCOORDN;
    [@_INSTANCED]
    float4 Transform2           :TEXCOORDN;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
#include "Lighting.fxh" 
LODVertexShadingOutput LODVertexShading(in VertexAssemble IN)
{
	LODVertexShadingOutput OUT;

	// World Position
	float3 EntityWPosition = float3(0, 0, 0);
	float3 EntityScale = float3(1, 1, 1);
	if(@_INSTANCED)
	{
		EntityWPosition = float3(IN.Transform0.w, IN.Transform1.w, IN.Transform2.w);
		EntityScale = float3(	length(float3(IN.Transform0.x, IN.Transform1.x, IN.Transform2.x)), 
								length(float3(IN.Transform0.y, IN.Transform1.y, IN.Transform2.y)),
								length(float3(IN.Transform0.z, IN.Transform1.z, IN.Transform2.z))
							);
	}
	else
	{
		EntityWPosition = float3(World[3]);
		if(@_METAL)
		{
			EntityScale = float3(
				length(float3(World[0].x, World[1].x, World[2].x)), 
				length(float3(World[0].y, World[1].y, World[2].y)),
				length(float3(World[0].z, World[1].z, World[2].z))
			  );
		}
		else
		{
			EntityScale = float3(length(World[0].xyz), length(World[1].xyz), length(World[2].xyz));
		}
	}

	float3 V = CameraPos.xyz - EntityWPosition;
	float3 Y = float3(0, 1, 0);
	float3 RotatedV = float3(0, 0, 1);

	if(@_INSTANCED)
	{
		RotatedV = float3(	dot(V, float3(IN.Transform0.x, IN.Transform1.x, IN.Transform2.x)),
							dot(V, float3(IN.Transform0.y, IN.Transform1.y, IN.Transform2.y)),
							dot(V, float3(IN.Transform0.z, IN.Transform1.z, IN.Transform2.z))
		);
		Y = normalize(float3(IN.Transform0.y, IN.Transform1.y, IN.Transform2.y));
	}
	else
	{
		RotatedV = float3(dot(V, World[0]), dot(V, World[1]), dot(V, World[2]));
		Y = normalize(World[1].xyz);
	}

	float3 X = normalize(cross(Y, V));
	OUT.WorldPosition = float4(EntityWPosition + Y * IN.Position.y * EntityScale.y - X * IN.Position.x * EntityScale.x, 1.0);
	// OUT.WorldPosition = float4(OUT.WorldPosition.x, OUT.WorldPosition.y, OUT.WorldPosition.z, 1);
	OUT.ViewDirection.xyz = -V;

	// TexCoord
	float2 StaticUV[8] = {
		float2(0.00, 0.5),
		float2(0.25, 0.5),
		float2(0.50, 0.5),
		float2(0.75, 0.5),
		float2(0.00, 0.0),
		float2(0.25, 0.0),
		float2(0.50, 0.0),
		float2(0.75, 0.0)
	};

	V = normalize(-float3(RotatedV.x, 0, RotatedV.z)); // Must Rotate 180

	float theta = acosFast4(V.z);
	float indexf = theta / (0.25 * PI);
	int index = (int)max(0, min(3, floor(indexf))); // [0 - 3)
	float weight = frac(indexf);

	index = (V.x >= 0) ? index : (7 - index);
	weight = (V.x >= 0) ? weight : (1.0 - weight);
	
	OUT.TexCoord.xy = StaticUV[index] + IN.TexCoord.xy * float2(0.25, 0.5);
	OUT.TexCoord.zw = StaticUV[int((index + 1.0) % 8.0)] + IN.TexCoord.xy * float2(0.25, 0.5);
	OUT.TexFrameWeight = 1.0 - weight;

	// Fog T
	if (@_FOG)
		OUT.ViewDirection.w = FogTransmittance(OUT.ViewDirection.xyz);

	// HPosition
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	OUT.LinearZ = OUT.HPosition.w * CameraInfo.z;
	if (@_INSTANCED)
	{
		OUT.Transform0 = IN.Transform0;
		OUT.Transform1 = IN.Transform1;
		OUT.Transform2 = IN.Transform2;
	}
	return OUT;
}
half DirectLightMap(in half4 GIMap, in float4 Transform0, in float4 Transform1, in float4 Transform2)
{
	half3 localRlight = half3(-0.816, 0.577, 0);
	half3 localGlight = half3(0.408, 0.577, -0.707);
	half3 localBlight = half3(0.408, 0.577, 0.707);
	half3 worldRlight = half3(1,0,0);
	half3 worldGlight = half3(0,1,0);
	half3 worldBlight = half3(0,0,1);
	if(@_INSTANCED)
	{
		worldRlight = normalize(Matrix4x3TransformN(Transform0, Transform1, Transform2, localRlight));
		worldGlight = normalize(Matrix4x3TransformN(Transform0, Transform1, Transform2, localGlight));
		worldBlight = normalize(Matrix4x3TransformN(Transform0, Transform1, Transform2, localBlight));
	}
	else
	{
		worldRlight = normalize(MulMat3x4((float3x3)World, localRlight));
		worldGlight = normalize(MulMat3x4((float3x3)World, localGlight));
		worldBlight = normalize(MulMat3x4((float3x3)World, localBlight));
	}
	half RGI = GIMap.r * saturate(dot(SunDirection.xyz, worldRlight));
	half GGI = GIMap.g * saturate(dot(SunDirection.xyz, worldGlight));
	half BGI = GIMap.b * saturate(dot(SunDirection.xyz, worldBlight));
	return RGI + GGI + BGI;
}
//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in LODVertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	//init MaterialParameters (always need)
	InitShadingParameters(MaterialParameters);
	half4 BaseColor1 = tBaseMap.SampleLevel(sBaseSampler, IN.TexCoord.xy, 0);
	half4 BaseColor2 = tBaseMap.SampleLevel(sBaseSampler, IN.TexCoord.zw, 0);
	half4 BaseColor = lerp(BaseColor2, BaseColor1, IN.TexFrameWeight);
	MaterialParameters.BaseColor = DecodeGamma(BaseColor.rgb);
	MaterialParameters.Roughness = 0.8;
	MaterialParameters.WorldNormal = normalize(-IN.ViewDirection.xyz);
	
	//Normal
    // {
	// 	half4 Normal1 = tDirLightMap.SampleLevel(sDirLightMapSampler, IN.TexCoord.xy, 0);
	// 	half4 Normal2 = tDirLightMap.SampleLevel(sDirLightMapSampler, IN.TexCoord.zw, 0);
	// 	half4 Normal = lerp(Normal2, Normal1, IN.TexFrameWeight);
	// 	half3 LocalNormal = 2 * Normal.rgb - 1;
	// 	LocalNormal = half3(-LocalNormal.x, LocalNormal.y, -LocalNormal.z);
	// 	if(@_INSTANCED)
	// 		MaterialParameters.WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, LocalNormal));
	// 	else
	// 		MaterialParameters.WorldNormal = normalize(MulMat3x4((float3x3)World, LocalNormal));
	// 	//MaterialParameters.AO = Normal.a;
    // }
	
	//Apply Weather
	MaterialParameters.BaseColor *= 1 - 0.3 * RainRate;
	MaterialParameters.Roughness *= 1 - 0.3 * RainRate;

	//Apply Burn
	half BurnRadius = GlobalBurnParam.w;
	half3 BurnPosition = GlobalBurnParam.xyz;
	if(BurnRadius > 0.01)
	{
		half BurnWidth = 80;
		half BurnArea = saturate((length(IN.WorldPosition-BurnPosition) - BurnRadius) / BurnWidth);
		MaterialParameters.BaseColor *= lerp(0.04 , 1 ,BurnArea);
	}

	//lighting
	half4 light1 = tDirLightMap.SampleLevel(sDirLightMapSampler, IN.TexCoord.xy, 0);
	half4 light2 = tDirLightMap.SampleLevel(sDirLightMapSampler, IN.TexCoord.zw, 0);
	half4 lightMap = lerp(light2, light1, IN.TexFrameWeight);
	MaterialParameters.AO = lightMap.a;
	float4 Transform0 = float4(0,0,0,0);
	float4 Transform1 = float4(0,0,0,0);
	float4 Transform2 = float4(0,0,0,0);
	if (@_INSTANCED)
	{
		Transform0 = IN.Transform0;
		Transform1 = IN.Transform1;
		Transform2 = IN.Transform2;
	}
	MaterialParameters.CustomData.r = DirectLightMap(lightMap, Transform0, Transform1, Transform2);
}

//////////////////////////////////PS////////////////////////////////////////
float4 LODPixelShading(in LODVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);
	MaterialShadingParameters MaterialParameters;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	OUT.a = MaterialParameters.Alpha;
	MaterialParameters.DiffuseColor = Diffuse_Lambert(MaterialParameters.BaseColor - MaterialParameters.BaseColor * MaterialParameters.Metallic);
	//GILighting
	MaterialParameters.GILighting.rgb = max(half3(0.01,0.01,0.01), SunFogColor.rgb/max(1, SunFogColor.w)) + AmbientColor.rgb;
	MaterialParameters.GILighting *= MaterialParameters.DiffuseColor * saturate(MaterialParameters.AO + 0.5);
	//Lighting
	half DirLightMap = MaterialParameters.CustomData.r;
	half3 Diffuselighting = MaterialParameters.DiffuseColor * DirLightMap * SunColor.xyz * ThunderRate;
	OUT.rgb = MaterialParameters.GILighting + Diffuselighting;
	// TODO: Fog shading.
	if (@_FOG)
	{
		float3 fogFinalColor = float3(0,0,0);
		FogLightingAndScatter(IN.ViewDirection.w, fogFinalColor, normalize( -IN.ViewDirection.xyz ), SunDirection.rgb);
		OUT.rgb = lerp(fogFinalColor, OUT.rgb, IN.ViewDirection.w);
	}
	//Correct HDR & EncodeHDR color
	EncodeHDR(OUT.rgb);
	return OUT;
}

technique Shading
{
	VertexShader = LODVertexShading();
	PixelShader = LODPixelShading();
}

//////////////////////////////////GBuffer////////////////////////////////////////
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	
	float4 GBufferB 	: SV_Target1;	
	float4 GBufferC 	: SV_Target2;	
	float4 GBufferD 	: SV_Target3;	
	float4 GBufferE 	: SV_Target4;	
	float4 GBufferF  	: SV_Target5;	
	float4 GBufferG 	: SV_Target6;	
};
GBufferOUTPUT GBufferGenPixelShading(in LODVertexShadingOutput IN)
{
	GBufferOUTPUT OUT;
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	half Alpha = MaterialParameters.Alpha;
	//GILighting
	MaterialParameters.GILighting.rgb = max(half3(0.01,0.01,0.01), SunFogColor.rgb/max(1, SunFogColor.w)) + AmbientColor.rgb;
	MaterialParameters.GILighting *= saturate(MaterialParameters.AO + 0.5) * LITMAP_GISCALE;
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	EncodeGBufferData(GBuffer, OUT.GBufferA, OUT.GBufferB, OUT.GBufferC, OUT.GBufferD, OUT.GBufferE, OUT.GBufferF);
	OUT.GBufferG = EncodeDepth(IN.LinearZ);
	OUT.GBufferA.w = EncodeDepth(IN.LinearZ);
	return OUT;
}

technique GBufferGen
{
	VertexShader = LODVertexShading();	
	PixelShader = GBufferGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
[barycentric("sample")]
float4 LODClipShading(in LODVertexShadingOutput IN) : SV_Target
{
	half4 BaseColor1 = tBaseMap.SampleLevel(sBaseSamplerClip, IN.TexCoord.xy, 0);
	half4 BaseColor2 = tBaseMap.SampleLevel(sBaseSamplerClip, IN.TexCoord.zw, 0);
	BaseColor1.a = saturate(BaseColor1.a + 0.2);//clip offset
	BaseColor2.a = saturate(BaseColor2.a + 0.2);//clip offset
	half4 BaseColor = lerp(BaseColor2, BaseColor1, IN.TexFrameWeight);
	
	clip(BaseColor.a - 0.5);

	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = LODVertexShading();
	PixelShader = LODClipShading();
}

/////////////////////////////////////ShadowGen/////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC 				: TEXCOORDN;			//	TexCoord
};

ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;

	// World Position
	float3 EntityWPosition = float3(0, 0, 0);
	float3 EntityScale = float3(1, 1, 1);
	if(@_INSTANCED)
	{
		EntityWPosition = float3(IN.Transform0.w, IN.Transform1.w, IN.Transform2.w);
		EntityScale = float3(	length(float3(IN.Transform0.x, IN.Transform1.x, IN.Transform2.x)), 
								length(float3(IN.Transform0.y, IN.Transform1.y, IN.Transform2.y)),
								length(float3(IN.Transform0.z, IN.Transform1.z, IN.Transform2.z))
							);
	}
	else
	{
		EntityWPosition = float3(World[3]);
		if(@_METAL)
		{
			EntityScale = float3(
				length(float3(World[0].x, World[1].x, World[2].x)), 
				length(float3(World[0].y, World[1].y, World[2].y)),
				length(float3(World[0].z, World[1].z, World[2].z))
			  );
		}
		else
		{
			EntityScale = float3(length(World[0].xyz), length(World[1].xyz), length(World[2].xyz));
		}
	}

	float3 V = SunDirection;
	float3 Y = float3(0, 1, 0);
	float3 RotatedV = float3(0, 0, 1);

	if(@_INSTANCED)
	{
		RotatedV = float3(	dot(V, float3(IN.Transform0.x, IN.Transform1.x, IN.Transform2.x)),
							dot(V, float3(IN.Transform0.y, IN.Transform1.y, IN.Transform2.y)),
							dot(V, float3(IN.Transform0.z, IN.Transform1.z, IN.Transform2.z))
		);
		Y = normalize(float3(IN.Transform0.y, IN.Transform1.y, IN.Transform2.y));
	}
	else
	{
		RotatedV = float3(dot(V, World[0]), dot(V, World[1]), dot(V, World[2]));
		Y = normalize(World[1].xyz);
	}

	float3 X = normalize(cross(Y, V));
	float4 WorldPosition = float4(EntityWPosition + Y * IN.Position.y * EntityScale.y - X * IN.Position.x * EntityScale.x, 1.0);

	// TexCoord
	float2 StaticUV[8] = {
		float2(0.00, 0.5),
		float2(0.25, 0.5),
		float2(0.50, 0.5),
		float2(0.75, 0.5),
		float2(0.00, 0.0),
		float2(0.25, 0.0),
		float2(0.50, 0.0),
		float2(0.75, 0.0)
	};

	V = normalize(-float3(RotatedV.x, 0, RotatedV.z)); // Must Rotate 180

	float theta = acosFast4(V.z);
	float indexf = theta / (0.25 * PI);
	int index = (int)max(0, min(3, floor(indexf))); // [0 - 3)
	float weight = frac(indexf);

	index = (V.x >= 0) ? index : (7 - index);
	weight = (V.x >= 0) ? weight : (1.0 - weight);
	
	OUT.TC = StaticUV[index] + IN.TexCoord.xy * float2(0.25, 0.5);

	// HPosition
	OUT.HPosition = MulMat4x4(ViewProj, float4(WorldPosition.xyz, 1.0));
	OUT.SceneDepth = OUT.HPosition.zw;

	return OUT;
}

[target("R32F")]
float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	half4 BaseColor = tBaseMap.SampleLevel(sBaseSampler, IN.TC, 0);
	half Alpha = BaseColor.a;
	
	if (@_CLIPPING)
	{
		clip(Alpha - 0.5);
	}

	float depth = IN.SceneDepth.x / IN.SceneDepth.y;	
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

#ifndef __FILE_DEBUGSHADER_FXH__
#define __FILE_DEBUGSHADER_FXH__
// Debug pixel shader.
float4 DebugPixelShading(in LODVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);
	GBufferData GBuffer;
	MaterialShadingParameters MaterialParameters;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	half Alpha = MaterialParameters.Alpha;
	//GILighting
	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
	MaterialParameters.GILighting *= MaterialParameters.AO;
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	if ((int)DebugFlag.x == 1)								//1-base color
	{
		OUT.rgb = MaterialParameters.BaseColor.rgb;
	}
	else if((int)DebugFlag.x == 2)							//2-Metallic
	{
		OUT.rgb = MaterialParameters.Metallic.rrr;
	}
	else if ((int)DebugFlag.x == 3)							//3-roughness
	{
		OUT.rgb = MaterialParameters.Roughness.rrr;
	}
	else if ((int)DebugFlag.x == 4)							//4-diffuse color
	{
		OUT.rgb = GBuffer.DiffuseColor.rgb;
	}
	else if ((int)DebugFlag.x == 5)							//5-specular color
	{
		OUT.rgb = GBuffer.SpecularColor.rgb;
	}
	else if ((int)DebugFlag.x == 6)							//6-normal
	{
		OUT.rgb = 0.5+0.5*GBuffer.WorldNormal;
	}else if((int)DebugFlag.x == 150)
	{
		OUT.rgb = Diffuse_Lambert(GBuffer.BaseColor.rgb - GBuffer.BaseColor.rgb * MaterialParameters.Metallic);
	}else if((int)DebugFlag.x == 151)
	{
		OUT.rgb = EncodeNormal(GBuffer.WorldNormal);
		OUT.a = IN.LinearZ;
	}
	return OUT;
}
technique Debug
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = DebugPixelShading();
}
#endif // __FILE_DEBUGSHADER_FXH__
