#ifndef __FILE_DEBUGSHADER_FXH__
#define __FILE_DEBUGSHADER_FXH__
//////////////////////////////////////////////////////////////////////////
// Debug pixel shader.
float4 DebugPixelShading(in VertexShadingOutput IN, in bool FrontFace : SV_IsFrontFace ) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	//1, Gen MaterialParameters & GBufferData
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	{
		PS_GetParameters(IN, MaterialParameters);
		//GBufferData(in deferred) doesn't need alpha, they are all opaque meshes.
		OUT.a = MaterialParameters.Alpha;	
		//only disable pre clip pass on es2, support clpping + transparent
		if (@_ES20 || @_TRANSPARENT)
		{
			// Clipping
			float clipValue = 1.0;
			if (@_CLIPPING)
			{
				clipValue = OUT.a - ALPHA_REF;
			}
			if (@_DISSOLVING)
			{
				clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(IN.HPosition.xy));
			}
			if (@_CLIPPING || @_DISSOLVING)
				clip(clipValue);
		}
		//apply weather
#if !IS_WATER
		SNOW_OVER(MaterialParameters.BaseColor, MaterialParameters.WorldNormal, MaterialParameters.Metallic, MaterialParameters.Roughness, IN);
#endif
		SetGBufferForShadingModel(GBuffer, MaterialParameters);
		RAIN_EFFECT(GBuffer.DiffuseColor, GBuffer.SpecularColor, GBuffer.Roughness, GBuffer.WorldNormal);
	}


	//2, lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = (MaterialParameters.GILighting + AmbientColor.rgb) * MaterialParameters.AO;
	LitResult.SpecularLighting = half3(0, 0, 0);
	half3 V = normalize(-IN.ViewDirection.xyz);
	half NoV = saturate(dot(GBuffer.WorldNormal, V));
	//GI lighting
	FetchPrecomputedLighing(IN, MaterialParameters);
	//InitIBLReflection(IN, MaterialParameters, V, NoV);
	//sun
	{
		int PBRQualityLevel = PBRQualityLevel_None;
		if (@_LOD_LEVEL <= LOD_Level_2 && @_EnableReflection)
			PBRQualityLevel = PBRQualityLevel_Normal;
		LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);
		LitData.InShadow = GetDynamicShadowForSun(IN);
		IntegrateBxDF(LitResult, GBuffer, LitData);
	}
	//vs lighting
	LitResult.DiffuseLighting += IN.DiffLighting.rgb;
	//emission
	OUT.rgb += GBuffer.Emission;
	//point lighting
	if(@_LOD_LEVEL <= LOD_Level_2)
	{
		if( @_LitCount >= 1 && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) >= 0 )
		{
			LightingData LitData = GetPointLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_None, Lights0);
			IntegrateBxDF(LitResult, GBuffer, LitData);
		}
		if( @_LitCount >= 2 && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) >= 1 )
		{
			LightingData LitData = GetPointLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_None, Lights1);
			IntegrateBxDF(LitResult, GBuffer, LitData);
		}
	}
	//hardcode virtual lit, feel free to change it.
#if VIRTUAL_LIT
	if (!@_LIGHT_MAP)//always has virtual light
	{
		half3 L = cVirtualLitDir.xyz;
		half NoL = saturate(dot(L, GBuffer.WorldNormal));
		LitResult.DiffuseLighting += cVirtualLitColor.rgb * NoL;
	}
#endif
	half3 DebugUV = MaterialParameters.GILighting;
	
	{
		if ((int)DebugFlag.x == 1)								//1-base color
		{
			OUT.rgb = MaterialParameters.BaseColor.rgb;
		}
		else if((int)DebugFlag.x == 2)							//2-Metallic
		{
			OUT.rgb = MaterialParameters.Metallic.rrr;
		}
		else if ((int)DebugFlag.x == 3)							//3-roughness
		{
			OUT.rgb = MaterialParameters.Roughness.rrr;
		}
		else if ((int)DebugFlag.x == 4)							//4-diffuse color
		{
			OUT.rgb = GBuffer.DiffuseColor.rgb;
		}
		else if ((int)DebugFlag.x == 5)							//5-specular color
		{
			OUT.rgb = GBuffer.SpecularColor.rgb;
		}
		else if ((int)DebugFlag.x == 6)							//6-normal
		{
			OUT.rgb = 0.5+0.5*GBuffer.WorldNormal;
		}
		else if ((int)DebugFlag.x == 7)							//7-tangent
		{
			OUT.rgb = half3(0,0,0);
#if IS_TERRAIN
			half3 t = normalize(cross(IN.WorldNormal.xyz, half3(0, 0, 1)));
			OUT.rgb = 0.5+0.5*t;
#else
			[@_TANGENT]
			OUT.rgb = 0.5+0.5*IN.WorldTangent.xyz;
#endif
		}
		else if ((int)DebugFlag.x == 8)							//8-binormal
		{
			OUT.rgb = half3(0,0,0);
#if IS_TERRAIN
			half3 t = normalize(cross(IN.WorldNormal.xyz, half3(0, 0, 1)));
			half3 b = normalize(cross(IN.WorldNormal.xyz, t));
			OUT.rgb = 0.5+0.5*b;
#else
			[@_TANGENT]
			OUT.rgb = 0.5+0.5*IN.WorldBinormal.xyz;
#endif
		}
		else if ((int)DebugFlag.x == 9)							//9-1uv
		{
			OUT.rgb = half3(BaseMapUV(IN.TexCoord),1.0);
		}
		else if ((int)DebugFlag.x == 10)						//10-2uv
		{
			OUT.rgb = half3(0,0,0);
#if IS_TERRAIN
			OUT.rgb = half3(IN.TexCoord.xy,1.0);
#else
			if(@_LIGHT_MAP)
			{
				OUT.rgb = half3(LightMapUV(IN.TexCoord),1.0);
			}
			else
			{
				OUT.rgb = DebugUV;
				return OUT;
			}
#endif
		}
		else if ((int)DebugFlag.x == 11)						//11-vertex color
		{
			OUT.rgb = half3(0,0,0);
			[@_VERTEXCOLOR]
			OUT.rgb = half3(IN.TintColor.xyz);
		}
		//lighting
		else if ((int)DebugFlag.x == 20)						//20-bake lighting
		{
			OUT.rgb = LitResult.DiffuseLighting * GBuffer.DiffuseColor;
		}
		else if ((int)DebugFlag.x == 21)						//21-specular lighting
		{
			if (@_EnableReflection)
			{
				half halfSize = 0.01*DiffuseMapSize;
				float2 AB = step(frac(halfSize.rr * IN.TexCoord.xy), 0.5.rr);
				OUT.rgb = (AB.x == AB.y)? LitResult.SpecularLighting : lerp(LitResult.SpecularLighting, half3(1,0,0), 0.1);
			}
			else
				OUT.rgb = half3(0, 0, 0);
		}
		else if ((int)DebugFlag.x == 22)						//22-emission lighting
		{
			OUT.rgb = MaterialParameters.Emission;
		}
		else if ((int)DebugFlag.x == 23)						//23-bake lighting only
		{
			OUT.rgb = MaterialParameters.GILighting * Diffuse_Lambert(half3(1,1,1)) / PI;
		}
		//density
		else if ((int)DebugFlag.x == 30)						//30-lightmap check
		{
			OUT.rgb = half3(0, 0, 0);
#if !IS_TERRAIN
			if (@_LIGHT_MAP)
#endif
			{
				half halfSize = 0.5*LightMapSize;
				float2 AB = step(frac(halfSize.rr * LightMapUV(IN.TexCoord)), 0.5.rr);
				half density = (AB.x == AB.y)? 1 : 0.5;
				OUT.rgb = LitResult.DiffuseLighting * Diffuse_Lambert(half3(1,1,1)) * density;
			}
		}
		else if ((int)DebugFlag.x == 31)						//31-lightmap density
		{
			OUT.rgb = half3(0, 0, 0);
#if !IS_TERRAIN
			if (@_LIGHT_MAP)
#endif
			{
				float WorldSpaceArea = length(cross(ddx(IN.WorldPosition.xyz), ddy(IN.WorldPosition.xyz)));
				float2 A = ddx(IN.TexCoord.zw * LightMapSize);
				float2 B = ddy(IN.TexCoord.zw * LightMapSize);
				float2 C = A.xy * B.yx;
				float area = abs(C.x - C.y);
				float density = sqrt(area / WorldSpaceArea);
				float bestDensity = 10;//[0-10-20]
				density /= bestDensity;
				if (density < 1)
					OUT.rgb = half3(0, density, 1 - density);
				else
				{
					density = (density - 1);
					OUT.rgb = half3(min(density, 1), max(1 - density, 0), 0);
				}
			}
		}
		else if ((int)DebugFlag.x == 32)						//32-texture check
		{
			half halfSize = 0.05*DiffuseMapSize;
			float2 AB = step(frac(halfSize.rr * IN.TexCoord.xy), 0.5.rr);
			OUT.rgb = (AB.x == AB.y)? GBuffer.DiffuseColor*half3(0.5,0.5,0.5) : half3(1,1,1);
		}
		else if ((int)DebugFlag.x == 33)						//33-texture density
		{
			float WorldSpaceArea = length(cross(ddx(IN.WorldPosition.xyz), ddy(IN.WorldPosition.xyz)));
			float2 A = ddx(IN.TexCoord.xy * DiffuseMapSize);
			float2 B = ddy(IN.TexCoord.xy * DiffuseMapSize);
			float2 C = A.xy * B.yx;
			float area = abs(C.x - C.y);
			float density = sqrt(area / WorldSpaceArea); // 越大表示贴图缓存越低
			float bestDensity = 100;//[0-100-200]
			density /= bestDensity;
			if (density < 1)
			{
				OUT.rgb = half3(0, 0, density); // 黑到偏蓝就是正常
			}	
			else
			{
				density = (density - 1);
				OUT.rgb = half3(min(density, 1), max(1 - density, 0), 0); // 绿到红就是渐渐不合理了
			}
		}
		else if ((int)DebugFlag.x == 34)						//34-random color
		{
			OUT.rgb = cParamter2.rgb;
		}
		//shader instructions
		else if ((int)DebugFlag.x == 40)						//40-ps instructions
		{
			float bestInstructions = 100;//[0-100-200]
			float val = PsInstructions / bestInstructions;
			if (val < 1)
				OUT.rgb = half3(0, val, 1 - val);
			else
			{
				val = (val - 1);
				OUT.rgb = half3(min(val, 1), max(1 - val, 0), 0);
			}
		}
		else if ((int)DebugFlag.x == 41)						//41-ps textures
		{
			float bestTextures = 3;//[0-3-6]
			float val = PsTextureLods / bestTextures;
			if (val < 1)
				OUT.rgb = half3(0, val, 1 - val);
			else
			{
				val = (val - 1);
				OUT.rgb = half3(min(val, 1), max(1 - val, 0), 0);
			}
		}
		else if ((int)DebugFlag.x == 42)						//42-triangle density
		{
			float bestDensity = 1;//[0-1-2]
			float val = TiangleDensity / bestDensity;
			if (val < 1)
				OUT.rgb = half3(0, val, 1 - val);
			else
			{
				val = (val - 1);
				OUT.rgb = half3(min(val, 1), max(1 - val, 0), 0);
			}
		}
		else if ((int)DebugFlag.x == 43)						//43-tiangle count
		{
			float bestCount = 2500;//[0-2500-5000]
			float val = TiangleCount / bestCount;
			if (val < 1)
				OUT.rgb = half3(0, val, 1 - val);
			else
			{
				val = (val - 1);
				OUT.rgb = half3(min(val, 1), max(1 - val, 0), 0);
			}
		}
		else
		{
			OUT.rgb = half3(1, 0, 0);
		}

		EncodeHDR(OUT.rgb);
	}

	if ((int)DebugFlag.x == 100)								//100-Capture base color for impostor
	{
		OUT.rgb = EncodeGamma(GBuffer.BaseColor.rgb);
	}
	else if ((int)DebugFlag.x == 101)							//101-Capture normal for impostor
	{
		OUT.rgb = 0.5 + 0.5 * GBuffer.WorldNormal;
	}else if((int)DebugFlag.x == 150)
	{
		OUT.rgb = Diffuse_Lambert(GBuffer.BaseColor.rgb - GBuffer.BaseColor.rgb * MaterialParameters.Metallic);
		if(MaterialParameters.Metallic > 0.5)
		{
			half3 R = reflect( -V, GBuffer.WorldNormal );
			OUT.rgb = IBL_Specular(GBuffer, R, NoV);
		}
		OUT.a = IN.LinearZ;
		return OUT;
	}else if((int)DebugFlag.x == 151)
	{
		OUT.rgb = EncodeNormal(normalize(GBuffer.WorldNormal));
		OUT.a = IN.LinearZ;
		return OUT;
	}

	// encode hdr
	if (@_TRANSPARENT)
	{
		if (@_HDR)
			OUT.rgb = OUT.rgb * OUT.a;
		else
			OUT.rgb = saturate(OUT.rgb) * OUT.a;
		OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	}
	else
	{
		OUT.a = EncodeSolidBloomFactor(0);
	}
	return OUT;
}

technique Debug
{
	VertexShader = PbrVertexShading();
	PixelShader = DebugPixelShading();
}
#endif // __FILE_DEBUGSHADER_FXH__
