#ifndef __FILE_EDGELINE_FXH__
#define __FILE_EDGELINE_FXH__
/////////////////////////////////////EdgeLine/////////////////////////////////////
struct EdgeLineVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		TC 				: TEXCOORDN;			//	TexCoord
};

//viewProjVS is light view now
EdgeLineVertexShadingOutput EdgeLineVertexShading(in VertexAssemble IN)
{
	EdgeLineVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (EdgeLineVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	//world normal
	float3 WorldNormal;
	if (@_INSTANCED)
	{
		WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
	}
	else
	{
		WorldNormal = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	}

	float ScreenRate = ScreenInfo.y / ScreenInfo.x;

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	
	float3 offsetPos = WorldPosition.xyz + WorldNormal;
	float4 offsetPosH = MulMat4x4(ViewProj, float4(offsetPos,1));
	float2 dir = offsetPosH.xy/offsetPosH.w - OUT.HPosition.xy/OUT.HPosition.w;

	dir = normalize(dir);
	dir.x *= ScreenRate;

	OUT.HPosition.xy += cParamter2.w * dir * OUT.HPosition.w;

//	OUT.HPosition.z -= 0.001 * OUT.HPosition.w;			//Not Sure

	OUT.TC = IN.TexCoord.xy;
	return OUT;
}

float4 EdgeLinePixelShading(in EdgeLineVertexShadingOutput IN) : SV_Target
{
	if ((@_CLIPPING||@_DISSOLVING) && @_BaseMapEnable)
	{
		half Alpha = tBaseMap.SampleRGBA(sBaseSampler, IN.TC).a;
		float clipValue = 1.0;
		if (@_CLIPPING)
		{
			clipValue = Alpha - ALPHA_REF;
		}
		if (@_DISSOLVING)
		{
			clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(IN.HPosition.xy));
		}
		if (@_CLIPPING || @_DISSOLVING)
			clip(clipValue);
	}
	return float4(cParamter.rgb, 1.0);
}

technique EdgeLine
{
	VertexShader = EdgeLineVertexShading();
	PixelShader = EdgeLinePixelShading();
}
#endif // __FILE_EDGELINE_FXH__
