///	@file	<FX_Lighting.fx>
///	@path	~/System/Shader/
///	@date	2019/01/--
///	@brief	FX shading model.

#ifndef __FILE_FX_SHADING_FXH__
#define __FILE_FX_SHADING_FXH__

// #ifndef FORCE_DISABLE_SHADOW
// #define FORCE_DISABLE_SHADOW 0
// #endif

#ifndef FORCE_ENABLE_FOG
#define FORCE_ENABLE_FOG false
#endif

@_PARTICLECLIPPING
{
	Name = ParticleClipping;
	Description = "Particle Clipping";
	Visible = false;
	Default = false;
}

#include "Lighting.fxh"

//Lighting Calculation using DynamicGI() in Lighting.fxh
//When using SH_IBL , Particle System should Undef SD_SH_IBL or Define as 0
// #ifdef SD_SH_IBL
// 	#undef SD_SH_IBL
// #endif

float4 UnlitPixelShading(in const VertexShadingOutput IN,in float4 screenCoord : SV_Position) : SV_Target
{

	// 1. Init Shading Parameters Struct
    EffectShadingParameters materialParamters;
    InitEffectShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(IN,materialParamters);

	// 3. Premultiply Alpha
	materialParamters.BaseColor.rgb *= materialParamters.BaseColor.a;
	materialParamters.BaseColor.a *= GetEffectParam(IN).z;

	// 4. Soft
	if (@_SOFT)
	{
		SoftParticle(IN,materialParamters,screenCoord);
	}	


	EncodeHDRParticle(materialParamters.BaseColor);

	if(@_TRANSPARENT || !@_DEFERRED_SHADING)
		InitFogLightingEffect(IN,materialParamters.BaseColor);

    return materialParamters.BaseColor;
}

float4 DistortionPixelShading(in const VertexShadingOutput IN,in const float4 screenCoord : SV_Position) : SV_Target
{

	// 1. Init Shading Parameters Struct	
	EffectShadingParameters materialParamters;
	InitEffectShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
	PS_GetParameters(IN,materialParamters);

	half2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

	float4 OUT = float4(0,0,0,0);
	// calcuate patch normal
	half3 P = IN.WorldPosition;
	half3 N = IN.WorldNormal;

	// hacked refraction
	half3 V = IN.ViewDirection;
	half3 R = cross(N, V);
	half3 D = cross(R, V);

	half3 VR = cross(float3(0, 1, 0), V);
	half3 VU = normalize(cross(V, VR));
	VR = cross(VU, V);

	half2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
	offset *= (500 / IN.TexCoord.z) * materialParamters.BaseColor.x;

	OUT = SampleColorMap(ScreenTexture, ScreenTexture_Sampler, stc + offset);

	OUT.a = 1;

	return OUT;

}

float4 LightingPixelShading(in const VertexShadingOutput IN,EffectShadingParameters materialParamters,in const float4 screenCoord : SV_Position) : SV_Target
{
	// 3. Get ScreenInfo For Shadow
	half2 shadowAndAO = half2(0,0);
	// if(FORCE_DISABLE_SHADOW == 0)
	// {	
	// 	float2 spos = IN.HPosition.xy;
	// 	[@_D3D9]
	// 	spos += 0.5;
	// 	float2 stc = spos * ScreenInfo.zw;
	// 	shadowAndAO = tShadowMask.Sample(sShadowMaskSampler, stc).rg;
	// }
	// There is No tShadowMask For Particle
	materialParamters.Roughness = max(materialParamters.Roughness,MIN_ROUGHNESS);

	half3 N = materialParamters.WorldNormal;
	half InShadow = shadowAndAO.x;
	half3 GILighting = float3(0,0,0); 

 	half3 L = normalize( SunDirection.xyz );
 	half3 V = normalize( IN.ViewDirection.xyz );
 	half3 H = normalize( V + L);

	half NoH = saturate(dot(N,H));
 	half NoL = dot(materialParamters.WorldNormal, L);
	half NoLS = saturate(NoL);
	half NoV = saturate(dot(N,V));
	// half VoH = saturate(dot(V,H));
	half VoL = saturate(-dot(L,V));

	half lightDensity = henyeyGreenstein(VoL,ScatteringPhaseG) * 5.0 ;
	half3 inputRadiance = saturate( NoL * Thickness + ( 1-Thickness ) ) * (1 - InShadow) + lightDensity * Scattering ; 

	half3 SpecularColor = lerp(half3(0.04,0.04,0.04) , materialParamters.BaseColor.xyz, (half)materialParamters.Metallic);
	half3 diffuseColor = Diffuse_Lambert(materialParamters.BaseColor.xyz - materialParamters.BaseColor.xyz * (half)materialParamters.Metallic);

	DynamicGILight(N,GILighting,InShadow);
	GILighting *= (diffuseColor /* *shadowAndAO.y */); // There is No Shadow For Particle

	half D = D_GGX(materialParamters.Roughness,NoH);
	half G = G_Schlick( materialParamters.Roughness, NoV, NoLS );
	// half3 F = F_Schlick( SpecularColor, VoH );
	half3 F = F_None(SpecularColor);

	half3 diffuseLighting = inputRadiance * SunColor.rgb * diffuseColor * ThunderRate;
	half3 SpecularLighting = D * G * F * inputRadiance * SunColor.rgb * ThunderRate;
	SpecularLighting = step(0,NoLS) * SpecularLighting;

	half4 OUT = half4(0,0,0,materialParamters.BaseColor.w);
	
	OUT.rgb += diffuseLighting;
	OUT.rgb += SpecularLighting;
	OUT.rgb += GILighting;

	OUT.rgb *= saturate(materialParamters.BaseColor.w * AlphaScale);

	if(@_TRANSPARENT || !@_DEFERRED_SHADING || FORCE_ENABLE_FOG)
		InitFogLightingEffect(IN,OUT);

	return OUT;
}

float4 EffectUnifyPixelShading(in const VertexShadingOutput IN, in const float4 screenCoord : SV_Position) : SV_Target
{

    if(@_DISTORTION)
	{
        return DistortionPixelShading(IN,screenCoord);
	}
    else
	{
        return UnlitPixelShading(IN,screenCoord);
	}

    return float4(1,0,1,1);
}

float4 EffectLightingPixelShading(in const VertexShadingOutput IN,in const float4 screenCoord : SV_Position) : SV_Target
{
	// 1. Init Shading Parameters Struct	
    EffectShadingParameters materialParamters;
    InitEffectShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(IN,materialParamters);

	// 2.1 Value Check
	materialParamters.Roughness = max(Roughness,MIN_ROUGHNESS);

	materialParamters.BaseColor = LightingPixelShading(IN,materialParamters,screenCoord);

	if (@_SOFT)
		SoftParticle(IN,materialParamters,screenCoord);

	float4 OUT = materialParamters.BaseColor;

	EncodeHDR(OUT.rgb);

	if (@_PARTICLECLIPPING)
		clip(OUT.a - ALPHA_REF);

	return OUT;
}

#endif
