#ifndef __FILE_FROZEN_FXH__
#define __FILE_FROZEN_FXH__
// !notice need PbrVertexShading
float4 FrozenPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	//1, Gen MaterialParameters & GBufferData
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	{
		PS_GetParameters(IN, MaterialParameters);
		//GBufferData(in deferred) doesn't need alpha, they are all opaque meshes.
		OUT.a = MaterialParameters.Alpha;	
		//only disable pre clip pass on es2, support clpping + transparent
		if (@_ES20 || @_TRANSPARENT)
		{
			// Clipping
			float clipValue = 1.0;
			if (@_CLIPPING)
			{
				clipValue = OUT.a - ALPHA_REF;
			}
			if (@_DISSOLVING)
			{
				clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(IN.HPosition.xy));
			}
			if (@_CLIPPING || @_DISSOLVING)
				clip(clipValue);
		}
		//apply weather
#if !IS_WATER
		SNOW_OVER(MaterialParameters.BaseColor, MaterialParameters.WorldNormal, MaterialParameters.Metallic, MaterialParameters.Roughness, IN);
#endif
		MaterialParameters.Metallic = 0.6;
		MaterialParameters.Roughness = 0.4;
		SetGBufferForShadingModel(GBuffer, MaterialParameters);
		RAIN_EFFECT(GBuffer.DiffuseColor, GBuffer.SpecularColor, GBuffer.Roughness, GBuffer.WorldNormal);

		//material parameters
		GBuffer.SpecularColor = half3(0.1,0.2,0.5);
		
	}

	//2, lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	half3 V = normalize(-IN.ViewDirection.xyz);
	half NoV = saturate(dot(GBuffer.WorldNormal, V));
	//GI lighting
	FetchPrecomputedLighing(IN, MaterialParameters);
	//InitIBLReflection(IN, MaterialParameters, V, NoV);
	//sun
	{
		int PBRQualityLevel = PBRQualityLevel_None;
		if (@_LOD_LEVEL <= LOD_Level_2 && @_EnableReflection)
			PBRQualityLevel = PBRQualityLevel_Normal;
		LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);
		LitData.InShadow = GetDynamicShadowForSun(IN);
		IntegrateBxDF(LitResult, GBuffer, LitData);
	}
	//vs lighting
	LitResult.DiffuseLighting += IN.DiffLighting.rgb;
	//emission
	OUT.rgb += GBuffer.Emission;
	//point lighting
	if(@_LOD_LEVEL <= LOD_Level_2)
	{
		if( @_LitCount >= 1 && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) >= 0 )
		{
			LightingData LitData = GetPointLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_None, Lights0);
			IntegrateBxDF(LitResult, GBuffer, LitData);
		}
		if( @_LitCount >= 2 && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) >= 1 )
		{
			LightingData LitData = GetPointLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_None, Lights1);
			IntegrateBxDF(LitResult, GBuffer, LitData);
		}
	}
	//hardcode virtual lit, feel free to change it.
#if VIRTUAL_LIT
	if (!@_LIGHT_MAP)//always has virtual light
	{
		half3 L = cVirtualLitDir.xyz;
		half NoL = saturate(dot(L, GBuffer.WorldNormal));
		LitResult.DiffuseLighting += cVirtualLitColor.rgb * NoL;
	}
#endif

	half3 iceColor = half3(0.1,0.2,0.5);
	half k = (1-NoV);
	OUT.rgb += k * iceColor * 3.8;
	
	//bloom scale dependends specular color only
	OUT.rgb += LitResult.SpecularLighting;
	half bloomScale = 0;
	if(@_LOD_LEVEL <= LOD_Level_2 && !@_DEFERRED_SHADING)
		bloomScale = cBloomScale * dot(LitResult.SpecularLighting + MaterialParameters.Emission, half3(0.6378, 2.1456, 1));
	OUT.rgb += LitResult.DiffuseLighting * GBuffer.DiffuseColor;

	// TODO: improve thunder effect
	OUT.rgb *= ThunderRate;

	//4, Fog shading.
#if !IS_WATER
	if (@_FOG && @_VolumetricFog && @_TRANSPARENT)
	{
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.DiffLighting.w;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
	}
	else if (@_FOG && !@_VolumetricFog)
	{
		OUT.rgb = FogLighting(IN.DiffLighting.w, OUT.rgb);
	}
#else
	if (@_FOG && @_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.DiffLighting.w;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz * OUT.a;
	}
	else if (@_FOG && !@_VolumetricFog)
	{
		OUT.rgb *= IN.DiffLighting.w;
		OUT.rgb += FogColor.rgb * OUT.a * (1-IN.DiffLighting.w);
	}
#endif

	//5, screen color tint
#if USE_LIGHTMAP
	if (@_LIGHT_MAP || LIGHTMAP_ENABLE)
		OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);
#endif

#if IS_WATER
	if(@_HDR)
	{
		OUT.rgb += DecodeHDR(GBuffer.CustomData.rgb);
	}
	else
	{
		OUT.rgb += GBuffer.CustomData.rgb;
	}
#endif

	//always calc fog in linear space
	EncodeHDR(OUT.rgb);
	if (@_TRANSPARENT)
	{
		if (@_HDR)
			OUT.rgb = OUT.rgb * OUT.a;
		else
			OUT.rgb = saturate(OUT.rgb) * OUT.a;
		OUT.a = EncodeTransparentBloomFactor(OUT.a, bloomScale);
	}
	else
	{
#if USE_LIGHTMAP
		if (LIGHTMAP_ENABLE)
		{
			OUT.a = EncodeSolidBloomFactor(bloomScale);
		}
		else
#endif
		{
			OUT.a = EncodeSolidBloomFactor(bloomScale + SunColor.w);
		}
	}

	//bug fix for ios device
	if (@_METAL && @_HDR && !@_ENCODEHDR)
	{
		EncodeHDR(OUT.rgb);
	}

	return OUT;
}
technique Frozen
{
	VertexShader = PbrVertexShading();
	PixelShader = FrozenPixelShading();
}
#endif // __FILE_FROZEN_FXH__
