#ifndef __FILE_GBUFFERGEN_FXH__
#define __FILE_GBUFFERGEN_FXH__
/////////////////////////////////////GBuffer/////////////////////////////////////
#ifndef NO_GBUFFER
#ifndef DEFERRED_PRE_AO
#define DEFERRED_PRE_AO 0
#endif
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	// GBufferA[R16G16B16A16F]	: Emission.rgb with lighting result
	float4 GBufferB 	: SV_Target1;	// GBufferB[R8G8B8A8]		: Roughness, Metallic, AO, ShadingModelID
	float4 GBufferC 	: SV_Target2;	// GBufferC[R8G8B8A8]		: BaseColor.rgb, Specular
	float4 GBufferD 	: SV_Target3;	// GBufferD[R8G8B8A8]		: CustomData
	float4 GBufferE 	: SV_Target4;	// GBufferE[R8G8B8A8]		: N.xyz, InShadow
	float4 GBufferF  	: SV_Target5;	// GBufferF[R8G8B8A8]		: TempRT for OnChipMem optimization
	float4 GBufferG 	: SV_Target6;	// GBufferG[R32F]			: Depth
};
[target("X16F", "X16F", "X16F", "X16F", "X16F", "X16F", "R32F")]
GBufferOUTPUT GBufferGenPixelShading(in VertexShadingOutput IN, in bool FrontFace : SV_IsFrontFace)
{
	GBufferOUTPUT OUT;
	//[@_HLSL]
	//OUT = (GBufferOUTPUT)0;

	if (@_DEFERRED_SHADING && !ENABLE_BURNT)
	{
		//1, Gen MaterialParameters & GBufferData
		MaterialShadingParameters MaterialParameters;
		GBufferData GBuffer;
		PS_GetParameters(IN, MaterialParameters);

	#if USE_ONE_SIDE_LIGHTING
		if(!@_OneSideLightingEnable)
	#endif	
		{
			if(!FrontFace)
			{
				MaterialParameters.WorldNormal = -MaterialParameters.WorldNormal;
			}
		}
		//screen gray
	#if USE_GRAY
		half3 GrayColor = GetLuminance(MaterialParameters.BaseColor);
		MaterialParameters.BaseColor = lerp(GrayColor, MaterialParameters.BaseColor, cSaturation);
	#endif
	#if USE_LIGHTMAP
		if (@_LIGHT_MAP || LIGHTMAP_ENABLE)
			MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, ScreenColor.rgb, ScreenColor.a);
	#endif
		InitScreenGray(IN, MaterialParameters.BaseColor);
		//apply weather
		SNOW_OVER(MaterialParameters.BaseColor, MaterialParameters.WorldNormal, MaterialParameters.Metallic, MaterialParameters.Roughness, IN);

		// half3 V = normalize(-IN.ViewDirection.xyz);
		// half NoV = saturate(dot(MaterialParameters.WorldNormal, V));
		//GI lighting
		FetchPrecomputedLighing(IN, MaterialParameters);
		// InitIBLReflection(IN, MaterialParameters, V, NoV);
		//MaterialParameters.GILighting.rgb = MaterialParameters.GILighting.rgb * SSGIandAO.a + SSGIandAO.rgb * MaterialParameters.DiffuseColor;
		//set GBuffer
		SetGBufferForShadingModel(GBuffer, MaterialParameters);
		//hardcode virtual lit, feel free to change it.
	#if VIRTUAL_LIT
		if (@_SKELETON)//always has virtual light
		{
			half3 V = normalize(-IN.ViewDirection.xyz);
			half NoV = saturate(dot(GBuffer.WorldNormal, V));
			InitVirtualLight(GBuffer, NoV, V);
		}
	#endif
		if(ENABLE_HIGHLIGHT)
		{
			half3 V = normalize(-IN.ViewDirection.xyz);
			half NoV = saturate(dot(GBuffer.WorldNormal, V));
			half k = lerp((1-NoV),1,saturate(cParamter2.z));
			k = k*k;
			k = k*k * 5;
			//in Script , cParamter2.y should be check in [0,1]
			float lightingRate = lerp(saturate(cParamter2.y) , 1 , 2 * abs(0.5-frac(GameTime * cParamter2.w)));
			GBuffer.Emission += k * cParamter.rgb * lightingRate ;		
		}
		
		EncodeGBufferData(GBuffer, OUT.GBufferA, OUT.GBufferB, OUT.GBufferC, OUT.GBufferD, OUT.GBufferE, OUT.GBufferF);
		//Temp RT
		//OUT.GBufferF = float4(0, 0, 0, 0);
		//linear z
		OUT.GBufferG = EncodeDepth(IN.LinearZ);
		//output z to GBufferA's w channel for efficient sss blur
		OUT.GBufferA.w = EncodeDepth(IN.LinearZ);
	}
	else
	{
		OUT.GBufferA = float4(0, 0, 0, 0);
		OUT.GBufferB = float4(1, 0, 0, 0); // roughness默认是1，免得ssr计算
		OUT.GBufferC = float4(0, 0, 0, 0);
		OUT.GBufferD = float4(0, 0, 0, 0);
		OUT.GBufferE = float4(0, 0, 0, 0);
		OUT.GBufferF = float4(0, 0, 0, 0);
		OUT.GBufferG = float4(0, 0, 0, 0);
	}
	return OUT;
}

technique GBufferGen
{
	VertexShader = PbrVertexShading();
#if USE_TESSELLATION
	HullShader = PbrVertexShading_HS();
	DomainShader = PbrVertexShading_DS();
#endif	
	PixelShader = GBufferGenPixelShading();
}

struct RSMBuffer
{
	float4 GBufferA 	: SV_Target0;	// GBufferA[R11G11B10]
	float4 GBufferB 	: SV_Target1;	// GBufferB[R8G8B8A8]
};
RSMBuffer RSMGenPixelShading(in VertexShadingOutput IN, in bool FrontFace : SV_IsFrontFace)
{
	RSMBuffer OUT;
	OUT.GBufferA = float4(0, 0, 0, 0);
	OUT.GBufferB = float4(1, 0, 0, 0);
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	PS_GetParameters(IN, MaterialParameters);
#if USE_ONE_SIDE_LIGHTING
	if(!@_OneSideLightingEnable)
#endif	
	{
		if(!FrontFace)
		{
			MaterialParameters.WorldNormal = -MaterialParameters.WorldNormal;
		}
	}
#if USE_GRAY
	half3 GrayColor = GetLuminance(MaterialParameters.BaseColor);
	MaterialParameters.BaseColor = lerp(GrayColor, MaterialParameters.BaseColor, cSaturation);
#endif
	InitScreenGray(IN, MaterialParameters.BaseColor);
	//apply weather
	SNOW_OVER(MaterialParameters.BaseColor, MaterialParameters.WorldNormal, MaterialParameters.Metallic, MaterialParameters.Roughness, IN);

	float3 diffuseColor = Diffuse_Lambert(MaterialParameters.BaseColor - MaterialParameters.BaseColor * MaterialParameters.Metallic);	
	float3 diffuseLighting = saturate(dot(MaterialParameters.WorldNormal.xyz, SunDirection.xyz)) * SunColor.rgb * diffuseColor;
	OUT.GBufferA.rgb = diffuseLighting;
	OUT.GBufferB.rgb = EncodeNormal(MaterialParameters.WorldNormal);
	return OUT;
}
technique ReflectiveShadowGen
{
	VertexShader = PbrVertexShading();
	PixelShader = RSMGenPixelShading();
}
#endif// NO_GBUFFER
#endif // __FILE_GBUFFERGEN_FXH__
