#ifndef __FILE_HEIGHTMAPGEN_FXH__
#define __FILE_HEIGHTMAPGEN_FXH__
/////////////////////////////////////HeightmapGen/////////////////////////////////////
struct HeightmapVertexShadingOutput
{
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		SceneDepthTC	: TEXCOORDN;			//	Scene Depth
};

//viewProjVS is light view now
HeightmapVertexShadingOutput HeightmapGenVertexShading(in VertexAssemble IN)
{
	HeightmapVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (HeightmapVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	//world normal
	float3 WorldNormal;
	if (@_INSTANCED)
	{
		WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
	}
	else
	{
		WorldNormal = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	}

#ifdef HAS_WORLD_POSITION_OFFSET
	WorldPosition.xyz += WorldPositionOffset(IN, WorldPosition.xyz, WorldNormal);
#endif

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	
	//SceneDepthTC
	OUT.SceneDepthTC.xy = OUT.HPosition.zw;
	OUT.SceneDepthTC.zw = IN.TexCoord.xy;

	return OUT;
}

[target("R32F")]
float4 HeightmapGenPixelShading(in HeightmapVertexShadingOutput IN) : SV_Target
{
#ifndef NO_SHADOW_CLIPPING
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		float alpha = tBaseMap.SampleRGBA( sBaseSampler, IN.SceneDepthTC.zw ).a;
		//clip(alpha - ALPHA_REF);
		clipValue = alpha - ALPHA_REF;
	}
	if (@_TRANSPARENT)
	{
		clipValue = tBaseMap.SampleRGBA( sBaseSampler, IN.SceneDepthTC.zw ).a - 0.3;
	}
	if (@_CLIPPING || @_TRANSPARENT)
		clip(clipValue);
#endif
	float depth = IN.SceneDepthTC.x / IN.SceneDepthTC.y;
	float4 rt = EncodeDepth(depth);
	return rt;
}

technique HeightmapGen
{ 
	VertexShader = HeightmapGenVertexShading();
	PixelShader = HeightmapGenPixelShading();
}
#endif // __FILE_HEIGHTMAPGEN_FXH__
