#ifndef __FILE_PBR_COMMON_FXH__
#define __FILE_PBR_COMMON_FXH__
#include "PBR_Definitions.fxh"
#include "Lighting.fxh"
#include "ShadingModel.fxh"
#include "ScreenSpaceReflection.fxh"

void VertexUnifyOutputCommon(in VertexAssemble IN, inout InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	InitVertexShadingOutput(OUT);

#if USE_UNIDIRECTIONAL_DISSOLVE
	if (@_UnidirectionalDissolveEnable)
	{
		OUT.ModelPositionY = IN.Position.y;
	}
#endif

	if (@_INSTANCED)
	{
		if (@_Billboard)
		{
			OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, float3(0, 0, 0)), 1);
		}
		else
		{
			OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
		}
	}
	else
	{
		OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	// TexCoord
	OUT.TexCoord = IN.TexCoord.xyxy;
#if IS_TERRAIN
	if (LIGHTMAP_ENABLE)
	{
		OUT.TexCoord.zw = IN.TexCoord.xy * cLightMapUVTransform.xy + cLightMapUVTransform.zw;
	}
#else
	if (HAS_2UV)
	{
		if (@_INSTANCED)
		{
#if USE_LIGHTMAP
			OUT.TexCoord.zw = IN.TexCoord2 * IN.LightMapTransform.xy + IN.LightMapTransform.zw;
#else
			OUT.TexCoord.zw = IN.TexCoord2.xy;
#endif
		}
		else
		{
// #if NO_LIGHT_MAP_UV_TRANSFORM
#if USE_LIGHTMAP
			OUT.TexCoord.zw = IN.TexCoord2 * cLightMapUVTransform.xy + cLightMapUVTransform.zw;
#else
			OUT.TexCoord.zw = IN.TexCoord2.xy;
#endif
		}
	}
#endif

	// Normal
	if (@_INSTANCED)
	{
		if (@_Billboard)
		{
			float3 dir = CameraPos.xyz - OUT.WorldPosition.xyz;
			dir.y = 0;	// project to xz plane.
			OUT.WorldNormal.xyz = normalize(dir);

			float scale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, float3(1, 0, 0)));
			OUT.WorldNormal.w = scale;
		}
		else
		{
			OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
			if (@_TANGENT)
			{
				OUT.WorldTangent.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalTangent));
				OUT.WorldBinormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalBinormal));
#if !IS_TERRAIN
				OUT.WorldTangent.w = 0;
				OUT.WorldBinormal.w = 0;
#endif
			}
		}
	}
	else
	{
		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
		if (@_TANGENT)
		{
			OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
			OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));
#if !IS_TERRAIN
				OUT.WorldTangent.w = 0;
				OUT.WorldBinormal.w = 0;
#endif
		}
	}

	// ViewDirection
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;

	// PerVertex Diffuse Lighting
	OUT.DiffLighting.xyz = half3(0, 0, 0);
	if (@_LOD_LEVEL < LOD_Level_3)
	{	
#if !IS_WATER		
		if(!@_DEFERRED_SHADING || @_TRANSPARENT)
#endif		
		{OUT.DiffLighting.xyz = LightingVS(OUT.WorldPosition.xyz, OUT.WorldNormal.xyz);}
	}

	// Fog T
	if (@_FOG && !@_VolumetricFog)
	{
#if !IS_WATER		
		if(!@_DEFERRED_SHADING || @_TRANSPARENT)
#endif
		{
			OUT.DiffLighting.w = FogTransmittance(OUT.ViewDirection.xyz);
		}
	}
#if USE_LIGHTMAP
	[@_INSTANCED]
	OUT.LightMapScale = IN.LightMapScale;
#endif

#if USE_LIGHTMAP_CONST
	[@_INSTANCED]
	OUT.LightMapTransform = IN.LightMapTransform;
#endif

	// DIY Output
	VS_DIY_OUTPUT(IN, INTM, OUT);

	if (@_VolumetricFog)
	{
		OUT.DiffLighting.w = OUT.HPosition.w;
	}

#if !IS_WATER
	[@_CastSunShadow]
	{
		OUT.LitUVZ = float4(MulMat4x4(ShadowViewProjTexs1, float4(OUT.WorldPosition.xyz + 
		OUT.WorldNormal.xyz * CSMShadowSlopBiases[1],1)).xyz,1);
		OUT.LitUVZ.z -= CSMShadowBiases[1];
	}
#endif
	[@_ES20] //es2 bug sample lightmap with .zw leads to mosaic
	OUT.TexCoord.xyzw = OUT.TexCoord.zwxy;

	//linear z
	OUT.LinearZ = OUT.HPosition.w * CameraInfo.z;
	OUT.CameraDistance = distance(CameraPos.xyz,OUT.WorldPosition.xyz);
}

// init MaterialParameters.GILighting and Inshadow
void FetchPrecomputedLighing(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	bool hasLightmap = false;
	MaterialParameters.GILighting = half3(0, 0, 0);
#if USE_LIGHTMAP
	if (@_LightMapEnable && !@_DEBUG)
	{
		hasLightmap = true;
		half3 lightMapScale;
		if (@_INSTANCED)
		{
			lightMapScale = IN.LightMapScale.xyz;
		}
		else
		{
#if NO_LIGHT_MAP_SCALE
			lightMapScale = half3(1,1,1);
#else
			lightMapScale = cLightMapScale.xyz;
#endif
		}
		half4 lightMapRaw = tLightMap.Sample(sLightMapSampler, LightMapUV(IN.TexCoord));
#if !IS_TERRAIN
		lightMapRaw.rgb = lightMapRaw.rgb * lightMapScale.xxx + lightMapScale.yyy;
#else
		lightMapRaw.rgb = lightMapRaw.rgb * cLightMapScale.rgb + cTintColor1.rgb;
#endif
		half LogL = dot(lightMapRaw.rgb, half3(0.3, 0.59, 0.11)) ;
		half Luma = exp2( LogL * 16 - 8 )-0.00390625;
		Luma = Luma ;
		MaterialParameters.CustomData = Luma ;
		MaterialParameters.InShadow = max(MaterialParameters.InShadow, 1 - lightMapRaw.a);
		MaterialParameters.GILighting = lightMapRaw.rgb * (MaterialParameters.AO * Luma / LogL);
		// MaterialParameters.GILighting *= 0.5;
	}
#if USE_LIGHTMAP_CONST
	else if (@_LightmapConstEnable && !@_DEBUG)
	{
		hasLightmap = true;
		if (@_INSTANCED)
		{
			MaterialParameters.GILighting = IN.LightMapTransform.xyz;
			MaterialParameters.InShadow = IN.LightMapTransform.w;
		}
		else
		{
			MaterialParameters.GILighting += cLightMapScale.xyz;
			MaterialParameters.InShadow = cLightMapScale.w;
		}
	}
#endif
#endif

	if (!hasLightmap)
	{
		DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);

#if MATERIAL_SHADINGMODEL_TWOSIDED_FOLIAGE
		half3 ScatterGI = max(half3(0.01,0.01,0.01), SunFogColor.rgb/max(1, SunFogColor.w));
		MaterialParameters.GILighting = lerp(MaterialParameters.GILighting, ScatterGI , saturate(MaterialParameters.Thickness));
#endif

	}
// #if MATERIAL_SHADINGMODEL_PREINTEGRATED_SKIN || MATERIAL_SHADINGMODEL_EYE
// 	half3 ScatterAO = pow(MaterialParameters.AO, half3(0.6,0.85,0.87));
// 	MaterialParameters.GILighting *=  ScatterAO * MaterialParameters.DiffuseColor;
// #else
	// if(@_SKELETON)
	// {
	// 	MaterialParameters.GILighting*=10;
	// }
	MaterialParameters.GILighting *=  MaterialParameters.AO;
// #endif

}

/*half3 IBL_Specular(in MaterialShadingParameters MaterialParameters, in half3 SpecularColor, in half3 R, in half NoV)
{
	half3 OUT = half3(0, 0, 0);
	half3 sampleEnvSpecular = half3(0,0,0);
	half MIP_ROUGHNESS = 0.17;
	half level = MaterialParameters.Roughness / MIP_ROUGHNESS;
	half fSign = R.z > 0;
	half fSign2 = fSign*2 - 1;
	R.xy /= R.z*fSign2 + 1;
	R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
	{
		half4 srcColor;
		srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy, level*RenderToMip);
		sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE);
	}
	sampleEnvSpecular *= MaterialParameters.AO;
	if (@_LOD_LEVEL <= LOD_Level_1)
		OUT = EnvBRDFApprox(SpecularColor, MaterialParameters.Roughness, NoV) * sampleEnvSpecular;
	else
		OUT = SpecularColor * sampleEnvSpecular;

#if MATERIAL_SHADINGMODEL_CLEAR_COAT
	half ClearCoat = MaterialParameters.CustomData.x;
	half ClearCoatRoughness = MaterialParameters.CustomData.y;
	float VoH = NoV;
	float F = F_SchlickGray(0.04, VoH) * ClearCoat;
	OUT *= 1-F;

	level = ClearCoatRoughness / MIP_ROUGHNESS;
	half4 srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy, level*RenderToMip);
	sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE);
	sampleEnvSpecular *= MaterialParameters.AO;
	OUT += F * sampleEnvSpecular;
#elif MATERIAL_SHADINGMODEL_HAIR
	OUT = EnvBRDFApprox(SpecularColor, 0.8, NoV) * sampleEnvSpecular;
#endif
	return OUT;
}

half3 IBL_SpecularRTM(in MaterialShadingParameters MaterialParameters, in VertexShadingOutput IN, in half3 SpecularColor, in half3 R, in half NoV)
{
	float2 spos = IN.HPosition.xy;
	if(@_D3D9) {spos += 0.5;}
	spos += MaterialParameters.WorldNormal.xz * 5;
	float2 stc = spos * ScreenInfo.zw;
	stc.x = 1-stc.x;
	half3 sampleEnvSpecular = tReflectionMap.SampleLevel(sReflectionSampler, stc, MaterialParameters.Roughness * MaterialParameters.Roughness * 10).xyz;
	sampleEnvSpecular *= MaterialParameters.AO;
	if (@_LOD_LEVEL <= LOD_Level_1)
		return EnvBRDFApprox(SpecularColor, MaterialParameters.Roughness, NoV) * sampleEnvSpecular;
	else
		return SpecularColor * sampleEnvSpecular;
}*/

#define RAIN_EFFECT(DiffuseColor, SpecularColor, Roughness, N)\
{\
	if (@_Weather == WEATHER_RAIN)\
	{\
		if (!LIGHTMAP_ENABLE)\
		{\
			DiffuseColor *= 1 - 0.15*RainRate;\
			Roughness *= 1 - 0.45*RainRate;\
		}\
		else\
		{\
			half DryRate = 1 - RainRate;\
			SpecularColor *= 1-0.9*Roughness;\
			DiffuseColor *= max(0.95,DryRate);\
			Roughness *= DryRate;\
		}\
	}\
}

#if MATERIAL_SHADINGMODEL_TWOSIDED_FOLIAGE
#define SNOW_OVER(BaseColor, N, Metallic, Roughness,IN)\
{\
	if (@_Weather == WEATHER_SNOW)\
	{\
		half overlayFactor = 0.6 * SnowRate;\
		BaseColor = lerp(BaseColor, half3(1,1,1), overlayFactor);\
		Metallic *= 1 - overlayFactor;\
		Roughness = lerp(Roughness, 1.0, overlayFactor);\
	}\
}
#else
#define SNOW_OVER(BaseColor, N, Metallic, Roughness,IN)\
{\
	if (@_Weather == WEATHER_SNOW && !IS_TERRAIN && LIGHTMAP_ENABLE)\
	{\
		half overlayFactor = saturate((N.y+0.2) * SnowRate) * SNOW_HEIGHT_MODIFIER;\
		BaseColor = lerp(BaseColor, half3(1,1,1), overlayFactor);\
		Metallic *= 1 - overlayFactor;\
		Roughness = lerp(Roughness, 1.0, overlayFactor);\
	}\
}
#endif

void ModifyReflectionDirection(in VertexShadingOutput IN, in out half3 R)
{
	R = R;
	//parallax fix, useful for low roughness reflection
#if !IS_TERRAIN
	#if IS_WATER
		if (ENABLE_IBL && @_EnableReflection)
			R = normalize(IN.WorldPosition.xyz + R * ReflectionPos.w - ReflectionPos.xyz);
	#else
		if (ENABLE_IBL && @_EnableReflection && LIGHTMAP_ENABLE)
			R = normalize(IN.WorldPosition.xyz + R * ReflectionPos.w - ReflectionPos.xyz);
	#endif
#endif
}

void GetHexShadow(float3 LitUVZ, int layerIndex, out half InDynamicShadow)
{
	float inShadow;
	float2 pInLitTC = LitUVZ.xy;
	float litZ = LitUVZ.z;
	half2 inRange = abs(pInLitTC - 0.5);
	half ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
	float2 vInvShadowMapWH = float2(SHADOW_PIXELSIZE,SHADOW_PIXELSIZE);
	if(!@_CSM)
	{
		if(@_LOD_LEVEL <= LOD_Level_1)
		{
			hardware_pcf_4x4(pInLitTC, litZ, vInvShadowMapWH, inShadow);
		}else
		{
			hardware_pcf_2x2(pInLitTC, litZ, vInvShadowMapWH, inShadow);
		}
	}
	else if(layerIndex == 0)
	{
		if(@_LOD_LEVEL <= LOD_Level_1)
		{
	#if MATERIAL_SHADINGMODEL_HAIR
			inShadow = hardware_pcf_array_4x4(pInLitTC, litZ, vInvShadowMapWH, 0);
	#else
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 0);
	#endif
		}
		else
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 0);
		}
	}else if(layerIndex == 1)
	{
		float tempShadow;
		if(@_LOD_LEVEL <= SD_SHADOW_HIGHQUALITY_LEVEL)
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 1);
			tempShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
		}
		else
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 1);
			tempShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
		}
		inShadow = max(inShadow, tempShadow);
	}else{
		inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
	}
	inShadow *= saturate((ShadowRate*10));
	InDynamicShadow = inShadow;
	
}

half GetDynamicShadowForSun(in VertexShadingOutput IN)
{
	half InShadow = 0;
	//get dynamic shadow
	if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		InShadow = tShadowMask.Sample(sShadowMaskSampler, tc).r;
	}
	else if (@_CastSunShadow && (@_REC_SHADOW || @_DEFERRED_SHADING || USE_SHADOWMAP_SAMPLE) && @_LOD_LEVEL <= LOD_Level_2)
	{
		float3 LitUVZ = IN.LitUVZ.xyz / IN.LitUVZ.w;
		LitUVZ.z = min(1.0-10*HALF_EPSILON, LitUVZ.z);
		GetHexShadow(LitUVZ, 1, InShadow);
	}
	InShadow = min(1.0, InShadow);
	if(!LIGHTMAP_ENABLE)
		InShadow *= SHADOW_SELFRATE;
	else
		InShadow *= SHADOW_RATE;
	return InShadow;
}

void InitRainOcclusion(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	half occlusionBias = 0.5;
	float2 shelterUV = 1 - (IN.WorldPosition.xz - SheltermapMinPos.xz) * SheltermapMaxPos.xz;
	float sceneDepth = tShelterMap.Sample(ShelterMap_Sampler, shelterUV).r;
	float sceneHeight = lerp(SheltermapMaxPos.y, SheltermapMinPos.y, sceneDepth);
	MaterialParameters.RainOcclusion = 1 - saturate(sceneHeight - IN.WorldPosition.y + occlusionBias);
	half2 mask = length(shelterUV - half2(0.5,0.5));
	mask = saturate(mask * 2 - 1);
	MaterialParameters.RainOcclusion = lerp(MaterialParameters.RainOcclusion, 1, mask);
}

void InitBaseColorCommon(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	if(@_BaseMapEnable)
	{
		half4 samplerBase;
		samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = cAlpha;
	}
}

void InitSmoothnessMetalAO(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	if (@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).rgb;
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_0)
		{
			MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else	
	{
		MaterialParameters.Metallic = cMetallic;
		if(ENABLE_MIXMAP || @_DEFERRED_SHADING)
			MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
}
//rain
float Hash( in float2 p ) 
{
	return frac(sin(dot(p, float2(127.1,311.7)))*43758.5453123);
}
float Noise( in float2 p ) 
{
	float2 i = floor( p );
	float2 f = frac( p );	
	float2 u = f*f*(3.0-2.0*f);
	return -1.0+2.0*lerp( lerp( Hash( i + float2(0.0,0.0) ), Hash( i + float2(1.0,0.0) ), u.x),lerp( Hash( i + float2(0.0,1.0) ), Hash( i + float2(1.0,1.0) ), u.x), u.y);
}
float SeaOctave(float2 uv) 
{
    uv += Noise(uv);        
    float2 wv = 1.0-abs(sin(uv));
    float2 swv = abs(cos(uv));    
    wv = lerp(wv,swv,wv);
    return 1.0-pow(wv.x * wv.y,0.65);
}
void RainFlowNormal(in out MaterialShadingParameters MaterialParameters, in float2 uv)
{
	if(MaterialParameters.RainOcclusion > 0.01)
	{
		half3 N = MaterialParameters.WorldNormal.xyz;
		float4 jitterUV;
		half worldscale = 1;
		jitterUV = uv.xyxy * float4(1.5, 5, 5, 1.5) * worldscale;
		float4 seed = clamp(N.xzxz * 10000, -1, 1) * float4(20, 20, 6, 6) * GameTime;
		float R1 = SeaOctave(jitterUV.yx * 10 - seed.x) + SeaOctave(float2(jitterUV.z*3-seed.z, jitterUV.w*3));
		float R3 = SeaOctave(float2(jitterUV.xy * 4 - seed.w)) + SeaOctave(jitterUV.zw * 8 -  seed.y);
		R3 *= 0.5;
		float R_D = (R1 * N.x * N.x + R3 * N.z * N.z) * 5+(R1 + R3) * 0.1 - 0.212;

		R_D *= step(0.5, RainRate) * RainRate * 1.3;
		MaterialParameters.WorldNormal.xyz = normalize(lerp(N + float3(0, 0, R_D), N, 1 - 0.2 * saturate(N.y)));
	}
}
void Ripple(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters, in out half3 tangentNormal, in half height)
{
	half YIntensity = pow(saturate(IN.WorldNormal.y), 155);
	
	//////////////Suo'S rain Ripple///////////////
	half4 raintexture =  tRainMap.Sample(sRainSampler, IN.WorldPosition.xz*0.3 + frac(GameTime*0.005));
	half timeTricker = (frac(GameTime*1.5+ raintexture.y*3)-1.2);
	half waveMask = sin(clamp(timeTricker+raintexture.x,0,3)*8 );
	half waveSmall =  pow(waveMask,5)*timeTricker * timeTricker.x ;
	half3 waveNormal = half3(raintexture.z,raintexture.w,1) ;
	waveNormal.xy*=waveSmall;
	waveNormal*=2-1;
	tangentNormal =BlendAngleCorrectedNormal(tangentNormal,waveNormal*RainRate);
}

void InitWorldNormal(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	if(@_TANGENT && @_NormalMapEnable)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		[@_DIRECTX]//always use bc5 for pc
		tangentNormal.z = 1;
		tangentNormal = (2 * tangentNormal) - 1;
		if (ENABLE_RIPPLE)
			Ripple(IN, MaterialParameters, tangentNormal, 1);
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		//tangentNormal.xy *= cBumpIns;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
		if (ENABLE_RIPPLE)
			RainFlowNormal(MaterialParameters, IN.WorldPosition.xz);
		//MaterialParameters.BaseColor = Noise(IN.WorldPosition.xz).xxx;
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
}

void InitEmission(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	if (@_EmissionMapEnable )//&& @_LOD_LEVEL <= LOD_Level_2
	{
		half2 EmissionUV = half2(IN.WorldNormal.w, IN.ViewDirection.w);
		MaterialParameters.Emission = tEmissionMap.Sample(sEmissionMapSampler, EmissionUV).rgb;
		MaterialParameters.Emission.rgb *= cEmissionScale.rgb * cEmissionScale.a;
		if(@_HDR){MaterialParameters.Emission.rgb *= MaterialParameters.BaseColor;}
	}
}
void InitWeaponHighlight(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
		if(@_WeaponHighLight)
	{
		float2 spos =(IN.HPosition.xy +0.5)*ScreenInfo.zw*200;
		float screenRoll =saturate( sin(spos*0.01+ GameTime*2))  ;  
		half3 N = normalize(IN.WorldNormal);
		half3 V = normalize(-IN.ViewDirection);
		float fresnel =1-saturate(dot(N,V));
		fresnel*=fresnel;
		MaterialParameters.Emission +=fresnel* cWeaponHighLightColor * screenRoll ;
	
	}
}

void InitWeaponPreview(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
		if(@_WeaponPreview)
	{	
		float2 spos =(IN.HPosition.xy +0.5)*ScreenInfo.zw;
		float Roll  = GameTime*2;
		float screenRoll =saturate( sin(spos+Roll))  ; 		
		MaterialParameters.Emission += half3(0.03,0.03,0.03) * cWeaponHighLightColor*screenRoll;
		
	}
	
}

void InitGrayEffect(half cGrayingin,in out MaterialShadingParameters MaterialParameters)
{		
		if(@_GrayEffect)
	{	 
		//cGrayingin = 0.8;
		MaterialParameters.BaseColor = MaterialParameters.BaseColor * (1.0 - cGrayingin);
		MaterialParameters.Specular = MaterialParameters.Specular * (1 - cGrayingin);
	}
}

//GlassReflectEffect
void InitReflectEffect(in VertexShadingOutput IN,in out MaterialShadingParameters MaterialParameters)
{
	if(@_LOD_LEVEL <= LOD_Level_2 && @_ReceiveCubeReflect)
	{
		
		half3 sampleEnvSpecular = half3(0,0,0);
		half MIP_ROUGHNESS = 0.45;
		half3 R = reflect( -IN.ViewDirection.xyz,IN.WorldNormal.xyz);
		half level = MaterialParameters.Roughness / MIP_ROUGHNESS;
		half fSign = R.z > 0;
		half fSign2 = fSign*2 - 1;
		R.xy /= (R.z+fSign2) * fSign2;
		R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
		R.xy += cReflectionOffset;
		sampleEnvSpecular = RGBMDecode( tReflectMap.SampleLevel( sReflectMapSampler, R.xy, level ), 16.0);
		sampleEnvSpecular *= cReflection;
		MaterialParameters.BaseColor += sampleEnvSpecular;
	}
}
void InitRadarEffect(in VertexShadingOutput IN,in out MaterialShadingParameters MaterialParameters)
{
	if(@_RadarEnable)
	{
	half3 N = normalize(IN.WorldNormal);
	half3 V = normalize(-IN.ViewDirection);
	float fresnel =1-saturate(dot(N,V)) ;
	fresnel*=fresnel;	
	// float camdistanceScale = distance(CameraPos.xyz,IN.WorldPosition.xyz)*0.007;
	float camdistanceScale = IN.CameraDistance*0.007;
	float3 povit = PlayerPos.xyz;
	float cosAngle = cos(cPlayerDirection);
 	float sinAngle = sin(cPlayerDirection);
	float2x2 rot = float2x2(cosAngle,-sinAngle,sinAngle,cosAngle);
	float2 pos = IN.WorldPosition.xz - povit.xz;
	pos = mul(rot,pos) ;
	half3 radaColor = cRadarColor*15;
	float time = -cRadarScrolling;
	// float time = 0;
	half redaline =  frac(saturate(pos.y*0.03+time+1.2) ) ;
	redaline*=redaline;
	redaline*=redaline;
	half redalineSlim =  ceil((sin(pos.y *floor((1/camdistanceScale*2)))*0.5+0.5)-0.99); 
	radaColor = radaColor*(redaline*redalineSlim+redaline*fresnel*0.2);
	
	MaterialParameters.Emission +=radaColor*saturate(1-camdistanceScale*1);
	}
}

		
void InitAnisotropic(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	if (ENABLE_ANISOTROPIC && @_TANGENT && @_LOD_LEVEL <= LOD_Level_1)
	{
		half RoughnessX = cRoughnessX * MaterialParameters.Roughness + HALF_EPSILON;
		half RoughnessY = cRoughnessY * MaterialParameters.Roughness + HALF_EPSILON;
		half3 L = SunDirection.xyz;
		half NoL = max(0.001, dot(MaterialParameters.WorldNormal, L));
		half3 inputRadiance = SunColor * NoL;
		half3 V = normalize( -IN.ViewDirection.xyz );
		float3 LPlusV = L + V;
		half len = max(0.001, length(LPlusV));
		half3 H = LPlusV / len;
		half VoN = max(0.001, dot(V, MaterialParameters.WorldNormal));
		half LoN = NoL;
		half HoN = dot(H, MaterialParameters.WorldNormal);
		half HoT = dot(H, IN.WorldTangent.xyz);
		half HoB = dot(H, IN.WorldBinormal.xyz);
		half beta_a = HoT / RoughnessX;
		beta_a *= beta_a;
		half beta_b = HoB / RoughnessY;
		beta_b *= beta_b;
		half beta = -2.0 * (beta_a + beta_b) / max(0.1, (1.0 + HoN)); // danteng
		half s_den = max(0.001, 12.5663706 * RoughnessX * RoughnessY * sqrt(LoN * VoN));
		MaterialParameters.Emission.rgb += IN.TintColor.rgb * cAnisotropicScale * lerp(half3(0.04,0.04,0.04), MaterialParameters.BaseColor, MaterialParameters.Metallic.xxx) * inputRadiance * exp(beta) / s_den;
	}
}

void InitRain(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	MaterialParameters.BaseColor *= 1 - (0.3 * RainRate * (1 - MaterialParameters.Metallic) * MaterialParameters.RainOcclusion);
	MaterialParameters.Roughness *= 1 - (0.85 * RainRate * MaterialParameters.RainOcclusion);
}

void InitBurnt(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters)
{
	half BurnRadius = GlobalBurnParam.w;
	half3 BurnPosition = GlobalBurnParam.xyz;
	// half BurnRadius = 40;
	// half3 BurnPosition = half3(1000,-210,1000);
	if (BurnRadius>0.01)
	{
		half BurnWidth = 20;
		half GrayColor = GetLuminance(MaterialParameters.BaseColor);
		half3 NoisePos = normalize(IN.WorldPosition - BurnPosition ) * Noise(IN.WorldPosition.xz * 0.1) * 10 + IN.WorldPosition;
		half BurnArea = saturate((length(NoisePos-BurnPosition) - BurnRadius) / BurnWidth);
		half OcclusionMask = lerp(0 , 0.35 , saturate((1-MaterialParameters.AO) * 5.0 ));
		half MaterialMaskOut = GrayColor>0.11;
		half MaterialMaskIn = GrayColor>0.05;
		half BurnMask = max(lerp(0.3 , 1 , MaterialMaskOut) * MaterialMaskIn , OcclusionMask);
		MaterialParameters.BaseColor = lerp(lerp(0.01,lerp(MaterialParameters.BaseColor,GrayColor,0.65)*0.7,BurnMask), MaterialParameters.BaseColor,BurnArea);
		MaterialParameters.Roughness = lerp(MaterialParameters.Roughness , 0.55 , max((1 - MaterialMaskIn) , BurnArea));
	}
}

void InitBurntSimple(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters, half Brightness)
{
	half BurnRadius = GlobalBurnParam.w;
	half3 BurnPosition = GlobalBurnParam.xyz;
	// half BurnRadius = 40;
	// half3 BurnPosition = half3(1000,-210,1000);
	if (BurnRadius>0.01)
	{
		half BurnWidth = 20;
		half GrayColor = GetLuminance(MaterialParameters.BaseColor);
		half3 NoisePos = normalize(IN.WorldPosition - BurnPosition ) * Noise(IN.WorldPosition.xz * 0.1) * 10 + IN.WorldPosition;
		half BurnArea = saturate((length(NoisePos-BurnPosition) - BurnRadius) / BurnWidth);
		MaterialParameters.BaseColor = lerp(lerp(GrayColor,MaterialParameters.BaseColor,0.1) * Brightness , MaterialParameters.BaseColor,BurnArea);
	}
}

half2 InitSunLighting(in VertexShadingOutput IN, in out GBufferData GBuffer, in out LightingResult LitResult, half3 V, half NoV)
{
	int PBRQualityLevel = PBRQualityLevel_None;
	if (ENABLE_MIXMAP)
		PBRQualityLevel = PBRQualityLevel_Normal;
	LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);

	half GTAO = 1;
	if(@_TRANSPARENT)
	{
		LitData.InShadow = GetDynamicShadowForSun(IN);
	}else
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		float2 shadowAndAO = tShadowMap.Sample(sShadowMapSampler, tc).rg;
		LitData.InShadow = shadowAndAO.x;
		GTAO = shadowAndAO.y;
	}
	
	IntegrateBxDF(LitResult, GBuffer, LitData);
	return half2(LitData.InShadow, GTAO);
}

void InitPointLighting(in VertexShadingOutput IN, in out GBufferData GBuffer, in out LightingResult LitResult, half3 V, half NoV)
{
	LitResult.DiffuseLighting += IN.DiffLighting.rgb;

	//point lighting
	// if(@_LOD_LEVEL <= LOD_Level_2)
	// {	
		if( @_LitCount >= 1 ) // && (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) >= 0
		{
			LightingData LitData;
			LitData = GetSpotLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_Normal, Lights0);
			if(Lights0[2].w > 0)
			{
				LitData	= GetPointLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_Normal, Lights0);
			}
			else
			{
			 	LitData = GetSpotLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_Normal, Lights0);
			}
			IntegrateBxDF(LitResult, GBuffer, LitData);
		}
		if( @_LitCount >= 2 ) //&& (SD_MAX_PS_LIGHTS-@_LOD_LEVEL) >= 1
		{	
			LightingData LitData;
			LitData = GetSpotLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_Normal, Lights1);
			if(Lights1[2].w > 0)
			{
				LitData = GetPointLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_Normal, Lights1);
			}
			else
			{
				LitData = GetSpotLightData(IN.WorldPosition.xyz, GBuffer, V, 0, PBRQualityLevel_Normal, Lights1);
			}
			IntegrateBxDF(LitResult, GBuffer, LitData);
		 }
	//}
}

void InitVirtualLight(in out GBufferData GBuffer, in half NoV, in float3 V)
{
	if (@_SKELETON)
	{
		half3 virtualLitSpec = half3(0,0,0);
		half3 virtualDirect =normalize(half3(0,1.5,0.0)-V) ;
		half dirUp =saturate( dot (virtualDirect,GBuffer.WorldNormal));
		if(@_LOD_LEVEL <=1)
		{
		
		 virtualLitSpec = SimpleBRDF_SPEC(virtualDirect, GBuffer.WorldNormal, virtualDirect, dirUp,dirUp, GBuffer.SpecularColor, GBuffer.Roughness) * cVirtualLitColor.rgb ;
#if MATERIAL_SHADINGMODEL_CLOTH
		virtualLitSpec *= (1 - GBuffer.CustomData.a);
#endif
		}
		half3 virtualLitDiff = cVirtualLitColor.rgb *dirUp * GBuffer.DiffuseColor ;
		GBuffer.Emission += virtualLitDiff;
		// half3 virtualLitDiff = cVirtualLitColor.rgb * NoV * GBuffer.DiffuseColor;
		// GBuffer.Emission += virtualLitDiff;
	}
}
// void InitVirtualLight(in out GBufferData GBuffer, in half NoV, in float3 V)
// {
//     if (@_SKELETON)
//     {
//         half3 virtualLitSpec = half3(0,0,0);
//         if(@_LOD_LEVEL <=1)
//         {
//          virtualLitSpec = SimpleBRDF_SPEC(V, GBuffer.WorldNormal, V,NoV,NoV, GBuffer.SpecularColor, GBuffer.Roughness) * cVirtualLitColor.rgb ;
// #if MATERIAL_SHADINGMODEL_CLOTH
//         virtualLitSpec *= (1 - GBuffer.CustomData.a);
// #endif
//         }
//         half3 virtualLitDiff = cVirtualLitColor.rgb * NoV * GBuffer.DiffuseColor ;
//         GBuffer.Emission += virtualLitDiff;
//         // half3 virtualLitDiff = cVirtualLitColor.rgb * NoV * GBuffer.DiffuseColor;
//         // GBuffer.Emission += virtualLitDiff;
//     }
// }



/*void InitIBLReflection(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters, half3 V, half NoV)
{
	half3 R = reflect(-V, MaterialParameters.WorldNormal);
	ModifyReflectionDirection(IN, R);
	if(ENABLE_MIXMAP)
	{
		half3 IBLSpec = half3(0, 0, 0);
#if RTM_REFLECTION
		if (ENABLE_IBL_REF)//[diff with deferred]deferred always calc ibl reflection
			IBLSpec = IBL_SpecularRTM(MaterialParameters, IN, MaterialParameters.SpecularColor, R, NoV);
#else
		if (ENABLE_IBL)
			IBLSpec = IBL_Specular(MaterialParameters, MaterialParameters.SpecularColor, R, NoV);
#endif
		MaterialParameters.GILighting.rgb += IBLSpec * ReflectionScale * cEnvStrength * MaterialParameters.AO;
	}
}*/

/*void InitReflectionSSR(in VertexShadingOutput IN, in out MaterialShadingParameters MaterialParameters, half3 V, half NoV)
{
	half3 R = reflect(-V, MaterialParameters.WorldNormal);
	ModifyReflectionDirection(IN, R);
	half3 SPEC = half3(0, 0, 0);
	if (ENABLE_MIXMAP && ENABLE_IBL)
	{
		half3 SSRColor = half3(0, 0, 0);
		float3 SSR = float3(0, 0, 0);
		SSR = ScreenSpaceReflection(IN.WorldPosition.xyz, IN.HPosition.xy, R, NoV);
		if(SSR.z > 0 && MaterialParameters.Roughness < 0.5)
		{
			SSRColor = ScreenTexture.Sample(ScreenTexture_Sampler, SSR.xy);
			// if(@_ENCODEHDR)
			// 	SSRColor *= SSRColor * SD_SCENE_MAX_LUMINANCE;
		}
		half3 sampleEnvSpecular = half3(0,0,0);
		half MIP_ROUGHNESS = 0.17;
		half level = MaterialParameters.Roughness / MIP_ROUGHNESS;
		half fSign = R.z > 0;
		half fSign2 = fSign*2 - 1;
		R.xy /= R.z*fSign2 + 1;
		R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
		{
			half4 srcColor;
			srcColor = tEnvMap.SampleLevel( sEnvSampler, R.xy, level * RenderToMip );
			sampleEnvSpecular = RGBMDecode( srcColor, MAX_PBR_RGBM_VALUE );
		}
		sampleEnvSpecular *= cEnvStrength * ReflectionScale * MaterialParameters.AO;
		if (@_LOD_LEVEL <= LOD_Level_1)
		{
			SPEC = lerp(sampleEnvSpecular, SSRColor, SSR.z);
			SPEC *= EnvBRDFApprox(MaterialParameters.SpecularColor, MaterialParameters.Roughness, NoV);
		}
		else
			SPEC = MaterialParameters.SpecularColor * sampleEnvSpecular;
	}
	MaterialParameters.GILighting += SPEC;
}*/

void InitFogLighting(in VertexShadingOutput IN, in LightingData LitData, in out float4 OUT)
{
	if (@_FOG && @_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.DiffLighting.w;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
	}
	else if (@_FOG && !@_VolumetricFog)
	{
		float3 fogFinalColor = float3(0,0,0);
		FogLightingAndScatter(IN.DiffLighting.w, fogFinalColor, LitData.V, LitData.L);
		OUT.rgb = lerp(fogFinalColor, OUT.rgb, IN.DiffLighting.w);
	}
}
// depend cBloomScale/
void InitBloomScale(in LightingResult LitResult, in MaterialShadingParameters MaterialParameters, in out float4 OUT)
{
	// half bloomScale = 0;
	// if(@_LOD_LEVEL <= LOD_Level_2 && !@_DEFERRED_SHADING)
	// 	bloomScale = cBloomScale * dot(LitResult.SpecularLighting + MaterialParameters.Emission, half3(0.6378, 2.1456, 1));
	if (@_TRANSPARENT)
	{
		if (@_HDR)
			OUT.rgb = OUT.rgb * OUT.a;
		else
			OUT.rgb = saturate(OUT.rgb) * OUT.a;
		// OUT.a = EncodeTransparentBloomFactor(OUT.a, bloomScale);
	}
// 	else
// 	{
// #if USE_LIGHTMAP
// 		if (LIGHTMAP_ENABLE)
// 		{
// 			OUT.a = EncodeSolidBloomFactor(bloomScale);
// 		}
// 		else
// #endif
// 		{
// 			OUT.a = EncodeSolidBloomFactor(bloomScale + SunColor.w);
// 		}
// 	}
}
void InitScreenGray(in VertexShadingOutput IN, in out half3 BaseColor)
{
	if(ScreenMGrayModeSpread)
	{
		half grayDist = (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - min(length(IN.WorldPosition.xyz - PlayerPos.xyz), ScreenMGrayMaxDistance);
		grayDist = step(0, grayDist);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);		
	}
	else if (ScreenMGrayModeShrink)
	{
		half grayDist = length(IN.WorldPosition.xyz - PlayerPos.xyz) + (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - ScreenMGrayMaxDistance;
		grayDist = step(grayDist, 0);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);
	}
}
#endif // __FILE_PBR_COMMON_FXH__
