//////////////////////////////
////////////////////////////// notice will not be use this
//////////////////////////////
//////////////////////////////
#ifndef __FILE_PBR_FXH__
#define __FILE_PBR_FXH__

#include "ShadingModel.fxh"

#ifndef OPACITY_MASK_VALUE
#define OPACITY_MASK_VALUE(IN,ALPHA) {\
	ALPHA = tBaseMap.SampleRGBA( sBaseSampler, IN.TexCoord ).a;\
}
#endif


#ifndef ENABLE_SCREEN_SHADOWMAP
	#define ENABLE_SCREEN_SHADOWMAP false
#endif
#ifndef ENABLE_HIGHLIGHT
	#define ENABLE_HIGHLIGHT false
#endif
#ifndef ENABLE_BURNT
	#define ENABLE_BURNT false
#endif

#include "Lighting.fxh"
#include "PBR_Common.fxh"


//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput PbrVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);
	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif

	//OUTPUT
	VertexUnifyOutputCommon(IN, INTM, OUT);

	return OUT;
}

half GetShelterRate(in VertexShadingOutput IN)
{
	if(SAMPLE_SHELTERMAP)
	{
		float2 uv = 1 - (IN.WorldPosition.xz - cSheltermapMinPos.xz) * cSheltermapMaxPos.xz;
		if (@_GLSL)
			uv.y = 1 - uv.y;
		float depth = tShelterMap.Sample(sShelterSampler, uv).r;
		// feel free to change the code below
		if(depth > 0.99)
			return 1;
		uv = abs(uv - 0.5);
		half rate = min(0.5-uv.x, 0.5-uv.y);
		rate = saturate(rate * 10);
		depth = lerp(cSheltermapMaxPos.y, cSheltermapMinPos.y, depth);
		float curDepth = IN.WorldPosition.y + IN.WorldNormal.y*0.02 + 0.02; // add bias
		half res = saturate(1 + (curDepth - depth));
		return lerp(1, res, rate);
	}
	else
	{
		return 1;
	}
}


float4 PbrPixelShading(in VertexShadingOutput IN, in bool FrontFace : SV_IsFrontFace) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	//1, Gen MaterialParameters & GBufferData
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	PS_GetParameters(IN, MaterialParameters);
	OUT.a = MaterialParameters.Alpha;
#if USE_UNIDIRECTIONAL_DISSOLVE
	if(@_UnidirectionalDissolveEnable && @_DISSOLVING){
		float ratio = min((GameTime - cParamter2.y) /  (abs(cParamter2.z) + EPSILON), 1);
		ratio = abs(cParamter2.w - ratio);
		float clipRatio = cParamter2.z > 0 ? (1 - ratio - IN.ModelPositionY/2 ) : (IN.ModelPositionY/2 - ratio );
		OUT.a = (clipRatio < 0) ? 0 : OUT.a;
	}
#endif
#if USE_ONE_SIDE_LIGHTING
	if(!@_OneSideLightingEnable)
#endif	
	{
		if(!FrontFace)
		MaterialParameters.WorldNormal = -MaterialParameters.WorldNormal;
	}

	if (@_TRANSPARENT && @_DISSOLVING)
	{
		OUT.a *= cParamter2.x;
	 }	
	if(@_LOD_LEVEL> LOD_Level_2)
	{

		MaterialParameters.Metallic = 0;
	}
	//apply weather
	SNOW_OVER(MaterialParameters.BaseColor, MaterialParameters.WorldNormal, MaterialParameters.Metallic, MaterialParameters.Roughness, IN);
	//RAIN_EFFECT(MaterialParameters.DiffuseColor, MaterialParameters.SpecularColor, MaterialParameters.Roughness, MaterialParameters.WorldNormal);//move to GetParameter function

	//screen gray
	InitScreenGray(IN, MaterialParameters.BaseColor);
	MaterialParameters.SpecularColor = lerp(0.08 * MaterialParameters.Specular.xxx, MaterialParameters.BaseColor, MaterialParameters.Metallic);
	MaterialParameters.DiffuseColor = Diffuse_Lambert(MaterialParameters.BaseColor - MaterialParameters.BaseColor * MaterialParameters.Metallic);
	//ShadowAndAO
	half GTAO = 1;
	if(USE_SHADOWMAP_SAMPLE)
	{
		MaterialParameters.InShadow = GetDynamicShadowForSun(IN);
#if USE_PLAYER_SHADOWMAP
		if(@_TRANSPARENT && @_CSM)
		{
			float3 LitUVZ2 = IN.LitUVZ2.xyz / IN.LitUVZ2.w;
			LitUVZ2.z = min(1.0-10*HALF_EPSILON, LitUVZ2.z);
			half calcShadow = 0;
			GetHexShadow(LitUVZ2, 0, calcShadow);
			MaterialParameters.InShadow = max(calcShadow, MaterialParameters.InShadow);
		}
#endif
	}else if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		float2 shadowAndAO = tShadowMask.Sample(sShadowMaskSampler, tc).rg;
		MaterialParameters.InShadow = max(shadowAndAO.x, MaterialParameters.InShadow);
		// GTAO = shadowAndAO.y;
	}

	//2, lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	int PBRQualityLevel = PBRQualityLevel_None;
	if (@_LOD_LEVEL <= LOD_Level_2 && @_EnableReflection || IS_TERRAIN)
		PBRQualityLevel = PBRQualityLevel_Normal;
	half3 V = normalize(-IN.ViewDirection.xyz);
	half NoV = saturate(dot(MaterialParameters.WorldNormal, V));
#if MATERIAL_SHADINGMODEL_PREINTEGRATED_SKIN
	half3 R = reflect(-V, MaterialParameters.CustomData.xyz);
#else
	half3 R = reflect(-V, MaterialParameters.WorldNormal);
#endif
	ModifyReflectionDirection(IN, R);
	//GI lighting
	FetchPrecomputedLighing(IN, MaterialParameters);

	//InitIBLReflection(IN, MaterialParameters, V, NoV);
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	//GI Specular lighting
	half3 IBLSpec = half3(0,0,0);
	if (ENABLE_MIXMAP && ENABLE_IBL)
		IBLSpec = IBL_Specular(GBuffer, R, NoV);
#if USE_LIGHTMAP
		if(@_LightMapEnable)
		{
			IBLSpec*= MaterialParameters.CustomData ;
		}
#endif	
	//sun lighting
	LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);
	IntegrateBxDF(LitResult, GBuffer, LitData);	
	//thunder shading
	LitResult.SpecularLighting *= ThunderRate;
	LitResult.DiffuseLighting *= ThunderRate;
	//point lighting calc include vs lighting and ps lighting
	// if (@_SKELETON)
	// {
		InitPointLighting(IN, GBuffer, LitResult, V, NoV);
	//}
	
	//virtual lit calc
#if VIRTUAL_LIT
	InitVirtualLight(GBuffer, NoV, V);
	if(@_SKELETON)
	{
		// half  distanceScale =clamp(IN.CameraDistance*LITMAP_GISCALE*0.03,1,5*cBloomScale) ;
		half distance =  IN.CameraDistance/cBloomScale;
	
		half  distanceScale = lerp(1, max(cEnvStrength,1) ,  distance) ;
	
	 GBuffer.Emission *=distanceScale;
	}
	 
#endif
	//accumulate specular and diffuse
	if(!@_HDR)
	{
	LitResult.SpecularLighting*=2;	
	 IBLSpec*=2;
	 LitResult.DiffuseLighting*=2;
	 GBuffer.GILighting*=2;
	}
	OUT.rgb += GBuffer.Emission;//emission without GI lighting
	OUT.rgb += LitResult.SpecularLighting + IBLSpec * GTAO;
	OUT.rgb += (LitResult.DiffuseLighting + GBuffer.GILighting * GTAO) * GBuffer.DiffuseColor;

	//3, Fog shading.
	InitFogLighting(IN, LitData, OUT);

	//4, screen color tint
#if USE_LIGHTMAP
	if (@_LIGHT_MAP || LIGHTMAP_ENABLE)
		OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);
#endif

	//always calc in linear space
	EncodeHDR(OUT.rgb);

	//5, bloom scale dependends specular color only and transparent will OUT.rgb * OUT.a
	InitBloomScale( LitResult, MaterialParameters, OUT);
	
// if(@_LOD_LEVEL== LOD_Level_1 )
// {
// OUT.rgb = half3(1,0,0);

// }
// if(@_LOD_LEVEL== LOD_Level_2)
// {
// OUT.rgb = half3(0,0,1);

 
 

// }
// if(@_LOD_LEVEL== LOD_Level_3)
// {
// OUT.rgb = half3(0,1,1);

// }
	
	// pos +=povit.xz;
	
	// OUT.rgb =camdistanceScale;

	return OUT;
}

// Output patch constant data.
struct HS_CONSTANT_DATA_OUTPUT
{
	float EdgeTessFactor[3]			: SV_TessFactor;
	float InsideTessFactor			: SV_InsideTessFactor;
};

// Patch Constant Function
HS_CONSTANT_DATA_OUTPUT CalcHSPatchConstants(
	InputPatch<VertexShadingOutput, 3> ip,
	uint PatchID : SV_PrimitiveID)
{
	HS_CONSTANT_DATA_OUTPUT Output;

	float4 tessFactors = HS_GetTessellationFactor(ip[0], ip[1], ip[2]);
	Output.EdgeTessFactor[0] = tessFactors.x;
	Output.EdgeTessFactor[1] = tessFactors.y;
	Output.EdgeTessFactor[2] = tessFactors.z;
	Output.InsideTessFactor = tessFactors.w;

	return Output;
}

VertexShadingOutput PbrVertexShading_VS(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutputCommon(IN, INTM, OUT);
	return OUT;
}

[domain("tri")]
[partitioning("fractional_odd")]
[outputtopology("triangle_cw")]
[outputcontrolpoints(3)]
[patchconstantfunc("CalcHSPatchConstants")]
VertexShadingOutput PbrVertexShading_HS(
	InputPatch<VertexShadingOutput, 3> ip,
	uint i : SV_OutputControlPointID,
	uint PatchID : SV_PrimitiveID)
{
	return ip[i];
}

[domain("tri")]
VertexShadingOutput PbrVertexShading_DS(
	HS_CONSTANT_DATA_OUTPUT input,
	const OutputPatch<VertexShadingOutput, 3> op,
	float3 domain : SV_DomainLocation)
{

	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);
//	[@_HLSL]
//	OUT = (VertexShadingOutput)0;

#define INTERPOLATE(field) OUT.field = op[0].field * domain.x + op[1].field * domain.y + op[2].field * domain.z

	INTERPOLATE(TexCoord);
	INTERPOLATE(WorldPosition);
	INTERPOLATE(WorldNormal);
	INTERPOLATE(ViewDirection.w);
	INTERPOLATE(DiffLighting);

#if !IS_TERRAIN
	if (@_TANGENT)
#endif
	{
		INTERPOLATE(WorldTangent);
		INTERPOLATE(WorldBinormal);
	}

	if (@_CastSunShadow)
	{
		INTERPOLATE(LitUVZ);
	}

	if (@_VERTEXCOLOR)
	{
		INTERPOLATE(TintColor);
	}

	// Phone tessellation
	DS_DIY_OUTPUT(op[0], op[1], op[2], domain, OUT);

	// recalc ViewDirection & HPosition
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	
	[@_CastSunShadow]
	OUT.LitUVZ = MulMat4x4(ShadowViewProjTexs1, float4(OUT.WorldPosition.xyz,1));

	return OUT;
}

technique Shading
{
	VertexShader = PbrVertexShading_VS();
#if USE_TESSELLATION
	HullShader = PbrVertexShading_HS();
	DomainShader = PbrVertexShading_DS();
#endif
	PixelShader = PbrPixelShading();
}

#include "GBufferGenPass.fxh"

#include "ShadowGenPass.fxh"

#include "HeightmapGenPass.fxh"

#include "DebugShaderPass.fxh"

#if USE_CLIPPING
#include "ClipPass.fxh"
#endif

#if USE_WIREFRAME
#include "WireframePass.fxh"
#endif

#if USE_FROZEN
#include "FrozenPass.fxh"
#endif

#if USE_BURNT
#include "BurntPass.fxh"
#endif

#if USE_HIGHLIGHT
#include "HighlightPass.fxh"
#endif

#if USE_XRAY
#include "XRayPass.fxh"
#endif

#if USE_EDGELINE
#include "EdgeLinePass.fxh"
#endif

#if USE_PLANAR_SHADOW
#include "PlanarShadowPass.fxh"
#endif

#if USE_VBufferGen
#include "VBufferGenPass.fxh"
#endif

#include "EdgeLineNormalGenPass.fxh"

#endif //__FILE_PBR_FXH__
