
#ifndef __FILE_PBR_VS_FXH__
#define __FILE_PBR_VS_FXH__

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 					;
	float4 cTintColor1				: TintColor1				; 
	float4 cTintColor2				: TintColor2				; 
	float4 cTintColor3				: TintColor3				; 

	float4 cShadowBias				: ShadowBias				; 
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
	float4 cPlaceHold											;
#endif
	float4 cParamter				: Param						;
	float4 cParamter2				: Param2					;
#if !IS_TERRAIN	
	float4 cVirtualLitDir			: VirtualLitDir				;
	float4 cVirtualLitColor			: VirtualLitColor			;
#else
	float4 cHeightmapMinPos			: HeightmapMinPos			;//w heightmap height
	float4 cHeightmapMaxPos			: HeightmapMaxPos			;//w heightmap range / texture size 
#endif
	float4x4 WorldViewProj			: WorldViewProj				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex		;
	float4 PlayerPos                : PlayerPos					;
	float4 FoliageCenter            : FoliageCenter				;
	float4	SheltermapMinPos		: SheltermapMinPos			;
	float4	SheltermapMaxPos		: SheltermapMaxPos			;
	float4x4 NormalWorld			: NormalWorld				;
	// WolrdToObject
	float4x3 Local 					: InvWorld;	
};

#define SHADOW_BIAS (cShadowBias.x)
#define SHADOW_SLOPEBIAS (cShadowBias.y)
#if SD_FORWARD_CSM
	#define SHADOW_SELFRATE (frac(cShadowBias.z))
#else
	#define SHADOW_SELFRATE (cShadowBias.z)
#endif
#define SHADOW_PIXELSIZE (cShadowBias.w)
#define LOD_RATE (cParamter.w)
#define SHADOW_BIAS2 (SHADOW_BIAS * floor(cShadowBias.z))
#define SHADOW_SLOPEBIAS2 (SHADOW_SLOPEBIAS * floor(cShadowBias.z))


//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_UV3UV4]
	float4		TexCoord3		: TEXCOORD2;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
	[@_INSTANCED]
	float4		LightMapScale	: TEXCOORD6;		// w : Dissolving Ratio
	[@_INSTANCED]
	float4		LightMapTransform: TEXCOORD7;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float		ModelPositionY	: TEXCOORDN;			//	Model PositionY

	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	
	float4 		DiffLighting 	: TEXCOORDN;			//  PerVertex Diffuse Lighting.
 	
#if IS_TERRAIN//terrain always need tangent/binormal
	half3		WorldTangent	: TEXCOORDN;
	half3		WorldBinormal	: TEXCOORDN;
#else	
	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;
#endif

	[@_CastSunShadow]
	float4 		LitUVZ			: TEXCOORDN; 

#if IS_WATER
	[@_CastSunShadow]
	float4 		LitUVZ2			: TEXCOORDN;
#elif USE_PLAYER_SHADOWMAP
	[@_CastSunShadow && @_TRANSPARENT]
	float4 		LitUVZ2			: TEXCOORDN;
#endif

	// Shader special output
	float4		TintColor 		: COLOR;				//	TintColor

	[@_INSTANCED]
	float4		LightMapScale	: TEXCOORDN;			//	Light map scale

#if USE_LIGHTMAP_CONST
	[@_INSTANCED]
	float4		LightMapTransform : TEXCOORDN;
#endif

	float 		LinearZ			: TEXCOORDN;
#ifdef HAS_WORLD_POSITION_OFFSET
	half3		WorldOffset		: TEXCOORDN;
#endif
	float 		CameraDistance  : TEXCOORDN;

	[@_UV3UV4]
	float4		MultiUVs		: TEXCOORDN;

};


float4 GetVAMultiUVS(in VertexShadingOutput arg)
{
    if (@_UV3UV4)
        return arg.MultiUVs;
    else
        return float4(0, 0, 0, 0);
}
#define VAMultiUVS(Arg)		 GetVAMultiUVS(Arg)
#define VSMultiUVS           VAMultiUVS(OUT)
#define PSMultiUVS           VAMultiUVS(IN)

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;

	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;
};


//////////////////////////////////Functions////////////////////////////////////////
void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{
	//[@_HLSL]
	//OUT = (VertexShadingOutput)0;

	//PBR_VS::struct VertexShadingOutput


	OUT.HPosition = float4(0, 0, 0, 0);

	OUT.TexCoord = float4(0,0,0,0);

	OUT.WorldPosition = float4(0,0,0,0);

	OUT.ModelPositionY = 0;

	OUT.WorldNormal = float4(0,0,0,0);

	OUT.ViewDirection = float4(0,0,0,0);

	OUT.DiffLighting = float4(0,0,0,1);

	#if IS_TERRAIN
		OUT.WorldTangent = half3(0,0,0);
		OUT.WorldBinormal = half3(0,0,0);
	#else
	[@_TANGENT]
	OUT.WorldTangent = half4(0,0,0,0);
	[@_TANGENT]	
	OUT.WorldBinormal = half4(0,0,0,0);
	#endif

	[@_CastSunShadow]
	OUT.LitUVZ = float4(0,0,0,0);

#if IS_WATER
	[@_CastSunShadow]
	OUT.LitUVZ2	 = float4(0,0,0,0);
#elif USE_PLAYER_SHADOWMAP
	[@_CastSunShadow && @_TRANSPARENT]
	OUT.LitUVZ2	 = float4(0,0,0,0);
#endif

	OUT.TintColor = float4(0,0,0,0);

	[@_INSTANCED]
	OUT.LightMapScale = float4(0,0,0,0);

#if USE_LIGHTMAP_CONST
	[@_INSTANCED]
	OUT.LightMapTransform = float4(0,0,0,0);
#endif

#ifdef HAS_WORLD_POSITION_OFFSET
	OUT.WorldOffset = half3(0, 0, 0);
#endif
	OUT.LinearZ = 0; 

	OUT.CameraDistance = 1;

	[@_UV3UV4]
	OUT.MultiUVs = float4(0,0,0,0);
	
}

void InitInterMediateVertex(inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(0,0,0,0);
	
	INTM.LocalNormal = float3(0,0,0);

	[@_TANGENT]
	INTM.LocalTangent = float3(0,0,0);

	[@_TANGENT]
	INTM.LocalBinormal = float3(0,0,0);
}

void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	//[@_HLSL]
	//INTM = (InterMediateVertex)0;

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand(IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		uint4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
			}
		}
	}
}

// Get 3x4 matrix from skeleton data.
void GetSkeletonMatrix4x3Tex( uint idx, out float4 Mx, out float4 My, out float4 Mz )
{
	Mx = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+0.5)*cParamter.z, 0), 0);
	My = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+1.5)*cParamter.z, 0), 0);
	Mz = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+2.5)*cParamter.z, 0), 0);
}

void VertexUnifySkinningTex(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		uint4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3Tex(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3Tex(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3Tex(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3Tex(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
			}
		}
		/*
		// Hard skinning.
		if (@_DF_SKIN == 2)
		{
			// Fetch.
			float4 Mx, My, Mz;
			GetSkeletonMatrix4x3(i[0], Mx, My, Mz);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalNormal);

			if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalTangent);
				INTM.LocalBinormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalBinormal);
			}
		}
		// Dual-quaternion skinning.
		if (@_DF_SKIN == 3)
		{
			// Fetch.
			float4 Q0[2], Q1[2], Q2[2], Q3[2];
			GetSkeletonDualQuaternion(i[0], Q0[0], Q0[1]);
			GetSkeletonDualQuaternion(i[1], Q1[0], Q1[1]);
			GetSkeletonDualQuaternion(i[2], Q2[0], Q2[1]);
			GetSkeletonDualQuaternion(i[3], Q3[0], Q3[1]);
			// Amend orientations.
			w *= float4(1.f, sign(dot(Q0[0], Q1[0])), sign(dot(Q0[0], Q2[0])), sign(dot(Q0[0], Q3[0])));
	
			// Blend.
			float4 Qn = LinearBlend(w, Q0[0], Q1[0], Q2[0], Q3[0]);
			float4 Qd = LinearBlend(w, Q0[1], Q1[1], Q2[1], Q3[1]); 

			// Apply.
			INTM.LocalPosition.xyz = DualQuaternionTransformP(Qn, Qd, INTM.LocalPosition.xyz);
			INTM.LocalNormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalNormal);

			if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = DualQuaternionTransformN(Qn, Qd, INTM.LocalTangent);
				INTM.LocalBinormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalBinormal);
			}
		}
		*/
	}
}

void InitTintColorCommon(in VertexAssemble IN, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR && !ENABLE_ANISOTROPIC)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
}

void InitHPositionCommon(in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}

	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);
		
		//overwrite Fog T
		// OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
		OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz));
	}

	

}
void InitDepthCommon(in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{

if(@_EnableDepthBias)
{

float depthBias = cDepthBias;
	OUT.HPosition.z -= CameraInfo.x *depthBias / (depthBias + OUT.HPosition.w);

	OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
	OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	
}

 		OUT.HPosition.w *= cFov;
	
	

}



void InitEmissionUVCommon(in VertexAssemble IN, inout VertexShadingOutput OUT)
{
	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;	
	if(@_EmissionMapEnable)
	{
		half2 EmissionUV = IN.TexCoord.xy * cEmissionScrolling.xy + GameTime * cEmissionScrolling.zw;
		OUT.WorldNormal.w = EmissionUV.x;
		OUT.ViewDirection.w = EmissionUV.y;
	}
}

#endif//__FILE_PBR_VS_FXH__
