///	@file	<FX_Lighting.fx>
///	@path	~/System/Shader/
///	@date	2019/01/--
///	@brief	FX shading model.

#ifndef __FILE_FX_SHADING_FXH__
#define __FILE_FX_SHADING_FXH__

// #ifndef FORCE_DISABLE_SHADOW
// #define FORCE_DISABLE_SHADOW 0
// #endif

@_PARTICLECLIPPING
{
	Name = ParticleClipping;
	Description = "Particle Clipping";
	Default = false;
}

float4 UnlitPixelShading(in const VertexShadingOutput IN,in const float4 screenCoord : SV_Position) : SV_Target
{
	// 1. Init Shading Parameters Struct
    ParticleShadingParameters materialParamters;
    InitParticleShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(IN,materialParamters);

	// 3. EncodeHDR
    EncodeHDRParticle(materialParamters.BaseColor);

	//All Particle is Transparent
	InitFogLightingParticle(IN,materialParamters.BaseColor);

	// 4. Soft Particle
	if (@_SOFT)
		SoftParticle(IN,materialParamters);
	

    return materialParamters.BaseColor;
}


float4 DistortionPixelShading(in const VertexShadingOutput IN,in const float4 screenCoord : SV_Position) : SV_Target
{
	// 1. Init Shading Parameters Struct	
    ParticleShadingParameters materialParamters;
    InitParticleShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(IN,materialParamters);

	// 3. Calc WorldNormal using ddx/ddy
	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

    float4 OUT = float4(0,0,0,0);
	// calcuate patch normal
	float3 P = IN.WorldPosition;
	float3 N = cross(ddy(P), ddx(P));
	if (@_GLSL || @_SPIRV)
	{
		// ddx/ddy is incorrect on some android devices.
		float len = length(N);
		N = (len == 0) ? float3(0, 1, 0) : N / len;
	}
	else
	{
		N = normalize(N);
	}

	// 3.2 hacked refraction
	float3 V = normalize(IN.ViewDirection);
	float3 R = cross(N, V);
	float3 D = cross(R, V);

	float3 VR = cross(float3(0, 1, 0), V);
	float3 VU = normalize(cross(V, VR));
	VR = cross(VU, V);

	float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
	offset *= (500 / IN.TexCoord.z) * (materialParamters.BaseColor.x);

	OUT = SampleColorMap(ScreenTexture, ScreenTexture_Sampler, stc + offset);

	//OUT.a = EncodeTransparentBloomFactor(1, 0);
	OUT.a = 1;

    return OUT;
}


float4 LightingShading(VertexShadingOutput IN,ParticleShadingParameters materialParamters)
{
	// 3. Get ScreenInfo For Shadow
	half2 shadowAndAO = half2(0,0);

	// There is No Shadow For Particle
	// if(FORCE_DISABLE_SHADOW == 0)
	// {	
	// 	float2 spos = IN.HPosition.xy;
	// 	[@_D3D9]
	// 	spos += 0.5;
	// 	float2 stc = spos * ScreenInfo.zw;
	// 	shadowAndAO = tShadowMask.Sample(sShadowMaskSampler, stc).rg;
	// }

	materialParamters.Roughness = max(materialParamters.Roughness,MIN_ROUGHNESS);

	half3 N = materialParamters.WorldNormal;
	half InShadow = shadowAndAO.x;
	half3 GILighting = half3(0,0,0); 
	half3 Emission = materialParamters.Emission;

 	half3 L = normalize( SunDirection.xyz );
 	half3 V = normalize( IN.ViewDirection.xyz );
 	half3 H = normalize( V + L);

	//using Half for Performance but May cause issue
	half NoH = saturate(dot(N,H));
 	half NoL = dot(materialParamters.WorldNormal, L);
	half NoLS = saturate(NoL);
	half NoV = saturate(dot(N,V));
	// half VoH = saturate(dot(V,H));
	half VoL = saturate(-dot(L,V));

	half lightDensity = henyeyGreenstein(VoL,ScatteringPhaseG) * 5.0 ;
	half3 inputRadiance = saturate( NoL * Thickness ) * (1 - InShadow) + lightDensity * Scattering ; 

	half3 SpecularColor = lerp(half3(0.04,0.04,0.04) , materialParamters.BaseColor.xyz, (half)materialParamters.Metallic);
	half3 diffuseColor = Diffuse_Lambert(materialParamters.BaseColor.xyz - materialParamters.BaseColor.xyz * (half)materialParamters.Metallic);

	DynamicGILight(N,GILighting,InShadow);
	GILighting *= (diffuseColor /* *shadowAndAO.y */); // There is No Shadow For Particle

	half D = D_GGX(materialParamters.Roughness,NoH);
	half G = G_Schlick( materialParamters.Roughness, NoV, NoLS );
	// half3 F = F_Schlick( SpecularColor, VoH );
	half3 F = F_None(SpecularColor);

	//Fake GI Lighting using diffuseColor
	half3 diffuseLighting = inputRadiance * SunColor.rgb * diffuseColor * ThunderRate;
	half3 SpecularLighting = D * G * F * inputRadiance * SunColor.rgb * ThunderRate;
	SpecularLighting = step(0,NoLS) * SpecularLighting;

	half4 OUT = half4(0,0,0,materialParamters.BaseColor.w);

	OUT.rgb += diffuseLighting;
	OUT.rgb += SpecularLighting;
	OUT.rgb += GILighting;
	OUT.rgb += Emission;

	OUT.rgb *= saturate(materialParamters.BaseColor.w * AlphaScale);
		
	InitFogLightingParticle(IN,OUT);

	return OUT;
}

float4 ParticleLightingPixelShading(in const VertexShadingOutput IN,in const float4 screenCoord : SV_Position) : SV_Target
{
	// 1. Init Shading Parameters Struct	
    ParticleShadingParameters materialParamters;
    InitParticleShadingParameters(materialParamters);

	// 2. Custom Get Shading Parameters
    PS_GetParameters(IN,materialParamters);

	// 3. Lighting Shading
	materialParamters.BaseColor = LightingShading(IN,materialParamters);

	if (@_SOFT)
		SoftParticle(IN,materialParamters);
	
	float4 OUT = materialParamters.BaseColor;

	EncodeHDR(OUT.rgb);

	if (@_PARTICLECLIPPING)
		clip(OUT.a - ALPHA_REF);

	return OUT;
}

float4 ParticleUnifyPixelShading(in const VertexShadingOutput IN, in const float4 screenCoord : SV_Position) : SV_Target
{
	// Src Particle Behavior
    if(@_DISTORTION)
	{
        return DistortionPixelShading(IN,screenCoord);
	}
    else
	{
        return UnlitPixelShading(IN,screenCoord);
	}

    return float4(1,0,1,1);
}


#endif //__FILE_FX_SHADING_FXH__
