#ifndef __FILE_PLANARSHADOW_FXH__
#define __FILE_PLANARSHADOW_FXH__
/////////////////////////////////////PlanarShadow/////////////////////////////////////
struct PlanarShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		TC 				: TEXCOORDN;			//	TexCoord
};

//viewProjVS is light view now
PlanarShadowVertexShadingOutput PlanarShadowVertexShading(in VertexAssemble IN)
{
	PlanarShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (PlanarShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif	

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}
	//planar equation
	float t = dot(cParamter2.xyz, WorldPosition.xyz - cParamter.xyz) / dot(cParamter2.xyz, SunDirection.xyz);
	WorldPosition.xyz -= max(t*0.95-0.02, 0) * SunDirection.xyz;

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//TC
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 PlanarShadowPixelShading(in PlanarShadowVertexShadingOutput IN) : SV_Target
{
#ifndef NO_SHADOW_CLIPPING
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		float alpha = tBaseMap.SampleRGBA( sBaseSampler, IN.TC ).a;
		clipValue = alpha - ALPHA_REF;
	}
	if (@_TRANSPARENT)
	{
		clipValue = tBaseMap.SampleRGBA( sBaseSampler, IN.TC ).a - 0.3;
	}
	if (@_CLIPPING || @_TRANSPARENT)
		clip(clipValue);
#endif
	return float4(float3(0,0,0), SHADOW_RATE);
}

technique PlanarShadow
{
	VertexShader = PlanarShadowVertexShading();
	PixelShader = PlanarShadowPixelShading();
}
#endif // __FILE_PLANARSHADOW_FXH__
