#ifndef _ScreenSpaceRayCast_
#define _ScreenSpaceRayCast_ 1

#include "Common.fxh"

half InterleavedRandomUV( in half2 xy)
{
	return frac(52.9829189 * frac(dot(xy, float2( 0.06711056, 0.00583715))));
}

#define SSR_HIGH_QUALITY		0

#if SSR_HIGH_QUALITY
#define SSR_STRIDE				16
#define SSR_SAMPLE_COUNT		32
#else
#define SSR_STRIDE 				8
#define SSR_SAMPLE_COUNT 		4
#endif

float3 ScreenSpaceRayCastWS(in float3 WorldPosition, in float2 PixelPos, in float3 R)
{
	float2 UVMul = float2(0.5, -0.5);
	if(@_GLSL)
	{
		UVMul.y = 0.5;
	}
	R = normalize(R);

	float4 RayStartClip = MulMat4x4(ViewProj, float4(WorldPosition, 1.0));
	float3 RayStartDevice;
	RayStartDevice.xy = RayStartClip.xy * UVMul + float2(0.5, 0.5) * RayStartClip.w;
	RayStartDevice.z = RayStartClip.w;


	float4 RayEndClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 16, 1.0));
	float3 RayEndDevice;
	RayEndDevice.xy = RayEndClip.xy * UVMul + float2(0.5, 0.5) * RayEndClip.w;
	RayEndDevice.z = RayEndClip.w;

	float3 RayDirDevice = (RayEndDevice - RayStartDevice) / SSR_SAMPLE_COUNT;

	// Reflection
	if(RayEndClip.w < RayStartClip.w)
	{
		return float3(0, 0, 0);
	}

	float StepOffset = 0;// InterleavedRandomUV(PixelPos) * 0.5;

	float LastRayDepth = RayStartClip.w;
	float SampleTime = StepOffset + 1;
	float3 LastSampleUVZ = RayStartDevice;
	[loop]
	for (int i = 0; i < SSR_SAMPLE_COUNT; ++i)
	{
		float3 SampleUVZ = RayStartDevice + RayDirDevice * SampleTime;
		SampleUVZ.xy /= SampleUVZ.z;
		if(SampleUVZ.x <= 0 || SampleUVZ.x >= 1 || SampleUVZ.y <= 0 || SampleUVZ.y >= 1)
		{
			break;
		}

		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ.xy, 0)) * CameraInfo.y;
 		
		float RayDepth = SampleUVZ.z;
		float DepthDiff = RayDepth - Depth;
		
		if(DepthDiff > 0 && abs(DepthDiff < 3.5))
		{
			float LastDiff = Depth - LastSampleUVZ.z;
			if(1)
			{
				float weight = LastDiff / (LastDiff + DepthDiff);
				float2 UV = SampleUVZ.xy * weight + LastSampleUVZ.xy * (1 - weight);
				float Depth2 = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, UV.xy, 0)) * CameraInfo.y;
				if(abs(Depth2 - Depth) < 3)
				{
			//		return float3(UV.xy, i+1);
				}				
			}
			else
			{
			//	return float3(SampleUVZ.xy, i+1);
			}
		}

		LastRayDepth = Depth;
		LastSampleUVZ = SampleUVZ;
		SampleTime += 1;
	}
	float4 RayInfClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 10000000, 1.0));
	float2 RayInfDevice = RayInfClip.xy / RayInfClip.w;
	if(RayInfDevice.x > -1 && RayInfDevice.x < 1 && RayInfDevice.y > -1 && RayInfDevice.y < 1)
	{
		float2 SampleUV = RayInfDevice * UVMul + 0.5;
		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUV, 0)) * CameraInfo.y;
		if(Depth > RayStartClip.w)
		{
			return float3(SampleUV, SSR_SAMPLE_COUNT);
		}
	}
	return float3(0, 0, 0);
}

float3 ScreenSpaceRayCast(in float3 WorldPosition, in float2 PixelPos, in float3 R)
{
	float2 UVMul = float2(0.5, -0.5);
	if(@_GLSL)
	{
		UVMul.y = 0.5;
	}

	float4 RayStartClip = MulMat4x4(ViewProj, float4(WorldPosition, 1.0));
	float3 RayStartDevice;
	float k0 = 1.0 / RayStartClip.w;
	RayStartDevice.xy = RayStartClip.xy * UVMul * k0 + 0.5;
	RayStartDevice.z = k0;

	float stride = lerp(0.125 * SSR_STRIDE, SSR_STRIDE, saturate(1.0 - RayStartClip.w / 300.0));

	float4 RayEndClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 500, 1.0));
	float3 RayEndDevice;
	float k1 = 1.0 / RayEndClip.w;
	RayEndDevice.xy = RayEndClip.xy * UVMul * k1 + 0.5;
	RayEndDevice.z = k1;

	// Reflection
	if(RayEndClip.w < RayStartClip.w)
	{
		return float3(0, 0, 0);
	}

	float3 RayDeviceDelta = RayEndDevice - RayStartDevice;
	if(RayDeviceDelta.x * RayDeviceDelta.x + RayDeviceDelta.y * RayDeviceDelta.y < 0.0001)
	{
		RayDeviceDelta.xy += float2(0.01, 0.01);
	}

	float2 PixelDelta = RayDeviceDelta.xy * ScreenInfo.xy;
	RayDeviceDelta /= max(abs(PixelDelta.x), abs(PixelDelta.y));
	RayDeviceDelta *= stride;

	float StepOffset = InterleavedRandomUV(PixelPos) * 0.5;

	float LastRayDepth = RayStartClip.w;
	float SampleTime = StepOffset + 1;
	float3 SampleUVZ = RayStartDevice + RayDeviceDelta * (StepOffset + 1);
	float3 LastSampleUVZ = SampleUVZ;
	[loop]
	for (int i = 0; i < SSR_SAMPLE_COUNT; ++i)
	{
		if(SampleUVZ.x <= 0 || SampleUVZ.x >= 1 || SampleUVZ.y <= 0 || SampleUVZ.y >= 1)
		{
			break;
		}

		float RayDepthNextHalfPixel = 1.0 / (SampleUVZ.z + RayDeviceDelta.z * 0.5);
		if(RayDepthNextHalfPixel >= CameraInfo.y)
		{
			break;
		}

		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ.xy, 0)) * CameraInfo.y;
 		

		float DepthDiff = RayDepthNextHalfPixel - Depth;
		if((Depth < 500) && (DepthDiff > 0) && (LastRayDepth < Depth + 1.5))
		{
			float weight = DepthDiff / abs(RayDepthNextHalfPixel - LastRayDepth);
			float2 UV = SampleUVZ.xy * (1 - weight) + LastSampleUVZ.xy * weight;
			float D = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, UV.xy, 0)) * CameraInfo.y;
			if(abs(D - Depth) < 0.1)
			{
				return float3(UV.xy, i+1);
			}
		}
		LastSampleUVZ = SampleUVZ;
		
		if((DepthDiff > 0) && (LastRayDepth < Depth + 1.5))
		{
			SampleUVZ -= RayDeviceDelta;
			RayDeviceDelta *= 0.5;
		}
		else
		{
			LastRayDepth = RayDepthNextHalfPixel;
		}

		SampleUVZ += RayDeviceDelta;
	}

	float4 RayInfClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 50000, 1.0));
	float2 RayInfDevice = RayInfClip.xy / RayInfClip.w;
	if((RayInfDevice.x > -1) && (RayInfDevice.x < 1) && (RayInfDevice.y > -1) && (RayInfDevice.y < 1))
	{
		float2 SampleUV = RayInfDevice * UVMul + 0.5;
		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUV, 0)) * CameraInfo.y;
		// No-Near, No-Sky
		if((Depth > 100))// && (Depth < CameraInfo.y))
		{
			return float3(SampleUV, SSR_SAMPLE_COUNT);
		}
	}
	return float3(0, 0, 0);
}
#endif
