#ifndef _ScreenSpaceRayCast_
#define _ScreenSpaceRayCast_ 1

#include "Common.fxh"

half InterleavedRandomUV( in half2 xy)
{
	return frac(52.9829189 * frac(dot(xy, float2( 0.06711056, 0.00583715))));
}

#define SSR_SAMPLE_COUNT 		32
#define SSR_TRACING_STEP		64
#define SSR_VECTORIZED 			0

#if SSR_VECTORIZED
float3 ScreenSpaceRayCast(in float3 WorldPosition, in half2 PixelPos, in float3 R)
{
	float2 UVMul = float2(0.5, -0.5);
	if(@_GLSL)
	{
		UVMul.y = 0.5;
	}
	float4 RayStartClip = MulMat4x4(ViewProj, float4(WorldPosition, 1.0));
	float3 RayStartDevice;
	float k = 1.0 / RayStartClip.w;
	RayStartDevice.xy = RayStartClip.xy * k * UVMul + 0.5;
	RayStartDevice.z = k;


	float4 RayEndClip = MulMat4x4(ViewProj, float4(WorldPosition + R, 1.0));
	float3 RayEndDevice;
	k = 1.0 / RayEndClip.w;
	RayEndDevice.xy = RayEndClip.xy * k * UVMul + 0.5;
	RayEndDevice.z = k;

	float3 RayDirDevice = RayEndDevice - RayStartDevice;

	// Reflection must from near to far
	if(RayEndClip.w < RayStartClip.w)
	{
		return float3(0, 0, 0);
	}

	// Normalize Ray Step to 1 pixel
	if(abs(RayDirDevice.x) > abs(RayDirDevice.y))
	{
		RayDirDevice *= ScreenInfo.z / abs(RayDirDevice.x);
	}
	else
	{
		RayDirDevice *= ScreenInfo.w / abs(RayDirDevice.y);
	}
	RayDirDevice *= SSR_TRACING_STEP;

	half StepOffset = InterleavedRandomUV(PixelPos) * 0.1;

	half LastRayDepth = RayStartClip.w;
	half4 SampleTime = StepOffset + float4(1, 2, 3, 4);
	float DepthRange = CameraInfo.y - CameraInfo.x;
	[loop]
	for (int i = 0; i < SSR_SAMPLE_COUNT / 4; ++i)
	{
		float4 SampleUVZ0 = RayStartDevice.xyxy + RayDirDevice.xyxy * SampleTime.xxyy;
		float4 SampleUVZ1 = RayStartDevice.xyxy + RayDirDevice.xyxy * SampleTime.zzww;
		float4 SampleW    = RayStartDevice.zzzz + RayDirDevice.zzzz * SampleTime.xyzw;
		
		float4 RayDepth = 1.0 / SampleW;

		// Clip to viewport
		if(SampleUVZ1.z <= 0 || SampleUVZ1.z >= 1 || SampleUVZ1.w <= 0 || SampleUVZ1.w >= 1)
		{
			break;
		}

		float4 SampleDepth = float4(0, 0, 0, 0);
		SampleDepth.x = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ0.xy, 0)) * DepthRange + CameraInfo.x;
		SampleDepth.y = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ0.zw, 0)) * DepthRange + CameraInfo.x;
		SampleDepth.z = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ1.xy, 0)) * DepthRange + CameraInfo.x;
		SampleDepth.w = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ1.zw, 0)) * DepthRange + CameraInfo.x;
 		
		float4 DepthDiff = RayDepth - SampleDepth;
		/* TODO: The follow codes Report ERROR in ShaderMan for  metal!!
		bool4 Hit = (DepthDiff > 0) && 
					(SampleDepth > RayStartClip.w) && 
					(DepthDiff < abs(RayDepth - float4(LastRayDepth, RayDepth.x, RayDepth.y, RayDepth.z)) * 1.5);

		float4 HitTime = Hit ? SampleTime : float4(99999, 99999, 99999, 99999);
		*/
		half4 Hit = step(0, DepthDiff) *
					step(RayStartClip.w, SampleDepth) *
					step(DepthDiff, abs(RayDepth - float4(LastRayDepth, RayDepth.x, RayDepth.y, RayDepth.z)) * 1.5);
		if(Hit.x > 0)
		{
			return float3(RayStartDevice.xy + RayDirDevice.xy * SampleTime.x, 1);			
		}
		if(Hit.y > 0)
		{
			return float3(RayStartDevice.xy + RayDirDevice.xy * SampleTime.y, 1);			
		}		
		if(Hit.z > 0)
		{
			return float3(RayStartDevice.xy + RayDirDevice.xy * SampleTime.z, 1);			
		}
		if(Hit.w > 0)
		{
			return float3(RayStartDevice.xy + RayDirDevice.xy * SampleTime.w, 1);			
		}		
		/*
		// convert 0 -> 99999, 1 -> 1
		float4 HitTime = (1 - Hit) * 99999 + 1;
		HitTime *= SampleTime;
		HitTime.xy = min(HitTime.xy, HitTime.zw);
		float MinHitTime = min(HitTime.x, HitTime.y);
		if(MinHitTime < 99999)
		{
			float2 HitUV = RayStartDevice.xy + RayDirDevice.xy * MinHitTime;
			return float3(HitUV, 1);
		}*/
		LastRayDepth = RayDepth.w;
		SampleTime += 4;
	}

	float4 RayInfClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 10000000, 1.0));
	float2 RayInfDevice = RayInfClip.xy / RayInfClip.w;
	if(RayInfDevice.x > -1 && RayInfDevice.x < 1 && RayInfDevice.y > -1 && RayInfDevice.y < 1)
	{
		float2 SampleUV = RayInfDevice * UVMul + 0.5;
		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUV, 0)) * CameraInfo.y;
		if(Depth > RayStartClip.w)
		{
			return float3(SampleUV, 1);
		}
	}
	return float3(0, 0, 0);
}

#else
float3 ScreenSpaceRayCastWS(in float3 WorldPosition, in float2 PixelPos, in float3 R)
{
	float2 UVMul = float2(0.5, -0.5);
	if(@_GLSL)
	{
		UVMul.y = 0.5;
	}
	R = normalize(R);

	float4 RayStartClip = MulMat4x4(ViewProj, float4(WorldPosition, 1.0));
	float3 RayStartDevice;
	RayStartDevice.xy = RayStartClip.xy * UVMul + float2(0.5, 0.5) * RayStartClip.w;
	RayStartDevice.z = RayStartClip.w;


	float4 RayEndClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 16, 1.0));
	float3 RayEndDevice;
	RayEndDevice.xy = RayEndClip.xy * UVMul + float2(0.5, 0.5) * RayEndClip.w;
	RayEndDevice.z = RayEndClip.w;

	float3 RayDirDevice = (RayEndDevice - RayStartDevice) / SSR_SAMPLE_COUNT;

	// Reflection
	if(RayEndClip.w < RayStartClip.w)
	{
		return float3(0, 0, 0);
	}

	float StepOffset = 0;// InterleavedRandomUV(PixelPos) * 0.5;

	float LastRayDepth = RayStartClip.w;
	float SampleTime = StepOffset + 1;
	float3 LastSampleUVZ = RayStartDevice;
	[loop]
	for (int i = 0; i < SSR_SAMPLE_COUNT; ++i)
	{
		float3 SampleUVZ = RayStartDevice + RayDirDevice * SampleTime;
		SampleUVZ.xy /= SampleUVZ.z;
		if(SampleUVZ.x <= 0 || SampleUVZ.x >= 1 || SampleUVZ.y <= 0 || SampleUVZ.y >= 1)
		{
			break;
		}

		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ.xy, 0)) * CameraInfo.y;
 		
		float RayDepth = SampleUVZ.z;
		float DepthDiff = RayDepth - Depth;
		
		if(DepthDiff > 0 && abs(DepthDiff < 3.5))
		{
			float LastDiff = Depth - LastSampleUVZ.z;
			if(1)
			{
				float weight = LastDiff / (LastDiff + DepthDiff);
				float2 UV = SampleUVZ.xy * weight + LastSampleUVZ.xy * (1 - weight);
				float Depth2 = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, UV.xy, 0)) * CameraInfo.y;
				if(abs(Depth2 - Depth) < 3)
				{
			//		return float3(UV.xy, i+1);
				}				
			}
			else
			{
			//	return float3(SampleUVZ.xy, i+1);
			}
		}

		LastRayDepth = Depth;
		LastSampleUVZ = SampleUVZ;
		SampleTime += 1;
	}
	float4 RayInfClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 100000, 1.0));
	float2 RayInfDevice = RayInfClip.xy / RayInfClip.w;
	if(RayInfDevice.x > -1 && RayInfDevice.x < 1 && RayInfDevice.y > -1 && RayInfDevice.y < 1)
	{
		float2 SampleUV = RayInfDevice * UVMul + 0.5;
		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUV, 0)) * CameraInfo.y;
		if(Depth > RayStartClip.w)
		{
			return float3(SampleUV, SSR_SAMPLE_COUNT);
		}
	}
	return float3(0, 0, 0);
}

float3 ScreenSpaceRayCast(in float3 WorldPosition, in float2 PixelPos, in float3 R)
{
	float2 UVMul = float2(0.5, -0.5);
	if(@_GLSL)
	{
		UVMul.y = 0.5;
	}
	R = normalize(R);

	float4 RayStartClip = MulMat4x4(ViewProj, float4(WorldPosition, 1.0));
	float3 RayStartDevice;
	float k0 = 1.0 / RayStartClip.w;
	RayStartDevice.xy = RayStartClip.xy * UVMul * k0 + 0.5;
	RayStartDevice.z = k0;

	float STRIDE = 16;
	float stride = lerp(0.125 * STRIDE, STRIDE, saturate(1.0 - RayStartClip.w / 100.0));

	float4 RayEndClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 500, 1.0));
	float3 RayEndDevice;
	float k1 = 1.0 / RayEndClip.w;
	RayEndDevice.xy = RayEndClip.xy * UVMul * k1 + 0.5;
	RayEndDevice.z = k1;

	// Reflection
	if(RayEndClip.w < RayStartClip.w)
	{
		return float3(0, 0, 0);
	}

	float3 RayDeviceDelta = RayEndDevice - RayStartDevice;
	if(RayDeviceDelta.x * RayDeviceDelta.x + RayDeviceDelta.y * RayDeviceDelta.y < 0.0001)
	{
		RayDeviceDelta.xy += float2(0.01, 0.01);
	}

	float2 PixelDelta = RayDeviceDelta.xy * ScreenInfo.xy;
	RayDeviceDelta /= max(abs(PixelDelta.x), abs(PixelDelta.y));
	RayDeviceDelta *= stride;

	float StepOffset = InterleavedRandomUV(PixelPos) * 0.5;

	float LastRayDepth = RayStartClip.w;
	float SampleTime = StepOffset + 1;
	float3 SampleUVZ = RayStartDevice + RayDeviceDelta * (StepOffset + 1);
	float3 LastSampleUVZ = SampleUVZ;
	[loop]
	for (int i = 0; i < SSR_SAMPLE_COUNT; ++i)
	{
		if(SampleUVZ.x <= 0 || SampleUVZ.x >= 1 || SampleUVZ.y <= 0 || SampleUVZ.y >= 1)
		{
			break;
		}

		float RayDepthNextHalfPixel = 1.0 / (SampleUVZ.z + RayDeviceDelta.z * 0.5);
		if(RayDepthNextHalfPixel >= CameraInfo.y)
		{
			break;
		}

		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ.xy, 0)) * CameraInfo.y;
 		

		float DepthDiff = RayDepthNextHalfPixel - Depth;

		if(DepthDiff > 0 && LastRayDepth < Depth + 1.5)
		{
			float weight = DepthDiff / abs(RayDepthNextHalfPixel - LastRayDepth);
			float2 UV = SampleUVZ.xy * (1 - weight) + LastSampleUVZ.xy * weight;
			float D = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, UV.xy, 0)) * CameraInfo.y;
			if(abs(D - Depth) < 0.1)
			{
				return float3(UV.xy, i+1);
			}
		//	return float3(SampleUVZ.xy, i+1);
		}
		LastSampleUVZ = SampleUVZ;
		
		if(DepthDiff > 0 && LastRayDepth < Depth + 1.5)
		{
			SampleUVZ -= RayDeviceDelta;
			RayDeviceDelta *= 0.5;
		}
		else
		{
			LastRayDepth = RayDepthNextHalfPixel;
		}

		SampleUVZ += RayDeviceDelta;
	}
	SampleUVZ = RayStartDevice + RayDeviceDelta * (SSR_SAMPLE_COUNT + 1);
	if(SampleUVZ.x > 0 && SampleUVZ.x < 1 && SampleUVZ.y > 0 && SampleUVZ.y < 1)
	{
		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUVZ.xy, 0)) * CameraInfo.y;
		if(Depth > RayStartClip.w)
		{
		//	return float3(SampleUVZ.xy, SSR_SAMPLE_COUNT);
		}
	}
	/*
	float4 RayInfClip = MulMat4x4(ViewProj, float4(WorldPosition + R * 100000, 1.0));
	float2 RayInfDevice = RayInfClip.xy / RayInfClip.w;
	if(RayInfDevice.x > -1 && RayInfDevice.x < 1 && RayInfDevice.y > -1 && RayInfDevice.y < 1)
	{
		float2 SampleUV = RayInfDevice * UVMul + 0.5;
		float Depth = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, SampleUV, 0)) * CameraInfo.y;
		if(Depth > RayStartClip.w)
		{
			return float3(SampleUV, SSR_SAMPLE_COUNT);
		}
	}*/
	return float3(0, 0, 0);
}
#endif

#endif
