// hex shader
#ifndef __FILE_SHADOWGEN_FXH__
#define __FILE_SHADOWGEN_FXH__

/////////////////////////////////////ShadowGen/////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		TC 				: TEXCOORDN;			//	TexCoord
};

void computeTangent(in half3 gatherNormal, out half3 tan0, out half3 tan1)
{
	tan0 = half3( gatherNormal.z, 0.f, -gatherNormal.x );
	half lg0 = dot( tan0, tan0 );
	if ( lg0 < HALF_EPSILON )
	{
		tan0 = half3(gatherNormal.y,-gatherNormal.x,0.f);
		//tan0-=dot(tan0,gatherNormal)*gatherNormal
		tan0 = normalize( tan0 );
	}
	else
		tan0 *= rsqrt( lg0 );
	tan1 = normalize( cross(gatherNormal,tan0) );
}

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	//[@_HLSL]
	//OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	//world normal
	float3 WorldNormal;
	if (@_INSTANCED)
	{
		WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
	}
	else
	{
		WorldNormal = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	}

	//TC
	OUT.TC = IN.TexCoord.xy;

#ifdef HAS_WORLD_POSITION_OFFSET
	if(NEED_SHADOWMAP_REFRESH)
		WorldPosition.xyz += WorldPositionOffset(IN, WorldPosition.xyz, WorldNormal);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
#else
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	}
	else
	{
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
#endif
	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
#ifndef NO_SHADOW_CLIPPING
	if (@_CLIPPING)
	{
		float alpha = tBaseMap.SampleRGBA( sBaseSampler, IN.TC ).a;
		clip(alpha - 0.1);
	}
#endif
	return float4(1, 1, 1, 1);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// PBR stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	
	OUT.HPosition.z += 0.001;

	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	if(@_TRANSPARENT && @_DISSOLVING)
	{
		return float4(cParamter.rgb, cParamter2.w);// 非常蛋疼的定制，因为又要dissolving，又要轮廓，而两者Param2是冲突的，所以如果有这种冲突情况，则替换为cParamter.rgb的颜色设置
	}
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

#endif // __FILE_SHADOWGEN_FXH__
