
#ifndef __FILE_TERRAIN_COMMON_FXH__
#define __FILE_TERRAIN_COMMON_FXH__

#define USE_REFINED_BLEND		1

#ifndef TERRAIN_BAKE
	#define TERRAIN_BAKE 0
#endif

#if TERRAIN_BAKE
	float ComputeFarFactor(float level)
	{
		if(level < 4)
			return 1;
		else if(level < 5)
			return 0.5;
		else
			return 0.2;
	}
#endif

void PS_Compute(inout int3 index, inout half3 blend, int i, half w)
{
	index[1] += (i - 257) * (int)(index[1] > 256) * (int)(i != index[0]) * (int)(w > 0.0);
	index[2] += (i - 257) * (int)(index[2] > 256) * (int)(i != index[0]) * (int)(i != index[1]) * (int)(w > 0.0);
	blend[1] += w * (half)(i == index[1]);
	blend[2] += w * (half)(i == index[2]);
}

void RefinedBlend(in float2 uv, out half3 blend, out int3 index)
{
	float2 uvbase = (floor(uv * cBlendMapSize.x - 0.5) + 0.5) * cBlendMapSize.y;
	float2 uvfrac = (uv - uvbase) * cBlendMapSize.x;
	float alpha00 = (1 - uvfrac.x) * (1 - uvfrac.y);
	float alpha01 = (1 - uvfrac.x) * (uvfrac.y);
	float alpha10 = (uvfrac.x) * (1 - uvfrac.y);
	float alpha11 = (uvfrac.x) * (uvfrac.y);
	half3 color00 = tIndexMap.Sample(sIndexMapSampler, uvbase).xyz;
	half3 color01 = tIndexMap.Sample(sIndexMapSampler, uvbase + float2(0, cBlendMapSize.y)).xyz;
	half3 color10 = tIndexMap.Sample(sIndexMapSampler, uvbase + float2(cBlendMapSize.y, 0)).xyz;
	half3 color11 = tIndexMap.Sample(sIndexMapSampler, uvbase + float2(cBlendMapSize.y, cBlendMapSize.y)).xyz;
	int2 index00 = (int2)floor(color00.xy * 256);
	int2 index01 = (int2)floor(color01.xy * 256);
	int2 index10 = (int2)floor(color10.xy * 256);
	int2 index11 = (int2)floor(color11.xy * 256);
	blend = half3(0, 0, 0);

	bool sameindex_x = (index01.x == index00.x) && (index10.x == index00.x) && (index11.x == index00.x);
	bool sameindex_y = (index01.y == index00.y) && (index10.y == index00.y) && (index11.y == index00.y);
	if (sameindex_x && sameindex_y)
	{
		index = int3(index00.x, index00.y, 0);
		blend[0] = color00.z * alpha00 + color01.z * alpha01 + color10.z * alpha10 + color11.z * alpha11;
		blend[1] = 1 - blend[0];
	}
	else
	{
		index = int3(index00.x, 257, 257);
		PS_Compute(index, blend, index00.y, alpha00 * (1 - color00.z));
		PS_Compute(index, blend, index01.x, alpha01 * (color01.z));
		PS_Compute(index, blend, index01.y, alpha01 * (1 - color01.z));
		PS_Compute(index, blend, index10.x, alpha10 * (color10.z));
		PS_Compute(index, blend, index10.y, alpha10 * (1 - color10.z));
		PS_Compute(index, blend, index11.x, alpha11 * (color11.z));
		PS_Compute(index, blend, index11.y, alpha11 * (1 - color11.z));
		index[1] = index[1] * (int)(index[1] < 257);
		index[2] = index[2] * (int)(index[2] < 257);
		blend[0] = 1 - blend[1] - blend[2];
	}
}

void NormalBlend(in float2 uv, out half3 blend, out int3 index)
{
	half2 color = tIndexMap.Sample(sIndexMapSampler, uv).xy;
	index.xy = (int2)floor(color * 256);
	index.z = 0;
	blend[0] = tBlendMap.Sample(sBlendMapSampler, uv).z;
	blend[1] = 1 - blend[0];
	blend[2] = 0;
}

half ComputeCliff(in float3 world_normal, out float3 w)
{
	// ensures that exactly 1 bit of w equals 1, w = ((r.x>r.y)*(r.x>r.z),(r.y>=r.z)*(r.y>=r.x),(r.z>=r.x)*(r.z>r.y)). 
	float3 r = abs(world_normal);
	float3 s = step(r.xzx, r.yyz); s.x = 1.0 - s.x;
	w = s * (1.0 - s.zxy);
	// make sure all faces look the same but not as in a mirror. 
	half dir = sign(dot(world_normal, half3(w.x, w.y, -w.z)));
	return dir;
}

void ComputeTexCoordCliff(in float3 t, in half dir, in float3 w, in float dd_factor, out float2 tc, out float2 dx, out float2 dy)
{
	half3 wc = cBlockShrinkage.y * frac(t * half3(dir, 1, dir));
	tc = float2(dot(wc.zxx, w), dot(wc.yzy, w)) + cBlockShrinkage.x;
	float3 dcx = ddx(t) * dd_factor;
	float3 dcy = ddy(t) * dd_factor;
	dx = cBlockShrinkage.y * min(abs(float2(dot(dcx.zxx, w), dot(dcx.yzy, w))), 0.125);
	dy = cBlockShrinkage.y * min(abs(float2(dot(dcy.zxx, w), dot(dcy.yzy, w))), 0.125);
}

void ComputeTexCoord(in float2 t, in float dd_factor, out float2 tc, out float2 dx, out float2 dy)
{
	tc = cBlockShrinkage.y * frac(t) + cBlockShrinkage.x;
	dx = cBlockShrinkage.y * min(abs(ddx(t) * dd_factor), 0.125);
	dy = cBlockShrinkage.y * min(abs(ddy(t) * dd_factor), 0.125);
}

void ComptueAllTexCoord(in float3 world_normal, in float3 world_position, 
	in float dd_factor, in int3 x, in int3 y, in float far_factor,
	out float2 tc0, out float2 dx0, out float2 dy0,
	out float2 tc0Far, out float2 dx0Far, out float2 dy0Far,
	out float2 tc1, out float2 dx1, out float2 dy1,
	out float2 tc2, out float2 dx2, out float2 dy2)
{
	float4x4 cBlockScaleList = float4x4(
	  cBlockScaleList0[0], cBlockScaleList0[1], cBlockScaleList0[2], cBlockScaleList0[3],
	  cBlockScaleList1[0], cBlockScaleList1[1], cBlockScaleList1[2], cBlockScaleList1[3],
	  cBlockScaleList2[0], cBlockScaleList2[1], cBlockScaleList2[2], cBlockScaleList2[3],
	  cBlockScaleList3[0], cBlockScaleList3[1], cBlockScaleList3[2], cBlockScaleList3[3]);

	if (@_EnableCliff)
	{
		float3 w;
		half dir = ComputeCliff(world_normal, w);

		float3 t0 = cBlockScaleList[y[0]][x[0]] * world_position;
		ComputeTexCoordCliff(t0, dir, w, dd_factor, tc0, dx0, dy0);

		//farBlockTile
		float3 t0Far = cBlockScaleList[y[0]][x[0]] * world_position * far_factor;
		ComputeTexCoordCliff(t0Far, dir, w, dd_factor, tc0Far, dx0Far, dy0Far);

		float3 t1 = cBlockScaleList[y[1]][x[1]] * world_position;
		ComputeTexCoordCliff(t1, dir, w, dd_factor, tc1, dx1, dy1);

#if USE_REFINED_BLEND
		float3 t2 = cBlockScaleList[y[2]][x[2]] * world_position;
		ComputeTexCoordCliff(t2, dir, w, dd_factor, tc2, dx2, dy2);
#endif
	}
	else
	{
		float2 t0 = cBlockScaleList[y[0]][x[0]] * world_position.xz;
		ComputeTexCoord(t0, dd_factor, tc0, dx0, dy0);

		//farBlockTile
		float2 t0Far = cBlockScaleList[y[0]][x[0]] * world_position.xz * far_factor;
		ComputeTexCoord(t0Far, dd_factor, tc0Far, dx0Far, dy0Far);

		float2 t1 = cBlockScaleList[y[1]][x[1]] * world_position.xz;
		ComputeTexCoord(t1, dd_factor, tc1, dx1, dy1);
#if USE_REFINED_BLEND
		float2 t2 = cBlockScaleList[y[2]][x[2]] * world_position.xz;
		ComputeTexCoord(t2, dd_factor, tc2, dx2, dy2);
#endif
	}
	float dl0 = max(dx0.x + dy0.x, dx0.y + dy0.y);
	tc0.x = clamp(tc0.x, dl0, cBlockShrinkage.z - dl0);
	tc0.y = clamp(tc0.y, dl0, cBlockShrinkage.z - dl0);
	tc0 = tc0 + cBlockShrinkage.z * float2(x[0], y[0]);

	float dl0Far = max(dx0Far.x + dy0Far.x, dx0Far.y + dy0Far.y);
	tc0Far.x = clamp(tc0Far.x, dl0Far, cBlockShrinkage.z - dl0Far);
	tc0Far.y = clamp(tc0Far.y, dl0Far, cBlockShrinkage.z - dl0Far);
	tc0Far = tc0Far + cBlockShrinkage.z * float2(x[0], y[0]);

	float dl1 = max(dx1.x + dy1.x, dx1.y + dy1.y);
	tc1.x = clamp(tc1.x, dl1, cBlockShrinkage.z - dl1);
	tc1.y = clamp(tc1.y, dl1, cBlockShrinkage.z - dl1);
	tc1 = tc1 + cBlockShrinkage.z * float2(x[1], y[1]);

#if USE_REFINED_BLEND
	float dl2 = max(dx2.x + dy2.x, dx2.y + dy2.y);
	tc2.x = clamp(tc2.x, dl2, cBlockShrinkage.z - dl2);
	tc2.y = clamp(tc2.y, dl2, cBlockShrinkage.z - dl2);
	tc2 = tc2 + cBlockShrinkage.z * float2(x[2], y[2]);
#endif
}

// index & blend.
void PS_IndexBlend(in float3 world_normal, in float3 world_position, in float2 uv, in half rate, inout half3 diff, inout half3 norm, 
	inout half height, inout half3 mixx, in float dd_factor, in float far_factor)
{
	half3 blend;
	int3 index;
#if USE_REFINED_BLEND
	RefinedBlend(uv, blend, index);
#else
	NormalBlend(uv, blend, index);
#endif

	// diff & norm & mixx.
	half3 diff0 = half3(0, 0, 0);
	half3 diff0Far = half3(0, 0, 0);
	half4 norm0 = half4(0.5,0.5,1,1);
	half4 norm0Far = half4(0, 0, 0, 0);
	half3 mixx0 = half3(1, 1, 1);
	half3 diff1 = half3(0, 0, 0);
	half4 norm1 = half4(0.5,0.5,1,1);
	half3 mixx1 = half3(1, 1, 1);
#if USE_REFINED_BLEND
	half4 norm2 = half4(0.5,0.5,1,1);
	half3 diff2 = half3(0, 0, 0);
	half3 mixx2 = half3(1, 1, 1);
#endif

	if(@_SurfaceTextureArrayEnable)
	{
		float2 surface_uv = cBlockScale.x * world_position.xz;
		float2 d_surface_uv_x = ddx(surface_uv) * dd_factor;
		float2 d_surface_uv_y = ddy(surface_uv) * dd_factor;
		
		if (blend[0] > 0) 
		{
			float2 far_surface_uv = surface_uv * far_factor;
			float2 d_far_surface_uv_x = ddx(far_surface_uv) * dd_factor;
			float2 d_far_surface_uv_y = ddy(far_surface_uv) * dd_factor;
#if TERRAIN_BAKE
			diff0 = tSurfaceBlockMap.SampleGrad(sSurfaceBlockSampler, float3(far_surface_uv, index[0]), d_far_surface_uv_x, d_far_surface_uv_y).xyz;
			norm0 = tSurfaceNormalMap.SampleGrad(sSurfaceNormalSampler, float3(far_surface_uv, index[0]), d_far_surface_uv_x, d_far_surface_uv_y);
#else
			diff0Far = tSurfaceBlockMap.SampleGrad(sSurfaceBlockSampler, float3(far_surface_uv , index[0]), d_far_surface_uv_x, d_far_surface_uv_y).xyz;
			norm0Far = tSurfaceNormalMap.SampleGrad(sSurfaceNormalSampler, float3(far_surface_uv, index[0]), d_far_surface_uv_x, d_far_surface_uv_y);
		
				diff0 = tSurfaceBlockMap.SampleGrad(sSurfaceBlockSampler, float3(surface_uv, index[0]), d_surface_uv_x, d_surface_uv_y).xyz;
				if(@_LOD_LEVEL <= LOD_Level_2)
			{
				norm0 = tSurfaceNormalMap.SampleGrad(sSurfaceNormalSampler, float3(surface_uv, index[0]), d_surface_uv_x, d_surface_uv_y);
				if(@_LOD_LEVEL <= LOD_Level_1)
				{
					diff0 = lerp(diff0Far, diff0, rate);
					norm0 = lerp(norm0Far, norm0, rate);
				}
							
			}
			// else
			// {
			// 	// diff0 = tSurfaceBlockMap.SampleGrad(sSurfaceBlockSampler, float3(surface_uv, index[0]), d_surface_uv_x, d_surface_uv_y).xyz;;
			// 	norm0 =  half4(0.5,0.5,1,1);
			// }
#endif
			mixx0 = tSurfaceMixtureMap.SampleGrad(sSurfaceMixtureSampler, float3(surface_uv, index[0]), d_surface_uv_x, d_surface_uv_y).xyz;
		}
		if (blend[1] > 0)
		{
			diff1 = tSurfaceBlockMap.SampleGrad(sSurfaceBlockSampler, float3(surface_uv, index[1]), d_surface_uv_x, d_surface_uv_y).xyz;
			if(@_LOD_LEVEL <= LOD_Level_2)
			{
			norm1 = tSurfaceNormalMap.SampleGrad(sSurfaceNormalSampler, float3(surface_uv, index[1]), d_surface_uv_x, d_surface_uv_y);
			if(@_LOD_LEVEL<=LOD_Level_1)
			{
			mixx1 = tSurfaceMixtureMap.SampleGrad(sSurfaceMixtureSampler, float3(surface_uv, index[1]), d_surface_uv_x, d_surface_uv_y).xyz;	
			}
			
			}
			
		}
#if USE_REFINED_BLEND
		if (blend[2] > 0) 
		{
			diff2 = tSurfaceBlockMap.SampleGrad(sSurfaceBlockSampler, float3(surface_uv, index[2]), d_surface_uv_x, d_surface_uv_y).xyz;
			if(@_LOD_LEVEL<=LOD_Level_2)
			{
			norm2 = tSurfaceNormalMap.SampleGrad(sSurfaceNormalSampler, float3(surface_uv, index[2]), d_surface_uv_x, d_surface_uv_y);
			if(@_LOD_LEVEL<=LOD_Level_1)
			{
			mixx2 = tSurfaceMixtureMap.SampleGrad(sSurfaceMixtureSampler, float3(surface_uv, index[2]), d_surface_uv_x, d_surface_uv_y).xyz;	
			}
			
			}
		
			
		}
#endif
	}
	else
	{
		int3 x, y;
		{
#if USE_REFINED_BLEND
			float3 f = (float3)index;
			y = (int3)floor(f / 16);
			x = (int3)(f - 16 * y);
#else
			float2 f = (float2)index.xy;
			y.xy = (int2)floor(f / 16);
			x.xy = (int2)(f - 16 * y);
			y.z = x.z = 0;
#endif
		}

		// texture coordinates.
		float2 dx0 = 0.0;
		float2 dy0 = 0.0;
		float2 dx1 = 0.0;
		float2 dy1 = 0.0;
		float2 dx2 = 0.0;
		float2 dy2 = 0.0;
		float2 tc0 = float2(0, 0);
		float2 tc1 = float2(0, 0);
		float2 tc2 = float2(0, 0);
		float2 dx0Far = 0.0;
		float2 dy0Far = 0.0;
		float2 tc0Far = float2(0, 0);

		ComptueAllTexCoord(world_normal, world_position, dd_factor, x, y, far_factor,
			tc0, dx0, dy0, tc0Far, dx0Far, dy0Far, tc1, dx1, dy1, tc2, dx2, dy2);

		if (@_BlockMapEnable)
		{
			if (blend[0] > 0)
			{
#if TERRAIN_BAKE
				diff0 = tBlockMap.SampleGrad(sBlockMapSampler, tc0Far, dx0Far, dy0Far).xyz;
#else
				diff0Far = tBlockMap.SampleGrad(sBlockMapSampler, tc0Far, dx0Far, dy0Far).xyz;
				if(@_LOD_LEVEL <= LOD_Level_1)
				{
					diff0 = tBlockMap.SampleGrad(sBlockMapSampler, tc0, dx0, dy0).xyz;
					diff0 = lerp(diff0Far, diff0, rate);
				}
				else
				{
					diff0 = tBlockMap.SampleGrad(sBlockMapSampler, tc0, dx0, dy0).xyz;
				}
#endif
			}
			if (blend[1] > 0) diff1 = tBlockMap.SampleGrad(sBlockMapSampler, tc1, dx1, dy1).xyz;
#if USE_REFINED_BLEND
			if (blend[2] > 0) diff2 = tBlockMap.SampleGrad(sBlockMapSampler, tc2, dx2, dy2).xyz;
#endif
		}
		if (@_NormalMapEnable)
		{
			if (blend[0] > 0)
			{
#if TERRAIN_BAKE
				norm0 = tNormalMap.SampleGrad(sNormalMapSampler, tc0Far, dx0Far, dy0Far);
#else
				norm0Far = tNormalMap.SampleGrad(sNormalMapSampler, tc0Far, dx0Far, dy0Far);
				if(@_LOD_LEVEL <= LOD_Level_1)
				{
					norm0 = tNormalMap.SampleGrad(sNormalMapSampler, tc0, dx0, dy0);
					norm0 = lerp(norm0Far, norm0, rate);
				}
				// else
				// {
				// 	norm0 =  half4(0,0,1,1);
				// }
#endif
				norm0.x = cheapContrast(norm0.x, 0.5);
				norm0.y = cheapContrast(norm0.y, 0.5);
			} 
			if (blend[1] > 0) norm1 = tNormalMap.SampleGrad(sNormalMapSampler, tc1, dx1, dy1);
#if USE_REFINED_BLEND
			if (blend[2] > 0) norm2 = tNormalMap.SampleGrad(sNormalMapSampler, tc2, dx2, dy2);
#endif
		}
		if (TERRAIN_BAKE || (@_LOD_LEVEL <= LOD_Level_1 && @_MixtureMapEnable))
		{
			if (blend[0] > 0) mixx0 = tMixtureMap.SampleGrad(sMixtureMapSampler, tc0, dx0, dy0).xyz;
			if (blend[1] > 0) mixx1 = tMixtureMap.SampleGrad(sMixtureMapSampler, tc1, dx1, dy1).xyz;
#if USE_REFINED_BLEND
			if (blend[2] > 0) mixx2 = tMixtureMap.SampleGrad(sMixtureMapSampler, tc2, dx2, dy2).xyz;
#endif
		}
	}

	// height
	float depth = max(cBlendSharpness.x, 0.01);
	depth =depth;
	norm0.a *= blend[0];
	norm1.a *= blend[1];
#if USE_REFINED_BLEND
	norm2.a *= blend[2];
	float ma = max(max(blend[0] + norm0.a, blend[1] + norm1.a), blend[2] + norm2.a) - depth;
#else
	float ma = max(blend[0] + norm0.a, blend[1] + norm1.a) - depth;
#endif
	blend[0] = max(blend[0] + norm0.a - ma, 0);
	blend[1] = max(blend[1] + norm1.a - ma, 0);
#if USE_REFINED_BLEND
	blend[2] = max(blend[2] + norm2.a - ma, 0);
	blend = blend / (blend[0] + blend[1] + blend[2]);
#else
	blend = blend / (blend[0] + blend[1]);
#endif

#if USE_REFINED_BLEND
	diff = blend[0] * diff0 + blend[1] * diff1 + blend[2] * diff2;
	norm = blend[0] * norm0.xyz + blend[1] * norm1.xyz + blend[2] * norm2.xyz;
	height = blend[0] * norm0.w + blend[1] * norm1.w + blend[2] * norm2.w;
	mixx = blend[0] * mixx0 + blend[1] * mixx1 + blend[2] * mixx2;
#else
	diff = blend[0] * diff0 + blend[1] * diff1;
	norm = blend[0] * norm0.xyz + blend[1] * norm1.xyz;
	height = blend[0] * norm0.w + blend[1] * norm1.w;
	mixx = blend[0] * mixx0 + blend[1] * mixx1;
#endif
}
#endif
