#ifndef __FILE_TESSELLATION_FXH__
#define __FILE_TESSELLATION_FXH__
struct HS_CONSTANT_DATA_OUTPUT
{
	float EdgeTessFactor[3]			: SV_TessFactor;
	float InsideTessFactor			: SV_InsideTessFactor;
};

// Patch Constant Function
HS_CONSTANT_DATA_OUTPUT CalcHSPatchConstants(
	InputPatch<VertexShadingOutput, 3> ip,
	uint PatchID : SV_PrimitiveID)
{
	HS_CONSTANT_DATA_OUTPUT Output;

	float4 tessFactors = HS_GetTessellationFactor(ip[0], ip[1], ip[2]);
	Output.EdgeTessFactor[0] = tessFactors.x;
	Output.EdgeTessFactor[1] = tessFactors.y;
	Output.EdgeTessFactor[2] = tessFactors.z;
	Output.InsideTessFactor = tessFactors.w;

	return Output;
}

[domain("tri")]
[partitioning("fractional_odd")]
[outputtopology("triangle_cw")]
[outputcontrolpoints(3)]
[patchconstantfunc("CalcHSPatchConstants")]
VertexShadingOutput PbrVertexShading_HS(
	InputPatch<VertexShadingOutput, 3> ip,
	uint i : SV_OutputControlPointID,
	uint PatchID : SV_PrimitiveID)
{
	return ip[i];
}

[domain("tri")]
VertexShadingOutput PbrVertexShading_DS(
	HS_CONSTANT_DATA_OUTPUT input,
	const OutputPatch<VertexShadingOutput, 3> op,
	float3 domain : SV_DomainLocation)
{
	
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

#define INTERPOLATE(field) OUT.field = op[0].field * domain.x + op[1].field * domain.y + op[2].field * domain.z

	INTERPOLATE(TexCoord);
	INTERPOLATE(WorldPosition);
	INTERPOLATE(WorldNormal);
	INTERPOLATE(ViewDirection.w);
	INTERPOLATE(DiffLighting);
	INTERPOLATE(LinearZ);

#if !IS_TERRAIN
	if (@_TANGENT)
#endif	
	{
		INTERPOLATE(WorldTangent);
		INTERPOLATE(WorldBinormal);
	}

	if (@_CastSunShadow)
	{
		INTERPOLATE(LitUVZ);
	}

	if (@_VERTEXCOLOR)
	{
		INTERPOLATE(TintColor);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}

	// Phone tessellation
	DS_DIY_OUTPUT(op[0], op[1], op[2], domain, OUT);

	// recalc ViewDirection & HPosition
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	
	[@_CastSunShadow]
	OUT.LitUVZ = MulMat4x4(ShadowViewProjTexs1, float4(OUT.WorldPosition.xyz,1));
	return OUT;
}

#endif // __FILE_TESSELLATION_FXH__
