#ifndef __FILE_VBUFFERGEN_FXH__
#define __FILE_VBUFFERGEN_FXH__
/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif
	OUT.HPosition = MulMat4x4(WorldViewProj, INTM.LocalPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	VertexUnifySkinningTex(IN, INTM2);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM2);
#endif	
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.LastTex.w = max(OUT.LastTex.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
		OUT.LastTex.z = min(OUT.LastTex.z, (1-OUT.LastTex.w*CameraInfo.z)*OUT.LastTex.w);
	}

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		float alpha = tBaseMap.SampleRGBA( sBaseSampler, IN.TC ).a;
		clip(alpha - ALPHA_REF);
	}

	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	currUV -= cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}
#endif // __FILE_VBUFFERGEN_FXH__
