
#ifndef __FILE_VT_COMMON_FXH__
#define __FILE_VT_COMMON_FXH__

void SampleVTIndInLevel(float level, float2 tc, out float2 vtBlockUV, out float2 vtUV)
{
	float totalSizeInLevelBlock;
	float2 levelCenter;
	if(level < VTLodLayerCount - 1)
	{
		totalSizeInLevelBlock = VTTerrainSizeInBlock / pow(2, level);
		levelCenter = round(VTViewpointInTerrain * totalSizeInLevelBlock) / totalSizeInLevelBlock;
	}
	else
	{
		level = VTLodLayerCount - 1;
		totalSizeInLevelBlock = VTLastLayerSizeInBlock;
		levelCenter = float2(0.5, 0.5) * (VTIndirectionLayerSizeInBlock / VTLastLayerSizeInBlock);
	}
	float2 levelUV = (tc - levelCenter) * totalSizeInLevelBlock + VTIndirectionLayerSizeInBlock * 0.5;
	float2 indUV = (floor(levelUV) + 0.5) / VTIndirectionLayerSizeInBlock;
	indUV.y = (level + indUV.y) / VTLodLayerCount;
	float3 indirectionIndex = tVTIndMap.Sample(sVTIndSampler, indUV).xyz;
	float2 vtStart = round(255 * indirectionIndex.xy) * float2(VTBlockSizeInUV, VTBlockSizeInUV);
	float realLevel = round(255 * indirectionIndex.z);
	if(realLevel > level)
	{
		if(realLevel < VTLodLayerCount - 1)
		{
			totalSizeInLevelBlock = VTTerrainSizeInBlock / pow(2, realLevel);
		}
		else
		{
			totalSizeInLevelBlock = VTLastLayerSizeInBlock;
		}
	}
	vtBlockUV = frac(tc * totalSizeInLevelBlock);
	float2 vtOffset = vtBlockUV * VTBlockContentSizeInUV + VTBlockMarginInUV;
	vtUV = vtStart + vtOffset;
}

void SampleVTInLevel(float level, float2 tc, out float2 vtBlockUV, out half3 diff, out half3 norm, out half3 mixx)
{
	float2 vtUV;
	SampleVTIndInLevel(level, tc, vtBlockUV, vtUV);
	half4 base_normalx = tVTBaseNormalXMap.Sample(sVTBaseNormalXSampler, vtUV);
	half3 normal = half3(0,0,1);
	diff = base_normalx.rgb;
	if(@_LOD_LEVEL <= LOD_Level_2)
	{
		half4 roughness_ao_height_normaly = tVTRoughnessAOHeightNormalYMap.Sample(sVTRoughnessAOHeightNormalYSampler, vtUV);
		normal.xy = float2(base_normalx.w, roughness_ao_height_normaly.w) * 2 - 1;
		normal.z = sqrt(saturate(1 - dot(normal.xy, normal.xy)));
		mixx = half3(roughness_ao_height_normaly.x, 0, roughness_ao_height_normaly.y);
	}
	else
	{
		mixx = half3(0,0,1);

	}
	
	
	
	
	
	norm = (normal + 1) * 0.5;
	
}

void SampleVTInLevelOnlyDiff(float level, float2 tc, out float2 vtBlockUV, out half3 diff)
{
	float2 vtUV;
	SampleVTIndInLevel(level, tc, vtBlockUV, vtUV);
	half4 base_normalx = tVTBaseNormalXMap.Sample(sVTBaseNormalXSampler, vtUV);
	diff = base_normalx.rgb;
}

void GetVTSampleParams(float3 worldPos, out float2 tc, out float distInBlock, out float screenMipLevel)
{
	tc = worldPos.xz * VTIndMapUVTransform.xy + float2(VTIndMapUVTransform.z, VTIndMapUVTransform.w);
	float dist = max(abs(tc.x - VTViewpointInTerrain.x), abs(tc.y - VTViewpointInTerrain.y)) * VTTerrainSizeInBlock;
	distInBlock = dist;
	float2 tc_abs_ddx = abs(ddx(tc));
	float2 tc_abs_ddy = abs(ddy(tc));
	screenMipLevel = max(0, log2(max(max(tc_abs_ddx.x, tc_abs_ddx.y), max(tc_abs_ddy.x, tc_abs_ddy.y)) * VTBlockContentSizeInPixel * VTTerrainSizeInBlock));
}

void SampleVT(float3 worldPos, out half3 diff, out half3 norm, out half3 mixx)
{
	float2 tc;
	float distInBlock;
	float screenMipLevel;
	GetVTSampleParams(worldPos, tc, distInBlock, screenMipLevel);
	float levelRatio = 0;
	if(distInBlock > (VTIndirectionLayerSizeInBlock * 0.5 - 0.5) * 0.5)
	{
		levelRatio = max(0, log2(distInBlock / (VTIndirectionLayerSizeInBlock * 0.5 - 0.5)) + 1);
	}
	//levelRatio = max(levelRatio, screenMipLevel);
	float level = floor(levelRatio);
	float2 vtBlockUV;
	
	SampleVTInLevel(level, tc, vtBlockUV, diff, norm, mixx);

	if(levelRatio - level > (1 - VTBlendRange))
	{
		float blendRatio = saturate((levelRatio - level - (1 - VTBlendRange)) / VTBlendRange);
		float2 vtBlockUVNext;
		float nextLevel = level + 1;
		half3 nextDiff, nextNorm, nextMixx;
		SampleVTInLevel(nextLevel, tc, vtBlockUVNext, nextDiff, nextNorm, nextMixx);
		diff = lerp(diff, nextDiff, blendRatio);
		norm = lerp(norm, nextNorm, blendRatio);
		mixx = lerp(mixx, nextMixx, blendRatio);
	}
	
	/*
	// if (@_VirtualTextureDebugEnable)
	{
		norm = half3(0.5, 0.5, 1);
		mixx = half3(0, 0, 1);
		if(abs(vtBlockUV.x - 0.5) > 0.475 || abs(vtBlockUV.y - 0.5) > 0.475)
		{
			diff = float3(1, 1, 1);
		}
		else if((floor(vtBlockUV.x * 10) + floor(vtBlockUV.y * 10)) % 2 == 1)
		{
			float levelValue = level / (VTLodLayerCount - 1);
			float3 levelColor = lerp(float3(1, 0, 0), lerp(float3(0, 1, 0), float3(0, 0, 1), max(0, levelValue * 2 - 1)), min(1, levelValue * 2));
			diff = levelColor;
		}
		else
		{
			diff = float3(0, 0, 0);
		}
	}
	*/
}

void SampleVTOnlyDiff(float3 worldPos, out half3 diff)
{
	float2 tc;
	float distInBlock;
	float screenMipLevel;
	GetVTSampleParams(worldPos, tc, distInBlock, screenMipLevel);
	float levelRatio = 0;
	if(distInBlock > (VTIndirectionLayerSizeInBlock * 0.5 - 0.5) * 0.5)
	{
		levelRatio = max(0, log2(distInBlock / (VTIndirectionLayerSizeInBlock * 0.5 - 0.5)) + 1);
	}
	levelRatio = max(levelRatio, screenMipLevel);
	float level = floor(levelRatio);
	float2 vtBlockUV;
	
	SampleVTInLevelOnlyDiff(level, tc, vtBlockUV, diff);
}

half SampleVTHeight(float3 worldPos)
{
	float2 tc;
	float distInBlock;
	float screenMipLevel;
	GetVTSampleParams(worldPos, tc, distInBlock, screenMipLevel);
	float levelRatio = 0;
	if(distInBlock > (VTIndirectionLayerSizeInBlock * 0.5 - 0.5) * 0.5)
	{
		levelRatio = max(0, log2(distInBlock / (VTIndirectionLayerSizeInBlock * 0.5 - 0.5)) + 1);
	}
	//levelRatio = max(levelRatio, screenMipLevel);
	float level = floor(levelRatio);

	float2 vtBlockUV;
	float2 vtUV;
	SampleVTIndInLevel(level, tc, vtBlockUV, vtUV);
	half4 roughness_ao_height_normaly = tVTRoughnessAOHeightNormalYMap.Sample(sVTRoughnessAOHeightNormalYSampler, vtUV);
	return roughness_ao_height_normaly.z;
}

half SampleVTBlendFactor(float3 worldPos, float blendHeight)
{
	float2 offset = worldPos.xz - VTViewpointInWorld;
	float2 uv = offset / (LocalHeightmapSize - 1) + 0.5;
	float terrain_height = lerp(LocalHeightmapMinHeight, LocalHeightmapMaxHeight, tLocalHeightmap.Sample(sLocalHeightmapSampler, uv).r);
	half result = saturate((terrain_height + blendHeight - worldPos.y) / blendHeight);
	float max_dist = (LocalHeightmapSize - 1) * 0.5;
	result *= 1 - smoothstep(max_dist - 5, max_dist, length(offset));
	return result;
}

float3 VTPOM(in float3 worldPos, in float3 viewDirTan, in float3 viewDirWorld, in float3 vertexNormal)
{
	//fresnel
	float fresnel = saturate(dot(viewDirWorld, vertexNormal));
	//fresnelScale
	float fresnelScale = saturate(fresnel * 4);
	//viewLength
	float heightScale = 0.25;
	float2 viewLength = viewDirTan.xy / viewDirTan.z * fresnelScale * heightScale;
	viewLength.x = -viewLength.x;
	//numSteps
	float minSteps = 8;
	float maxStep = 16;
	float smoothSteps = lerp(maxStep, minSteps, fresnel);
	float stepSize = 1 / smoothSteps;
	float numSteps = floor(smoothSteps);
	//uv
	float refPlane = -0.5;
	worldPos.xz = worldPos.xz + viewLength * refPlane;
	float2 UVDist = viewLength * stepSize;
	//rayMarching
	float rayHeight = 1;
	float oldRay = 1;
	float2 offset = 0;
	float oldTex = 1;
	float heightMap = 1;
	if(fresnelScale > 0.1)
	{
		for( int i = 0; i < numSteps + 2; i++ )
		{
			heightMap = SampleVTHeight(worldPos + float3(offset.x, 0, offset.y));
			if (rayHeight < heightMap)
			{
				float xintersect = (oldRay - oldTex) + (heightMap - rayHeight);
				xintersect = (heightMap - rayHeight) / xintersect;
				offset -= (xintersect * UVDist);
				break;
			}
			oldRay = rayHeight;
			rayHeight -= stepSize;
			offset += UVDist;
			oldTex = heightMap;
		}
	}
	worldPos.xz += offset;

	return worldPos;
}

#endif
