
void VertexUnifyOutput(in VertexAssemble IN, inout InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	// TexCoord
	OUT.TexCoord = IN.TexCoord;

	//world * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput VolumetricVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);
	OUT.DepthInView = OUT.HPosition.w / CameraInfo.y;

	//disable farplane culling
	OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

	return OUT;
}

half3 LightingVolumetric(in VertexShadingOutput IN, in float3 P, in InterMediatePixel IMP)
{
	half3 DiffLit = SunColor.rgb + IMP.IndirectLighing.rgb;
#if SD_TILE_BASED_LIGHTING
	if (@_DEFERRED_SHADING)
#endif
	{
		half3 N = half3(1,1,1);
		if( @_LitCount >= 1)
		{
			PointLighting(P, N, Lights0, DiffLit);
		}
		if( @_LitCount >= 2)
		{
			PointLighting( P, N, Lights1, DiffLit);
		}
		if( @_LitCount >= 3)
		{
			PointLighting( P, N, Lights2, DiffLit);
		}
	}
#if SD_TILE_BASED_LIGHTING
	else
	{
		TileBasedLightingPS(IN.HPosition.xy, P, float3(1,1,1), DiffLit);
	}
#endif
	return DiffLit;
}

float4 VolumetricFogLightingPS(in VertexShadingOutput IN) : SV_Target
{
	//material parameters
	InterMediatePixel IMP;
	PS_GetParameters(IN, IMP);

    //OUT.rgb = Distance * ∫[0-1] Tr(startPos->Pos_t) * [sun intensity * Phase(theta) * VolumetricShadow(Pos_t) + IndirectLighing] * ScatteringCoefficient dt
	float3 OutScattering = float3(0, 0, 0);
    float OutAlpha = 1.0;
	float stepDistance = IMP.RaymarchingDistance / IMP.NSamples;
	float3 stepPos = (IMP.EndPos-IMP.StartPos) / IMP.NSamples;
	float3 iterPos = IMP.StartPos + 0.5 * stepPos;
	float3 ConstantValue = IMP.RaymarchingDistance;// * IsotropicPhase();
	[loop] for (int i = 0; i < IMP.NSamples; ++i)
	{
        float ExtinctionCoefficent;
        float3 ScatteringCoefficient;
        GetVolumetricCoefficient(iterPos, ScatteringCoefficient, ExtinctionCoefficent);
		float T = exp(-ExtinctionCoefficent*stepDistance);
        OutAlpha *= T;
        OutScattering += ScatteringCoefficient * LightingVolumetric(IN, iterPos, IMP) * OutAlpha;
		iterPos += stepPos;
	}
	OutScattering = OutScattering * ConstantValue / IMP.NSamples;
	return float4(OutScattering, OutAlpha);
}

technique Shading
{
	VertexShader = VolumetricVertexShading();
	PixelShader = VolumetricFogLightingPS();
}


//////////////////////////////////////////technique FogVolumeDepth////////////////////////////////////
float4 FogVolumeDepthPS(in VertexShadingOutput IN) : SV_Target
{
	return float4(IN.DepthInView.rrrr);
}

technique FogVolumeDepth
{
	VertexShader = VolumetricVertexShading();
	PixelShader = FogVolumeDepthPS();
}

