//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 					;
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
	float4 cPlaceHold											;
#endif	
	float4 cWBasisX					: WBasisX					;
	float4 cWBasisY					: WBasisY					;
	float4 cWBasisZ					: WBasisZ					;
	float4 cMinBB					: BBMin						;
	float4 cMaxBB					: BBMax						;
	float4x4 WorldViewProj			: WorldViewProj				;
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
};

struct InterMediatePixel
{
	half3 IndirectLighing;
	float3 StartPos;
	float3 EndPos;
	float RaymarchingDistance;
	float NSamples;
};

void InitInterMediatePixel(inout InterMediatePixel IMP)
{
	IMP.IndirectLighing = float3(0, 0, 0);
	IMP.StartPos = float3(0, 0, 0);
	IMP.EndPos = float3(0, 1, 0);
	IMP.RaymarchingDistance = 1;
	IMP.NSamples = 32;
}

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	[@_HLSL]
	INTM = (InterMediateVertex)0;

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
		}
	}
}

//https://tavianator.com/fast-branchless-raybounding-box-intersections/
float2 IntersctionAABB(in float3 Org, in float3 Dir, in float3 minXYZ, in float3 maxXYZ)
{
	//change (x - a)/b to x * c - d, make it fast, but will lead NaN (if D.x = INF, INF + INF = NAN)
	//D.x = 0, if Org.x in [minXYZ.x, maxXYZ.x]: has intersection, t1=-INF, t2=+INF
	//D.x = 0, if Org.x not in [minXYZ.x, maxXYZ.x]: no intersection, t1,t2 has same value, +INF or -INF

	float t1 = (minXYZ.x - Org.x) / Dir.x;
	float t2 = (maxXYZ.x - Org.x) / Dir.x;
	float tMin = min(t1, t2);
	float tMax = max(t1, t2);

	t1 = (minXYZ.y - Org.y) / Dir.y;
	t2 = (maxXYZ.y - Org.y) / Dir.y;
	tMin = max(tMin, min(t1, t2));
	tMax = min(tMax, max(t1, t2));

	t1 = (minXYZ.z - Org.z) / Dir.z;
	t2 = (maxXYZ.z - Org.z) / Dir.z;
	tMin = max(tMin, min(t1, t2));
	tMax = min(tMax, max(t1, t2));
	return float2(tMin, tMax);

	// float3 t1 = (minXYZ - Org) / Dir;
	// float3 t2 = (maxXYZ - Org) / Dir;
	// float3 tMin = min(t1, t2);
	// float3 tMax = max(t1, t2);
	// return float2(min(tMin.x, min(tMin.y, tMin.z)), max(tMax.x, max(tMax.y, tMax.z)));
}

