#ifndef __FILE_WIREFRAME_FXH__
#define __FILE_WIREFRAME_FXH__
VertexShadingOutput WireframeVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif
	//OUTPUT
	VertexUnifyOutputCommon(IN, INTM, OUT);

	if (cParamter2.w < 0.1)
	{
		OUT.HPosition.z += CameraInfo.x * 0.01 / (0.01 + OUT.HPosition.w);
	}
	//[0.1-0.2) 1uv
	else if (cParamter2.w < 0.2)
	{
		OUT.HPosition.xy = IN.TexCoord.xy * float2( 1.f, -1.f ) - float2(1.f, 0.f);
		OUT.HPosition.z = 1.f;
		OUT.HPosition.w = 1.f;
	}
	//[0.2-0.3) 2uv
	else if (cParamter2.w < 0.3)
	{
		if (@_LIGHT_MAP)
		{
			OUT.HPosition.xy = IN.TexCoord2.xy * float2( 1.f, -1.f ) - float2(1.f, 0.f);
			OUT.HPosition.z = 1.f;
			OUT.HPosition.w = 1.f;
		}
	}
	//[0.3-0.4) 3uv
	else if (cParamter2.w < 0.4)
	{
		if (@_UV3UV4)
		{
			OUT.HPosition.xy = IN.TexCoord3.xy * float2( 1.f, -1.f ) - float2(1.f, 0.f);
			OUT.HPosition.z = 1.f;
			OUT.HPosition.w = 1.f;
		}
	}
	//[0.4-0.5) 4uv
	else if (cParamter2.w < 0.5)
	{
		if (@_UV3UV4)
		{
			OUT.HPosition.xy = IN.TexCoord3.zw * float2( 1.f, -1.f ) - float2(1.f, 0.f);
			OUT.HPosition.z = 1.f;
			OUT.HPosition.w = 1.f;
		}
	}
	//[0.5-0.6) lightmap
	else if (cParamter2.w < 0.6)
	{
		if (LIGHTMAP_ENABLE)
		{
// #if NO_LIGHT_MAP_UV_TRANSFORM
#if USE_LIGHTMAP
			float2 lightmapUV = IN.TexCoord2 * cLightMapUVTransform.xy + cLightMapUVTransform.zw;
#else
			float2 lightmapUV = IN.TexCoord2.xy;
#endif		
			OUT.HPosition.xy = lightmapUV * float2(1.f, -1.f) - float2(1.f, 0.f);
			OUT.HPosition.z = 1.f;
			OUT.HPosition.w = 1.f;
		}
	}
	return OUT;
}

[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].TexCoord = float4(1, 0, 0, 0);
	OUT.Append(IN[0]);
	IN[1].TexCoord = float4(0, 1, 0, 0);
	OUT.Append(IN[1]);
	IN[2].TexCoord = float4(0, 0, 1, 0);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN, float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT;
	if (cParamter2.w < 0)	// Quad overdraw
	{
		uint2 pos = uint2(screenCoord.xy) & 1;
		uint index = pos.x + 2 * pos.y;
		bool inside = all(IN.TexCoord.xyz >= 0);
		uint mask0 = inside << index;

		int2 dir = pos ? -1 : 1;
		uint mask1 = mask0 + dir.x * ddx_fine(mask0);
		uint mask2 = mask0 + dir.y * ddy_fine(mask0);
		uint mask3 = mask2 + dir.x * ddx_fine(mask2);
		uint mask = mask0 | mask1 | mask2 | mask3;

		float value = firstbitlow(mask) == index ? 1 : 0;
		OUT = float4(value / 255.0, 0, 0, 0);
	}
	else					// Anti-aliased wireframe.
	{
		float3 difference = fwidth(IN.TexCoord.xyz);
		float3 distance = IN.TexCoord.xyz / difference;
		float d = min(min(distance.x, distance.y), distance.z);
		float a = exp2(-1.5 * d * d);
		OUT = float4(cParamter2.xyz * a, a);
	}
	return OUT;
}

technique Wireframe
{
	VertexShader = WireframeVertexShading();
	GeometryShader = WireframeGeometryShading();
#if USE_TESSELLATION	
	HullShader = PbrVertexShading_HS();
	DomainShader = PbrVertexShading_DS();
#endif
	PixelShader = WireframePixelShading();
};
#endif // __FILE_WIREFRAME_FXH__
