#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 		1
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow)
#define DEFERRED_PRE_AO						1 // 是否走ao
#define VIRTUAL_LIT 						1
#include "Constants.fxh"
#include "Common.fxh"
#include "BRDF.fxh"
#include "DeferredShadingCommon.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_EmissionMapEnable
{
	Name = EmissionMapEnable;
	Description = "Do we have EmissionMap?";
	AffectedTex = tEmissionMap; 
	Fallback = true;
	Visible = true;
};
@_EnableNormalMap
{
	Name = EnableNormalMap;
	Description = "EnableNormalMap?";
	AffectedTex = tNormalMap; 
	Default = false;
	Visible = true;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;
	float3 cAdjustCenter			 = {0, 0, 0};
	float4 cEmissionScale			 = {1, 1, 1, 0};
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "1"; Max = "100"; Step = "0.01"; Precision = "2"; Visible = "0";> = 80.0;

};
//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
	Precision = Medium;
};
// Emission map
texture2D tEmissionMap : layout(t2) = "Transparent";
sampler sEmissionMapSampler : layout(s2)
{
	Texture = tEmissionMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
	Precision = Medium;
};

texture2D tNormalMap : layout(t3) = "Transparent";
sampler sNormalMapSampler : layout(s3)
{
	Texture = tNormalMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
	Precision = Medium;
};

texture2D tShadowMask : ShadowMask : layout(t4) = "White";
sampler sShadowMaskSampler : layout(s4)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct LODVertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.

	float4 		WorldPivot	 	: TEXCOORDN;			//  WorldPivot

	float 		LinearZ			: TEXCOORDN;
	[@_TANGENT]
	float4		Tangent			: TEXCOORDN;
	[@_TANGENT]
	float4		Binormal		: TEXCOORDN;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
#include "Lighting.fxh" 
LODVertexShadingOutput LODVertexShading(in VertexAssemble IN)
{
	LODVertexShadingOutput OUT;
//	[@_HLSL]
//	OUT = (LODVertexShadingOutput)0;

	//position
	if (@_INSTANCED)
	{
		OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, IN.Position.xyz), 1);
	}
	else
	{
		OUT.WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, 1);
	}

	//WorldPivot
	if (@_INSTANCED)
	{
		OUT.WorldPivot = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, (cAdjustCenter.xzy*0.01) ), 1);
	}
	else
	{
		OUT.WorldPivot = float4(MulMat3x4(World, float4( (cAdjustCenter.xzy*0.01) , 1)).xyz, 1);
	}

	//TexCoord
	OUT.TexCoord = IN.TexCoord.xyxy;
	//OUT.TexCoord.zw = IN.TexCoord2.xy;

	float3 LocalNormal = (float3)Expand((half3)IN.Normal.xyz);
	if (@_INSTANCED)
	{
		OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, LocalNormal));
		//Tangent&Binormal
		[@_TANGENT]
		OUT.Tangent.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, IN.Tangent));
		[@_TANGENT]
		OUT.Binormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, IN.Binormal));
	}
	else
	{
		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, LocalNormal));
		//Tangent&Binormal
		[@_TANGENT]
		OUT.Tangent.xyz =  normalize(MulMat3x4((float3x3)World, IN.Tangent.xyz));
		[@_TANGENT]
		OUT.Binormal.xyz = normalize(MulMat3x4((float3x3)World, IN.Binormal.xyz));
	}
	OUT.WorldNormal.w = 0;

	//ViewDirection
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;

	//Fog T
	//OUT.ViewDirection.w = FogTransmittance(normalize(OUT.ViewDirection.xyz)*cFogDistance);
	if (@_FOG)
		OUT.ViewDirection.w = FogTransmittance(OUT.ViewDirection.xyz);

	//diy output
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	OUT.LinearZ = OUT.HPosition.w * CameraInfo.z;
	return OUT;
}

//////////////////////////////////Function////////////////////////////////////////
/*half3 SphereLighting(in LODVertexShadingOutput IN, half3 V, half3 L)
{
    half3 CoSphere = (normalize(IN.WorldPosition.xyz - IN.WorldPivot.xyz));        //center of sphere(CoSphere)
	half NdotL =  saturate(dot(CoSphere,L)) * 0.5;
	half NdotL2 =  saturate(dot(CoSphere,L) * 0.5 + 0.5) * 0.5;
	half VdotL = dot(-V,L);
	half _NdotL = lerp(NdotL2, NdotL, VdotL);
	return _NdotL * SunColor.xyz + AmbientColor.xyz;
}*/
void InitScreenGray(in LODVertexShadingOutput IN, in out half3 BaseColor)
{
	if(ScreenMGrayModeSpread)
	{
		half grayDist = (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - min(length(IN.WorldPosition.xyz - PlayerPos.xyz), ScreenMGrayMaxDistance);
		grayDist = step(0, grayDist);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);		
	}
	else if (ScreenMGrayModeShrink)
	{
		half grayDist = length(IN.WorldPosition.xyz - PlayerPos.xyz) + (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - ScreenMGrayMaxDistance;
		grayDist = step(grayDist, 0);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);
	}
}
void PS_GetParameters(in LODVertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	//init MaterialParameters (always need)
	InitShadingParameters(MaterialParameters);
	half4 samplerBase = tBaseMap.SampleLevel(sBaseSampler, IN.TexCoord.xy, 0);
	MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
	MaterialParameters.Alpha = samplerBase.a;
	MaterialParameters.Roughness = 0.8;
	//screen gray
	InitScreenGray(IN, MaterialParameters.BaseColor);
	//Normal
	if (@_EnableNormalMap && @_TANGENT)
    {
		half4 samplerNormal = tNormalMap.SampleLevel(sNormalMapSampler, IN.TexCoord.xy, 0);
		half3 tangentNormal = (2 * samplerNormal.rgb) - 1;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.Tangent + tangentNormal.y * IN.Binormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
		MaterialParameters.AO = samplerNormal.a;
    } 
	else
	{ MaterialParameters.WorldNormal = IN.WorldNormal.xyz; }
	//Emission
	if (@_EmissionMapEnable)
	{
		MaterialParameters.Emission = tEmissionMap.SampleBias(sEmissionMapSampler, IN.TexCoord.xy, -3).rgb;
		MaterialParameters.Emission *= cEmissionScale.rgb * MaterialParameters.BaseColor;
	}
	//Apply Weather
	MaterialParameters.BaseColor *= 1 - 0.3 * RainRate;
	MaterialParameters.Roughness *= 1 - 0.3 * RainRate;
	//Apply Burn
	half BurnRadius = GlobalBurnParam.w;
	half3 BurnPosition = GlobalBurnParam.xyz;
	if(BurnRadius>0.01)
	{
		half BurnWidth = 80;
		half BurnArea = saturate((length(IN.WorldPosition-BurnPosition) - BurnRadius) / BurnWidth);
		MaterialParameters.BaseColor *= lerp(0.04 , 1 ,BurnArea);
	}
}

//////////////////////////////////PS////////////////////////////////////////
float4 LODPixelShading(in LODVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);
	MaterialShadingParameters MaterialParameters;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	OUT.a = MaterialParameters.Alpha;
	MaterialParameters.DiffuseColor = Diffuse_Lambert(MaterialParameters.BaseColor - MaterialParameters.BaseColor * MaterialParameters.Metallic);
	//GILighting
	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
	MaterialParameters.GILighting *= MaterialParameters.DiffuseColor;
	//Lighting
	half _NoL = dot(MaterialParameters.WorldNormal, SunDirection.xyz);
	half NoL = saturate(_NoL);
	half MicroShadow = saturate(MaterialParameters.AO * MaterialParameters.AO * 2 + abs(_NoL) - 1);
	half3 Diffuselighting = MaterialParameters.DiffuseColor * NoL * MicroShadow * SunColor.xyz;
	OUT.rgb = MaterialParameters.GILighting + Diffuselighting;
	//thunder shading
	OUT.rgb *= ThunderRate;
	// TODO: Fog shading.
	if (@_FOG)
	{
		float3 fogFinalColor = float3(0,0,0);
		FogLightingAndScatter(IN.ViewDirection.w, fogFinalColor, normalize( -IN.ViewDirection.xyz ), SunDirection.rgb);
		OUT.rgb = lerp(fogFinalColor, OUT.rgb, IN.ViewDirection.w);
	}
	//Correct HDR & EncodeHDR color
	EncodeHDR(OUT.rgb);
	return OUT;
}

technique Shading
{
	VertexShader = LODVertexShading();
	PixelShader = LODPixelShading();
}

//////////////////////////////////GBuffer////////////////////////////////////////
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	
	float4 GBufferB 	: SV_Target1;	
	float4 GBufferC 	: SV_Target2;	
	float4 GBufferD 	: SV_Target3;	
	float4 GBufferE 	: SV_Target4;	
	float4 GBufferF  	: SV_Target5;	
	float4 GBufferG 	: SV_Target6;	
};
GBufferOUTPUT GBufferGenPixelShading(in LODVertexShadingOutput IN)
{
	GBufferOUTPUT OUT;
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	half Alpha = MaterialParameters.Alpha;
	//GILighting
	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
	MaterialParameters.GILighting *= MaterialParameters.AO;
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	EncodeGBufferData(GBuffer, OUT.GBufferA, OUT.GBufferB, OUT.GBufferC, OUT.GBufferD, OUT.GBufferE, OUT.GBufferF);
	OUT.GBufferG = EncodeDepth(IN.LinearZ);
	OUT.GBufferA.w = EncodeDepth(IN.LinearZ);
	return OUT;
}

technique GBufferGen
{
	VertexShader = LODVertexShading();	
	PixelShader = GBufferGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
[barycentric("sample")]
float4 LODClipShading(in LODVertexShadingOutput IN) : SV_Target
{
	// Clipping
	float clipValue = 1.0;

	if (@_CLIPPING)
	{
		half Alpha = tBaseMap.SampleLevel(sBaseSampler, IN.TexCoord.xy, 0).a;
		clipValue = Alpha - ALPHA_REF;
	}
	if (@_DISSOLVING)
	{
		clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(IN.HPosition.xy));
	}
	if (@_CLIPPING || @_DISSOLVING)
		clip(clipValue);

	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = LODVertexShading();
	PixelShader = LODClipShading();
}

/////////////////////////////////////ShadowGen/////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC 				: TEXCOORDN;			//	TexCoord
};

ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
//	[@_HLSL]
//	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.SceneDepth = OUT.HPosition.zw;
	//TC
	OUT.TC = IN.TexCoord.xy;
	return OUT;
}

[target("R32F")]
float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float Alpha = tBaseMap.SampleRGBA( sBaseSampler, IN.TC ).a;
	float clipValue = 1.0;
	
	if (@_CLIPPING)
	{
		clipValue = Alpha - ALPHA_REF;
	}
	if (@_DISSOLVING)
	{
		clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(IN.HPosition.xy));
	}
	if (@_CLIPPING || @_DISSOLVING)
		clip(clipValue);

	float depth = IN.SceneDepth.x / IN.SceneDepth.y;	
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

#if SD_ENABLE_VXGI
#define VSOutput LODVertexShadingOutput
#define VoxelizeVS LODVertexShading
#include "VXGI/VXGIVoxelize.fxh"
#endif

#ifndef __FILE_DEBUGSHADER_FXH__
#define __FILE_DEBUGSHADER_FXH__
// Debug pixel shader.
float4 DebugPixelShading(in LODVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);
	GBufferData GBuffer;
	MaterialShadingParameters MaterialParameters;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	half Alpha = MaterialParameters.Alpha;
	//GILighting
	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
	MaterialParameters.GILighting *= MaterialParameters.AO;
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	if ((int)DebugFlag.x == 1)								//1-base color
	{
		OUT.rgb = MaterialParameters.BaseColor.rgb;
	}
	else if((int)DebugFlag.x == 2)							//2-Metallic
	{
		OUT.rgb = MaterialParameters.Metallic.rrr;
	}
	else if ((int)DebugFlag.x == 3)							//3-roughness
	{
		OUT.rgb = MaterialParameters.Roughness.rrr;
	}
	else if ((int)DebugFlag.x == 4)							//4-diffuse color
	{
		OUT.rgb = GBuffer.DiffuseColor.rgb;
	}
	else if ((int)DebugFlag.x == 5)							//5-specular color
	{
		OUT.rgb = GBuffer.SpecularColor.rgb;
	}
	else if ((int)DebugFlag.x == 6)							//6-normal
	{
		OUT.rgb = 0.5+0.5*GBuffer.WorldNormal;
	}else if((int)DebugFlag.x == 150)
	{
		OUT.rgb = Diffuse_Lambert(GBuffer.BaseColor.rgb - GBuffer.BaseColor.rgb * MaterialParameters.Metallic);
	}else if((int)DebugFlag.x == 151)
	{
		OUT.rgb = EncodeNormal(GBuffer.WorldNormal);
		OUT.a = IN.LinearZ;
	}
	return OUT;
}
technique Debug
{
	VertexShader = LODVertexShading();
	PixelShader = DebugPixelShading();
}
#endif // __FILE_DEBUGSHADER_FXH__



