//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				1
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define USE_SNOW 				1
#define FOLIAGE_SUB_SURFACE
#define FOLIAGE_SCATTER_OCC saturate(cScatterOcc + MaterialParameters.Thickness)

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh" 

#pragma shader predict("@_DF_SKIN", "0")
#pragma shader predict("@_TRANSPARENT", "0")
#pragma shader predict("@_CLIPPING", "1")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?"; 
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	Fallback = true;
	Default = false;
	Visible = true; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Predict = false;
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Predict = true;
	Visible = true;
};
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Default = true;
	Fallback = true;
	Visible = true;
};
@_LightmapConstEnable
{
	Name = LightmapConstEnable;
	Default = false;
	Visible = false;
};
@_OverlayEnable
{
	Name = OverlayEnable;
	Default = false;
	Fallback = false;
	Visible = false;
};
@_SnowEnable
{
	Name = SnowEnable;
	Description = "Use Snow?";
	Default = true;
	Visible = true;
}
@_SubsurfaceEnable
{
	Name = SubsurfaceEnable;
	Fallback = false;
	Default = false;
	Visible = true;
};
@_LodTexSetEnable
{
	Name = LodTexSetEnable;
	Fallback = false;
	Default = false;
	AffectedTex = tLodBaseMap;
	Visible = false;
};
@_NoiseMapEnable
{
	Name = NoiseMapEnable;
	Fallback = false;
	Default = false;
	AffectedTex = tNoiseMap;
	Visible = false;
};
@_VertexMove
{
	Name = VertexMove;
	Description = "Use Vertex Animation?";
	Default = false;
	Visible = true;
}

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < IsColor = "true"; > = {1, 1, 1};
	float cSnowHeight	  		     < Widget = "FloatSlider"; Min = "0"; Max = "3"; Step = "0.01"; Precision = "3"; > = 2;
 
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cScatterOcc				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cScatterTran		         < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cScatterDiff				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cScatterAmbient			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.1;
	float3 cScatterColor		     < Widget = "LinearColor3"; > = {0.0, 0.0, 0.0};
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};
	
	float cBaseMapBias			     < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;
	float cOverlayBlend			     = 0.0;
	float cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;

	float cAlphaCut 				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.001"; Precision = "3"; > = 0.49;
	float cBillboardAlphaCut		 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.001"; Precision = "3"; > = 0.278;
	float cAlphaDither	    		 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "3"; > = 0.5;
	float cBillboardLit		         < Widget = "FloatSlider"; Min = "0"; Max = "2"; Step = "0.001"; Precision = "1"; > = 1;

	float cWindDirectionX			 = 1;
	float cWindDirectionZ			 = 0;
	float cGlobalMoveFrequency		 = 1;
	float cGlobalMoveFactor			 = 0;
	float cGlobalMovePower			 = 1;
	float cBranchMoveFrequency		 = 1;
	float cBranchMoveFactor			 = 0;
	float cBranchMovePower			 = 1;
	float cDistanceRange			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1.5;
	float cStepForce				 < Widget = "FloatSlider"; Min = "-10"; Max = "40"; Step = "0.01"; Precision = "2"; > = 0;
	float cStepOnPower				 < Widget = "FloatSlider"; Min = "-1"; Max = "5"; Step = "0.001"; Precision = "3"; > = 1;
	float cRandomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "3"; > = 1;

	float cBranchScale				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.001"; Precision = "3"; > = 1;

	float cAlpha					 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;	// Clipping Pass needs this
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t2) = "White";
sampler sEnvSampler : layout(s2)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t3) = "NormalMap";
sampler sNormalSampler : layout(s3)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t4);
sampler sLightMapSampler : layout(s4)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;  
	SRGB = false;  
	MaxMipLevel = 0;  
}; 

texture2D tShadowMask : ShadowMask : layout(t5) = "White";
sampler sShadowMaskSampler : layout(s5)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
}; 

texture2D tLodBaseMap : layout(t7) = "White";
sampler sLodBaseSampler : layout(s7)
{
	Texture = tLodBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tNoiseMap : layout(t8) = "Black";
sampler sNoiseSampler : layout(s8)
{
	Texture = tNoiseMap;
	Filter = Point;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
///////////////////////////////////////////////////////////////////////  
//  TrigApproximate

float CubicSmooth(float x)
{
	return x * x * (3.0 - 2.0 * x);
}

float TriangleWave(float x)
{
	return abs((frac(x + 0.5) * 2.0) - 1.0);
}

float TrigApproximate(float x)
{
	return (CubicSmooth(TriangleWave(x)) - 0.5) * 2.0;
}

float Random(float2 v)
{
	return frac(sin(dot(v, float2(12.9898, 78.233))) * 43758.5453);
}

float3 WindWaveMovement(float3 worldPos, float3 worldNormal, float3 localPos, float scale, half branchScale, float randomPhase)
{
	float3 windDir = float3(cWindDirectionX, 0, cWindDirectionZ);
	float3 branchMoveDir = dot(windDir, worldNormal) * worldNormal;
	float globalMoveAdjust = saturate(localPos.y * cGlobalMoveFactor);
	float branchMoveAdjust = saturate(localPos.y * cBranchMoveFactor);
	
	globalMoveAdjust = pow(globalMoveAdjust, cGlobalMovePower);
	branchMoveAdjust = pow(branchMoveAdjust, cBranchMovePower);

	float globalPhase = GameTime * cGlobalMoveFrequency;
	float g = TrigApproximate(randomPhase + globalPhase);

	worldPos *= cBranchScale;
	
	float branchPhase = GameTime * cBranchMoveFrequency;
	float b1 = TrigApproximate(worldPos.x + branchPhase);
	float b2 = TrigApproximate(worldPos.z + branchPhase * 0.72);
	float b = b1 + b2 * b2;


	if (1) // (@_IsImpactByPlayer)
	{
		float StepOnPower = cStepOnPower * PlayerPos.w;
		float3 worldPositionT = PlayerPos.xyz;
		float effectRange = distance(worldPositionT,worldPos);
		effectRange = smoothstep(0, cDistanceRange / 2, cDistanceRange - effectRange);
		g *= 1 + StepOnPower * effectRange;
		b *= 1 + StepOnPower * effectRange;
		float2 pushDirectionXZ = normalize(worldPos.xz - worldPositionT.xz)* effectRange * cStepForce;
		float3 pushDirection = float3(pushDirectionXZ.x, 1, pushDirectionXZ.y);
		float3 globalMove = windDir * globalMoveAdjust * (g + pushDirection);
		float3 branchMove = branchMoveDir * branchMoveAdjust* (b + pushDirection) ;
		return (globalMove + branchMove * branchScale) * scale;
	}
	else
	{
		float3 globalMove = windDir * g * globalMoveAdjust;
		float3 branchMove = branchMoveDir * b * branchMoveAdjust;
		return (globalMove + branchMove * branchScale) * scale;
	}
	
}


float3 WorldPositionOffset(in VertexAssemble IN, in float3 WorldPosition, in float3 WorldNormal, in half branchScale)
{
	if (@_LOD_LEVEL <= LOD_Level_2 && !@_ES20 && @_VertexMove)
	{
		float scale = 1.0f;
		float randomPhase = 0;
		if (@_INSTANCED)
		{
			scale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, float3(1, 0, 0)));
		}
		else
		{
			scale = length(MulMat3x4((float3x3)World, float3(1, 0, 0)));
			randomPhase = frac(World[3].z * 53.2475) * cRandomScale;
		}
		return WindWaveMovement(WorldPosition, WorldNormal, IN.Position.xyz, scale * LOD_RATE, branchScale, randomPhase);
	}
	else if (@_LOD_LEVEL == LOD_Level_2 && @_INSTANCED && @_VertexMove)
	{
		float r = Random(float2(IN.Transform0.w, IN.Transform2.w));
		if (r > LOD_RATE)
		{
			return -WorldPosition;
		}
		else
		{
			return float3(0, 0, 0);
		}
	}
	else
	{
		return float3(0, 0, 0);
	}
}

void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
	{
		if (@_LightmapConstEnable && !@_INSTANCED)
		{
			// Expect batched foliage here, for those the lightmap info is stored in vertex color.
			OUT.TintColor = GetVertexColor(IN.Color) * 8.0;
		}
		else
		{
			OUT.TintColor = GetVertexColor(IN.Color);
		}
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	if (@_Billboard && @_INSTANCED)
	{
		float3 yAxis = float3(0, 1, 0);
		float3 zAxis = OUT.WorldNormal.xyz;
		float3 xAxis = cross(yAxis, zAxis);
		float3 offset = IN.Position.x * xAxis + IN.Position.y * yAxis + IN.Position.z * zAxis;
		OUT.WorldPosition.xyz += offset * OUT.WorldNormal.w;
	}
	else
	{
		if (@_VERTEXCOLOR)
		{
			OUT.WorldPosition.xyz += WorldPositionOffset(IN, OUT.WorldPosition.xyz, OUT.WorldNormal.xyz, 1);//OUT.TintColor.g);
		}
		else
		{
			OUT.WorldPosition.xyz += WorldPositionOffset(IN, OUT.WorldPosition.xyz, OUT.WorldNormal.xyz, 1);
		}
	}

	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
}

//////////////////////////////////PS////////////////////////////////////////
float GetRegularPattern(float2 tc)
{
	int2 sp = (int2)(tc + float2(TemporalAASample, TemporalAASample));//frac(floor(frac(GameTime)*60) * 0.25));
	return fmod(sp.x + 2 * sp.y, 5);
}

void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	half4 samplerBase;
	
	if(@_LOD_LEVEL > LOD_Level_1 && @_LodTexSetEnable)
	{
		samplerBase =  tLodBaseMap.SampleBias(sLodBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
	}
	else
	{
		samplerBase =  tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
	}
	
	MaterialParameters.BaseColor = samplerBase.rgb;
	MaterialParameters.Alpha =samplerBase.a;

    if(@_LOD_LEVEL <= LOD_Level_1)
    {
    	if(@_NoiseMapEnable)
    	{
    		half4 samplerNoise = tNoiseMap.Sample(sNoiseSampler, IN.HPosition.xy * 64);
			MaterialParameters.Alpha += cAlphaCut - cAlphaDither * ((GetRegularPattern(IN.HPosition.xy) + samplerNoise.r)  / 6);
    	}
    	else
    	{
    		MaterialParameters.Alpha += cAlphaCut - cAlphaDither * GetRegularPattern(IN.HPosition.xy) / 6;
    	}
    	
    }
	else
	{
		if(@_NoiseMapEnable)
    	{
    		half4 samplerNoise = tNoiseMap.Sample(sNoiseSampler, IN.HPosition.xy * 64);
			MaterialParameters.Alpha += cBillboardAlphaCut - cAlphaDither * ((GetRegularPattern(IN.HPosition.xy) + samplerNoise.r)  / 6);
    	}
    	else
    	{
    		MaterialParameters.Alpha += cBillboardAlphaCut - cAlphaDither * GetRegularPattern(IN.HPosition.xy) / 6;
    	}
	}


	if (@_OverlayEnable && @_INSTANCED)
	{
		float brightness = dot(MaterialParameters.BaseColor, half3(0.3, 0.6, 0.1));
		MaterialParameters.BaseColor.rgb = lerp(MaterialParameters.BaseColor.rgb, IN.LightMapScale.rgb * brightness, cOverlayBlend);
	}

	half3 vertexColor = half3(1,1,1);

	if(!@_LightmapConstEnable || @_INSTANCED)
	{
		vertexColor = IN.TintColor.rgb;
		MaterialParameters.AO = IN.TintColor.a;
	}

	MaterialParameters.BaseColor = DecodeGamma(MaterialParameters.BaseColor);

	if(@_IsSameDirection)
	{
		MaterialParameters.BaseColor *= cBillboardLit;
	}

    if(@_TANGENT && @_NormalMapEnable)
    {
		if(@_LOD_LEVEL <= LOD_Level_1 || @_SubsurfaceEnable)
		{
			half4 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias);
			MaterialParameters.Roughness = tangentNormal.z;
			MaterialParameters.Thickness = tangentNormal.a;
			tangentNormal = (2 * tangentNormal) - 1;
			if(@_IsSameDirection)
			{
				MaterialParameters.WorldNormal = normalize(float3(tangentNormal.x, 1, tangentNormal.y));
			}
			else if(@_LOD_LEVEL <= LOD_Level_1)
			{
				
				MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + IN.WorldNormal.xyz;
				MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
			}
			else
			{
				MaterialParameters.WorldNormal = normalize(IN.WorldNormal);
			}
		}
		else
		{
			if(@_IsSameDirection)
			{
				MaterialParameters.WorldNormal = float3(0, 1, 0);
			}
			else
			{
				MaterialParameters.WorldNormal = normalize(IN.WorldNormal);
			}
		}
    }
    else
    {
    	if(@_IsSameDirection)
	    {
	    	MaterialParameters.WorldNormal = float3(0, 1, 0);
	    }
	    else
	    {
	    	MaterialParameters.WorldNormal = normalize(IN.WorldNormal);
	    }
    }
}

float3 GetFoliageSubsurfaceScattering(in half3 L, in half3 V, in half3 N, in half Thickness, in half invShadow, in half AO)
{
	if (@_SubsurfaceEnable && !@_ES20)
	{

		half3 SubsurfaceColor = cScatterColor.rgb * (1 - Thickness);

        // GGX scatter distribution
		half VoL = saturate(dot(V, -L));
		half a = 0.6;
		half a2 = a * a;
		half d = (VoL * a2 - VoL) * VoL + 1;	// 2 mad
		half GGX = (a2 / PI) / (d * d);			// 2 mul, 1 rcp

        if(@_IsSameDirection)
        {
			return (GGX * cScatterTran + cScatterDiff) * SunColor.rgb * invShadow * SubsurfaceColor;
        }
        else
        {
        	// http://blog.stevemcauley.com/2011/12/03/energy-conserving-wrapped-diffuse/
			half Wrap = 0.5;
			half NoL = saturate((abs(dot(N, L)) + Wrap) / ((1 + Wrap) * (1 + Wrap)));
			return (GGX * cScatterTran * NoL + cScatterDiff) * SunColor.rgb * invShadow * SubsurfaceColor + cScatterAmbient * AO * SubsurfaceColor;
        }
		
	}
	return float3(0, 0, 0);
}
#include "PBR_PS.fxh"
