//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : szn3878@corp.netease.com
//	Author: Si Zheng
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR_Character.fx>
///	@path	~/System/Shader/
///	@date	2017/11/14
///	@brief	Improved Character Shader.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				1
#define USE_HIGHLIGHT			1
#define USE_FROZEN				1
#define USE_BURNT				1
#define USE_EDGELINE			1
#define USE_PLANAR_SHADOW		1
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			1
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define VIRTUAL_LIT 			1
#define HAS_RIMLIGHT 			1
///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_LIGHT_MAP", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	Default = false;
	Visible = true;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = false;	
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Predict = false;
	Visible = true;
};
@_EmissionMapEnable
{
	Name = EmissionMapEnable;
	Description = "Do we have EmissionMap?";
	AffectedTex = tEmissionMap;
	Fallback = true;
	Visible = true;
};
@_AnisotropicEnable
{
	Name = AnisotropicEnable;
	Description = "Do we need calc anisotropic lighting?";
	Visible = true;
}
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};
@_PatternEnable
{
	Name = ISPatternEnable;
	Description = "PatternEnable?";
	Visible = true;
};
@_RimLight
{
	Name = RimLight;
	Description = "EnableRimLight";
	Visible = true;
}
@_SnowEnable
{
	Name = SnowEnable;
	Description = "Use Snow?";
	Fallback = false;
	Visible = true;
}
@_AniMetal
{
	Name = AniMetal;
	Description = "Use Metallic on Anisotropic?";
	Fallback = false;
	Visible = true;
}
@_LightmapConstEnable
{
	Name = LightmapConstEnable;
	Predict = false;
	Visible = false;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cEmission					 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;

	float cAliasingFactor			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;//抗锯齿因子
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1.5;	//环境强度[0:8]
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "15"; > = -1;

	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float4 cEmissionScrolling		 = {1, 1, 0, 0};

	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;
	float3 cLDirection				 = {-1, -1, -1};

	float4 cPatternUV				 = {0, 0, 1, 1};
	float3 cPatternColor 			 < Widget = "LinearColor3"; > = {1, 1, 1};

	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;

	float cSnowHeight	  			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "2"; > = 0.6;
	float cEdgeThinckness			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.9;
	float cEdgeDarkness				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.67;
	float cEdgeCover				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = -0.64;
	float3 cRimColor 				 < Widget = "LinearColor3"; > = {0.63, 0.75, 1};
	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1;
	float3 cAniColor1 				 < Widget = "LinearColor3"; > = {1, 0.8, 0.81};
	float cRoughnessX				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.1;
	float cRoughnessY				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 1;
	float cAnisotropicScale			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0;
	float cAniPos					 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cAbumpIn					 < Widget = "FloatSlider"; Min = "0"; Max = "0.5"; Step = "0.01"; Precision = "2"; > = 0.2;
	float cRoughnessY2				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.1;
	float cAnisotropicScale2		 < Widget = "FloatSlider"; Min = "0"; Max = "20"; Step = "0.01"; Precision = "2"; > = 0.3;
	float cAnishift					 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cAbumpIn2					 < Widget = "FloatSlider"; Min = "0"; Max = "0.5"; Step = "0.01"; Precision = "2"; > = 0.2;
	float cAlpha	      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;
	float cDepthOffset     			 < Widget = "FloatSlider"; Min = "-5"; Max = "5"; Step = "0.1"; Precision = "2"; > = 0;	
	float cTinColorAlpha  			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "2"; > = 1;	
	float cDepthBias	  			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "2"; > = 0;	
	float cEdgeThinckness2			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1.11;
	float cEdgeDarkness2			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.2;
	float cEdgeCover2				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = -0.5;
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.0;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tShadowMask : ShadowMask : layout(t5) = "White";
sampler sShadowMaskSampler : layout(s5)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

// Emission map
texture2D tEmissionMap : layout(t7) = "Transparent";
sampler sEmissionMapSampler : layout(s7)
{
	Texture = tEmissionMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
};

texture2D tRainMap : RainMap: layout(t8) = "NormalMap";
sampler sRainSampler : layout(s8)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh" 
float3 GetAnisotropic(half aniLscale,half anishift,half aniscale,half aniwide, half3 anicolor,half Ashift,half3 N,half3 suncolor,half sunpower,half3 basecolor,VertexShadingOutput IN)
{
	half RoughnessX = cRoughnessX + EPSILON;
	half RoughnessY = aniwide + EPSILON;
	half3 L = SunDirection.xyz + half3(0,Ashift * aniLscale*10-1.2 +anishift,0);
	half NoL = max(0.001, dot(N, L));
	half3 inputRadiance = suncolor * anicolor * sunpower * NoL;
	half3 V = normalize( -IN.ViewDirection.xyz );
	float3 LPlusV = L + V;
	half len = max(0.001, length(LPlusV));
	half3 H = LPlusV / len;
	half VoN = max(0.001, dot(V, N));
	half LoN = NoL;
	half HoN = dot(H, N);
	half HoT = dot(H, IN.WorldTangent.xyz);
	half HoB = dot(H, IN.WorldBinormal.xyz);
	half beta_a = HoT / RoughnessX;
	beta_a *= beta_a;
	half beta_b = HoB / RoughnessY;
	beta_b *= beta_b;
	half beta = -2.0 * (beta_a + beta_b) / (1.0 + HoN);
	half s_den = max(0.001, 12.5663706 * RoughnessX * RoughnessY * sqrt(LoN * VoN));
	half3 outAni = aniscale * basecolor * inputRadiance * exp(beta) / s_den;
	return outAni;
}
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR && !@_AnisotropicEnable)
	{
		half3 mask = GetVertexColor(IN.Color) > 0.1;
		OUT.TintColor = mask.r*cTintColor1 + mask.g*cTintColor2 + mask.b*cTintColor3;
		OUT.TintColor += 1 - (half)(mask.r + mask.g + mask.b > 0.001);
	}
	else if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}

	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	
	if(@_EmissionMapEnable)
	{
		half2 EmissionUV = IN.TexCoord.xy * cEmissionScrolling.xy + GameTime * cEmissionScrolling.zw;
		OUT.WorldNormal.w = EmissionUV.x;
		OUT.ViewDirection.w = EmissionUV.y;
	}

	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}

	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

		//overwrite Fog T
		OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
	}
}  

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	float4 pattern = float4(1,1,1,1);
	float4 UsePatternA = float4(1,1,1,1);	
	float2 PatternUV;
	float3 tangentNormal = float3(1,0,0);
	float4 samplerBase = float4(1,1,1,1);
	float4 TintColor = IN.TintColor;
	float3 normalSamp = float3(1,1,0);;
	half3 mBasecolor = cBaseColor;
	if(@_AniMetal)
		mBasecolor = half3(1,1,1);

	if (@_BaseMapEnable)
	{
		if (@_CLIPPING || @_DISSOLVING || @_TRANSPARENT || @_EmissionMapEnable || @_ES20)
			samplerBase = tBaseMap.SampleRGBA(sBaseSampler, BaseMapUV(IN.TexCoord));
		else
			samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
		MaterialParameters.BaseColor = samplerBase.rgb;
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = half3(1,1,1);
		MaterialParameters.Alpha = cAlpha;
	}

	MaterialParameters.AO = 1.0;
	float anisotropicCtr = 0.5;
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		float4 nsamp = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias);
		anisotropicCtr = nsamp.a;
		normalSamp = nsamp.xyz;
		tangentNormal = float3(normalSamp.xy,1);
		tangentNormal = (2 * tangentNormal) - 1;
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		tangentNormal.xy *= cBumpIns;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
		MaterialParameters.Metallic = normalSamp.b;
		MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
		MaterialParameters.Metallic = cMetallic;
	}

	if (@_EnableReflection)
	{
		half Smoothness = samplerBase.a;
		MaterialParameters.Roughness = 1 - Smoothness;
	}

	if(@_PatternEnable)
	{
		float patternmask = 1;
		if(!@_AnisotropicEnable && @_NormalMapEnable)
		{
			patternmask = anisotropicCtr;
		}
		PatternUV = float2(IN.TexCoord.x * cPatternUV.z+ cPatternUV.x, IN.TexCoord.y * cPatternUV.w + cPatternUV.y) ;
		pattern = tMixMap.SampleRGBA(sMixSampler,PatternUV);
		MaterialParameters.BaseColor = lerp(DecodeGamma(MaterialParameters.BaseColor) * mBasecolor, cPatternColor * pattern.rgb , pattern.a * TintColor.r * cTinColorAlpha * patternmask);
	}
	else
	{
		MaterialParameters.BaseColor = DecodeGamma(MaterialParameters.BaseColor) * mBasecolor;
	}
	
	MaterialParameters.Emission = half3(0, 0, 0);
	if (@_EmissionMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half2 EmissionUV = half2(IN.WorldNormal.w, IN.ViewDirection.w);
		MaterialParameters.Emission = tEmissionMap.Sample(sEmissionMapSampler, EmissionUV).rgb;
		MaterialParameters.Emission.rgb *= cEmission;
		if(@_HDR){MaterialParameters.Emission.rgb *= MaterialParameters.BaseColor;}
		if(!@_TRANSPARENT && !@_CLIPPING)
		{
			MaterialParameters.Emission.rgb *= MaterialParameters.Alpha;
			MaterialParameters.Alpha = 1.0;
		}
	}
	if (@_AnisotropicEnable && @_TANGENT && @_LOD_LEVEL <= LOD_Level_2 && !@_ES20)
	{
		half sunPower = GetLuminance(SunColor.rgb);
		half3 sunNcolor = SunColor.rgb/sunPower;
		half animask = 1;
		half3 secColor = cAniColor1;
		if(@_AniMetal)
		{
			animask *= normalSamp.b;
			secColor = cBaseColor;
		}
		animask *= TintColor.b;
		half3 anicolor1 = GetAnisotropic(cAbumpIn,cAniPos * 5,cAnisotropicScale * MaterialParameters.Alpha,cRoughnessY, cAniColor1,anisotropicCtr,MaterialParameters.WorldNormal,sunNcolor,sunPower,MaterialParameters.BaseColor.rgb,IN);
		MaterialParameters.Emission += anicolor1 * animask;
		if((@_D3D11||@_D3D9) && @_LOD_LEVEL <= LOD_Level_1)
		{
			half3 anicolor2 = GetAnisotropic(cAbumpIn2, cAnishift*5,cAnisotropicScale2 * MaterialParameters.Alpha, cRoughnessY2,secColor,anisotropicCtr,MaterialParameters.WorldNormal,sunNcolor,sunPower,MaterialParameters.BaseColor.rgb,IN);
			MaterialParameters.Emission += anicolor2 * animask;
		}
	}
}
#include "PBR_PS.fxh"
