//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_CLEAR_COAT 1
#define PBR_NOMIXMAP			1
#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				1
#define USE_HIGHLIGHT			1
#define USE_FROZEN				1
#define USE_BURNT				1
#define USE_EDGELINE			1
#define USE_PLANAR_SHADOW		1
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			1
#define VIRTUAL_LIT 1
#define DOUBLE_REFLECTCOLOR 0

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = false;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
#ifndef PBR_NOMIXMAP
	AffectedTex = tMixMap;
	Fallback = true;
	Visible = false;
#else
	Predict = false;
	Visible = true;
#endif
	Default = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = true;	
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = false;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = false;
};
@_EmissionMapEnable
{
	Name = EmissionMapEnable;
	Description = "Do we have EmissionMap?";
	AffectedTex = tEmissionMap; 
	Fallback = true;
	Visible = false;
};
@_AnisotropicEnable
{
	Name = AnisotropicEnable;
	Description = "Do we need calc anisotropic lighting?";
	Visible = true;
}
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};
@_IsFoliageBake
{
	Name = IsFoliageBake;
	Description = "is foliage bake?";
	Visible = true;
};
@_SnowEnable
{
	Name = SnowEnable;
	Description = "Use Snow?";
	Default = false;
	Visible = true;
};
@_LightmapConstEnable
{
	Name = LightmapConstEnable;
	Default = false;
	Visible = false;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 

	float cAliasingFactor			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;//抗锯齿因子
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float4 cEmissionScale			 = {1, 1, 1, 0};
	float4 cEmissionScrolling		 = {1, 1, 0, 0};

	float cRoughnessX				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.1;
	float cRoughnessY				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 1;
	float cAnisotropicScale			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0;
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;

	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1;
	float cAlpha	      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;
	float cDepthOffset     			 < Widget = "FloatSlider"; Min = "-20"; Max = "20"; Step = "0.1"; Precision = "2"; > = 0;
	float cDepthBias	  			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "2"; > = 0;
	float cSnowHeight	  			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.1"; Precision = "2"; > = 0.6;
	float cSnowDetail	  			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "2"; > = 1;
	float cClearCoat				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cClearCoatRoughness		 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tShadowMask : ShadowMask : layout(t6) = "White";
sampler sShadowMaskSampler : layout(s6)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

// Emission map
texture2D tEmissionMap : layout(t7) = "Transparent";
sampler sEmissionMapSampler : layout(s7)
{
	Texture = tEmissionMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
};

texture2D tRainMap : RainMap: layout(t8) = "NormalMap";
sampler sRainSampler : layout(s8)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR && !@_AnisotropicEnable)
	{
		half3 mask = GetVertexColor(IN.Color) > 0.1;
		OUT.TintColor = mask.r*cTintColor1 + mask.g*cTintColor2 + mask.b*cTintColor3;
		OUT.TintColor += 1 - (half)(mask.r + mask.g + mask.b > 0.001);
	}
	else if (@_VERTEXCOLOR && @_AnisotropicEnable)
		OUT.TintColor = GetVertexColor(IN.Color);
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}

	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	OUT.WorldPosition.w = IN.Position.y;
	if(@_EmissionMapEnable)
	{
		half2 EmissionUV = IN.TexCoord.xy * cEmissionScrolling.xy + GameTime * cEmissionScrolling.zw;
		OUT.WorldNormal.w = EmissionUV.x;
		OUT.ViewDirection.w = EmissionUV.y;
	}

	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}

	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);
		
		//overwrite Fog T
		OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
	}
}

//////////////////////////////////HS////////////////////////////////////////
float4 HS_GetTessellationFactor(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2)
{
	return float4(4, 4, 4, 4);
}

//////////////////////////////////DS////////////////////////////////////////
float3 Project(float3 p, float3 c, float3 n)
{
	return p - dot(p - c, n) * n;
}
void DS_DIY_OUTPUT(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2, float3 domain, inout VertexShadingOutput OUT)
{
	float3 c0 = Project(OUT.WorldPosition.xyz, p0.WorldPosition.xyz, normalize(p0.WorldNormal.xyz));
	float3 c1 = Project(OUT.WorldPosition.xyz, p1.WorldPosition.xyz, normalize(p1.WorldNormal.xyz));
	float3 c2 = Project(OUT.WorldPosition.xyz, p2.WorldPosition.xyz, normalize(p2.WorldNormal.xyz));
	float3 pos = c0 * domain.x + c1 * domain.y + c2 * domain.z;
	OUT.WorldPosition.xyz = lerp(OUT.WorldPosition.xyz, pos, 0.5);
}

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	if (@_BaseMapEnable)
	{
		half4 samplerBase;
		if (@_CLIPPING || @_DISSOLVING || @_TRANSPARENT || @_EmissionMapEnable)
			samplerBase = tBaseMap.SampleRGBA(sBaseSampler, BaseMapUV(IN.TexCoord));
		else
			samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = cAlpha;
	}
	if (!@_AnisotropicEnable)
	{
		MaterialParameters.BaseColor *= IN.TintColor.rgb;
	}
	MaterialParameters.AO = 1.0;
#ifndef PBR_NOMIXMAP	
	if (@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).rgb;
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_2)
		{
			MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else
#endif	
	{
		MaterialParameters.Metallic = cMetallic;
		if(@_EnableReflection || @_DEFERRED_SHADING)
			MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		[@_DIRECTX]//always use bc5 for pc
		tangentNormal.z = 1;
		if (@_RIPPLE)
		{
			half2 tc = IN.WorldPosition.xz * RainNormalXZScale;
			half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
			half2 ttc = tc + delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			ttc = tc - delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			tangentNormal /= (1 + 2*RainNormalRate);
		}
		tangentNormal = (2 * tangentNormal) - 1;
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		tangentNormal.xy *= cBumpIns;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
	MaterialParameters.Emission = half3(0, 0, 0);
	if (@_EmissionMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half2 EmissionUV = half2(IN.WorldNormal.w, IN.ViewDirection.w);
		MaterialParameters.Emission = tEmissionMap.Sample(sEmissionMapSampler, EmissionUV).rgb;
		MaterialParameters.Emission.rgb *= cEmissionScale.rgb;
		if(@_HDR){MaterialParameters.Emission.rgb *= MaterialParameters.BaseColor;}
		if(!@_TRANSPARENT && !@_CLIPPING)
		{
			MaterialParameters.Emission.rgb *= MaterialParameters.Alpha;
			MaterialParameters.Alpha = 1.0;
		}
	}
	if (@_AnisotropicEnable && @_TANGENT && @_LOD_LEVEL <= LOD_Level_1 && !@_ES20)
	{
		half RoughnessX = cRoughnessX * MaterialParameters.Roughness + HALF_EPSILON;
		half RoughnessY = cRoughnessY * MaterialParameters.Roughness + HALF_EPSILON;
		half3 L = SunDirection.xyz;
		half NoL = max(0.001, dot(MaterialParameters.WorldNormal, L));
		half3 inputRadiance = SunColor * NoL;
		half3 V = normalize( -IN.ViewDirection.xyz );
		float3 LPlusV = L + V;
		half len = max(0.001, length(LPlusV));
		half3 H = LPlusV / len;
		half VoN = max(0.001, dot(V, MaterialParameters.WorldNormal));
		half LoN = NoL;
		half HoN = dot(H, MaterialParameters.WorldNormal);
		half HoT = dot(H, IN.WorldTangent.xyz);
		half HoB = dot(H, IN.WorldBinormal.xyz);
		half beta_a = HoT / RoughnessX;
		beta_a *= beta_a;
		half beta_b = HoB / RoughnessY;
		beta_b *= beta_b;
		half beta = -2.0 * (beta_a + beta_b) / (1.0 + HoN);
		half s_den = max(0.001, 12.5663706 * RoughnessX * RoughnessY * sqrt(LoN * VoN));
		MaterialParameters.Emission.rgb += IN.TintColor.rgb * cAnisotropicScale * lerp(half3(0.04,0.04,0.04), MaterialParameters.BaseColor, MaterialParameters.Metallic.xxx) * inputRadiance * exp(beta) / s_den;
	}
	MaterialParameters.CustomData.x = cClearCoat;
	MaterialParameters.CustomData.y = cClearCoatRoughness;
}
#include "PBR_PS.fxh"


