//	--------------------------------------------------------------------
//	Copyright(C) 2006~2016 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Project Messiah
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Lava.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_UNLIT 1

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

@_EnableDepthBias
{
	Name = EnableDepthBias;
	Description = "for EnableDepthBias";
	Default = false;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 					;
	float4 cTintColor1				: TintColor1				; 
	float4 cTintColor2				: TintColor2				; 
	float4 cTintColor3				: TintColor3				; 

	float4 cShadowBias				: ShadowBias				; 
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
	
#endif
	float4 cParamter				: Param						;
	float4 cParamter2				: Param2					;
#if !IS_TERRAIN	
	float4 cVirtualLitDir			: VirtualLitDir				;
	float4 cVirtualLitColor			: VirtualLitColor			;
#else
	float4 cHeightmapMinPos			: HeightmapMinPos			;//w heightmap height
	float4 cHeightmapMaxPos			: HeightmapMaxPos			;//w heightmap range / texture size 
#endif
	float4x4 WorldViewProj			: WorldViewProj				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex		;
	float4 PlayerPos                : PlayerPos					;
	float4 FoliageCenter            : FoliageCenter				;
};


cbuffer Shader	
{
	float4 cCursorColor					 = {1, 1 ,1,1};
	float cScale						 = { 1 };
	float3 cOffsets						 = { 0,0,0 };
	float cFov							 <Visible = "0";> = { 1 };
	float cAlpha						 = { 1 };
	float cDepthBias					 = { 1 };


};

// Surface textures.
texture2D tDiffuseMap : layout(t0);
sampler sDiffuseSampler : layout(s0, s1)
{
	Texture = tDiffuseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};



texture2D DepthBuffer : SceneDepth : layout(t4);
sampler DepthBuffer_Sampler : layout(s4)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};


//////////////////////////////////Properties////////////////////////////////////////

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	
	float4 		DiffLighting 	: TEXCOORDN;			//  PerVertex Diffuse Lighting.
 	
#if IS_TERRAIN//terrain always need tangent/binormal
	half3		WorldTangent	: TEXCOORDN;
	half3		WorldBinormal	: TEXCOORDN;
#else	
	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;
#endif

	[@_CastSunShadow]
	float4 		LitUVZ			: TEXCOORDN; 

#if IS_WATER
	[@_CastSunShadow]
	float4 		LitUVZ2			: TEXCOORDN;
#elif USE_PLAYER_SHADOWMAP
	[@_CastSunShadow && @_TRANSPARENT]
	float4 		LitUVZ2			: TEXCOORDN;
#endif

	// Shader special output
	float4		TintColor 		: COLOR;				//	TintColor

	[@_INSTANCED]
	float4		LightMapScale	: TEXCOORDN;			//	Light map scale

	float 		LinearZ			: TEXCOORDN;
#ifdef HAS_WORLD_POSITION_OFFSET
	half3		WorldOffset		: TEXCOORDN;
#endif
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD2;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD4;
	[@_INSTANCED]
	float4		LightMapScale	: TEXCOORD5;
	[@_INSTANCED]
	float4		LightMapTransform: TEXCOORD6;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
	float3 LocalNormal;
	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;

};

void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{
	//[@_HLSL]
	//OUT = (VertexShadingOutput)0;

	//PBR_VS::struct VertexShadingOutput

	OUT.HPosition = float4(0, 0, 0, 0);

	OUT.TexCoord = float4(0,0,0,0);

	OUT.WorldPosition = float4(0,0,0,0);

	OUT.WorldNormal = float4(0,0,0,0);

	OUT.ViewDirection = float4(0,0,0,0);

	OUT.DiffLighting = float4(0,0,0,1);

	#if IS_TERRAIN
		OUT.WorldTangent = half3(0,0,0);
		OUT.WorldBinormal = half3(0,0,0);
	#else
	[@_TANGENT]
	OUT.WorldTangent = half4(0,0,0,0);
	[@_TANGENT]	
	OUT.WorldBinormal = half4(0,0,0,0);
	#endif

	[@_CastSunShadow]
	OUT.LitUVZ = float4(0,0,0,0);

#if IS_WATER
	[@_CastSunShadow]
	OUT.LitUVZ2	 = float4(0,0,0,0);
#elif USE_PLAYER_SHADOWMAP
	[@_CastSunShadow && @_TRANSPARENT]
	OUT.LitUVZ2	 = float4(0,0,0,0);
#endif

	OUT.TintColor = float4(0,0,0,0);

	[@_INSTANCED]
	OUT.LightMapScale = float4(0,0,0,0);

#ifdef HAS_WORLD_POSITION_OFFSET
	OUT.WorldOffset = half3(0, 0, 0);
#endif
	OUT.LinearZ = 0; 
}

void InitInterMediateVertex(inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(0,0,0,0);
	
	INTM.LocalNormal = float3(0,0,0);

	[@_TANGENT]
	INTM.LocalTangent = float3(0,0,0);

	[@_TANGENT]
	INTM.LocalBinormal = float3(0,0,0);
}

void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	//[@_HLSL]
	//INTM = (InterMediateVertex)0;

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}
void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		uint4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			if(@_TANGENT && @_LOD_LEVEL <= LOD_Level_1)
			{
				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
			}
		}
	}
}
// void GetSkeletonMatrix4x3Tex( uint idx, out float4 Mx, out float4 My, out float4 Mz )
// {
// 	Mx = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+0.5)*cParamter.z, 0), 0);
// 	My = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+1.5)*cParamter.z, 0), 0);
// 	Mz = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+2.5)*cParamter.z, 0), 0);
// }
// void VertexUnifySkinningTex(in VertexAssemble IN, inout InterMediateVertex INTM)
// {
// 	if (@_DF_SKIN)
// 	{
// 		half4 w = IN.BlendWeights;
// 		uint4 i = IN.BlendIndices;
// 		// Linear-blend skinning.
// 		if (@_DF_SKIN == 1)
// 		{
// 			// Fetch.
// 			float4 M0[3], M1[3], M2[3], M3[3];
// 			GetSkeletonMatrix4x3Tex(i[0], M0[0], M0[1], M0[2]);
// 			GetSkeletonMatrix4x3Tex(i[1], M1[0], M1[1], M1[2]);
// 			GetSkeletonMatrix4x3Tex(i[2], M2[0], M2[1], M2[2]);
// 			GetSkeletonMatrix4x3Tex(i[3], M3[0], M3[1], M3[2]);

// 			// Blend.
// 			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
// 			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
// 			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

// 			// Apply.
// 			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
// 			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

// 			if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
// 			{
// 				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
// 				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
// 			}
// 		}
		
// 	}
// }



void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 
	INTM.LocalPosition.xyz *=cScale;
	INTM.LocalPosition.xyz+= cOffsets*0.01;

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	
	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}

	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);
		
		//overwrite Fog T
		// OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
	
	}
	// if(@_DisableZdepth)
	// {
	// 	OUT.HPosition.w = 0.1;
	// }

	
	//TexCoord
	OUT.TexCoord.xy = IN.TexCoord.xy;
	OUT.HPosition.w*=cFov;
		if(@_EnableDepthBias)
{

float depthBias = cDepthBias;
	OUT.HPosition.z -= depthBias / (depthBias + OUT.HPosition.w);

	OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
	OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	
}
	//ViewDirection
	// OUT.ViewDirection.xyz = WorldPosition.xyz - CameraPos.xyz;

	



}


//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput BillboardVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	
	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	//OUTPUT
	VertexUnifySkinning(IN, INTM);
	VertexUnifyOutput(IN, INTM, OUT);

	
	
	return OUT;
}

float4 BillboardPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	
	
	half4 color = tDiffuseMap.SampleRGBA(sDiffuseSampler,IN.TexCoord.xy);

	OUT.a = color.a;	
			if (@_CLIPPING)
	{
		
		clip(OUT.a - 0.5);
	}

	OUT.rgb = color.rgb*color.rgb *saturate(cCursorColor) * color.a*10*cAlpha	;
	OUT.a = color.a*cAlpha;
	// OUT.rgb *=OUT.a*cCursorColor.a;
	

		

	


	return OUT;
}

technique Shading
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}
#ifndef __FILE_GBUFFERGEN_FXH__
#define __FILE_GBUFFERGEN_FXH__
/////////////////////////////////////GBuffer/////////////////////////////////////
#ifndef NO_GBUFFER
#ifndef DEFERRED_PRE_AO
#define DEFERRED_PRE_AO 0
#endif
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	// GBufferA[R16G16B16A16F]	: Emission.rgb with lighting result
	float4 GBufferB 	: SV_Target1;	// GBufferB[R8G8B8A8]		: Roughness, Metallic, AO, ShadingModelID
	float4 GBufferC 	: SV_Target2;	// GBufferC[R8G8B8A8]		: BaseColor.rgb, Specular
	float4 GBufferD 	: SV_Target3;	// GBufferD[R8G8B8A8]		: CustomData
	float4 GBufferE 	: SV_Target4;	// GBufferE[R8G8B8A8]		: N.xyz, InShadow
	float4 GBufferF  	: SV_Target5;	// GBufferF[R8G8B8A8]		: TempRT for OnChipMem optimization
	float4 GBufferG 	: SV_Target6;	// GBufferG[R32F]			: Depth
};
[target("X16F", "X16F", "X16F", "X16F", "X16F", "X16F", "R32F")]
GBufferOUTPUT GBufferGenPixelShading(in VertexShadingOutput IN )
{
	GBufferOUTPUT OUT;
	//[@_HLSL]
	//OUT = (GBufferOUTPUT)0;
		half4 color = tDiffuseMap.SampleRGBA(sDiffuseSampler,IN.TexCoord.xy);
	


	 color.rgb*= cCursorColor*cCursorColor.a;

		OUT.GBufferA = color;
		OUT.GBufferB = float4(1, 0, 0, 0); // roughness默认是1，免得ssr计算
		OUT.GBufferC = float4(0, 0, 0, 0);
		OUT.GBufferD = float4(0, 0, 0, 0);
		OUT.GBufferE = float4(0, 0, 0, 0);
		OUT.GBufferF = float4(0, 0, 0, 0);
		OUT.GBufferG = float4(0, 0, 0, 0);

	return OUT;
}

technique GBufferGen
{
	VertexShader = BillboardVertexShading();

	PixelShader = GBufferGenPixelShading();
}
#endif// NO_GBUFFER
#endif // __FILE_GBUFFERGEN_FXH__
//////////////////////////////////////////////////////////////////////////
// Clip

[barycentric("sample")][RE_Z]
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
	if (@_CLIPPING)
	{
		half a = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TexCoord.xy).a;
		clip(a - 0.5);
	}
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = BillboardVertexShading();
	PixelShader = ClipPixelShading();
}



//////////////////////////////////////////////////////////////////////////
// Wireframe
[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].ViewDirection = float4(1, 0, 0,0);
	OUT.Append(IN[0]);
	IN[1].ViewDirection = float4(0, 1, 0,0);	
	OUT.Append(IN[1]);
	IN[2].ViewDirection = float4(0, 0, 1,0);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.ViewDirection);
	float3 distance = IN.ViewDirection / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(cParamter2.xyz, a);
}

technique Wireframe
{
	VertexShader = BillboardVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = WireframePixelShading();
};

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);
	
	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
		OUT.HPosition.w *=cFov;
	if(@_EnableDepthBias)
{

float depthBias = cOffsets.z*0.01;
	OUT.HPosition.z -= depthBias / (depthBias + OUT.HPosition.w);

	OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
	OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	
}
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		half a = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TC.xy).a;
		clip(a - 0.5);
	}

	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}


