#pragma shader material
#pragma shader predict("@_DF_SKIN", "0")
#pragma shader predict("@_TRANSPARENT", "0")
#pragma shader predict("@_CLIPPING", "1")
// #pragma shader predict("@_VERTEXCOLOR", "1")

#define MATERIAL_SHADINGMODEL_TWOSIDED_FOLIAGE 1
#define USE_LIGHTMAP						1
#define USE_LIGHTMAP_CONST					0
#define USE_CLIPPING						1
#define USE_WIREFRAME						1
#define USE_FROZEN							0
#define USE_BURNT							0
#define USE_HIGHLIGHT						0
#define USE_XRAY							0
#define USE_EDGELINE						0
#define USE_PLANAR_SHADOW					0
#define USE_VBufferGen						0
#define VIRTUAL_LIT 						0
#define USE_ONE_SIDE_LIGHTING				1

#define HAS_WORLD_POSITION_OFFSET
#define USE_HISM							1

#define NEED_SHADOWMAP_REFRESH	(@_ShadowMapRefresh)

#define HAS_2UV								(@_LIGHT_MAP)
#define ENABLE_MIXMAP						0
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow && @_LOD_LEVEL <= LOD_Level_2 && !@_TRANSPARENT)
#define DEFERRED_PRE_AO						1
///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"  
//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?"; 
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Default = false;
	Visible = false; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	// Predict = false;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Predict = false;
	Visible = true;
};
@_LightmapConstEnable
{
	Name = LightmapConstEnable;
	Default = false;
	Visible = false;
	Predict = false;
};
@_ShadowMapRefresh
{
	Name = ShadowMapRefresh;
	Description = "Do we need shadow map real time refresh?"; 
	Default = false;
	Visible = true;
};
@_OneSideLightingEnable
{
	Name = OneSideLightingEnable;
	Description = "Do we need calc One side lighting?";
	Default = true;
	Visible = true;	
	Predict = true;
};
@_UsePivotAnimL1
{
	Name = UsePivotAnimL1;
	Default = false;
	Visible = true;
	Predict = false;
};
@_UsePivotAnimL2
{
	Name = UsePivotAnimL2;
	Default = false;
	Visible = false;
	Predict = false;
};
// @_UsePivotAnimL3
// {
// 	Name = UsePivotAnimL3;
// 	Default = true;
// 	Visible = true;
// };
@_WindNoiseMapEnable
{
	Name = WindNoiseMapEnable;
	Description = "Do we have WindNoiseMap?"; 
	// AffectedTex = tWindNoiseMap;
	Fallback = false;
	Visible = false;
};
@_IsHISM
{
	Name = IsHISM;
	Description = "Is HISM?"; 
	Fallback = false;
	Visible = false;
};
@_RadarEnable
{
	Name = RadarEnable;
	Description = "RadarEnable";
	Default = false;
	Visible = false;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < IsColor = "true"; > = {1, 1, 1};
	float cRoughnessLow				 = 0;	//粗糙度下界
	float cRoughnessHigh			 = 1;	//粗糙度上界
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = 0;
	float cAlpha					 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.1"; Precision = "2"; > = 1.0;
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};
	///VertexFoliage
	// float cGlobalMoveFrequency		 = 1;
	// float cGlobalMoveFactor			 = 0;
	// float cGlobalMovePower			 = 1;
	//PivotFoliage
	float cMaxRotationL1_Piv		 = 0.3;
	float cLocalWindScaleL1_Piv    	 = 1;
	float cSoftnessL1_Piv    	 	 = 0.5;
	float cMaxRotationL2_Piv     	 = 0.5;
	float cLocalWindScaleL2_Piv 	 = 5;
	float cSoftnessL2_Piv    	 	 = 0.5;
	//float cPlayerPushDistance		 = 1.5;
	float cPlayerPushIntencity		 = 0.5;
	float3 cRadarColor				< Widget = "LinearColor3"; > = {1, 1, 1};
	float	cRadarScrolling			={0};
	float 	cPlayerDirection		= {1};
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tMixMap : layout(t1) = "Reflection";
sampler sMixSampler : layout(s1)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tNormalMap : layout(t2) = "NormalMap";
sampler sNormalSampler : layout(s2)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tFoliageMap : FoliageMap : layout(t3) = "Black";
sampler sFoliageMapSampler : layout(s3)
{
	Texture = tFoliageMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tFoliageBurnMap : FoliageBurnMap : layout(t4) = "Black";
sampler sFoliageBurnMapSampler : layout(s4)
{
	Texture = tFoliageBurnMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};


texture2D tEnvMap : EnvMap : layout(t5) = "White";
sampler sEnvSampler : layout(s5)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tLightMap : layout(t6);
sampler sLightMapSampler : layout(s6)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;  
	SRGB = false;  
	MaxMipLevel = 0;  
};

// ShadowMap 
texture2D tShadowMask : ShadowMask : layout(t7) = "White";
sampler sShadowMaskSampler : layout(s7)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
}; 

texture2D tShadowMap : ShadowMap : layout(t7) = "White";
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
}; 

texture2D tPivPosMap : layout(t8) = "White";
sampler sPivPosMapSampler : layout(s8)
{
	Texture = tPivPosMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tPivAxisMap : layout(t9) = "White";
sampler sPivAxisMapSampler : layout(s9)
{
	Texture = tPivAxisMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tWindNoiseMap : layout(t10) = "White";
sampler sWindNoiseMapSampler : layout(s10)
{
	Texture = tWindNoiseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"

void checkIfShakeTree(inout float shakeParam,float checkTime,float shakeParamMultiplier)
{
	if( abs(GameTime -  checkTime) < 1  )
		 shakeParam *= shakeParamMultiplier;
}

float3 WindWaveMovementTree(float3 worldPos, float3 localPos, half scale, float3 vertexColor,float3 worldVerPos,float LightMapScaleX )
{
	//windparam xyz 方向 w 强度

	float cWindPush= 5;
	worldPos = abs(worldPos) * sign(worldPos);
	float3 windDir = normalize(float3(WindParam.x, 0, WindParam.z));
	float treeSoftWave = 0.01*WindParam.w*5;
	float3 _treePostion_Instance = worldPos * 0.25;
	float timeSpeed = GameTime*0.5;
	checkIfShakeTree(timeSpeed , LightMapScaleX, 10);

	//6.5  for unreal to  excharge num
	float timeSpeed2 = timeSpeed * 6.5;
	float temp =_treePostion_Instance.x + _treePostion_Instance.z * 0.5 + timeSpeed2;
	float temp2 =_treePostion_Instance.x + _treePostion_Instance.z * 0.5 + timeSpeed;
	vertexColor.g*=500;

	float treeWoldPosWave = sin(temp);//+ windTemp.x
	treeWoldPosWave =  pow(treeWoldPosWave + 1.0, 1.5) +2.0 +WindParam.w;
	float3 windDirDown = float3(0, -1, 0);
	// float3 vertexPos = normalize(localPos);
	float3 windFace = normalize(worldVerPos - worldPos);
	float winDotVert = dot(windDir, windFace);
	float winDotVertDetail = clamp(winDotVert * -0.5, 0.0, 1.0);
	half heightMask = vertexColor.r* vertexColor.r*vertexColor.r;
	float3 waveHorizontal = windDir * treeSoftWave * treeWoldPosWave * heightMask;
	float3 waveDown = treeSoftWave * 0.2 * windDirDown * abs(treeWoldPosWave) * heightMask;
	float3 waveDownDetail =  windDirDown * sin((temp +  (1-winDotVert)) * 2 +  vertexColor.g) * treeSoftWave *  vertexColor.b * vertexColor.r;
	// TODO: timeSpeed2 leads to black pixels on iOS11&iPhone8Plus
	float3 waveHorizontalDetail = vertexColor.b * winDotVertDetail * ( sin(( temp + winDotVert) * 2 + vertexColor.g) + cWindPush*WindParam.w +WindParam.w) * treeSoftWave * windDir* vertexColor.r;
	float3 waveforce = waveHorizontal + waveDown + waveDownDetail + waveHorizontalDetail ; //+ waveDown + waveDownDetail + waveHorizontalDetail
	return waveforce * scale;
}

float3 RotateAboutAxis(float4 NormalizedRotationAxisAndAngle, float3 PivotWorldPos, float3 WorldPos)
{
	// Project Position onto the rotation axis and find the closest point on the axis to Position
	float3 ClosestPointOnAxis = PivotWorldPos + NormalizedRotationAxisAndAngle.xyz * dot(NormalizedRotationAxisAndAngle.xyz, WorldPos - PivotWorldPos);
	// Construct orthogonal axes in the plane of the rotation
	float3 UAxis = WorldPos - ClosestPointOnAxis;
	float3 VAxis = cross(NormalizedRotationAxisAndAngle.xyz, UAxis);
	float CosAngle;
	float SinAngle;
	sincos(NormalizedRotationAxisAndAngle.w, SinAngle, CosAngle);
	// Rotate using the orthogonal axes
	float3 R = UAxis * CosAngle + VAxis * SinAngle;
	// Reconstruct the rotated world space position
	float3 RotatedPosition = ClosestPointOnAxis + R;
	// Convert from position to a position offset
	return RotatedPosition - WorldPos;
}

float3 FixRotateAboutAxisNormals(float4 NormalizedRotationAxisAndAngle, float3 WorldNormal)
{
	float3 rotateNormal = RotateAboutAxis(NormalizedRotationAxisAndAngle, float3(0,0,0), WorldNormal);
	WorldNormal = normalize(WorldNormal + rotateNormal);
	return WorldNormal;
}

float CalculateMeshElementIndex(float2 uvDeimesion,float2 uv)
{
	float2 uvCoordinate = floor(uvDeimesion * uv);
	return float (uvDeimesion.x*uvCoordinate.g + uvCoordinate.r);
}	

float2 ArrayLookUpByIndex (float index,float2 uvDim)
{
	return float2( (float2(fmod(index,uvDim.x) ,floor( index/uvDim.x)) + float2(0.5,0.5) ) / uvDim) ;
}

void ReturnParentTextureInfo( float parentIndex, float2 textSize ,float3 currentIndex, out float2 ParentUV, out float isChild )
{
	ParentUV = ArrayLookUpByIndex(parentIndex, textSize );
	if( parentIndex == currentIndex.r  )
		isChild= 0.0;
	else
		isChild = 1.0;
}

float UnPackIntegerAsFloat(float integerAsFloat)
{
	uint uRes32 = asuint(integerAsFloat);
	uint sign2 = ((uRes32>>16) & (uint)0x8000);
	uint exp2  = ((((uRes32>>23) & (uint)0xff)-127+15) << 10);
	uint mant2 = ((uRes32>>13) & (uint)0x3ff);
	uint bits = (sign2 | exp2 | mant2);
	uint result = bits - 1024;
	return (float)result;
}

float DecodeAlphaAixsExtent(float AlphaExtentVal)
{
	return clamp( AlphaExtentVal * 2048, 8, AlphaExtentVal * 128);
}

void calclulatHierachyInfor(float2 parentUV, float2 textSize, float3 currentIndex, out float2 ParentUVOUT, out float isChildOUT, out float CurrentIndexOUT)
{
	float4 PositionAndIndex = tPivPosMap.SampleLevel(sPivPosMapSampler, parentUV,0);
	float UnPackIntAsFloat =  UnPackIntegerAsFloat( PositionAndIndex.a);
	CurrentIndexOUT= UnPackIntAsFloat;
	ReturnParentTextureInfo(UnPackIntAsFloat,textSize,currentIndex, ParentUVOUT, isChildOUT);
}

void lerpMultipleFloat2(float2 L1,float2 L2, float2 L3, float2 L4, float alphaScale , out float2 Lerp3Input , out float2 Lerp4Input)
{
	float fractTime = alphaScale;	
	Lerp3Input= lerp( lerp( L1, L2, clamp(fractTime*2 -1,0,1)) , L3 ,  clamp(fractTime-1,0,1)  );
	Lerp4Input = lerp(  lerp(lerp(  L1 ,L2 ,  clamp(fractTime*3,0,1)) ,L3, clamp(alphaScale*3-1,0,1)) , L4,  clamp( (fractTime*3 -1)-1 ,0,1) );
}

float SphereMask(float3 A, float3 B, float AttenuationRadius, float Hardness)
{
	float Distance = length((A-B));
	float ArgInvRadius = 1/AttenuationRadius;
	float NormalizeDistance = Distance * ArgInvRadius;
	float ArgInvHardness = 1 - Hardness;
	float NegNormalizedDistance=1-NormalizeDistance;
	float MaskUnclamped = NegNormalizedDistance * ArgInvHardness;
	return clamp(MaskUnclamped, 0, 1);
}

float3 PlayerImpactForPiv( float3 modelPos ,float3 WorldPosition,float4 PlayerImpact)
{
	float3 PlayerRoateAxis =  normalize(cross(normalize( PlayerPos.xyz - modelPos), float3(0,1,0)));
	float PlayerRotateAngle = saturate( distance(PlayerPos.xyz, modelPos) / 1.5);//cPlayerPushDistance );
	PlayerRotateAngle *= PlayerRotateAngle * PlayerRotateAngle;
	PlayerRotateAngle = (1 - PlayerRotateAngle) * cPlayerPushIntencity;
	return RotateAboutAxis( float4(PlayerRoateAxis, PlayerRotateAngle), modelPos, WorldPosition);
}

float3 pivotFoliageAnim(float LocalWindScale, float MaxRoation, float2 uv2, float3 modelPos, float3 WorldPosition, float layerMask, float softness,
						float4 Transform0, float4 Transform1, float4 Transform2, float objectScale, float4 impact, float ScarTime, inout float3 WorldNormal)
{
	float3 PivotRotateOffsetOUT = float3(0,0,0);
	float windRandomRotationRate = 10;
	float windRadomRotationInfluence = 3;
	float DampeningRadius = 8;
	float3 windDirection = normalize(-WindParam.xyz);
	float4 pivotPosSampler = tPivPosMap.SampleLevel(sPivPosMapSampler, uv2, 0);
	float4 pivotAxisSampler = tPivAxisMap.SampleLevel(sPivAxisMapSampler, uv2, 0);
	//decode pivotPos and axis
	pivotAxisSampler.xyz = (pivotAxisSampler.xyz - 0.5) * 2;
	if(@_INSTANCED){
		pivotPosSampler.xyz = Matrix4x3TransformP( Transform0, Transform1, Transform2,  float3(-pivotPosSampler.x,pivotPosSampler.z,-pivotPosSampler.y) * 0.01 ).xyz ;
		pivotAxisSampler.xyz = normalize( Matrix4x3TransformN( Transform0, Transform1, Transform2, float3(-pivotAxisSampler.x,pivotAxisSampler.z, -pivotAxisSampler.y) )  ); 
	}
	else{
		pivotPosSampler.xyz = MulMat3x4( World, float4( float3(-pivotPosSampler.x, pivotPosSampler.z, -pivotPosSampler.y) * 0.01, 0) ).xyz + modelPos;
		pivotAxisSampler.xyz = normalize( MulMat3x4( World, float4( float3(-pivotAxisSampler.x, pivotAxisSampler.z, -pivotAxisSampler.y),0) ).xyz ); 
	}
	
	//mask motion
	float PivotAxisLenght = DecodeAlphaAixsExtent(pivotAxisSampler.a);
	float3 localToPiotPosIncludScale = (WorldPosition - pivotPosSampler.xyz) / objectScale;
	float MaskMotionVal = saturate(dot(localToPiotPosIncludScale,  pivotAxisSampler.xyz ) / PivotAxisLenght * DampeningRadius);

	//WindDirMask
	float2 windShelter = float2(-1, -0.5);
	float windDirMask = (dot(-windDirection, pivotAxisSampler.xyz) + windShelter.x) * windShelter.y;

	//windMotion
	float3 windXaxis = float3(1,0,0);
	float3 windZaxis = float3(0,0,1);
	float shakeParam = 1;
	checkIfShakeTree( shakeParam, ScarTime, 10);
	float WindSpeed = 0.2;
	float WindHorziontalSpeed = 0.2;
	float2 WindVector = float2(WindSpeed * LocalWindScale, WindHorziontalSpeed * LocalWindScale) * WindParam.w * GameTime * shakeParam;
	WindVector += SphereMask(WorldPosition, pivotPosSampler.xyz, PivotAxisLenght, 0.5) * softness;
	float2 WindDotParam = float2( dot(windXaxis, (pivotPosSampler.xyz + pivotAxisSampler.xyz * PivotAxisLenght)), dot(windZaxis, (pivotPosSampler.xyz + pivotAxisSampler.xyz * PivotAxisLenght)) );
	float2 WindDotUV = (WindDotParam + WindVector + modelPos.xz);
	float4 AmbientWindSamplerAngleJitter = tWindNoiseMap.SampleLevel(sWindNoiseMapSampler, WindDotUV, 0);
	float AmbientWindAngleJitter = AmbientWindSamplerAngleJitter.a;

	float3 AmbientWindAxisJitter = float3(0,0,0);

	//RoateAxis
	float3 temp = pow( dot( -windDirection, pivotAxisSampler.xyz), 5) * float3(0,0, -0.2) - windDirection;
	float3 PivoAxisPrecotZero = cross( pivotAxisSampler.xyz, temp );
	float3 NormlaizeRoateAxis = normalize( PivoAxisPrecotZero + AmbientWindAxisJitter);

	//RoateAngle
	float RoateAngleINT = ( AmbientWindAngleJitter * MaxRoation * saturate(WindParam.w) ) * windDirMask; //TODO *5
	RoateAngleINT *= MaskMotionVal;

	PivotRotateOffsetOUT += RotateAboutAxis( float4( NormlaizeRoateAxis , RoateAngleINT * 2 * PI ) , pivotPosSampler.xyz, WorldPosition ) * layerMask;
	PivotRotateOffsetOUT += PlayerImpactForPiv( pivotPosSampler.xyz, WorldPosition, impact);

	return PivotRotateOffsetOUT;
}

float3 PivotTreeALL(float3 modelPos, inout float3 WorldNormal, float3 WorldPosition, float Scale, float2 uv2, float4 Transform0, float4 Transform1, float4 Transform2, float ScarTime, float4 impact)
{
	//cal pivot painter textureCoordinate
	float2 texSize = float2(1,1);
	tPivPosMap.GetDimensions(texSize.x, texSize.y);
	float firstIndex = CalculateMeshElementIndex(texSize, uv2);

	float2 Lev0_parentUV = float2(0,0);
	float Lev0_isChild  = 0;
	float Lev0_CurrentUV = 0;
	calclulatHierachyInfor(uv2, texSize, firstIndex, Lev0_parentUV, Lev0_isChild, Lev0_CurrentUV);

	float2 Lev1_parentUV = float2(0,0);
	float Lev1_isChild  = 0;
	float Lev1_CurrentUV = 0;
	calclulatHierachyInfor(Lev0_parentUV, texSize, Lev0_CurrentUV, Lev1_parentUV, Lev1_isChild, Lev1_CurrentUV);

	float2 Lev2_parentUV = float2(0,0);
	float Lev2_isChild  = 0;
	float Lev2_CurrentUV = 0;
	calclulatHierachyInfor(Lev1_parentUV, texSize, Lev1_CurrentUV, Lev2_parentUV, Lev2_isChild, Lev2_CurrentUV);
	float ReturnHtotalDepth = ceil( Lev0_isChild + Lev1_isChild + Lev2_isChild );
	
	//turnk PivotUV
	float trunkLayerMask = 1.0;
	float2 turnkPivotUV =float2(0,0);
	float2 branch0PivotUV= float2(0,0);
	lerpMultipleFloat2( uv2, Lev0_parentUV , Lev1_parentUV, Lev2_parentUV , ReturnHtotalDepth/3 , branch0PivotUV, turnkPivotUV );
	float3 pivotLev1 = pivotFoliageAnim(cLocalWindScaleL1_Piv, cMaxRotationL1_Piv, turnkPivotUV, modelPos, WorldPosition, trunkLayerMask, cSoftnessL1_Piv, Transform0, Transform1, Transform2, Scale, impact, ScarTime, WorldNormal);
	float3 pivotLev2 = float3(0,0,0);
	float3 pivotLev3 = float3(0,0,0);
	if(@_UsePivotAnimL2)
	{
		float L2LayerMask = saturate(ReturnHtotalDepth);
		float2 Leve2LeafPivotUV;
		lerpMultipleFloat2(uv2, Lev0_parentUV , Lev1_parentUV, Lev2_parentUV , saturate((ReturnHtotalDepth - 1) * 0.5), Leve2LeafPivotUV, branch0PivotUV );
		pivotLev2 = pivotFoliageAnim(cLocalWindScaleL2_Piv, cMaxRotationL2_Piv, Leve2LeafPivotUV, modelPos, WorldPosition, L2LayerMask, cSoftnessL2_Piv, Transform0, Transform1, Transform2, Scale, impact, ScarTime, WorldNormal);
	}
	// else if(@_UsePivotAnimL3)
	// {
	// 	float L3LayerMask = saturate(ReturnHtotalDepth - 1);
	// 	float L4LayerMask = saturate(ReturnHtotalDepth - 2);//use lev1 lev3
	// 	float2 Leve3LeafPivotUV = lerp(uv2, Lev0_parentUV, L4LayerMask);
	// 	pivotLev3 = pivotFoliageAnim(cLocalWindScaleL2_Piv, cMaxRotationL2_Piv, Leve3LeafPivotUV, modelPos, WorldPosition, trunkLayerMask, cSoftnessL2_Piv, Transform0, Transform1, Transform2, Scale, impact, ScarTime, WorldNormal);
	// }
	return pivotLev1 + pivotLev2 + pivotLev3;
}

float3 WorldPositionOffset(in VertexAssemble IN, in float3 WorldPosition, inout float3 WorldNormal)
{
	float3 WPoffsetOUT = float3(0,0,0);
	if(@_LOD_LEVEL <= LOD_Level_3)
	{
		half scale = 1.0;
		half objectScale = 1;
		half3 modelPos = half3(0,0,0);
		float ScarTime = 0;
		float4 Transform0 = half4(0,0,0,0);
		float4 Transform1 = half4(0,0,0,0);
		float4 Transform2 = half4(0,0,0,0);
		if (@_INSTANCED)
		{
			Transform0 = IN.Transform0; Transform1 = IN.Transform1; Transform2 = IN.Transform2;
			modelPos = half3(IN.Transform0.w, IN.Transform1.w, IN.Transform2.w);
			ScarTime = IN.LightMapScale.x;
			scale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, half3(1, 0, 0)));
			objectScale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, half3(1, 1, 1)));
		}
		else
		{
			modelPos = MulMat3x4(World, half4(0,0,0,1)).xyz;
			ScarTime = 0;
			scale = length(MulMat3x4((float3x3)World, half3(1, 0, 0)));
			objectScale = length(MulMat3x4((float3x3)World, half3(1, 1, 1)));
		}		
		// if(@_UsePivotAnimL1 && @_LIGHT_MAP)
		// {
		// 	float4 PlayerImpact = tFoliageMap.SampleLevel(sFoliageMapSampler, (WorldPosition.xz - FoliageCenter.xy) / 128.0 + 0.5, 0);
		// 	WPoffsetOUT = PivotTreeALL(modelPos, WorldNormal, WorldPosition.xyz, objectScale, IN.TexCoord2, Transform0, Transform1, Transform2, ScarTime, PlayerImpact);
		// }
		if(@_VERTEXCOLOR&&@_LOD_LEVEL< LOD_Level_3)
		{
			WPoffsetOUT = WindWaveMovementTree(modelPos, IN.Position.xyz, scale * LOD_RATE, GetVertexColor(IN.Color).xyz, WorldPosition, ScarTime);
		}
	}
	return WPoffsetOUT;
}

void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
		OUT.TintColor = GetVertexColor(IN.Color);
	float3 Offset = WorldPositionOffset(IN, OUT.WorldPosition.xyz, OUT.WorldNormal.xyz);
	OUT.WorldPosition.xyz += Offset;
	OUT.WorldOffset.xyz = Offset;
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	if(@_METAL)
		OUT.WorldOffset += half3(0.001, 0.001, 0.001); // for flicker
}
#include "PBR_Common.fxh"
//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	// 1. init MaterialParameters (always need)
	InitShadingParameters(MaterialParameters);
	MaterialParameters.Thickness = 0.1;

	// 2. init MaterialParameters.BaseColor and MaterialParameters.Alpha by sample basemap
	InitBaseColorCommon(IN, MaterialParameters);

	//R:Roughness G:Thickness B:AO
	// if(@_VERTEXCOLOR)
	// 	MaterialParameters.AO *= saturate(IN.TintColor.a);
	if(@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		float roughsfade =saturate(distance(CameraPos.xyz,IN.WorldPosition.xyz)*0.05) ;
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).rgb;
		half  Smoothness =lerp(samplerMis.r,0,roughsfade) ;
		MaterialParameters.Roughness = 1-Smoothness;
		MaterialParameters.Thickness = samplerMis.g * 0.3;
		MaterialParameters.AO *= samplerMis.b;
	}
	else
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);

	//Normal
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 tangentNormal = tNormalMap.Sample(sNormalSampler, BaseMapUV(IN.TexCoord)).xyz;
		tangentNormal = (2 * tangentNormal) - 1;
		MaterialParameters.WorldNormal = normalize(tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz);
	}
	else
		MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);

	//Apply Burnt
	InitBurntSimple(IN, MaterialParameters, 0.45);
	//Burnt
	float2 uv = (IN.WorldPosition.xz - FoliageCenter.zw) / 256 + 0.5;
	half burnIntensity = tFoliageBurnMap.Sample(sFoliageBurnMapSampler, uv).x;
	half3 burntColor = GetLuminance(MaterialParameters.BaseColor) * 0.5;
	MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, burntColor, saturate(burnIntensity));
	//Apply Weather
	InitRain(IN, MaterialParameters);
	InitRadarEffect(IN, MaterialParameters);
	MaterialParameters.WorldOffset = clamp(IN.WorldOffset, half3(-8,-8,-8), half3(8,8,8));
	

	
}
#include "PBR_PS.fxh"

#if SD_ENABLE_VXGI
#include "VXGI/VXGIVoxelize.fxh"
#endif

