#pragma shader material
#pragma shader predict("@_DF_SKIN", "0")
#pragma shader predict("@_TRANSPARENT", "0")
#pragma shader predict("@_CLIPPING", "1")
#pragma shader predict("@_VERTEXCOLOR", "1")

#define MATERIAL_SHADINGMODEL_TWOSIDED_FOLIAGE 1
#define USE_LIGHTMAP						0
#define USE_LIGHTMAP_CONST					1
#define USE_CLIPPING						1
#define USE_WIREFRAME						1
#define USE_FROZEN							0
#define USE_BURNT							0
#define USE_HIGHLIGHT						0
#define USE_XRAY							0
#define USE_EDGELINE						0
#define USE_PLANAR_SHADOW					0
#define USE_VBufferGen						0
#define VIRTUAL_LIT 						0
#define USE_ONE_SIDE_LIGHTING				1

#define HAS_WORLD_POSITION_OFFSET
#define USE_HISM							1

#define HAS_2UV								(@_LIGHT_MAP)
#define ENABLE_MIXMAP						0
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow && @_LOD_LEVEL <= LOD_Level_2 && !@_TRANSPARENT)
#define DEFERRED_PRE_AO						1
///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"
//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?"; 
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_EnableDepthBias
{
	Name = EnableDepthBias;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = false;	
};
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Default = false;
	Visible = false; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	
};
@_LightmapConstEnable
{
	Name = LightmapConstEnable;
	Default = false;
	Visible = false;
	Predict = false;
};
@_OneSideLightingEnable
{
	Name = OneSideLightingEnable;
	Description = "Do we need calc One side lighting?";
	Default = true;
	Visible = true;	
	Predict = true;
};
@_UsePivotAnim
{
	Name = UsePivotAnim;
	Default = false;
	Visible = true;
	Predict = false;
};
@_IsHISM
{
	Name = IsHISM;
	Description = "Is HISM?"; 
	
	Visible = false;
};
// @_WindNoiseMapEnable
// {
// 	Name = WindNoiseMapEnable;
// 	Description = "Do we have WindNoiseMap?"; 
// 	AffectedTex = tWindNoiseMap;
// 	Fallback = true;
// 	Visible = false;
// };
@_RadarEnable
{
	Name = RadarEnable;
	Description = "RadarEnable";
	Default = false;
	Visible = false;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < IsColor = "true"; > = {1, 1, 1};
	float cRoughnessLow				 = 0;	//粗糙度下界
	
	float cRoughnessHigh			 = 1;	//粗糙度上界
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = 0;
	float cAlpha					 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.1"; Precision = "2"; > = 1.0;
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	///VertexFoliage
	float cLocalWindPower			 = 1;
	float cGrassImpactPlayer		 = 10;
	//PivotFoliage
	float cMaxRotation_Piv		 	 = 0.3;
	float cLocalWindScale_Piv    	 = 1;
	//float cSoftness_Piv    	 	 = 0.5;
	//float cPlayerPushDistance		 = 1.5;
	
	float3 cRadarColor				< Widget = "LinearColor3"; > = {1, 1, 1};
	float cPlayerPushIntencity		 = 0.5;
	
	float	cRadarScrolling			={0};
	float 	cPlayerDirection			 = {1};
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tMixMap : layout(t1) = "Reflection";
sampler sMixSampler : layout(s1)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tNormalMap : layout(t2) = "NormalMap";
sampler sNormalSampler : layout(s2)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tFoliageMap : FoliageMap : layout(t3) = "Black";
sampler sFoliageMapSampler : layout(s3)
{
	Texture = tFoliageMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tFoliageBurnMap : FoliageBurnMap : layout(t4) = "Black";
sampler sFoliageBurnMapSampler : layout(s4)
{
	Texture = tFoliageBurnMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};


texture2D tEnvMap : EnvMap : layout(t5) = "White";
sampler sEnvSampler : layout(s5)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tLightMap : layout(t6);
sampler sLightMapSampler : layout(s6)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;  
	SRGB = false;  
	MaxMipLevel = 0;  
};

// ShadowMap 
texture2D tShadowMask : ShadowMask : layout(t7) = "White";
sampler sShadowMaskSampler : layout(s7)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
}; 

texture2D tShadowMap : ShadowMap : layout(t7) = "White";
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
}; 

texture2D tPivPosMap : layout(t8) = "White";
sampler sPivPosMapSampler : layout(s8)
{
	Texture = tPivPosMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tPivAxisMap : layout(t9) = "White";
sampler sPivAxisMapSampler : layout(s9)
{
	Texture = tPivAxisMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tWindNoiseMap : layout(t10) = "Black";
sampler sWindNoiseMapSampler : layout(s10)
{
	Texture = tWindNoiseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};




//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"

float CubicSmooth(float x)
{
	return x * x * (3.0 - 2.0 * x);
}

float TrigApproximate(float x)
{
	float triangleWave = abs((frac(x + 0.5) * 2.0) - 1.0);
	return (CubicSmooth(triangleWave) - 0.5) * 2.0;
}

float3 WindWaveMovementGrass(float3 worldPos, float3 localPos ,float attenMask)
{
	float2 windDir = -WindParam.xz * WindParam.w;
	float2 globalPase = GameTime * windDir*0.05 + worldPos.xz * 0.1;

	// float g1 =  TrigApproximate(globalPase.x * 1.975);
	// float g2 =  TrigApproximate(globalPase.y * 1.975);
	// float g3 =  TrigApproximate(globalPase.x * 0.375);
	// float g4 =  TrigApproximate(globalPase.y * 0.375);
	// float globalWave = g1 + g3 * g3;
	// float sencondaryWave = g2 + g4 * g4;
	// float waveNoise = globalWave + sencondaryWave;
	// float3 vNewPos = float3(waveNoise * windDir.x, max(1, WindParam.w) - 1, waveNoise * windDir.y) * localPos.y;
	float waveNoise = tWindNoiseMap.SampleLevel(sWindNoiseMapSampler, globalPase, 0).a * 2 - 1;
	 waveNoise *=0.1;
	float3 vNewPos = float3(waveNoise * windDir.x, waveNoise * WindParam.w, waveNoise * windDir.y);

	float4 impact = tFoliageMap.SampleLevel(sFoliageMapSampler, (worldPos.xz - FoliageCenter.xy)/128.0 + 0.5, 0);
	vNewPos += float3(impact.x, 0, impact.y) * 300;
	return vNewPos * attenMask;
}

float3 RotateAboutAxis(float4 NormalizedRotationAxisAndAngle, float3 PivotWorldPos, float3 WorldPos)
{
	// Project Position onto the rotation axis and find the closest point on the axis to Position
	float3 ClosestPointOnAxis = PivotWorldPos + NormalizedRotationAxisAndAngle.xyz * dot(NormalizedRotationAxisAndAngle.xyz, WorldPos - PivotWorldPos);
	// Construct orthogonal axes in the plane of the rotation
	float3 UAxis = WorldPos - ClosestPointOnAxis;
	float3 VAxis = cross(NormalizedRotationAxisAndAngle.xyz, UAxis);
	float CosAngle;
	float SinAngle;
	sincos(NormalizedRotationAxisAndAngle.w, SinAngle, CosAngle);
	// Rotate using the orthogonal axes
	float3 R = UAxis * CosAngle + VAxis * SinAngle;
	// Reconstruct the rotated world space position
	float3 RotatedPosition = ClosestPointOnAxis + R;
	// Convert from position to a position offset
	return RotatedPosition - WorldPos;
}

float3 FixRotateAboutAxisNormals(float4 NormalizedRotationAxisAndAngle, float3 WorldNormal)
{
	float3 rotateNormal = RotateAboutAxis(NormalizedRotationAxisAndAngle, float3(0,0,0), WorldNormal);
	WorldNormal = normalize(WorldNormal + rotateNormal);
	return WorldNormal;
}

float3 RotateAboutAxisNormal(float3 Axis, float Angle, float3 WorldNormal)
{
	float s = sin(Angle);
	float c = cos(Angle);
	float one_minus_c = 1.0 - c;
	float3x3 rot_mat = float3x3(
		one_minus_c * Axis.x * Axis.x + c, one_minus_c * Axis.x * Axis.y - Axis.z * s, one_minus_c * Axis.z * Axis.x + Axis.y * s,
		one_minus_c * Axis.x * Axis.y + Axis.z * s, one_minus_c * Axis.y * Axis.y + c, one_minus_c * Axis.y * Axis.z - Axis.x * s,
		one_minus_c * Axis.z * Axis.x - Axis.y * s, one_minus_c * Axis.y * Axis.z + Axis.x * s, one_minus_c * Axis.z * Axis.z + c
	);
	WorldNormal = mul(rot_mat, WorldNormal);
	return normalize(WorldNormal);
}

float DecodeAlphaAixsExtent(float AlphaExtentVal)
{
	return clamp( AlphaExtentVal * 2048, 8, AlphaExtentVal * 128);
}

float SphereMask(float3 A, float3 B, float AttenuationRadius, float Hardness)
{
	float Distance = length((A-B));
	float ArgInvRadius = 1/AttenuationRadius;
	float NormalizeDistance = Distance * ArgInvRadius;
	float ArgInvHardness = 1 - Hardness;
	float NegNormalizedDistance=1-NormalizeDistance;
	float MaskUnclamped = NegNormalizedDistance * ArgInvHardness;
	return clamp(MaskUnclamped, 0, 1);
}

float3 PlayerImpactForPiv( float3 modelPos ,float3 WorldPosition,float4 PlayerImpact)
{
	float3 PlayerRoateAxis =  normalize(cross(normalize( PlayerPos.xyz - modelPos), float3(0,1,0)));
	float PlayerRotateAngle = saturate( distance(PlayerPos.xyz, modelPos) / 1.5);//cPlayerPushDistance );
	PlayerRotateAngle *= PlayerRotateAngle * PlayerRotateAngle;
	PlayerRotateAngle = (1 - PlayerRotateAngle) * cPlayerPushIntencity;
	return RotateAboutAxis( float4(PlayerRoateAxis, PlayerRotateAngle), modelPos, WorldPosition);
}

float3 pivotFoliageAnim(float LocalWindScale, float MaxRoation, float2 uv2, float3 modelPos, float3 WorldPosition, float layerMask, float softness,
						float4 Transform0, float4 Transform1, float4 Transform2, float objectScale, float4 impact, float ScarTime, inout float3 WorldNormal)
{
	float3 PivotRotateOffsetOUT = float3(0,0,0);
	float windRandomRotationRate = 10;
	float windRadomRotationInfluence = 3;
	float DampeningRadius = 8;
	float3 windDirection = normalize(-WindParam.xyz);
	float4 pivotPosSampler = tPivPosMap.SampleLevel(sPivPosMapSampler, uv2, 0);
	float4 pivotAxisSampler = tPivAxisMap.SampleLevel(sPivAxisMapSampler, uv2, 0);
	//decode pivotPos and axis
	pivotAxisSampler.xyz = (pivotAxisSampler.xyz - 0.5) * 2;
	if(@_INSTANCED){
		pivotPosSampler.xyz = Matrix4x3TransformP( Transform0, Transform1, Transform2,  float3(-pivotPosSampler.x,pivotPosSampler.z,-pivotPosSampler.y) * 0.01 ).xyz ;
		pivotAxisSampler.xyz = normalize( Matrix4x3TransformN( Transform0, Transform1, Transform2, float3(-pivotAxisSampler.x,pivotAxisSampler.z, -pivotAxisSampler.y) )  ); 
	}
	else{
		pivotPosSampler.xyz = MulMat3x4( World, float4( float3(-pivotPosSampler.x, pivotPosSampler.z, -pivotPosSampler.y) * 0.01, 0) ).xyz + modelPos;
		pivotAxisSampler.xyz = normalize( MulMat3x4( World, float4( float3(-pivotAxisSampler.x, pivotAxisSampler.z, -pivotAxisSampler.y),0) ).xyz ); 
	}

	//rotateRoot
	float3 pivotAxis = float3(0,1,0);
	float3 xAxis = float3(-1,0,0);
	if(@_INSTANCED)
		pivotAxis = normalize( Matrix4x3TransformN( Transform0, Transform1, Transform2, float3(0,1,0) ).xyz );
	else
		pivotAxis = normalize( MulMat3x4( World, float4( float3(0,1,0),0) ).xyz );
	float atanXAxis = atan2(xAxis.x , xAxis.z);
	float atanWindAxis = atan2(windDirection.x ,windDirection.z);
	float angleOffset = 0;
	PivotRotateOffsetOUT += RotateAboutAxis(float4(pivotAxis,  atanWindAxis - atanXAxis + angleOffset), pivotPosSampler.xyz, WorldPosition);
	WorldNormal = RotateAboutAxisNormal(pivotAxis,  atanWindAxis - atanXAxis + angleOffset, WorldNormal);
	pivotAxisSampler.xyz = RotateAboutAxisNormal(pivotAxis,  atanWindAxis - atanXAxis + angleOffset, pivotAxisSampler.xyz);
	WorldPosition += PivotRotateOffsetOUT;
	
	//mask motion
	float PivotAxisLenght = DecodeAlphaAixsExtent(pivotAxisSampler.a);
	float3 localToPiotPosIncludScale = (WorldPosition - pivotPosSampler.xyz) / objectScale;
	float MaskMotionVal = saturate(dot(localToPiotPosIncludScale,  pivotAxisSampler.xyz ) / PivotAxisLenght * DampeningRadius);

	//WindDirMask
	float2 windShelter = float2(-1, -0.5);
	float windDirMask = (dot(-windDirection, pivotAxisSampler.xyz) + windShelter.x) * windShelter.y;

	//windMotion
	float3 windXaxis = float3(1,0,0);
	float3 windZaxis = float3(0,0,1);
	float shakeParam = 1;

	float WindSpeed = 0.2;
	float WindHorziontalSpeed = 0.2;
	float2 WindVector = float2(WindSpeed * LocalWindScale, WindHorziontalSpeed * LocalWindScale) * WindParam.w * GameTime * shakeParam;
	WindVector += SphereMask(WorldPosition, pivotPosSampler.xyz, PivotAxisLenght, 0) * 0.5;//softness;
	float2 WindDotParam = float2( dot(windXaxis, (pivotPosSampler.xyz + pivotAxisSampler.xyz * PivotAxisLenght)), dot(windZaxis, (pivotPosSampler.xyz + pivotAxisSampler.xyz * PivotAxisLenght)) );
	float2 WindDotUV = (WindDotParam + WindVector + modelPos.xz);
	float4 AmbientWindSamplerAngleJitter = tWindNoiseMap.SampleLevel(sWindNoiseMapSampler, WindDotUV, 0);
	float AmbientWindAngleJitter = AmbientWindSamplerAngleJitter.a;

	float3 AmbientWindAxisJitter = float3(0,0,0);

	float2 AmbientMove = windDirection.xz * GameTime * 0.5;
	float3 AmbientWindSamplerAxis = tWindNoiseMap.SampleLevel(sWindNoiseMapSampler, ((WindDotParam + WindVector + modelPos.xz) / windRandomRotationRate) + AmbientMove, 0).xyz;
	AmbientWindAxisJitter = (AmbientWindSamplerAxis - float3(0.5,0.5,0.5)) * windRadomRotationInfluence;


	//RoateAxis
	float3 temp = pow( dot( -windDirection, pivotAxisSampler.xyz), 5) * float3(0,0, -0.2) - windDirection;
	float3 PivoAxisPrecotZero = cross( pivotAxisSampler.xyz, temp );
	float3 NormlaizeRoateAxis = normalize( PivoAxisPrecotZero + AmbientWindAxisJitter);

	//RoateAngle
	float RoateAngleINT = ( AmbientWindAngleJitter * MaxRoation * saturate(WindParam.w) ) * windDirMask; //TODO *5
	RoateAngleINT *= MaskMotionVal;

	PivotRotateOffsetOUT += RotateAboutAxis( float4( NormlaizeRoateAxis , RoateAngleINT * 2 * PI ) , pivotPosSampler.xyz, WorldPosition ) * layerMask;

	WorldNormal = FixRotateAboutAxisNormals(float4( NormlaizeRoateAxis , RoateAngleINT * 2 * PI ), WorldNormal);
	PivotRotateOffsetOUT += PlayerImpactForPiv( pivotPosSampler.xyz, WorldPosition, impact);

	return PivotRotateOffsetOUT;
}

float3 WorldPositionOffset(in VertexAssemble IN, in float3 WorldPosition, inout float3 WorldNormal)
{
	float3 WPoffsetOUT = float3(0,0,0);
	if(@_LOD_LEVEL <= LOD_Level_3)
	{
		half scale = 1.0;
		half objectScale = 1;
		half3 modelPos = half3(0,0,0);
		float ScarTime = 0;
		float4 Transform0 = half4(0,0,0,0);
		float4 Transform1 = half4(0,0,0,0);
		float4 Transform2 = half4(0,0,0,0);
		if (@_INSTANCED)
		{
			Transform0 = IN.Transform0; Transform1 = IN.Transform1; Transform2 = IN.Transform2;
			modelPos = half3(IN.Transform0.w, IN.Transform1.w, IN.Transform2.w);
			ScarTime = IN.LightMapScale.x;
			scale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, half3(1, 0, 0)));
			objectScale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, half3(1, 1, 1)));
		}
		else
		{
			modelPos = MulMat3x4(World, half4(0,0,0,1)).xyz;
			ScarTime = 0;
			scale = length(MulMat3x4((float3x3)World, half3(1, 0, 0)));
			objectScale = length(MulMat3x4((float3x3)World, half3(1, 1, 1)));
		}		
		if(@_UsePivotAnim  && @_LOD_LEVEL <= LOD_Level_1)
		{
			float4 PlayerImpact = tFoliageMap.SampleLevel(sFoliageMapSampler, (WorldPosition.xz - FoliageCenter.xy) / 128.0 + 0.5, 0);
			WPoffsetOUT = pivotFoliageAnim(cLocalWindScale_Piv, cMaxRotation_Piv, IN.TexCoord2, modelPos, WorldPosition.xyz, 1, 0.5, Transform0, Transform1, Transform2, objectScale, PlayerImpact, ScarTime, WorldNormal); 
		}
		 if(@_VERTEXCOLOR && @_LOD_LEVEL <= LOD_Level_1)
		{
			WPoffsetOUT = WindWaveMovementGrass(WorldPosition.xyz, IN.Position.xyz, GetVertexColor(IN.Color).x);
		}
	}
	return WPoffsetOUT;
}

void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
		OUT.TintColor = GetVertexColor(IN.Color);
	OUT.WorldPosition.xyz += WorldPositionOffset(IN, OUT.WorldPosition.xyz, OUT.WorldNormal.xyz);
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	if(@_METAL)
		OUT.WorldOffset += half3(0.001, 0.001, 0.001); // for flicker
}
#include "PBR_Common.fxh"
//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	// 1. init MaterialParameters (always need)
	InitShadingParameters(MaterialParameters);
	MaterialParameters.Thickness = 0.1;

	// 2. init MaterialParameters.BaseColor and MaterialParameters.Alpha by sample basemap
	InitBaseColorCommon(IN, MaterialParameters);

	//R:Roughness G:Thickness B:AO
	if(@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		float roughsfade = distance(CameraPos.xyz,IN.WorldPosition.xyz)*0.05;
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).rgb;		
		//MaterialParameters.Specular *= samplerMis.r;// Specular Mask
		half  Smoothness =lerp(samplerMis.r,0,roughsfade) ;
		MaterialParameters.Roughness = 1-Smoothness;
		MaterialParameters.Thickness = samplerMis.g * 0.15;
		MaterialParameters.AO *= samplerMis.b;
	}
	else
	{
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);

	}
		

	//Normal
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.Sample(sNormalSampler, BaseMapUV(IN.TexCoord)).xyz;
		tangentNormal = (2 * tangentNormal) - 1;
		MaterialParameters.WorldNormal = normalize(tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz);
	}
	else
		MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);

	//Apply Burnt
	InitBurntSimple(IN, MaterialParameters, 0.45);
	//Burnt
	float2 uv = (IN.WorldPosition.xz - FoliageCenter.zw) / 256 + 0.5;
	half burnIntensity = tFoliageBurnMap.Sample(sFoliageBurnMapSampler, uv).x;
	half3 burntColor = GetLuminance(MaterialParameters.BaseColor) * 0.5;
	MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, burntColor, saturate(burnIntensity));
	//Apply Weather
	InitRain(IN, MaterialParameters);
	InitRadarEffect(IN, MaterialParameters);

	MaterialParameters.WorldOffset = clamp(IN.WorldOffset, half3(-8,-8,-8), half3(8,8,8));
}
#include "PBR_PS.fxh"

#if SD_ENABLE_VXGI
#include "VXGI/VXGIVoxelize.fxh"
#endif
