//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : szn3878@corp.netease.com
//	Author: Si Zheng		
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR_HeightBlend.fx>
///	@path	~/System/Shader/
///	@date	2017/02/20
///	@brief	PBR shader designed for structures.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			0

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = true;	
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_DetailMapEnable  
{
	Name = DetailMapEnable;
	Description = "Do we have DetailMap?"; 
	AffectedTex = tDetailMap;
	Fallback = true;
	Visible = false;
};
@_DetailNormalEnable  
{
	Name = DetailNormalEnable;
	Description = "Do we have Detail Normal?"; 
	AffectedTex = tDetailNormal;
	Fallback = true;
	Visible = true;
};
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cSelfEmit					 < Widget = "FloatSlider"; Min = "0"; Max = "2"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 

	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;
	float cdetailSmoothCtl			 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	
	float cdetaiRoughness			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cSmoothCtl				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cRoughness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float cMixShapness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.2;
	float cMixVtxControl			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.5;
	float cMixShift					 < Widget = "FloatSlider"; Min = "-4"; Max = "4"; Step = "0.01"; Precision = "2"; > = 0;
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;
	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.65;
	float3 cDetailColor 			 < Widget = "LinearColor3"; > = {1, 1, 1};
	float UVscaleU 					 < Widget = "FloatSlider"; Min = "0.1"; Max = "10"; Step = "0.001"; Precision = "2"; > = 1;
	float UVscaleV 					 < Widget = "FloatSlider"; Min = "0.1"; Max = "10"; Step = "0.001"; Precision = "2"; > = 1;
	float cDetailU					 < Widget = "FloatSlider"; Min = "0.1"; Max = "10"; Step = "0.001"; Precision = "2"; > = 1;
	float cDetailV					 < Widget = "FloatSlider"; Min = "0.1"; Max = "10"; Step = "0.001"; Precision = "2"; > = 1;
	float cBaseColorBrightness		 < Widget = "FloatSlider"; Min = "0"; Max = "3"; Step = "0.01"; Precision = "2"; > = 1;
	float cDetailColorBrightness	 < Widget = "FloatSlider"; Min = "0"; Max = "3"; Step = "0.01"; Precision = "2"; > = 1;
	float cAlpha	      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;
	float cSnowHeight	  			 < Widget = "FloatSlider"; Min = "0"; Max = "2"; Step = "0.001"; Precision = "3"; > = 0.6;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t1) = "White";
sampler sEnvSampler : layout(s1)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};
// Light map
texture2D tLightMap : layout(t2);
sampler sLightMapSampler : layout(s2)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tShadowMask : ShadowMask : layout(t3) = "White";
sampler sShadowMaskSampler : layout(s3)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t3) = "White";
SamplerComparisonState sShadowMapSampler : layout(s3)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};


texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Detail Map
texture2D tDetailNormal : layout(t5) = "NormalMap";
sampler sDetailNormalSampler : layout(s5)
{
	Texture = tDetailNormal;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tDetailMap : layout(t6) = "White";
sampler sDetailSampler : layout(s6, s7)
{
	Texture = tDetailMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	
	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	OUT.WorldPosition.w = IN.Position.y;

	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));

	OUT.TexCoord.x *= UVscaleU;
	OUT.TexCoord.y *= UVscaleV;
	
	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

		//overwrite Fog T
		OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
	}
}  
half colorMask(half3 incolor)
{
	return (incolor.x+incolor.y+incolor.z)/3;
}
//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	half detailAlpha = 1;
	half alphaA = 0;
	half alphaB = 0;
	if (@_BaseMapEnable)
	{
		half4 samplerBase;
		if (@_CLIPPING || @_DISSOLVING || @_TRANSPARENT)
			{samplerBase = tBaseMap.Sample(sBaseSampler, BaseMapUV(IN.TexCoord));}
		else
			{samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);}
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb) * lerp(1, cBaseColor ,colorMask(samplerBase.rgb));
		MaterialParameters.BaseColor *= cBaseColorBrightness;
		if(@_EnableReflection || @_DEFERRED_SHADING)
		{
			MaterialParameters.Roughness = cRoughness + saturate(samplerBase.r - cSmoothCtl);
		}
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = 1;
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		alphaA = tangentNormal.z;
		tangentNormal = (2*tangentNormal)-1;
		if (@_D3D11 || @_D3D9)
			tangentNormal.z = sqrt(1-dot(tangentNormal.xy,tangentNormal.xy));
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		MaterialParameters.WorldNormal = tangentNormal.x * cBumpIns * IN.WorldTangent + tangentNormal.y *cBumpIns* IN.WorldBinormal + IN.WorldNormal.xyz;
		float k = length(MaterialParameters.WorldNormal);
		MaterialParameters.WorldNormal /= k;
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
	float3 detailN = MaterialParameters.WorldNormal;
	float2 DetailTexCoord = float2(IN.TexCoord.x * cDetailU, IN.TexCoord.y * cDetailV);
	if(@_TANGENT && @_DetailNormalEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 detailTangent = tDetailNormal.SampleBias(sDetailNormalSampler, DetailTexCoord.xy, cNormalMapBias).xyz;
		alphaB = detailTangent.z;
		detailTangent = (2*detailTangent)-1;
		if (@_D3D11 || @_D3D9)
			{detailTangent.z = sqrt(1-dot(detailTangent.xy,detailTangent.xy));}
		if (@_LOD_LEVEL == LOD_Level_1)
			{detailTangent.xy *= LOD_RATE;}
		detailN = detailTangent.x * cBumpIns * IN.WorldTangent + detailTangent.y *cBumpIns* IN.WorldBinormal + IN.WorldNormal.xyz;
		float k = length(detailN);
		detailN /= k;
	}
	if (@_VERTEXCOLOR)
	{
		alphaB = alphaB + (IN.TintColor.r-0.5) * 4 * cMixVtxControl;
	}
	alphaB += cMixShift;
	detailAlpha = saturate((alphaB - alphaA) * 10 * cMixShapness);
	MaterialParameters.WorldNormal = lerp(MaterialParameters.WorldNormal,detailN,detailAlpha);
	if (@_DetailMapEnable)
	{
		half4 samplerDetail;
		if (@_CLIPPING || @_DISSOLVING || @_TRANSPARENT)
			{samplerDetail = tDetailMap.SampleRGBA(sDetailSampler, DetailTexCoord.xy);}
		else
			{samplerDetail = tDetailMap.SampleBias(sDetailSampler, DetailTexCoord.xy, cBaseMapBias);}
		samplerDetail.rgb = DecodeGamma(samplerDetail.rgb)* lerp(1, cDetailColor ,colorMask(samplerDetail.rgb)) * cDetailColorBrightness;
		MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, samplerDetail.rgb,detailAlpha);
		half detailRoughness =cdetaiRoughness + saturate(samplerDetail.r - cdetailSmoothCtl);
		MaterialParameters.Roughness = lerp(MaterialParameters.Roughness,detailRoughness,detailAlpha);
	}
	MaterialParameters.Emission = half3(0, 0, 0);
	MaterialParameters.Emission += cSelfEmit;
}
#include "PBR_PS.fxh"
