//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			1
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define USE_TESSELLATION		1
#define IS_HEIGHTMAP 			1

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Predict = true;
	Fallback = true;
	Visible = false;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	Predict = true;
	Fallback = true;
	Default = true;
	Visible = true;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = true;
	Predict = true;
	Fallback = true;
	Visible = true;	
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	Predict = false;
	Fallback = false;
	Default = false;
	Visible = true;
};
@_DetailNormalMapEnable
{
	Name = DetailNormalMapEnable;
	Description = "Do we have DetailNormalMap?"; 
	AffectedTex = tDetailNormalMap;
	Predict = true;
	Fallback = true;
	Default = true;
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_VSHeightmapEnable
{
	Name = VSHeightmapEnable;
	Description = "Apply heightmap to vs?";
	Predict = true;
	Fallback = true;
	Default = true;
	Visible = true;
};
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 

	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;

	float2 cBaseMapRepeat			  = {1, 1};
	float2 cDetailMapRepeat			  = {1, 1};

	float3 cWorldScale				 = {1, 1, 1};
	float cBloomScale	  			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1;

	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;
	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1;
	float cDepthOffset     			 < Widget = "FloatSlider"; Min = "-5"; Max = "5"; Step = "0.1"; Precision = "2"; > = 0;
	float cDepthBias	  			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "2"; > = 0;
	float2 cHeightmapSize			 = {256, 256};
	float cSunShadowZRange			 = 1.0;
	float cNearTessDis				 = 10.0;
	float cFarTessDis				 = 100.0;
	float cNearTriPerMeter			 < Widget = "FloatSlider"; Min = "0"; Max = "100"; Step = "0.1"; Precision = "2"; > = 10.0;
	float cFarTriPerMeter			 = 0.25;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tHeightMap : layout(t2) = "White";
sampler sHeightSampler : layout(s2)
{
	Texture = tHeightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tDetailNormalMap : layout(t4) = "NormalMap";
sampler sDetailNormalSampler : layout(s4)
{
	Texture = tDetailNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tShadowMask : ShadowMask : layout(t6) = "White";
sampler sShadowMaskSampler : layout(s6)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tRainMap : RainMap: layout(t7) = "NormalMap";
sampler sRainSampler : layout(s7)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
#define HEIGHTMAP_MODIFIER(IN, INTM)\
{\
	if (@_VSHeightmapEnable)\
	{\
		float localHeight = tHeightMap.SampleLevel(sHeightSampler, INTM.LocalPosition.xz, 0).x;\
		INTM.LocalPosition.y = localHeight * cWorldScale.y;\
	}\
}

void VS_DIY_OUTPUT(in VertexAssemble IN, inout InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
		OUT.TintColor = GetVertexColor(IN.Color);
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	
	//world * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	OUT.TexCoord.xy = INTM.LocalPosition.xz;
	OUT.ViewDirection.w = INTM.LocalPosition.y;
	
	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}

	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

		//overwrite Fog T
		OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
	}
}

//////////////////////////////////HS////////////////////////////////////////
float4 HS_GetTessellationFactor(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2)
{
	float dis0 = length(p0.ViewDirection.xyz);
	float dis1 = length(p1.ViewDirection.xyz);
	float dis2 = length(p2.ViewDirection.xyz);
	float dis = min(min(dis0, dis1), dis2);
	float triPerMeter = lerp(cNearTriPerMeter, cFarTriPerMeter, saturate((dis-cNearTessDis)/(cFarTessDis-cNearTessDis)));
	float len0 = length(p1.WorldPosition.xyz-p2.WorldPosition.xyz);
	float len1 = length(p2.WorldPosition.xyz-p0.WorldPosition.xyz);
	float len2 = length(p0.WorldPosition.xyz-p1.WorldPosition.xyz);
	return float4(len0*triPerMeter, len1*triPerMeter, len2*triPerMeter, (len0+len1+len2)*(0.333*triPerMeter));
}

//////////////////////////////////DS////////////////////////////////////////
void DS_DIY_OUTPUT(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2, float3 domain, inout VertexShadingOutput OUT)
{
	float localHeight = tHeightMap.SampleLevel(sHeightSampler, OUT.TexCoord.xy, 0).x * cWorldScale.y;
	OUT.WorldPosition.y += localHeight-OUT.ViewDirection.w;
}

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	float2 uv = BaseMapUV(IN.TexCoord);
	//Metallic, Roughness
	MaterialParameters.Metallic = cMetallic;
	MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	//N
	float localHeight;
	if (@_LOD_LEVEL <= LOD_Level_2)
	{
		float2 texelSize = 1/cHeightmapSize;
		float4 h;
		h[0] = tHeightMap.Sample(sHeightSampler, uv + float2(-texelSize.x, 0)).x;
		h[1] = tHeightMap.Sample(sHeightSampler, uv + float2(+texelSize.x, 0)).x;
		h[2] = tHeightMap.Sample(sHeightSampler, uv + float2(0, -texelSize.y)).x;
		h[3] = tHeightMap.Sample(sHeightSampler, uv + float2(0, +texelSize.y)).x;
		MaterialParameters.WorldNormal.x = (h[0] - h[1]) * cWorldScale.y / (cWorldScale.x * texelSize.x);
		MaterialParameters.WorldNormal.z = (h[2] - h[3]) * cWorldScale.y / (cWorldScale.z * texelSize.y);
		MaterialParameters.WorldNormal.y = 2;
		localHeight = 0.25 * (h[0] + h[1] + h[2] + h[3]);
		float3 detailNormal = tDetailNormalMap.SampleLevel(sDetailNormalSampler, uv*cDetailMapRepeat, 0).xyz;
		MaterialParameters.WorldNormal.xz += (detailNormal.xy*2-1) * cBumpIns;
		MaterialParameters.Roughness *= detailNormal.z * (1-0.5*localHeight);
		if (@_LOD_LEVEL == LOD_Level_2)
			MaterialParameters.WorldNormal.xz *= LOD_RATE;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
		localHeight = IN.ViewDirection.w  / cWorldScale.y;
	}
	//BaseColor
	if (@_BaseMapEnable)
	{
		half4 samplerBase = tBaseMap.Sample(sBaseSampler, uv.xy * cBaseMapRepeat);
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
		MaterialParameters.Alpha = samplerBase.a;
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = 1.0;
	}

	//snow overlay
	half overlayFactor = saturate(MaterialParameters.WorldNormal.y * (localHeight+0.2));
	MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, half3(1,1,1), overlayFactor);
}
#include "PBR_PS.fxh"

