//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : jiawei@corp.netease.com
//	Author: jiawei
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR_ice.fx>
///	@path	~/System/Shader/
///	@date	2016/06/19
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define NO_SHADOW_CLIPPING 		1

#pragma shader predict("@_TANGENT", "true")
#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "true")
#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")
 
///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Visible = true; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = true;
	Visible = true;
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
}; 
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{		
	float3 cSnowColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;

	float3 cIeColor					 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0;

	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float  cSnowAlpha				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	 
	float  cSnowness				 < Widget = "FloatSlider"; Tips = "test"; Min = "0"; Max = "5"; Step = "0.01"; Precision = "2"; > = 1.5;
	float  cAffectBaseUV			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;
	float  cSnowRough    			 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;

	float2 cUVReppet				 = {1, 1};
	float  cIceHeight				 < Widget = "FloatSlider"; Min = "-10"; Max = "30"; Step = "0.01"; Precision = "2"; > = 0.3291;
	float  cAlphaTrans				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5) = "White";
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;  
	SRGB = false;  
	MaxMipLevel = 0;  
}; 

texture2D tShadowMask : ShadowMask : layout(t6) = "White";
sampler sShadowMaskSampler : layout(s6)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tRainMap : RainMap: layout(t7) = "NormalMap";
sampler sRainSampler : layout(s7)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	
	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;

	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
}

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	half2 UVT = BaseMapUV(IN.TexCoord) * cUVReppet;
	half3 samplerMis = half3(0, 0, 1);
	if (@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		samplerMis = tMixMap.Sample(sMixSampler, UVT).xyz;
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = 0;
		MaterialParameters.AO = 1.0f;
		if (@_LOD_LEVEL == LOD_Level_2)
		{ 
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = Smoothness;
	}
	else
	{ 
		MaterialParameters.Roughness = cSmoothness;
	}
	half heightRatioInput =  cIceHeight ;
	half heightFactor = heightRatioInput * -0.5 + heightRatioInput * samplerMis.g;
	half3 V = normalize( IN.ViewDirection.xyz );
	V.xy = -V.xy ;
	half3 oN = normalize( IN.WorldNormal.xyz );
	half NoV = dot(V, oN);
	half2 offsetUv = BaseMapUV(IN.TexCoord);
	[@_TANGENT]
	offsetUv += half3(dot(IN.WorldTangent.xyz, V), dot(IN.WorldBinormal.xyz, V), NoV).xy  * heightFactor;
	half IceRate = 0;
	if(@_VERTEXCOLOR)
	{
		IceRate = 1- IN.TintColor.r;
	}
	else
	{ 
		IceRate = 1 - saturate(IN.WorldNormal.y * cSnowness);
	}
	half DryRate = 1.0 - IceRate;
	half3 newNov = (V - oN * NoV) * (1 - IceRate);
	DryRate = smoothstep(0.58, cSnowness, IceRate);
	half snowShine = samplerMis.b * -IceRate * ( -sin(frac(newNov)- 2.0f ) - 1.2f ) * cSmoothness;
	MaterialParameters.Roughness = lerp(cSnowRough, MaterialParameters.Roughness, IceRate);
	if(@_TANGENT && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		if (@_RIPPLE)
		{
			half2 tc = IN.WorldPosition.xz * RainNormalXZScale;
			half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
			half2 ttc = tc + delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			ttc = tc - delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			tangentNormal /= (1 + 2 * RainNormalRate);
		}
		tangentNormal = (2*tangentNormal)-1;
		half3 waterNormal = half3(0, 0, 1);
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		MaterialParameters.WorldNormal = tangentNormal.x*IN.WorldTangent.xyz + tangentNormal.y*IN.WorldBinormal.xyz + tangentNormal.z*IN.WorldNormal.xyz;\
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else 
	{
		MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
	}
	half4 samplerBase;
	samplerBase = tBaseMap.SampleBias(sBaseSampler, offsetUv, -1);
	MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb) / max(0.001, samplerMis.r) * cIeColor;
	MaterialParameters.Alpha = lerp( 1 , cAlphaTrans, IceRate);
	MaterialParameters.BaseColor = lerp(cSnowColor+snowShine, MaterialParameters.BaseColor, clamp(DryRate, cSnowAlpha, 1.0));
	MaterialParameters.Emission = half3(0, 0, 0);
}

#include "PBR_PS.fxh"

