#pragma shader material
#define USE_LIGHTMAP			0
#define USE_CLIPPING			0
#define USE_WIREFRAME			0
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			0

#define NEED_SCREEN_SHADOWMAP	(1)
// #define MATERIAL_SHADINGMODEL_DEFAULT_LIT    1
#define MATERIAL_SHADINGMODEL_IMP_LIT    1
// #define MATERIAL_SHADINGMODEL_TWOSIDED_FOLIAGE  1


#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_EnableSphereLighting
{
	Name = EnableSphereLighting;
	Description = "Enable SphereLighting?";
	Default = true;
};
@_EmissionMapEnable
{
	Name = EmissionMapEnable;
	Description = "Do we have EmissionMap?";
	AffectedTex = tEmissionMap; 
	Fallback = true;
	Visible = false;
};

@_ImpostorFullSphere
{
    Name = ImpostorFullSphere;
    Description = "Impostor Full Sphere?";
    Default = false;
}


@_EnableNormalMap
{
	Name = EnableNormalMap;
	Description = "EnableNormalMap?";
	AffectedTex = tNormalMap; 
	Default = false;
	Visible = true;
};


cbuffer Batch
{
	// WolrdToObject
	float4x3 Local 					: InvWorld				;
	//worldviewproj
	float4x3 World 					: World		            ;
	float4 cParamter2				: Param2				;
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
	float4 cPlaceHold											;
#endif	
	float4 PlayerPos                : PlayerPos					;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	// float4 ImpostorParam		;
	float4 cImpostorOffset		 =  { 0,0 ,100,1}; // x z y
	float  cImpostorMeshSize	 = 100;
    float cGlobalMoveFrequency	 = 1;
	float cGlobalMoveFactor		 = 5;
	float Padding_VS			 = 0;

	float  cImpostorBorderClamp   = 0;
	float  cCutoff  			  = 0.01;
    float  cImpotorMapSize        = 512;  
	float  cShadowHeight		  =  2;
	float  cImpostorFrames		 = 16;
    float  cBoudingBoxSize       = 5;
    float  cRentangeTree         = 0;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};
// Emission map
texture2D tEmissionMap : layout(t2) = "Transparent";
sampler sEmissionMapSampler : layout(s2)
{
	Texture = tEmissionMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
};

texture2D tNormalMap : layout(t3) = "NormalMap";
sampler sNormalMapSampler : layout(s3)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};


// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t4) = "White";
SamplerComparisonState sShadowMapSampler : layout(s4)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};


struct VertexAssemble
{
	float4		Position 		: POSITION;
	
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
    [@_INSTANCED]
    float4      InvTransform0   : TEXCOORD6;
    [@_INSTANCED]
    float4      InvTransform1   : TEXCOORD7;
    [@_INSTANCED]
    float4      InvTransform2   : TEXCOORD8;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position

    float4 		frame0			: TEXCOORDN;

    float4 		frame1			: TEXCOORDN;

    float4 		frame2			: TEXCOORDN;

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.

    [@_INSTANCED]
    float4 Transform0           :TEXCOORDN;
    [@_INSTANCED]
    float4 Transform1           :TEXCOORDN;
    [@_INSTANCED]
    float4 Transform2           :TEXCOORDN;

	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4 		WorldPivot	 	: TEXCOORDN;			//  WorldPivot

	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;

	float4      LocalPosition    :TEXCOORDN;

	// // float3      LocalNormal       :TEXCOORDN;
    // float4 ViewProjPS_0 		 : TEXCOORDN; 
    			
    // float4 ViewProjPS_1 		 : TEXCOORDN;

    // float4 ViewProjPS_2 		 : TEXCOORDN; 

    // float4 ViewProjPS_3 		 : TEXCOORDN; 		
};


//////////////////////////////////VS////////////////////////////////////////
#include "Lighting.fxh" 

float3 SpriteProjection( float3 pivotToCameraRayLocal, float frames, float2 size, float2 coord )
{    
    //octahedron vector, pivot to camera
    float3 y = normalize(pivotToCameraRayLocal);
    float3 x = normalize( cross( y, float3(0.0, 1.0, 0.0) ) );
    float3 z = normalize( cross( x, y ) );
    float2 uv = ((coord*frames)-0.5) * 2.0; //-1 to 1 
    float3 newX = x * uv.x;
    float3 newZ = z * uv.y;    
    newX *= size.x*0.5;
    newZ *= size.y*0.5;  
    float3 res = newX + newZ;  
    return res;
}


float3 WindParamNormlize()
{
    return float3( normalize( WindParam.xyz + float3( 0.01, 0.01, 0.01)));
}

//for hemisphere
float2 VecToHemiOct( float3 vec )
{
	vec.xz /= dot( 1.0, abs(vec) );
	return float2( vec.x + vec.z, vec.x - vec.z );
}

float2 VecToSphereOct( float3 vec )
{
    vec.xz /= dot( 1,  abs(vec) );
    if ( vec.y <= 0 )
    {
        float2 flip = sign(vec.xz);//vec.xz >= 0 ? float2(1,1) : float2(-1,-1);
        vec.xz = (1-abs(vec.zx)) * flip;
    }
    return vec.xz;
}

float2 VectorToGrid( float3 vec )
{
    float2 coord =float2(0,0);
    if ( @_ImpostorFullSphere )
    {
        coord = VecToSphereOct( vec );
    }
    else
    {
        vec.y = max(0.001,vec.y);
        vec = normalize(vec);
        coord = VecToHemiOct( vec );
    }
    return coord;
}

float4 TriangleInterpolate( float2 uv )
{
    uv = frac(uv);

    float2 omuv = float2(1.0,1.0) - uv.xy;
    
    float4 res = float4(0,0,0,0);
    //frame 0
    res.x = min(omuv.x,omuv.y); 
    //frame 1
    res.y = abs( dot( uv, float2(1.0,-1.0) ) );
    //frame 2
    res.z = min(uv.x,uv.y); 
    //mask
    res.w = saturate(ceil(uv.x-uv.y));
    
    return res;
}

//for hemisphere
float3 OctaHemiEnc( float2 coord )
{
 	coord = float2( coord.x + coord.y, coord.x - coord.y ) * 0.5;
 	float3 vec = float3( coord.x, 1.0 - dot( float2(1.0,1.0), abs(coord.xy) ), coord.y  );
 	return vec;
}
//for sphere
float3 OctaSphereEnc( float2 coord )
{
    float3 vec = float3( coord.x, 1-dot(1,abs(coord)), coord.y );
    if ( vec.y < 0 )
    {
        float2 flip = sign(vec.xz);//vec.xz >= 0 ? float2(1,1) : float2(-1,-1);
        vec.xz = (1-abs(vec.zx)) * flip;
    }
    return vec;
}

float3 GridToVector( float2 coord )
{
    float3 vec;
    if ( @_ImpostorFullSphere )
    {
        vec = OctaSphereEnc(coord);
    }
    else
    {
        vec = OctaHemiEnc(coord);
    }
    return vec;
}

//frame and framecout, returns 
float3 FrameXYToRay( float2 frame, float2 frameCountMinusOne )
{
    //divide frame x y by framecount minus one to get 0-1
    float2 f = frame.xy / frameCountMinusOne;
    //bias and scale to -1 to 1
    f = (f-0.5)*2.0; 
    float3 vec = float3(0,0,0);
    //convert to vector, either full sphere or hemi sphere
    vec = GridToVector( f );
    vec = normalize(vec);
    return vec;
}

float3 ITBasis( float3 vec, float3 basedX, float3 basedY, float3 basedZ )
{
    return float3( dot(basedX,vec), dot(basedY,vec), dot(basedZ,vec) );
}

float3 FrameTransform( float3 projRay, float3 frameRay, out float3 worldX, out float3 worldZ  )
{
    worldX = normalize( float3(-frameRay.z, 0, frameRay.x) );
    worldZ = normalize( cross(worldX, frameRay ) ); 
    
    projRay *= -1.0; 
    
    float3 local = normalize( ITBasis( projRay, worldX, frameRay, worldZ ) );
    return local;
}

float2 VirtualPlaneUV( float3 planeNormal, float3 planeX, float3 planeZ, float3 center, float2 uvScale, float3 Origin, float3 Direction )
{
    float normalDotOrigin = dot(planeNormal,Origin);
    float normalDotCenter = dot(planeNormal,center);
    float normalDotRay = dot(planeNormal,Direction);
    
    float planeDistance = normalDotOrigin-normalDotCenter;
    planeDistance *= -1.0;  
    float intersect = planeDistance / normalDotRay; 
    float3 intersection = ((Direction * intersect) + Origin) - center;
    float dx = dot(planeX,intersection);
    float dz = dot(planeZ,intersection);
    float2 uv = float2(0,0);
  
    if( intersect > 0 ){ uv = float2(dx,dz); }
    else{ uv = float2(0,0);}
    
    uv /= uvScale;
    uv += float2(0.5,0.5);
    return uv;
}

float3 WindWaveMovementTree( float3 modelVertexPos ,float3 WorldPosition,  float3 worldPos,  float scale, float vertexColor)
{
    worldPos = abs(worldPos) * sign(worldPos);
	float3 windDir = float3( WindParamNormlize().x, 0, WindParamNormlize().z);
	float treeSoftWave = 0.01 * cGlobalMoveFactor * WindParam.w*2;
	float3 _treePostion_Instance = worldPos * 0.25;
	float WaveSpeed = GameTime * cGlobalMoveFrequency*2;  		
	float WaveSpeedSecond = WaveSpeed * 6.5/3;//6.5  for unreal to  excharge num
	float WavePos =_treePostion_Instance.x + _treePostion_Instance.z  + WaveSpeedSecond;
	float treeWoldPosWave = sin(WavePos);//+ windTemp.x
	treeWoldPosWave =  pow(treeWoldPosWave + 1.0, 1.5) + 30.0 + WindParam.w;
	float3 windDirDown = float3(0, -1, 0);

	float3 waveHorizontal = windDir * treeSoftWave * treeWoldPosWave * vertexColor;
	float3 waveDown = treeSoftWave * 0.1 * windDirDown * abs(treeWoldPosWave) * vertexColor;
	float3 waveforce = waveHorizontal + waveDown ;//+ waveDown
	return waveforce * scale;
}


void GetDynamicShadowSimple(float3 LitUVZ, float3 LitUVZ2, float2 vInvShadowMapWH, out half InDynamicShadow, out half ShadowRate)
{
	float inShadow;
	float2 pInLitTC = LitUVZ.xy;
	float litZ = (LitUVZ.z);
	half2 inRange = abs(pInLitTC - 0.5);
	ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
    float depth;
    float depth2;
    SampleShadowMap( pInLitTC , depth);
    inShadow = litZ >depth?1:0;
	InDynamicShadow = inShadow;
}

void ImpostorVS ( in float4 vertexPos,in float3 objectSpaceCameraPos, in float2 texcoordIN, 
out float3 vertexOUT, out float2 texcoord,out float2 grid,out float2 vUv0, out float2 vUv1, out float2 vUv2, 
out float3 frame0local, out float3 frame1local, out float3 frame2local , out float vertexColor,out float3 impostorPivotOffset)
{
    impostorPivotOffset = float3(-cImpostorOffset.x,cImpostorOffset.z,-cImpostorOffset.y)*0.01; //flip Unreal offsetParam
    // impostorPivotOffset = float3(0,cImpostorOffset.z,0)*0.01; //flip Unreal offsetParam
    
    float framesMinusOne = cImpostorFrames-1;

    //pivot to camera ray
    float3 pivotToCameraRay = normalize(objectSpaceCameraPos.xyz-impostorPivotOffset.xyz);

    //scale uv to single frame  
    texcoord = float2(texcoordIN.x,texcoordIN.y)*(1.0/cImpostorFrames.x);   

    //radius 
    float2 size = (cImpostorMeshSize.xx*0.01) ; //  unreal to messiah *0.01   objectScale.xx;

    float3 projected = SpriteProjection( pivotToCameraRay, cImpostorFrames, size, texcoord.xy );

    //this creates the proper offset for vertices to camera facing billboard
    float3 vertexOffset = projected + impostorPivotOffset;

    // camera to projection vector
    float3 rayDirectionLocal = (impostorPivotOffset + projected) - objectSpaceCameraPos;
	//projected position to camera ray
    float3 projInterpolated = normalize( objectSpaceCameraPos - (projected + impostorPivotOffset) ); 

    float3 Origin = objectSpaceCameraPos-impostorPivotOffset; 
    float3 Direction = rayDirectionLocal; 

    grid = VectorToGrid( pivotToCameraRay );
    // float2 gridRaw = grid;
 	grid = saturate((grid+1.0)*0.5); //bias and scale to 0 to 1 
    grid *= framesMinusOne;
    
    float2 gridFrac = frac(grid);
    float2 gridFloor = floor(grid);
    float4 weights = TriangleInterpolate( gridFrac ); 

	//3 nearest frames
    float2 frame0 = gridFloor;
    float2 frame1 = gridFloor + lerp(float2(0,1),float2(1,0),weights.w); //lerp(float2(0,1),float2(1,0),weights.w)
    float2 frame2 = gridFloor + float2(1,1);

    //convert frame coordinate to octahedron direction
    float3 frame0ray = FrameXYToRay(frame0, framesMinusOne.xx);
    float3 frame1ray = FrameXYToRay(frame1, framesMinusOne.xx);
    float3 frame2ray = FrameXYToRay(frame2, framesMinusOne.xx);
    
    float3 planeCenter = float3(0,0,0);
    float3 plane0x;
    float3 plane0normal = frame0ray;
    float3 plane0z;
    frame0local = FrameTransform( projInterpolated, frame0ray, plane0x, plane0z );
    frame0local.xz /= cImpostorFrames.xx; //for displacement
    
    //virtual plane UV coordinates
    vUv0 = VirtualPlaneUV( plane0normal, plane0x, plane0z, planeCenter, size, Origin, Direction );
    vUv0 /= cImpostorFrames.xx;   
    
    float3 plane1x; 
    float3 plane1normal = frame1ray;
    float3 plane1z;
    frame1local = FrameTransform( projInterpolated, frame1ray, plane1x, plane1z);
    frame1local.xz = frame1local.xz/cImpostorFrames.xx; //for displacement
    
    //virtual plane UV coordinates
    vUv1 = VirtualPlaneUV( plane1normal, plane1x, plane1z, planeCenter, size, Origin, Direction );
    vUv1 /= cImpostorFrames.xx;
    
    float3 plane2x;
    float3 plane2normal = frame2ray;
    float3 plane2z;
    frame2local = FrameTransform( projInterpolated, frame2ray, plane2x, plane2z );
    frame2local.xz = frame2local.xz/cImpostorFrames.xx; //for displacement

    //virtual plane UV coordinates
    vUv2 = VirtualPlaneUV( plane2normal, plane2x, plane2z, planeCenter, size, Origin, Direction );
    vUv2 /= cImpostorFrames.xx;

    //vertexColor 
    vertexColor = step(0,-vertexPos.y);

    //subtract from camera pos 
    vertexOffset = normalize(objectSpaceCameraPos-vertexOffset);
    // //then add the original projected world
    vertexOffset += projected;
    // //remove position of vertex
    vertexOffset -= vertexPos.xyz;
    // //add pivot
    vertexOffset += impostorPivotOffset;
    //add offset here
    vertexOUT = vertexPos.xyz + vertexOffset;
}


void ImpostorPS( in float2 TexCoordIN, in float2 fram0_IN, in float2 fram1_IN, in float2 fram2_IN  ,out float2 vp0uv , out float2 vp1uv , out float2 vp2uv, out float4 weights)
{
    float4 _impostorBaseTex_TexelSize = float4( 1/cImpotorMapSize ,1/cImpotorMapSize ,cImpotorMapSize, cImpotorMapSize);

    // float2 fracGrid = frac(IN.TexCoord.zw);
    float2 fracGrid = TexCoordIN;
    
    weights = TriangleInterpolate( fracGrid );
      
    float2 gridSnap = floor(TexCoordIN) / cImpostorFrames.xx;


	float2 frame0 = gridSnap;
    float2 frame1 = gridSnap + (lerp(float2(0,1),float2(1,0),weights.w)/cImpostorFrames.xx);
    float2 frame2 = gridSnap + (float2(1,1)/cImpostorFrames.xx);


    vp0uv = frame0 + fram0_IN;
    vp1uv = frame1 + fram1_IN; 
    vp2uv = frame2 + fram2_IN;

	//resolution of atlas (Square)
    float textureDims = _impostorBaseTex_TexelSize.z;
    //fractional frame size, ex 2048/12 = 170.6
    float frameSize = textureDims/cImpostorFrames; 
    //actual atlas resolution used, ex 170*12 = 2040
    float actualDims = floor(frameSize) * cImpostorFrames; 
    //the scale factor to apply to UV coordinate, ex 2048/2040 = 0.99609375
    float scaleFactor = actualDims / textureDims;

	vp0uv *= scaleFactor;
    vp1uv *= scaleFactor;
    vp2uv *= scaleFactor;


	//clamp out neighboring frames
    float2 gridSize = 1.0/cImpostorFrames.xx;
    gridSize *= _impostorBaseTex_TexelSize.zw;
    gridSize *= _impostorBaseTex_TexelSize.xy;
    float2 border = _impostorBaseTex_TexelSize.xy*cImpostorBorderClamp;


    //clamp out neighboring frames 
    vp0uv = clamp(vp0uv,frame0+border,frame0+gridSize-border);
    vp1uv = clamp(vp1uv,frame1+border,frame1+gridSize-border);
    vp2uv = clamp(vp2uv,frame2+border,frame2+gridSize-border);
	
    vp0uv = 1- vp0uv;   //flip uv dir
    vp1uv = 1- vp1uv;
    vp2uv = 1- vp2uv;
}  

void InitScreenGray(in VertexShadingOutput IN, in out half3 BaseColor)
{
	if(ScreenMGrayModeSpread)
	{
		half grayDist = (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - min(length(IN.WorldPosition.xyz - PlayerPos.xyz), ScreenMGrayMaxDistance);
		grayDist = step(0, grayDist);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);		
	}
	else if (ScreenMGrayModeShrink)
	{
		half grayDist = length(IN.WorldPosition.xyz - PlayerPos.xyz) + (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - ScreenMGrayMaxDistance;
		grayDist = step(grayDist, 0);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);
	}
}  

void ImpClip(float alpha , float2 HPositionXY)
{
    float clipValue =  saturate(pow(alpha + cCutoff,2)) - 0.3333;//for match unreal param MaskOffset
    // float clipValue =  alpha - 0.3333;//for match unreal param MaskOffset

    if (@_DISSOLVING){
        clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(HPositionXY));
    }
    clip(clipValue); 
}

void VertexUnifyOutput(in VertexAssemble IN, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	//incoming vertex, object space
    float4 vertex =IN.Position;
    
    //camera in object space
    float3 objectSpaceCameraPos = float3(0, 0, 0);

    if (@_INSTANCED)
    {
        objectSpaceCameraPos =Matrix4x3TransformP(IN.InvTransform0, IN.InvTransform1, IN.InvTransform2, CameraPos.xyz).xyz;
    }
    else
    {
        objectSpaceCameraPos = MulMat3x4( Local, float4(CameraPos.xyz,1) ).xyz;
    }

    //inital param
    float2 texcoord = IN.TexCoord.xy;
    float3 impostorPivotOffset  = float3(0,0,0);
    float2 vUv0 =float2(0,0);
    float2 vUv1 =float2(0,0);
    float2 vUv2 =float2(0,0);
    float3 frame0local= float3(0,0,0);
    float3 frame1local= float3(0,0,0);
    float3 frame2local= float3(0,0,0);
    float2 grid =float2(0,0);
    float2 texcoordOUT =float2(0,0);
    float vertexColor = 0;
    float3 vertexOUT =float3(0,0,0);

  //impostorVS
    ImpostorVS(IN.Position,objectSpaceCameraPos , texcoord ,
    vertexOUT ,texcoordOUT , grid , vUv0 ,vUv1 ,vUv2 ,frame0local ,frame1local ,frame2local,vertexColor ,impostorPivotOffset);

	//position
	if (@_INSTANCED)
	{
		OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, vertexOUT.xyz), 1); //IN.Position.xyz
    }
	else
	{
	    OUT.WorldPosition = float4(MulMat3x4 (World, float4(vertexOUT.xyz, 1)).xyz, 1);  //float4(IN.Position.xyz, 1)  
	}

    //Movement  
    float3 modelPos =float3(0,0,0);
	if (@_INSTANCED)
		modelPos = float3(IN.Transform0.w, IN.Transform1.w, IN.Transform2.w);
	else
		modelPos = MulMat3x4(World, float4(0,0,0,1)).xyz;
    float3 WindMovementTree = WindWaveMovementTree( vertexOUT.xyz , OUT.WorldPosition, modelPos, 1.5, vertexColor);

    //HPosition
    OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz + WindMovementTree.xyz, 1.0));


    if(@_INSTANCED)
    {
        OUT.Transform0 = IN.Transform0;
        OUT.Transform1 = IN.Transform1;
        OUT.Transform2 = IN.Transform2;
    }

	OUT.TexCoord.xy = texcoordOUT;
    OUT.TexCoord.zw = grid;
    OUT.frame0 = float4(vUv0.xy,frame0local.xz);
    OUT.frame1 = float4(vUv1.xy,frame1local.xz);
    OUT.frame2 = float4(vUv2.xy,frame2local.xz);

	//ViewDirection
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;

	//Fog T
	//OUT.ViewDirection.w = FogTransmittance(normalize(OUT.ViewDirection.xyz)*cFogDistance);
	OUT.ViewDirection.w = FogTransmittance(OUT.ViewDirection.xyz);

	//WorldPivot
	if (@_INSTANCED)
	{
		OUT.WorldPivot = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2,(cImpostorOffset.xzy*0.01)), 1);
	}
	else
	{
		OUT.WorldPivot = float4(MulMat3x4(World, float4((cImpostorOffset.xzy*0.01),1)).xyz, 1);
	}


    //normal
	float3 LocalNormal = (float3)Expand((float3)IN.Normal.xyz);
	if (@_INSTANCED)
	{
		OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, LocalNormal));
	}
	else
	{
		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, LocalNormal));
	}

    if (@_TANGENT)
    {
        if(@_INSTANCED)
        {
            OUT.WorldTangent.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, IN.Tangent));
            OUT.WorldBinormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, IN.Binormal));
        }
        else
        {
            OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, IN.Tangent.xyz));
            OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, IN.Binormal.xyz));
        }
    }

    OUT.LocalPosition  = IN.Position;
}


void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	//init MaterialParameters (always need)
	InitShadingParameters(MaterialParameters);

    float2 vp0uv =float2(0,0);
    float2 vp1uv =float2(0,0);
    float2 vp2uv =float2(0,0);
    float4 weights=float4(0,0,0,0);
    float2 TexCoordIN = IN.TexCoord.zw;
    float2 fram0_IN = IN.frame0.xy;
    float2 fram1_IN = IN.frame1.xy;
    float2 fram2_IN = IN.frame2.xy;
    ImpostorPS( TexCoordIN, fram0_IN, fram1_IN, fram2_IN, vp0uv,vp1uv,vp2uv, weights);

    //impostor tex 
    float4 samplerBase =float4(0,0,0,0);
    float4 worldNormalTex  = float4(0,0,0,0);

    float4 baseSamp0 = tBaseMap.SampleLevel( sBaseSampler, vp0uv, 0 );
    float4 baseSamp1 = tBaseMap.SampleLevel( sBaseSampler, vp1uv, 0 );
    float4 baseSamp2 = tBaseMap.SampleLevel( sBaseSampler, vp2uv, 0 );
    samplerBase = baseSamp0*weights.x + baseSamp1*weights.y + baseSamp2*weights.z; //ori

	MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
	MaterialParameters.Alpha = samplerBase.a;
	MaterialParameters.Roughness = 0.8;

	//screen gray
	InitScreenGray(IN, MaterialParameters.BaseColor);

	//Normal
	if (@_EnableNormalMap && @_TANGENT)
    {
        float4 samp0 = tNormalMap.SampleLevel( sNormalMapSampler, vp0uv, 0 );
        float4 samp1 = tNormalMap.SampleLevel( sNormalMapSampler, vp1uv, 0 );
        float4 samp2 = tNormalMap.SampleLevel( sNormalMapSampler, vp2uv, 0 );
        worldNormalTex = samp0*weights.x + samp1*weights.y + samp2*weights.z; 

        float3  N = normalize( float3(worldNormalTex.r,1-worldNormalTex.b,worldNormalTex.g)*2-1);  //ok
	    N = normalize(N.x * IN.WorldBinormal + N.y * IN.WorldTangent + N.z * IN.WorldNormal.xyz);
		MaterialParameters.WorldNormal = N;
		MaterialParameters.AO = worldNormalTex.a;
    } 
	else
	{ MaterialParameters.WorldNormal = IN.WorldNormal.xyz; }

    // //TODO
	// //Apply Weather
	// MaterialParameters.BaseColor *= 1 - 0.3 * RainRate;
	// MaterialParameters.Roughness *= 1 - 0.3 * RainRate;
	// //Apply Burn
	// half BurnRadius = GlobalBurnParam.w;
	// half3 BurnPosition = GlobalBurnParam.xyz;
	// if(BurnRadius>0.01)
	// {
	// 	half BurnWidth = 80;
	// 	half BurnArea = saturate((length(IN.WorldPosition-BurnPosition) - BurnRadius) / BurnWidth);
	// 	MaterialParameters.BaseColor *= lerp(0.04 , 1 ,BurnArea);
	// }
}


VertexShadingOutput ImpostorVertexShading_VS(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//OUTPUT
	VertexUnifyOutput(IN, OUT);
	return OUT;
}

//////////////////////////////////PS////////////////////////////////////////
float4 ImpostorPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	MaterialShadingParameters MaterialParameters;
	PS_GetParameters(IN, MaterialParameters);

    MaterialParameters.DiffuseColor = Diffuse_Lambert(MaterialParameters.BaseColor - MaterialParameters.BaseColor * MaterialParameters.Metallic);
	float NoL = dot(MaterialParameters.WorldNormal, SunDirection.xyz);
    float MicroShadow = saturate( MaterialParameters.AO*MaterialParameters.AO*2 + abs(NoL) -1 );

    DynamicGILight( MaterialParameters.WorldNormal, MaterialParameters.GILighting , MaterialParameters.InShadow );
	MaterialParameters.GILighting *= MaterialParameters.DiffuseColor;

    //////// merge Color ////////
    half3 Diffuselighting = MaterialParameters.DiffuseColor * NoL * MicroShadow * SunColor.xyz;
	OUT.rgb = MaterialParameters.GILighting + Diffuselighting;

    // ////////    clip    ////////
    ImpClip( MaterialParameters.Alpha , IN.HPosition.xy);

    ////////    Weather    ////////
	OUT.rgb *= ThunderRate;

    if (@_Weather == WEATHER_RAIN){
        MaterialParameters.DiffuseColor  *= 1 - 0.3*RainRate;
        Diffuselighting *= 0.5;
    }

    if(@_Weather == WEATHER_SNOW){
        half overlayFactor = 0.6 * SnowRate;
        MaterialParameters.DiffuseColor  = lerp( MaterialParameters.DiffuseColor , half3(1,1,1) * NoL , overlayFactor );
    }

    //TODO Volm Fog
    if (@_FOG)
    {    float3 fogFinalColor = float3(0,0,0);
        FogLightingAndScatter(IN.ViewDirection.w, fogFinalColor, normalize( -IN.ViewDirection.xyz ), SunDirection.rgb);
        OUT.rgb = lerp(fogFinalColor, OUT.rgb, IN.ViewDirection.w);
    }
	EncodeHDR(OUT.rgb);

	return OUT;
}

technique Shading
{
	VertexShader = ImpostorVertexShading_VS();
	PixelShader = ImpostorPixelShading();
}




struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	// GBufferA: Emission.rgb
	float4 GBufferB 	: SV_Target1;	// GBufferB: Metallic, Specular, Roughness, ShadingModelID
	float4 GBufferC 	: SV_Target2;	// GBufferC: BaseColor.rgb, AO
	float4 GBufferD 	: SV_Target3;	// GBufferD: CustomData
	float4 GBufferE 	: SV_Target4;	// GBufferE: N.xyz, InShadow
	float4 GBufferF  	: SV_Target5;	// GBufferF: CustomData1
	float4 GBufferG 	: SV_Target6;	// GBufferG: Depth(r32f)
};


GBufferOUTPUT GBufferGenPixelShading(in VertexShadingOutput IN)
{
	GBufferOUTPUT OUT;

	//[@_HLSL]
	//OUT = (GBufferOUTPUT)0;

	if (@_DEFERRED_SHADING)
	{
		//1, Gen MaterialParameters & GBufferData
		MaterialShadingParameters MaterialParameters;
        InitShadingParameters(MaterialParameters);

		GBufferData GBuffer;
        PS_GetParameters(IN, MaterialParameters);


        float4 hPosition = MulMat4x4( ViewProj, float4(IN.WorldPosition.xyz,1) );
        DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting ,MaterialParameters.InShadow);
    
        //////// clip ////////
        ImpClip( MaterialParameters.Alpha , IN.HPosition.xy);

        if (@_Weather == WEATHER_RAIN)
        {
            MaterialParameters.BaseColor *= 1 - 0.3*RainRate;
            MaterialParameters.GILighting *= 0.5;
        }

        if(@_Weather == WEATHER_SNOW)
        {
            half overlayFactor = 0.6 * SnowRate;
		    MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, half3(1,1,1), overlayFactor);
        }

        SetGBufferForShadingModel(GBuffer, MaterialParameters);
		EncodeGBufferData(GBuffer, OUT.GBufferA, OUT.GBufferB, OUT.GBufferC, OUT.GBufferD, OUT.GBufferE, OUT.GBufferF);
        
        // float LinearZ = IN.HPosition.w * CameraInfo.z ; //ori
        float LinearZ = (hPosition.w) * CameraInfo.z; // for ios Z bug
		
        //linear z
		OUT.GBufferG = EncodeDepth(LinearZ);

        //Correct HDR & EncodeHDR color
        EncodeHDR(OUT.GBufferA.rgb);
	}
	else
	{
		OUT.GBufferA = float4(0, 0, 0, 0);
		OUT.GBufferB = float4(1, 0, 0, 0); // roughness默认是1，免得ssr计算
		OUT.GBufferC = float4(0, 0, 0, 0);
		OUT.GBufferD = float4(0, 0, 0, 0);
		OUT.GBufferE = float4(0, 0, 0, 0);
		OUT.GBufferF = float4(0, 0, 0, 0);
		OUT.GBufferG = float4(0, 0, 0, 0);
	}

	return OUT;
}

technique GBufferGen
{
	VertexShader = ImpostorVertexShading_VS();
	PixelShader = GBufferGenPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Debug pixel shader.
float4 ImpostorDebugPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
    float2 vp0uv =float2(0,0);
    float2 vp1uv =float2(0,0);
    float2 vp2uv =float2(0,0);
    float4 weights=float4(0,0,0,0);
    float2 TexCoordIN = IN.TexCoord.zw;
    float2 fram0_IN = IN.frame0.xy;
    float2 fram1_IN = IN.frame1.xy;
    float2 fram2_IN = IN.frame2.xy;
    ImpostorPS( TexCoordIN, fram0_IN, fram1_IN, fram2_IN, vp0uv,vp1uv,vp2uv, weights);


    float4 baseTex =float4(0,0,0,0);
    float4 worldNormalTex  = float4(0,0,0,0);
	float4 baseSamp0 = tBaseMap.SampleLevel( sBaseSampler, vp0uv, 0 );
    float4 baseSamp1 = tBaseMap.SampleLevel( sBaseSampler, vp1uv, 0 );
    float4 baseSamp2 = tBaseMap.SampleLevel( sBaseSampler, vp2uv, 0 );

    float4 samp0 = tNormalMap.SampleLevel( sNormalMapSampler, vp0uv, 0 );
    float4 samp1 = tNormalMap.SampleLevel( sNormalMapSampler, vp1uv, 0 );
    float4 samp2 = tNormalMap.SampleLevel( sNormalMapSampler, vp2uv, 0 );

    baseTex = baseSamp0*weights.x + baseSamp1*weights.y + baseSamp2*weights.z; //ori
    worldNormalTex = samp0*weights.x + samp1*weights.y + samp2*weights.z; 

    // float3 N = normalize(float3(worldNormalTex.gbr*2-1));
    float3 N = normalize(float3(worldNormalTex.rbg*2-1));//ori
    float3 V = normalize( IN.ViewDirection.xyz );
	float3 L = normalize(SunDirection.rgb);
	float NoL = saturate(dot(N,L)); 
	// float VdotL = dot(V,L);

    half3 DiffuseColor = Diffuse_Lambert(( baseTex.rgb) );

    if ((int)DebugFlag.x == 1)								//1-base color
    {
        OUT.rgb = baseTex.rgb;
    }
    else if ((int)DebugFlag.x == 4)							//4-diffuse color
    {
        OUT.rgb = DiffuseColor.rgb;
    }
    else if ((int)DebugFlag.x == 6)							//6-normal
    {
        OUT.rgb = 0.5+0.5*N;
    }
    else
    {
        OUT = float4( DiffuseColor, baseTex.a);
    }

    float clipValue =  saturate(pow(baseTex.a + cCutoff,2)) - 0.333;//for match unreal param MaskOffset
	clip(baseTex.a - cCutoff - 0.333); //for match unreal param MaskOffset
	
	EncodeHDR(OUT.rgb);
	return OUT;
}
technique Debug
{
	VertexShader = ImpostorVertexShading_VS();
	PixelShader = ImpostorDebugPixelShading();
}



/////////////////////////////////////ShadowGen/////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth

    float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

    float4 		WorldPosition 	: TEXCOORDN;			//  World Position

    float4 		frame0			: TEXCOORDN;

    float4 		frame1			: TEXCOORDN;

    float4 		frame2			: TEXCOORDN;

    [@_INSTANCED]
    float4 Transform0           :TEXCOORDN;
    [@_INSTANCED]
    float4 Transform1           :TEXCOORDN;
    [@_INSTANCED]
    float4 Transform2           :TEXCOORDN;
	
};


//ViewProj is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//incoming vertex, object space
    float4 vertex =IN.Position ;
    
    //camera in object space
    float3 objectSpaceCameraPos = float3(0, 0, 0);

    if (@_INSTANCED)
    {
        objectSpaceCameraPos =Matrix4x3TransformP(IN.InvTransform0, IN.InvTransform1, IN.InvTransform2, CameraPos.xyz).xyz;
    }
    else
    {
        objectSpaceCameraPos = MulMat3x4( Local, float4(CameraPos.xyz,1) ).xyz;
    }


    float2 texcoord = IN.TexCoord.xy;
    float3 impostorPivotOffset  = float3(0,0,0);
    float2 vUv0 =float2(0,0);
    float2 vUv1 =float2(0,0);
    float2 vUv2 =float2(0,0);
    float3 frame0local= float3(0,0,0);
    float3 frame1local= float3(0,0,0);
    float3 frame2local= float3(0,0,0);
    float2 grid =float2(0,0);
    float2 texcoordOUT =float2(0,0);
    float vertexColor = 0;
    float3 vertexOUT =float3(0,0,0);

    //impostor vertx
    ImpostorVS(IN.Position,objectSpaceCameraPos , texcoord ,
    vertexOUT ,texcoordOUT , grid , vUv0 ,vUv1 ,vUv2 ,frame0local ,frame1local ,frame2local,vertexColor,impostorPivotOffset);


    if (@_INSTANCED)
    {
        OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, vertexOUT.xyz), 1); //IN.Position.xyz
    }
    else
    {
        OUT.WorldPosition = float4(MulMat3x4 (World, float4(vertexOUT.xyz, 1)).xyz, 1);  //float4(IN.Position.xyz, 1)
    }
    OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
    
    
    if(@_INSTANCED)
    {
        OUT.Transform0 = IN.Transform0;
        OUT.Transform1 = IN.Transform1;
        OUT.Transform2 = IN.Transform2;
    }
    
	OUT.SceneDepth = OUT.HPosition.zw;
	return OUT;
}



[target("R32F")]
float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
    float4 OUT = float4(0, 0, 0, 0);
    float2 vp0uv =float2(0,0);
    float2 vp1uv =float2(0,0);
    float2 vp2uv =float2(0,0);
    float4 weights=float4(0,0,0,0);
    float2 TexCoordIN = IN.TexCoord.zw;
    float2 fram0_IN = IN.frame0.xy;
    float2 fram1_IN = IN.frame1.xy;
    float2 fram2_IN = IN.frame2.xy;
    ImpostorPS( TexCoordIN, fram0_IN, fram1_IN, fram2_IN, vp0uv,vp1uv,vp2uv, weights);

    float4 worldNormalTex  = float4(0,0,0,0);
	float4 baseSamp0 = tBaseMap.SampleLevel( sBaseSampler, vp0uv, 0 );
    float4 baseSamp1 = tBaseMap.SampleLevel( sBaseSampler, vp1uv, 0 );
    float4 baseSamp2 = tBaseMap.SampleLevel( sBaseSampler, vp2uv, 0 );
    float4 baseTex = baseSamp0*weights.x + baseSamp1*weights.y + baseSamp2*weights.z; //ori

    float clipValue =  saturate(pow(baseTex.a + cCutoff,2)) - 0.3333;//for match unreal param MaskOffset
    clip(clipValue);
    return half4 (1,1,1,1);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader  = ShadowMapGenPixelShading();
}



// //////////////////////////////////////////////////////////////////////////
// // Wireframe
// void VertexUnifyOutput(in VertexAssemble IN, inout InterMediateVertex INTM, out VertexShadingOutput OUT)
// {
// 	[@_HLSL]
// 	OUT = (VertexShadingOutput)0;

// 	if (@_INSTANCED)
// 	{
// 		if (@_Billboard)
// 		{
// 			OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, float3(0, 0, 0)), 1);
// 		}
// 		else
// 		{
// 			OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
// 		}
// 	}
// 	else
// 	{
// 		OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
// 	}

// 	// TexCoord
// 	OUT.TexCoord = IN.TexCoord.xyxy;

// 	// Normal
// 	if (@_INSTANCED)
// 	{
// 		if (@_Billboard)
// 		{
// 			float3 dir = CameraPos.xyz - OUT.WorldPosition.xyz;
// 			dir.y = 0;	// project to xz plane.
// 			OUT.WorldNormal.xyz = normalize(dir);

// 			float scale = length(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, float3(1, 0, 0)));
// 			OUT.WorldNormal.w = scale;
// 			if (@_TANGENT)
// 			{
// 				OUT.WorldTangent.xyz = float3(0,0,0);
// 				OUT.WorldBinormal.xyz = float3(0,0,0);
// 			}
// 		}
// 		else
// 		{
// 			OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
// 			if (@_TANGENT)
// 			{
// 				OUT.WorldTangent.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalTangent));
// 				OUT.WorldBinormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalBinormal));
// 			}
// 		}
// 	}
// 	else
// 	{
// 		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
// 		if (@_TANGENT)
// 		{
// 			OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
// 			OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));
// 		}
// 	}

// 	// ViewDirection
// 	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;

// 	// PerVertex Diffuse Lighting
// 	OUT.DiffLighting.xyz = float3(0, 0, 0);
// #if !SD_TILE_BASED_LIGHTING
// 	if (@_LOD_LEVEL < LOD_Level_3)
// 		OUT.DiffLighting.xyz = LightingVS(OUT.WorldPosition.xyz, OUT.WorldNormal.xyz);
// #endif		

// 	// Fog T
// 	if (!@_NoNearPlaneClip)
// 	{
// 		OUT.DiffLighting.w = FogTransmittance(OUT.ViewDirection.xyz);  

// 	}
// 	else
// 		OUT.DiffLighting.w = 0.999;

// 	// [@_INSTANCED]
// 	// OUT.LightMapScale = IN.LightMapScale;

// 	// DIY Output
// 	// VS_DIY_OUTPUT(IN, INTM, OUT);




// 	[@_ES20] //es2 bug sample lightmap with .zw leads to mosaic
// 	OUT.TexCoord.xyzw = OUT.TexCoord.zwxy;

// 	[@_Billboard] // add some bias to avoid depth fighting
// 	OUT.HPosition.z += 0.0001;
// }



// VertexShadingOutput WireframeVertexShading(in VertexAssemble IN)
// {
// 	VertexShadingOutput OUT;

// 	//InterMediate
// 	InterMediateVertex INTM;
// 	VertexUnifyInterMediate(IN, INTM);
// 	// VertexUnifySkinning(IN, INTM);



// 	//OUTPUT
// 	VertexUnifyOutput(IN, INTM, OUT);

// 	if (cParamter2.w < 0.1)
// 	{
// 	}
// 	//[0.1-0.2) 1uv
// 	else if (cParamter2.w < 0.2)
// 	{
// 		OUT.HPosition.xy = IN.TexCoord.xy * float2( 1.f, -1.f ) - float2(1.f, 0.f);
// 		OUT.HPosition.z = 1.f;
// 		OUT.HPosition.w = 1.f;
// 	}
// 	//[0.2-0.3) 2uv
// 	else if (cParamter2.w < 0.3)
// 	{
// 		if (@_LIGHT_MAP)
// 		{
// 			OUT.HPosition.xy = IN.TexCoord2.xy * float2( 1.f, -1.f ) - float2(1.f, 0.f);
// 			OUT.HPosition.z = 1.f;
// 			OUT.HPosition.w = 1.f;
// 		}
// 	}
// 	return OUT;
// }

// [maxvertexcount(3)]
// void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
// {
// 	IN[0].TexCoord = float4(1, 0, 0, 0);
// 	OUT.Append(IN[0]);
// 	IN[1].TexCoord = float4(0, 1, 0, 0);
// 	OUT.Append(IN[1]);
// 	IN[2].TexCoord = float4(0, 0, 1, 0);
// 	OUT.Append(IN[2]);

// 	OUT.RestartStrip();
// }

// float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
// {
// 	float3 difference = fwidth(IN.TexCoord.xyz);
// 	float3 distance = IN.TexCoord.xyz / difference;
// 	float d = min(min(distance.x, distance.y), distance.z);
// 	float a = exp2(-1.5 * d * d);
// 	return float4(cParamter2.xyz, a);
// }

// technique Wireframe
// {
// 	VertexShader = WireframeVertexShading();
// 	GeometryShader = WireframeGeometryShading();
// 	PixelShader = WireframePixelShading();
// };
