//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				1
#define USE_BURNT				1
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define VIRTUAL_LIT 			1
#define DOUBLE_REFLECTCOLOR 	0

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_IndexMapEnable 
{ 
	Name = IndexMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tIndexMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Default = false;
	Visible = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = true;	
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 

	float cAliasingFactor			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;//抗锯齿因子
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float cRoughnessX				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.1;
	float cRoughnessY				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 1;
	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.65;	
	float cAlpha					 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;

	float2 cBaseMapSize				  = {4, 4};
	float2 cAtlasMapSize			  = {4, 4};
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

// IndexMap
texture2D tIndexMap : layout(t2) = "White";
sampler sIndexSampler : layout(s2)
{
	Texture = tIndexMap;
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
};

texture2D tMixMap : layout(t3) = "Reflection";
sampler sMixSampler : layout(s3)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t4) = "White";
sampler sEnvSampler : layout(s4)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t5) = "NormalMap";
sampler sNormalSampler : layout(s5)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t6);
sampler sLightMapSampler : layout(s6)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tShadowMask : ShadowMask : layout(t7) = "White";
sampler sShadowMaskSampler : layout(s7)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t7) = "White";
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

@_LightmapConstEnable
{
	Name = LightmapConstEnable;
	Default = false;
	Visible = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
		OUT.TintColor = GetVertexColor(IN.Color);
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	//rotate ViewDirection
	//float sinF = cRotateSin;
	//float cosF = sqrt(1-sinF*sinF) * sign(cRotateCos);
	//float4 rotMatr = float4(cosF, -sinF, sinF, cosF);
	//float2 rotateXZ;
	//rotateXZ.x = dot(rotMatr.xy, OUT.ViewDirection.xz);
	//rotateXZ.y = dot(rotMatr.zw, OUT.ViewDirection.xz);
	//OUT.ViewDirection.xz = rotateXZ;

	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
}  

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	if (@_BaseMapEnable)
	{
		float2 uv = BaseMapUV(IN.TexCoord);
		float2 dx = float2(0, 0);
		float2 dy = float2(0, 0);
		if (@_IndexMapEnable)
		{
			float4 raw = tIndexMap.Sample(sIndexSampler, uv);
			float2 index = floor(raw.xy * cAtlasMapSize);
			float2 subUv = frac(uv * cBaseMapSize);
			if (raw.z > 0.5)//rotate 90
			{
				subUv = float2(subUv.y, 1-subUv.x);
			}
			dx = ddx(uv) * cBaseMapSize / cAtlasMapSize;
			dy = ddy(uv) * cBaseMapSize / cAtlasMapSize;
			uv = (index + subUv) / cAtlasMapSize;
		}
		half4 samplerBase;
		if (@_CLIPPING || @_TRANSPARENT)
			samplerBase = tBaseMap.SampleRGBA(sBaseSampler, uv);
		else
			samplerBase = tBaseMap.SampleGrad(sBaseSampler, uv, dx, dy);
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = cAlpha;
	}
	MaterialParameters.BaseColor *= IN.TintColor.rgb;
	MaterialParameters.AO = 1.0;
	if (@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).rgb;
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_2)
		{
			MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else
	{
		MaterialParameters.Metallic = cMetallic;
		if(@_EnableReflection || @_DEFERRED_SHADING)
			MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		// if (@_RIPPLE)
		// {
		// 	half2 tc = IN.WorldPosition.xz * RainNormalXZScale;
		// 	half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
		// 	half2 ttc = tc + delta;
		// 	tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
		// 	ttc = tc - delta;
		// 	tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
		// 	tangentNormal /= (1 + 2*RainNormalRate);
		// }
		tangentNormal = (2 * tangentNormal) - 1;
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		tangentNormal.xy *= cBumpIns;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
	MaterialParameters.Emission = half3(0, 0, 0);
}
#include "PBR_PS.fxh"
