//	--------------------------------------------------------------------
//	Copyright(C) 2006~2016 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Project Messiah
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Lava.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.
#define USE_SHADOWMAP_SAMPLE				(@_LOD_LEVEL <= LOD_Level_2 && @_CastSunShadow)
#define ENABLE_SCREEN_SHADOWMAP				(@_ReceiveShadow && @_CSM && @_CastSunShadow && @_LOD_LEVEL <= LOD_Level_2)
#pragma shader material
#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "false")
#pragma shader predict("@_INSTANCED", "false")
#include "Constants.fxh"
#include "Common.fxh"

@_EnableDepthBias
{
	Name = EnableDepthBias;
	Description = "for EnableDepthBias";
	Default = false;
};
@_ReceiveShadow
{
	Name = ReceiveShadow;
	Description = "for WeaponPreview";
	Default = true;
	Visible = false;
};

cbuffer Shader	
{
	
	float Lens					 = { 1 };
	float cFov					 <Visible = "0";> = { 1 };
	float cDepthBias			 <Visible = "0";> = { 1 };
	float cImageScale			 = { 1 };
	float cDarkCorner			 = { 0.5 };
	float cAlpha				 = { 1.0 };
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {0.15, 0.2, 0.15};
	float2 cLensMask			= {1, 1};
	float cIBLintensity			= {1};


};


// Surface textures.
texture2D tDiffuseMap : layout(t0);
sampler sDiffuseSampler : layout(s0, s1)
{
	Texture = tDiffuseMap;
	Filter =Linear ;
	Address = Wrap;
	SRGB = false;
};
texture2D tNormalMap : layout(t2) = "NormalMap";
sampler sNormalSampler : layout(s2)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};


texture2D tEnvMap : EnvMap : layout(t3) =  "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};
texture2D DepthBuffer : SceneDepth : layout(t4);
sampler DepthBuffer_Sampler : layout(s4)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};
texture2D ScreenTexture : SceneColor : layout(t5) = "Transparent";
sampler ScreenTexture_Sampler : layout(s5)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Clamp;
};
texture2D tShadowMask : ShadowMask : layout(t6) = "White";
sampler sShadowMaskSampler : layout(s6)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tShadowMap : ShadowMap : layout(t7) = "White";
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
//////////////////////////////////Properties////////////////////////////////////////

//////////////////////////////////Vertex Output////////////////////////////////////////
//////////////////////////////////Constant buffer////////////////////////////////////////

cbuffer Batch
{
	float4x3 World 					: World 					;
	float4 cTintColor1				: TintColor1				; 
	float4 cTintColor2				: TintColor2				; 
	float4 cTintColor3				: TintColor3				; 

	float4 cShadowBias				: ShadowBias				; 
#if SD_SH_IBL
	float4 cSHCoefficients[7]		: SHCoefficentsData			;
#else	
	float4 cPointCloud[6]			: PointCloudData			;
	
	float4 cPlaceHold											;
#endif
	float4 cParamter				: Param						;
	float4 cParamter2				: Param2					;
#if !IS_TERRAIN	
	float4 cVirtualLitDir			: VirtualLitDir				;
	float4 cVirtualLitColor			: VirtualLitColor			;
#else
	float4 cHeightmapMinPos			: HeightmapMinPos			;//w heightmap height
	float4 cHeightmapMaxPos			: HeightmapMaxPos			;//w heightmap range / texture size 
#endif
	float4x4 WorldViewProj			: WorldViewProj				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex		;
	float4 PlayerPos                : PlayerPos					;
	float4 FoliageCenter            : FoliageCenter				;
};
#define SHADOW_PIXELSIZE (cShadowBias.w)
#include "Lighting.fxh"
//////////////////////////////////Vertex Output////////////////////////////////////////

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	// [@_DF_SKIN]
	// float4		BlendWeights	: BLENDWEIGHT;
	// [@_DF_SKIN]
	// uint4		BlendIndices	: BLENDINDICES;

	// [@_INSTANCED]
	// float4		Transform0		: TEXCOORD2;
	// [@_INSTANCED]
	// float4		Transform1		: TEXCOORD3;
	// [@_INSTANCED]
	// float4		Transform2		: TEXCOORD4;
	// [@_INSTANCED]
	// float4		LightMapScale	: TEXCOORD5;
	// [@_INSTANCED]
	// float4		LightMapTransform: TEXCOORD6;
};
struct VertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	
	float4 		DiffLighting 	: TEXCOORDN;			//  PerVertex Diffuse Lighting.
 	


	[@_TANGENT] 
	half4		WorldTangent	: TEXCOORDN;
	[@_TANGENT]
	half4		WorldBinormal	: TEXCOORDN;


	[@_CastSunShadow]
	float4 		LitUVZ			: TEXCOORDN; 



	// Shader special output
	float4		TintColor 		: COLOR;				//	TintColor

// 	[@_INSTANCED]
// 	float4		LightMapScale	: TEXCOORDN;			//	Light map scale

// 	float 		LinearZ			: TEXCOORDN;
// #ifdef HAS_WORLD_POSITION_OFFSET
// 	half3		WorldOffset		: TEXCOORDN;
// #endif
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
	float3 LocalNormal;
	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;

};

void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{
	//[@_HLSL]
	//OUT = (VertexShadingOutput)0;

	//PBR_VS::struct VertexShadingOutput

	OUT.HPosition = float4(0, 0, 0, 0);

	OUT.TexCoord = float4(0,0,0,0);

	OUT.WorldPosition = float4(0,0,0,0);

	OUT.WorldNormal = float4(0,0,0,0);

	OUT.ViewDirection = float4(0,0,0,0);

	OUT.DiffLighting = float4(0,0,0,1);


	[@_TANGENT]
	OUT.WorldTangent = half4(0,0,0,0);
	[@_TANGENT]	
	OUT.WorldBinormal = half4(0,0,0,0);


	[@_CastSunShadow]
	OUT.LitUVZ = float4(0,0,0,0);

// #if IS_WATER
// 	[@_CastSunShadow]
// 	OUT.LitUVZ2	 = float4(0,0,0,0);
// #elif USE_PLAYER_SHADOWMAP
// 	[@_CastSunShadow && @_TRANSPARENT]
// 	OUT.LitUVZ2	 = float4(0,0,0,0);
// #endif

	OUT.TintColor = float4(0,0,0,0);

// 	[@_INSTANCED]
// 	OUT.LightMapScale = float4(0,0,0,0);

// #ifdef HAS_WORLD_POSITION_OFFSET
// 	OUT.WorldOffset = half3(0, 0, 0);
// #endif
// 	OUT.LinearZ = 0; 
}

void InitInterMediateVertex(inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(0,0,0,0);
	
	INTM.LocalNormal = float3(0,0,0);

	[@_TANGENT]
	INTM.LocalTangent = float3(0,0,0);

	[@_TANGENT]
	INTM.LocalBinormal = float3(0,0,0);
}
void InitDepthCommon(inout VertexShadingOutput OUT)
{

if(@_EnableDepthBias)
{

float depthBias = cDepthBias;
	OUT.HPosition.z -= depthBias / (depthBias + OUT.HPosition.w);

	OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
	OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	
}

}
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	//[@_HLSL]
	//INTM = (InterMediateVertex)0;

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}
// void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
// {
// 	if (@_DF_SKIN)
// 	{
// 		half4 w = IN.BlendWeights;
// 		uint4 i = IN.BlendIndices;
// 		// Linear-blend skinning.
// 		if (@_DF_SKIN == 1)
// 		{
// 			// Fetch.
// 			float4 M0[3], M1[3], M2[3], M3[3];
// 			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
// 			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
// 			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
// 			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

// 			// Blend.
// 			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
// 			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
// 			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

// 			// Apply.
// 			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
// 			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

// 			if(@_TANGENT  && @_LOD_LEVEL <= LOD_Level_1)
// 			{
// 				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
// 				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
// 			}
// 		}
// 	}
// }
// void GetSkeletonMatrix4x3Tex( uint idx, out float4 Mx, out float4 My, out float4 Mz )
// {
// 	Mx = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+0.5)*cParamter.z, 0), 0);
// 	My = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+1.5)*cParamter.z, 0), 0);
// 	Mz = tSkeletonMap.SampleLevel(sSkeletonSampler, cParamter.xy + float2((3*idx+2.5)*cParamter.z, 0), 0);
// }
// void VertexUnifySkinningTex(in VertexAssemble IN, inout InterMediateVertex INTM)
// {
// 	if (@_DF_SKIN)
// 	{
// 		half4 w = IN.BlendWeights;
// 		uint4 i = IN.BlendIndices;
// 		// Linear-blend skinning.
// 		if (@_DF_SKIN == 1)
// 		{
// 			// Fetch.
// 			float4 M0[3], M1[3], M2[3], M3[3];
// 			GetSkeletonMatrix4x3Tex(i[0], M0[0], M0[1], M0[2]);
// 			GetSkeletonMatrix4x3Tex(i[1], M1[0], M1[1], M1[2]);
// 			GetSkeletonMatrix4x3Tex(i[2], M2[0], M2[1], M2[2]);
// 			GetSkeletonMatrix4x3Tex(i[3], M3[0], M3[1], M3[2]);

// 			// Blend.
// 			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
// 			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
// 			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

// 			// Apply.
// 			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
// 			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

// 			if(@_TANGENT  && @_LOD_LEVEL <= LOD_Level_1)
// 			{
// 				INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
// 				INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
// 			}
// 		}
		
// 	}
// }


void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	InitVertexShadingOutput(OUT);
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 
	// INTM.LocalPosition.xyz *=cScale;
	// INTM.LocalPosition.xyz+= cOffsets*0.01;

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
		//ViewDirection
	
	 OUT.WorldPosition = WorldPosition ;
	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World,INTM.LocalNormal));
	// OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World,float3(0,0,-1,)));
	if (@_TANGENT)
		{
			OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
			OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));
			
		}

		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));


// OUT.HPosition.w +=0.002;


	if(@_EnableDepthBias)
{

float depthBias = cDepthBias;
	OUT.HPosition.z -= CameraInfo.x* depthBias / (depthBias + OUT.HPosition.w);

	OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
	OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	
}

OUT.HPosition.w *=cFov;
		
	//TexCoord
	OUT.TexCoord.xy = IN.TexCoord.xy;


}



//////////////////////////////////Shading////////////////////////////////////////

VertexShadingOutput BillboardVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	
	VertexUnifyInterMediate(IN, INTM);
	//OUTPUT
	// VertexUnifySkinning(IN, INTM);
	
	VertexUnifyOutput(IN, INTM, OUT);
	

	
	
	return OUT;
}
void GetHexShadow(float3 LitUVZ, int layerIndex, out half InDynamicShadow)
{
	float inShadow;
	float2 pInLitTC = LitUVZ.xy;
	float litZ = LitUVZ.z;
	half2 inRange = abs(pInLitTC - 0.5);
	half ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
	float2 vInvShadowMapWH = float2(SHADOW_PIXELSIZE,SHADOW_PIXELSIZE);
	if(!@_CSM)
	{
		if(@_LOD_LEVEL <= LOD_Level_1)
		{
			hardware_pcf_4x4(pInLitTC, litZ, vInvShadowMapWH, inShadow);
		}else
		{
			hardware_pcf_2x2(pInLitTC, litZ, vInvShadowMapWH, inShadow);
		}
	}
	else if(layerIndex == 0)
	{
		if(@_LOD_LEVEL <= LOD_Level_1)
		{
	#if MATERIAL_SHADINGMODEL_HAIR
			inShadow = hardware_pcf_array_4x4(pInLitTC, litZ, vInvShadowMapWH, 0);
	#else
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 0);
	#endif
		}
		else
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 0);
		}
	}else if(layerIndex == 1)
	{
		float tempShadow;
		if(@_LOD_LEVEL <= SD_SHADOW_HIGHQUALITY_LEVEL)
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 1);
			tempShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
		}
		else
		{
			inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, 1);
			tempShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
		}
		inShadow = max(inShadow, tempShadow);
	}else{
		inShadow = hardware_pcf_array_2x2(pInLitTC, litZ, vInvShadowMapWH, CSMCacheIndexs[layerIndex]);
	}
	inShadow *= saturate((ShadowRate*10));
	InDynamicShadow = inShadow;
}

half GetDynamicShadowForSun(in VertexShadingOutput IN)
{
	half InShadow = 0;
	//get dynamic shadow
	if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		InShadow = tShadowMask.Sample(sShadowMaskSampler, tc).r;
	}
	else if (@_CastSunShadow && (@_REC_SHADOW || @_DEFERRED_SHADING || USE_SHADOWMAP_SAMPLE) && @_LOD_LEVEL <= LOD_Level_2)
	{
		float3 LitUVZ = IN.LitUVZ.xyz / IN.LitUVZ.w;
		LitUVZ.z = min(1.0-10*HALF_EPSILON, LitUVZ.z);
		GetHexShadow(LitUVZ, 1, InShadow);
	}
	InShadow = min(1.0, InShadow);

		InShadow *= SHADOW_RATE;
	return InShadow;
}
float4 BillboardPixelShading(in VertexShadingOutput IN,in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);


	// float3 V =normalize(IN.ViewDirection);
	float3 V =normalize(CameraPos.xyz-IN.WorldPosition.xyz) ;
	float3 N = normalize(IN.WorldNormal);
	// float2 spos = IN.HPosition.xy;
	
	half4 maskMap = tDiffuseMap.SampleRGBA(sDiffuseSampler,IN.TexCoord.xy);

	// float UVScale = ScreenInfo.x/ScreenInfo.y;
	// IN.TexCoord.x*=UVScale;
	half3 screenUVDistorb =tNormalMap.SampleRGBA(sNormalSampler,IN.TexCoord.xy);
	screenUVDistorb.xy =screenUVDistorb.xy*2-1;
	screenUVDistorb.xyz*=0.1;
	if(@_TANGENT)
	{
		screenUVDistorb.xy*=3;
		// screenUVDistorb = half3(0,0,1);
		N = screenUVDistorb.x * IN.WorldTangent + screenUVDistorb.y * IN.WorldBinormal + screenUVDistorb.z * IN.WorldNormal.xyz;
		N = normalize(N);
		
	}
		half3 GILighting = half3(1,1,1);
	
half InShadow = 0;
	if(USE_SHADOWMAP_SAMPLE)
	{
		InShadow = GetDynamicShadowForSun(IN);
#if USE_PLAYER_SHADOWMAP
		if(@_TRANSPARENT && @_CSM)
		{
			float3 LitUVZ2 = IN.LitUVZ2.xyz / IN.LitUVZ2.w;
			LitUVZ2.z = min(1.0-10*HALF_EPSILON, LitUVZ2.z);
			half calcShadow = 0;
			GetHexShadow(LitUVZ2, 0, calcShadow);
			InShadow = max(calcShadow, InShadow);
		}
#endif
	}else if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		float2 shadowAndAO = tShadowMask.Sample(sShadowMaskSampler, tc).rg;
		InShadow = max(shadowAndAO.x, InShadow);
		// GTAO = shadowAndAO.y;
	}

	DynamicGILight(N, GILighting, InShadow);
	half2 sphereUV = (IN.TexCoord.xy-half2(0.5,0.5))+ (cLensMask) ;
	half sphereUVMask =saturate(pow( distance(sphereUV,0),10)) ;
	float fresnelMask  = sphereUVMask*cAlpha;
	////////////lensDirtColor/////////
	half NoL = saturate(dot(N,SunDirection.xyz));
	float VoL = 0;
	if(USE_SHADOWMAP_SAMPLE)
	{
	VoL =saturate(dot(-V,SunDirection)) ;
	VoL = pow(VoL,55);
	}

	half3 lensDirt = 0;
	
	// ///cpt
	// if(0)
	// {
	// lensDirt  = (NoL+VoL)*(1-InShadow)*SunColor.rgb+GILighting;
	// maskMap.rgb*=maskMap.rgb/PI;
	// lensDirt*=maskMap.rgb*maskMap.a;
	// }
	
	
	half Roughness = maskMap.r+0.1;
////////IBL/////////////
half3 R = reflect(-V, N);
		half3 sampleEnvSpecular = half3(0,0,0);
		
		half level =10*Roughness;
		half fSign = R.z > 0;
		half fSign2 = fSign*2 - 1;
		R.xy /= R.z*fSign2 + 1;
		R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
		
		half4 srcColor;
		srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy,level);
		sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) ;
		

	
	sampleEnvSpecular = EnvBRDFApprox(cBaseColor*0.08, Roughness, 1) * sampleEnvSpecular*cIBLintensity;


///////////////finalColor//////////////////////

	if(@_TRANSPARENT)
	{
			
	OUT.a = saturate((fresnelMask *cAlpha)+(1-maskMap.a)) ;

	}
	// else
	// {
	// float2 ScreenUV = spos * ScreenInfo.zw; 
	// // ScreenUV =(ScreenUV-0.5)*cNormalIntensity + 0.5;
	// screenUVDistorb.y*=-UVScale;
	// screenUVDistorb.xy =  screenUVDistorb.xy *cImageScale ;
	// half UVdistobSum = 0.04;	
	// float2 UV1 = ScreenUV + screenUVDistorb.xy*UVdistobSum ;
	// float2 UV2 = ScreenUV + screenUVDistorb.xy*UVdistobSum*0.9+0.001 ;
	
	
	// float3 color1 = ScreenTexture.SampleRGBA(ScreenTexture_Sampler,UV1);
	// float3 color2 = ScreenTexture.SampleRGBA(ScreenTexture_Sampler,UV2);
	
	// OUT.rgb = half3(color1.r,color1.g,color2.b) *(1-fresnelMask)  ;	

	// }	
	OUT.rgb += lensDirt;
	OUT.rgb +=sampleEnvSpecular;
	
	


	// OUT.rgb = sphereUVMask.xxx;
	// OUT.a=1;

	return OUT;
}

technique Shading
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
	if (@_CLIPPING)
	{
		half a = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TexCoord.xy).a;
		clip(a - 0.5);
	}
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = BillboardVertexShading();
	PixelShader = ClipPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Wireframe
[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].ViewDirection = float4(1, 0, 0,0);
	OUT.Append(IN[0]);
	IN[1].ViewDirection = float4(0, 1, 0,0);	
	OUT.Append(IN[1]);
	IN[2].ViewDirection = float4(0, 0, 1,0);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.ViewDirection);
	float3 distance = IN.ViewDirection / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(cParamter2.xyz, a);
}

technique Wireframe
{
	VertexShader = BillboardVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = WireframePixelShading();
};

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	// VertexUnifySkinning(IN, INTM);
	//position
	
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	
	OUT.HPosition.w +=0.001;


if(@_EnableDepthBias)
{

float depthBias = cDepthBias;
	OUT.HPosition.z -=  CameraInfo.x*depthBias / (depthBias + OUT.HPosition.w);

	OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
	OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	
}
	
OUT.HPosition.w *=cFov;


	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);

	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	



	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);

	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		half a = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TC.xy).a;
		clip(a - 0.5);
	}

	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

