//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_VBufferGen			1
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define VIRTUAL_LIT 			0
#define DOUBLE_REFLECTCOLOR 	0
#define RTM_REFLECTION 			1
#define ENABLE_MIXMAP			(@_EnableReflection)

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Default = false;
	Visible = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = true;	
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_EmissionMapEnable
{
	Name = EmissionMapEnable;
	Description = "Do we have EmissionMap?";
	AffectedTex = tEmissionMap;
	Fallback = true;
	Visible = true;
};
@_IsSameDirection
{
	Name = IsSameDirection;
	Description = "same direction?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 

	float cAliasingFactor			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;//抗锯齿因子
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float4 cEmissionScale			 = {1, 1, 1, 0};
	float4 cEmissionScrolling		 = {1, 1, 0, 0};

	float cRoughnessX				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.1;
	float cRoughnessY				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 1;
	float cAnisotropicScale			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0;
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;

	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.65;
	float cAlpha	      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;
	//float cRotateSin				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.1"; Precision = "2"; > = 0;
	//float cRotateCos				 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.1"; Precision = "2"; > = 1;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tReflectionMap : ReflectionMap : layout(t3) = "White";
sampler sReflectionSampler : layout(s3)
{
	Texture = tReflectionMap;
	Filter = Linear;
	Address = Mirror;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tShadowMask : ShadowMask : layout(t6) = "White";
sampler sShadowMaskSampler : layout(s6)
{
	Texture = tShadowMask; 
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

// Emission map
texture2D tEmissionMap : layout(t7) = "Transparent";
sampler sEmissionMapSampler : layout(s7)
{
	Texture = tEmissionMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
};

texture2D tRainMap : RainMap: layout(t8) = "NormalMap";
sampler sRainSampler : layout(s8)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	if(@_EmissionMapEnable)
	{
		half2 EmissionUV = IN.TexCoord.xy * cEmissionScrolling.xy + GameTime * cEmissionScrolling.zw;
		OUT.WorldNormal.w = EmissionUV.x;
		OUT.ViewDirection.w = EmissionUV.y;
	}

	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}
}  

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	if (@_BaseMapEnable)
	{
		half4 samplerBase;
		if (@_CLIPPING || @_DISSOLVING || @_TRANSPARENT || @_EmissionMapEnable)
			samplerBase = tBaseMap.SampleRGBA(sBaseSampler, BaseMapUV(IN.TexCoord));
		else
			samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = cAlpha;
	}
	MaterialParameters.BaseColor *= IN.TintColor.rgb;
	MaterialParameters.AO = 1.0;
	if (@_EnableReflection && @_ReflectionMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).xyz;
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_2)
		{
			MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else
	{
		MaterialParameters.Metallic = cMetallic;
		if(@_EnableReflection || @_DEFERRED_SHADING)
			MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		if (@_RIPPLE)
		{
			half2 tc = IN.WorldPosition.xz * RainNormalXZScale;
			half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
			half2 ttc = tc + delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			ttc = tc - delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			tangentNormal /= (1 + 2 * RainNormalRate);
		}
		tangentNormal = (2*tangentNormal)-1;
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		MaterialParameters.WorldNormal = tangentNormal.x * cBumpIns * IN.WorldTangent + tangentNormal.y *cBumpIns* IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
	MaterialParameters.Emission = half3(0, 0, 0);
	if (@_EmissionMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half2 EmissionUV = half2(IN.WorldNormal.w, IN.ViewDirection.w);
		MaterialParameters.Emission = tEmissionMap.Sample(sEmissionMapSampler, EmissionUV).rgb;
		MaterialParameters.Emission.rgb *= cEmissionScale.rgb;
		if(@_HDR){MaterialParameters.Emission.rgb *= MaterialParameters.BaseColor;}
		if(!@_TRANSPARENT && !@_CLIPPING)
		{
			MaterialParameters.Emission.rgb *= MaterialParameters.Alpha;
			MaterialParameters.Alpha = 1.0;
		}
	}
}
#include "Lighting.fxh"
#include "PBR_Common.fxh"
half3 IBL_SpecularRTM(in GBufferData GBuffer, in VertexShadingOutput IN, in half3 R, in half NoV)
{
	float2 spos = IN.HPosition.xy;
	if(@_D3D9) {spos += 0.5;}
	spos += GBuffer.WorldNormal.xz * 5;
	float2 stc = spos * ScreenInfo.zw;
	stc.x = 1-stc.x;
	half3 sampleEnvSpecular = tReflectionMap.SampleLevel(sReflectionSampler, stc, GBuffer.Roughness * GBuffer.Roughness * 10).xyz;
	if (@_LOD_LEVEL <= LOD_Level_1)
		return EnvBRDFApprox(GBuffer.SpecularColor, GBuffer.Roughness, NoV) * sampleEnvSpecular;
	else
		return GBuffer.SpecularColor * sampleEnvSpecular;
}
//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput PbrVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);
	
	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif

	//OUTPUT
	VertexUnifyOutputCommon(IN, INTM, OUT);

	return OUT;
}

float4 PbrPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	//1, Gen MaterialParameters & GBufferData
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	PS_GetParameters(IN, MaterialParameters);
	OUT.a = MaterialParameters.Alpha;

	MaterialParameters.SpecularColor = lerp(0.08 * MaterialParameters.Specular.xxx, MaterialParameters.BaseColor, MaterialParameters.Metallic);
	MaterialParameters.DiffuseColor = Diffuse_Lambert(MaterialParameters.BaseColor - MaterialParameters.BaseColor * MaterialParameters.Metallic);
		
	//ShadowAndAO
	half GTAO = 1;
	if(USE_SHADOWMAP_SAMPLE)
	{
		MaterialParameters.InShadow = GetDynamicShadowForSun(IN);
	}else if(ENABLE_SCREEN_SHADOWMAP)
	{
		float2 tc = IN.HPosition.xy * ScreenInfo.zw;
		float2 shadowAndAO = tShadowMask.Sample(sShadowMaskSampler, tc).rg;
		MaterialParameters.InShadow = max(shadowAndAO.x, MaterialParameters.InShadow);
		GTAO = shadowAndAO.y;
	}

	//2, lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	int PBRQualityLevel = PBRQualityLevel_None;
	if (@_LOD_LEVEL <= LOD_Level_2)
		PBRQualityLevel = PBRQualityLevel_Normal;
	half3 V = normalize(-IN.ViewDirection.xyz);
	half NoV = saturate(dot(MaterialParameters.WorldNormal, V));
	half3 R = reflect(-V, MaterialParameters.WorldNormal);
	ModifyReflectionDirection(IN, R);
	//GI lighting
	FetchPrecomputedLighing(IN, MaterialParameters);
	//InitIBLReflection(IN, MaterialParameters, V, NoV);
	MaterialParameters.GILighting *= GTAO;
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	//GI Specular lighting
	half3 IBLSpec = half3(0, 0, 0);
	if(ENABLE_MIXMAP)
	{
		IBLSpec = IBL_SpecularRTM(GBuffer, IN, R, NoV);
		IBLSpec *= GTAO  * GBuffer.AO;
	}
	//sun lighting
	LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);
	IntegrateBxDF(LitResult, GBuffer, LitData);	
	//thunder shading
	LitResult.SpecularLighting *= ThunderRate;
	LitResult.DiffuseLighting *= ThunderRate;
	//point lighting calc include vs lighting and ps lighting
	InitPointLighting(IN, GBuffer, LitResult, V, NoV);

	//accumulate specular and diffuse
	OUT.rgb += GBuffer.Emission;//emission without GI lighting
	OUT.rgb += LitResult.SpecularLighting + IBLSpec;
	OUT.rgb += (LitResult.DiffuseLighting + GBuffer.GILighting) * GBuffer.DiffuseColor;

	//3, Fog shading.
	InitFogLighting(IN, LitData, OUT);

	//4, screen color tint
#if USE_LIGHTMAP
	if (@_LIGHT_MAP || LIGHTMAP_ENABLE)
		OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);
#endif

	//always calc in linear space
	EncodeHDR(OUT.rgb);

	return OUT;
}

technique Shading
{
	VertexShader = PbrVertexShading();
#if USE_TESSELLATION	
	HullShader = PbrVertexShading_HS();
	DomainShader = PbrVertexShading_DS();
#endif
	PixelShader = PbrPixelShading();
}

#include "GBufferGenPass.fxh"
#include "ShadowGenPass.fxh"
#include "HeightmapGenPass.fxh"
#include "DebugShaderPass.fxh"
#include "ClipPass.fxh"
#include "WireframePass.fxh"
#include "EdgeLineNormalGenPass.fxh"

