//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_SSS 1

#define USE_LIGHTMAP			0
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				1
#define USE_HIGHLIGHT			1
#define USE_FROZEN				1
#define USE_BURNT				1
#define USE_EDGELINE			1
#define USE_PLANAR_SHADOW		1
#define USE_LIGHTMAP_CONST		0
#define USE_VBufferGen			1
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define VIRTUAL_LIT 			1
#define DOUBLE_REFLECTCOLOR 	0
#define HAS_2UV					0

#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"
#include "BlendNormal.fxh"

#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_CLIPPING", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	Fallback = false;
	Visible = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Fallback = false;
	Visible = true;	
};
@_MixMapEnable
{
	Name = MixMapEnable;
	Description = "Has mix map?";
	AffectedTex = tMixMap;
	Fallback = true;
	Visible = false;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_DetailNormalMapEnable
{
	Name = DetailNormalMapEnable;
	Description = "Do we have DetailNormalMap?"; 
	AffectedTex = tDetailNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};


//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
    float cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;
	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1;
	float cAlpha	      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;
	float cDepthOffset     			 < Widget = "FloatSlider"; Min = "-5"; Max = "5"; Step = "0.1"; Precision = "2"; > = 0;
	float cDetailNormalUVScale		 < Widget = "FloatSlider"; Min = "0"; Max = "100"; Step = "0.01"; Precision = "2"; > = 50;
    float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tMixMap : layout(t1) = "Reflection";
sampler sMixSampler : layout(s1)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t2) = "White";
sampler sEnvSampler : layout(s2)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t3) = "NormalMap";
sampler sNormalSampler : layout(s3)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tDetailNormalMap : layout(t4) = "NormalMap";
sampler sDetailNormalSampler : layout(s4)
{
	Texture = tDetailNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t5) = "White";
SamplerComparisonState sShadowMapSampler : layout(s5)
{
	Texture = tShadowMap; 
	Filter = Linear; 
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	//MaxMipLevel = 0;
};

texture2D tRainMap : RainMap: layout(t6) = "NormalMap";
sampler sRainSampler : layout(s6)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////Function/////////////////////////////////////

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}

	//rain
	OUT.WorldPosition.w = IN.Position.y;
	
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}

	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}
}
#include "PBR_Common.fxh"
//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	if (@_BaseMapEnable)
	{
		half4 samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);
		MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
		MaterialParameters.Alpha = min(cAlpha,samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = cAlpha;
	}
	MaterialParameters.AO = 1.0;
	MaterialParameters.CustomData.r = 0;
	if (@_MixMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		//todo: cavity
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord)).rgb;
		half Smoothness = samplerMis.r;
		MaterialParameters.CustomData.r = samplerMis.g * samplerMis.g * samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_2)
		{
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else
	{
		if(@_EnableReflection || @_DEFERRED_SHADING)
			MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		if (@_RIPPLE)
		{
			half2 tc = IN.WorldPosition.xz * RainNormalXZScale;
			half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
			half2 ttc = tc + delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			ttc = tc - delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			tangentNormal /= (1 + 2*RainNormalRate);
		}
		if (@_DetailNormalMapEnable)
		{
			half3 detailNormal = tDetailNormalMap.Sample(sDetailNormalSampler, BaseMapUV(IN.TexCoord) * cDetailNormalUVScale).xyz;
			tangentNormal = blend_rnm(tangentNormal, detailNormal);
		}
        else
		    tangentNormal = (2 * tangentNormal) - 1;
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		tangentNormal.xy *= cBumpIns;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent + tangentNormal.y * IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		if (@_LOD_LEVEL <= LOD_Level_2)
			MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
		else
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
}
#include "PBR_PS.fxh"

