//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : jiawei@corp.netease.com
//	Author: jiawei
//	--------------------------------------------------------------------
///	@file	<PBR_TerrainBlend.fx>
///	@path	~EngineShaders
///	@date	2015/09/18
///	@brief	brief Terrain Blend material

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1

#pragma shader predict("@_TANGENT", "true")
#pragma shader predict("@_VERTEXCOLOR", "false")
#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"  

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?"; 
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Visible = true; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = true;	
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
}; 

@_DetailNormalMapEnable  
{
	Name = DetailNormalMapEnable;
	Description = "Do we have DetailNormalMap?"; 
	AffectedTex = tDetailNormalMap;
	Fallback = true;
	Visible = true;
}; 

@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};

@_OverlayMapEnable
{
	Name = OverlayMapEnable;
	Description = "Do we have OverlayMap?";
	AffectedTex = tOverlayMap;
	Fallback = true;
	Visible = true;
};
@_IsDynamicDirection
{
	Name = IsDynamicDirection;
	Description = "IsDynamicDirection?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < IsColor = "true"; > = {1, 1, 1};
	float cMetallic					 = 0;	//金属度[0:1] 

	float cBumpStrength				 = 1;	//法线强度  
	float cEnvStrength				 = 1;	//环境强度[0:8]
	float cRoughnessLow				 = 0;	//粗糙度下界
	float cRoughnessHigh			 = 1;	//粗糙度上界
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;
	float  cBlendRough			     < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;	//overlay roughness

	float2 cDetailNormalUV			 < Widget = "Vector2"; > = { 3, 3 };
	float2 cDetailTexUV 			 < Widget = "Vector2"; > = { 2, 2 };

	float3 cOverlayColor			 < IsColor = "true"; > = {1, 1, 1}; 
	float  cOverlayRough			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;	//overlay roughness

	float  cBlendTilling			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;	//环境强度[0:8]
	float  cBlendDensity			 < Widget = "FloatSlider"; Min = "-500"; Max = "500"; Step = "0.05"; Precision = "1"; > = 20;
	float  cBlendContrast		     < Widget = "FloatSlider"; Min = "0"; Max = "120"; Step = "0.01"; Precision = "2"; > = 1;

	float2 cBlendNorPower   	     < Widget = "Vector2"; > = { 1.5, 1.5 };
	float2 cBlendNormalUV			 < Widget = "Vector2"; > = { 3, 3 };		
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;  
	SRGB = false;  
	MaxMipLevel = 0;  
}; 

// ShadowMap
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
}; 

texture2D tDetailNormalMap: layout(t7) = "White";
sampler sDetailNormalSampler : layout(s7)
{
	Texture = tDetailNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tOverlayMap : layout(t8) = "White";
sampler sOverlayMapSampler : layout(s8)
{
	Texture = tOverlayMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tRainMap : RainMap: layout(t9) = "NormalMap";
sampler sRainSampler : layout(s9)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	OUT.LocalPosition =  float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if (@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}
//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	float3 worldUV = IN.WorldPosition.xyz * (cBlendTilling/2.5);
	half3 wc = frac(worldUV);
	half3 w = abs(IN.WorldNormal.xyz);
	w = step(w.yzx, w.xyz) * step(w.zxy, w.xyz);
	wc.y *= sign(dot(IN.WorldNormal.xyz, w));
	half2 tc = half2(dot(wc.yxx, w), dot(wc.zzy, w));
	half2 normalOffset = half2(0,0);
	half4 diffuse;
	if ( @_BaseMapEnable )
	{
		diffuse = tBaseMap.SampleRGBA( sBaseSampler, IN.TexCoord.xy);
		diffuse.rgb = DecodeGamma(diffuse.rgb);
	} 
	else
	{
		diffuse.rgb = cBaseColor;
		diffuse.a = 0.5;
	}
	half overlayFactor;
	if ( @_IsDynamicDirection  )
	{
		overlayFactor = saturate(dot(cBlendContrast,IN.WorldNormal.y ) * clamp(0.3,1.0,diffuse.r)-cBlendDensity/20);
	} 
	else
	{
		overlayFactor = saturate(dot(cBlendContrast,IN.LocalPosition.y/20 ) * clamp(0.3,1.0,diffuse.r)-cBlendDensity/20);
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, IN.TexCoord.xy, cNormalMapBias).xyz;
		if (@_RIPPLE)
		{
			half2 tc0 = IN.WorldPosition.xz * RainNormalXZScale;
			half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
			half2 ttc = tc0 + delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			ttc = tc0 - delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			tangentNormal /= (1 + 2 * RainNormalRate);
		}
		tangentNormal = (2*tangentNormal)-1;
		tangentNormal.xy *= cBumpStrength;
		if (@_DetailNormalMapEnable)
		{
			half2 detailNormal = tDetailNormalMap.SampleBias(sDetailNormalSampler, tc.xy * cBlendNormalUV, cNormalMapBias).xy;
			detailNormal = (2*detailNormal)-1;
			detailNormal *= cBlendNorPower;
			tangentNormal = lerp( half3(detailNormal.xy,1) ,  tangentNormal ,  overlayFactor);
		}
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		MaterialParameters.WorldNormal = tangentNormal.x * IN.WorldTangent.xyz + tangentNormal.y * IN.WorldBinormal.xyz + tangentNormal.z*IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = lerp(float3(0,1,0) ,MaterialParameters.WorldNormal,overlayFactor);
	}
	else
	{
		MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}
	MaterialParameters.WorldNormal = normalize( MaterialParameters.WorldNormal );
	half om_overlayFactor = overlayFactor;
	half3 overlayColor;
	if ( @_OverlayMapEnable )
	{
		overlayColor = tOverlayMap.SampleRGBA( sOverlayMapSampler, tc.xy * cDetailTexUV).rgb;
		overlayColor = DecodeGamma(overlayColor);
	}
	else
	{
		overlayColor = cOverlayColor.xyz;
	}
	half3 lerpedBaseColor = lerp(overlayColor, lerp(0.01, diffuse.rgb, om_overlayFactor), om_overlayFactor);
	MaterialParameters.BaseColor = lerpedBaseColor;
	MaterialParameters.Alpha = 1;
	if ( @_EnableReflection )  
	{  
		half4 samplerMis = tMixMap.SampleRGBA( sMixSampler, BaseMapUV(IN.TexCoord) );
		samplerMis = DecodeGamma(samplerMis);
		MaterialParameters.Roughness = lerp( samplerMis.r, cOverlayRough, overlayFactor );
		MaterialParameters.Metallic = samplerMis.g;
	} 
	else
	{
		MaterialParameters.Roughness = cRoughnessLow; 
		MaterialParameters.Metallic = cMetallic;
	}
}

#include "PBR_PS.fxh"
