//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define NO_SHADOW_CLIPPING 		1
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1

#pragma shader predict("@_TANGENT", "true")
#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "true")
#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")
 
///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	AffectedTex = tMixMap;
	Fallback = true;
	Visible = true; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = true;
	Visible = true;
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
}; 
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_WaterNormalMapEnable  
{
	Name = WaterNormalMapEnable;
	Description = "Do we have WaterNormalMap?"; 
	AffectedTex = tWaterNormalMap;
	Fallback = true;
	Visible = true;
};
@_AffectedBySystemRain  
{
	Name = AffectedBySystemRain  ;
	Description = "Affected by system rain"; 
	Default = false;
	Visible = true;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cAliasingFactor			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2; 
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "8"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float4 cWaterScrolling1			 < Precision = "2"; > = {1, 1, 1, 1};
	float4 cWaterScrolling2			 < Precision = "2"; > = {1, 1, 1, 1};

	float3 cWaterColor				 < Widget = "LinearColor3"; > = {0, 0, 0};
	float  cWaterNormalBumpStrength  < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1.0;

	float  cMetallic				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 
	float  cWaterAlpha				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;	 
	float  cWaterness				 < Widget = "FloatSlider"; Tips = "test"; Min = "0"; Max = "5"; Step = "0.01"; Precision = "2"; > = 1.5;
	float  cAffectBaseUV			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;

	float  cWaterRough    			 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "Black";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// Light map
texture2D tLightMap : layout(t5) = "White";
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;  
	SRGB = false;  
	MaxMipLevel = 0;  
}; 

// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t6) = "White";
SamplerComparisonState sShadowMapSampler : layout(s6)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
}; 

// WaterNormalMap
texture2D tWaterNormalMap : layout(t7) = "Black";
sampler sWaterNormalSampler : layout(s7)
{
	Texture = tWaterNormalMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
		OUT.TintColor = GetVertexColor(IN.Color);
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}

	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	if(@_WaterNormalMapEnable && @_TANGENT)
	{
		half2 WaterUV1 = IN.TexCoord.xy * cWaterScrolling1.xy + GameTime * cWaterScrolling1.zw*0.1;
		half2 WaterUV2 = IN.TexCoord.xy * cWaterScrolling2.xy + GameTime * cWaterScrolling2.zw*0.1;
		OUT.WorldPosition.w = WaterUV1.x;
		OUT.WorldNormal.w = WaterUV1.y;
		OUT.ViewDirection.w = WaterUV2.x;
		OUT.WorldTangent.w = WaterUV2.y;
	}

	//world * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
}  

//////////////////////////////////PS//////////////////////////////////////// 
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	MaterialParameters.AO = 1.0;
	if (@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord));
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_2)
		{ 
			MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else
	{ 
		MaterialParameters.Metallic = cMetallic;
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
	half WaterRate = 0;
	WaterRate = IN.TintColor.r;
	half DryRate = 1.0 - WaterRate;
	DryRate = smoothstep(0.58, cWaterness, DryRate);
	MaterialParameters.Roughness = lerp(cWaterRough, MaterialParameters.Roughness, DryRate);
	half2 baseUV = BaseMapUV(IN.TexCoord);
	if(@_TANGENT && @_LOD_LEVEL <= LOD_Level_1)
	{	
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		tangentNormal = (2*tangentNormal)-1;
		half3 waterNormal = half3(0, 0, 1);
		if (@_WaterNormalMapEnable)
		{
			half2 WaterUV1 = half2(IN.WorldPosition.w, IN.WorldNormal.w);
			half2 WaterUV2 = half2(IN.ViewDirection.w, IN.WorldTangent.w);
			half3 waterNormal1 = tWaterNormalMap.Sample(sWaterNormalSampler, WaterUV1).xyz;
			waterNormal1 = (2*waterNormal1) - 1;
			half3 waterNormal2 = tWaterNormalMap.Sample(sWaterNormalSampler, WaterUV2).xyz;
			waterNormal2 = (2*waterNormal2) - 1;
			waterNormal = (waterNormal1+waterNormal2) * half3(0.5*cWaterNormalBumpStrength, 0.5*cWaterNormalBumpStrength, 0.5);
		}
		tangentNormal = lerp(waterNormal, tangentNormal, DryRate);
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		MaterialParameters.WorldNormal = tangentNormal.x*IN.WorldTangent + tangentNormal.y*IN.WorldBinormal + tangentNormal.z*IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
		baseUV += waterNormal.xy * (1-DryRate) * 5 * cAffectBaseUV;
	}
	else 
	{
		MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
	}
	half4 samplerBase;
	samplerBase = tBaseMap.SampleBias(sBaseSampler, baseUV, cBaseMapBias);
	MaterialParameters.BaseColor = DecodeGamma(samplerBase.rgb);
	MaterialParameters.Alpha = samplerBase.a;
	MaterialParameters.BaseColor = lerp(cWaterColor, MaterialParameters.BaseColor, clamp(DryRate, cWaterAlpha, 1.0));
	if (@_AffectedBySystemRain)
	{
		MaterialParameters.Roughness *= 1-RainRate;
		MaterialParameters.BaseColor *= 1-RainRate;
	}
}
#include "PBR_PS.fxh"

