#pragma shader material
#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "true")
#pragma shader predict("@_INSTANCED", "false")

///////////////choice define start///////////////////////////
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 	1 // for deferred
#define USE_LIGHTMAP						1
#define NO_SHADOW_CLIPPING					1
#define USE_LIGHTMAP_CONST					0
#define USE_CLIPPING						1
#define USE_WIREFRAME						1
#define USE_FROZEN							0
#define USE_BURNT							0
#define USE_HIGHLIGHT						0
#define USE_XRAY							0
#define USE_EDGELINE						0
#define USE_PLANAR_SHADOW					0
#define USE_VBufferGen						0
#define VIRTUAL_LIT 						0

#define DEFERRED_PRE_AO						1

#define HAS_2UV								(@_LIGHT_MAP)
#define ENABLE_MIXMAP						(@_EnableReflection)
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow)
///////////////choice define end//////////////////////////

#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"
//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "Enable reflection?";
	Default = true;
	Visible = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable image based lighting?";
	Default = true;
	Visible = false;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have light map?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = false;
};
@_SurfaceTextureArrayEnable
{
	Name = SurfaceTextureArrayEnable;
	Description = "Enable texture array for terrain?";
	AffectedTex = tSurfaceBlockMap;
	Visible = false;
};
@_SurfaceTextureArrayMixtureEnable
{
	Name = SurfaceTextureArrayMixtureEnable;
	Description = "Enable texture mixture array for terrain?";
	AffectedTex = tSurfaceMixtureMap;
	Visible = false;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have normal map?";
	AffectedTex = tSurfaceNormalMap;
	Visible = false;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cEnvStrength				 = 1;   //环境强度[0:8]
	float cRoughnessLow				 = 0;   //粗糙度下界
	float cRoughnessHigh			 = 1;   //粗糙度上界

	float cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;

	float4 cBlockScale				 = { 1, 0, 0, 0 };

	float4 cLightMapScale			 = { 1, 1, 1, 1 };
	float4 cLightMapUVTransform		 = { 1, 1, 0, 0 };
};

texture2DArray tSurfaceBlockMap : layout(t2);
sampler2DArray sSurfaceBlockSampler : layout(s2)
{
	Texture = tSurfaceBlockMap;
	Filter = Linear;
	Address = Wrap;
};

texture2DArray tSurfaceNormalMap : layout(t3);
sampler2DArray sSurfaceNormalSampler : layout(s3)
{
	Texture = tSurfaceNormalMap;
	Filter = Linear;
	Address = Wrap;
};

texture2DArray tSurfaceMixtureMap : layout(t4);
sampler2DArray sSurfaceMixtureSampler : layout(s4)
{
	Texture = tSurfaceMixtureMap;
	Filter = Linear;
	Address = Wrap;
};

texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tEnvMap : EnvMap : layout(t6);
sampler sEnvSampler : layout(s6)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tShadowMask : ShadowMask : layout(t7);
sampler sShadowMaskSampler : layout(s7)
{
	Texture = tShadowMask;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tShadowMap : ShadowMap : layout(t7);
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tRainMap : RainMap : layout(t9) = "NormalMap";
sampler sRainSampler : layout(s9)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	OUT.ViewDirection.w = 0;

	// World * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));

	if(@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 0, 0, 0);
	}
}

#include "PBR_Common.fxh"
//////////////////////////////////PS////////////////////////////////////////

#define SampleBlend(blend_result, SurfaceTexture, SurfaceSampler, uv, blend_weights, blend_indices) \
{ \
	blend_result = float3(0, 0, 0); \
	blend_result += blend_weights[0] * SurfaceTexture.Sample(SurfaceSampler, float3(uv, blend_indices[0])).xyz; \
	blend_result += blend_weights[1] * SurfaceTexture.Sample(SurfaceSampler, float3(uv, blend_indices[1])).xyz; \
	blend_result += blend_weights[2] * SurfaceTexture.Sample(SurfaceSampler, float3(uv, blend_indices[2])).xyz; \
	blend_result += blend_weights[3] * SurfaceTexture.Sample(SurfaceSampler, float3(uv, blend_indices[3])).xyz; \
}

void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);

	float dist = length(CameraPos.xyz - IN.WorldPosition.xyz);
	half rateSpec = saturate((1 - dist / cParamter.w * 0.8) / 0.8);

	float4 weights = IN.TintColor;
	weights /= weights.x + weights.y + weights.z + weights.w;
	float4 indices;
	indices.x = round(IN.TexCoord.x / 256);
	indices.y = round(IN.TexCoord.x - indices.x * 256);
	indices.z = round(IN.TexCoord.y / 256);
	indices.w = round(IN.TexCoord.y - indices.z * 256);

	float2 uv_x = cBlockScale.x * IN.WorldPosition.yz;
	float2 uv_y = cBlockScale.x * IN.WorldPosition.xz;
	float2 uv_z = cBlockScale.x * IN.WorldPosition.xy;

	float3 ratio = abs(IN.WorldNormal.xyz);
	ratio = ratio / (ratio.x + ratio.y + ratio.z);

	if(@_SurfaceTextureArrayEnable)
	{
		float3 diff_x, diff_y, diff_z;
		SampleBlend(diff_x, tSurfaceBlockMap, sSurfaceBlockSampler, uv_x, weights, indices);
		SampleBlend(diff_y, tSurfaceBlockMap, sSurfaceBlockSampler, uv_y, weights, indices);
		SampleBlend(diff_z, tSurfaceBlockMap, sSurfaceBlockSampler, uv_z, weights, indices);

		half3 diff = ratio.x * diff_x + ratio.y * diff_y + ratio.z * diff_z;
		MaterialParameters.BaseColor = DecodeGamma(diff);	
	}

	if (@_LOD_LEVEL <= LOD_Level_1 && @_SurfaceTextureArrayMixtureEnable)
	{
		float3 mixx_x, mixx_y, mixx_z;
		SampleBlend(mixx_x, tSurfaceMixtureMap, sSurfaceMixtureSampler, uv_x, weights, indices);
		SampleBlend(mixx_y, tSurfaceMixtureMap, sSurfaceMixtureSampler, uv_y, weights, indices);
		SampleBlend(mixx_z, tSurfaceMixtureMap, sSurfaceMixtureSampler, uv_z, weights, indices);

		half3 mixx = ratio.x * mixx_x + ratio.y * mixx_y + ratio.z * mixx_z;
		MaterialParameters.Metallic = 0.0;
		MaterialParameters.Roughness = 1 - lerp(0.2, mixx.r, rateSpec);
		MaterialParameters.AO = lerp(1, mixx.b, rateSpec);
	}
	else
	{
		MaterialParameters.Metallic = 0;
		MaterialParameters.Roughness = 1 - 0.2;
	}
	MaterialParameters.Specular = 0.2;

	if(@_NormalMapEnable)
	{
		half3 is_pos = lerp(-1, 1, step(half3(0, 0, 0), IN.WorldNormal.xyz));

		float3 tangent = normalize(cross(IN.WorldNormal.xyz, float3(0, 0, -1))) * is_pos.x;
		float3 binormal = cross(tangent, IN.WorldNormal.xyz) * is_pos.x;

		float3 norm;
		SampleBlend(norm, tSurfaceNormalMap, sSurfaceNormalSampler, uv_x, weights, indices);
		norm = 2 * norm - 1;
		MaterialParameters.WorldNormal = ratio.x * (norm.z * IN.WorldNormal.xyz + norm.x * tangent + norm.y * binormal);

		tangent = normalize(cross(IN.WorldNormal.xyz, float3(0, 0, 1))) * is_pos.y;
		binormal = cross(IN.WorldNormal.xyz, tangent) * is_pos.y;
		SampleBlend(norm, tSurfaceNormalMap, sSurfaceNormalSampler, uv_y, weights, indices);
		norm = 2 * norm - 1;
		MaterialParameters.WorldNormal += ratio.y * (norm.z * IN.WorldNormal.xyz + norm.x * tangent + norm.y * binormal);

		tangent = normalize(cross(IN.WorldNormal.xyz, float3(0, -1, 0))) * is_pos.z;
		binormal = cross(tangent, IN.WorldNormal.xyz) * is_pos.z;
		SampleBlend(norm, tSurfaceNormalMap, sSurfaceNormalSampler, uv_z, weights, indices);
		norm = 2 * norm - 1;
		MaterialParameters.WorldNormal += ratio.z * (norm.z * IN.WorldNormal.xyz + norm.x * tangent + norm.y * binormal);

		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
	}

	// Apply Burnt
	InitBurntSimple(IN, MaterialParameters, 0.35);
	//Apply Weather
	InitRain(IN, MaterialParameters);
}
#include "PBR_PS.fxh"

#if SD_ENABLE_VXGI
#include "VXGI/VXGIVoxelize.fxh"
#endif
