//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : szn3878@corp.netease.com
//	Author: Si Zheng
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<PBR_Wall.fx>
///	@path	~/System/Shader/
///	@date	2016/09/20
///	@brief	PBR_Feather + PBR_PowerLine
#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			0
#define USE_CLIPPING			1
#define USE_WIREFRAME			1
#define USE_XRAY				1
#define USE_HIGHLIGHT			1
#define USE_FROZEN				1
#define USE_BURNT				1
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			1
#define USE_EdgeLineNormalGen	0
#define DOUBLE_REFLECTCOLOR 	1
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

//////////////////////////////////Properties////////////////////////////////////////
@_BaseMapEnable 
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};
@_EnableReflection 
{
	Name = EnableReflection;
	Description = "EnableReflection?";
	Default = true;
	Visible = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = false;
	Visible = true;	
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Default = false;
	Description = "Do we have LightMap?";
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_EmissionMapEnable
{
	Name = EmissionMapEnable;
	Description = "Do we have EmissionMap?";
	AffectedTex = tEmissionMap;
	Fallback = true;
	Visible = true;
};
@_HasFeatherMovement
{
	Name = HasFeatherMovement;
	Description = "Does it shake?";
	Visible = true;
};
@_PowerLineMapEnable
{
	Name = PowerLineMapEnable;
	Description = "Do we have PowerLine?";
	AffectedTex = tPowerLineMap;
	Fallback = true;
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cBaseColor 				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//金属度[0:1] 

	float cAliasingFactor			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.2;//抗锯齿因子
	float cEnvStrength				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughnessLow				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;	//粗糙度下界
	float cRoughnessHigh			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 1;	//粗糙度上界
	
	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float  cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float  cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float  cBaseMapBias				 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -1;
	float  cNormalMapBias			 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.01"; Precision = "2"; > = -0.5;

	float4 cEmissionScale			 = {0, 0, 0, 0};
	float4 cEmissionScrolling		 = {1, 1, 1, 1};

	float cRoughnessX				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 0.1;
	float cRoughnessY				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "2"; > = 1;
	float cAnisotropicScale			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1;
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;

	float cBumpIns      			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.65;
	float3 cSecondColor				 < Widget = "LinearColor3"; > = {1, 1, 1};

	float3 cThirdColor				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cBlendRate   				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	
	float cWindFrequency			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.1;
	float cMoveIntensity			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 0.1;
	float cMoveRandomness			 < Widget = "FloatSlider"; Min = "0"; Max = "20"; Step = "0.1"; Precision = "2"; > = 5;

	float cPowerLineAbs    			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.1;
	float cPowerLineInt    			 < Widget = "FloatSlider"; Min = "0"; Max = "50"; Step = "0.1"; Precision = "2"; > = 0.1;
	float cAlpha					 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "2"; > = 1.0;

	float3 cLineColorA				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cPowerLineSpd				 < Widget = "FloatSlider"; Min = "0"; Max = "5"; Step = "0.01"; Precision = "2"; > = 0.1;

	float3 cLineColorB				 < Widget = "LinearColor3"; > = {1, 1, 1};
	float cSubLight					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.0;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tMixMap : layout(t2) = "Reflection";
sampler sMixSampler : layout(s2)
{
	Texture = tMixMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t5) = "White";
SamplerComparisonState sShadowMapSampler : layout(s5)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
}; 

// Emission map
texture2D tEmissionMap : layout(t6) = "Black";
sampler sEmissionMapSampler : layout(s6)
{
	Texture = tEmissionMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
};

texture2D tPowerLineMap : layout(t7) = "Black";
sampler sPowerLineMapSampler : layout(s7)
{
	Texture = tPowerLineMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tRainMap : RainMap: layout(t8) = "NormalMap";
sampler sRainSampler : layout(s8)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tSkeletonMap : SkeletonMap : layout(t3) = "Black";
sampler sSkeletonSampler : layout(s3)
{
	Texture = tSkeletonMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
float CubicSmooth(float x)
{
	return x * x * (3.0 - 2.0 * x);
}

float TriangleWave(float x)
{
	return abs((frac(x + 0.5) * 2.0) - 1.0);
}

float TrigApproximate(float x)
{
	return (CubicSmooth(TriangleWave(x)) - 0.5) * 2.0;
}

float3 FeatherMovement(float3 worldNormal, float3 localPos, float scale, float vpaint)
{
	float randomWind = (0.3 + saturate(TriangleWave(GameTime * cWindFrequency))*0.7);

	float g = TrigApproximate(localPos.x*cMoveRandomness + GameTime);
	
	float3 globalMove = worldNormal * g * randomWind * cMoveIntensity* vpaint * scale * 0.01;

	return globalMove;
}


void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	OUT.TintColor = float4(1, 1, 1, 1);
	if(@_HasFeatherMovement)
	{
		float scale = 1.0f;
		scale = length(MulMat3x4((float3x3)World, float3(1, 0, 0)));
		if (@_VERTEXCOLOR)
		{
			OUT.TintColor = GetVertexColor(IN.Color);
			OUT.WorldPosition.xyz += FeatherMovement(OUT.WorldNormal.xyz, IN.Position.xyz, scale * LOD_RATE,OUT.TintColor.r);
		}
		else
			OUT.WorldPosition.xyz += FeatherMovement(OUT.WorldNormal.xyz, IN.Position.xyz, scale * LOD_RATE, 1);
	}

	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	if(@_EmissionMapEnable)
	{
		half2 EmissionUV = IN.TexCoord.xy * cEmissionScrolling.xy + GameTime * cEmissionScrolling.zw;
		OUT.WorldNormal.w = EmissionUV.x;
		OUT.ViewDirection.w = EmissionUV.y;
	}
	
	if (@_INSTANCED)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	}
	else
	{
		//world * VP leads to jitter issues(>2000), we use local * WVP instead.
		OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	}

	if (@_NoNearPlaneClip)
	{
		OUT.HPosition.w = max(OUT.HPosition.w, EPSILON);
		OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);
	}

	if(@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);
		//overwrite Fog T
		OUT.DiffLighting.w = FogTransmittance(normalize(OUT.ViewDirection.xyz) * cFogDistance);
	}
}  

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	if (@_BaseMapEnable)
	{
		half4 samplerBase;
		if (@_CLIPPING || @_TRANSPARENT || @_EmissionMapEnable)
			{samplerBase = tBaseMap.SampleRGBA(sBaseSampler, BaseMapUV(IN.TexCoord));}
		else
			{samplerBase = tBaseMap.SampleBias(sBaseSampler, BaseMapUV(IN.TexCoord), cBaseMapBias);}
		MaterialParameters.BaseColor = samplerBase.rgb;
		MaterialParameters.BaseColor = pow(MaterialParameters.BaseColor, 2.2f);
		MaterialParameters.Alpha = min(cAlpha, samplerBase.a);
	}
	else
	{
		MaterialParameters.BaseColor = cBaseColor;
		MaterialParameters.Alpha = cAlpha;
	}
	MaterialParameters.AO = 1.0;
	if (@_EnableReflection && @_LOD_LEVEL <= LOD_Level_2)
	{
		half3 samplerMis = tMixMap.Sample(sMixSampler, BaseMapUV(IN.TexCoord));
		half Smoothness = samplerMis.r;
		MaterialParameters.Metallic = samplerMis.g;
		MaterialParameters.AO = samplerMis.b;
		if (@_LOD_LEVEL == LOD_Level_2)
		{
			MaterialParameters.Metallic = lerp(cMetallic, MaterialParameters.Metallic, LOD_RATE);
			Smoothness = lerp(cSmoothness, Smoothness, LOD_RATE);
		}
		MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, Smoothness);
	}
	else
	{
		MaterialParameters.Metallic = cMetallic;
		if(@_EnableReflection || @_DEFERRED_SHADING)
			MaterialParameters.Roughness = lerp(cRoughnessHigh, cRoughnessLow, cSmoothness);
	}
	if(@_TANGENT && @_NormalMapEnable && @_LOD_LEVEL <= LOD_Level_1)
	{
		half3 tangentNormal = tNormalMap.SampleBias(sNormalSampler, BaseMapUV(IN.TexCoord), cNormalMapBias).xyz;
		if (@_RIPPLE)
		{
			half2 tc = IN.WorldPosition.xz * RainNormalXZScale;
			half delta = IN.WorldPosition.y * RainNormalYScale + GameTime * RainNormalTimeScale;
			half2 ttc = tc + delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			ttc = tc - delta;
			tangentNormal += RainNormalRate * tRainMap.SampleBias(sRainSampler, ttc, cNormalMapBias).xyz;
			tangentNormal /= (1 + 2 * RainNormalRate);
		}
		tangentNormal = (2*tangentNormal)-1;
		if (@_LOD_LEVEL == LOD_Level_1)
			tangentNormal.xy *= LOD_RATE;
		MaterialParameters.WorldNormal = tangentNormal.x * cBumpIns * IN.WorldTangent + tangentNormal.y *cBumpIns* IN.WorldBinormal + tangentNormal.z * IN.WorldNormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
	}
	if (@_EmissionMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half2 EmissionUV = half2(IN.WorldNormal.w, IN.ViewDirection.w);
		MaterialParameters.Emission = tEmissionMap.Sample(sEmissionMapSampler, EmissionUV).rgb;
		MaterialParameters.Emission.rgb *= cEmissionScale.rgb;
		if(@_HDR)
			MaterialParameters.Emission.rgb *= MaterialParameters.BaseColor;
		if(!@_TRANSPARENT && !@_CLIPPING)
		{
			MaterialParameters.Emission.rgb *= MaterialParameters.Alpha;
			MaterialParameters.Alpha = 1.0;
		}
	}
	if (@_PowerLineMapEnable && @_LOD_LEVEL <= LOD_Level_2)
	{
		half4 plguide = tPowerLineMap.Sample(sPowerLineMapSampler,BaseMapUV(IN.TexCoord));
		half powerLineGUIDE = plguide.r;
		half3 powerColor = lerp(cLineColorA,cLineColorB,powerLineGUIDE);
		half templine = powerLineGUIDE;
		half timeCheck = frac(GameTime*cPowerLineSpd);
		powerLineGUIDE = saturate( cPowerLineAbs - abs(timeCheck - templine));
		powerLineGUIDE += saturate( cPowerLineAbs - (1 - abs(timeCheck - templine)));
		powerLineGUIDE = saturate(powerLineGUIDE*10 + cSubLight) * sign(saturate(0.99 - templine)) * plguide.a;
		powerLineGUIDE *= cPowerLineInt;
		MaterialParameters.Emission.rgb += powerColor * powerLineGUIDE;
	}
}
#include "PBR_PS.fxh"
