#pragma shader material

///////////////choice define start///////////////////////////
#pragma shader predict("@_CLIPPING", "0")
#pragma shader predict("@_INSTANCED", "0")
#pragma shader predict("@_DF_SKIN", "0")

#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 	1 // for deferred
#define NO_LIGHT_MAP_SCALE 					1
#define USE_LIGHTMAP						0
#define USE_LIGHTMAP_CONST					0
#define VIRTUAL_LIT 						0
#define IS_WATER 							1

#define ENABLE_RIPPLE 						(@_RIPPLE)
#define HAS_2UV								(@_LIGHT_MAP)
#define ENABLE_MIXMAP						1
#define ENABLE_SCREEN_SHADOWMAP				0
#define USE_SHADOWMAP_SAMPLE				(@_LOD_LEVEL <= LOD_Level_2)
///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh" 
//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "EnableReflection?";
	Default = true;
	Visible = true; 
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable Image based lighting?";
	Default = true;
	Visible = true;	
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have LightMap?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = true;
};
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
}; 

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Do we have DepthMap?"; 
	AffectedTex = tDepthMap;
	Visible = true;
}; 
@_FlowMapEnable
{
	Name = FlowMapEnable;
	Description = "Do We have FlowMap?";
	AffectedTex = tFlowMap;
	Fallback = true;
	Visible = false;
}

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cEnvStrength	 		 < Widget = "FloatSlider"; Min = "0"; Max = "8"; Step = "0.01"; Precision = "2"; > = 1;	//环境强度[0:8]
	float cRoughness	 		 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.1;	//粗糙度下界
	float cLowQualityDepth  	 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "3"; > = 0.01;
	float cLowQualityT 	 		 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.1"; Precision = "3"; > = 0.8;

	float4 cLightMapScale			 = {1, 1, 1, 1};
	float4 cLightMapUVTransform		 = {1, 1, 0, 0};

	float4 cMainWater	     	 = { 0.1, 0.1, 5, 5 };
	float4 cSubWater     	 	 = { 0.15, 0.15, 5, 5 };
	float4 cThirdWater	     	 = { 0.3, 0.3,-15, 15 };

	float  cMainWaterBump  	 	 < Widget = "FloatSlider"; Min = "0"; Max = "3"; Step = "0.1"; Precision = "2"; > = 1.0;
	float  cSubWaterBump     	 < Widget = "FloatSlider"; Min = "0"; Max = "3"; Step = "0.1"; Precision = "3"; > = 0.5;  	
	float  cFlowSpeed		 	 < Widget = "FloatSlider"; Min = "-2"; Max = "2"; Step = "0.1"; Precision = "3"; > = 1;
	float  cDistortU	 	 	 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "2"; > = 1.0;  

	float3 cWaterColor	     	 < Widget = "LinearColor3"; > = {0.0, 0.2, 0.2};
	float  cDeepWaterRange   	 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.4;
	float3 cDeepWaterColor	 	 < Widget = "LinearColor3"; > = {0.01, 0.02, 0.02};
	float cBiasFarAwayShadow	 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.1"; Precision = "3"; > = 5;

	
};

//////////////////////////////////Input Texture////////////////////////////////////////
// Normalmap
texture2D tNormalMap : layout(t0) = "White";
sampler sNormalSampler : layout(s0)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tDepthMap : SceneDepth : layout(t1);
sampler sDepthSampler : layout(s1)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Mirror;
};

texture2D ScreenTexture : SceneColorOpaque : layout(t2);
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D tEnvMap : EnvMap : layout(t3) = "White";
sampler sEnvSampler : layout(s3)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};

// Light map
texture2D tLightMap : layout(t4) = "White";
sampler sLightMapSampler : layout(s4)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
}; 

// ShadowMap 
texture2D tShadowMap : ShadowMap : layout(t5) = "White";
SamplerComparisonState sShadowMapSampler : layout(s5)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};

// Flow map
// texture2D tFlowMap : layout(t6) = "Black";
// sampler sFlowMapSampler : layout(s6)
// {
// 	Texture = tFlowMap;
// 	Filter = Linear;
// 	Address = Clamp;
// 	SRGB = false;
// };
// but is last frame color
texture2D LastFrameRT : SceneColor : layout(t6) = "Black";
sampler LastFrameRT_Sampler : layout(s6)
{
	Texture = LastFrameRT;
	Filter = Linear;
	Address = Mirror;
};

texture2D tRainMap : RainMap: layout(t7) = "NormalMap";
sampler sRainSampler : layout(s7)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D tSkyProbeMap : SkyProbe : layout(t8) = "White";
sampler sSkyProbeSampler : layout(s8)
{
	Texture = tSkyProbeMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false; 
};
// Caustics map
// texture2D tCausticsMap : layout(t10) = "Black";
// sampler sCausticsSampler : layout(s10)
// {
// 	Texture = tCausticsMap;
// 	Filter = Linear;
// 	Address = Wrap;
// 	SRGB = false;
// };
#include "PBR_VS.fxh"
//////////////////////////////////VS////////////////////////////////////////
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	if(@_NormalMapEnable && @_TANGENT)
	{
		float2 WaterUV1 = OUT.WorldPosition.xz * cMainWater.xy + GameTime * cMainWater.zw*0.01;
		float2 WaterUV2 = OUT.WorldPosition.xz * cSubWater.xy + GameTime * cSubWater.zw*0.01;
		float2 WaterUV3 = OUT.WorldPosition.xz* cThirdWater.xy + GameTime * cThirdWater.zw*0.01;

		OUT.WorldPosition.w = WaterUV1.x;
		OUT.WorldNormal.w = WaterUV1.y;

		OUT.ViewDirection.w = WaterUV2.x;
		OUT.WorldTangent.w = WaterUV2.y;

		OUT.TexCoord.z = WaterUV3.x;
		OUT.WorldBinormal.w = WaterUV3.y;
	}
	//world * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	OUT.TexCoord.w = OUT.HPosition.w;

	if (@_VERTEXCOLOR)
		OUT.TintColor = GetVertexColor(IN.Color);
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
	[@_CastSunShadow]
	{
		OUT.LitUVZ = MulMat4x4(ShadowViewProjTexs2, float4(OUT.WorldPosition.xyz + 
			OUT.WorldNormal.xyz * CSMShadowSlopBiases[2], 1));
		OUT.LitUVZ.z -= CSMShadowBiases[2];
		// for far away shadow		
		OUT.LitUVZ2 = MulMat4x4(ShadowViewProjTexs3, float4(OUT.WorldPosition.xyz + 
		OUT.WorldNormal.xyz * cBiasFarAwayShadow, 1));
		OUT.LitUVZ2.z -= CSMShadowBiases[3];
	}
}
#include "Lighting.fxh"
#include "PBR_Common.fxh"
//////////////////////////////////PS//////////////////////////////////////// 
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	MaterialParameters.Roughness = cRoughness;
	MaterialParameters.BaseColor = cDeepWaterColor;

		half waterFoam = 0;
		half mudRange = 0;
		if(@_TANGENT && @_NormalMapEnable)
		{
			half3 waterNormal = half3(0,0,1);
			half3 waterNormal1 = half3(0,0,1);
			half3 waterNormal2 = half3(0,0,1);
			half4 waterNormal3 = half4(0,0,1,0);

		
				// float2 WaterUV1 = float2(IN.WorldPosition.w, IN.WorldNormal.w);
				// waterNormal1 = tNormalMap.Sample(sNormalSampler, WaterUV1).xyz;
				// waterNormal1.xy = waterNormal1.xy * 2 - 1;
			float2 WaterUV2 = float2(IN.ViewDirection.w, IN.WorldTangent.w);
			waterNormal2 = tNormalMap.Sample(sNormalSampler, WaterUV2).xyz;
			waterNormal2.xy = waterNormal2.xy * 2 - 1;
			
			float2 WaterUV3 = float2(LightMapUV(IN.TexCoord).x, IN.WorldBinormal.w);
			waterNormal3 = tNormalMap.SampleBias(sNormalSampler, WaterUV3 + waterNormal1.xy * 0.05, -0.5);
			waterNormal3.xy = waterNormal3.xy * 2 - 1;
			waterNormal.xy = (waterNormal1.xy * cMainWaterBump + (waterNormal2.xy + waterNormal3.xy) * cSubWaterBump) * 0.33;
			if (ENABLE_RIPPLE)
			{
				half dist = length(CameraPos.xyz - IN.WorldPosition.xyz);
				half occlusionRate = saturate((1 - dist / cParamter.w * 0.08) / 0.1);
				if(occlusionRate > 0.1)
				{
					half YIntensity = pow(saturate(IN.WorldNormal.y), 10);
					float2 RippleMove01 = -WindParam.xz * WindParam.w * GameTime * RainNormalTimeScale;
					float2 RippleUV01 = IN.WorldPosition.xz * RainNormalXZScale + RippleMove01;
					half4 Ripple01 = tRainMap.SampleLevel(sRainSampler, RippleUV01, 0);
					Ripple01.xy *= 2 - 1;
					half DropFrac01 = frac(Ripple01.w + GameTime);
					half TimeFrac01 = DropFrac01 - 1 + Ripple01.z;
					half DorpFactor01 = saturate(RainRate - DropFrac01);
					half FinalFactor01 = DorpFactor01 * Ripple01.z * sin(clamp(TimeFrac01 * 9, 0, 3) * PI);
					half2 FinalRipple01 = Ripple01.xy * FinalFactor01;
					waterNormal.xy += FinalFactor01 * RainNormalRate * YIntensity;
				}
			}
			MaterialParameters.WorldNormal = waterNormal.x*IN.WorldTangent.xyz + waterNormal.y*IN.WorldBinormal.xyz + IN.WorldNormal.xyz;
			MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
		}
		else
		{
			MaterialParameters.WorldNormal = IN.WorldNormal.xyz;
		}	
		float2 spos = IN.HPosition.xy;
		if(@_D3D9)
			{spos += 0.5;}

		float3 V = normalize( -IN.ViewDirection.xyz );
		half viewCoef = abs(V.y) * 50;
		half2 deltaUV = MaterialParameters.WorldNormal.xz * (cDistortU  * (abs(V.y)) / IN.TexCoord.w);	
		half2 baseUV = spos * ScreenInfo.zw ; 

 		float depth = DecodeDepth(tDepthMap.Sample(sDepthSampler, baseUV.xy));
		
		depth = (depth * CameraInfo.y - IN.TexCoord.w) ;
			
		half edgeSoft = saturate(depth);
		
		 
		 if(@_DepthMapEnable&& @_LOD_LEVEL <= LOD_Level_2)
		{

			baseUV = baseUV +deltaUV*edgeSoft;
			half deepWaterRange = saturate(depth * cDeepWaterRange * 0.1);
			deepWaterRange = sqrt(deepWaterRange);

			MaterialParameters.CustomData.rgb = ScreenTexture.Sample(ScreenTexture_Sampler, baseUV).rgb;
			
			half3 deepWaterColor = lerp(cDeepWaterColor, cWaterColor, mudRange);
			MaterialParameters.BaseColor = deepWaterColor + waterFoam;
			MaterialParameters.CustomData.a = deepWaterRange;
			MaterialParameters.Alpha = edgeSoft;
		}
		else
		{
		MaterialParameters.CustomData.a=1;	
		MaterialParameters.BaseColor =cDeepWaterColor;
		half nov =  dot(V,normalize(IN.WorldNormal.xyz) );
		nov = 1-nov;
		MaterialParameters.Alpha = edgeSoft*nov;
		// MaterialParameters.Alpha = edgeSoft;
		}
		
		
		// {
		// 	baseUV += MaterialParameters.WorldNormal.xy * cLowQualityDepth * half2(cDistortU * 10, 0);
		// 	MaterialParameters.CustomData.rgb = ScreenTexture.Sample(ScreenTexture_Sampler, baseUV).rgb;
		// 	MaterialParameters.CustomData.rgb = lerp(cDeepWaterColor, MaterialParameters.CustomData.rgb, cLowQualityT);
		// }
		//MaterialParameters.Roughness = saturate(MaterialParameters.Roughness + (1 - lodRate));
	
}
#include "BRDF.fxh"
#include "ScreenSpaceReflection.fxh"
half GetWaterDynamicShadow(in VertexShadingOutput IN)
{
	half inShadow = 0;
	//get dynamic shadow
	if (@_CastSunShadow && @_LOD_LEVEL <= LOD_Level_2)
	{
		float3 LitUVZ = IN.LitUVZ.xyz / IN.LitUVZ.w;
		LitUVZ.z = min(1.0-10*HALF_EPSILON, LitUVZ.z);

		float3 LitUVZ2 = IN.LitUVZ2.xyz / IN.LitUVZ2.w;
		LitUVZ2.z = min(1.0-10*HALF_EPSILON, LitUVZ2.z);

		float2 vInvShadowMapWH = float2(SHADOW_PIXELSIZE, SHADOW_PIXELSIZE);
		half ShadowRate = 0.5;


		half2 inRange = abs(LitUVZ.xy - 0.5);
		ShadowRate = min(0.5-inRange.x, 0.5-inRange.y);
		float2 pInLitTC = LitUVZ.xy;
		float litZ = LitUVZ.z;
		half isInRange = max(inRange.x, inRange.y);

		if(isInRange > 0.49)
		{
			GetHexShadow(LitUVZ2, 3, inShadow);
		}
		else
		{
			GetHexShadow(LitUVZ, 2, inShadow);
		}
	}
	inShadow = min(1.0, inShadow);
	
	return inShadow;
}
//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput PbrVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);
	
	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);
#if IS_HEIGHTMAP
	HEIGHTMAP_MODIFIER(IN, INTM);
#endif

	//OUTPUT
	VertexUnifyOutputCommon(IN, INTM, OUT);

	return OUT;
}

float4 PbrPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);

	//1, Gen MaterialParameters & GBufferData
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	PS_GetParameters(IN, MaterialParameters);
	OUT.a = MaterialParameters.Alpha;
	//screen gray
	// InitScreenGray(IN, MaterialParameters.BaseColor);
	// MaterialParameters.InShadow = GetWaterDynamicShadow(IN);
	//2, lighting
	LightingResult LitResult;
	LitResult.DiffuseLighting = half3(0, 0, 0);
	LitResult.SpecularLighting = half3(0, 0, 0);
	int PBRQualityLevel = PBRQualityLevel_Normal;
	half3 V = normalize(-IN.ViewDirection.xyz);
	half NoV = saturate(dot(MaterialParameters.WorldNormal, V));
	half3 R = reflect(-V, MaterialParameters.WorldNormal);
	//GI lighting
	FetchPrecomputedLighing(IN, MaterialParameters);
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	//GI Specular lighting
	// ModifyReflectionDirection(IN, R);
	half3 GISpecular = half3(0, 0, 0);
	float3 SSR =float3(0,0,0);
	
		

	
	if(@_LOD_LEVEL <= LOD_Level_1)
	{
		SSR = ScreenSpaceReflection(IN.WorldPosition.xyz, IN.HPosition.xy, R, NoV);
	if(SSR.z > 0 && GBuffer.Roughness < 0.5)
		GISpecular = LastFrameRT.Sample(LastFrameRT_Sampler, SSR.xy);
	}
	 
	half3 sampleEnvSpecular = half3(0,0,0);
		half MIP_ROUGHNESS = 0.17;
		half level = GBuffer.Roughness / MIP_ROUGHNESS;;
		half fSign = R.z > 0;
		half fSign2 = fSign*2 - 1;
		R.xy /= R.z*fSign2 + 1;
		R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
		
		half4 srcColor;
		srcColor = tEnvMap.SampleLevel(sEnvSampler, R.xy, level*RenderToMip);
		sampleEnvSpecular = RGBMDecode(srcColor, MAX_PBR_RGBM_VALUE) ;
		

	
	sampleEnvSpecular = EnvBRDFApprox(GBuffer.SpecularColor, GBuffer.Roughness, NoV) * sampleEnvSpecular;
	half3 IBLSpec = sampleEnvSpecular;
	GISpecular = lerp(IBLSpec, GISpecular, saturate(SSR.z));
	
	//sun lighting
	LightingData LitData = GetSunLightData(V, GBuffer, PBRQualityLevel);
	half3 InputRadiance = LitData.FalloffColor * LitData.Atten * (1 - LitData.InShadow);

	
		if(@_LOD_LEVEL<=LOD_Level_1)
	{
	half VoL = saturate(dot(LitData.V, -SunDirection.xyz));
	half Wrap = 0.5;
	half WrapNoL = saturate( ( -LitData._NoL + Wrap ) / Square( 1 + Wrap ) );
	half Scatter = D_GGX( 0.8, VoL );
	OUT.rgb += InputRadiance * WrapNoL * Scatter * GBuffer.BaseColor * PI;
	}
	
	else if(@_LOD_LEVEL>=LOD_Level_2)
	{
		 half3 H = normalize(LitData.V );
		 half NoH = saturate(dot(MaterialParameters.WorldNormal.rgb, H));
		LitResult.SpecularLighting += pow(1-NoH,5)*GBuffer.SpecularColor*InputRadiance.x/PI;
	}
	
	//sun subsurface lighting


	//thunder shading
	LitResult.SpecularLighting *= ThunderRate;
	LitResult.DiffuseLighting *= ThunderRate;
	//point lighting calc include vs lighting and ps lighting
	InitPointLighting(IN, GBuffer, LitResult, V, NoV);
	//accumulate specular and diffuse
	OUT.rgb += GBuffer.Emission;//emission without GI lighting
	OUT.rgb += (LitResult.DiffuseLighting + GBuffer.GILighting) * GBuffer.DiffuseColor;
	OUT.rgb += (LitResult.SpecularLighting + GISpecular) * MaterialParameters.Alpha;
	OUT.rgb = lerp(GBuffer.CustomData.rgb, OUT.rgb,  MaterialParameters.CustomData.a);
	

	//4, Fog shading.
	if (!@_DEFERRED_SHADING)
		InitFogLighting(IN, LitData, OUT);
	
	//5, screen color tint
#if USE_LIGHTMAP
	if (@_LIGHT_MAP || LIGHTMAP_ENABLE)
		OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);
#endif
	
	EncodeHDR(OUT.rgb);
	return OUT;
}

// Output patch constant data.
#if USE_TESSELLATION
//////////////////////////////////Tessellation diy function start////////////////////////////////////////
float4 HS_GetTessellationFactor(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2)
{
	return float4(4, 4, 4, 4);
}

float3 Project(float3 p, float3 c, float3 n)
{
	return p - dot(p - c, n) * n;
}
void DS_DIY_OUTPUT(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2, float3 domain, inout VertexShadingOutput OUT)
{
	float3 c0 = Project(OUT.WorldPosition.xyz, p0.WorldPosition.xyz, normalize(p0.WorldNormal.xyz));
	float3 c1 = Project(OUT.WorldPosition.xyz, p1.WorldPosition.xyz, normalize(p1.WorldNormal.xyz));
	float3 c2 = Project(OUT.WorldPosition.xyz, p2.WorldPosition.xyz, normalize(p2.WorldNormal.xyz));
	float3 pos = c0 * domain.x + c1 * domain.y + c2 * domain.z;
	OUT.WorldPosition.xyz = lerp(OUT.WorldPosition.xyz, pos, 0.5);
}
//////////////////////////////////Tessellation diy function end////////////////////////////////////////
#include "TessellationPass.fxh"
#endif

technique Shading
{
	VertexShader = PbrVertexShading();
#if USE_TESSELLATION	
	HullShader = PbrVertexShading_HS();
	DomainShader = PbrVertexShading_DS();
#endif
	PixelShader = PbrPixelShading();
}

#include "DebugShaderPass.fxh"
#include "WireframePass.fxh"
#include "EdgeLineNormalGenPass.fxh"

