//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle_Twisted.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Particle shading model.

#pragma shader material
//#pragma shader category("Effect")

//////////////////////////////////Properties////////////////////////////////////////
@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = true;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask map enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

@_Mask02Enable
{
	Name = Mask02Enable;
	Description = "Is mask02 map enabled?";
	AffectedTex = MaskTexture02;
	Visible = true;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
};

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = false;
}

#include "Constants.fxh"
#include "Common.fxh"
#define FORCE_DISABLE_ENCODE 1

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World				: layout(c0);
//	float4x4 WorldViewProj 		: WorldViewProj		: layout(c3);
	float4   InstanceColor		: Color				: layout(c3);
//	float4   cTintColor1		: TintColor1 	: layout(c4);
};

cbuffer Shader
{
	float4 ModulateColor		= { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	= { 1, 1, 0, 0 };
	float4 TwistUVScolling		= { 1, 1, 0, 0 };
	float4 MaskUVScolling		= { 1, 1, 0, 0 };
	float2 TwistStrength		= { 0.1, 0.1 };
	float  BloomScale			= 0.5;
	float  InvFadeDistance		= 1;
	float  FogStrength			= 0;
	float  DepthBias			= 0;
	float  HDRScale				= 1;
	float2 UVScale				= {1,1};
	float  InvertCenter			= 0.5;
	float  InvertExtend			= 0.5;
	float4 InverterColor		= { 0.5, 0.5, 0.5, 1 };
	float  InverterScale		= 1.0;
	float  cFov					 = 1.0;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D TwistTexture : layout(t2) = "Black";
sampler TwistTexture_Sampler : layout(s2)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t3) = "White";
sampler MaskTexture_Sampler : layout(s3)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture02 : layout(t4) = "White";
sampler MaskTexture02_Sampler : layout(s4)
{
	Texture = MaskTexture02;
	Filter = Linear;
	Address = FromTexture;
};

texture2D ScreenTexture : SceneColor : layout(t6) = "Transparent";
sampler ScreenTexture_Sampler : layout(s6)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t7);
sampler DepthBuffer_Sampler : layout(s7)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
#include "Particle_Common.fxh"

void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float4 OUT = float4(0,0,0,0);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + GameTime * TwistUVScolling.zw;
	float2 maskTC = IN.TexCoord.xy * MaskUVScolling.xy + GameTime * MaskUVScolling.zw;
	float2 twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;
	diffuseTC += twist * TwistStrength;
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);

	float4 DColor = float4(1,0,1,0.2);
	float lerpValue = 0;

	OUT.xyz = (OUT.xyz * IN.VertexColor.xyz * InstanceColor.xyz);
	if(@_DISTORTION)
	{	
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float4 screen = SampleColorMap(ScreenTexture, ScreenTexture_Sampler, stc);
		screen.rgb = EncodeGamma(screen.rgb);
		DColor.xyz = (screen.x + screen.y + screen.z)*(InverterScale*0.333);
		DColor.xyz += OUT.xyz * OUT.a;
		DColor = clamp(DColor,0,MAX_EMISSIVE_VALUE);
		lerpValue = smoothstep((InvertCenter-InvertExtend)*(InverterScale+1),(InvertCenter+InvertExtend)*(InverterScale+1),max(DColor.x,max(DColor.y,DColor.z)));
	}
	else
	{
		DColor.xyz = OUT.xyz * OUT.a;
		lerpValue = smoothstep(InvertCenter-InvertExtend,InvertCenter+InvertExtend,max(DColor.x,max(DColor.y,DColor.z)));
	}

	
	float4 MultipleColor = lerp(ModulateColor,InverterColor,(lerpValue));
	OUT *= MultipleColor;

	OUT.xyz *= OUT.a;
	OUT.a = OUT.a * IN.VertexColor.a;
 	materialParamters.BaseColor = OUT;

}	

#include "Particle_Shading.fxh"

float4 ParticleCustomPixelShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	//Here Distortion is Different
	//Using ScreenColor to Invert Color 
	//Not Offset ScreenUV
	//Particle_Shading.fx
    return UnlitPixelShading(IN,screenCoord);
}

technique Shading
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleCustomPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

