//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : lnn4047@corp.netease.com
//	Author: liuning
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<RadarEffect.fx>
///	@path	~/System/Shader/
///	@date	2020/11/10
///	@brief	Radar effect.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////

cbuffer Batch
{
	float cRadarEffectIntensity;
};

///////////////// samplers //////////////////
// Backbuffer target.
texture2D tSourceMap : layout(t0);
sampler sSourceMapSampler : layout(s0)
{
	Texture = tSourceMap;
	Filter = Linear;
	Address = Clamp;
};

// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput RadarEffectVS(VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 RadarEffectPS(VertexScreenOutput IN) : SV_Target
{
	half3 Source = tSourceMap.Sample(sSourceMapSampler, IN.TC);
	half3 Grayed = GetLuminance(Source) * 0.5;
	// Darken corner.
	Grayed *= saturate(1 - abs(0.5*0.8 - IN.TC.x*0.8) * abs(0.5*0.8 - IN.TC.y*0.8) * 5);
	half3 Final = lerp(Source, Grayed, cRadarEffectIntensity);
	return float4(Final, 1.0);
}

technique RadarEffect
{
	VertexShader = RadarEffectVS();
	PixelShader = RadarEffectPS();
}
