
//- Import from libraries.
import lib-env.glsl
import lib-vectors.glsl
import lib-alpha.glsl
import lib-pbr.glsl

// Link Metal/Roughness MDL for Iray
//: metadata {
//:   "mdl":"mdl::alg::materials::physically_metallic_roughness::physically_metallic_roughness"
//: }

//- Show back faces as there may be holes in front faces.
//: state cull_face off

//- Channels needed for metal/rough workflow are bound here.
//: param auto channel_basecolor
uniform sampler2D basecolor_tex;
//: param auto channel_roughness
uniform sampler2D roughness_tex;
//: param auto channel_metallic
uniform sampler2D metallic_tex;
//: param auto channel_specularlevel
uniform sampler2D specularlevel_tex;


//: param custom {
//:   "default": 12,
//:   "label": "Time",
//:   "min": 0.0,
//:   "max": 24.0,
//:   "group": "Messiah Parameters"
//: }
uniform float SunTime;

//: param custom {
//:   "default": 0,
//:   "label": "Longitude",
//:   "min": 0.0,
//:   "max": 360.0,
//:   "group": "Messiah Parameters"
//: }
uniform float SunLongitude;

//: param custom {
//:   "default": 1,
//:   "label": "Sun Color",
//:   "widget": "color",
//:   "group": "Messiah Parameters"
//: }
uniform vec3 SunColor;

//: param custom {
//:   "default": 1,
//:   "label": "Sun Color Intensity",
//:   "min": 0.0,
//:   "max": 30.0,
//:   "group": "Messiah Parameters"
//: }
uniform float SunColorIntensity;

//: param custom {
//:   "default": 6500,
//:   "label": "Temperature",
//:   "min": 0.0,
//:   "max": 12000.0,
//:   "group": "Messiah Parameters"
//: }
uniform float Temperature;

//: param custom {
//:   "default": 1,
//:   "label": "Sun Ambient",
//:   "widget": "color",
//:   "group": "Messiah Parameters"
//: }
uniform vec3 SunAmbient;

//: param custom {
//:   "default": 0,
//:   "label": "Sun Ambient Intensity",
//:   "min": 0.0,
//:   "max": 100.0,
//:   "group": "Messiah Parameters"
//: }
uniform float SunAmbientIntensity;

//: param custom {
//:   "default": 0,
//:   "label": "IBL Specular",
//:   "min": 0.0,
//:   "max": 10.0,
//:   "group": "Messiah Parameters"
//: }
uniform float IBLSpe;

//: param custom {
//:   "default": 0,
//:   "label": "IBL Diffuse",
//:   "min": 0.0,
//:   "max": 10.0,
//:   "group": "Messiah Parameters"
//: }
uniform float IBLDiffuse;


//: param custom {
//:   "default": 0,
//:   "label": "Anisotropic Scale",
//:   "min": 0.0,
//:   "max": 10.0,
//:   "group": "Messiah Parameters"
//: }
uniform float AnisotropicScale;

//: param custom {
//:   "default": 0,
//:   "label": "RoughnessX",
//:   "min": 0.0,
//:   "max": 1.0,
//:   "group": "Messiah Parameters"
//: }
uniform float cRoughnessX;

//: param custom {
//:   "default": 0,
//:   "label": "RoughnessY",
//:   "min": 0.0,
//:   "max": 1.0,
//:   "group": "Messiah Parameters"
//: }
uniform float cRoughnessY;

//: param custom {
//:   "default": 1,
//:   "label": "Virtual Lit Color",
//:   "widget": "color",
//:   "group": "Messiah Parameters"
//: }
uniform vec3 VlitColor;

//: param custom {
//:   "default": 0,
//:   "label": "Virtual Lit Intensity",
//:   "min": 0.0,
//:   "max": 100.0,
//:   "group": "Messiah Parameters"
//: }
uniform float VlitIntense;

//: param custom { "default": false, "label": "HDR Tonemapping", "group": "Messiah Parameters"}
uniform bool HDRTonemapping;

//: param custom { "default": true, "label": "Auto Exposure", "group": "Messiah Parameters"}
uniform bool AutoExposure;

//: param custom {
//:   "default": 1,
//:   "label": "Eye Adaption Level",
//:   "min": 0,
//:   "max": 1,
//:   "group": "Messiah Parameters"
//: }
uniform float EyeAdaptionLevel;

//: param custom {
//:   "default": 4,
//:   "label": "Calibration Constant",
//:   "min": 0,
//:   "max": 20,
//:   "group": "Messiah Parameters"
//: }
uniform float EyeAdaptionScale;

//: param custom {
//:   "default": 0,
//:   "label": "Exposure Compensation",
//:   "min": -4,
//:   "max": 4,
//:   "group": "Messiah Parameters"
//: }
uniform float ExposureCompensation;

//: param custom {
//:   "default": 1,
//:   "label": "Fix Adaption Lum",
//:   "min": 0,
//:   "max": 4,
//:   "group": "Messiah Parameters"
//: }
uniform float FixAdaptionLum;

vec3 EnvBRDFApprox(vec3 SpecularColor, float Roughness, float NoV)
{
  // [ Lazarov 2013, "Getting More Physical in Call of Duty: Black Ops II" ]
  // Adaptation to fit our G term.
  const vec4 c0 = { -1, -0.0275, -0.572, 0.022 };
  const vec4 c1 = { 1, 0.0425, 1.04, -0.04 };
  vec4 r = Roughness * c0 + c1;
  float a004 = min( r.x * r.x, exp2( -9.28 * NoV ) ) * r.x + r.y;
  vec2 AB = vec2( -1.04, 1.04 ) * a004 + r.zw;
  return SpecularColor * AB.x + AB.y;
}

float D_GGX(float Roughness, float NoH )
{
  float m = Roughness * Roughness;
  float m2 = m * m;
  float d = ( NoH * m2 - NoH ) * NoH + 1; // 2 mad
  return m2 / ( 3.1415926535897932384626433832795f*d*d );         // 2 mul, 1 rcp
}

vec3 aniso(vec3 inputRadiance, float Roughness, vec3 SpecularColor, float NoL, float NoV, float NoH, float HoT, float HoB)
{
    float RoughnessX = cRoughnessX * Roughness + 0.0001;
    float RoughnessY = cRoughnessY * Roughness + 0.0001;
    float beta_a = HoT / RoughnessX;
    beta_a *= beta_a;
    float beta_b = HoB / RoughnessY;
    beta_b *= beta_b;
    float beta = -2.0 * (beta_a + beta_b) / (1.0 + NoH);
    float s_den = max(0.001, 12.5663706 * RoughnessX * RoughnessY * sqrt(NoL * NoV));
    return AnisotropicScale * SpecularColor * inputRadiance * exp(beta) / s_den;
}

float DegreeToRadian(float degree)
{
    return degree / 180 * 3.1415926535897932384626433832795f;
}


//Temperature
vec3 MakeFromColorTemperature(float Temp)
{
  Temp = clamp(Temp, 1000.0f, 15000.0f);

  // Approximate Planckian locus in CIE 1960 UCS
  float u = (0.860117757f + 1.54118254e-4f * Temp + 1.28641212e-7f * Temp*Temp) / (1.0f + 8.42420235e-4f * Temp + 7.08145163e-7f * Temp*Temp);
  float v = (0.317398726f + 4.22806245e-5f * Temp + 4.20481691e-8f * Temp*Temp) / (1.0f - 2.89741816e-5f * Temp + 1.61456053e-7f * Temp*Temp);

  float x = 3.0f * u / (2.0f * u - 8.0f * v + 4.0f);
  float y = 2.0f * v / (2.0f * u - 8.0f * v + 4.0f);
  float z = 1.0f - x - y;

  float Y = 1.0f;
  float X = Y / y * x;
  float Z = Y / y * z;

  // XYZ to RGB with BT.709 primaries
  float R = 3.2404542f * X + -1.5371385f * Y + -0.4985314f * Z;
  float G = -0.9692660f * X + 1.8760108f * Y + 0.0415560f * Z;
  float B = 0.0556434f * X + -0.2040259f * Y + 1.0572252f * Z;

  return vec3(R, G, B);
}

vec3 SunColorTemp(vec3 sunColor, float temp)
{
  return sunColor * MakeFromColorTemperature(temp);
}


vec3 ACES_Knarkowicz( vec3 x )
{
    float a = 2.51;
    float b = 0.03;
    float c = 2.43;
    float d = 0.59;
    float s = 0.14;
    return (x*(a*x+b))/(x*(c*x+d)+s);
}

vec3 Tonemapping(vec3 HDRColor)
{
  float AdaptionLevel = EyeAdaptionLevel;
  float AdaptionScale = EyeAdaptionScale;
  if (AutoExposure)
  {
    AdaptionLevel = 1.0;
    AdaptionScale = 4.0;
  }

  //+- EV
  HDRColor *= pow(2.0, ExposureCompensation);

  // Exposure
  float AdaptedLum = FixAdaptionLum * AdaptionScale;
  float AdaptedLumDest = 1.0 / (AdaptionLevel + AdaptedLum);
  HDRColor = AdaptedLumDest * HDRColor;
  

  // Tonemapping
  vec3 LDRColor = ACES_Knarkowicz(HDRColor);

  return LDRColor;
}

vec3 pbrMessiah(V2F inputs, vec3 diffColor, vec3 specColor, float roughness)
{
    LocalVectors vectors = computeLocalFrame(inputs);
   
    //IBL specular
    //float lodS = roughness < 0.01 ? 0.0 : (roughness * 13.5);
    float nov = dot(vectors.eye, vectors.normal);
    //vec3 Ln = -reflect(vectors.eye, vectors.normal);
    //vec3 srcIBL = envSampleLOD(Ln, lodS);
    //vec3 IBL_spe = EnvBRDFApprox(specColor, roughness, nov) * srcIBL;
    vec3 IBL_spe = pbrComputeSpecular(vectors, specColor, roughness);

    //Dir specular/diffuse
    float latitude = DegreeToRadian(SunTime * 360.0 / 24.0 + 180.0);
    float longitude = DegreeToRadian(SunLongitude);
    float sinLat = sin(latitude);
    float cosLat = cos(latitude);
    float sinLong = sin(longitude);
    float cosLong = cos(longitude);
    vec3 SunDir = vec3(sinLat * cosLong, cosLat, sinLat * sinLong);
    vec3 l = normalize(SunDir);

    float nol = max(dot(l, vectors.normal),0);
    vec3 h = normalize(vectors.eye + l);
    float noh = max(dot(h, vectors.normal), 0);
    vec3 dirLit = nol * SunColorTemp(pow(SunColor,vec3(2.2)) * SunColorIntensity, Temperature);
    vec3 spe = D_GGX(roughness, noh) * 0.25 * specColor * dirLit;

    float HoT = dot(h, vectors.tangent);
    float HoB = dot(h, vectors.bitangent);
    vec3 Anisotropic = aniso(dirLit, roughness, specColor, nol, nov, noh, HoT, HoB);

    vec3 vLit = max(dot(-camera_dir, vectors.normal), 0) * VlitColor * VlitIntense;

    vec3 outColor = (envIrradiance(vectors.normal) * IBLDiffuse + dirLit + SunAmbient * SunAmbientIntensity + vLit) * diffColor + IBL_spe * IBLSpe + spe + Anisotropic;
    
    if(HDRTonemapping)
    {
      outColor = Tonemapping(outColor);
    }

    return outColor;
}

//- Shader entry point.
void shade(V2F inputs)
{
  alphaKill(inputs.tex_coord);

  float roughness = getRoughness(roughness_tex, inputs.tex_coord);

  vec3 baseColor = getBaseColor(basecolor_tex, inputs.tex_coord);
  //baseColor = baseColor * baseColor;

  float metallic = getMetallic(metallic_tex, inputs.tex_coord);

  vec3 diffColor = (baseColor - baseColor * metallic) / 3.1415926535897932384626433832795f;

  vec3 specColor = mix(vec3(0.04, 0.04, 0.04), baseColor, metallic);

  // Get detail (ambient occlusion) and global (shadow) occlusion factors
  //float occlusion = getAO(inputs.tex_coord) * getShadowFactor();

  // Feed parameters for a physically based BRDF integration
  vec3 final = pbrMessiah(inputs, diffColor, specColor, roughness) * getShadowFactor();
  diffuseShadingOutput(final);
}

//- Entry point of the shadow pass.
void shadeShadow(V2F inputs)
{
  // Discard current fragment on the basis of the opacity channel
  // and a user defined threshold
  alphaKill(inputs.tex_coord);
}