#pragma shader material
#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "false")
#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

///////////////choice define start///////////////////////////
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 	1 // for deferred
#define USE_LIGHTMAP						0
#define NO_SHADOW_CLIPPING					1
#define USE_LIGHTMAP_CONST					0
#define USE_CLIPPING						1
#define USE_WIREFRAME						1
#define USE_FROZEN							0
#define USE_BURNT							0
#define USE_HIGHLIGHT						0
#define USE_XRAY							0
#define USE_EDGELINE						0
#define USE_PLANAR_SHADOW					0
#define USE_VBufferGen						0
#define VIRTUAL_LIT 						0

#define IS_TERRAIN 							1
#define DEFERRED_PRE_AO						1

#define half  float
#define half2 float2
#define half3 float3
#define half4 float4

#define HAS_2UV								(@_LIGHT_MAP)
#define ENABLE_MIXMAP						(@_EnableReflection)
#define ENABLE_RIPPLE 						(@_RIPPLE && @_LOD_LEVEL <= LOD_Level_1)
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow)
///////////////choice define end//////////////////////////

#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"
//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "Enable reflection?";
	Default = true;
	Visible = false;
};
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable image based lighting?";
	Default = true;
	Visible = false;
};
@_EnableCliff
{
	Name = EnableCliff;
	Description = "Enable cliff?";
	Default = false;
	Visible = false;
	Predict = false;
};
@_EnableBlendSharpness
{
	Name = EnableBlendSharpness;
	Description = "Enable blend sharpness?";
	Default = false;
	Visible = false;
};
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have light map?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = false;
};
@_BlockMapEnable
{
	Name = BlockMapEnable;
	Description = "Do we have block map?";
	AffectedTex = tBlockMap;
	Visible = false;
	Predict = true;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have normal map?";
	AffectedTex = tNormalMap;
	Fallback = true;
	Visible = false;
};
@_MixtureMapEnable
{
	Name = MixtureMapEnable;
	Description = "Do we have mixture map?";
	AffectedTex = tMixtureMap;
	Fallback = true;
	Visible = false;
};
@_OverlayMapEnable
{
	Name = OverlayMapEnable;
	Description = "Do we have overlay map?";
	AffectedTex = tOverlayMap;
	Visible = false;
	
};
@_AuxiliaryMapEnable
{
	Name = AuxiliaryMapEnable;
	Description = "Do we have auxiliary map?";
	AffectedTex = tAuxiliaryMap;
	Visible = false;
};
@_SurfaceTextureArrayEnable
{
	Name = SurfaceTextureArrayEnable;
	Description = "Enable texture array for terrain?";
	AffectedTex = tSurfaceBlockMap;
	Visible = false;
};
@_SurfaceTextureArrayNormalEnable
{
	Name = SurfaceTextureArrayNormalEnable;
	Description = "Enable texture normal array for terrain?";
	AffectedTex = tSurfaceNormalMap;
	Visible = false;
};
@_SurfaceTextureArrayMixtureEnable
{
	Name = SurfaceTextureArrayMixtureEnable;
	Description = "Enable texture mixture array for terrain?";
	AffectedTex = tSurfaceMixtureMap;
	Visible = false;
};
@_RadarEnable
{
	Name = ReceiveCubeReflect;
	Description = "for Glass";
	Default = false;
	Visible = false;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cAliasingFactor			 = 0;   //抗锯齿因子
	float cEnvStrength				 = 1;   //环境强度[0:8]
	float cRoughnessLow				 = 0;   //粗糙度下界
	float cRoughnessHigh			 = 1;   //粗糙度上界

	float cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cBlendSharpness			 < Widget = "FloatSlider"; Min = "0.1"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.3;

	float4 cBlockScale				 = { 1, 0, 0, 0 };
	float4 cBlockShrinkage			 = { 0, 0, 1, 1 };

	float4 cBlendMapSize			 = { 1, 1, 0, 0 };

	float4 cLightMapScale			 = { 1, 1, 1, 1 };
	float4 cLightMapUVTransform		 = { 1, 1, 0, 0 };
	float4 cOverlayMapUVTransform	 = { 1, 1, 0, 0 };
	float4 cBlockScaleList0			 = {1, 1, 1, 1};
	float4 cBlockScaleList1			 = {1, 1, 1, 1};
	float4 cBlockScaleList2			 = {1, 1, 1, 1};
	float4 cBlockScaleList3			 = {1, 1, 1, 1};

	float 	cParallaxStrength			 = 0.5;
	float3 	cRadarColor				< Widget = "LinearColor3"; > = {1, 1, 1};
	float	cRadarScrolling			={0};
	float 	cPlayerDirection			 = {1};
};

texture2D tIndexMap : layout(t0) = "Black";
sampler sIndexMapSampler : layout(s0)
{
	Texture = tIndexMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlendMap : layout(t1) = "Black";
sampler sBlendMapSampler : layout(s1)
{
	Texture = tBlendMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlockMap : layout(t2);
sampler sBlockMapSampler : layout(s2)
{
	Texture = tBlockMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tMixtureMap : layout(t3) = "Reflection";
sampler sMixtureMapSampler : layout(s3)
{
	Texture = tMixtureMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tNormalMap : layout(t4) = "NormalMap";
sampler sNormalMapSampler : layout(s4)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tVTIndMap : VTIndMap : layout(t0) = "Black";
sampler sVTIndSampler : layout(s0)
{
	Texture = tVTIndMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tVTBaseNormalXMap : VTBaseNormalXMap: layout(t1) = "Black";
sampler sVTBaseNormalXSampler : layout(s1)
{
	Texture = tVTBaseNormalXMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tVTRoughnessAOHeightNormalYMap : VTRoughnessAONormalYMap : layout(t2) = "Black";
sampler sVTRoughnessAOHeightNormalYSampler : layout(s2)
{
	Texture = tVTRoughnessAOHeightNormalYMap;
	Filter = Linear;
	Address = Clamp;
};

texture2DArray tSurfaceBlockMap : layout(t2);
sampler2DArray sSurfaceBlockSampler : layout(s2)
{
	Texture = tSurfaceBlockMap;
	Filter = Linear;
	Address = Wrap;
};

texture2DArray tSurfaceNormalMap : layout(t3);
sampler2DArray sSurfaceNormalSampler : layout(s3)
{
	Texture = tSurfaceNormalMap;
	Filter = Linear;
	Address = Wrap;
};

texture2DArray tSurfaceMixtureMap : layout(t4);
sampler2DArray sSurfaceMixtureSampler : layout(s4)
{
	Texture = tSurfaceMixtureMap;
	Filter = Linear;
	Address = Wrap;
};

texture2D tLightMap : layout(t5);
sampler sLightMapSampler : layout(s5)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tEnvMap : EnvMap : layout(t6);
sampler sEnvSampler : layout(s6)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tShadowMask : ShadowMask : layout(t7);
sampler sShadowMaskSampler : layout(s7)
{
	Texture = tShadowMask;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tShadowMap : ShadowMap : layout(t7);
SamplerComparisonState sShadowMapSampler : layout(s7)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tOverlayMap : layout(t8) = "White";
sampler sOverlayMapSampler : layout(s8)
{
	Texture = tOverlayMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tRainMap : RainMap : layout(t9) = "NormalMap";
sampler sRainSampler : layout(s9)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D tAuxiliaryMap : layout(t10) = "NormalMap";
sampler sAuxiliaryMapSampler : layout(s10)
{
	Texture = tAuxiliaryMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	OUT.ViewDirection.w = 0;

	// World * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));

	if (@_LOD_LEVEL <= LOD_Level_3)
	{
		if (@_EnableCliff)
		{
			half up = step(OUT.WorldNormal.x, OUT.WorldNormal.y) * step(OUT.WorldNormal.z, OUT.WorldNormal.y);
			OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, half3(0, 1 - up, up)));
		}
		else
		{
			OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, half3(0, 0, 1)));
		}
		OUT.WorldBinormal.xyz = cross(OUT.WorldNormal.xyz, OUT.WorldTangent.xyz);
	}
}

#include "Terrain_Common.fxh"
#include "VT_Common.fxh"
#include "PBR_Common.fxh"
//////////////////////////////////PS////////////////////////////////////////

void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	float3 V = normalize( -IN.ViewDirection.xyz );
	float3 tangentV = float3(dot(IN.WorldTangent,V),dot(IN.WorldBinormal,V),dot(IN.WorldNormal,V));
	float2 uv = BaseMapUV(IN.TexCoord);
	half3 diff = half3(1, 1, 1);
	half3 norm = half3(0.5, 0.5, 1);
	half height  = 0.5;
	half3 mixx = half3(0, 0, 1);
	float2 tc = float2(0, 0);

	float dist = length(CameraPos.xyz - IN.WorldPosition.xyz);
	half rateSpec = saturate((1 - dist / cParamter.w * 0.8) / 0.8);
	half rateBase = saturate((1 - dist / cParamter.w * 0.2) / 0.5);

	float3 worldPos = IN.WorldPosition.xyz;
	if (@_VTEnable)
	{
		if(IS_WINDOWS)
		{
			if(cParallaxStrength > 0)
			{
				float3 worldPosPOM = VTPOM(IN.WorldPosition.xyz, tangentV, V, IN.WorldNormal.xyz);
				worldPos = lerp(worldPos, worldPosPOM, cParallaxStrength);
			}
		}	
		SampleVT(worldPos, diff, norm, mixx);
	}
	else
	{
		PS_IndexBlend(IN.WorldNormal.xyz, IN.WorldPosition.xyz, uv, rateSpec, diff, norm, height, mixx, 1, 0.2);
	}

	//overlay.
	if (@_OverlayMapEnable && @_LOD_LEVEL >= LOD_Level_1)
	{
		tc = BaseMapUV(IN.TexCoord).yx * cOverlayMapUVTransform.xy + cOverlayMapUVTransform.zw;
		half4 lodBaseColor = tOverlayMap.Sample(sOverlayMapSampler, tc);
		//diff = lerp(lodBaseColor.rgb,diff,rateBase);
	}

	MaterialParameters.BaseColor = DecodeGamma(diff);
	
	if (@_LOD_LEVEL <= LOD_Level_1 && (@_MixtureMapEnable || @_SurfaceTextureArrayMixtureEnable || @_VTEnable))
	{
		MaterialParameters.Metallic = 0.0;
		MaterialParameters.Roughness = 1 - lerp(0.2, mixx.r, rateSpec);
		MaterialParameters.AO = lerp(1, mixx.b, rateSpec);
	}
	else
	{
		MaterialParameters.Metallic = 0;
		MaterialParameters.Roughness = 0.9;
	}
	//MaterialParameters.Specular = 0.2;
	half3 auxTangentNormal = half3(0,0,1);
	if(@_AuxiliaryMapEnable)
	{
		tc = BaseMapUV(IN.TexCoord) * cOverlayMapUVTransform.xy + cOverlayMapUVTransform.zw;
		half4 AuxiliaryTex = tAuxiliaryMap.SampleBias(sAuxiliaryMapSampler, tc, -0.5);
		auxTangentNormal = AuxiliaryTex.xyz;
		auxTangentNormal.xyz = auxTangentNormal.xyz * 2 - 1;
	}
	if (@_LOD_LEVEL <= LOD_Level_2 && @_NormalMapEnable || @_SurfaceTextureArrayNormalEnable || @_VTEnable )
	{
		norm = 2 * norm - 1;
		if (ENABLE_RIPPLE)
			Ripple(IN, MaterialParameters, norm, height);
		norm.xy += auxTangentNormal.xy;
		if (@_D3D11 || @_D3D9)
			norm.z = saturate(norm.z);
		MaterialParameters.WorldNormal = norm.z * IN.WorldNormal.xyz + norm.x * IN.WorldTangent.xyz + norm.y * IN.WorldBinormal.xyz;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	else
	{
		MaterialParameters.WorldNormal = IN.WorldTangent.xyz * auxTangentNormal.x + IN.WorldBinormal.xyz * auxTangentNormal.y + IN.WorldNormal.xyz * 1;
		MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal);
	}
	// Apply Burnt
	InitBurntSimple(IN, MaterialParameters, 0.35);
	//Apply Weather
	InitRain(IN, MaterialParameters);
	InitRadarEffect(IN, MaterialParameters);
	
	
}
#include "PBR_PS.fxh"

#if SD_ENABLE_VXGI
#include "VXGI/VXGIVoxelize.fxh"
#endif
