//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzlixianying@corp.netease.com
//	Author: sinelee 
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Terrain.fx>
///	@path	~/System/Shader/
///	@date	2015/03/26
///	@brief	Terrain material.

#pragma shader material
#define MATERIAL_SHADINGMODEL_DEFAULT_LIT 1

#define USE_LIGHTMAP			1
#define USE_CLIPPING			0
#define USE_WIREFRAME			1
#define USE_XRAY				0
#define USE_HIGHLIGHT			0
#define USE_FROZEN				0
#define USE_BURNT				0
#define USE_EDGELINE			0
#define USE_PLANAR_SHADOW		0
#define USE_VBufferGen			0
#define USE_EdgeLineNormalGen	0
#define USE_RAIN_EFFECT			1
#define USE_SNOW_EFFECT			1
#define NO_SHADOW_CLIPPING
#define USE_TESSELLATION		1
#define IS_TERRAIN 1

#pragma shader predict("@_CLIPPING", "false")
#pragma shader predict("@_TRANSPARENT", "false")
#pragma shader predict("@_VERTEXCOLOR", "false")
#pragma shader predict("@_INSTANCED", "false")
#pragma shader predict("@_DF_SKIN", "0")

///////////////choice define end//////////////////////////
#include "PBR_Definitions.fxh"
#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_EnableReflection 
{
	Name = EnableReflection ;
	Description = "Enable reflection?";
	Default = false;
	Visible = false;
}
@_EnableIBL
{
	Name = EnableIBL;
	Description = "Enable image based lighting?";
	Default = true;
	Visible = true;
}
@_EnableCliff
{
	Name = EnableCliff;
	Description = "Enable cliff?";
	Default = true;
	Visible = true;
}
@_LightMapEnable
{
	Name = LightMapEnable;
	Description = "Do we have light map?";
	AffectedTex = tLightMap;
	Fallback = "@_LIGHT_MAP";
	Visible = false;
}
@_BlockMapEnable
{
	Name = BlockMapEnable;
	Description = "Do we have block map?";
	AffectedTex = tBlockMap;
	Visible = false;
}
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have normal map?";
	AffectedTex = tNormalMap;
	Fallback = true;
	Visible = false;
}
@_MixtureMapEnable
{
	Name = MixtureMapEnable;
	Description = "Do we have mixture map?";
	AffectedTex = tMixtureMap;
	Fallback = true;
	Visible = false;
}
@_OverlayMapEnable
{
	Name = OverlayMapEnable;
	Description = "Do we have overlay map?";
	AffectedTex = tOverlayMap;
	Visible = false;
}
@_HeightmapEnable
{
	Name = HeightmapEnable;
	Description = "Apply heightmap to vs?";
	Predict = false;
	Fallback = false;
	Default = false;
	Visible = false;
};

cbuffer Shader
{
	float cAliasingFactor			 = 0;   //抗锯齿因子
	float cEnvStrength				 = 1;   //环境强度[0:8]
	float cRoughnessLow				 = 0;   //粗糙度下界
	float cRoughnessHigh			 = 1;   //粗糙度上界

	float cBloomScale				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cMetallic					 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.5;
	float cSmoothness				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;
	float cBlendSharpness			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0;

	float4 cBlockScale				 = { 1, 0, 0, 0 };
	float4 cBlockShrinkage			 = { 0, 0, 1, 1 };

	float4 cLightMapScale			 = { 1, 1, 1, 1 };
	float4 cLightMapUVTransform		 = { 1, 1, 0, 0 };
	float4 cOverlayMapUVTransform	 = { 1, 1, 0, 0 };
};

texture2D tIndexMap : layout(t0);
sampler sIndexMapSampler : layout(s0)
{
	Texture = tIndexMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlendMap : layout(t1);
sampler sBlendMapSampler : layout(s1)
{
	Texture = tBlendMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlockMap : layout(t2);
sampler sBlockMapSampler : layout(s2)
{
	Texture = tBlockMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tMixtureMap : layout(t4) = "Reflection";
sampler sMixtureMapSampler : layout(s4)
{
	Texture = tMixtureMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tNormalMap : layout(t5) = "NormalMap";
sampler sNormalMapSampler : layout(s5)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D tLightMap : layout(t6);
sampler sLightMapSampler : layout(s6)
{
	Texture = tLightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tEnvMap : EnvMap : layout(t7);
sampler sEnvSampler : layout(s7)
{
	Texture = tEnvMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tShadowMap : ShadowMap : layout(t8);
SamplerComparisonState sShadowMapSampler : layout(s8)
{
	Texture = tShadowMap;
	Filter = Linear;
	Address = Clamp;
	ComparisonFunc = Greater;
	SRGB = false;
	MaxMipLevel = 0;
};
texture2D tOverlayMap : layout(t9);
sampler sOverlayMapSampler : layout(s9)
{
	Texture = tOverlayMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};
texture2D tRainMap : RainMap : layout(t10) = "NormalMap";
sampler sRainSampler : layout(s10)
{
	Texture = tRainMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
//Vertex Texture Fetch
texture2D tHeightMap : HeightMap : layout(t3) = "White";
sampler sHeightSampler : layout(s3)
{
	Texture = tHeightMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
void VS_DIY_OUTPUT(in VertexAssemble IN, in InterMediateVertex INTM, inout VertexShadingOutput OUT)
{
	OUT.ViewDirection.w = 0;

	//world * VP leads to jitter issues(>2000), we use local * WVP instead.
	OUT.HPosition = MulMat4x4(WorldViewProj, float4(INTM.LocalPosition.xyz, 1));
	OUT.ViewDirection.w = OUT.WorldPosition.y;//INTM.LocalPosition.w;

	if (@_LOD_LEVEL <= LOD_Level_2)
	{
		if (@_EnableCliff)
		{
			half upward = step(OUT.WorldNormal.x, OUT.WorldNormal.y) * step(OUT.WorldNormal.z, OUT.WorldNormal.y);
			OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, half3(0, 1 - upward, upward)));
		}
		else
		{
			OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, half3(0, 0, 1)));
		}
		OUT.WorldBinormal.xyz = cross(OUT.WorldNormal.xyz, OUT.WorldTangent.xyz);
	}
}

//////////////////////////////////HS////////////////////////////////////////
#define cNearTessDis 10.0
#define cFarTessDis	 100.0
#define cNearTriPerMeter 5.0
#define cFarTriPerMeter	 0.25
float4 HS_GetTessellationFactor(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2)
{
	float dis0 = length(p0.ViewDirection.xyz);
	float dis1 = length(p1.ViewDirection.xyz);
	float dis2 = length(p2.ViewDirection.xyz);
	float dis = min(min(dis0, dis1), dis2);
	float triPerMeter = lerp(cNearTriPerMeter, cFarTriPerMeter, saturate((dis-cNearTessDis)/(cFarTessDis-cNearTessDis)));
	float len0 = length(p1.WorldPosition.xyz-p2.WorldPosition.xyz);
	float len1 = length(p2.WorldPosition.xyz-p0.WorldPosition.xyz);
	float len2 = length(p0.WorldPosition.xyz-p1.WorldPosition.xyz);
	return float4(len0*triPerMeter, len1*triPerMeter, len2*triPerMeter, (len0+len1+len2)*(0.333*triPerMeter));
}

float GetYPos(float2 xzPos, in VertexShadingOutput OUT)
{
	float2 uv = (xzPos - cHeightmapMinPos.xz) / (cHeightmapMaxPos.xz - cHeightmapMinPos.xz);
	if (!@_GLSL) {uv.y = 1 - uv.y;}
	float currDepth = DecodeDepth(tHeightMap.SampleLevel(sHeightSampler, uv, 0));
	float2 inRange = abs(uv - 0.5);
	float RATE = smoothstep(0.475, 0.5, max(inRange.x, inRange.y));
	float currY = lerp(cHeightmapMinPos.y, cHeightmapMaxPos.y, currDepth);
	return lerp(clamp(currY, OUT.ViewDirection.w, OUT.ViewDirection.w+cHeightmapMinPos.w), OUT.ViewDirection.w+cHeightmapMinPos.w, RATE);
}
//////////////////////////////////DS////////////////////////////////////////
void DS_DIY_OUTPUT(in VertexShadingOutput p0, in VertexShadingOutput p1, in VertexShadingOutput p2, float3 domain, inout VertexShadingOutput OUT)
{
	OUT.WorldPosition.y = GetYPos(OUT.WorldPosition.xz, OUT);
	
	// float texSizeInWorld = cHeightmapMaxPos.w;
	// float4 h;
	// h[0] = GetYPos(OUT.WorldPosition.xz + float2(-texSizeInWorld, 0), OUT);
	// h[1] = GetYPos(OUT.WorldPosition.xz + float2(+texSizeInWorld, 0), OUT);
	// h[2] = GetYPos(OUT.WorldPosition.xz + float2(0, -texSizeInWorld), OUT);
	// h[3] = GetYPos(OUT.WorldPosition.xz + float2(0, +texSizeInWorld), OUT);
	// float3 N;
	// N.x = (h[0] - h[1]) / texSizeInWorld;
	// N.z = (h[2] - h[3]) / texSizeInWorld;
	// N.y = 2;
	// OUT.WorldNormal.xyz = normalize(N);
	// float upward = step(OUT.WorldNormal.x, OUT.WorldNormal.y) * step(OUT.WorldNormal.z, OUT.WorldNormal.y);
	// OUT.WorldTangent.xyz = normalize(cross(OUT.WorldNormal.xyz, float3(0, 1 - upward, upward)));
	// OUT.WorldBinormal.xyz = cross(OUT.WorldNormal.xyz, OUT.WorldTangent.xyz);
}

//////////////////////////////////PS////////////////////////////////////////
void PS_GetParameters(in VertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	InitShadingParameters(MaterialParameters);
	half2 index = tIndexMap.Sample(sIndexMapSampler, BaseMapUV(IN.TexCoord)).xy;
	half2 blend = tBlendMap.Sample(sBlendMapSampler, BaseMapUV(IN.TexCoord)).zz;
	blend[1] = 1 - blend[0];
	half2 y = floor(index * 16);
	half2 x = floor(index * 256) - 16 * y;
	half4 c = half4(x[0], y[0], x[1], y[1]);
	c = c * cBlockShrinkage.z;
	half2 tc = half2(0, 0);
	float2 dx = 0.0;
	float2 dy = 0.0;
	if (@_EnableCliff)
	{
		/* ensures that exactly 1 bit of w equals 1, w = ((r.x>r.y)*(r.x>r.z),(r.y>=r.z)*(r.y>=r.x),(r.z>=r.x)*(r.z>r.y)). */
		float3 r = abs(IN.WorldNormal.xyz);
		float3 s = step(r.xzx, r.yyz); s.x = 1.0 - s.x;
		float3 w = s * (1.0 - s.zxy);
		/* make sure all faces look the same but not as in a mirror. */
		float3 t = cBlockScale.x * IN.WorldPosition.xyz;
		half dir = sign(dot(IN.WorldNormal.xyz, half3(w.x, w.y, -w.z)));
		half3 wc = cBlockShrinkage.y * frac(t * half3(dir, 1, dir));
		tc = half2(dot(wc.zxx, w), dot(wc.yzy, w)) + cBlockShrinkage.x;
		float3 dcx = ddx(t);
		float3 dcy = ddy(t);
		dx = cBlockShrinkage.y * min(abs(float2(dot(dcx.zxx, w), dot(dcx.yzy, w))), 0.125);
		dy = cBlockShrinkage.y * min(abs(float2(dot(dcy.zxx, w), dot(dcy.yzy, w))), 0.125);
	}
	else
	{
		float2 t = cBlockScale.x * IN.WorldPosition.xz;
		tc = cBlockShrinkage.y * frac(t) + cBlockShrinkage.x;
		dx = cBlockShrinkage.y * min(abs(ddx(t)), 0.125);
		dy = cBlockShrinkage.y * min(abs(ddy(t)), 0.125);
	}
	float dl = max(dx.x + dy.x, dx.y + dy.y);
	tc.x = clamp(tc.x, dl, cBlockShrinkage.z - dl);
	tc.y = clamp(tc.y, dl, cBlockShrinkage.z - dl);
	half2 tc0 = tc + c.xy;
	half2 tc1 = tc + c.zw;
	half3 diff0 = half3(0, 0, 0);
	half3 diff1 = half3(0, 0, 0);
	half3 mixx0 = half3(0, 0, 0);
	half3 mixx1 = half3(0, 0, 0);
	if (@_BlockMapEnable)
	{
		diff0 = tBlockMap.SampleGrad(sBlockMapSampler, tc0, dx, dy);
		diff1 = tBlockMap.SampleGrad(sBlockMapSampler, tc1, dx, dy);
	}
	if (@_LOD_LEVEL <= LOD_Level_2 && @_MixtureMapEnable && @_EnableReflection)
	{
		mixx0 = tMixtureMap.SampleGrad(sMixtureMapSampler, tc0, dx, dy);
		mixx1 = tMixtureMap.SampleGrad(sMixtureMapSampler, tc1, dx, dy);
		half sharpblend = step(blend[1] + mixx1.b, blend[0] + mixx0.b);
		blend[0] = lerp(blend[0], sharpblend, cBlendSharpness);
		blend[1] = 1 - blend[0];
	}
	half3 diff = blend[0] * diff0 + blend[1] * diff1;
	half3 mixx = blend[0] * mixx0 + blend[1] * mixx1;
	if (@_OverlayMapEnable)
	{
		tc = BaseMapUV(IN.TexCoord) * cOverlayMapUVTransform.xy + cOverlayMapUVTransform.zw;
		half3 overlay = 2 * tOverlayMap.Sample(sOverlayMapSampler, tc);
		diff = diff * overlay - step(1.0, overlay) * (1 - 2 * diff) * (1 - overlay);
	}
	MaterialParameters.BaseColor = DecodeGamma(diff);
	float rate = 0;
	if (@_LOD_LEVEL <= LOD_Level_2 && @_MixtureMapEnable && @_EnableReflection || @_LOD_LEVEL <= LOD_Level_1 && @_NormalMapEnable)
	{
		float dist = length(CameraPos - IN.WorldPosition.xyz);
		rate = saturate((1 - dist / cParamter.w) / 0.2);
	}
	if (@_LOD_LEVEL <= LOD_Level_2 && @_MixtureMapEnable && @_EnableReflection)
	{
		MaterialParameters.Metallic = lerp(cMetallic, mixx.g, rate);
		MaterialParameters.Roughness = 1 - lerp(cSmoothness, mixx.r, rate);
	}
	else
	{
		MaterialParameters.Metallic = cMetallic;
		if (@_EnableReflection || @_DEFERRED_SHADING)
		{
			MaterialParameters.Roughness = 1 - cSmoothness;
		}
	}
	float snowRate = (IN.WorldPosition.y - IN.ViewDirection.w) / cHeightmapMinPos.w;
	if (@_LOD_LEVEL <= LOD_Level_2 && @_NormalMapEnable)
	{
		float texSizeInWorld = cHeightmapMaxPos.w;
		float4 h;
		h[0] = GetYPos(IN.WorldPosition.xz + float2(-texSizeInWorld, 0), IN);
		h[1] = GetYPos(IN.WorldPosition.xz + float2(+texSizeInWorld, 0), IN);
		h[2] = GetYPos(IN.WorldPosition.xz + float2(0, -texSizeInWorld), IN);
		h[3] = GetYPos(IN.WorldPosition.xz + float2(0, +texSizeInWorld), IN);
		MaterialParameters.WorldNormal.x = (h[0] - h[1]) / texSizeInWorld;
		MaterialParameters.WorldNormal.z = (h[2] - h[3]) / texSizeInWorld;
		MaterialParameters.WorldNormal.y = 2;

		float3 detailNormal = tNormalMap.Sample(sNormalMapSampler, IN.WorldPosition.xz*cAliasingFactor).xyz * 2 - 1;
		detailNormal.xy *= 0.5; 
		float3 DN = detailNormal.z * IN.WorldNormal.xyz + detailNormal.x * IN.WorldTangent + detailNormal.y * IN.WorldBinormal;
		// MaterialParameters.WorldNormal = normalize(MaterialParameters.WorldNormal + DN*1.5);
		MaterialParameters.WorldNormal = normalize(DN);
	}
	else
	{
		MaterialParameters.WorldNormal = normalize(IN.WorldNormal.xyz);
	}
	
	//snow overlay
	//half overlayFactor = saturate(MaterialParameters.WorldNormal.y * (snowRate+0.2));
	half overlayFactor = clamp(snowRate, 0.3, 0.8);
	MaterialParameters.BaseColor = lerp(MaterialParameters.BaseColor, half3(1,1,1), overlayFactor);
	MaterialParameters.Alpha = 1.0;
}

#include "PBR_PS.fxh"
