///	@file	<UIBasic.fx>
///	@path	~/Shader/UI/

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 	;
	float4 ModulateColor			 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling		 = { 1, 1, 0, 0 };
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
	MaxMipLevel = 0;
};

texture2D MaskTexture 					: layout(t1) = "White";
sampler_state MaskTexture_Sampler		: layout(s1)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Wrap;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	[@_VERTEXCOLOR]
	float3		Color					: COLOR;
	float2		TexCoord				: TEXCOORDN;
	[@_LIGHT_MAP]
	float2		TexCoord2				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float2		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UICustomMaskDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;

	return OUT;
}

float4 UICustomMaskDrawPS(in VertexOutput IN) : SV_Target
{
	float4 OUT = float4(1,1,1,1);
	//uv
	float2 diffuseTC = IN.TexCoord * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	//sample	
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	float Mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord).a;

	OUT *= ModulateColor;
	OUT.a *= Mask;
	return OUT;
}

technique Shading
{
	VertexShader = UICustomMaskDrawVS();
	PixelShader = UICustomMaskDrawPS();
}
