//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIBlurNode.fx>
///	@path	~/Shader/UI/
///	@date	2019/12/03
///	@brief	Blur UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_BGBlur
{
	Name = BGBlur;
	Description = "Is blur?";
	Default = false;
}

cbuffer Batch
{
	float4x4 WorldViewProj		 		: layout(vs);
	float4   UVTransform				: layout(vs) = { 1.0f, 0.0f, 0.0f, 1.0f };
};

texture2D SceneTexture					: layout(t0) = "NormalMap";
sampler_state SceneTexture_Sampler		: layout(s0)
{
	Texture = SceneTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UIBlurNodeDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	//OUT.TexCoord.xy = IN.TexCoord;
	//OUT.TexCoord.zw = IN.TexCoord2;
	//OUT.Color = GetVertexColor(IN.Color);

	float2 screenUV = OUT.HPosition.xy / (2.0 * OUT.HPosition.w) + 0.5;
	[!@_GLSL]
		screenUV.y = 1.0 - screenUV.y;

	if (@_VULKAN)
	{
		float2 t = (screenUV - 0.5) * 2.0;
			screenUV.x = dot(t, UVTransform.xy);
			screenUV.y = dot(t, UVTransform.zw);
			screenUV = screenUV * 0.5 + 0.5;
	}
	OUT.TexCoord.xy = screenUV;

		return OUT;
}

float Gaussian(float sigma, float x)
{
	return exp(-(x*x) / (2.0 * sigma*sigma));
}

float3 BlurredPixel(float2 uv)
{
	int   c_samplesX = 15;  // must be odd
	int   c_samplesY = 18;  // must be odd
	float c_textureSize = 200.0;

	int   c_halfSamplesX = c_samplesX / 2;
	int   c_halfSamplesY = c_samplesY / 2;
	float c_pixelSize = (1.0 / c_textureSize);

	float total = 0.0;
	float3 ret = float3(0.0, 0.0, 0.0);

	//float4 texColor = texSampler.Sample(tex, uv );

	[loop]for (int iy = 0; iy < c_samplesY; ++iy)
	{
		float fy = Gaussian(5.0, float(iy) - float(c_halfSamplesY));
		float offsety = float(iy - c_halfSamplesY) * c_pixelSize;

		[loop]for (int ix = 0; ix < c_samplesX; ++ix)
		{

			float fx = Gaussian(5.0, float(ix) - float(c_halfSamplesX));
			float offsetx = float(ix - c_halfSamplesX) * c_pixelSize;
			total += fx * fy;
			float2 finalUV = uv + float2(offsetx, offsety);

			float4 texColor = SampleColorMap(SceneTexture, SceneTexture_Sampler, finalUV);  // tex.Sample(texSampler, finalUV ); 

			ret += texColor * fx*fy;

		}
	}

	return ret / total;

	//return SampleColorMap(SceneTexture, SceneTexture_Sampler,uv);
}

float4 UIBlurNodeDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = float4(0, 0, 0, 1);
	if (@_BGBlur)
	{

		float2 stc = IN.TexCoord.xy;
		[@_GLSL]
		stc.y = 1 - stc.y;

		half2 offset = 1 / 512.0h;//ScreenInfo.zw;
		//stc += offset * 0.5;

		static const float kernel[5] = {0.2659616,
							0.3223055,
							0.0435913,
							0.0011174,
							0.0000050};
		static const float newoffset[5] = {0,
							1.9392436678865236,
							3.774284777008256,
							5.079828172543404,
							7};


		//diffuse.rgb = SampleColorMap(SceneTexture, SceneTexture_Sampler, stc).rgb * kernel[0];


		diffuse.rgb = BlurredPixel(stc);


	}

	return diffuse;
}


technique Shading
{
	VertexShader = UIBlurNodeDrawVS();
	PixelShader = UIBlurNodeDrawPS();
}
