//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIBlurNode.fx>
///	@path	~/Shader/UI/
///	@date	2019/12/03
///	@brief	Blur UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

@_BGBlur
{
	Name = BGBlur;
	Description = "Is blur?";
	Default = false;
}

@_ScreenCoord
{
	Name = ScreenCoord;
	Description = "Is Use Screen Coord?";
	Default = true;
}

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float  BlurSize = 1					;
};

texture2D SceneTexture					: layout(t0) = "Transparent";
sampler_state SceneTexture_Sampler		: layout(s0)
{
	Texture = SceneTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture 					: layout(t1) = "Transparent";
sampler_state MaskTexture_Sampler		: layout(s1)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Wrap;
	MaxMipLevel = 0;
};


struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UIBlurNodeDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.zw = IN.TexCoord2;
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UIBlurNodeDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = float4(0, 0, 0, 1);
	if (@_BGBlur)
	{
		float2 stc = IN.TexCoord;
		if (@_ScreenCoord)
		{
			float2 spos = IN.HPosition.xy;
				stc = spos * ScreenInfo.zw;
		}
		float2 scaleSize = float2(1.0 / BlurSize, 1.0 / BlurSize);
		

		diffuse.a = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.xy).a;
		
		diffuse.rgb = SampleColorMap(SceneTexture, SceneTexture_Sampler, stc).rgb * 0.1;
		diffuse.rgb += (SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(5.629332, 0) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(-5.629332, 0) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(0, 5.629332) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(0, -5.629332) * ScreenInfo.zw * scaleSize).rgb) * 0.08;

		diffuse.rgb += (SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(6.928204, 6.928204) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(-6.928204, 6.928204) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(6.928204, -6.928204) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(-6.928204, -6.928204) * ScreenInfo.zw * scaleSize).rgb) * 0.065;

		diffuse.rgb += (
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(13.17686, 0) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(-13.17686, 0) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(0, 13.17686) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(0, -13.17686) * ScreenInfo.zw * scaleSize).rgb) * 0.05;

		diffuse.rgb += (
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(14.422204, 14.422204) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(-14.422204, 14.422204) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(14.422204, -14.422204) * ScreenInfo.zw * scaleSize).rgb +
			SampleColorMap(SceneTexture, SceneTexture_Sampler, stc + float2(-14.422204, -14.422204) * ScreenInfo.zw * scaleSize).rgb) * 0.03;
	}
	return diffuse;
}

technique Shading
{
	VertexShader = UIBlurNodeDrawVS();
	PixelShader = UIBlurNodeDrawPS();
}
