//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UICircularImage.fx>
///	@path	~/Shader/UI/
///	@date	2015/01/16
///	@brief	Translucent circular UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
};

texture2D MaskTexture					: layout(t0) = "Transparent";
sampler_state MaskTexture_Sampler	: layout(s0)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UICustomMaskDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.zw = IN.TexCoord2;
	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UICustomMaskDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse = float4(0, 0, 0, 0);
	diffuse.a = 1.0 - SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.xy).a;
	return diffuse;
}

technique Shading
{
	VertexShader = UICustomMaskDrawVS();
	PixelShader = UICustomMaskDrawPS();
}
