﻿///	@file	<UIBasic.fx>
///	@path	~/Shader/UI/

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 	;
	float  BlurSize = 0.6;
	float  DrawSize = 0.2;
	float  Refresh = 1;
	float4 Mouse = { 0, 0, 0, 0 };
	float2  Resolution = { 1920, 1080 };
	float  actionTs = 0;

};

texture2D	 MainTex				: layout(t0) = "Black"; //暂用
sampler_state MainTex_Sampler		: layout(s0)
{
	Texture = MainTex;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};


struct VertexAssemble
{
	float4		Position 				: POSITION;
	[@_VERTEXCOLOR]
	float3		Color					: COLOR;
	float2		TexCoord				: TEXCOORDN;
	[@_LIGHT_MAP]
	float2		TexCoord2				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float2		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.



float4 draw(float2 uv, float4 iMouse, float2 iResolution, float4 old, float drawSize, float4 maintexture)
{
	float4 mouseT = iMouse;

//	if (iMouse.z < 1)
//	{
//		mouseT.w =2- (GameTime  -actionTs)* 0.001;  
//	}
	float brushOpacity = clamp(mouseT.w*6 , 0, 2);

	float brushSize = (drawSize*brushOpacity*length(iResolution.xy)/50000);//*length(iResolution.xy) 

	float2 currDelta = iMouse.xy - uv;
	
	float linee;
	float brush;
	
	
	if (mouseT.w >= 0) {
	//	float x = step(0.1, iMouse.w);	
	//	old.xy = lerp(old.xy, iMouse.xy, iMouse.z);
		if (0.1 > iMouse.w && iMouse.w > 0)
		{
			old.xy = lerp(old.xy, iMouse.xy, iMouse.z);
		}
		
//		if (iMouse.z < 0)
//		{
//			float2 distance = mouseT.xy - old.xy;
//			old.xy = old.xy + distance * (GameTime - actionTs)*0.06;
//			mouseT.xy = mouseT.xy + distance * ((GameTime - actionTs + 1)*0.06);
//		}

		float2 oldDelta = uv - old.xy;
		float2 newDelta = mouseT.xy - old.xy;

		
	
	//	float2 lon = old.xy - iMouse.xy;
		
		float linlon = min(1000, pow(newDelta.x * newDelta.x + newDelta.y * newDelta.y, 0.5));
		
		float perlon =pow(( 1 - linlon / 1.5),8);
	

		linee = length(oldDelta - newDelta * clamp(dot(oldDelta, newDelta) / dot(newDelta, newDelta), 0, 1));// -brushSize * perlon;
		
		
		brush = brushOpacity * clamp(perlon - linee / brushSize, 0, 1) * perlon;
	    

	}


	float bout = smoothstep(0, 0.8, brush);
	float w = clamp(iMouse.w , 0, 8);
	maintexture.xyz = max(maintexture.x, bout*(1 - smoothstep(0,1.26,w*0.11)));
//	maintexture.y = max(maintexture.y, brush);
//	maintexture.xz = max(maintexture.xz, bout);

//	float4 fragColor = maintexture;  // float4(old.xyz,1);

	



/*test
//	float4 lee = 0;
	if (iMouse.z >=1)
	{
		float str = smoothstep(-0.5, 1, length(iMouse.xy - old.xy ));
		maintexture += str * 0.0009 / (pow(length(uv - iMouse.xy), 1.7) + 0.002);
		

//		float le = length(uv - iMouse.xy);
//		if (le < 0.01)
//		{
//			lee = 1;
//		}
		
	}
//	maintexture = maintexture + lee;
*/

	return maintexture;
}

float hash(float n)
{
	return frac(sin(dot(float2(n, n), float2(12.9898, 78.233))) * 43758.5453);
}

float2 turbulence(float2 uv, float2 speed)// , float iTime)
{
	float2 turb;
	turb.x = sin(uv.x);
	turb.y = cos(uv.y);
	for (int i = 0; i < 10; i++)
	{
		float ifloat = 1.0 + float(i);
		float ifloat1 = ifloat + 1;
		float ifloat2 = ifloat + 100;

		float r1 = hash(ifloat1)*2.0 - 1.0;
		float r2 = hash(ifloat2)*2.0 - 1.0;

		float2 turb2;
		turb2.x = sin(uv.x*(1.0 + r1 * 30) + turb.y*0.3*ifloat);// +iTime * speed.x*r2);
		turb2.y = cos(uv.y*(1.0 + r1 * 30) + turb.x*0.3*ifloat);// +iTime * speed.y*r2);

		turb.x = lerp(turb.x, turb2.x, 0.5);
		turb.y = lerp(turb.y, turb2.y, 0.5);
	}
	return turb;
}

float4 saveMouse(float4 Tex, float4 iMouse, float2 uv, float2 iResolution)
{

//	float4 fragColor = Tex;
	if	(uv.y <= 4/iResolution.y && uv.x <= 4 / iResolution.x) {
		Tex.xyz =  iMouse.xyz;
	}
	return Tex;

}

VertexOutput UIDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;

	return OUT;
}

float4 UIDrawPS(in VertexOutput IN) : SV_Target
{
	float4 mouse = Mouse;
    mouse.y = 1 - mouse.y;
//	float2 fragCoord = IN.HPosition +0.5;//test
	float2 uv = IN.TexCoord;

	float4 maintexture = MainTex.Sample( MainTex_Sampler, uv);
	maintexture.w = 1;
//	maintexture.xyz = uv.x;
	float2 uvtest = (0.1,0.1);
	float4 old = MainTex.Sample(MainTex_Sampler, float2(0.5/ Resolution.x, 0.5 / Resolution.y));


	maintexture = draw(uv, mouse, Resolution, old, DrawSize, maintexture);

	float2 offsetUV = MainTex.Sample(MainTex_Sampler, uv).rg;
	float offsetUVRT = MainTex.Sample(MainTex_Sampler, uv + float2(0.002, 0.001)).g;
	float offsetUVLT = MainTex.Sample(MainTex_Sampler, uv + float2(-0.002, 0.001)).g;
	float offsetUVRD = MainTex.Sample(MainTex_Sampler, uv + float2(0.002, -0.001)).g;
	float offsetUVLD = MainTex.Sample(MainTex_Sampler, uv + float2(-0.002, -0.001)).g;

	offsetUV.r = offsetUV.r + offsetUVRT - offsetUVLD;
	offsetUV.r = offsetUV.r + offsetUVLT - offsetUVRD;
	float2 speed = 20;
	float2 uvTiling = uv * float2(2, 1);

	float2 turUV = (turbulence(uvTiling, speed)*0.5 + 0.5)*0.01 - 0.005; //test
	turUV = turUV + uv + offsetUV * 0.01;

	float4 turColor = 1;//test
	turColor.r = maintexture.g* 0.568;
	turColor.g = maintexture.g* BlurSize;
	turColor.b = MainTex.Sample(MainTex_Sampler, turUV).b* BlurSize;
	maintexture = max(turColor, maintexture);

	maintexture = maintexture * Refresh;

	maintexture = saveMouse(maintexture, mouse, uv, Resolution);
//	float4 fragColor = maintexture;

//	if (uv.y >= 0.8 && uv.x >= 0.8) {
//		maintexture = old;
//	}

	return maintexture;
}

technique Shading
{             
	VertexShader = UIDrawVS();
	PixelShader = UIDrawPS();
}
