﻿///	@file	<UIBasic.fx>
///	@path	~/Shader/UI/

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 	;

//	float4 Mouse = { 0, 0, 0, 0 };
	float2 Tiling = { 0.1, 0.2};
	float  Time = 0;
	float Gray =0.7;

};

texture2D	 MainTex				: layout(t0) = "Black"; //暂用
sampler_state MainTex_Sampler		: layout(s0)
{
	Texture = MainTex;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D	 NoiseTex				: layout(t1) = "Black"; //暂用
sampler_state NoiseTex_Sampler		: layout(s1)
{
	Texture = NoiseTex;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};


struct VertexAssemble
{
	float4		Position 				: POSITION;
	[@_VERTEXCOLOR]
	float3		Color					: COLOR;
	float2		TexCoord				: TEXCOORDN;
	[@_LIGHT_MAP]
	float2		TexCoord2				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float2		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.




float hash(float n)
{
	return frac(sin(dot(float2(n, n), float2(12.9898, 78.233))) * 43758.5453);
}

float2 turbulence(float2 uv, float2 speed, float iTime)
{
	float2 turb;
	turb.x = sin(uv.x);
	turb.y = cos(uv.y);
	for (int i = 0; i < 10; i++)
	{
		float ifloat = 1.0 + float(i);
		float ifloat1 = ifloat + 1;
		float ifloat2 = ifloat + 100;

		float r1 = hash(ifloat1)*2.0 - 1.0;
		float r2 = hash(ifloat2)*2.0 - 1.0;

		float2 turb2;
		turb2.x = sin(uv.x*(1.0 + r1 * 30) + turb.y*0.3*ifloat);// +iTime * speed.x*r2);
		turb2.y = cos(uv.y*(1.0 + r1 * 30) + turb.x*0.3*ifloat);// +iTime * speed.y*r2);

		turb.x = lerp(turb.x, turb2.x, 0.5);
		turb.y = lerp(turb.y, turb2.y, 0.5);
	}
	return turb;
}



VertexOutput UIDrawOutputVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;

	return OUT;
}

float4 UIDrawOutputPS(in VertexOutput IN) : SV_Target
{
	
	float2 uv = IN.TexCoord;

	float3 noiseTex = NoiseTex.Sample( NoiseTex_Sampler, uv).xyz;
	float2 noiseUV = (uv + ((noiseTex.x + noiseTex.y + noiseTex.z)*0.2 - 0.3))*Tiling;
	
	float speed = 20;
	float noise = clamp(turbulence(noiseUV, speed, Time).x + 0.5, 0, 1);

	float3 mainTex = MainTex.Sample(MainTex_Sampler, uv).xyz;

	float4 fragColor = 1;


	fragColor.x = clamp(((mainTex.x + mainTex.y*noise) / 2 + mainTex.z * 4) / 5, 0, 1);
	fragColor.w = clamp(fragColor.x * 5, 0, 1); 
	fragColor.xyz = smoothstep(0,1.26,fragColor.x);
	return fragColor;
}

technique Shading
{             
	VertexShader = UIDrawOutputVS();
	PixelShader = UIDrawOutputPS();
}
