﻿///	@file	<UIBasic.fx>
///	@path	~/Shader/UI/

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 	;
	float4 ModulateColor			 = { 1, 1, 1, 1 };
	float4 DiffuseUV				 = { 1, 1, 0, 0 };
	float4 MaskUV					 = { 1, 1, 0, 0 };
	float  ExtraAlpha				 = 0;
	float  DirRange					 = 0;
	float  LinkRange				 = 0;
	float  DirEdgeSoft				 = 2;

};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskDirectionTexture 					: layout(t1) = "Black";
sampler_state MaskDirectionTexture_Sampler		: layout(s1)
{
	Texture = MaskDirectionTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};


texture2D MaskLinkTexture 					: layout(t2) = "Black";
sampler_state MaskLinkTexture_Sampler		: layout(s2)
{
	Texture = MaskLinkTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	[@_VERTEXCOLOR]
	float3		Color					: COLOR;
	float2		TexCoord				: TEXCOORDN;
	[@_LIGHT_MAP]
	float2		TexCoord2				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float2		TexCoord				: TEXCOORD0;
};

////////////////////////////////////////////////////////////////////////
// Translucent  UI draw.

VertexOutput UIFontDissipatedDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
		OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.TexCoord = IN.TexCoord;

	return OUT;
}

float4 UIFontDissipatedDrawPS(in VertexOutput IN) : SV_Target
{
	
	//uv
	float2 diffuseUV = IN.TexCoord * DiffuseUV.xy + DiffuseUV.zw;
	float2 maskUV = IN.TexCoord * MaskUV.xy + MaskUV.zw;

	//sample	
	float4 diffuse = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseUV);
	float linkMask = clamp(SampleColorMap(MaskLinkTexture, MaskLinkTexture_Sampler, maskUV).r / 2.2, 0, 1);
	float dirMask = clamp(SampleColorMap(MaskDirectionTexture, MaskDirectionTexture_Sampler, maskUV).r / 2.2, 0, 1);

	dirMask = lerp(dirMask, 0, DirRange);
	linkMask = lerp(linkMask, 0, LinkRange);

	dirMask = pow(dirMask, DirEdgeSoft);
	
	float alpha = clamp((linkMask + dirMask), 0, 1);
	float4 OUT = diffuse;
	alpha = diffuse.a * alpha + ExtraAlpha;

	OUT *= ModulateColor;
	OUT.a *= alpha;
	return OUT;
}

technique Shading
{             
	VertexShader = UIFontDissipatedDrawVS();
	PixelShader = UIFontDissipatedDrawPS();
}
