//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : szn3878@corp.netease.com
//	Author: Si Zheng
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Unlit_BasicCloud.fx>
///	@path	Shader/
///	@date	2017/05/15
///	@brief	Basic Shader support UV tiling and UV scroll
//////////////////////////////////Properties////////////////////////////////////////

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

@_VSCloudMapEnable 
{ 
	Name = CloudMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tRayMap;
	Fallback = true;
	Visible = true;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 			;
	float4 cParamter2				: Param2			;
};

cbuffer Shader
{
	float3 cCloudColor				 < Widget = "LinearColor3"; > = 1;
	float cEnvAffector				 < Widget = "FloatSlider"; Min = "0"; Max = "1.5"; Step = "0.01"; Precision = "3"; > = 1;
	float cBrightness				 < Widget = "FloatSlider"; Min = "0"; Max = "3"; Step = "0.01"; Precision = "3"; > = 1;
	float cTransparentShift			 < Widget = "FloatSlider"; Min = "-1"; Max = "1"; Step = "0.01"; Precision = "3"; > = 0;
	float UVscaleU 					 < Widget = "FloatSlider"; Min = "0.1"; Max = "20"; Step = "0.01"; Precision = "2"; > = 1;
	float UVscaleV 					 < Widget = "FloatSlider"; Min = "0.1"; Max = "20"; Step = "0.01"; Precision = "2"; > = 1;
	float cSpdU						 < Widget = "FloatSlider"; Min = "-10"; Max = "10"; Step = "0.001"; Precision = "3"; > = 0;
	float cSpdV						 < Widget = "FloatSlider"; Min = "-10"; Max = "10"; Step = "0.001"; Precision = "3"; > = 0;
	float cMixColor					 < Widget = "FloatSlider"; Min = "-0.5"; Max = "0.5"; Step = "0.001"; Precision = "3"; > = 0;
	float3 cSecColor				 < Widget = "LinearColor3"; > = 1;
	float cDepthMovement			 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.001"; Precision = "3"; > = 1;
};

texture2D tRayMap : layout(t0) = "White";
sampler sRaySampler : layout(s0)
{
	Texture = tRayMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		TexCoord		: TEXCOORDN;
	float4		LocalPos		: TEXCOORDN;
	float4      WorldPosition	: TEXCOORDN;
	half3 		WorldNormal		: NORMAL;
	half3 		WorldBinormal 	: TEXCOORDN;
	float		FogIntensity	: TEXCOORDN;
	[@_VERTEXCOLOR]
	float4		TintColor 		: COLOR;				//	TintColor 
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	float3		Normal			: NORMAL;
	float2		TexCoord		: TEXCOORDN;
	float4		Tangent			: TANGENT;
	float4		Binormal		: BINORMAL;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
	float3 LocalNormal;
	float3 LocalBinormal;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);
	INTM.LocalBinormal = normalize(IN.Binormal.xyz);
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.FogIntensity = FogTransmittance(OUT.WorldPosition.xyz - CameraPos.xyz);
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);
	OUT.TexCoord = IN.TexCoord;
	OUT.LocalPos = IN.Position;
	OUT.WorldNormal = MulMat3x4((float3x3)World, INTM.LocalNormal);
	OUT.WorldBinormal = MulMat3x4((float3x3)World, INTM.LocalBinormal);
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput BasicCloudVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float TriangleWave(float x)
{
	return abs((frac(x + 0.5) * 2.0) - 1.0);
}

float4 BasicCloudPixelShading(in VertexShadingOutput IN, in float4 sc: SV_Position) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	half4 samplerBase;
	float2 texSample = IN.TexCoord;
	texSample.x = texSample.x * UVscaleU + GameTime * cSpdU;
	texSample.y = texSample.y * UVscaleV + GameTime * cSpdV;
	OUT = tRayMap.SampleRGBA(sRaySampler, texSample);
	half sAlpha= saturate(OUT.r + cTransparentShift);
	half mixAlpha = 0;
	if(@_LOD_LEVEL <= LOD_Level_1)
	{
		half shifttime =  GameTime * cDepthMovement * 0.1;
		mixAlpha = lerp(OUT.r,OUT.g, saturate((TriangleWave(shifttime)-0.3)*2.5));
	}
	OUT.a = mixAlpha * sAlpha;
	OUT.rgb = lerp(cCloudColor,cSecColor,saturate( IN.LocalPos.x * cMixColor)) * lerp(half3(1,1,1), SunColor.rgb/4,cEnvAffector) * cBrightness;
	if (@_FOG)
		{OUT.rgb = FogLighting(IN.FogIntensity, OUT.rgb);}
	
	
	if(@_VERTEXCOLOR)
	{
		OUT.a *= IN.TintColor.r;
	}
	
	OUT.a = EncodeTransparentBloomFactor(OUT.a,0);
	OUT.rgb *= OUT.a;
	
	return OUT;
}

technique Shading
{
	VertexShader = BasicCloudVertexShading();
	PixelShader = BasicCloudPixelShading();
}

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);

	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = BasicCloudVertexShading();
	PixelShader = BasicCloudPixelShading();
}
