//	--------------------------------------------------------------------
//	Copyright(C) 2006~2016 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Project Messiah
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Lava.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_UNLIT 1

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

@_UseDiyFog
{
	Name = UseDiyFog;
	Description = "use diy fog?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 				: layout(c0);
	float4 cParamter2				: Param2				: layout(c3);
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex	: layout(c4);
};

cbuffer Shader	
{
	float cIntensity			: layout(c0.x) = 2;
	float4 cColor				: layout(c1) = { 1, 1, 1, 1 };


	float4 _padding2				: layout(c2) = { 1, 1, 1, 1 };
};

// Surface textures.
texture2D tBaseMap : layout(t0);
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.

	float		T				: TEXCOORDN;
	float4		WorldPosition	: TEXCOORDN;
	float3		WorldNormal		: TEXCOORDN;				
	//float4 		screenCoord		: SV_Position;									//	FogTransmitance
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal			: NORMAL;
	float4		TexCoord		: TEXCOORDN;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
	float3 LocalNormal;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal = (float3)Expand((half3)IN.Normal.xyz);
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.WorldNormal = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));;
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.WorldPosition = WorldPosition;
	//TexCoord
	OUT.TexCoord.xy = IN.TexCoord.xy;

	//ViewDirection
	OUT.ViewDirection = WorldPosition.xyz - CameraPos.xyz;
	
	//Fog T
	
		OUT.T = FogTransmittance(OUT.ViewDirection.xyz);

	// if (@_VolumetricFog)
	// {
	// 	OUT.T = OUT.HPosition.w;
	// }

	// if (@_NoFarPlaneClip)
	// {
	// 	OUT.HPosition.z = max(EPSILON*OUT.HPosition.w, OUT.HPosition.z);
	// }
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput BillboardVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	//OUTPUT
	
	VertexUnifyOutput(IN, INTM, OUT);
	OUT.TexCoord.z = OUT.HPosition.w;
	
	return OUT;
}

float4 BillboardPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	float lineargradient = IN.TexCoord.y;
	half3 baseColor = half3(1,1,1);

	float uvY =IN.TexCoord.y + GameTime*3;
	uvY =(sin(uvY*55));	
	uvY = ceil((uvY)) ;
	float uvYsecond = IN.TexCoord.x + GameTime*6;
	uvYsecond =(sin(uvYsecond*25));
	uvYsecond = ceil((uvYsecond)) ;
	uvY +=uvYsecond;
	uvY *=0.01;
	uvY+= IN.TexCoord.y;

	float2 GlitchUV = float2(IN.TexCoord.x,uvY);
	float Framecount = frac(sin(GameTime*5));   
	GlitchUV.xy =  lerp(GlitchUV.xy,IN.TexCoord.xy,Framecount);
	// float2 SmallTC = IN.TexCoord.xy*half2(25,90);
	float2 SmallTC = IN.WorldPosition.xy * 0.05;
	float2 SmallTCz = IN.WorldPosition.yz *0.05;
	half normalMask =saturate( pow( abs(normalize(IN.WorldNormal).z),1)*5) ;
	
	baseColor =tBaseMap.SampleRGBA(sBaseSampler, IN.TexCoord) ;
	baseColor*=baseColor;
	baseColor = half3(baseColor.x,0,0);
	half3 OrinColor =tBaseMap.SampleRGBA(sBaseSampler, GlitchUV).rgb ;
	OrinColor*=OrinColor;
	OrinColor = half3(0,OrinColor.y,OrinColor.z);
	baseColor = half3(baseColor.x ,OrinColor.y+OrinColor.z*0.1,OrinColor.z) ;
	

	OUT.rgb = baseColor *cColor *cIntensity ;
	
// OUT =  lerp (tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.WorldPosition.xy),tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.WorldPosition.xy),saturate());
//FADElINE

//////////////////////////

	// Clipping
	if (@_CLIPPING && @_ES20)
		clip(OUT.a - 0.5);

	OUT.a =OUT.r*0.2+OUT.g*0.7+OUT.b*0.1;
	OUT.a =0;

	return OUT;
}

technique Shading
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
	if (@_CLIPPING)
	{
		half a = tBaseMap.SampleRGBA(sBaseSampler, IN.TexCoord.xy).a;
		clip(a - 0.5);
	}
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = BillboardVertexShading();
	PixelShader = ClipPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Wireframe
[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].ViewDirection = float3(1, 0, 0);
	OUT.Append(IN[0]);
	IN[1].ViewDirection = float3(0, 1, 0);
	OUT.Append(IN[1]);
	IN[2].ViewDirection = float3(0, 0, 1);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.ViewDirection);
	float3 distance = IN.ViewDirection / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(cParamter2.xyz, a);
}

technique Wireframe
{
	VertexShader = BillboardVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = WireframePixelShading();
};

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);
	
	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		half a = tBaseMap.SampleRGBA(sBaseSampler, IN.TC.xy).a;
		clip(a - 0.5);
	}

	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

