//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : szn3878@corp.netease.com
//	Author: SiZheng
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Unlit_GodRay.fx>
///	@path	~/System/Shader/
///	@date	2016/05/17
///	@brief	GodRay for G55
/// Removing unnecessary elements

#pragma shader material
#define MATERIAL_SHADINGMODEL_UNLIT 1

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

@_GodRayMapEnable 
{ 
	Name = GodRayMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tRayMap;
	Visible = true;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 				;
	float4 cParamter2				: Param2				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex	;
};

cbuffer Shader
{

	float3 cGodRayColor				 < Widget = "LinearColor3"; > = 1;
	float cRayRipple				 < Widget = "FloatSlider"; Min = "0.01"; Max = "5"; Step = "0.01"; Precision = "3"; > = 2.88;
	float cBloomRate				 < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "3"; > = 0.5;
	float cRayAngle					 < Widget = "FloatSlider"; Min = "0"; Max = "2"; Step = "0.01"; Precision = "3"; > = 1;
	float cRayLength				 < Widget = "FloatSlider"; Min = "0"; Max = "2"; Step = "0.01"; Precision = "3"; > = 1;
	float cRaySpeed					 < Widget = "FloatSlider"; Min = "0.01"; Max = "1"; Step = "0.01"; Precision = "3"; > = 0.01;
	float cNoiseMSPD				 < Widget = "FloatSlider"; Min = "-1";Max = "1"; Step = "0.01"; Precision = "3"; > = 0.1;
	float cRayLight					 < Widget = "FloatSlider"; Min = "0.01"; Max = "1"; Step = "0.01"; Precision = "3"; > = 0.1;	
	float cRayMaxLight				 < Widget = "FloatSlider"; Min = "0.01"; Max = "1"; Step = "0.01"; Precision = "3"; > = 0.8;	
};

texture2D tRayMap : layout(t0);
sampler sRaySampler : layout(s0)
{
	Texture = tRayMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float		BloomZ			: TEXCOORDN;
	float2		TexCoord		: TEXCOORDN;		
	float4		WorldPosition	: TEXCOORDN;
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal			: NORMAL;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	//position
	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);
	OUT.TexCoord = IN.TexCoord;
	OUT.BloomZ = OUT.HPosition.w * CameraInfo.z;
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput GodRayVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float Rand(float2 tc)
{
	return frac(tc.x * 3029.34 + tc.y * 4.325) - 0.001;
}

float randomMovieNoise(float2 screenxy,float randomseed)
{
	return sin(screenxy.x * cRayRipple * 8 - GameTime*0.2 ) - sin(7.253 * cRayRipple * screenxy.x * (0.5 + 0.5*sin(GameTime*cRaySpeed)) + randomseed ) ;
}

float4 GodRayPixelShading(in VertexShadingOutput IN, in float4 sc: SV_Position) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	half4 samplerBase;
	
	float rayRipple = randomMovieNoise(IN.TexCoord,IN.BloomZ*42.513) * (1 - IN.TexCoord.y) * 2;

	float2 texSample = IN.TexCoord;
	// 贴图平移
	texSample.x += (GameTime + IN.BloomZ) * cNoiseMSPD;
	
	samplerBase = tRayMap.SampleRGBA(sRaySampler, texSample);

	//光线张角控制
	float shapedAlpha = 1 - saturate((abs (0.5 - IN.TexCoord.x ) - IN.TexCoord.y *cRayAngle)*4);
	//光线长度控制
	shapedAlpha *= saturate(cRayLength - IN.TexCoord.y) * cRayMaxLight;	
	// 边界柔化
	shapedAlpha *= 1 - (IN.TexCoord.x*2 - 1) * (IN.TexCoord.x*2 - 1);
	shapedAlpha *= saturate(IN.TexCoord.y * (1 - IN.TexCoord.y)*25);

	OUT.rgb = lerp(0.1,0.3,samplerBase.a * samplerBase.rgb * rayRipple) * cGodRayColor * shapedAlpha*(1 + cRayLight )  ;
	OUT.a = EncodeTransparentBloomFactor(0 , Rand(sc.xy) - IN.TexCoord.y + cBloomRate * shapedAlpha);

	return OUT;
}

technique Shading
{
	VertexShader = GodRayVertexShading();
	PixelShader = GodRayPixelShading();
}

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);

	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = GodRayVertexShading();
	PixelShader = GodRayPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

