//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Unlit_UI.fx>
///	@path	Shader/
///	@date	2018/10/09
///	@brief	Shader for UI model.
//////////////////////////////////Properties////////////////////////////////////////

#pragma shader material
#define MATERIAL_SHADINGMODEL_UNLIT 1

@_BaseMapEnable
{ 
	Name = BaseMapEnable;
	Description = "Do we have BaseMap?";  
	AffectedTex = tBaseMap;
	Fallback = true;
	Visible = true;
};

@_IsBaseMapUpsideDown
{
	Name = IsBaseMapUpsideDown;
	Description = "Is BaseMap UpsideDown?";
	Default = false;
}

@_IsUIMapUpsideDown
{
	Name = IsUIMapUpsideDown;
	Description = "Is UIMap UpsideDown?";
	Default = true;
}

@_UIMapEnable
{ 
	Name = UIMapEnable;
	Description = "Do we have UIMap?";  
	AffectedTex = tUIMap;
	Fallback = true;
	Visible = false;
};

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 				;
	float4 cParamter2				: Param2				;
	float4 cParamter				: Param					;
	float4 cShadowBias				: ShadowBias			; 
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex	;
};

#define SHADOW_BIAS (cShadowBias.x)
#define SHADOW_SLOPEBIAS (cShadowBias.y)
#define SHADOW_SELFRATE (cShadowBias.z)
#define SHADOW_PIXELSIZE (cShadowBias.w)

cbuffer Shader
{
	float4 cUIColor					 = { 1, 1, 1, 1 };
	float4 cTextureParam0			 = { 0, 0, 1, 1 };
	float4 cTextureParam1			 = { 0, 0, 1, 1 };
};

texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

texture2D tUIMap : layout(t1) = "White";
sampler sUISampler : layout(s1)
{
	Texture = tUIMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;
	float4		WorldPosition	: TEXCOORDN;
	float3 		WorldNormal		: NORMAL;
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;

	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;

	float3		Normal			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
	float3 LocalNormal;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);
}
void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

		}
		/*
		// Hard skinning.
		if (@_DF_SKIN == 2)
		{
			// Fetch.
			float4 Mx, My, Mz;
			GetSkeletonMatrix4x3(i[0], Mx, My, Mz);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalNormal);
		}
		// Dual-quaternion skinning.
		if (@_DF_SKIN == 3)
		{
			// Fetch.
			float4 Q0[2], Q1[2], Q2[2], Q3[2];
			GetSkeletonDualQuaternion(i[0], Q0[0], Q0[1]);
			GetSkeletonDualQuaternion(i[1], Q1[0], Q1[1]);
			GetSkeletonDualQuaternion(i[2], Q2[0], Q2[1]);
			GetSkeletonDualQuaternion(i[3], Q3[0], Q3[1]);
			// Amend orientations.
			w *= float4(1.f, sign(dot(Q0[0], Q1[0])), sign(dot(Q0[0], Q2[0])), sign(dot(Q0[0], Q3[0])));

			// Blend.
			float4 Qn = LinearBlend(w, Q0[0], Q1[0], Q2[0], Q3[0]);
			float4 Qd = LinearBlend(w, Q0[1], Q1[1], Q2[1], Q3[1]);

			// Apply.
			INTM.LocalPosition.xyz = DualQuaternionTransformP(Qn, Qd, INTM.LocalPosition.xyz);
			INTM.LocalNormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalNormal);
		}
		*/
	}
}
void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);
	OUT.TexCoord.xy = IN.TexCoord;
	if (@_IsBaseMapUpsideDown && @_GLSL)
	{
		OUT.TexCoord.y = 1 - OUT.TexCoord.y;
	}
	if (@_LIGHT_MAP)
	{
		OUT.TexCoord.zw = IN.TexCoord2;
		if (@_GLSL && @_IsUIMapUpsideDown)
		{
			OUT.TexCoord.w = 1 - OUT.TexCoord.w;
		}
	}
	OUT.WorldNormal = MulMat3x4((float3x3)World, INTM.LocalNormal);
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput UIVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float4 UIPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	float alpha = 0;

	if (@_BaseMapEnable)
	{
		OUT = tBaseMap.Sample(sBaseSampler, cTextureParam0.xy + IN.TexCoord.xy * cTextureParam0.zw);
		alpha = OUT.a;
	}

	if (@_UIMapEnable && @_LIGHT_MAP)
	{
		float4 ui = tUIMap.Sample(sUISampler, cTextureParam0.xy + IN.TexCoord.zw * cTextureParam0.zw);
		ui *= cUIColor;
		OUT.rgb = ui.rgb + OUT.rgb * (1.0 - ui.a);
		OUT.a = ui.a + OUT.a * (1 - ui.a);
		alpha += ui.a;
	}

	if (@_ES20 && @_CLIPPING)
	{
		clip(alpha - 0.5);
	}

	if (@_TRANSPARENT)
	{
		if (@_BaseMapEnable)
		{
			OUT.rgb = OUT.rgb * OUT.a;
		}
		OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	}
	else
	{
		OUT.a = 0;
	}

	return OUT;
}

technique Shading
{
	VertexShader = UIVertexShading();
	PixelShader = UIPixelShading();
}

/////////////////////////////////////ShadowGen/////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
};

void computeTangent(in half3 gatherNormal, out half3 tan0, out half3 tan1)
{
	tan0 = half3( gatherNormal.z, 0.f, -gatherNormal.x );
	half lg0 = dot( tan0, tan0 );
	if ( lg0 < HALF_EPSILON )
	{
		tan0 = half3(gatherNormal.y,-gatherNormal.x,0.f);
		//tan0-=dot(tan0,gatherNormal)*gatherNormal
		tan0 = normalize( tan0 );
	}
	else
		tan0 *= rsqrt( lg0 );
	tan1 = normalize( cross(gatherNormal,tan0) );
}

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate 
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);

	//world normal
	float3 WorldNormal = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	
#ifdef HAS_WORLD_POSITION_OFFSET
	WorldPosition.xyz += WorldPositionOffset(IN, WorldPosition.xyz, WorldNormal);
#endif

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	return float4(1, 1, 1, 1);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = UIVertexShading();
	PixelShader = UIPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float alpha = 0;
	if (@_BaseMapEnable)
	{
		alpha = tBaseMap.Sample(sBaseSampler, IN.TexCoord.xy).a;
	}
	if (@_UIMapEnable && @_LIGHT_MAP)
	{
		alpha += tUIMap.Sample(sUISampler, IN.TexCoord.zw).a;
	}
	clip(alpha - 0.5);
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = UIVertexShading();
	PixelShader = ClipPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Wireframe
[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].WorldPosition = float4(1, 0, 0, 0);
	OUT.Append(IN[0]);
	IN[1].WorldPosition = float4(0, 1, 0, 0);
	OUT.Append(IN[1]);
	IN[2].WorldPosition = float4(0, 0, 1, 0);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.WorldPosition.xyz);
	float3 distance = IN.WorldPosition.xyz / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(cParamter2.xyz, a);
}

technique Wireframe
{
	VertexShader = UIVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = WireframePixelShading();
};

/////////////////////////////////////PlanarShadow/////////////////////////////////////
struct PlanarShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		TC 				: TEXCOORDN;			//	TexCoord 
};

//viewProjVS is light view now
PlanarShadowVertexShadingOutput PlanarShadowVertexShading(in VertexAssemble IN)
{
	PlanarShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (PlanarShadowVertexShadingOutput)0;

	//InterMediate 
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	//planar equation
	float t = dot(cParamter2.xyz, WorldPosition.xyz - cParamter.xyz) / dot(cParamter2.xyz, SunDirection.xyz);
	WorldPosition.xyz -= max(t*0.95-0.02, 0) * SunDirection.xyz;

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//TC
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 PlanarShadowPixelShading(in PlanarShadowVertexShadingOutput IN) : SV_Target
{
	return float4(lerp(float3(1,1,1), ShadowColor.rgb, SHADOW_RATE), SHADOW_RATE);
}

technique PlanarShadow
{
	VertexShader = PlanarShadowVertexShading();
	PixelShader = PlanarShadowPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}
