#pragma shader material
#define MATERIAL_SHADINGMODEL_LOD_LIT 		1
#define ENABLE_SCREEN_SHADOWMAP				(@_CSM && @_CastSunShadow)
#define DEFERRED_PRE_AO						1 // 是否走ao

#include "Constants.fxh"
#include "Common.fxh"
#include "FastMath.fxh"
#include "BRDF.fxh"
#include "DeferredShadingCommon.fxh"

//////////////////////////////////Properties////////////////////////////////////////

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 100;
	float3 cAdjustCenter			 = {0, 0, 0};
	float2 cScale 					 = {1, 1};
};
//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tBaseMap : layout(t0) = "White";
sampler sBaseSampler : layout(s0, s1)
{
	Texture = tBaseMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
	Precision = Medium;
};

texture2D tNormalMap : layout(t2) = "Transparent";
sampler sNormalMapSampler : layout(s2)
{
	Texture = tNormalMap;
	Filter = Linear; 
	Address = Wrap;
	SRGB = false;
	Precision = Medium;
};

// cbuffer Batch2
// {
// 	// WolrdToObject
// 	float4x3 Local 					: InvWorld;
// };


//////////////////////////////////Vertex Output////////////////////////////////////////
struct LODVertexShadingOutput 
{ 
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float4 		WorldPosition 	: TEXCOORDN;			//  World Position
  
	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.

//	float4 		WorldPivot	 	: TEXCOORDN;			//  WorldPivot

	float 		LinearZ			: TEXCOORDN;

	float 		TexFrameWeight  : TEXCOORDN;			// nointerpolation not supported in glopt

};

//////////////////////////////////VS////////////////////////////////////////
#include "PBR_VS.fxh"
#include "Lighting.fxh" 
half2 Index2UV(int index)
{
	int i = (int)((float)index % 4.0);
	int j = index / 4;

	return half2(i * 0.25, 1.0 - j * 0.5);
}

LODVertexShadingOutput LODVertexShading(in VertexAssemble IN)
{
	LODVertexShadingOutput OUT;
		half3 objectX = MulMat3x4((float3x3)World, half3(1, 0, 0));
		half3 objectY = MulMat3x4((float3x3)World, half3(0, 1, 0));
		half3 objectZ = MulMat3x4((float3x3)World, half3(0, 0, 1));
		half3 objectScale = half3(length(objectX),length(objectY),length(objectZ));	

	// World Position
	// {
		float3 EntityWPosition = float3(0, 0, 0);
		if (@_INSTANCED)
		{
			EntityWPosition = float3(IN.Transform0.w, IN.Transform1.w, IN.Transform2.w);
		}
		else
		{
			EntityWPosition = float3(World[3]);
		}

		float3 V = CameraPos.xyz - EntityWPosition;

		float3 Y = float3(0, 1, 0);
		float3 X = normalize(cross(Y, V));

		OUT.WorldPosition = float4(EntityWPosition + Y * objectScale.y * IN.Position.y - X * objectScale.x * IN.Position.x , 1.0);
		//OUT.WorldPosition = float4(0,0,0,0);
		OUT.ViewDirection.xyz = -V;
	// }

	// TexCoord
	// {
		float3 Vlocal = 1;
		//V = normalize(V);
		float3 objectSpaceCameraPos = float3(0, 0, 0);
			if (@_INSTANCED)
    	{
	        objectSpaceCameraPos =Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, CameraPos.xyz).xyz;
	    }
	    else
	    {
	        objectSpaceCameraPos = MulMat3x4( Local, float4(CameraPos.xyz,1) ).xyz;
	    }
		Vlocal =  normalize( objectSpaceCameraPos ) ;
		V = normalize(float3(Vlocal.x, 0, Vlocal.z));

		float theta = acos(V.z);// acosFast4(V.z);
		float indexf = theta / (0.25 * PI);
		int index = (int)max(0, min(3, floor(indexf))); // [0 - 3)
		float weight = 1.0 - frac(indexf);
		if(V.x < 0)
		{
			index = (7 - index);
			weight = (1.0 - weight);
		}

		OUT.TexCoord.xy = Index2UV(index) + IN.TexCoord.xy * float2(0.25, 0.5);
		OUT.TexCoord.zw = Index2UV((index + 1.0) % 8) + IN.TexCoord.xy * float2(0.25, 0.5);
		OUT.TexFrameWeight = weight;		
	// }

	// Fog T
	if (@_FOG)
		OUT.ViewDirection.w = FogTransmittance(OUT.ViewDirection.xyz);

	// HPosition
	OUT.HPosition = MulMat4x4(ViewProj, float4(OUT.WorldPosition.xyz, 1.0));
	OUT.LinearZ = OUT.HPosition.w * CameraInfo.z;
	return OUT;
}

//////////////////////////////////Function////////////////////////////////////////
void InitScreenGray(in LODVertexShadingOutput IN, in out half3 BaseColor)
{
	if(ScreenMGrayModeSpread)
	{
		half grayDist = (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - min(length(IN.WorldPosition.xyz - PlayerPos.xyz), ScreenMGrayMaxDistance);
		grayDist = step(0, grayDist);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);		
	}
	else if (ScreenMGrayModeShrink)
	{
		half grayDist = length(IN.WorldPosition.xyz - PlayerPos.xyz) + (GameTime - ScreenMGrayStartTime) * ScreenMGraySpeed - ScreenMGrayMaxDistance;
		grayDist = step(grayDist, 0);
		half GrayColor = GetLuminance(BaseColor) * 1.5;
		BaseColor = lerp(BaseColor, half3(GrayColor, GrayColor, GrayColor), saturate(grayDist) * 0.5);
	}
}
void PS_GetParameters(in LODVertexShadingOutput IN, out MaterialShadingParameters MaterialParameters)
{
	float weightA = IN.TexFrameWeight;
	float weightB = 1.0 - weightA;

	//init MaterialParameters (always need)
	InitShadingParameters(MaterialParameters);
	half4 BaseColor1 = tBaseMap.SampleLevel(sBaseSampler, IN.TexCoord.xy, 0);
	half4 BaseColor2 = tBaseMap.SampleLevel(sBaseSampler, IN.TexCoord.zw, 0);
	half4 BaseColor = BaseColor1 * weightA + BaseColor2 * weightB;
	MaterialParameters.BaseColor = DecodeGamma(BaseColor.rgb);
	MaterialParameters.Alpha = BaseColor.a;
	MaterialParameters.Roughness = 0.8;
	
	//screen gray
	InitScreenGray(IN, MaterialParameters.BaseColor);
	
	//Normal
    {
		half4 Normal1 = tNormalMap.SampleLevel(sNormalMapSampler, IN.TexCoord.xy, 0);
		half4 Normal2 = tNormalMap.SampleLevel(sNormalMapSampler, IN.TexCoord.zw, 0);
		half4 Normal = Normal1 * weightA + Normal2 * weightB;
		half3 LocalNormal = normalize((2 * Normal.rgb) - 1);
		//MaterialParameters.WorldNormal = LocalNormal;		
		half3 Normalfix = half3(-1,1,-1);
		if (@_INSTANCED)
			Normalfix = half3(1,1,1);
		MaterialParameters.WorldNormal = normalize(MulMat3x4((float3x3)World, Normalfix*LocalNormal));
		MaterialParameters.AO = sqrt(sqrt(Normal.a));
    }
	
	//Apply Weather
	MaterialParameters.BaseColor *= 1 - 0.3 * RainRate;
	MaterialParameters.Roughness *= 1 - 0.3 * RainRate;
	//MaterialParameters.Thickness = 100000;

	//Apply Burn
	half BurnRadius = GlobalBurnParam.w;
	half3 BurnPosition = GlobalBurnParam.xyz;
	if(BurnRadius > 0.01)
	{
		half BurnWidth = 80;
		half BurnArea = saturate((length(IN.WorldPosition-BurnPosition) - BurnRadius) / BurnWidth);
		MaterialParameters.BaseColor *= lerp(0.04 , 1 ,BurnArea);
	}
}

//////////////////////////////////PS////////////////////////////////////////
float4 LODPixelShading(in LODVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);
	MaterialShadingParameters MaterialParameters;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	OUT.a = MaterialParameters.Alpha;
	
	{
		half NoL = saturate(dot(MaterialParameters.WorldNormal, SunDirection.xyz));
		NoL = 0.9 * NoL + 0.1;

	    float3 DiffuseLight = (NoL * SunColor.xyz / PI + AmbientColor.xyz) * MaterialParameters.BaseColor;

	    OUT.rgb = DiffuseLight;

		//Correct HDR & EncodeHDR color
		EncodeHDR(OUT.rgb);
		return OUT;		
	}
}

technique Shading
{
	VertexShader = LODVertexShading();
	PixelShader = LODPixelShading();
}

//////////////////////////////////GBuffer////////////////////////////////////////
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	
	float4 GBufferB 	: SV_Target1;	
	float4 GBufferC 	: SV_Target2;	
	float4 GBufferD 	: SV_Target3;	
	float4 GBufferE 	: SV_Target4;	
	float4 GBufferF  	: SV_Target5;	
	float4 GBufferG 	: SV_Target6;	
};
GBufferOUTPUT GBufferGenPixelShading(in LODVertexShadingOutput IN)
{
	GBufferOUTPUT OUT;
	MaterialShadingParameters MaterialParameters;
	GBufferData GBuffer;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	half Alpha = MaterialParameters.Alpha;
	//GILighting
	DynamicGILight(MaterialParameters.WorldNormal, MaterialParameters.GILighting, MaterialParameters.InShadow);
	MaterialParameters.GILighting *= MaterialParameters.AO;
	//set GBuffer
	SetGBufferForShadingModel(GBuffer, MaterialParameters);
	EncodeGBufferData(GBuffer, OUT.GBufferA, OUT.GBufferB, OUT.GBufferC, OUT.GBufferD, OUT.GBufferE, OUT.GBufferF);
	OUT.GBufferG = EncodeDepth(IN.LinearZ);
	OUT.GBufferA.w = EncodeDepth(IN.LinearZ);
	return OUT;
}

technique GBufferGen
{
	VertexShader = LODVertexShading();	
	PixelShader = GBufferGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
[barycentric("sample")]
float4 LODClipShading(in LODVertexShadingOutput IN) : SV_Target
{
	// Clipping
	float clipValue = 1.0;

	MaterialShadingParameters MaterialParameters;
	//material parameters
	PS_GetParameters(IN, MaterialParameters);
	
	clip(MaterialParameters.Alpha - 0.33);

	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = LODVertexShading();
	PixelShader = LODClipShading();
}

/////////////////////////////////////ShadowGen/////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC 				: TEXCOORDN;			//	TexCoord
};

ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
//	[@_HLSL]
//	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.SceneDepth = OUT.HPosition.zw;
	//TC
	OUT.TC = IN.TexCoord.xy;
	return OUT;
}

[target("R32F")]
float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float Alpha = tBaseMap.SampleRGBA( sBaseSampler, IN.TC ).a;
	float clipValue = 1.0;
	
	if (@_CLIPPING)
	{
		clipValue = Alpha - ALPHA_REF;
	}
	if (@_DISSOLVING)
	{
		clipValue = min(clipValue, cParamter2.x - sign(cParamter2.x) * GetRandom(IN.HPosition.xy));
	}
	if (@_CLIPPING || @_DISSOLVING)
		clip(clipValue);

	float depth = IN.SceneDepth.x / IN.SceneDepth.y;	
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}
